#
# Copyright (c) 1991 University College London
# All rights reserved.
#
# Redistribution and use in source and binary forms are permitted
# provided that the above copyright notice and this paragraph are
# duplicated in all such forms and that any documentation,
# advertising materials, and other materials related to such
# distribution and use acknowledge that the software was developed
# by the Department of Computer Science, University College London.
# The name of the University may not be used to
# endorse or promote products derived from this software without
# specific prior written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#

# Configuration information for all Makefiles
#
# By George Pavlou, June 1991

# The first part of this file serves to configure the OSIMIS tree
# to cope with optional components. These are considered optional
# either because special compiler versions / libraries are needed to 
# build them or because they may not be needed in many
# environments - see the relevant comments for each one later.
# These optional components are:
#
#	SIMTIME		- the simulated time support service
#	LT		- the X.500-based location transparency feature
#	MONMET		- the generic monitor metric objects
#	MONSUP		- the generic monitoring support objects
#	IQA		- the Internet Q-Adaptor (CMIS/P->SNMPv1 proxy)
#	XMP		- the XOM/XMP CMIS/P stack
#	TCLRMIB		- the TCL-based generic manager support service
#	TPMIB		- the Transport Protocol MIB for ISODE
#	ODPOBJ		- the example ODP objects
#
# If you do not want to build any of these, you will have to comment out
# all the relevant lines in the following section. Typically, out of all these
# the most likely to need is the MONMET and TCLRMIB.


# SIMTIME - useful for running management systems over simulated networks
#
#SIMTIME		= simtime

# LT - X.500-based location transparency service, you need to know how to
# set-up the X.500 directory to use it (see relevant document in doc/),
# better avoided unless you know what you are doing
#
#CCONFIG		+= -DUSE_X500
#LT		= lt
#LT_LIB		= $(TOP)/lt/liblt.a
#X500_LIB	= $(ISODE)/libdsap-osisec.a
#X500_LIB	= $(ISODE)/libdsap.a

# MONMET - generic monitor metric classes as in X.739
#
CCONFIG		+= -DMONMET
MONMET_MIB	= monmet_mib
MONMET_MIB_LIB	= $(TOP)/agent/monmet_mib/libmonmetmib.a
MONMET_SNTX_LIB	= $(TOP)/agent/monmet_mib/libmonmetsntx.a

# MONSUP - generic monitoring support classes combining X.738/X.739,
# see relevant document in doc/
#
#CCONFIG         += -DMONSUP
#MONSUP_MIB	= monsup_mib
#MONSUP_MIB_LIB  = $(TOP)/agent/monsup_mib/libmonsupmib.a

# TCLRMIB - TCL-based generic manager infrastructure for interpreted managers
#
TCLRMIB	= tcl-cmis
TKCOORD	= TkCoord.o

# IQA - the Internet Q-Adaptor (generic CMIS/P->SNMPv1 proxy)
#
CCONFIG		+= -DIQA
IQA		= iqa
IQA_SNTX_LIB	= $(TOP)/proxy/iqa/proxy/libiqasyntaxes.a

# XMP - support for Sun XOM/XMP CMIS/P stack version 8.1
#
#XMP		= msap-xmp

# TPMIB - the Transport Protocol MIB for the ISODE TP0 implementation;
# you will also need to build the misode/ stuff for this to be of any use
#
CCONFIG         += -DTPMIB
TP_MIB	= isode_mib
TP_MIB_LIB	= $(TOP)/agent/isode_mib/libisodemib.a
TP_SNTX_LIB	= $(TOP)/agent/isode_mib/libisodesntx.a

# ODPOBJ - the example ODP statistical object, an example of using
# OSI/TMN management as a general purpose distributed management framework 
#
#CCONFIG         += -DODPOBJ
#ODP_OBJ	= odp
#ODP_SNTX_LIB	= $(TOP)/examples/odp/libodpsntx.a




# The following part configures the local environment.
# Please read carefully the ./README file before you proceed.
#

HOME	=	/cs/research/icm/proj2/athena/gpavlou/solaris/tmp/osimis-4.0.7

# LSOCKET should be as in the ISODE CONFIG.make
# e.g. /usr/sunlink/osi/lib/libosi.a for SunLink OSI, -lbsd for AIX etc.
# You may also put any other needed system libraries in it
LSOCKET	= -lsocket -lnsl -lelf

# GNU C version 2.7.2
CC	=	/opt/ucl/bin/gcc -traditional
# Sun C version 4.0
CC	=	/opt/SUNWspro/bin/cc

# GNU C++ version 2.7.2
CCPLUS	=	/opt/ucl/bin/g++
# Sun C++ version 4.1
CCPLUS	=	/opt/SUNWspro/bin/CC


PEPSY	=	/cs/research/icm/proj2/athena/gpavlou/solaris/isode-8.0/pepsy

ISODE	=	/cs/research/icm/proj2/athena/gpavlou/solaris/isode-8.0

X11	=	-L/usr/openwin/lib	# for Sun OpenWindows X11
TCLTK	=	/cs/research/race/sol/prepare/common/tcl-support/sun5/lib

# none of the next four should be /usr/include, see README file
X11INC	=	/usr/openwin/include	# for Sun OpenWindows X11
# the next is only needed for building the optional kernel/TkCoord.o
# (see kernel/Makefile)
TCLINC	=	/cs/research/race/sol/prepare/common/tcl-support/src/tcl7.4
TKINC	=	/cs/research/race/sol/prepare/common/tcl-support/src/tk4.0

#LPP	=	-lpp
CMPL	=
#CMPL	=	.$(CCPLUS)
ARCH	=
#ARCH	=	.sun3
#ARCH	=	.sun4


# do not change the following for building OSIMIS
#
LIB	=	$(TOP)/lib$(CMPL)$(ARCH)
BIN	=	$(TOP)/bin$(ARCH)
OSIMIS-INC =	$(TOP)/include
H	=	$(TOP)/h
ETC	=	$(TOP)/etc
MAN	=	$(TOP)/man

# configure the following in order to install outside the OSIMIS tree
# (they are currently configured as at UCL) - you will obviously need
# to uncomment them to overwrite the options above
#
#LIB	=	/usr/local/lib
#BIN	=	/usr/local/bin
#OSIMIS-INC =	/usr/local/include/osimis
#H	=	/usr/local/include/osimis
#ETC	=	/usr/local/etc/osimis
#MAN	=	/usr/local/man


CCONFIG +=	-DDEBUG -DLOGGING -DPERSISTENCY -DSOLARIS -DSOLARIS24 -DPOSIX_SIGNALS # -DTIMING # -DLPP
INC	=	-I$(TCLINC) -I$(TKINC) -I$(X11INC)
OPTIONS =	-I. -I$(OSIMIS-INC) \
		-I$(OSIMIS-INC)/isode/pepsy \
		-I$(OSIMIS-INC)/isode/quipu \
		$(INC) -I$(H) $(CCONFIG)
CCFLAGS	=	$(OPTIONS) -g # -O
LDFLAGS	=	-s
PFLAGS	=	-A -h -m	 # pepsy flags, do not change
LFLAGS	=	-bhuz $(OPTIONS) # lint flags, do not change


YACC		= $(TOP)/compilers/mocompiler/byacc/byacc 
YACCFLAGS	= -v
LEX		= $(TOP)/compilers/mocompiler/flex/flex 
LEXFLAGS	= 


# the following paths may need to start with /usr/bin instead of /bin,
# check and change accordingly
AR              = ar
ARFLAGS         = rv
RANLIB          = ranlib


ASN-CMPL	= $(TOP)/compilers/asncompiler/asn-cmpl
GDMO		= $(TOP)/compilers/mocompiler/gdmo/gdmo
GDMO-CMPL	= $(TOP)/compilers/mocompiler/gdmo/gdmo-cmpl
RUN-GDMO	= $(GDMO-CMPL)	# for backwards compatibility
GDMOASN-CMPL	= $(TOP)/compilers/mocompiler/gdmo/gdmoasn-cmpl


.SUFFIXES:	.py .c .cc .o .asn -types.h _tables.c

.asn_tables.c .asn-types.h:;	$(PEPSY) $(PFLAGS) $<

.c.o:;		$(CC) $(CCFLAGS) -c $*.c

.cc.o:;		$(CCPLUS) $(CCFLAGS) -c $*.cc
