//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// CmisAgent.h - CMIS Agent object declaration

// By George Pavlou, February 1991


#ifndef _CMIS_AGENT_
#define _CMIS_AGENT_


#ifndef _BOOLEAN_
#include "Boolean.h"
#endif

#ifndef _LIST_ITERATOR_
#include "ListIterator.h"
#endif

#ifndef _ISODE_AGENT_
#include "IsodeAgent.h"
#endif

#ifndef _SMI_ASN_
#include "SmiAsn.h"
#endif

#ifndef _AVA_
#include "AVA.h"
#endif

#if	defined(SECURITY) && !defined(_ASSOCIATION_MANAGER_)
#include "AssocMgr.h"
#endif


class CMISAgent : public ISODEAgent
{
    friend class AssociationMgr;

#define AGENT_FUNUNITS	0x17		/* CMIS agent capabilities:
					   scope, filter, cancelGet */
#define DFLT_MAXASSOCS	10		/* default maximum cmis associations */
#define CANCELGETBLOCK	5		/* to check for M-Cancel-Get */
#define	CONTEXT		"management"	/* the application context */

protected:
    static CMISAgent* _instance;
    OID		 _systemIdOID;
    OID		 _processingFailureOID;
    char*	 _operationClass;
    int		 _maxAssocs;		// max cmis associations
    int		 _curAssocs;		// current cmis associations
    int		 _thisAssoc;		// association being served
    char*	 _aei;			// application entity title
    List	 _requests;
    ListIterator _iterator;
    int		 _moClassIndex;
    Bool	 _createOperation;
    int		 _invokeId;
    timeval	 _prevTime;
#ifdef USE_X500
    char*	_agentService;
    Bool	_singleInstanceAgent;	// for true location transparency
    Bool	_ltFailed;		// in case the DSA is down 
#endif

    void	doTime (char* = NULLCP);
    void	init ();
    CMISErrors	checkParms (MID, MN, MN,
			CMISScope*, CMISFilter*, CMISSync);
    int		serviceCmisRequest (int, MSAPIndication*, PE);
    int		checkForCancelGet (int, int);

    int		Get (int, int, MID, MN, CMISScope*, CMISFilter*,
			External*, CMISSync, int, MID);
    int		CancelGet (int, int, int, int);

    int		Set (int, Bool, int, MID, MN, CMISScope*, CMISFilter*,
			External*, CMISSync, int, CMISSetAttr*);
    int		Action (int, Bool, int, MID, MN, CMISScope*, CMISFilter*,
			External*, CMISSync, CMISParam*);
    int		Create (int, int, MID, MN, int, MN,
			External*, int, CMISParam*);
    int		Delete (int, int, MID, MN, CMISScope*, CMISFilter*,
			External*, CMISSync);
    int		MError (int, int, int, int, CMISErrors, MID, MN, CMISScope*,
			CMISFilter*, CMISSync, MIDListVal*, PE, AVA*&);

public:
    int		initialise (int, char**, char*, TSAPdisconnect*, Bool = False);
    int		associationRequest (int, char**);
    int		readCommEndpoint (int);	// read association's data
    int		shutdown (int);		// shutdown association
    int		sendEventReport (MID, MN, MID, PE, char*, DestAddressVal*);
					// event notification
    void	setMaxAssocs (int);
    int		getThisAssoc ();
    char*	getAEI ();
    DN		mn2localdn (MN);

    static CMISAgent* getInstance ();

    CMISAgent (int, char**, char*, Bool = False);
    CMISAgent ();
    ~CMISAgent ();
};


inline void	CMISAgent::setMaxAssocs (int assocs)	{ _maxAssocs = assocs; }
inline int	CMISAgent::getThisAssoc ()	{ return _thisAssoc; }
inline char*	CMISAgent::getAEI ()		{ return _aei; }
inline CMISAgent* CMISAgent::getInstance ()	{ return _instance; }
inline		CMISAgent::CMISAgent ()		{ init(); }


#endif
