//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Create.h - create MO class methods

// By George Pavlou, May 1991


#ifndef _CREATE_
#define _CREATE_


#ifndef _GENERIC_MO_
#include "GenericMO.h"
#endif

#ifndef _GEN_ARRAY_
#include "GenArray.h"
#endif


#define MIBINIT_FILE	"mib.init"

typedef MOClassInfo* (*CLASSINFOFP) ();
typedef int	     (*INTFP) ();


typedef struct
{
    char*	className;	// class name as registered in oidtable.gen
    CLASSINFOFP	getClassInfo;	// meta-class information object access
    INTFP	classInit;	// class initialisation method

} MOClass;

extern  MOClass moclasses[];
extern  int     NCLASSES;


typedef struct NotificationMapping_s
{
    const char *notification;	// notification name
    const char *record;		// name of logRecord type

} NotificationMapping;

extern NotificationMapping notificationMappingTable[];
extern int                 NNOTIFICATIONS;


class MOClassArray : public Array
{
#define STRCMP	1
#define OIDCMP	2

    int		_cmpMode;

protected:
    void	deleteElem (void*);
    int		compare (void*, void*);

public:
    MOClass*	findClass (char*);
    MOClass*	findClass (OID);
    void	addClass (char*, CLASSINFOFP);
    int		removeClass (char*);
};


inline void  MOClassArray::deleteElem (void* e)
{ delete ((MOClass*) e) -> className; delete e; }

inline int  MOClassArray::compare (void* e1, void* e2)
{ return _cmpMode == OIDCMP ?
    (oid_equal((OID) ((MOClass*) e1) -> className,
	((MOClass*) e2) -> getClassInfo() -> getClass()) ? 0 : -1) :
    strcasecmp(((MOClass*) e1) -> className, ((MOClass*) e2) -> className); }

inline MOClass*  MOClassArray::findClass (char* cl)
{ int i; MOClass moclass; moclass.className = cl;
  _cmpMode = STRCMP;
  return (i = find(&moclass)) != -1 ? (MOClass*) get(i) : (MOClass*) NULL; }

inline MOClass*  MOClassArray::findClass (OID cl)
{ int i; MOClass moclass; moclass.className = (char*) cl;	// small hack
  _cmpMode = OIDCMP;
  return (i = find(&moclass)) != -1 ? (MOClass*) get(i) : (MOClass*) NULL; }

inline void  MOClassArray::addClass (char* cl, CLASSINFOFP info)
{ MOClass* moclass = new MOClass; moclass->className = strdup(cl);
  moclass->getClassInfo = info; moclass->classInit = NULLIFP; add(moclass); }

inline int  MOClassArray::removeClass (char* cl)
{ MOClass moclass; moclass.className = cl;
  _cmpMode = STRCMP; return remove(&moclass); }


#endif
