//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Log.h - Log MO

// Saleem N. Bhatti
// April 1992

#if !defined(_Log_h_)
#define _Log_h_

#include "GenericKS.h"
#if defined(PERSISTENCY)
#include "dbm-lock.h"
#endif
#include <signal.h>

#include "Top.h"

class Integer;
class MFilter;
class AdministrativeState;
class OperationalState;
class AvailabilityStatus;
class LogFullAction;


/*
** The default action is to rename the gdbm files when a log is deleted.
** Comment out the next line to enable deleteing of log files.
*/
#define LOG_RenameOldLogFiles

#define LOG_Directory     "logs"
#define LOG_DirectoryMode 0775
#define LOG_FileMode      0664

#define LOG_TidyInterval  10 /* secodns */

class LogRecord;

class Log : public Top, public KS {

    char    *_logName;
#if defined(PERSISTENCY)
    DbmFile *_db;
#endif

    static
    MOClassInfo _info;

    static
    int _logId; // monotonically increasing

    int _recordId; // monotonically increasing
    int _loRecordId;
    UTCtime _time; // loggingTime

    int _mask;

    Bool _scheduledWakeUp;

protected:

    static const int
    _level;                   // 1
    enum {
    I_logId,                  // 0 Integer
    I_discriminatorConstruct, // 1 Filter [GET-REPLACE]
    I_administrativeState,    // 2 AdministrativeState [GET-REPLACE]
    I_operationalState,       // 3 OpertaionalState
    I_availabilityStatus,     // 4 AvailabilityStatus
    I_logFullAction,          // 5 LogFullAction [GET-REPLACE]
    I_maxLogSize,             // 6 Integer [GET-REPLACE]
    I_currentLogSize,         // 7 Integer
    I_numberOfRecords,        // 8 Integer
    I_logNAttrs};

    Attr *_attrs[I_logNAttrs];

    int set(CMISModifyOp, int, int, void*, AVA*&, Bool, int);
    CMISErrors cmisDelete();

    int wakeUp(char *);

    static
    int getNewLogId(); // For RDN/logId

    int getNewRecordId(); // for logRecordId
    UTCtime *getTime();   // for loggingTime

    Log(RDN, MO *, Bool, MO *, int, CMISParam *,
        CMISErrors *, CMISErrorInfo *, OID *);

public:
    static
    int restartLogs(); // When restarting agent

    static
    int initialiseClass();
    static
    MOClassInfo *getClassInfo();
    static
    MO *create(RDN, MO *);

    int get(int, int, AVA*&, Bool = False, int = -1);

    static
    MO *cmisCreate(RDN, MO *, MO *, int, CMISParam *, CMISErrors *,
                   CMISErrorInfo *, OID *);
    int refreshSubordinates(AVA*&, int);
    int refreshSubordinate(RDN, AVA*&, int);

    char *getLogName();  // name of log

    int store(LogRecord *);

    /* Disc LogRecord access */
#define LOG_FetchFirst -1
#define LOG_FetchNext  -2
    int fetch(int, char *, int *);
    int remove(int);

    Integer *logId();
    MFilter *discriminatorConstruct();
    AdministrativeState *administrativeState();
    OperationalState *operationalState();

    ~Log();
};


inline int
Log::getNewLogId()
{ return(++_logId); }

inline MOClassInfo *
Log::getClassInfo()
{ return(&_info); }

inline MO *
Log::create(RDN rdn, MO *superior)
{ return new Log(rdn, superior, True, (MO *) 0, 0, (CMISParam *) 0,
                 (CMISErrors *) 0, (CMISErrorInfo *) 0, (OID *) 0); }

inline char *
Log::getLogName()
{ return(_logName); }

inline UTCtime *
Log::getTime()
{ time_t now = time(NULL); tm2ut(localtime(&now), &_time); return(&_time); }

inline int
Log::getNewRecordId()
{ return(++_recordId); }

inline Integer *
Log::logId()
{ return (Integer *) _attrs[I_logId]; }

inline MFilter *
Log::discriminatorConstruct()
{ return (MFilter *) _attrs[I_discriminatorConstruct]; }

inline AdministrativeState *
Log::administrativeState()
{ return (AdministrativeState *) _attrs[I_administrativeState]; }

inline OperationalState *
Log::operationalState()
{ return (OperationalState *) _attrs[I_operationalState]; }


#endif /* _Log_h_ */
