//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Top.cc - top managed object class realisation

// By George Pavlou, February 1991


#include <stdlib.h>
#include "msap.h"
#include "mtime.h"
#include "Error.h"
#include "Top.h"
#include "eventForwardingDiscriminator.h"
#ifdef	LOGGING
#include "Log.h"
#endif
#include "EventLog.h"
#include "RecordInfo.h"
#include "ListIterator.h"

#include "ObjectClass.h"
#include "ObjectClassList.h"
#include "ObjId.h"
#include "ObjIdList.h"

#include "SimpleNameType.h"
#include "Filter.h"
#include "AdministrativeState.h"
#include "OperationalState.h"
#include "Destination.h"


MOClassInfo Top::_top;
MO*	    Top::_testEventRecord;

List*	    Top::_attrValueList;
int	    Top::_classLevel;
int	    Top::_attrId;
void*	    Top::_oldAttrValue;
int	    Top::_source;
char*	    Top::_addText;
type_SMI_AttrValChangeInfo*	Top::_attrValChangeInfo;
type_SMI_ObjectInfo*		Top::_objectInfo;
CMISParamListVal*		Top::_objectInfoListCurPos;


extern Bool		   globalNames;
extern NotificationMapper *notificationMapper;


Top::Top ()
{
    if (this == getRoot())
	_attrValChangeInfo = AttrValChangeInfo_alloc();
    _objectMgmtMask = 0;

    // initialise attributes
    _attrs[OBJECTCLASS]	= new ObjectClass;
    _attrs[NAMEBINDING]	= NULLATTR;
    _attrs[PACKAGES]	= NULLATTR;
    _attrs[ALLOMORPHS]	= NULLATTR;

    // register class
    registerClass(1, _attrs, &_top, False);

    // dummy checkNameBinding
    checkNameBinding();
}


Top::Top (RDN rdn, MO* superior)
: MO(rdn, superior)
{
    _objectMgmtMask = 0;

    // initialise attributes
    _attrs[OBJECTCLASS]	= new ObjectClass;
    _attrs[NAMEBINDING]	= new ObjId;
    _attrs[PACKAGES]	= new ObjIdList;
    _attrs[ALLOMORPHS]	= new ObjectClassList;

    // register class
    registerClass(TOPNATTRS, _attrs, &_top, False);
}


Top::Top (RDN rdn, MO* superior, int nattrs, CMISParam* attrs,
				CMISErrors* err, CMISErrorInfo* errInfo)
: MO(rdn, superior)
{
    _objectMgmtMask = 0;

    // initialise attributes
    _attrs[OBJECTCLASS]	= new ObjectClass;
    _attrs[NAMEBINDING]	= new ObjId;
    _attrs[PACKAGES]	= new ObjIdList;
    _attrs[ALLOMORPHS]	= new ObjectClassList;

    // register class
    registerClass(TOPNATTRS, _attrs, &_top, False);

    // check for a possible packages attribute initial value
    for(register int i = 0; i < nattrs; i++) {
	switch (_top.getAttrId(attrs[i].mp_id.mid_global)) {
	case PACKAGES:
	    *err = _attrs[PACKAGES] -> set(attrs[i].mp_val);
	    break;

	case OBJECTCLASS:
	case NAMEBINDING:
	case ALLOMORPHS:
	    *err = m_invalidAttributeValue;
	    break;

	default:
	    continue;	// derived class attribute
	}

	switch (*err) {
	case m_invalidAttributeValue:
	    errInfo -> ei_invalidAttributeValue.mp_id = attrs[i].mp_id;
								// struct copy
	    errInfo -> ei_invalidAttributeValue.mp_val = attrs[i].mp_val;
	    return;

	default:	// m_noError
	    mid_free(&attrs[i].mp_id);	// for derived classes
	    break;
	}
    }
}


Top::~Top ()
{
    if (this == getRoot()) {
	register type_SMI_AttrValChangeDefinition* cur =
		_attrValChangeInfo->attrValChangeDefinition;
	for ( ; cur; cur = cur->next)
	    if (cur->attrValChangeDefinition_member->oldAttrValue) {
		pe_free(cur->attrValChangeDefinition_member->oldAttrValue);
		cur->attrValChangeDefinition_member->oldAttrValue = NULLPE;
	    }
	_attrValChangeInfo->additionalText = NULL;
	AttrValChangeInfo_free_aux(_attrValChangeInfo);

        // free the ObjectInfo structure (the contents are already freed)
	if (_objectInfo) {
	    register CMISParamListVal** cur =
			(CMISParamListVal**) &_objectInfo->attributeList;
	    for (; *cur; cur = &(*cur)->mpl_next);
	    *cur = (CMISParamListVal*) _objectInfoListCurPos;

	    _objectInfo->additionalText = NULL;
	    ObjectInfo_free_aux(_objectInfo);
	}

	if (_testEventRecord)
	    delete _testEventRecord;
    }
}


int  Top::initialiseClass ()
{
    if (_top.setClass("top", 0, TOPNATTRS, 0, 0, 0, 1) == NOTOK ||

	_top.setAttr("objectClass", OBJECTCLASS) == NOTOK ||

	_top.setAttr("nameBinding", NAMEBINDING) == NOTOK ||

	_top.setAttr("packages", PACKAGES) == NOTOK ||

	_top.setAttr("allomorphs", ALLOMORPHS) == NOTOK ||

	_top.setPackage("allomorphicPackage", ALLOMPACKAGE) == NOTOK)

	return NOTOK;

    return OK;
}


int  Top::addAllomorphicClass (OID clss)
{
    for (register int i = 1; i < _nclasses-1; i++)
	if (oid_equal(clss, _classes[i]->getClass()))
	    break;

    if (i == _nclasses-1) {
#ifdef DEBUG
	advise(NULLCP, "Top::addAllomorphicClass: wrong allomorphic class \
`%s' for class `%s'!",
			sprintoid(clss),
			_classes[_nclasses-1] -> getClassName());
#endif
	return NOTOK;
    }

    if (!allomorphs())
	_attrs[ALLOMORPHS] = new ObjectClassList;
    allomorphs() -> addval(clss);

    initialiseSpecialAttrs();

    return OK;
}


int  Top::initialiseSpecialAttrs ()
{
    int nattrs = TOPNATTRS;

    if (oid_equal(objectClass() -> getOID(), DEFAULT_OID))
	objectClass() -> setOID(oid_cpy(getClass()));

    if (this == getRoot()) {
	MO::_nattrs[TOP_LEVEL] = 1;
	return OK;
    }

    MO* superior = getSuperior();
    if (oid_equal((OID) nameBinding() -> get(), DEFAULT_OID))
	if (superior && superior->getSuperior()) // not the top and dummy MOs
	    _attrs[NAMEBINDING] ->
			set(oid_cpy(getNameBinding()->getNameBinding()));

    if (! allomorphs() -> get()) {
        delete _attrs[ALLOMORPHS];
        _attrs[ALLOMORPHS] = NULLATTR;
	--nattrs;
    }

    if (! allomorphs()) {
	if (! packages() -> get()) {
	    delete _attrs[PACKAGES];
	    _attrs[PACKAGES] = NULLATTR;
	    --nattrs;
	}
    }
    else {
	if (! packages())
	    _attrs[PACKAGES] = new ObjIdList;
	packages() -> addval(oid_cpy(_top.getPackageOid(ALLOMPACKAGE)));
    }

    MO::_nattrs[TOP_LEVEL] = nattrs;
    return OK;
}


type_SMI_ObjectInfo*  Top::getObjectInfo ()
{
    CMISParam	attrs[MO_MAXATTRS];
    int		nattrs = 0;
    register int i;
    register CMISParamListVal** cur;

    Get(NULLOID, &nattrs, attrs);	// fills in attrs

    if (!_objectInfo) {		// the very first time
	_objectInfo = new type_SMI_ObjectInfo;
	_objectInfo->sourceIndicator = new type_SMI_SourceIndicator;
	_objectInfo->sourceIndicator->parm = int_SMI_SourceIndicator_unknown;
	_objectInfo->additionalText = NULL;

	cur = (CMISParamListVal**) &_objectInfo->attributeList;
        for (i = 0; i < MO_MAXATTRS; i++) {
	    *cur = new CMISParamListVal;
	    (*cur)->mpl_id.mid_type = MID_GLOBAL;
	    cur = &(*cur)->mpl_next;
        }
	_objectInfoListCurPos = (CMISParamListVal*) _objectInfo->attributeList;
	_objectInfo->attributeList = NULL;
        *cur = NULL;	// end of full list (MO_MAXATTRS)
    }

    // restore the list first (remove "end" of the last attribute list)
    cur = (CMISParamListVal**) &_objectInfo->attributeList;
    for (; *cur; cur = &(*cur)->mpl_next);
    *cur = _objectInfoListCurPos;

    // fill it with info (OIDs and PEs are NOT copied)
    cur = (CMISParamListVal**) &_objectInfo->attributeList;
    for (i = 0; i < nattrs && i < MO_MAXATTRS; i++) {
	(*cur)->mpl_id.mid_global = attrs[i].mp_id.mid_global;
	(*cur)->mpl_val = attrs[i].mp_val;
	cur = &(*cur)->mpl_next;
    }

    _objectInfoListCurPos = *cur;	// used to restore list
    *cur = NULL;			// end of current list (nattrs)

    return _objectInfo;
}


OID  Top::getObjectManagementEvent (int eventId)
{
    switch (eventId) {

    case MO_objectCreation:
	return findEvent("objectCreation");

    case MO_objectDeletion:
	return findEvent("objectDeletion");

    case MO_attributeValueChange:
	return findEvent("attributeValueChange");

    case MO_stateChange:
	return findEvent("stateChange");

    default:
	return NULLOID;
    }
}


int  Top::buildReport (int eventId, int classLevel, void*&info, Bool& freeFlag)
{
    if (classLevel != TOP_LEVEL)
	return NOTOK;

    freeFlag = False;

    switch (eventId) {

    case MO_objectCreation:
    case MO_objectDeletion:
    {
	type_SMI_ObjectInfo* _objectInfo = getObjectInfo(); // MO class method
	_objectInfo->sourceIndicator->parm = _source;
	_attrValChangeInfo->additionalText = _addText;
        info = (void *) _objectInfo;
	// the actual information will be free'd by the MO class destructor
    }
	break;

    case MO_attributeValueChange:
    case MO_stateChange:
    {
	// free oldAttrValue for the first attribute of previous invocation
	register type_SMI_AttrValChangeDefinition
	    *cur = _attrValChangeInfo->attrValChangeDefinition, **curr, *prev;
	if (cur->attrValChangeDefinition_member->oldAttrValue) {
	    pe_free(cur->attrValChangeDefinition_member->oldAttrValue);
	    cur->attrValChangeDefinition_member->oldAttrValue = NULLPE;
	}

	// memory is allocated once when constructing the root
	// apart from multiple attributes (only one remains always allocated)

	// free first any additional attributes from previous call
	for (cur = cur->next; cur; ) {
	    if (cur->attrValChangeDefinition_member->oldAttrValue)
		pe_free(cur->attrValChangeDefinition_member->oldAttrValue);
	    prev = cur;
	    cur = cur->next;
	    free((char*) prev->attrValChangeDefinition_member->attributeId);
	    free((char*) prev->attrValChangeDefinition_member);
	    free((char*) prev);
	}
	_attrValChangeInfo->attrValChangeDefinition->next = NULL;

	// fill-in the first value
	AttrValue *attrValue, av(_attrId, _classLevel, _oldAttrValue);
	curr = &_attrValChangeInfo->attrValChangeDefinition;
	if (_attrValueList)
	    attrValue = (AttrValue*) _attrValueList->first();
	else
	    attrValue = &av;
	if (attrValChangeDefinition_fill(*curr, attrValue) == NOTOK)
	    return NOTOK;

	// iterate through the list and fill in the rest
	if (_attrValueList) {
	    ListIterator iterator(_attrValueList);
	    iterator.getNext();	// we have done the first one
	    curr = &_attrValChangeInfo->attrValChangeDefinition->next;
	    while (attrValue = (AttrValue*) iterator.getNext()) {
	        *curr = new type_SMI_AttrValChangeDefinition;
		(*curr)->attrValChangeDefinition_member = new type_SMI_AttrValChangeDefinition::element_SMI_4;
		(*curr)->attrValChangeDefinition_member->attributeId = new MIDentifier;
	        if (attrValChangeDefinition_fill(*curr, attrValue) == NOTOK)
		    return NOTOK;
	        curr = &(*curr) -> next;
	        *curr = NULL;
	    }
	    delete _attrValueList;	// this could be left to the user
	    _attrValueList = NULL;
	}

	_attrValChangeInfo->sourceIndicator->parm = _source;
	_attrValChangeInfo->additionalText = _addText;

        info = (void *) _attrValChangeInfo;
    }
	break;

    default:
	return NOTOK;
    }

    _source = int_SMI_SourceIndicator_unknown;	// "initialise" for next time
    return OK;
}


int  Top::attrValChangeDefinition_fill(type_SMI_AttrValChangeDefinition* avcd,
					AttrValue* av)
{
    if (av->_attrName) {
	findAttr(av->_attrName, &av->_attrId, &av->_classLevel);
	if (av->_attrId == -1)
	    return NOTOK;
    }
    avcd->attrValChangeDefinition_member->attributeId->mid_type = MID_GLOBAL;
    avcd->attrValChangeDefinition_member->attributeId->mid_global =
	_classes[av->_classLevel] -> getAttrOid(av->_attrId);
    avcd->attrValChangeDefinition_member->newAttrValue =
	MO::_attrs[av->_classLevel][av->_attrId] -> encode();
    avcd->attrValChangeDefinition_member->oldAttrValue = av->_attrValue ?
	MO::_attrs[av->_classLevel][av->_attrId] -> encode(av->_attrValue) :
	NULLPE;
    if (av->_attrValue) {
	MO::_attrs[av->_classLevel][av->_attrId] -> ffree(av->_attrValue);
	av->_attrValue = NULLVD;
    }
    return OK;
}


int  Top::processNotification (MO* mo,
                        OID eventType, int eventId, int classLevel,
                        OID moClass, MN moInst, char* eventTime, PE eventInfo)
{
    if (!_notificationFunction) {
	advise(NULLCP, "Top::processNotification: not notification function \
but triggered event(s)!");
	return NOTOK;
    }

    /* Set up testRecord info */
    struct MIDentifier mo_class, event_type;
    MName       mo_inst;
    char*       event_time;
    PE          event_info = mo ? NULLPE : eventInfo;
    Bool        firstLog = True, freeFlag = True;
    void        *notification = (void *) 0;

    mo_class.mid_type = event_type.mid_type = MID_GLOBAL;
    mo_class.mid_global = mo ? mo -> getClass() : moClass;
    event_type.mid_global = eventType;

    mo_inst.mn_type = mo ? MN_DN : moInst -> mn_type;
    mo_inst.mn_dn = mo ? mo -> getDN(globalNames ? True : False)
		       : moInst -> mn_dn;

    event_time = mo ? mtime() : eventTime;

    if (mo != NULLMO) {
        if (mo -> buildReport(eventId, classLevel, notification, freeFlag)
		== NOTOK) {
#ifdef DEBUG
            advise(NULLCP, "MO Class %s: failed to make report for event %s!",
			mo->getClassName(), mo->getEventName(eventType));
#endif
            return NOTOK;
        }
    }

    /* Get the record to be used for this notification */
    register
    char *notificationName;

    if (mo)
	notificationName = mo->getEventName(eventType);
    else
    if ((notificationName = oid2name(eventType, OIDPART)) == NULLCP) {
	advise(NULLCP, "Top::processNotification: cannot forward unknown \
notification - oid %s !",
			sprintoid(eventType));
	return NOTOK;
    }


    /* Changes need to be made below */
    register
    LogRecord *testRecord;

    testRecord = notificationMapper->mapNotification(notificationName);
    if (testRecord == (LogRecord *) 0) {
        advise((char *) 0, "Top::processNotification(): could not map notification %s", notificationName);
        return NOTOK;
    }

    /* Put the notification information into the logRecord */
    if (event_info == (PE) 0) {

        event_info = notificationMapper->encodeNotification(notificationName, notification);
        if (event_info == (PE) 0) {
            advise((char *) 0, "Top::processNotification(): could not encode notification %s", notificationName);
            return NOTOK;
        }

        switch ((int) testRecord->setInfo(notification)) {

        case (int) LogRecordInfo_ok:
            /* All OK */
            break;

        case (int) LogRecordInfo_failed:
            /* Something went wrong */
            advise((char *) 0, "Top::processNotification(): could not set notification info for %s", notificationName);
            return NOTOK;

        case (int) LogRecordInfo_wrongRecordType:
            /* The wrong record type for this notification was used */
            advise((char *) 0, "Top::processNotification(): Wrong record type for notification %s", notificationName);
            return NOTOK;

        case (int) LogRecordInfo_noRecordType:
            /* No record types exists make ad hoc usage of eventLogRecord */
            ((EventLogRecord *) testRecord)->setInfo(event_info);
            break;
        }
    }

    /* Encoded notification was given - make ad hoc usage of eventLogRecord */
    else ((EventLogRecord *) testRecord)->setInfo(event_info);

    /* Set class, name and event type */
    testRecord->setInfo(mo_class.mid_global, mo_inst.mn_dn,
                        event_type.mid_global);


    // find the MIT level of efds and logs (assume always bound to top MO)

    MO* topMO;
    register MO* cur;
    if (mo) {
	// there may be more than one MITs in an OSIMIS process in the future,
	// so we find the topMO for mo
	for (cur = mo; cur->_superior->_superior; cur = cur->_superior);
	topMO = cur;
    }
    else
	// in this case the event report was generated somewhere else
	// and forwarded; some indication which AET/MIT to be used
	// will be needed in the future, this is a temporary solution
	topMO = getRoot()->getSubordinate();
	    
	
    // check all event forwarding discriminators

    for (cur = topMO->_subordinate; cur; cur = cur->_peer) {
	if (cur->checkClass("eventForwardingDiscriminator") != OK)
	    continue;

        register eventForwardingDiscriminator* discr =
        			(eventForwardingDiscriminator*) cur;

        if (*(AdminStateVal*) discr -> administrativeState() -> getval()
            	== as_locked ||
            *(OperStateVal*) discr -> operationalState() -> getval()
            	== os_disabled ||
            testRecord -> Filter((CMISFilter*)
		discr -> discriminatorConstruct() -> getval(), DISCARDAVA)
            	== False)
            continue;

        // send the report to the destination(s) specified in the EFD
	discr -> sendEventReport(&mo_class, &mo_inst, &event_type,
				 event_info, event_time);
    }

#ifdef	LOGGING
    // check all logs

    for (cur = topMO->_subordinate; cur; cur = cur->_peer) {
	if (cur->checkClass("log") != OK)
	    continue;

	register Log* llog = (Log*) cur;

        if (*(AdminStateVal*) llog -> administrativeState() -> getval()
                        != as_unlocked ||
                *(OperStateVal*) llog -> operationalState() -> getval()
                        != os_enabled ||
                testRecord -> Filter((CMISFilter*)
			llog -> discriminatorConstruct() -> getval(), DISCARDAVA)
                        != True)
            continue;

        /* (Kind of) Place the testRecord into the MIT for a while ... */
        testRecord->setSuperior((Log *) llog);

        /* ... store the event report in the log ... */
        ((Log *) llog) -> store(testRecord);

        /* ... now take it back out of the MIT */
        testRecord->setSuperior(NULLMO);
    }
#endif

    if (mo && event_info)
        pe_free(event_info);

    if (notification != NULLVD && freeFlag)
        notificationMapper->freeNotification(notificationName, notification);

    return OK;
}
