#include "CmisAgent.h"
#include "Coordinator.h"
#include "OperationalState.h"
#include "SimpleNameType.h"


// eventForwardingDiscriminator methods

int  eventForwardingDiscriminator::createRR (AVA*& err, void* v, int)
{
    // if no destination was supplied, default to this association
    if (*(CreationType*) v == ct_managementOperation &&
				!destination() -> wasSet()) {
	struct type_SMI_Destination* dest = new type_SMI_Destination;
	char* cp;
	dest->offset = type_SMI_Destination_singleAET;
	dest->un.singleAET = dn_comp_new(NULLRDN);
	destination()->set((void*) dest);
	destination()->setMsd(CMISAgent::getInstance()->getThisAssoc());
    }
    destination()->resetWasSet();

    // call the parent class createRR method
    if (discriminator::createRR(err) == NOTOK)
        return NOTOK;

    // register with the Coordinator to optimise wake-ups
    Coordinator::getInstance()->registerReportObj();

    return OK;
}


int  eventForwardingDiscriminator::deleteRR (AVA*& err, void* info,
						Bool checkOnly, int)
{
    // deregister with the Coordinator
    if (!checkOnly)
	Coordinator::getInstance()->deregisterReportObj();

    return discriminator::deleteRR(err, info, checkOnly);
}


int  eventForwardingDiscriminator::set
	(CMISModifyOp mode, int attrId, int classLevel, void* value,
	 AVA*& err, Bool checkOnly, int)
{
    OperStateVal* os;

    if (classLevel != _level)
	return discriminator::set(mode, attrId, classLevel, value,
				  err, checkOnly, -1);
    if (checkOnly)
	return OK;
 
    switch (attrId) {
    case I_destination:
	if (*(os = (OperStateVal*) operationalState()->getval()) == os_disabled)
	    *os = os_enabled;
	break;

    default:
	break;
    }

    return m_noError;
}


int  eventForwardingDiscriminator::sendEventReport
			(MO* mo, MID moClass, MN moInst,
			 MID eventType, PE eventInfo, char* eventTime)
{
    Bool succeededAtLeastOnce = False, *b;
    register type_SMI_Destination* dest =
		(type_SMI_Destination*) destination()->getval();
    struct type_SMI_DistinguishedNameList first, *cur;

    // set up aet so that we treat single and multiple in the same loop
    if (dest->offset == type_SMI_Destination_singleAET) {
	first.DistinguishedName = dest->un.singleAET;
	first.next = NULL;
	cur = &first;
    }
    else
	cur = dest->un.multipleAET;
    AetStateArray* aetStateArray = destination()->getAetState();

    for (register int i = 0; cur; i++, cur = cur->next) {
	AetState* aetState = (AetState*) aetStateArray->get(i);
	if (destination()->getMsd() == -1 &&	// no "thisAssociation"
		aetState->state == False)
	    continue;
        if (CMISAgent::getInstance()->
			sendEventReport(mo, moClass, moInst, eventType,
                	eventInfo, eventTime,
			aetState ? &aetState->aet : (AET*) NULL,
			destination()->getMsd()) == OK) {
	    succeededAtLeastOnce = True;
	    continue;
	}
	aetState->state = False; // it is set in the array as it points to it
    }

    if (!succeededAtLeastOnce) {
        // event report failed for all destinations, disable the EFD
	void* oldStateVal = operationalState()->copy();
	destination()->setMsd(-1);
	operationalState()->set(os_disabled);
	triggerStateChange("operationalState", oldStateVal);
#ifdef DEBUG
        advise(NULLCP, "EFD::sendEventReport \
failed, makes eFDiscr with discriminatorId=%d disabled",
			*(int*) discriminatorId() -> getval());
#endif
	return NOTOK;
    }
    return OK;
}


void  eventForwardingDiscriminator::checkOnDisconnect (int msd)
{
    // check in case any discriminator with "thisAssociation" destination
    // sends event reports to this association and disable it

    if (!notificationFunction())
	return;

    // check all event forwarding discriminators

    MO* topMO = getRoot()->getSubordinate();

    for (register MO* cur = topMO->getSubordinate(); cur; cur = cur->getPeer()){
	if (cur->checkClass("eventForwardingDiscriminator") != OK)
	    continue;

        register eventForwardingDiscriminator* discr =
        			(eventForwardingDiscriminator*) cur;
        struct type_SMI_Destination* dest = (struct type_SMI_Destination*)
                                        discr -> destination() -> getval();
	DN title = dest->offset == type_SMI_Destination_singleAET ?
		   dest->un.singleAET :
		   dest->un.multipleAET->DistinguishedName;
	if (dn_cmp(title, CMISAgent::getInstance()->getEmptyDN()) == 0 &&
		discr->destination()->getMsd() == msd &&
		*((OperStateVal*) discr->operationalState()->get())
							== os_enabled) {
	    void* oldStateVal = discr->operationalState()->copy();
	    discr->operationalState()->set(os_disabled);
	    discr->triggerStateChange("operationalState", oldStateVal);
#ifdef DEBUG
	    advise(NULLCP, "EFD::checkOnDisconect: makes eFDiscr with \
discriminatorId=%d disabled",
			*(int*) discr -> discriminatorId() -> getval());
#endif
	}
    }
}
