#include "CmisAgent.h"
#include "Coordinator.h"
#include "OperationalState.h"
#include "SimpleNameType.h"


// eventForwardingDiscriminator methods

int  eventForwardingDiscriminator::createRR (AVA*& err, void*, int)
{
    // if no destination was supplied, default to this association
    if (!destination() -> wasSet()) {
	struct type_SMI_Destination* dest = new type_SMI_Destination;
	dest->offset = type_SMI_Destination_singleAET;
	dest->un.singleAET = new type_SMI_AETitle;
	dest->un.singleAET->title = oid_cpy(str2oid("0.0"));
	dest->un.singleAET->qualifier = strdup("thisAssociation");
	destination()->set((void*) dest);
	destination()->setMsd(CMISAgent::getInstance()->getThisAssoc());
    }

    // call the parent class createRR method
    if (discriminator::createRR(err) == NOTOK)
        return NOTOK;

    // register with the Coordinator to optimise wake-ups
    Coordinator::getInstance()->registerReportObj();

    return OK;
}


int  eventForwardingDiscriminator::deleteRR (AVA*& err, void* info,
						Bool checkOnly, int)
{
    // deregister with the Coordinator
    if (!checkOnly)
	Coordinator::getInstance()->deregisterReportObj();

    return discriminator::deleteRR(err, info, checkOnly);
}


int  eventForwardingDiscriminator::set
	(CMISModifyOp mode, int attrId, int classLevel, void* value,
	 AVA*& err, Bool checkOnly, int)
{
    OperStateVal* os;

    if (classLevel != _level)
	return discriminator::set(mode, attrId, classLevel, value,
				  err, checkOnly, -1);
    if (checkOnly)
	return OK;
 
    switch (attrId) {
    case I_destination:
	if (*(os = (OperStateVal*) operationalState()->getval()) == os_disabled)
	    *os = os_enabled;
	break;

    default:
	break;
    }

    return m_noError;
}


int  eventForwardingDiscriminator::sendEventReport
	(MID moClass, MN moInst, MID eventType, PE eventInfo, char* eventTime)
{
    Bool succeededAtLeastOnce = False, *b;
    register type_SMI_Destination* dest =
		(type_SMI_Destination*) destination()->getval();
    struct type_SMI_AETitleList first, *aet;

    // set up aet so that we treat single and multiple in the same loop
    if (dest->offset == type_SMI_Destination_singleAET) {
	first.AETitle = dest->un.singleAET;
	first.next = NULL;
	aet = &first;
    }
    else
	aet = dest->un.multipleAET;
    Array* multipleAetState = destination()->getMultipleAetState();
    
    for (register int i = 0; aet; i++, aet = aet->next) {
	if (multipleAetState && *(b = (Bool*)multipleAetState->get(i)) == False)
	    continue;
        if (CMISAgent::getInstance()->
			sendEventReport(moClass, moInst, eventType,
                	eventInfo, eventTime, aet->AETitle,
			destination()->getMsd()) == OK) {
	    succeededAtLeastOnce = True;
	    continue;
	}
	if (multipleAetState)
	    *b = False;		// it is set in the array as it points to it
    }

    if (!succeededAtLeastOnce) {
        // event report failed for all destinations, disable the EFD
	void* oldStateVal = operationalState()->copy();
	destination()->setMsd(-1);
	operationalState()->set(os_disabled);
	triggerStateChange("operationalState", oldStateVal);
#ifdef DEBUG
        advise(NULLCP, "EFD::sendEventReport \
failed, makes eFDiscr with discriminatorId=%d disabled",
			*(int*) discriminatorId() -> getval());
#endif
	return NOTOK;
    }
    return OK;
}


void  eventForwardingDiscriminator::checkOnDisconnect (int msd)
{
    // check in case any discriminator with "thisAssociation" destination
    // sends event reports to this association and disable it

    if (!notificationFunction())
	return;

    // check all event forwarding discriminators

    MO* topMO = getRoot()->getSubordinate();

    for (register MO* cur = topMO->getSubordinate(); cur; cur = cur->getPeer()){
	if (cur->checkClass("eventForwardingDiscriminator") != OK)
	    continue;

        register eventForwardingDiscriminator* discr =
        			(eventForwardingDiscriminator*) cur;
        struct type_SMI_Destination* dest = (struct type_SMI_Destination*)
                                        discr -> destination() -> getval();
	OID title = dest->offset == type_SMI_Destination_singleAET ?
		   dest->un.singleAET->title :
		   dest->un.multipleAET->AETitle->title;
	if (oid_equal(title, str2oid("0.0")) &&
		discr->destination()->getMsd() == msd &&
		*((OperStateVal*) discr->operationalState()->get())
							== os_enabled) {
	    void* oldStateVal = discr->operationalState()->copy();
	    discr->operationalState()->set(os_disabled);
	    discr->triggerStateChange("operationalState", oldStateVal);
#ifdef DEBUG
	    advise(NULLCP, "EFD::checkOnDisconect: makes eFDiscr with \
discriminatorId=%d disabled",
			*(int*) discr -> discriminatorId() -> getval());
#endif
	}
    }
}
