#include "CmisAgent.h"
#include "Coordinator.h"
#include "OperationalState.h"
#include "SimpleNameType.h"
#include "CorbaAgent.h"


// eventForwardingDiscriminator methods

RDN eventForwardingDiscriminator::makeRdn (MO* mo)
{
    return discriminator::makeRdn(mo);
}


int  eventForwardingDiscriminator::createRR (AVA*& err, void* v, int)
{
    CMISAgent* cmisAgent = CMISAgent::getInstance();

    // if no destination was supplied and the creation is through
    // the OSI agent, default to this association, otherwise abort

    if (*(CreationType*) v == ct_managementOperation &&
					!destination() -> wasSet())
	if (cmisAgent && cmisAgent->getThisAssoc() >= 0) {
	    struct type_SMI_Destination* dest = new type_SMI_Destination;
	    char* cp;
	    dest->offset = type_SMI_Destination_singleAET;
	    dest->un.singleAET = dn_comp_new(NULLRDN);
	    destination()->set((void*) dest);
	    destination()->setMsd(cmisAgent->getThisAssoc());
	}
	else {	// creation through the corba agent
	    err = new AVA(m_missingAttributeValue, "destination");
	    advise(NULL, "EFD being created is missing destination!\n");
	    return NOTOK;
	}
    destination()->resetWasSet();

    // call the parent class createRR method
    if (discriminator::createRR(err) == NOTOK)
        return NOTOK;

    // register with the Coordinator to optimise wake-ups
    Coordinator::getInstance()->registerReportObj();

    return OK;
}


int  eventForwardingDiscriminator::deleteRR (AVA*& err, void* info,
						Bool checkOnly, int)
{
    // deregister with the Coordinator
    if (!checkOnly)
	Coordinator::getInstance()->deregisterReportObj();

    return discriminator::deleteRR(err, info, checkOnly);
}


int  eventForwardingDiscriminator::set
	(CMISModifyOp mode, int attrId, int classLevel, void* value,
	 AVA*& err, Bool checkOnly, int)
{
    OperStateVal* os;

    if (classLevel != _level)
	return discriminator::set(mode, attrId, classLevel, value,
				  err, checkOnly, -1);
    if (checkOnly)
	return OK;
 
    switch (attrId) {
    case I_destination:
	if (*(os = (OperStateVal*) operationalState()->getval()) == os_disabled)
	    *os = os_enabled;
	break;

    default:
	break;
    }

    return m_noError;
}


int  eventForwardingDiscriminator::sendEventReport
			(MO* mo, MID moClass, MN moInst,
			 MID eventType, PE eventInfo, char* eventTime)
{
    register int i;
    Bool succeededAtLeastOnce = False, *b;
    register type_SMI_Destination* dest =
		(type_SMI_Destination*) destination()->getval();
    struct type_SMI_DistinguishedNameList first, *cur;

    if (dest->offset != type_SMI_Destination_corbaObject) {

        CMISAgent* cmisAgent = CMISAgent::getInstance();
	if (cmisAgent) {

            // set up aet so that we treat single and multiple in the same loop
            if (dest->offset == type_SMI_Destination_singleAET) {
	        first.DistinguishedName = dest->un.singleAET;
	        first.next = NULL;
	        cur = &first;
            }
            else
	        cur = dest->un.multipleAET;
            AetStateArray* aetStateArray = destination()->getAetState();

            for (i = 0; cur; i++, cur = cur->next) {
	        AetState* aetState = (AetState*) aetStateArray->get(i);
	        if (destination()->getMsd() == -1 &&	// no "thisAssociation"
		        aetState->state == False)
	            continue;
                if (cmisAgent -> sendEventReport(mo, moClass, moInst,
				 eventType, eventInfo, eventTime,
				 aetState ? &aetState->aet : (AET*) NULL,
				 destination())
			== OK) {
	            succeededAtLeastOnce = True;
	            continue;
	        }
		// it is set in the array, it points to it
	        aetState->state = False;
            }
	}
    }
    else {

        CorbaAgent* corbaAgent = CorbaAgent::getInstance();
	if (corbaAgent && corbaAgent -> sendEventReport(mo, moClass, moInst,
				 eventType, eventInfo, eventTime,
				 (AET*) NULL, destination())
			== OK)
	    succeededAtLeastOnce = True;
    }

    if (!succeededAtLeastOnce) {
        // event report failed for all destinations, disable the EFD
	void* oldStateVal = operationalState()->copy();
	destination()->setMsd(-1);
	operationalState()->set(os_disabled);
	triggerStateChange("operationalState", oldStateVal);
#ifdef DEBUG
	char* cp = discriminatorId() -> print();
        advise(NULLCP, "EFD::sendEventReport \
failed, makes eFDiscr with discriminatorId=%s disabled", cp);
	free(cp);
#endif
	return NOTOK;
    }
    return OK;
}


void  eventForwardingDiscriminator::checkOnDisconnect (int msd)
{
    // check in case any discriminator with "thisAssociation" destination
    // sends event reports to this association and disable it

    if (!notificationFunction())
	return;

    // check all event forwarding discriminators

    MO* topMO = getRoot()->getSubordinate();

    for (register MO* cur = topMO->getSubordinate(); cur; cur = cur->getPeer()){
	if (cur->checkClass("eventForwardingDiscriminator") != OK)
	    continue;

        register eventForwardingDiscriminator* discr =
        			(eventForwardingDiscriminator*) cur;
        struct type_SMI_Destination* dest = (struct type_SMI_Destination*)
                                        discr -> destination() -> getval();

	if (dest->offset == type_SMI_Destination_corbaObject)
	    continue;

	DN title = dest->offset == type_SMI_Destination_singleAET ?
		   dest->un.singleAET :
		   dest->un.multipleAET->DistinguishedName;
	if (dn_cmp(title, MO::getEmptyDN()) == 0 &&
		discr->destination()->getMsd() == msd &&
		*((OperStateVal*) discr->operationalState()->get())
							== os_enabled) {
	    void* oldStateVal = discr->operationalState()->copy();
	    discr->operationalState()->set(os_disabled);
	    discr->triggerStateChange("operationalState", oldStateVal);
#ifdef DEBUG
	    advise(NULLCP, "EFD::checkOnDisconect: makes eFDiscr with \
discriminatorId=%d disabled",
			*(int*) discr -> discriminatorId() -> getval());
#endif
	}
    }
}
