#include "Integer.h"

extern Bool globalNames;	/* in CmisAgent.cc */


int
eventLogRecord::createRR (AVA*& err, void*, int)
{
    if (logRecord::createRR(err) == NOTOK)
	return NOTOK;

    // the following is needed for correct filtering on global names
    managedObjectInstance() -> setContainingMO(this);
    return OK;
}

int
eventLogRecord::deleteRR (AVA*& err, void* info, Bool checkOnly, int)
{
    if (checkOnly)
        return OK;

    /*
    ** This delereRR is only for the test object used by the GMS to process
    ** a notification for generating Event Reports or for logging.
    **
    ** The values in the test object could be in use elsewhere.
    ** To prevent them being deleted when the test object is deleted,
    ** the _attrs[] pointers for the test object are set to NULL.
    */
    if (logRecordId()->getint() <= 0) {
        managedObjectClass()->replace(NULLVD);
        managedObjectInstance()->replace(NULLVD);
        eventType()->replace(NULLVD);
    }

    return logRecord::deleteRR(err, info, checkOnly);
}

void
eventLogRecord::setInfo(OID object, DN name, OID event)
{
    /*
    ** This method is for the use of the test object!
    */
    if (logRecordId()->getint() > 0)
        return;

    static Bool first = True;
    static struct MIDentifier objClass;
    static struct MIDentifier evType;
    static MName       objInst;


    if (first) {
	managedObjectClass()->setval(NULL);
	managedObjectInstance()->setval(NULL);
	eventType()->setval(NULL);
	first = False;
    }

    objClass.mid_type = evType.mid_type = MID_GLOBAL;

    objClass.mid_global = object;
    managedObjectClass()->replval(&objClass);

    evType.mid_global = event;
    eventType()->replval(&evType);

    objInst.mn_type = globalNames ? MN_DN : MN_LOCALDN;
    objInst.mn_dn = name;

    managedObjectInstance()->replval(&objInst);

    eventTime()->set();
}

LogRecordInfo
eventLogRecord::setInfo(const void *)
{
    /*
    ** If this is a derived class, then this method should have been
    ** redefined to map the notification to the particular record type!
    */
    if (strcmp("eventLogRecord", getClassName()) != 0) {
        advise((char *) 0, "eventLogRecord::setInfo(): This should be redefined for class %s", getClassName());
        additionalText()->set((void *) "<Unknown Value>");
        return LogRecordInfo_wrongRecordType;
    }

    /* This notification has no associated record type */
    return LogRecordInfo_noRecordType;
}

void
eventLogRecord::setInfo(const PE info)
{
    if (info == (PE) 0) {
        advise((char *) 0, "eventLogRecord::setInfo(): NULL info!");
        additionalText()->set((void *) "<NULL>");
        return;
    }
    /* Make a hexdump copy of the BER and store it */
    /* Non-standard behaviour */

    char *str = (char *) 0;
    int length;

    if (pe2mem(info, &str, &length) == NOTOK)
        additionalText()->set((void *) "<Bad value>");
    else {
        register
        char *bp, *sp, *buffer = new char[2*length + 1];
        register
        int i;

        for(i = 0, bp = buffer, sp = str; i < length; ++i, bp += 2, ++sp)
            sprintf(bp, "%02x", (u_int) *((u_char *) sp));
        *bp = '\0';

        additionalText()->set((void *) buffer);
	delete buffer;
    }

    if (str != (char *) 0)
        free(str);
}

