//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// IsodeICC.cc - ISODE ICC realisation

// By George Pavlou, October 1991


#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>

#ifdef  __GNUG__
#include <std.h>
#include <netdb-gnu.h>
#include <netinet/in-gnu.h>
#elif	__cplusplus
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#endif

#include "msap.h"
#include "Error.h"
#include "transportEntity.h"
#include "transportConnection.h"
#include "IsodeICC.h"


IsodeICC* IsodeICC::_instance;
static transportEntity* _tpEntity;


int  IsodeICC::initialisePort ()
{
    struct servent * sp;
    struct sockaddr_in sin;

    if ((sp = getservbyname("manager", (char*) 0)) == 0) {
	advise(NULLCP, "no \"manager\" port for isode management!");
	return NOTOK;
    }
    if ((_port = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("sma"); return NOTOK;
    }
    sin.sin_family = AF_INET;
    sin.sin_port = sp->s_port;
    sin.sin_addr.s_addr = INADDR_ANY;
    if (bind(_port, (struct sockaddr*) &sin, sizeof(sin)) < 0) {
	perror("sma"); return NOTOK;
    }
    listen(_port, 5);
    return OK;
}


int  IsodeICC::registerMO (MO* mo, char* moclass, RDN)
{
    if (++_nobjs > 1)
	return _nobjs;

    // first object
    if (strcmp(moclass, "transportEntity") == 0)
	_tpEntity = (transportEntity*) mo;
    else {
	advise(NULLCP, "IsodeICC::registerMO: unknown class %s", moclass);
	return NOTOK;	// we do not expect anything else to register...
    }

    if (initialisePort() == NOTOK)
	return NOTOK;

    startListen(_port);

    return 1;
}


int  IsodeICC::deregisterMO (MO*)
{
    if (_nobjs == 0) {
	advise(NULLCP, "IsodeICC::deregisterMO: no objects registered!");
	return NOTOK;
    }

    if (--_nobjs > 0)
	return _nobjs;

    // last object gone...
    stopListen(_port);
    closePort();

    return 0;
}


int  IsodeICC::readCommEndpoint (int port)
{
    MReport minfo;
    transportConnection* tpConn;
    char id[32];
    RDN rdn;

    // only the transport layer is managed at present

    read(port, (char*) &minfo, sizeof(MReport));	// read mgmt info

    switch(minfo.type) {
    case OPREQIN:
    case OPREQOUT:
	sprintf(id, "%s%d%d", "connectionId=", minfo.id, minfo.cid);
	tpConn = (transportConnection*)
			transportConnection::create(str2rdn(id), _tpEntity);
	tpConn -> createRR(NULLAVAREF);
	tpConn -> update(&minfo);
	break;

    case SOURCEADDR:
    case USERDT:
    case USERDR:
    case DISCREQ:
	sprintf(id, "%s%d%d", "connectionId=", minfo.id, minfo.cid);
	rdn = str2rdn(id);
	if ((tpConn = (transportConnection*)
		transportConnection::getClassInfo()->getMO(rdn)) == NULLMO) {
	    advise(NULLCP, "cannot find tpConn %s", id);
	    rdn_free(rdn);
	    return NOTOK;
	}
	tpConn -> update(&minfo);

	// trigger the objectCreation/Deletion notification
	if (minfo.type == SOURCEADDR)
	    tpConn -> triggerObjectCreation();
	else
	if (minfo.type == DISCREQ) {
	    tpConn -> deleteRR(NULLAVAREF);
	    tpConn -> triggerObjectDeletion();
	    delete tpConn;
	}

	rdn_free(rdn);
	break;

    case OPREQINBAD:
    case OPREQOUTBAD:
    case CONGEST:
    case CONFIGBAD:
    case PROTERR:
	_tpEntity -> update(&minfo);
	break;

    default:
	advise(NULLCP, "unknown isode transport MReport type %d",
			minfo.type);
	return NOTOK;
    }
    return OK;
}
