#if     defined(__cplusplus) && !defined(__GNUG__)
#include <sysent.h>
#endif

#include "SmiSntx.h"
#include "IsodeICC.h"
#include "GraphicString.h"


int transportEntity::createRR (AVA*& err, void*, int)
{
    // this is a single-instanced class, so check if this is the 2nd instance
    if (MO::getClassInfo() -> getFirst() -> getClassNext()) {
        err = new AVA(m_processingFailure, "singleInstanceClass");
        return NOTOK;
    }

    // call first the parent class createRR method
    if (entity::createRR(err) == NOTOK)
	return NOTOK;

    static Bool classInitialised;

    // initialise Entity localSapName attribute
    char hostname[128];
    gethostname(hostname, 128);
    entity::localSapName() -> set(hostname);

    // initialise the ISODE KS (if not initialised)
    IsodeICC* isodeICC;
    if (!(isodeICC = IsodeICC::getInstance()))
        isodeICC = new IsodeICC;

    // register with it
    isodeICC->registerMO(this, "transportEntity", getRDN());


    // set the type of the gauge (integer) as the GDMO compiler does not know
    openConnections() -> setType(OV_INTEGER);

    if (!classInitialised)
	classInitialised = True;

    return OK;
}


int transportEntity::deleteRR (AVA*& err, void* info, Bool checkOnly, int)
{
    IsodeICC* isodeICC = IsodeICC::getInstance();

    if (!checkOnly) {
        // deregister from the isode knowledge source
        isodeICC->deregisterMO(this);

        // kill the latter if there are no other registered objects
        if (isodeICC->getNObjects() == 0)
            delete isodeICC;
    }

    // the following will do nothing but for correct OO style...
    return entity::deleteRR(err, info, checkOnly);
}


void  transportEntity::update (MReport* minfo)
{
    switch (minfo -> type) {
    case OPREQINBAD:
	totalNumberTpduReceived() -> increment();
	localUnsuccessfulConnections() -> increment();
	break;

    case OPREQOUTBAD:
	totalNumberTpduSent() -> increment();
	remoteUnsuccessfulConnections() -> increment();
	break;

    case CONGEST:
    case PROTERR:
	localUnsuccessfulConnections() -> increment();
	totalNumberTpduReceived() -> increment();
	localCongestionErrors() -> increment();
	break;

    case CONFIGBAD:
	localUnsuccessfulConnections() -> increment();
	totalNumberTpduReceived() -> increment();
	localCongestionErrors() -> increment();
	break;
    }
}

