/*
* NAME
*   TimePeriod.h
*
* DESCRIPTION
*   ISO/IEC 10164-11 TimePeriod Attribute from supporting ASN.1 defns
*
* AUTHOR
*   Jim Reilly
*   Technical Research Centre of Finland
*   icm.group@tel.vtt.fi
*/

#ifndef _TIMEPERIOD_
#define _TIMEPERIOD_


#ifndef _GENERIC_ATTR_
#include "GenericAttr.h"
#endif
 
#ifndef _MONITOR_SNTX_
#include "MonitorSntx.h"
#endif


class TimePeriod : public Attr
{
protected:
    PE			_encode ();
    void*		_decode (PE pe);
    void		_free ();
    void*		_parse (char* str);
    char*		_print ();
    int			_compare (void* v1, void* v2);

public:
    int set(void* newval);
				// redefine Attr::set(void*) to check if
				// new value units is supported
				// i.e. units secs ... days

    TimePeriod ();
    TimePeriod (type_Monitor_TimePeriod* tp);
};


inline PE TimePeriod::_encode () {
  return TimePeriod_encode((type_Monitor_TimePeriod*) getval());
}

inline void* TimePeriod::_decode (PE pe) { return TimePeriod_decode(pe); }

inline void  TimePeriod::_free () {
  TimePeriod_free((type_Monitor_TimePeriod*)getval());
}

inline void* TimePeriod::_parse (char* str) {
  return TimePeriod_parse(str);
}

inline char* TimePeriod::_print () {
  return attrv2str(getval(), (PRINT_FNX) TimePeriod_print);
}

inline int TimePeriod::_compare (void* v1, void* v2) {
  return TimePeriod_cmp((type_Monitor_TimePeriod*) v1,
                        (type_Monitor_TimePeriod*) v2);
}

inline int TimePeriod::set(void* newval) {
  type_Monitor_TimePeriod* tp = (type_Monitor_TimePeriod*)newval;
  if ( (tp->offset == type_Monitor_TimePeriod_milliSeconds) ||
       (tp->offset == type_Monitor_TimePeriod_microSeconds) ||
       (tp->offset == type_Monitor_TimePeriod_nanoSeconds)  ||
       (tp->offset == type_Monitor_TimePeriod_picoSeconds)
  ) {
    return NOTOK;
	// units < seconds are not supported
  } else {
    return Attr::set(newval);
  }
}

inline TimePeriod::TimePeriod() {
  type_Monitor_TimePeriod* tp = new type_Monitor_TimePeriod;
  tp->offset=type_Monitor_TimePeriod_seconds;
  tp->un.seconds=60;
  setval(tp);
}

inline TimePeriod::TimePeriod(type_Monitor_TimePeriod* tp) { set(tp); }


#endif /* _TIMEPERIOD */
