/*
* NAME
*   TimePeriod.inc.c
*
* DESCRIPTION
*   ISO/IEC 10164-11 Monitor Objects attribute syntaxes 
*
* AUTHOR
*   Jim Reilly
*   Technical Research Centre of Finland
*   icm.group@tel.vtt.fi
*/


struct type_Monitor_TimePeriod*  TimePeriod_parse (str)
char* str;
{
    int value;
    char units[5];
    struct type_Monitor_TimePeriod* tp = NULL;

    if (sscanf(str, "%4s:%d", units, &value) != 2) {
        return NULL; /* unexpected EOF */
    }

    if ((tp = (struct type_Monitor_TimePeriod*)
		malloc(sizeof(struct type_Monitor_TimePeriod)))
		== (struct type_Monitor_TimePeriod*) NULL) {
        return NULL;
    }

    if (!strcmp(units, "days")) {
        tp->offset = type_Monitor_TimePeriod_days;
        tp->un.days = value;
    } else if (!strcmp(units, "hour")) {
        tp->offset = type_Monitor_TimePeriod_hours;
        tp->un.hours = value;
    } else if (!strcmp(units, "mins")) {
        tp->offset = type_Monitor_TimePeriod_minutes;
        tp->un.minutes = value;
    } else if (!strcmp(units, "secs")) {
        tp->offset = type_Monitor_TimePeriod_seconds;
        tp->un.seconds = value;
    } else if (!strcmp(units, "msec")) {
        tp->offset = type_Monitor_TimePeriod_milliSeconds;
        tp->un.milliSeconds = value;
    } else if (!strcmp(units, "usec")) {
        tp->offset = type_Monitor_TimePeriod_microSeconds;
        tp->un.microSeconds = value;
    } else if (!strcmp(units, "nsec")) {
        tp->offset = type_Monitor_TimePeriod_nanoSeconds;
        tp->un.nanoSeconds = value;
    } else if (!strcmp(units, "psec")) {
        tp->offset = type_Monitor_TimePeriod_picoSeconds;
        tp->un.picoSeconds = value;
    } else {
        /* by default assume "seconds" are units */
        tp->offset = type_Monitor_TimePeriod_seconds;
        tp->un.seconds = value;
    }

    return tp;
}

int  TimePeriod2seconds (tp)
struct type_Monitor_TimePeriod* tp;
{
    if (tp) {
        switch (tp->offset) {
        case type_Monitor_TimePeriod_days:
            return (86400 * tp->un.seconds);	/* 24h * 60m * 60s */
        case type_Monitor_TimePeriod_hours:
            return (3600 * tp->un.seconds);	/* 60m * 60s */
        case type_Monitor_TimePeriod_minutes:
            return (60 * tp->un.seconds);	/* 60s */
        case type_Monitor_TimePeriod_seconds:
            return (tp->un.seconds);
        default:
	    /* fall thru */
            break;
        };
    }
    return 0;
}

int  TimePeriod_print (ps, tp, format)
PS ps;
register struct type_Monitor_TimePeriod* tp;
int format;
{
    char buf[5];

    if (tp) {

        switch (tp->offset) {
        case type_Monitor_TimePeriod_days: strcpy (buf, "days"); break;
        case type_Monitor_TimePeriod_hours: strcpy (buf, "hour"); break;
        case type_Monitor_TimePeriod_minutes: strcpy (buf, "mins"); break;
        case type_Monitor_TimePeriod_seconds: strcpy (buf, "secs"); break;
        case type_Monitor_TimePeriod_milliSeconds: strcpy (buf, "msec"); break;
        case type_Monitor_TimePeriod_microSeconds: strcpy (buf, "usec"); break;
        case type_Monitor_TimePeriod_nanoSeconds: strcpy (buf, "nsec"); break;
        case type_Monitor_TimePeriod_picoSeconds: strcpy (buf, "psec"); break;
        }

        ps_printf(ps, "%4s:%d", buf, tp->un.seconds);
    }
}

int  TimePeriod_compare (tp1, tp2)
struct type_Monitor_TimePeriod *tp1, *tp2;
{
    /*
     * 0 for equality, 1 for a > b, -1 for a < b, 2 for non-comparable
     */
    int secs1 = TimePeriod2seconds(tp1), secs2 = TimePeriod2seconds(tp2);
    return ( secs1 == secs2 ? 0 : (secs1 > secs2 ? 1 : -1) );
}

