//
// NAME
//   scanner.inc.cc
//
// DESCRIPTION
//   for scanner MOC, via OSIMIS GDMO compiler
//   (based on ISO/IEC 10164-11 Monitor Objects MIB)
//
// AUTHOR
//   Jim Reilly
//   Technical Research Centre of Finland
//   icm.group@tel.vtt.fi
//


int scanner::_scannerId = 0;	// initialize

RDN scanner::makeRdn(MO*) {
  char    buf[64];
  sprintf(buf, "scannerId=%d", getNewScannerId());
  return str2rdn(buf);

} //makeRdn


int scanner::createRR(AVA*& err, void*, int) {

  if (strcmp(getClassName(), "scanner") == 0) {
    err = new AVA(m_processingFailure, "abstractClass");
    return NOTOK;
  }

  //
  // MonitorMetric MOs should poll other MOs with a polling
  // interval kept in the attribute "granularityPeriod".
  // Wakeups should be scheduled using the "granularityPeriod".
  //

  int iPollInterval           = 0;
  type_Monitor_TimePeriod* tp = (type_Monitor_TimePeriod*)
    scanner::granularityPeriod()->getval();

  if ((iPollInterval = TimePeriod2seconds(tp)) == 0) {
    iPollInterval = 60; /* default value */
  }

  scheduleWakeUps(
        iPollInterval,     /* how often to be woken up (secs) */
        "MonitorObjects"   /* unique identifier string */
  );

  return OK;

} //scanner::createRR


int scanner::deleteRR(AVA*& err, void* info, Bool checkOnly, int) {
  if (!checkOnly)
    cancelWakeUps ("MonitorObjects");
  return Top::deleteRR(err, info, checkOnly); // correct OO-style
} //scanner::deleteRR


int  scanner::set (
  CMISModifyOp op, int attrId, int classLevel, void* newval, AVA*& err,
  Bool checkOnly, int
) {

    if (classLevel != _level) {
        return Top::set(op, attrId, classLevel, newval, err, checkOnly);
    } 

#if DEBUG
    advise(NULL, "!!! scanner::set");
#endif

    int iPollInterval = 0;
    type_Monitor_TimePeriod *tp = (type_Monitor_TimePeriod*) newval;

    AdminStateVal admState = *(AdminStateVal*)
      administrativeState()->getval();

    switch (attrId) {
    case I_administrativeState:
      // this is in order to be able to set it to locked, set other attrs
      // and then unlock it again in one Set request -
      // if we do not do that the GMS sets the value after it has finished
      // with all the attributes in which case the request will fail (unlocked)

      administrativeState() -> set(*(AdminStateVal*) newval);
      break;

    case I_granularityPeriod:
      if (admState == as_locked) {
	if (checkOnly)
	  return OK;

        // The granularityPeriod in the scanner managed object
        // class shall not be modified unless the value of the
        // administrative state is "locked"

        // granularityPeriod()->set((type_Monitor_TimePeriod*)newval);
        if ((iPollInterval = TimePeriod2seconds(tp)) == 0) {
          iPollInterval = 60; /* default value */
        }
  
        //
        // cancel old wakeups and schedule new ones
        // using new granularityPeriod polling interval
        //
        cancelWakeUps ("MonitorObjects");
        scheduleWakeUps(
              iPollInterval,    /* how often to be woken up (seconds) */
              "MonitorObjects"  /* unique identifier string */
        );
        advise (NULL,
            "scanner::set() new granularityPeriod=%d %d:%d\n",
            iPollInterval, tp->offset, tp->un.seconds );

      } else {
	if (checkOnly)
	  return NOTOK;

	err = new AVA(m_processingFailure, "adminStateUnlocked");
        return NOTOK;
      }

      break;

    default:
      break;
    }

    return OK;

} //scanner::set

