//
// NAME
//   MSscanner.inc.cc
//
// DESCRIPTION
//   for MSscanner MOC, via OSIMIS GDMO compiler
//   (based on ISO/IEC 10164-11 Monitor Objects MIB) 
//   but with some changes
//
// IMPLEMENTORS
//   George Mykoniatis
//   NTUA
//   mykoniat@ektor.ntua.gr
//

#include "Integer.h"
#include "iostream.h"

int MSscanner::_scannerId = 0;	// initialize

//-------------------------------------------------------------------
RDN MSscanner::makeRdn(MO*) {
  char    buf[64];
  sprintf(buf, "scannerId=%d", getNewScannerId());
  return str2rdn(buf);

} //makeRdnAndSuperior 


//-------------------------------------------------------------------
int MSscanner::createRR(AVA*& err, void*, int) {
  //
  // Intelligent Summarizers MOs should poll other MOs with a polling
  // interval kept in the attribute "granularityPeriod".
  // Wakeups should be scheduled using the "granularityPeriod".
  //

  wkup_status = 0;

  return OK;

} //MSscanner::createRR


//-------------------------------------------------------------------
int MSscanner::deleteRR(AVA*& err, void* info, Bool checkOnly, int) {
  return Top::deleteRR(err, info, checkOnly); // correct OO-style
} //MSscanner::deleteRR


//-------------------------------------------------------------------
int  MSscanner::set ( CMISModifyOp op, int attrId, int classLevel, 
				void* newval, AVA*& err, Bool checkOnly, int) 
{

    if (classLevel != _level) {
        return Top::set(op, attrId, classLevel, newval, err, checkOnly);
    } 

#if DEBUG
    advise(NULL, "!!! MSscanner::set");
#endif

    int iPollInterval = 0;

    AdminStateVal admState = *(AdminStateVal*)
      ((Integer*)MSscanner::_attrs[I_administrativeState])->getval();

    switch (attrId) {
    case I_granularityPeriod:
    {
      if (admState == as_locked) 
      {
        if (checkOnly)
	   return OK;    

        // The granularityPeriod in the MSscanner managed object
        // class shall not be modified unless the value of the
        // administrative state is "locked"

	type_Monitor_TimePeriod *tp = (type_Monitor_TimePeriod*)newval;
       	if ((iPollInterval = TimePeriod2seconds(tp)) == 0) {
       		iPollInterval = 60; /* default value */
       	}

        advise (NULL,
            "MSscanner::set() new granularityPeriod=%d %d:%d\n",
            iPollInterval, tp->offset, tp->un.seconds );
      }  
      else 
      {
         if (checkOnly)
	   return NOTOK;
	 err = new AVA(m_processingFailure, "adminStateUnlocked");
	 return NOTOK;
      }
    } //case I_granularityPeriod
    break;

   // If I_administrativeState is set to unlocked while it was
   // locked then start the wakeUps.
   // If I_administrativeState is set to locked while it was
   // unlocked then stop the wakeUps.
   case I_administrativeState:
    {
          int nval = *(int *)newval;
          switch (nval)
          {
          case as_locked:
                {
                  if (admState == as_unlocked) // If stop when active ...
		  {            
                        ShutDown();                   
                       	// Set AdministrativeState to locked
                  }
                  else                                    // If stop when inactive
                    	fprintf(stderr, "sum::set - Already inactive ...\n");
                }; 
		break;
           case as_unlocked:
                {
                  if (admState == as_unlocked)             // If start when active ...
                           fprintf(stderr, "sum::set - Already active ...\n");
                  else                                     // If start when inactive
		  if (wkup_status == 0)
		   {
                   	// Just schedule wakeups...
                   	// Initialisation will be done the first time sum::update
                   	// is executed ....

  			type_Monitor_TimePeriod* tp =
    			(type_Monitor_TimePeriod*)
    			((TimePeriod*)(MSscanner::_attrs[I_granularityPeriod]))->getval();
			
  			if ((iPollInterval = TimePeriod2seconds(tp)) == 0) {
    				iPollInterval = 60; /* default value */
  			}


#ifdef DEBUG
   cout << "scheduleWakeUps - MSscanner::set" << endl; 
   cout << "granulPeriod -> " << iPollInterval << endl;
#endif
        		if ( scheduleWakeUps(
              		   iPollInterval,
              		   NULLCP,
              		   False ) == NOTOK ) 
			{
          		   advise(NULL,
            		   "Error: MSscanner::set -- scheduleWakeUps failed");
        		} 
			else 
			{
			   wkup_status = 1;
          		   advise (NULL,
            		   "MSscanner::set() new granularityPeriod=%d %d:%d\n",
            		   iPollInterval, tp->offset, tp->un.seconds );

			   // Re-set the operationalState
      			   void* oldStateVal = operationalState()->copy();
      			   operationalState() -> set(os_enabled);
			   // Emit the StateChange notification
      			   triggerStateChange(I_operationalState, _level, oldStateVal);

        		}
                    }; // endif as_unlocked
                }; // as_unlocked
		break;  
            }; // Switch (nval);
      }; // case I_administrativeState
      break; 
    default:
      break;
    }

    return OK;

} //MSscanner::set

