//
// NAME
//   buffered.h
//
// DESCRIPTION
//   definition of the class for the statistics
//   of the Monitoring Support objects
//
// IMPLEMENTORS
//   George Mykoniatis
//   NTUA
//   mykoniat@ektor.ntua.gr
//

#include <stdio.h>
#include <math.h>
#include "msap.h"

#include "sumdefs.h"

#ifndef _BUFF_
#define _BUFF_

class buffered
{
private:
  double *array;		    	// Pointer to buffer
  double result;	        	// Result
  int    period;		    	// sampling period
  int    window;		    	// Window size
  double threshold;			// Threshold value for comparing with
					// the observed attribute value
  int    no_of_samples;			// Holds the number of samples
  int    buffer_flag;			// If true, shows that the buffer
				        // has been filled with samples
				        // and the derived result is correct
  int    index;             		// Index in the array
  double find_probability();		// For finding the probability

public:
  int initialize(int window, int period, double thres, double initial_value);
  ~buffered();
  int append_value(double sample);
  void print_buffered();
  double get_last_value();    

  double get_result()        { return result; }
  int    get_status()        { return buffer_flag; }
  int    get_no_of_samples() { return no_of_samples; }
  int    get_period()        { return period; }
};

#endif
