//
// NAME
//   monitoringSupportMode2.inc.cc
//
// DESCRIPTION
//   for monitoringSupportMode2 MOC, via OSIMIS GDMO compiler
//
// IMPLEMENTORS
//   George Mykoniatis
//   NTUA
//   mykoniat@ektor.ntua.gr
//


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "AdministrativeState.h"
#include "Integer.h"
#include "AVA.h"
#include "sumstat.h"    
#include "observed_attr.h"      
#include <CmisAgent.h>

//-------------------------------------------------------
int monitoringSupportMode2::GetAttributes()
{
  	//
  	// initialize class instance variables first
  	//
	_pBaseMO                   = NULL;
  	_pObservedAttribute        = NULL;

    	//
    	// Get Distinguished Name of watched MO in this agent
    	//
  	MN mnbaseManagedObject = (MN) baseManagedObject()->get();
  	if (!mnbaseManagedObject ||
    		(_dnBaseMO = CMISAgent::getInstance() ->
         		mn2localdn(mnbaseManagedObject)) == NULLDN) 
	{
    	  cout << "NULL or wrong baseManagedObject in monSupMode2::FindObservedMOs()"
		<< endl;
    	  return NOTOK;
  	}

  	_typeBaseMO = mnbaseManagedObject->mn_type;

       	//
	// Get OID of attribute inside MO to be watched 
       	//
  	_oidObservedAttribute = attributeId()->getOID();
  	if (!_oidObservedAttribute)
  	{
    	  advise (NULL, "NULL observedAttributeId in monSupMode2::FindObservedMOs()\n");
    	  return NOTOK;
  	}

  	MO* top = _typeBaseMO == MN_DN ?
                        MO::getRoot() : MO::getRoot() -> getSubordinate();

  	_pBaseMO = top -> getMO(_dnBaseMO, True, DISCARDAVA);

  	char* inst;

  	if (_pBaseMO == NULLMO) 
	{
    	  advise (
      	     NULL,
      	     "Error: NULLMO for DN=\"%s\" in monitoringSupportMode2::FindObservedMOs()\n",
      	     inst = dn2str(_dnBaseMO)
    	  );
    	  free(inst);
    	  _pObservedAttribute = NULL;
    	  return NOTOK;
  	}
	return OK;

} // monitoringSupportMode2::GetAttributes


//-------------------------------------------------------
int monitoringSupportMode2::FindObservedMOs()
{
	CMISScope* sc = (CMISScope *) scope()->get();

	CMISFilter* fi = (CMISFilter *) filter()->getval();

        MO** sub_arr = _pBaseMO->Scope(sc, TRUE, DISCARDAVA);

        // find the size of sub_arr
	int i = 0;
        no_of_attributes = 0;

        cout << endl;

	// Find the number of the observed attributes
        Bool val;
        while (sub_arr[i] != NULLMO)
        {
           if (fi != NULL)
           {
                val = sub_arr[i]->Filter(fi, DISCARDAVA);
                if (val) no_of_attributes++;
           }
           i++;
        }

	// sub_arr is an array of MO* 
	// pointing at the observed attributes
	_passed = new MO* [no_of_attributes];
	int j = 0;
	i = 0;

        while (sub_arr[i] != NULLMO)
        {
           if (fi != NULL)
           {
                val = sub_arr[i]->Filter(fi, DISCARDAVA);
                if (val) 
		   _passed[j++] = sub_arr[i];
           }
           i++;
        }

	return OK;

} //monitoringSupportMode2::FindObservedMOs


//-------------------------------------------------------
int monitoringSupportMode2::InitObservedAttrs()
{

   // oa_array is the array of the observed attrs
   oa_array = new observed_attr * [no_of_attributes];
   if (oa_array == NULL)
   {
     	  cout << "monitoringSupportMode1 : Failed to allocate memory for oa_array\n";
          return NOTOK;
   }

   // allocate memory for the observed attrs
   for(int i=0;i<no_of_attributes;i++)
           oa_array[i] = NULL;             // Initialise observed attrs to NULL

   for(i=0;i<no_of_attributes;i++)
   {
     oa_array[i]=new observed_attr();

     if (oa_array[i] == NULL)
     {
       fprintf( stderr, "monitoringSupportMode2 : Failed to allocate memory\n");
       return NOTOK;
     }
     // Initialise
     if ((oa_array[i])->initialise(_passed[i], _oidObservedAttribute) == NOTOK)
           return NOTOK;
#ifdef DEBUG
     else 
       oa_array[i]->print_oa();
#endif
   }
   // Now, all the class attributes are initialized...
   // so, return ok
   return OK;
} //monitoringSupportMode2::InitObservedAttrs


//-------------------------------------------------------
int monitoringSupportMode2::CheckObservedMOs()
{
	CMISScope* sc = (CMISScope *) scope()->get();

	CMISFilter* fi = (CMISFilter *) filter()->getval();

        MO** sub_arr = _pBaseMO->Scope(sc, TRUE, DISCARDAVA);

        // find the size of sub_arr
	int i = 0;
        int counter_of_attr = 0;

        cout << endl;

	// Find the number of the observed attributes
        Bool val;
        while (sub_arr[i] != NULLMO)
        {
           if (fi != NULL)
           {
                val = sub_arr[i]->Filter(fi, DISCARDAVA);
                if (val) counter_of_attr++;
           }
           i++;
        }

	// sub_arr is an array of MO* 
	// pointing at the observed attributes
	MO** passed_cp = new MO* [counter_of_attr];
	int j = 0;
	i = 0;

        while (sub_arr[i] != NULLMO)
        {
           if (fi != NULL)
           {
                val = sub_arr[i]->Filter(fi, DISCARDAVA);
                if (val) 
		   passed_cp[j++] = sub_arr[i];
           }
           i++;
        }

	// time for checking has come!
	if (counter_of_attr != no_of_attributes)
	   return NOTOK;
	// We make the assumption that the scope and
	// filter methods return the MO* with the same 
	// sequence every time.
	for (int k = 0;k<counter_of_attr;k++)
	   if (passed_cp[k] != _passed[k])
	     return NOTOK;
	
	return OK;
	
} //monitoringSupportMode2::CheckObservedMOs


//-------------------------------------------------------
int monitoringSupportMode2::CheckFirstTime()
{
    if (GetAttributes() == NOTOK)
       return NOTOK;

    if (FindObservedMOs() == NOTOK)
       return NOTOK;
    
    if (InitObservedAttrs() == NOTOK)
       return NOTOK;

    return OK;  // everything went OK..

} // monitoringSupportMode2::CheckFirstTime


//-------------------------------------------------------
int monitoringSupportMode2::CheckConsistency()
{
    if (GetAttributes() == NOTOK)
       return NOTOK;

    if (CheckObservedMOs() == NOTOK)
       return NOTOK;

    return OK;
   
} //monitoringSupportMode2::CheckConsistency


// --------------------------------------------------------------
int monitoringSupportMode2::StartUp()
{
 int window, algorithm ;
 

 // create the observed_attr instances
 if (CheckFirstTime() == OK)
 {
   // Allocate memory for the operation matrix
   oaval = new double[no_of_attributes];   
   if (oaval == NULL)
   {
	 fprintf( stderr, "StartUp: Failed to allocate memory \n");
	 return NOTOK;
   }

   // Allocate memory for the result monitoringSupportMode2 stat class
   // but get the window and algorithm first.

   window = *(int *)collectionPeriod()->get();

   int iPollInterval           = 0;
   type_Monitor_TimePeriod* tp =
     (type_Monitor_TimePeriod*)
     ((TimePeriod*)(MSscanner::_attrs[I_granularityPeriod]))->getval();

   if ((iPollInterval = TimePeriod2seconds(tp)) == 0) {
     iPollInterval = 60; /* default value */
   }

   if ( (window == 0) || (window < iPollInterval) )             // Check window also
   {
        fprintf( stderr, "monitoringSupportMode1 : Invalid movingTimePeriod\n");
        return NOTOK;
   }

   algorithm = *(int *) algorithmType()->get();

   stat_p = new sumstats;
   if (stat_p == NULL)
   {
         fprintf( stderr, "StartUp: Failed to allocate memory \n");
         return NOTOK;
   }
   // Set initial value to 0 (?)
   if (stat_p->initialize(window, iPollInterval, algorithm, 0) == NOTOK)
   {  
        fprintf( stderr, "StartUp: Failed to initialize \n");
        return NOTOK;
   }

   // DO NOT schedule wake-ups yet.. 
   // Since everything went ok
   return OK;
 }
 // Failed to initialise the attributes (DecomposeAttrList failed)...
 return NOTOK;
} //monitoringSupportMode2::StartUp


//-------------------------------------------------------
int monitoringSupportMode2::createRR(AVA*& err, void* pv, int) 
{

  // correct OO style
  if (monitoringSupport::createRR(err, pv) == NOTOK)
	return NOTOK;

  // the attributeId must be initialized!

  OID OID_attribName		= NULL;

  if (OID_attribName == NULL) {
    OID_attribName = name2oid("attributeId");
  }

  attributeId() -> setOID(OID_attribName);

  return OK;

} //monitoringSupportMode2::createRR


// --------------------------------------------------------------
int monitoringSupportMode2::set (CMISModifyOp mode, 
					int attrId, 
					int classLevel,
                      			void* value,
					AVA*& err,
					Bool checkOnly,
					int)
{
    
    if (classLevel != _level ) 
        return monitoringSupport::set(mode, attrId, classLevel, value, err, checkOnly);

#ifdef DEBUG
  fprintf(stdout, "monitoringSupportMode2::set-Executing.....\n");
#endif

    AdminStateVal admState = *(AdminStateVal*)
        MSscanner::administrativeState()->getval();

    switch (attrId)
    {
      case I_scope:
      case I_filter:
      case I_baseManagedObject:
      case I_attributeId:
        if (admState != as_locked)
	{
           if (checkOnly)
             return NOTOK;
           err = new AVA(m_invalidOperation);
           return NOTOK;
	}
        break;
      default:
        break;

    } // Switch attrid

    return OK;
}//monitoringSupportMode2::set

