//
// NAME
//   monitoringSupportMode3.inc.cc
//
// DESCRIPTION
//   for monitoringSupportMode3 MOC, via OSIMIS GDMO compiler
//
// IMPLEMENTORS
//   Kmtm - temporary polling object for the PON.
//


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream.h>

#include "AdministrativeState.h"
#include "OperationalState.h"
#include "Integer.h"
#include "AVA.h"
#include <CmisAgent.h>
#include "String.h"
#include "MSscanner.h"

//-------------------------------------------------------
int monitoringSupportMode3::GetAttributes()
{
  	//
  	// initialize class instance variables first
  	//
	_pBaseMO                   = NULL;

    	//
    	// Get Distinguished Name of watched MO in this agent
    	//
  	MN mnbaseManagedObject = (MN) baseManagedObject()->get();
  	if (!mnbaseManagedObject ||
    		(_dnBaseMO = CMISAgent::getInstance() ->
         		mn2localdn(mnbaseManagedObject)) == NULLDN) 
	{
    	  cout << "NULL or wrong baseManagedObject in monSupMode3::FindObservedMOs()"
		<< endl;
    	  return NOTOK;
  	}

  	_typeBaseMO = mnbaseManagedObject->mn_type;

  	MO* top = _typeBaseMO == MN_DN ?
                        MO::getRoot() : MO::getRoot() -> getSubordinate();

  	_pBaseMO = top -> getMO(_dnBaseMO, True, DISCARDAVA);

  	char* inst;

  	if (_pBaseMO == NULLMO) 
	{
    	  advise (
      	     NULL,
      	     "Error: NULLMO for DN=\"%s\" in monitoringSupportMode3::FindObservedMOs()\n",
      	     inst = dn2str(_dnBaseMO)
    	  );
    	  free(inst);
    	  return NOTOK;
  	} else {
	    CMISScope* sc = (CMISScope *) scope()->get();

            MO** sub_arr = _pBaseMO->Scope(sc, TRUE, DISCARDAVA);
	}
	return OK;

} // monitoringSupportMode3::GetAttributes


//-------------------------------------------------------
int monitoringSupportMode3::CheckFirstTime()
{
    if (GetAttributes() == NOTOK)
       return NOTOK;

    return OK;  // everything went OK..

} // monitoringSupportMode3::CheckFirstTime


// --------------------------------------------------------------
int monitoringSupportMode3::StartUp()
{
 // create the observed_attr instances
 if (CheckFirstTime() == OK)
 {
   int iPollInterval           = 0;
   type_Monitor_TimePeriod* tp =
     (type_Monitor_TimePeriod*)
     ((TimePeriod*)(MSscanner::_attrs[I_granularityPeriod]))->getval();

   if ((iPollInterval = TimePeriod2seconds(tp)) == 0) {
     iPollInterval = 60; /* default value */
   }

   // DO NOT schedule wake-ups yet.. 
   // Since everything went ok
   return OK;
 }
 // Failed to initialise the attributes (DecomposeAttrList failed)...
 return NOTOK;
} //monitoringSupportMode3::StartUp


//-------------------------------------------------------
int monitoringSupportMode3::createRR(AVA*& err, void* pv, int) 
{
  status = 0 ;
 
  if (MSscanner::createRR(err, pv) == NOTOK)
	return NOTOK;

  return OK;

} //monitoringSupportMode3::createRR


// --------------------------------------------------------------
int monitoringSupportMode3::set (CMISModifyOp mode, 
					int attrId, 
					int classLevel,
                      			void* value,
					AVA*& err,
					Bool checkOnly,
					int)
{
    
    if (classLevel != _level ) 
        return MSscanner::set(mode, attrId, classLevel, value, err, checkOnly);

#ifdef DEBUG
  fprintf(stdout, "monitoringSupportMode3::set-Executing.....\n");
#endif

    AdminStateVal admState = *(AdminStateVal*)
        MSscanner::administrativeState()->getval();

    switch (attrId)
    {
      case I_scope:
      case I_baseManagedObject:
        if (admState != as_locked)
	{
           if (checkOnly)
             return NOTOK;
           err = new AVA(m_invalidOperation);
           return NOTOK;
	}
        break;
      default:
        break;

    } // Switch attrid

    return OK;
}//monitoringSupportMode3::set


// --------------------------------------------------------------

int monitoringSupportMode3::deleteRR (AVA*& err, void* info, Bool checkOnly, int)
{

#ifdef DEBUG
  fprintf(stdout, "monitoringSupportMode3::deleteRR-Executing.....\n");
#endif

  ShutDown();

  return MSscanner::deleteRR(err, info, checkOnly);
} //monitoringSupportMode3::deleteRR

// --------------------------------------------------------------
// Stops the monitoringSupportMode3 instance
// --------------------------------------------------------------
int monitoringSupportMode3::ShutDown()
{
  if (wkup_status)
  {
	  wkup_status = 0;
	  cancelWakeUps(NULLCP);    // Disable wake ups
	  //MSscanner::cancelWakeUps(NULLCP);    // Disable wake ups 
  }
  status = 0;

#ifdef DEBUG
   advise(NULL,"monitoringSupportMode3::ShutDown : just executed");
#endif
  return OK;
}

// --------------------------------------------------------------
void monitoringSupportMode3::update(void)
{
 AdminStateVal admState = *(AdminStateVal*)
    MSscanner::administrativeState()->get();
 OperStateVal operState = *(OperStateVal*)
    MSscanner::operationalState()->get();

 if (admState == as_unlocked && operState == os_enabled) 
 {
  if (status == 0)  				// Not correctly initialized yet
  {
     // StartUp is a virtual function used to specialize
     // the two derived MOCs in the initialization phase.
     if (StartUp() == NOTOK) 			// Not properly initialized
     {
	fprintf(stderr, "monitoringSupport::update initialisation of object failed \n");
	return;					// Failed to initialize...
     } // endif  StartUp
     status = 1;
  }	// endif status
  else if (GetAttributes() == NOTOK)
  {
     void* oldStateVal = MSscanner::operationalState()->copy();
     MSscanner::operationalState() -> set(os_disabled);
     triggerStateChange(I_operationalState, MSscanner::_level, oldStateVal);

     advise (NULL, "Error: Changed observed Attributes in monitoringSupport::update()\n");

     return;
  }

  return;
 } // if unlocked (scan)
}
