//
// NAME
//   probabilityEstimator.inc.cc
//
// DESCRIPTION
//   for probabilityEstimator MOC, via OSIMIS GDMO compiler
//
// IMPLEMENTORS
//   George Mykoniatis
//   NTUA
//   mykoniat@ektor.ntua.gr
//
// HISTORY
//   A first version was produced by
//   Kostas Kasapakis (Alpha), Panos Georgatsos (Alpha),
//   George Mykoniatis (NTUA), Jorge Sanchez (NTUA)
//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "sumstat.h"
#include "buffered.h"

#include "observed_attr.h"    
#include "AdministrativeState.h" /* for MSscanner::operationalState */
#include "OperationalState.h"    /* for MSscanner::operationalState */

static OID OID_ObservedObjectInstance = NULL;
static OID OID_ObservedAttributeId    = NULL;
static OID OID_ResourceRequestRate    = NULL;
        //
        // These are initialized (just once) in monitorMetric::createRR below.
        //
        // We can *NOT* use: name2oid("~~~") here, since it depends
        // on an OID hashtable being initialized first to work properly.
        // Hashtable is not initialized before above OID handles are.
        //

// --------------------------------------------------------------
// deallocates memory 
// --------------------------------------------------------------
int probabilityEstimator::RemoveOAs()
{
  if (oa)
  {
        delete oa;
        oa = NULL;
  }
  return OK;
}
	
// --------------------------------------------------------------
// Stops the probabilityEstimator instance
// --------------------------------------------------------------
int probabilityEstimator::ShutDown()
{
  if (wkup_status)
  {
	  wkup_status = 0;
	  cancelWakeUps(NULLCP);    // Disable wake ups
  }

  status = 0;

  RemoveOAs();				// Release Memory 

  if (buff_p) 
  {
	delete buff_p;
	buff_p = NULL;
  }
  if (stat_p) 
  {
	delete stat_p;
	stat_p = NULL;
  }

#ifdef DEBUG
   advise(NULL,"probabilityEstimator::ShutDown : just executed");
#endif
  return OK;
}

// --------------------------------------------------------------
int probabilityEstimator::StartUp()
{
  int window, period, algorithm;
  double threshold ;
  // create the observed_attr instances
  if ( InitObservedAttr() == OK )
  {

   int iPollInterval           = 0;
   type_Monitor_TimePeriod* tp =
     (type_Monitor_TimePeriod*)
     MSscanner::granularityPeriod()->get();

   if ((iPollInterval = TimePeriod2seconds(tp)) == 0) {
     iPollInterval = 60; /* default value */
   }

   // Buffering stuff...
   int iFreqPollInterval           = 0;
   type_Monitor_TimePeriod* tp1 = (type_Monitor_TimePeriod*)
     reportingPeriod()->get();

   if ((iFreqPollInterval = TimePeriod2seconds(tp1)) == 0) {
     iFreqPollInterval = 60; /* default value */
   }

   threshold = * (double *) (thresholdValue()->get());

   buff_p = new buffered;
   if (buff_p == NULL)
   {
         fprintf( stderr, "StartUp: Failed to allocate memory \n");
         return NOTOK;
   }
   // Set initial value to 0 (?)
   if (buff_p->initialize(iFreqPollInterval, iPollInterval, 
					threshold, 0) == NOTOK)
   {
	fprintf( stderr, "StartUp: Failed to initialize \n");
	return NOTOK;
   }

   // Statistical stuff...
   algorithm= * (int *) (algorithmType()->get());
   window = * (int *) (collectionPeriod()->get());

   stat_p = new sumstats;
   if (stat_p == NULL)
   {
         fprintf( stderr, "StartUp: Failed to allocate memory \n");
         return NOTOK;
   }
   // Set initial value to 0 (?)
   if (stat_p->initialize(window, iPollInterval, algorithm, 0) == NOTOK)
   {
        fprintf( stderr, "StartUp: Failed to initialize \n");
        return NOTOK;
   }

   // Since everything went ok
   return OK;
 }
 // Failed to initialise the attributes ()...
 return NOTOK;
} //probabilityEstimator::StartUp

// --------------------------------------------------
int probabilityEstimator::InitObservedAttr()
{

  // Allocate space for the oa
  oa = new observed_attr;
  if (oa == NULL)
  {
       cout << "probabilityEstimator : Failed to allocate memory for oa\n";
       return NOTOK;
  }

  MO  *pMO = NULL;

  MN mnObservedObjectInstance = (MN) observedObjectInstance()->get();
  if (!mnObservedObjectInstance ||
         (dnMO = mn2localdn(mnObservedObjectInstance)) == NULLDN) 
  {
    if (!mnObservedObjectInstance)
    cout << "NULL or wrong observedObjectInstance in " 
	 << "probabilityEstimator::InitObservedAttr()" << endl;
    return NOTOK;
  }

  _typeMO = mnObservedObjectInstance->mn_type;
        //
        // Distinguished Name (local or global) of watched MO in this agent
        //

  oid = observedAttributeId()->getOID();
        //
        // OID of attribute inside MO to be watched
        //

  if (!oid)
  {
    cout <<  "NULL observedAttributeId in probabilityEstimator::InitObservedAttr()\n";
    return NOTOK;
  }

  MO* top = _typeMO == MN_DN ?
                        MO::getRoot() : MO::getRoot() -> getSubordinate();

  pMO = top -> getMO(dnMO, True, DISCARDAVA);

  char* inst;

  if (!pMO) 
  {
    inst = dn2str(dnMO);
    cout  
      << "Error: NULLMO for DN=" << inst 
      << " in probabilityEstimator::InitObservedAttr()\n";
    free(inst);
    return NOTOK;
  } 

   if (oa->initialise(pMO, oid) == NOTOK)
         return NOTOK;   // Initialise

#ifdef DEBUG
     if (oa->get_flag() != NOTOK)
        oa->print_oa();
#endif
   // Now, all the class attributes are initialized...
   // so, return ok
   return OK;
} //probabilityEstimator::InitObservedAttr

// --------------------------------------------------------------
int probabilityEstimator::CheckObservedMO()
{

   MO *pMO = NULL;

   //Find the MO* from the DN from dnlist
   MO* top = _typeMO == MN_DN ?
           MO::getRoot() : MO::getRoot() -> getSubordinate();

   pMO = top -> getMO(dnMO, True, DISCARDAVA);
   if (!pMO)
   {
           cout << "Observed MO not found" << endl;
           return NOTOK;
   }

   return OK;

} //probabilityEstimator::CheckObservedMO

// --------------------------------------------------------------
// Used by KS::wakeup method. 
// --------------------------------------------------------------
void probabilityEstimator::update()
{
 AdminStateVal admState = *(AdminStateVal*)
    MSscanner::administrativeState()->get();
 OperStateVal operState = *(OperStateVal*)
    MSscanner::operationalState()->get();

 if (admState == as_unlocked && operState == os_enabled) 
 {
  if (status == 0)  				// Not correctly initialized yet
  {
     // StartUp is a virtual function used to specialize
     // the two derived MOCs in the initialization phase.
     if (StartUp() == NOTOK) 			// Not properly initialized
     {
	fprintf(stderr, 
	   "probabilityEstimator::update initialisation of object failed \n");
	return;					// Failed to initialize...
     } // endif  StartUp
     status = 1;
  }	// endif status
  else if (CheckObservedMO() == NOTOK)
  {
     void* oldStateVal = MSscanner::operationalState()->copy();
     MSscanner::operationalState() -> set(os_disabled);
     triggerStateChange(I_operationalState, MSscanner::_level, oldStateVal);
     return;
  }

#ifdef DEBUG
  char *str = (char *)((String *)MSscanner::_attrs[I_scannerId]->get());
  fprintf(stdout,"\nprobabilityEstimator::update - Observing the object : %s\n", str);
#endif

  // Update the value of the observed attribute 
  // and store it in the oaval 
  oaval = oa->getval_attr();

  // Now find the probability and store it in the derived result attribute
  buff_p->append_value(oaval);
  buff_p->print_buffered();
  double result = buff_p->get_result();
  result = max(result,0); 
  ((Gauge *)_attrs[I_derivedResult])->setreal(result);

  // Now, apply on the result of the operation of the algorithm
  // stored in algorithmType attribute
  stat_p->append_value(oaval);
  stat_p->print_sumstats();
  double dtmp = stat_p->get_result();

  // Store the result in attribute estimateOfMean
  ((Gauge *)_attrs[I_estimateOfMean])->setreal(dtmp); 

    if (derivedResult()->triggeredThreshold() == On) {
      _qosAlarmGauge   = derivedResult();
      _qosAlarmThldOid = _classInfo ->
                                getAttrOid(I_severityIndicatingGaugeThreshold);
      triggerEvent(I_qualityofServiceAlarm, _level);
    }

    if (estimateOfMean()->triggeredThreshold() == On) {
      _qosAlarmGauge   = estimateOfMean();
      _qosAlarmThldOid = _classInfo ->
                                getAttrOid(I_estimateOfMeanSeverityIndicatingGaugeThreshold);
      triggerEvent(I_qualityofServiceAlarm, _level);
    }

#ifdef DEBUG
cout << "probabilityEstimator::update - Derived result-probability : "
     << result << endl;
cout << "probabilityEstimator::update - Derived result from sumstat   : "
     << dtmp << endl;
#endif

  return;
 } // if unlocked (scan)
} //probabilityEstimator::update

int probabilityEstimator::buildReport (int eventId, int classLevel,
                                void*& info, Bool& freeFlag)
{
    if (classLevel != _level) {
        return MSscanner::buildReport(eventId, classLevel, info, freeFlag);
        // for correct OO style...
    }

    freeFlag = False;   // we take control for efficiency

    switch (eventId) {

    case I_qualityofServiceAlarm:
      if((_thldAlarmInfo == NULL) &&
         (_thldAlarmInfo=this->_thldAlarmInfo_alloc())==NULL
      ) {
        return NOTOK;   // not alloc'd and can't alloc
      }

      _thldAlarmInfo->thresholdInfo->triggeredThreshold->mid_global =
                        _qosAlarmThldOid;       // NOT to be free'd

      _thldAlarmInfo->thresholdInfo->observedValue->offset = OV_REAL;
      _thldAlarmInfo->thresholdInfo->observedValue->un.real =
                        ((GaugeReal*) _qosAlarmGauge) -> getreal();

      info = _thldAlarmInfo;
    break;

    default:
      return NOTOK;

    } /* end switch */

    return OK;

} //probabilityEstimator::buildReport


//
// used to initialize _thldAlarmInfo pointer during createRR()
// freed by deleteRR()
//
type_SMI_AlarmInfo*
probabilityEstimator::_thldAlarmInfo_alloc ()
{
    struct type_SMI_AlarmInfo* elem;
    if ((elem = ThresholdAlarmInfo_ProbsAttrs_alloc (NULL, 1, 2))== NULL) {
       return NULL;
    }

    /* specificProblems */
    ((MIDListVal*) elem->specificProblems)->midl_mid.mid_global =
	::OID_ResourceRequestRate;
    /* monitoredAttrs */

    /* insert observedObjectInstance OID/value */
    ((CMISParamListVal*) elem->monitoredAttrs)->mpl_id.mid_global =
       ::OID_ObservedObjectInstance;
    ((CMISParamListVal*) elem->monitoredAttrs)->mpl_val =
        observedObjectInstance()->encode();

    /* insert observedAttributeId OID/value */
    ((CMISParamListVal*) elem->monitoredAttrs)->mpl_next->mpl_id.mid_global =
       ::OID_ObservedAttributeId;
    ((CMISParamListVal*) elem->monitoredAttrs)->mpl_next->mpl_val =
        observedAttributeId()->encode();

    return elem;
}

// --------------------------------------------------------------
// Initializes all the attributes of the new probabilityEstimator 
// object instance to initial values
// --------------------------------------------------------------
int probabilityEstimator::createRR (AVA*& err, void* pv, int)
{

  // correct OO style
  if (MSscanner::createRR(err, pv) == NOTOK)
    return NOTOK;

  static Bool bClassInitialised = False;

  // Initialise the variables for this instance
  _thldAlarmInfo                = NULL;

  status = 0;              // Set the status to inactive
  buff_p = NULL;           // buffered instance pointer,
                           // initialised to NULL
  stat_p = NULL;           // sumstats instance pointer,
                           // initialised to NULL

  oa = NULL;         	   // Initiailise NULL

#ifdef DEBUG
   fprintf(stdout, "probabilityEstimator::createRR - Creating an MO with MO id : %s \n",
                   (char *)MSscanner::_attrs[I_scannerId]->get());
#endif

//Set types...

   derivedResult()->setType(OV_REAL);

   estimateOfMean()->setType(OV_REAL);

   severityIndicatingGaugeThreshold() ->setType(OV_REAL);

   estimateOfMeanSeverityIndicatingGaugeThreshold()->setType(OV_REAL);

   severityIndicatingTideMarkMax()->setType(TM_MAX);

   severityIndicatingTideMarkMin()->setType(TM_MIN);

//Associate and set initial values....

   severityIndicatingGaugeThreshold()
	->add(MSDERIVEDGAUGETHLDLOW, Off, MSDERIVEDGAUGETHLDHIGH, Off);

   derivedResult() ->associateThreshold(severityIndicatingGaugeThreshold());

   derivedResult()
	->associateTideMark(severityIndicatingTideMarkMax());

   derivedResult()
	->associateTideMark(severityIndicatingTideMarkMin());

   estimateOfMean()->setreal(0.0);
   derivedResult()->setreal(0.0);

//Set types...

   estimateOfMeanSeverityIndicatingTideMarkMax()->setType(TM_MAX);

   estimateOfMeanSeverityIndicatingTideMarkMin()->setType(TM_MIN);


//Associate and set initial values....

   estimateOfMeanSeverityIndicatingGaugeThreshold()
	->add(MSDERIVEDGAUGETHLDLOW, Off, MSDERIVEDGAUGETHLDHIGH, Off);

  estimateOfMean()->associateThreshold(estimateOfMeanSeverityIndicatingGaugeThreshold());

  estimateOfMean()->associateTideMark(estimateOfMeanSeverityIndicatingTideMarkMax());

  estimateOfMean()->associateTideMark(estimateOfMeanSeverityIndicatingTideMarkMin());


   if (!bClassInitialised)
   { 
	_info.setDefault ( 
			   I_severityIndicatingGaugeThreshold,
	    		   new GaugeThresholdReal ( 
				 MSDERIVEDGAUGETHLDLOW, 
				 Off, 
				 MSDERIVEDGAUGETHLDHIGH, 
				 Off)
			 );
	_info.setDefault ( 
			   I_estimateOfMeanSeverityIndicatingGaugeThreshold,
	    		   new GaugeThresholdReal ( 
				 MSDERIVEDGAUGETHLDLOW, 
				 Off, 
				 MSDERIVEDGAUGETHLDHIGH, 
				 Off)
			 );
   }				   

  if (!bClassInitialised) {

    ::OID_ObservedObjectInstance = name2oid("observedObjectInstance");
    ::OID_ObservedAttributeId = name2oid("observedAttributeId");
    ::OID_ResourceRequestRate = name2oid("resourceRequestRate");

    bClassInitialised = True;
  }


#ifdef DEBUG
// Print a message that initialisation has finished
   fprintf( stdout, "Finished createRR\n");
#endif
   return OK;
} //probabilityEstimator::createRR

int probabilityEstimator::deleteRR (AVA*& err, void* info, Bool checkOnly, int)
{

#ifdef DEBUG
  fprintf(stdout, "probabilityEstimator::deleteRR-Executing.....\n");
#endif

  ShutDown();

  return MSscanner::deleteRR(err, info, checkOnly);
} //probabilityEstimator::deleteRR

// --------------------------------------------------------------
int probabilityEstimator::set (CMISModifyOp mode, int attrId, int classLevel,
                      void* value, AVA*& err, Bool checkOnly, int)
{
    
  if (classLevel != _level) 
      return MSscanner::set(mode, attrId, classLevel, value, err, checkOnly);

  int pval;             // Passed value (if integer ...)
    int iPollInterval = 0;

#ifdef DEBUG
  fprintf(stdout, "probabilityEstimator::set-Executing.....\n");
#endif

    AdminStateVal admState =  *(AdminStateVal*)
	MSscanner::administrativeState()->getval();

  switch (attrId)
  {
    case I_algorithmType:
      if (admState == as_locked) {
         if ( (*(int *)value) != 0 && (*(int *)value) != 1 && (*(int *)value) != 2
                && (*(int *)value) != 3 )
          {
           advise (NULL,
              "probabilityEstimator::set(): invalid algorithmType value, set failed \n");
           
           if (checkOnly)            
              return NOTOK;

           err = new AVA(m_invalidAttributeValue, "invalidValue");
           return NOTOK;
          }

         if (checkOnly)
           return OK;

         }
      else {
        if (checkOnly)
          return NOTOK;

         err = new AVA(m_processingFailure, "adminStateUnlocked");
         return NOTOK;
         }
      break;

    case I_observedAttributeId:
    case I_observedObjectInstance:
    case I_thresholdValue:
    case I_reportingPeriod:
    case I_severityIndicatingGaugeThreshold:
    case I_estimateOfMeanSeverityIndicatingGaugeThreshold:
    case I_severityIndicatingTideMarkMax:
    case I_severityIndicatingTideMarkMin:
    case I_estimateOfMeanSeverityIndicatingTideMarkMax:
    case I_estimateOfMeanSeverityIndicatingTideMarkMin:
      if (admState == as_locked) {
        // The attributes :
        //
        // reportingPeriod
        // thresholdValue
        // observedAttributeId
        // observedObjectInstance
	// severityIndicatingGaugeThreshold
	// estimateOfMeanSeverityIndicatingGaugeThreshold
	// severityIndicatingTideMarkMax
	// severityIndicatingTideMarkMin
	// estimateOfMeanSeverityIndicatingTideMarkMax
	// estimateOfMeanSeverityIndicatingTideMarkMin
        //
        // in the probabilityEstimator managed object
        // class shall not be modified unless the value of the
        // administrative state is "locked"

        if (checkOnly)
          return OK;

        advise (NULL,
            "probabilityEstimator::set() new collectionPeriod=%d \n",*(int *) value);
        }
      else {
        if (checkOnly)
          return NOTOK;

         err = new AVA(m_processingFailure, "adminStateUnlocked");
         return NOTOK;
      }
      break;


    case I_collectionPeriod:
      if (admState == as_locked) {
	// collectionPeriod 
        // shall not be modified unless the value of the
        // administrative state is "locked"

        if (checkOnly)
          return OK;

        advise (NULL,
            "probabilityEstimator::set() new collectionPeriod=%d \n",*(int *) value);
        }
      else {
        if (checkOnly)
          return NOTOK;

         err = new AVA(m_processingFailure, "adminStateUnlocked");
         return NOTOK;
      }
      break;

    default:
      break;

  } // Switch attrid

  return OK;
} //probabilityEstimator::set

