//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Destination.cc

// By George Pavlou, May 1995


#include <stdlib.h>
#include <string.h>
#include "msap.h"
#include "Error.h"
#include "Destination.h"

extern int globalNameDomainRdns;	// in msap/syntax.c


AetState*  aetStateDNSetUp (DN dn)
{
    AetState* aetState = new AetState;
    aetState->aet.aet = dn;		// NOT a copy
    char *suffix /* last 2 RDNs */, *cp, *cp2, *cp3;
    register DN cur; register int rdnCnt;
    DN oneBfreLast, twoBfreLast;
    int cse = 0;

    suffix = cp = cp2 = cp3 = NULLCP;
    cur = oneBfreLast = twoBfreLast = NULLDN;


    for (cur = dn, rdnCnt = 0; cur ; cur = cur->dn_parent, rdnCnt++) {
	twoBfreLast = oneBfreLast;
        oneBfreLast = cur;
    }


    if (rdnCnt < globalNameDomainRdns ||
	oneBfreLast == NULLDN || twoBfreLast == NULLDN ||
    	    	(suffix = dn2str( twoBfreLast )) == NULLCP)
	goto failed;


    // we cater here for three types of DN suffixes:
    // 1. cn=SMA@cn=athena
    // 2. X=Y@cn=SMA:athena	   or X=Y@cn=SMA-athena
    // 3. X=Y@cn=SMA:athena@athena or X=Y@cn=SMA-athena@athena


    if ( ((cp = strrchr(suffix, ':')) != NULLCP ||
    	     (cp = strrchr(suffix, '-')) != NULLCP ) &&
    	 (cp2 = strchr(suffix, '@')) != NULLCP && cp > cp2 &&
	    strncasecmp( cp2 + 1, "cn=", 3 ) == 0 ) {

	cse = ((cp3 = strchr( cp, '@' )) == NULLCP ) ? 2 : 3;

    } else {

	if( (cp = strchr(suffix, '@')) != NULLCP &&
	    strncasecmp( cp + 1, "cn=", 3 ) == 0 &&
	    strncasecmp( suffix, "cn=", 3 ) == 0   ) {
    	        cse = 1;
    	}
    }

    if (cp==NULLCP || cse == 0)
    	goto failed;


    switch( cse ) {
    	case 1 : *cp = '\0'; // extract cn=SMA
		 aetState->aet.title = strdup( suffix + 3 ); // SMA
		 aetState->aet.qualifier = strdup( cp + 4 ); // athena
	break ;

	case 2 :
	case 3 : *cp = '\0'; // extract cn=SMA
		 aetState->aet.title = strdup( cp2 + 4 );
		 if (cp3 != NULLCP) *cp3 = '\0'; // lose the @athena
                 aetState->aet.qualifier = strdup( cp + 1 );
	break ;
    }

    aetState->state = True;
    free(suffix);
    return aetState;

failed:

#ifdef DEBUG
    if( suffix != NULLCP )
	advise( NULLCP, "Destination.c FAILURE for suffix %s\n", suffix);
#endif
    aetState->aet.title = aetState->aet.qualifier = NULLCP;
    aetState->state = False;
    if (suffix)
	(void)free(suffix);
    return aetState;
}


void  DestinationAddress::aetStateSetUp ()
{
    type_SMI_Destination* dest =
		(struct type_SMI_Destination*) getval();
    type_SMI_DistinguishedNameList first;
    register type_SMI_DistinguishedNameList *cur;

    // set up aet so that we treat single and multiple in the same loop
    if (dest->offset == type_SMI_Destination_singleAET) {
        first.DistinguishedName = dest->un.singleAET;
        first.next = NULL;
        cur = &first;
    }
    else
        cur = dest->un.multipleAET;

    _aetStateArray = new AetStateArray(5);

    for (register int i = 0; cur; i++, cur = cur->next) {
	AetState* aetState = aetStateDNSetUp(cur->DistinguishedName);
	_aetStateArray->add(aetState);
    }
}


void  DestinationAddress::aetStateReset ()
{
    if (_aetStateArray) {
	_aetStateArray->clear();
	delete _aetStateArray;
	aetStateSetUp();
    }
}


int  DestinationAddress::set (void* newVal)
{
    Attr::set(newVal);
    aetStateReset();
    if (_msd > 0)
	_msd = -1;
    _wasSet = True;
    return OK;
}


DestinationAddress::DestinationAddress (type_SMI_Destination* myAddr)
{
    setval(myAddr);
    _msd = -1;
    _wasSet = False;
    aetStateSetUp();
}


DestinationAddress::DestinationAddress ()
{
    struct type_SMI_Destination* dest = new type_SMI_Destination;
    dest->offset = type_SMI_Destination_singleAET;
    dest->un.singleAET = dn_comp_new(NULLRDN); // Empty DN {}
    setval(dest);
    _msd = -1;
    _wasSet = False;
    aetStateSetUp();
}


DestinationAddress::~DestinationAddress ()
{
    _aetStateArray->clear();
    delete _aetStateArray;
}

