//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Gauge.cc

// By George Pavlou, February 1991


#include <stdlib.h>
#include "msap.h"
#include "Error.h"
#include "Gauge.h"


// Gauge methods

Gauge::Gauge (int type, double val)
{
    ObservedValue* v = new ObservedValue;
    if ((v -> ov_type = type) == OV_INTEGER)
	v -> ov_integer = (int) val;
    else			// OV_REAL
	v -> ov_real = val;
    setval(v);

    threshold = NULL;
    tideMarkMax = tideMarkMin = NULL;
    thldTrigger = tmrkMaxTrigger = tmrkMinTrigger = Off;
}


Gauge::Gauge ()
{
    ObservedValue* v = new ObservedValue;
    v -> ov_type = OV_INTEGER;	// default
    v -> ov_integer = 0;
    setval(v);

    threshold = NULL;
    tideMarkMax = tideMarkMin = NULL;
    thldTrigger = tmrkMaxTrigger = tmrkMinTrigger = Off;
}


int  Gauge::setType (int type)
{
    ObservedValue* v = (ObservedValue*) getval();

    switch (type) {
    case OV_INTEGER:
    case OV_REAL:
	v -> ov_type = type;
	return OK;

    default:
	advise(NULLCP, "Gauge: setType: invalid type!");
	return NOTOK;
    }
}


Switch  Gauge::increment (double incr)
{
    ObservedValue* val = (ObservedValue*) getval();

    if (val->ov_type == OV_INTEGER)
	return setreal(val->ov_integer + incr);
    else		// OV_REAL
	return setreal(val->ov_real + incr);
}


Switch  Gauge::decrement (double decr)
{
    ObservedValue* val = (ObservedValue*) getval();

    if (val->ov_type == OV_INTEGER)
	return setreal(val->ov_integer - decr);
    else		// OV_REAL
	return setreal(val->ov_real - decr);
}


double  Gauge::getreal ()
{
    ObservedValue* val = (ObservedValue*) getval();

    if (val->ov_type == OV_INTEGER)
	return (double) val->ov_integer;
    else		// OV_REAL
	return val->ov_real;
}


Switch  Gauge::setreal (double newVal)
{
    double prevVal;
    ObservedValue* val = (ObservedValue*) getval();

    if (val->ov_type == OV_INTEGER) {
	prevVal = (double) val->ov_integer;
	val->ov_integer = (int) newVal;
    }
    else {		// OV_REAL
	prevVal = val->ov_real;
	val->ov_real = newVal;
    }
    replval(val);	// we use the same allocated memory

    if (threshold)
	thldTrigger = threshold -> check(newVal, prevVal);
    if (tideMarkMax)
	tmrkMaxTrigger = tideMarkMax -> check(newVal);
    if (tideMarkMin)
	tmrkMinTrigger = tideMarkMin -> check(newVal);

    return thldTrigger;	// for compatibility with previous versions
}

