//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Gauge.h

// By George Pavlou, February 1991


#ifndef _Gauge_h_
#define _Gauge_h_


#ifndef _GENERIC_ATTR_
#include "GenericAttr.h"
#endif

#ifndef _SMI_SNTX_
#include "SmiSntx.h"
#endif


class Gauge;


class GThreshold : public Attr
{
protected:
    inline	GThreshold ()		{ setval(NULL); }

public:
    inline virtual void	  associateGauge (Gauge*) { }
    inline virtual Switch check (double, double) { return Off; }
};


class TMark : public Attr
{
protected:
    inline	TMark ()		{ setval(NULL); }

public:
    inline virtual void	  associateGauge (Gauge*) { }
    inline virtual Switch check (double) { return Off; }
};


class Gauge : public Attr
{
    GThreshold*	threshold;	// threshold applied to the gauge
    TMark*	tideMarkMax;	// max tide-mark applied to the gauge
    TMark*	tideMarkMin;	// max tide-mark applied to the gauge
    Switch	thldTrigger;
    Switch	tmrkMaxTrigger;
    Switch	tmrkMinTrigger;

protected:
    inline PE 	 _encode ()		{ return ObservedValue_encode(
					  (ObservedValue*) getval()); }
    inline void* _decode (PE pe)	{ return ObservedValue_decode(pe); }
    inline void	 _free ()		{ free((char*) getval()); }
    inline char* _print ()		{ return attrv2str(getval(),
					  (PRINT_FNX) ObservedValue_print); }
    inline int	 _compare (void* v1, void* v2)
					{ return ObservedValue_cmp(
					(ObservedValue*)v1,(ObservedValue*)v2);}

    inline GThreshold* getThreshold ()	{ return threshold; }
    inline TMark* getTideMarkMax ()	{ return tideMarkMax; }
    inline TMark* getTideMarkMin ()	{ return tideMarkMin; }
    inline TMark* getTideMark ()	{ return getTideMarkMax(); }

    Gauge (int, double);

public:
    inline int	getType ()		{ return
					  ((ObservedValue*)getval())->ov_type; }
    int		setType (int);
    inline void	associateThreshold (GThreshold* thld)
					{ (threshold = thld) ->
						associateGauge(this); }
    inline void	associateTideMarkMax (TMark* tm)
					{ (tideMarkMax = tm) ->
						associateGauge(this); }
    inline void	associateTideMarkMin (TMark* tm)
					{ (tideMarkMin = tm) ->
						associateGauge(this); }
    inline void	associateTideMark (TMark* tm)
					{ associateTideMarkMax(tm); }

    inline Switch triggeredThreshold ()	{ return thldTrigger; }
    inline Switch triggeredTideMarkMax () { return tmrkMaxTrigger; }
    inline Switch triggeredTideMarkMin () { return tmrkMinTrigger; }
    inline Switch triggeredTideMark ()	{ return triggeredTideMarkMax(); }

    inline char*  getSyntax ()		{ return "ObservedValue"; }

    double	getreal ();
    Switch	setreal (double);

    // not set(double) anymore as ATT C++ does not like it -
    // please use setreal(double) instead

    Switch increment (double);
    Switch decrement (double);

    Gauge ();
};


class GaugeInt : public Gauge
{
public:
    inline Switch increment (int incr)	{ return
					  Gauge::increment((double) incr); }
    inline Switch decrement (int decr)	{ return
					  Gauge::decrement((double) decr); }
    inline Switch increment ()		{ return increment(1); }
    inline Switch decrement ()		{ return decrement(1); }
    inline int	  getint ()		{ return ((ObservedValue*) getval())
					  ->ov_integer; }
    inline	GaugeInt ()		: Gauge(OV_INTEGER, 0) {}
};


class GaugeReal : public Gauge
{
public:
    inline double getreal ()		{ return ((ObservedValue*) getval())
					  ->ov_real; }
    inline	GaugeReal ()		: Gauge(OV_REAL, 0) {}
};


#endif
