//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// GaugeThld.cc

// By George Pavlou, February 1991


#include <stdlib.h>
#include "msap.h"
#include "Error.h"
#include "GaugeThld.h"


void  GaugeThreshold::associateGauge (Gauge* g)
{
    _type = g -> getType();	// cannot be inline because of
}				// forward referencing


int  GaugeThreshold::setType (int type)
{
    switch (type) {
    case OV_INTEGER:
    case OV_REAL:
	_type = type;
	return OK;

    default:
	advise(NULLCP, "GaugeThreshold: setType: invalid type!");
	return NOTOK;
    }
}


int  GaugeThreshold::set (void* newThld)
{
    if (!newThld) {
	setval(NULLVD);
	return OK;
    }

    // check if all the new threshold elements have valid observed values
    // and low < high

    register GaugeThresholdVal* cur = (GaugeThresholdVal*) newThld;

    for (; cur; cur = cur -> gt_next)
	if (cur -> gt_notifyLow.nt_threshold.ov_type != _type ||
		cur -> gt_notifyHigh.nt_threshold.ov_type != _type ||
		(_type == OV_INTEGER &&
		cur -> gt_notifyLow.nt_threshold.ov_integer >=
		cur -> gt_notifyHigh.nt_threshold.ov_integer) ||
		(_type == OV_REAL &&
		cur -> gt_notifyLow.nt_threshold.ov_real >=
		cur -> gt_notifyHigh.nt_threshold.ov_real))
	    return NOTOK;
	    
    setval(newThld);
    return OK;
}


int  GaugeThreshold::add
		(double low, Switch onofflow, double high, Switch onoffhigh)
{
    GaugeThresholdVal* elem = new GaugeThresholdVal;

    elem -> gt_notifyLow.nt_threshold.ov_type	= _type;
    elem -> gt_notifyLow.nt_notifyOnOff		= onofflow;
    elem -> gt_notifyLow.nt_trigger		= Off;

    elem -> gt_notifyHigh.nt_threshold.ov_type	= _type;
    elem -> gt_notifyHigh.nt_notifyOnOff	= onoffhigh;
    elem -> gt_notifyHigh.nt_trigger		= Off;
    elem -> gt_next				= NULL;

    if (_type == OV_INTEGER) {
	elem -> gt_notifyLow.nt_threshold.ov_integer  = (int) low;
	elem -> gt_notifyHigh.nt_threshold.ov_integer = (int) high;
    }
    else {	// OV_REAL
	elem -> gt_notifyLow.nt_threshold.ov_real  = low;
	elem -> gt_notifyHigh.nt_threshold.ov_real = high;
    }

    return Attr::add(elem);
}


Switch  GaugeThreshold::check (double val, double prev)
{
    GaugeThresholdVal* gthld = (GaugeThresholdVal*) getval();
    double gval;
    Switch trigger = Off;

    advise(NULLCP, "\nGaugeThreshold::check");

    for (GaugeThresholdVal* cur = gthld; cur; cur = cur -> gt_next)
	if (val > prev) {
	    gval = _type == OV_INTEGER ?
			(double) cur -> gt_notifyHigh.nt_threshold.ov_integer :
			cur -> gt_notifyHigh.nt_threshold.ov_real;
	    if (val >= gval && prev < gval) {
	        if (cur -> gt_notifyLow.nt_trigger == On)
		    cur -> gt_notifyLow.nt_trigger = Off;
					// trigger subsequent 'low' crossings
	        if (cur -> gt_notifyHigh.nt_trigger == Off) {
		    cur -> gt_notifyHigh.nt_trigger = On;
		    if (cur -> gt_notifyHigh.nt_notifyOnOff == On)
		        trigger = On;		// triggered high
	        }
		/*
		advise(NULLCP, "newVal %.1f \
prevVal %.1f, lowTrigger Off, highTrigger On, TRIGGER %s\n",
			val, prev, trigger == On ? "On" : "Off");
		*/
	    }
	}
	else
	if (val < prev) {
	    gval = _type == OV_INTEGER ?
			(double) cur -> gt_notifyLow.nt_threshold.ov_integer :
			cur -> gt_notifyLow.nt_threshold.ov_real;
	    if (val <= gval && prev > gval) {
	        if (cur -> gt_notifyHigh.nt_trigger == On)
		    cur -> gt_notifyHigh.nt_trigger = Off;
					// trigger subsequent 'high' crossings
	        if (cur -> gt_notifyLow.nt_trigger == Off) {
		    cur -> gt_notifyLow.nt_trigger = On;
		    if (cur -> gt_notifyLow.nt_notifyOnOff == On)
		        trigger = On;		// triggered low
	        }
		/*
		advise(NULLCP, "newVal %.1f \
prevVal %.1f, highTrigger Off, LowTrigger On, TRIGGER %s\n",
			val, prev, trigger == On ? "On" : "Off");
		*/
	    }
	}

    return trigger;
}

