/*
 * Copyright (c) 1990 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* SmiAsn.h - standard SMI type ideclarations in C
 *            for parsing through the ASN.1 compiler
 */

/* By George Pavlou / Saleem Bhatti, February 1991 */


#ifndef _SMI_ASN_
#define _SMI_ASN_


#ifndef _MSAP_
#include "msap.h"
#endif

#ifndef _BOOLEAN_
#include "Boolean.h"
#endif


/* SimpleNameType */

typedef struct SimpleNameTypeVal
{
    int         sn_type;
#define	SN_NUMBER	1	/* same as type_SMI_SimpleNameType_number */
#define	SN_STRING	2	/* same as type_SMI_SimpleNameType_string */

    union {
        int	sn_un_number;
        char*	sn_un_string;
    } sn_un;

#define sn_number	sn_un.sn_un_number
#define sn_string	sn_un.sn_un_string

} SimpleNameTypeVal;


/* String List */

typedef struct StringListVal
{
    char*			string;
    struct StringListVal*	next;
} StringListVal;


/* Integer List */

typedef struct IntegerListVal
{
    int				integer;
    struct IntegerListVal*	next;
} IntegerListVal;


/* Real List */

typedef struct RealListVal
{
    double			real;
    struct RealListVal*		next;
} RealListVal;


/* Object Identifier List */

typedef struct ObjIdListVal
{
    OID				oidl_oid;
    struct ObjIdListVal*	oidl_next;

} ObjIdListVal;


/* Management Name List (Object Instances) */

typedef struct MNListVal {
    MName		mnl_mn;
    struct MNListVal*	mnl_next;

} MNListVal, Member, ObjectInstanceListVal;


/* Counter Threshold */

typedef struct CounterThresholdVal
{
    int		ct_compLevel;
    int		ct_offsetVal;
    Switch	ct_notifyOnOff;

    struct	CounterThresholdVal* ct_next;

} CounterThresholdVal;


/* Gauge Threshold */

/*
 * this is exactly the same in layout/tags as the type_SMI_ObservedValue
 * produced automatically by pepsy in SMI-types.h
 */
typedef struct ObservedValue
{
    int	ov_type;
#define	OV_INTEGER	1	/* same as type_SMI_ObservedValue_integer */
#define	OV_REAL		2	/* same as type_SMI_ObservedValue_real */

    union {
	int	ov_un_integer;
	double	ov_un_real;
    } ov_un;

#define ov_integer	ov_un.ov_un_integer
#define ov_real		ov_un.ov_un_real

} ObservedValue;


typedef struct NotifyThreshold
{
    ObservedValue	nt_threshold;
    Switch		nt_notifyOnOff;
    Switch		nt_trigger;		/* for MO use only */

} NotifyThreshold;


typedef struct GaugeThresholdVal
{
    NotifyThreshold     gt_notifyLow;
    NotifyThreshold	gt_notifyHigh;

    struct  GaugeThresholdVal* gt_next;

} GaugeThresholdVal;


/* Tide Mark */

typedef struct TideMarkVal
{
    int			tm_type;
#define	TM_MAX		1		/* not 0 and 1 as in the past */
#define	TM_MIN		2

    ObservedValue	tm_curValue;
    ObservedValue	tm_prevValue;
    char*		tm_resetTime;	/* no UTCtime as in the past */

} TideMarkVal;


/* Administrative State */

typedef enum
{
    as_locked,
    as_unlocked,
    as_shuttingDown

} AdminStateVal;


/* Operational State */

typedef enum
{
    os_disabled,
    os_enabled

} OperStateVal;


/* Destination Address */

typedef struct DestAddressVal
{
    int	da_type;
#define DA_PSAP	1		/* PSAP address */
#define DA_AEI	2		/* Application Entity Information */
#define DA_MSD	3		/* Management Association Descriptor */

    union {
	struct PSAPaddr* da_un_psap;
	AEI  da_un_aei;
	int  da_un_msd;
    } da_un;
#define da_psap	da_un.da_un_psap
#define da_aei	da_un.da_un_aei
#define da_msd	da_un.da_un_msd

    struct DestAddressVal* da_next;

} DestAddressVal;


typedef enum {
    lfa_wrap,
    lfa_halt
} LogFullActionValue;
 

typedef struct {
    int nValues;
    int *values;
#define AvailabilityStatus_inTest       0
#define AvailabilityStatus_failed       1
#define AvailabilityStatus_powerOff     2
#define AvailabilityStatus_offLine      3
#define AvailabilityStatus_offDuty      4
#define AvailabilityStatus_dependency   5
#define AvailabilityStatus_degraded     6
#define AvailabilityStatus_notInstalled 7
#define AvailabilityStatus_logFull      8

} AvailabilityStatusValue;


#endif
