/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	GDMONot.h

	Class that represents a	GDMO Notification

	Author:	J.Cowan		(j.cowan@.cs.ucl.ac.uk)

*/
/****************************************************************/

#ifndef	_SYMBOL
#include "Symbol.h"
#endif

#ifndef	_GENERIC_LIST_
#include "GenList.h"
#endif

#ifndef	_LIST_ITERATOR_
#include "ListIterator.h"
#endif

#ifndef	_GENERIC_STRING_
#include "GenStr.h"
#endif

/****************************************************************/

class GDMONotification : public MetaClass
{
	StringSymbol	_name, _infoSyntax, _replySyntax;
	IntegerSymbol	_mode;
	GDMOOid		_oid;
public:
	GDMONotification(char *name) :
			MetaClass("GDMONotification"),
			_name(name, "NotificationName"),
			_mode(False, "Mode"),
			_infoSyntax("",	"InfoSyntax"),
			_replySyntax("", "ReplySyntax"),
			_oid("Oid")
	{
		AddSymbol(&_name);
		AddSymbol(&_mode);
		AddSymbol(&_infoSyntax);
		AddSymbol(&_replySyntax);
		AddSymbol(&_oid);
	}

	void	AddInfoSyntax(char *syntax)
	{
		_infoSyntax.StrReplace(syntax);
	}

	void	AddReplySyntax(char *syntax)
	{
		_replySyntax.StrReplace(syntax);
	}

	char	*GetName() { return (char *) _name;}

	GDMOOid	*GetOid() { return &_oid;}
};

/****************************************************************/

class	GDMONotificationList : public ListSymbol
{
protected:
	void	deleteElem (void*);
public:
	GDMONotificationList(Bool del, char *name)
	: ListSymbol(name, "GDMONotification", ObjectTypeDef, del) {}
	GDMONotificationList(char *name)
	: ListSymbol(name, "GDMONotification", ObjectTypeDef) {}
	~GDMONotificationList()	{ clear();}

	GDMONotification *FindNotification(char *notificationName)
	{
	GDMONotification	*notification;
	ListIterator	notificationIterator(this);

		while ((notification=(GDMONotification *)
				notificationIterator.getNext())
				!= (GDMONotification *) 0)

			if (!strcmp(notification->GetName(), notificationName))
				break;

		return notification;
	}
};

/****************************************************************/

class GDMONotificationRef
{
	GString		_name;
	GDMONotification	*_notification;
public:
	GDMONotificationRef(char *name)
	{
		_name = name;
		_notification = (GDMONotification *) 0;
	}

	void	UpdateRef(GDMONotification *notification)
	{
		_notification = notification;
	}
	char	*GetName()
	{
		return (char *)_name;
	}
	GDMONotification *GetNotification()
	{
		if (!_notification)
			printf("\nNotification %s not updated",(char *) _name);

		return _notification;
	}
};

/****************************************************************/

class	GDMONotificationRefList : public List
{
protected:
	void	deleteElem (void*);
public:
	~GDMONotificationRefList() { clear();}
};

/****************************************************************/
// inline methods for GDMONotificationList

inline void  GDMONotificationList::deleteElem (void* val)
{  if (del) delete (GDMONotification *) val; }

/****************************************************************/
// inline methods for GDMONotificationRefList

inline void  GDMONotificationRefList::deleteElem (void* val)
{  if (del) delete (GDMONotificationRef *) val; }

/****************************************************************/





