/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	SymExec.h

	Code generator Interpreter Header file

	Author:	J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

#ifndef	_SYMEXEC_

#define	_SYMEXEC_

#ifndef	_SYMBOL_
#include "Symbol.h"
#endif

#ifndef	_NODE_
#include "Node.h"
#endif

/****************************************************************/

extern	FILE	*yyin;

/****************************************************************/

/****************************************************************/

struct	ControlStackVal
{
	Node	*_node;
	FnPtr	_fnPtr;

}; /* ControlStackVal */

/****************************************************************/

struct	ValueStackVal
{
	int	_type;
	union
	{
		char	*_strValue;
		int	_intValue;
		Symbol	*_sym;
	}
	_data;

}; /* ValueStackVal */

#define	STK_CHAR	1
#define	STK_INT		2
#define	STK_SYM		3
#define	MAXSTACK	100

/****************************************************************/

class	InterPreter
{
	ValueStackVal	*_valueStack;
	int		_valStkPtr, _ctrlStkPtr;
	ControlStackVal	*_controlStack;
	SymbolTable	*_symTable;
	SymbolTable	_localSymbols;
	SymbolTable	_globalSymbols;
	FILE		*_inFile, *_outFile, *_yyinSave;
	char		*_inFileName, *_outFileName;
	Node		*_firstNode;
	int		_runError;

	int		Parse();
	Bool		Run();
	Bool		Compile();
	void		LexInitialise(char *);
	void		InitialiseOutputFile(char *);
	void		Initialise()
	{
		_valStkPtr = _ctrlStkPtr = 0;
		_controlStack =	(NEW ControlStackVal[MAXSTACK]);
		_valueStack = (NEW ValueStackVal[MAXSTACK]);
	}
	int	Lex();
	Symbol	*Lookup(char *name);
	void	FirstNode(Node *node)
	{
		_firstNode = node;
	}
	void	Output(FILE *, int);
	void	Output(FILE *, char *);
public:
	InterPreter(char *inFileName, char *outFileName, SymbolTable *symTable)
		: _symTable(symTable), _localSymbols(True), 
		  _globalSymbols(True), _inFileName(inFileName), 
		  _outFileName(outFileName)
	{
		Initialise();

		LexInitialise(inFileName);
		InitialiseOutputFile(outFileName);
	}

	InterPreter(char *inFileName, FILE *outFile, SymbolTable *symTable)
		: _symTable(symTable), _localSymbols(True), 
		  _globalSymbols(True), _inFileName(inFileName), 
		  _outFileName((char *) 0)
	{
		Initialise();
		LexInitialise(inFileName);
		_outFile = outFile;	// appending to another file
	}

	~InterPreter()
	{
		DELETE [] _controlStack;
		DELETE [] _valueStack;
		_localSymbols.Clear();
		_globalSymbols.Clear();
		yyin = _yyinSave;	// Restore lex stream
	}

	Bool		AddSymbol(Symbol *, Bool, Bool);
	Bool		Execute();
	void		Stop()	{_runError = True;}
	void		PushControl(Node *, FnPtr);
	ControlStackVal	*PopControl();

	void		PushValue(int);
	void		PushValue(char *);
	void		PushValue(Symbol *);
	ValueStackVal	*PopValue();

	void	Print(int ch)		{Output(_outFile, ch);}
	void	Print(char *text)	{Output(_outFile, text);}

	void	Display(int ch)		{Output(stdout,	ch);}
	void	Display(char *text)	{Output(stdout,	text);}

	FILE 	*GetOutputFile()	{return _outFile;}

}; /* InterPreter */

/****************************************************************/

#endif

/****************************************************************/
/* End of module */
/****************************************************************/

