#
# attrgen.awk - parses the generated attrcmpl.log file 
#               and produces the attribute source files
#
# By George Pavlou - November 1994
#


{
    syntax[NR] = $1
    if ($2 && $3) {
        struct[NR] = $2
        member[NR] = $3
        if (index(member[NR], "*")) {
	    member[NR] = substr(member[NR], 2)
	    ptr[NR] = "*"
        }
    }
    else
    if ($2)
	ptr[NR] = "*"
}


END {
    # constant definitions
    # ENC = "encode"
    # DEC = "decode"
    ENC = "build"
    DEC = "parse"

    printf "\nmodule: " syntax[1] "\n\n"

    oidtable_header_file_start(syntax[1])
    oidtable_source_file_start(syntax[1])

    for (i = 2; i <= NR; i++) {
	# syntax_info(syntax[i], struct[i], member[i], ptr[i])

	file = syntax[i]".inc.h"
	include = 0
	inch[i] = ""
        if (getline < file > 0) {
	    include = 1
	    close(file)
	    inch[i] = file
	}
	printf "producing " syntax[i] ".h\n"
	class_header_file(syntax[1], syntax[i],
				struct[i], member[i], ptr[i], include)

	file = syntax[i]".inc.c"
	include = 0
	incc[i] = ""
        if (getline < file > 0) {
	    include = 1
	    close(file)
	    incc[i] = file
	}
	oidtable_header_file(syntax[1], syntax[i],
				struct[i], member[i], ptr[i], include)
	oidtable_source_file(syntax[1], syntax[i],
				struct[i], member[i], ptr[i], include)
    }

    oidtable_header_file_end(syntax[1])
    oidtable_source_file_end(syntax)
    makefile_attr(syntax, inch, incc)

    printf "\nproducing " syntax[1] "-Sntx.h\n"
    printf "producing " syntax[1] "-Sntx.c\n"
    printf "producing Makefile.asn\n\n"
}


function syntax_info (syntax, struct, member, ptr)
{
    printf "syntax: " syntax "\n"
    if (struct) {
	printf "struct: " struct "\n"
	if (ptr)
	    printf "ptr_"
	printf "member: " member "\n"
    }
}


function oidtable_header_file (module, syntax, struct, member, ptr, include)
{
    file = module"-Sntx.h"
    type = "type_"module"_"syntax

    printf "/* " syntax " */\n\n" >> file
    printf "#ifdef FunctionPrototypes\n\n" >> file

    printf "PE " syntax "_encode (struct " type "*);\n" >> file
    printf "struct " type "* " syntax "_decode (PE);\n" >> file
    printf "int " syntax "_free (struct " type "*);\n" >> file
    printf "int " syntax "_print (PS, struct " type "*, int);\n" >> file

    parse = syntax"_parse"
    if (find_file_proc(syntax".inc.c", parse, 0))
        printf "struct " type "* " syntax "_parse (char*);\n" >> file

    compare = syntax"_compare"
    if (find_file_proc(syntax".inc.c", compare, 0))
        printf "int " syntax "_compare (struct "type"*, struct "type"*);\n" >> file

    if (struct) {
	if (ptr && member != "char")
	    printf "struct " >> file
        printf struct "** " syntax "_getelem (struct " type "*);\n" >> file
        printf "struct " type "** " syntax "_getnext (struct " type "*);\n" >> file
    }

    printf "\n#else\n\n" >> file

    printf "PE " syntax "_encode ();\n" >> file
    printf "struct " type "* " syntax "_decode ();\n" >> file
    printf "int " syntax "_free ();\n" >> file
    printf "int " syntax "_print ();\n" >> file

    parse = syntax"_parse"
    if (find_file_proc(syntax".inc.c", parse, 0))
        printf "struct " type "* " syntax "_parse ();\n" >> file

    compare = syntax"_compare"
    if (find_file_proc(syntax".inc.c", compare, 0))
        printf "int " syntax "_compare ();\n" >> file

    if (struct) {
	if (ptr && member != "char")
	    printf "struct " >> file
        printf struct "** " syntax "_getelem ();\n" >> file
        printf "struct " type "** " syntax "_getnext ();\n" >> file
    }

    printf "\n#endif\n\n\n" >> file
}


function oidtable_header_file_start (module)
{
    file = module"-Sntx.h"

    printf "/* automatically generated by the OSIMIS attribute compiler, do not edit! */\n\n" > file

    printf "#ifndef _" module "_Sntx_h_\n" >> file
    printf "#define _" module "_Sntx_h_\n\n\n" >> file

    printf "#ifndef _module_" module "_defined_\n" >> file
    printf "#include \"" module "-types.h\"\n" >> file
    printf "#endif\n\n\n" >> file

    printf "#ifndef _FunctionPrototypes\n" >> file
    printf "#if defined(__STDC__) || defined(__cplusplus)\n" >> file
    printf "#define FuntionPrototypes\n" >> file
    printf "#endif\n" >> file
    printf "#endif\n\n" >> file

    printf "#ifdef __cplusplus\n" >> file
    printf "extern \"C\" {\n" >> file
    printf "#endif\n\n\n" >> file
}


function oidtable_header_file_end (module)
{
    file = module"-Sntx.h"
    incfile = module"-Sntx.inc.h"

    mod = tolower(module)

    printf "/* " module " syntaxes */\n\n" >> file
    printf "void " mod"_syntaxes ();\n\n" >> file

    if (getline < incfile > 0) {
        printf "#include \""incfile"\"\n\n" >> file
	close(incfile)
    }

    printf "\n#ifdef  __cplusplus\n" >> file
    printf "}\n" >> file
    printf "#endif\n\n" >> file
    printf "#endif\n" >> file

    close(file)

}


function oidtable_source_file (module, syntax, struct, member, ptr, include)
{
    file = module"-Sntx.c"
    type = "type_"module"_"syntax

    printf "/* " syntax " */\n\n" >> file

    printf "PE " syntax "_encode (val)\n" >> file
    printf "struct " type "* val;\n{\n    PE pe;\n" >> file
    printf "    if (" ENC "_" module "_" syntax  >> file
    printf "(&pe, 1, 0, NULLCP, val) == NOTOK)\n" >> file
    printf "\treturn NULLPE;\n    return pe;\n}\n\n" >> file

    printf "struct " type "* " syntax "_decode (pe)\n" >> file
    printf "PE pe;\n{\n    struct " type "* val;\n" >> file
    printf "    if (" DEC "_" module "_" syntax  >> file
    printf "(pe, 1, NULLIP, NULLVP, &val) == NOTOK)\n" >> file
    printf "\treturn NULL;\n    return val;\n}\n\n" >> file

    free = syntax"_free"
    free_fnx = 0
    if (find_file_proc(syntax".inc.c", free, 0))
	free_fnx = 1
    else {
        printf "int " syntax "_free (val)\n" >> file
        printf "struct " type "* val;\n{\n" >> file
        printf "    free_" module "_" syntax "(val);\n"  >> file
        printf "    return OK;\n}\n\n" >> file
    }

    prnt = syntax"_print"
    print_fnx = 0
    if (find_file_proc(syntax".inc.c", prnt, 0))
	print_fnx = 1
    else {
        printf "/* ARGSUSED */\n" >> file
        printf "int " syntax "_print (ps, val, format)\n" >> file
        printf "PS ps;\nregister struct " type "* val;\nint format;\n{\n" >> file
        printf "    IFP prevfnx; PE pe = NULLPE; int vlen; char* vcp;\n\n" >> file
        printf "    if (!prnt_pe && " ENC "_" module "_" syntax  >> file
        printf "(&pe, 1, 0, NULLCP, val) == NOTOK)\n" >> file
        printf "\treturn NOTOK;\n\n" >> file
        printf "    prevfnx = vfnx, vfnx = bwrite;\n" >> file
        printf "    (void) print_" module "_" syntax >> file
        printf "(pe ? pe : prnt_pe, 1, &vlen, &vcp, NULLCP);\n" >> file
        printf "    vfnx = prevfnx;\n\n" >> file
        printf "    if (pe) pe_free(pe);\n" >> file
        printf "    if (vcp) ps_printf(ps, \"%%s\", vcp);\n" >> file
        printf "    return OK;\n}\n\n" >> file
    }

    parse = syntax"_parse"
    parse_fnx = 0
    if (find_file_proc(syntax".inc.c", parse, 0))
	parse_fnx = 1

    compare = syntax"_compare"
    compare_fnx = 0
    if (find_file_proc(syntax".inc.c", compare, 0))
	compare_fnx = 1

    if (struct) {
	if (ptr && member != "char")
	    printf "struct " >> file
        printf struct "** " syntax "_getelem (val)\n" >> file
        printf "struct " type "* val;\n{\n" >> file
	if (ptr)
            printf "    return &val->" member ";\n}\n\n" >> file
	else {
            printf "    getelem = &val->" member ";\n" >> file
            printf "    return (" struct "**) &getelem;\n}\n\n" >> file
	}

        printf "struct " type "** " syntax "_getnext (val)\n" >> file
        printf "struct " type "* val;\n{\n" >> file
        printf "    return &val->next;\n}\n\n" >> file
    }

    # printf syntax " print_fnx " print_fnx " parse_fnx " parse_fnx "\n"
    if (print_fnx || parse_fnx || compare_fnx || free_fnx)
        printf "#include \"" syntax ".inc.c\"\n\n" >> file
	

    printf "void " syntax "_syntax ()\n{\n" >> file
    printf "    (void) add_attribute_syntax(\n" >> file
    printf "\t\t\"" syntax "\",\n" >> file
    printf "\t\t (IFP) " syntax "_encode,\n" >> file
    printf "\t\t (IFP) " syntax "_decode,\n" >> file

    if (parse_fnx)
        printf "\t\t (IFP) " syntax "_parse,\n" >> file
    else
        printf "\t\t NULLIFP,\n" >> file

    printf "\t\t " syntax "_print,\n" >> file
    printf "\t\t NULLIFP," >> file

    if (compare_fnx)
        printf "\n\t\t " syntax "_compare,\n" >> file
    else
        printf " NULLIFP,\n" >> file

    printf "\t\t " syntax "_free,\n" >> file

    if (struct) {
        printf "\t\t (IFP) " syntax "_getelem,\n" >> file
        printf "\t\t (IFP) " syntax "_getnext,\n" >> file
    }
    else
        printf "\t\t NULLIFP, NULLIFP,\n" >> file

    printf "\t\t FALSE);\n}\n\n\n" >> file
}


function oidtable_source_file_start (module)
{
    file = module"-Sntx.c"

    printf "/* automatically generated by the OSIMIS attribute compiler, do not edit! */\n\n" > file

    printf "#include <stdio.h>\n" >> file
    printf "#include \"msap.h\"\n" >> file
    printf "#include \"" module "-Sntx.h\"\n\n\n" >> file
    printf "extern IFP   vfnx;\t\t\t\t/* in util/sntx/prnt.c */\n" >> file
    printf "extern int   bwrite();\t\t\t\t/*  ..  */\n" >> file
    printf "extern PE    prnt_pe;\t\t\t\t/*  ..  */\n\n" >> file
    printf "/* the following may be used in user-defined <syntax>_parse methods */\n" >> file
    printf "extern int  empty_asn1_type;\t\t\t/* in msap/attrv.c */\n" >> file
    printf "extern char sntxbuf [];\t\t\t\t/* in agent/smi/SmiSntx.c */\n" >> file
    printf "extern void* getelem;\t\t\t\t/*  ..  */\n" >> file
#    printf "#ifdef __linux__\n" >> file
#    printf "#include \"Boolean.h\"\n" >> file
#    printf "#endif\n\n\n" >> file
    printf "\n\n" >> file
}


function oidtable_source_file_end (syntax)
{
    file = syntax[1]"-Sntx.c"
    incfile = syntax[1]"-Sntx.inc.c"

    mod = tolower(syntax[1])

    printf "/* " syntax[1] " syntaxes */\n\n" >> file
    printf "void " mod"_syntaxes ()\n{\n" >> file

    for (i = 2; i <= NR; i++)
	printf "    " syntax[i]"_syntax();\n" >> file
    printf "}\n" >> file

    if (getline < incfile > 0) {
        printf "\n#include \""incfile"\"\n" >> file
	close(incfile)
    }

}


function class_header_file (module, syntax, struct, member, ptr, include)
{
    type = "type_"module"_"syntax
    file = syntax".h"

    constructor = 1
    if ( ptr && ! (struct && member) )
	constructor = 0		# non-multivalued syntax with pointer members

    printf "// automatically generated by the OSIMIS attribute compiler, do not edit!\n\n" > file

    printf "#ifndef _" syntax "_h_\n" >> file
    printf "#define _" syntax "_h_\n\n\n" >> file

    printf "#ifndef _GENERIC_ATTR_\n" >> file
    printf "#include \"GenericAttr.h\"\n" >> file
    printf "#endif\n\n" >> file

    printf "#ifndef _" module "_Sntx_h_\n" >> file
    printf "#include \"" module "-Sntx.h\"\n" >> file
    printf "#endif\n\n" >> file

    printf "#ifndef _ucbext_h\n" >> file
    printf "#include \"ucbext.h\"\n" >> file
    printf "#endif\n\n\n" >> file

    printf "class " syntax " : public Attr\n" >> file    
    printf "{\nprotected:\n" >> file

    printf "    inline PE     _encode ()\n" >> file
    printf "    { return " syntax "_encode((" type >> file
    printf "*) getval()); }\n\n" >> file

    printf "    inline void*  _decode (PE pe)\n" >> file
    printf "    { return " syntax "_decode(pe); }\n\n" >> file

    printf "    inline void   _free ()\n" >> file
    printf "    { " syntax "_free((" type "*) getval()); }\n\n" >> file


    prnt = syntax"_print"
    print_fnx = find_file_proc(syntax".inc.c", prnt, 0)

    printf "    inline char*  _print ()\n" >> file
    if (print_fnx)
        printf "    { return attrv2str(getval(), (PRINT_FNX) " syntax "_print); }\n\n" >> file
    else {
        printf "    { prnt_pe = getPE();\n" >> file
	printf "      char* str = attrv2str(getval(), (PRINT_FNX) " syntax "_print);\n" >> file
        printf "      prnt_pe = NULLPE; return str; }\n\n" >> file
    }

    if (find_file_proc(syntax".inc.c", syntax"_parse", 0)) {
        printf "    inline void*  _parse (char* str)\n" >> file
        printf "    { return " syntax "_parse(str); }\n\n" >> file
    }

    if (find_file_proc(syntax".inc.c", syntax"_compare", 0)) {
        printf "    inline int    _compare (void* v1, void* v2)\n" >> file
        printf "    { return " syntax "_compare(("type"*) v1, ("type"*) v2); }\n\n" >> file
    }

    if (struct) {
        printf "    inline void**  _getelem (void* v)\n" >> file
        printf "    { return (void**) " syntax "_getelem((" type >> file
        printf "*) v); }\n\n" >> file

        printf "    inline void** _getnext (void* v)\n" >> file
        printf "    { return (void**) " syntax "_getnext((" type >> file
        printf "*) v); }\n\n" >> file

    }

    printf "public:\n" >> file

    printf "    inline char*  getSyntax ()\n    { return \""syntax"\"; }\n\n" \
							>> file
    if (struct) {
	if (! index(struct, "element_"module"_")) {
            printf "    inline int    addval (" struct ptr " v)\n" >> file
            printf "    { " type "* m = new " type ";\n" >> file
            printf "      m->" member " = v; m->next = NULL; return add(m); }\n\n" >> file

            printf "    inline int    remval (" struct ptr " v)\n" >> file
            printf "    { " type " m;\n" >> file
            printf "      m." member " = v; m.next = NULL; return remove(&m); }\n\n" >> file
	}
    }

    if (! (include && find_file_proc(syntax".inc.h", syntax, 1))) # not supplied
	if (struct && member) {		# multi-valued
            printf "    inline " syntax " ()\n" >> file
            printf "    { setval(NULL); }\n\n" >> file
	}
        else
	if (!ptr) {
	    struct = "type_"module"_"syntax
            printf "    inline " syntax " ()\n" >> file
            printf "    { " struct "* v = new " struct ";\n" >> file
            printf "      bzero((char*) v, sizeof(" struct ")); setval(v); }\n\n" >> file
        }
	else {
            printf "    // you will need to supply the inline constructor in " \
			syntax".inc.h -\n" >> file
	    printf "    // see Attr class documentation for template\n" \
			>> file
            printf "    "syntax " ();\n" >>file
	    printf "\nWARNING: you need to supply a constructor for class \"" \
			syntax "\"\n\n"
	}

    if (include == 1)
        printf "#include \"" syntax ".inc.h\"\n\n" >> file

    printf "};\n\n\n#endif\n" >> file
    close(file)
}


function find_file_proc (file, proc, noargs)
{
    if (noargs)
	regexpr = proc" *\\( *\\)"
    else
	regexpr = proc" *\\("

    # printf file " " proc "\n"
    while (getline < file > 0) {
	if (match($0, regexpr)) {
	    # printf "found " regexpr "\n\n"
            close(file)
	    return 1
	}
    }
    close(file)
    return 0
}


function makefile_attr(syntax, inch, incc)
{
    file = "Makefile.asn"

    inchfile = syntax[1]"-Sntx.inc.h"
    hasinchfile = 0
    if (getline < inchfile > 0) {
        hasinchfile = 1
	close(inchfile)
    }
    inccfile = syntax[1]"-Sntx.inc.c"
    hasinccfile = 0
    if (getline < inccfile > 0) {
        hasinccfile = 1
	close(inccfile)
    }

    printf "########################################################################\n#\n" >> file
    printf "# (Part of) Sample Makefile produced by the OSIMIS attribute compiler.\n" >> file
    printf "# It should be merged with the one produced by the GDMO compiler.\n" >> file
    printf "# Note that manual augmentation is necessary (for targets etc.)\n" >> file
    printf "#\n########################################################################\n\n" >> file
    printf "# ASN.1 Module:  "syntax[1]"\n\n\n" >> file

    printf "# header files\n\n" >> file
    printf "HFILES =\t" syntax[1]"-types.h " syntax[1]"-Sntx.h" >> file
    if (hasinchfile)
	printf " \\\n\t\t"inchfile >> file
    for (i = 2; i <= NR; i++) {
	printf " \\\n\t\t" syntax[i]".h" >> file
	if (inch[i])
	    printf " "inch[i] >> file
    }

    printf "\n\n# object files\n\n" >> file
    printf "OSNTXFILES =\t" syntax[1]"_tables.o " syntax[1]"-Sntx.o" >> file

    printf "\n\n\n# files to be \"cleaned\"\n\n" >> file
    printf "CLEANFILES =\t" syntax[1]"_tables.c " syntax[1]"-types.h" >> file
    printf " \\\n\t\t" syntax[1]"-Sntx.c " syntax[1]"-Sntx.h" >> file
    for (i = 2; i <= NR; i++)
	printf " \\\n\t\t" syntax[i]".h" >> file
    printf "  \\\n\t\tMakefile.* *.gen *.log\n\n" >> file

    printf "# source file dependencies\n\n" > file
    printf syntax[1]"_tables.c " syntax[1]"-types.h \\\n" >> file
    printf syntax[1]"-Sntx.c " syntax[1]"-Sntx.h \\\n" >> file
    for (i = 2; i <= NR-1; i++)
	printf syntax[i]".h \\\n" >> file
    printf syntax[NR]".h:\t"syntax[1]".asn "syntax[1]".gdmo" >> file
	
    for (i = 2; i <= NR; i++) {
	# printf "inch " inch[i] " incc " incc[i] "\n"
	if (inch[i] || incc[i]) {
	    printf " \\\n\t\t" >> file
	    if (inch[i]) {
		printf inch[i] >> file
		if (incc[i])
		    printf " " >> file
	    }
	    if (incc[i])
		printf incc[i] >> file
	}
    }
    printf "\n\t\t$(GDMOASN-CMPL) "syntax[1]".py "syntax[1]".gdmo\n\n" >> file

    printf "# object file dependencies\n\n" > file
    printf syntax[1]"_tables.o:\t" syntax[1]"_tables.c "syntax[1]"-types.h \n" >> file
    printf syntax[1]"-Sntx.o:\t" syntax[1]"-types.h "syntax[1]"-Sntx.h" >> file
    if (hasinchfile || hasinccfile) {
	printf " \\\n\t\t" >> file
	if (hasinchfile)
	    printf inchfile " " >> file
	if (hasinccfile)
	    printf inccfile >> file
    }
    for (i = 2; i <= NR; i++)
	if (incc[i])
	    printf " \\\n\t\t" incc[i] >> file
    printf "\n" >> file
}
