#
# attrtbl.awk - parses the <module>.attr and <module>-types.h files and
#               produces the asn-cmpl.log file to be parsed by attrgen.awk
#               which subsequently generates the attribute source files
#
# By George Pavlou - November 1994
#

BEGIN {
    nsntx = 1
}

{
    if (! index($1, "#"))	# ignore comment lines
	syntax[nsntx++] = $1
}


END {
    # for debugging...
    # for (i = 1; i < nsntx; i++)
	# printf syntax[i] "\n"
    # printf "\n"

    pepsyhfile = syntax[1]"-types.h"
    if (getline < pepsyhfile <= 0) {
	printf "pepsy generated file " pepsyhfile " does not exit!\n"
	printf "you may check Makefile dependencies\n\n"
	exit(-1)
    }
    close(pepsyhfile)

    # printf "\nmodule: " syntax[1] "\n\n"
    attr_table_start(syntax[1])

    for (i = 2; i < nsntx; i++) {
        # printf "syntax: " syntax "\n"

	syntax_table_entry(syntax[1], syntax[i])
    }
}


function attr_table_start (module)
{
    file = module".tbl"
    printf module "\n" # > file
}


function attr_table_entry (module, syntax, struct, member, ptr)
{
    file = module".tbl"

    printf syntax # >> file
    # printf "syntax: " syntax "\n"

    if (struct && member) {
	printf " " struct " " member "\n" # >> file
	# printf "struct: " struct "\n"
	# printf "member: " member "\n"
    }
    else
    if (ptr) {
        printf " *\n" # >> file
	# printf "ptr: *\n"
    }
    else
	printf "\n" # >> file
}


function syntax_table_entry (module, syntax)
{
    type = "type_"module"_"syntax
    ptr = ""

    while (getline < pepsyhfile > 0) {
	if (! match($0, "^struct "type))
	    continue
	nbracket = 1

	# found the type declaration

	getline < pepsyhfile	# next line
	# printf $0 "\n"

	if ($1 == "struct")
	    strct = $2
	else
	    strct = $1

	if (! index($0, "{")) {
	    member = $NF
	    sub(";", "", member)
	}
	else {
	    ++nbracket
	    member = ""
	}
	# printf "nbracket " nbracket " struct " strct " member " member "\n"

	prevline = $0
	while (getline < pepsyhfile > 0) {
	    # printf $0 "\n"

	    # check if there is a pointer member (non multi-valued syntaxes)
	    if (!ptr && index($0, "*"))
		ptr = "*"

	    if (index($0, "{")) {
		++nbracket
		# printf "nbracket " nbracket "\n"
	    }
	    else
	    if (index($0, "}")) {
		--nbracket
		# printf "nbracket " nbracket "\n"

		if (!member && nbracket == 1) {
		    member = $NF
		    sub(";", "", member)
		    prevline = $0
		    # printf "member " member "\n"
		    continue
		}
		if (nbracket == 0) {

		    if (! index(prevline, "*next")) {
			strct = ""
			member = ""
		    }
		    attr_table_entry(module, syntax, strct, member, ptr)
		    close(pepsyhfile)
		    return;
		}
		# printf "nbracket " nbracket "\n"
	    }
	    prevline = $0
	}
    }

    printf "\nERROR: struct " type " not found in "  pepsyhfile "!\n" > "/dev/stderr"
    printf "check the ASN.1 in " module ".py\n" > "/dev/stderr"
    printf "if " syntax " is a string/OID type or an \"alias\", remove it from that file - \n" > "/dev/stderr"
    printf "you may drive the GDMO compiler through a MIB.syntax.dat file\n\n" > "/dev/stderr"

    close(pepsyhfile)
}

