//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Counter.h

// By George Pavlou, February 1991


#ifndef _Counter_h_
#define _Counter_h_


#ifndef _GENERIC_ATTR_
#include "GenericAttr.h"
#endif

#ifndef _SMI_SNTX_
#include "SmiSntx.h"
#endif


class Counter;


class CThreshold : public Attr
{
protected:
    Counter* counter;

    inline	CThreshold ()		{ setval(NULL); counter = NULL; }

public:
    inline void		  associateCounter (Counter* cntr) { counter = cntr; }
    inline virtual Switch check (int, int) { return Off; }
};


class Counter : public Attr
{
    CThreshold* threshold;		// threshold applied to the counter

protected:
    inline PE	 _encode ()		{ return int_enc((int*) getval());}
    inline void* _decode (PE pe)	{ return int_dec(pe); }
    inline void  _free ()		{ free((char*) getval()); }
    inline char* _print ()		{ return attrv2str(getval(),
					  (PRINT_FNX) int_print); }
    inline int	 _compare (void* v1, void* v2)
					{ return int_cmp((int*) v1, (int*) v2);}
public:
    inline int	getint ()		{ return *(int*) getval(); }
    Switch	set (int);
    Switch	increment (int);
    inline int	set (void* val)		{ set(*(int*) val); return OK; }
    inline Switch increment ()		{ return increment(1); }
    inline void	associateThreshold (CThreshold* thld)
					{ (threshold = thld) ->
						associateCounter(this); }
    inline char* getSyntax ()		{ return "Count"; }

    inline	Counter ()		{ int val = 0; setval(int_cpy(&val));
					  threshold = NULL; }
};


#endif
