//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and	binary forms are permitted
// provided that the above copyright notice and	this paragraph are
// duplicated in all such forms	and that any documentation,
// advertising materials, and other materials related to such
// distribution	and use	acknowledge that the software was developed
// by the Department of	Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS	SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/******************************************************************/
/*

	File:		GenString.h

		String Class written because some
		people do not use the excellent	GNU C++
		String class (sigh!).


	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/*****************************************************************/

#include "DebugMem.h"

#ifndef	_GENERIC_STRING_

#define	_GENERIC_STRING_

/*****************************************************************/

#include <string.h>

/*****************************************************************/
/* Constants */

#define	RXwhite		1

/*****************************************************************/
/* Debugging constants etc */
/*
#define	DEBUG_STRING	1
*/

/*****************************************************************/

class	GString_chars
{
	char	*_str;
	int	_ref_count;
	friend	class	GString;
	GString_chars(char *str, int lgth)
	{
		_str = (NEW char[lgth+1]);
		strncpy(_str, str, lgth);
		_str[lgth] = '\0';
		_ref_count = 1;

#ifdef	DEBUG_STRING

	printf("\nGString_chars: --%s--%d--%d--",
		 _str, lgth, _ref_count);
#endif
	}
	~GString_chars()
	{

		DELETE	_str;
	}
};

/*****************************************************************/

class	GString
{
	GString_chars	*_str_chars;

	void	deleteStr()
	{
		if (!--_str_chars->_ref_count)
			DELETE _str_chars;
	}

protected:
	void	initialise(char	*chars,	int lgth)
	{
		_str_chars = (NEW GString_chars(chars, lgth));

#ifdef	DEBUG_STRING
		printf("\nGString: Initialise: --%s--%d--", chars, lgth);
#endif
	}

public:
	GString()
	{
		initialise("", 0);
	}

	GString(char *str)
	{
		initialise(str,	strlen(str));
	}

	GString(char *str, int lgth)
	{
		initialise(str,	lgth);
	}

	GString(char achar)
	{
/*
	char temp[2];

		temp[0]	= achar;
		temp[1]	= '\0';
*/
		initialise(&achar, 1);
	}

	GString(GString	&init)
	{
		_str_chars = init._str_chars;
		_str_chars->_ref_count++;
	}

	virtual	~GString()
	{
		deleteStr();
	}

	void	StrReplace(char	*str)
	{
		deleteStr();
		initialise(str,	strlen(str));
	}

	friend	GString	operator+(GString &, GString &);
	friend	GString	operator+=(GString &, GString &);
	friend	GString	operator+=(GString &, char *);
	friend	GString	operator+=(GString &, char);
	GString	&operator=(GString &);
	GString	&operator=(char	*);
	GString	after(int);
	GString	after(char *);
	GString	before(int);
	GString	before(char *);
	char	*search(char *,	char *);
	operator char *();
	char	*print()
	{
		return(_str_chars->_str);
	}
};

/*****************************************************************/

typedef	GString	*GStringPtr;

/*****************************************************************/

#endif

/*****************************************************************/

