/*
 * Copyright (c) 1993 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* */

/* */

/* @(#)Library.h 	 01/04/1993  */

/*M*******************************************************************


  Module name:       Library.h
     
  Description:       This module contains the include file of the 
                     library that comes with the Attribute Compiler
		     and that implements some basic comparing functions.

  Creation date:     April 1st, 1993

  Coded by:          Luca Deri


  Modification history: 

**********************************************************************/

#ifndef _LIBRARY_H_
#define _LIBRARY_H_

#include <stdlib.h>
#include <string.h>
#include "Boolean.h"
#include "mparm.h"

#define A_EQUAL_B                    0
#define A_GREATER_EQUAL_B            1
#define A_LESS_EQUAL_B               2
#define A_SUBSET_OF_B                3
#define A_SUPERSET_OF_B              4
#define A_NON_NULL_INTERSECTION_B    5
#define A_NOT_EQUAL_B                6


#if defined(__STDC__) || defined(__cplusplus)
typedef void   (*FreeType)    (void*);
typedef int    (*CompareType) (void*, void*);
typedef int    (*PrintType)   (PS, void*, int);
typedef void** (*NextElement) (void*);
#else
typedef void   (*FreeType)    ();
typedef int    (*CompareType) ();
typedef int    (*PrintType)   ();
typedef void** (*NextElement) ();
#endif


#ifndef _LIBRARY_C_

/* Prototypes of functions defined in the file Library.cc */

#if defined(__STDC__) || defined(__cplusplus)
#ifdef __cplusplus
extern "C" {
#endif
int   SeqOfPrint(PS, void*, int,  Bool, PrintType, NextElement);
int   SeqOfCompare(void*, void*, Bool, CompareType, NextElement);
void* SeqOfAdd(void*, void*, FreeType, Bool, CompareType, NextElement);
void* SeqOfRemove(void*,  void*, FreeType, Bool, CompareType, NextElement);
Bool  SeqOfSetCompare(int, void*, void*, Bool, CompareType, NextElement);
int   SetOfCompare(void*, void*, Bool, CompareType, NextElement);
void* SetOfAdd(void*, void*, FreeType, Bool, CompareType, NextElement);
void* SetOfRemove(void*, void*, FreeType, Bool, CompareType, NextElement);
Bool  SetOfSetCompare(int, void*, void*, Bool, CompareType, NextElement);
Bool  MemoryContainment(char*, int, char*, int, int);
#ifdef __cplusplus
}
#endif
#else
int   SeqOfPrint();
int   SeqOfCompare();
void* SeqOfAdd();
void* SeqOfRemove();
Bool  SeqOfSetCompare();
int   SetOfCompare();
void* SetOfAdd();
void* SetOfRemove();
Bool  SetOfSetCompare();
Bool  MemoryContainment();
#endif

/* Same function for SEQUENCE OF and SET OF*/
#define SetOfPrint SeqOfPrint

#define SubstringContainment(a, b, c) \
        MemoryContainment(a, strlen(a), b, strlen(b), c)

#define String_free qb_free	/* qbuf free */

/* *** Other library functions */

#if defined(__STDC__) || defined(__cplusplus)
#ifdef __cplusplus
extern "C" {
#endif
int compareInt(void*, void*);
int compareChar(void*, void*);
int compareReal(void*, void*);
int printInt(PS, int*, int);
int printChar(PS, int*, int);
int printReal(PS, int*, int);

int String_compare(void *elem_a, void *elem_b);
int String_print(PS ps, struct qbuf* elem, int quoted);

char* printFilter(struct qbuf* qb);
#ifdef __cplusplus
}
#endif
#else
int compareInt();
int compareChar();
int compareReal();
int printInt();
int printChar();
int printReal();

int String_compare();
int String_print();

char* printFilter();
#endif

#endif
#endif



