#ifndef _PROXYALARMKS_H_
#define _PROXYALARMKS_H_



extern "C" {

#include "SNMP-types.h"
#include "PrxAlarmSntx-types.h"

int	qb_cmp( struct qbuf*, struct qbuf* ) ;
struct 	qbuf* qb_cpy( struct qbuf* ) ;

} ;

#ifndef _SMI_SNTX_
#include "SmiSntx.h"
#endif

#ifndef _GENERIC_KS_
#include "GenericKS.h"
#endif

#ifndef _GEN_ARRAY_
#include "GenArray.h"
#endif

#ifndef _BOOLEAN_ 
#include "Boolean.h"
#endif

struct	PortDetails {
public:
	int 	sdPort ;
	int	udpPort ;
} ;

class	PortArray : public Array {

protected:
	int	compare(void*, void*) ;
public:
	PortArray( Bool ) ;
} ;

inline PortArray::PortArray( Bool flag ) : Array( flag ) { ; }


// Comparison based on EITHER the socket port OR the udp port, depending of the socket port..
// on whether the respective port in the first parameter is set to -1...
inline	int PortArray::compare( void* port1p, void* port2p )
{
	if( ((struct PortDetails*)port1p)->sdPort == -1 )
	    return int_cmp( (int*)&((struct PortDetails*)port1p)->udpPort, 
			    (int*)&((struct PortDetails*)port2p)->udpPort ) ;
	else
	    return int_cmp( (int*)&((struct PortDetails*)port1p)->sdPort, 
			    (int*)&((struct PortDetails*)port2p)->sdPort ) ;
}


struct	ProxyAgentTrapDetails {

	struct  qbuf * 	trapCommunityStringqb ;	
	int		udpTrapPort ;
} ;

class	ProxyAgentTrapDetailsArray : public Array {

protected:
	int	compare(void*, void*) ;
	void 	deleteElem (void*);
public:
	ProxyAgentTrapDetailsArray( Bool ) ;
} ;


inline ProxyAgentTrapDetailsArray::ProxyAgentTrapDetailsArray( Bool flag ) : Array( flag ) { ; }

// compare of the UDP port..
inline int ProxyAgentTrapDetailsArray::compare( void* paTrapDetails1p, void* paTrapDetails2p )
{
	return int_cmp( (int*)&(((struct ProxyAgentTrapDetails*)paTrapDetails1p)->udpTrapPort),
			(int*)&(((struct ProxyAgentTrapDetails*)paTrapDetails2p)->udpTrapPort) ) ;
}

inline void ProxyAgentTrapDetailsArray::deleteElem( void* paTrapDetailsp )
{
struct ProxyAgentTrapDetails* patdsp = (struct ProxyAgentTrapDetails*)paTrapDetailsp ;

	if( patdsp->trapCommunityStringqb )
	    qb_free( patdsp->trapCommunityStringqb ) ;

	delete patdsp ;
}


#define PROXYALARMDEFTRAPPORT		162
#define PROXYALARMNONDEFMINTRAPPORT	1500

#define PROXYALARMTRAPPSLEN	8192


class cmipsnmpProxyAgent ;


class ProxyAlarmKS : public KS {

    static ProxyAlarmKS* _instance ;
    PortArray *		 _portArrayp ;
    int		     	 _nobjs;

    void	     	 closePorts( void );

    OID			 formEventType( struct type_SNMP_Trap__PDU * ) ;
    PE 			 formEventInfo( struct type_SNMP_Trap__PDU *, OID, 
    					struct qbuf*,
					cmipsnmpProxyAgent*,
					struct type_PrxIIMCSntx_TransportAddresses * ) ;

    OID			 convertInternetOIDtoIIMCOID( OID ) ;

    DN 	determineObjectInstanceDN( cmipsnmpProxyAgent*, OID ) ;

public:

    int		     	 initialisePort( int );

    static ProxyAlarmKS* getInstance( void );
    int		     	 getNObjects( void );
    int		     	 registerMO( MO*, char* );
    int		     	 deregisterMO(MO*);
    int		     	 readCommEndpoint(int);
    int		     	 shutdown(int);

    ProxyAlarmKS();
    ~ProxyAlarmKS();
};

inline ProxyAlarmKS*  ProxyAlarmKS::getInstance (){ return _instance; }
inline int  ProxyAlarmKS::getNObjects (){ return _nobjs; }
inline int ProxyAlarmKS::shutdown (int)	{ closePorts(); return OK; }
inline ProxyAlarmKS::ProxyAlarmKS ()	{ _instance = this; _nobjs = 0; _portArrayp = new PortArray( True ) ; }
inline ProxyAlarmKS::~ProxyAlarmKS ()	{ closePorts(); if( _portArrayp ) { _portArrayp->clear(), delete _portArrayp ; } ; _instance = NULL ; }
						

#endif
