/*
 * Copyright (c) 1990 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* SmiSntx.h - SMI types syntax handling method declarations */

/* By George Pavlou / Saleem Bhatti, December 1991 */


#ifndef _SMI_SNTX_
#define _SMI_SNTX_


#ifndef _SMI_ASN_
#include "SmiAsn.h"
#endif

#ifndef _module_SMI_defined_
#include "SMI-types.h"
#endif


#ifndef FunctionPrototypes
#if defined(__STDC__) || defined(__cplusplus)
#define FunctionPrototypes
#endif
#endif

#ifdef	__cplusplus
extern "C" {
#endif


/*
 * to avoid having to change hand-written code every time a new type
 * is added somewhere in the middle of SmiAsn.py - pepsy is a drag...
 */

#define attrValChangeDefinition_element	struct element_SMI_5
#define attrValChangeDefinition_member	member_SMI_8


#ifdef FunctionPrototypes

PE	r_octenc (qbuf*);
qbuf*	r_octsdec (PE);
int	r_octprint (PS, qbuf*, int);
int	strprint (PS, char*, int);
char*	hexprint (char*, int);
int	pstrcmp (char*, char*);

int	oidprint (PS, OID, int);

PE	time_enc (char*);
char*	time_dec (PE);
int	time_cmp (char*, char*);

PE	dn_enc (DN);
DN	dn_dec (PE);

OID	dup_prim2oid (PE);

PE	bool_enc (Bool*);
Bool*	bool_dec (PE);
Bool*	bool_parse (char*);
int	bool_print (PS, Bool*, int);

PE	int_enc (int*);
int*	int_dec (PE);
int*	int_parse (char*);
int	int_print (PS, int*, int);
int	int_free (int*);
int*	int_cpy (int*);
int	int_cmp (int*, int*);

PE	IntegerList_encode (IntegerListVal*);
IntegerListVal*	IntegerList_decode (PE);
int	IntegerList_print (PS, IntegerListVal*, int);
int	IntegerList_free (IntegerListVal*);
int	IntegerListVal_cmp (IntegerListVal*, IntegerListVal*);
int*	IntegerList_getelem (IntegerListVal*);
IntegerListVal**	IntegerList_getnext (IntegerListVal*);

PE	real_enc (double*);
double*	real_dec (PE);
double*	real_parse (char*);
int	real_print (PS, int*, int);
int	real_free (double*);
double*	real_cpy (double*);
int	real_cmp (double*, double*);

PE	RealList_encode (RealListVal*);
RealListVal*	RealList_decode (PE);
int	RealList_print (PS, RealListVal*, int);
int	RealList_free (RealListVal*);
int	RealListVal_cmp (RealListVal*, RealListVal*);
double*	RealList_getelem (RealListVal*);
RealListVal**	RealList_getnext (RealListVal*);

StringListVal*	StringList_parse (char*);
int	StringList_print (PS, StringListVal*, int);
int	StringList_free (StringListVal*);
int	StringListVal_cmp (StringListVal*, StringListVal*);
char*	StringList_getelem (StringListVal*);
StringListVal**	StringList_getnext (StringListVal*);

PE	GraphicStringList_encode (StringListVal*);
StringListVal*	GraphicStringList_decode (PE);

PE	IA5StringList_encode (StringListVal*);
StringListVal*	IA5StringList_decode (PE);

PE	PrintableStringList_encode (StringListVal*);
StringListVal*	PrintableStringList_decode (PE);

PE	SimpleNameType_encode (SimpleNameTypeVal*);
SimpleNameTypeVal*	SimpleNameType_decode (PE);
SimpleNameTypeVal*	SimpleNameType_parse (char*);
int	SimpleNameType_print (PS, SimpleNameTypeVal*, int);
int	SimpleNameType_free (SimpleNameTypeVal*);
int	SimpleNameType_cmp (SimpleNameTypeVal*, SimpleNameTypeVal*);

MID		MID_parse (char*);
int		MID_print (PS, MID, int);
int		MID_free (MID);
int		MID_cmp (MID, MID);

MIDListVal*	MIDList_parse (char*);
int		MIDList_print (PS, MIDListVal*, int);
int		MIDList_free (MIDListVal*);
int		MIDListVal_cmp (MIDListVal*, MIDListVal*);
MID		MIDList_getelem (MIDListVal*);
MIDListVal**	MIDList_getnext (MIDListVal*);

PE		ObjectClass_encode (MID);
MID		ObjectClass_decode (PE);

PE		ObjectClassList_encode (MIDListVal*);
MIDListVal*	ObjectClassList_decode (PE);

PE		AttributeId_encode (MID);
MID		AttributeId_decode (PE);

PE		AttributeIdList_encode (MIDListVal*);
MIDListVal*	AttributeIdList_decode (PE);

PE		EventType_encode (MID);
MID		EventType_decode (PE);

PE		ObjectInstance_encode (MN);
MN		ObjectInstance_decode (PE);
MN		ObjectInstance_parse (char*);
int		ObjectInstance_print (PS, MN, int);
int		ObjectInstance_free (MN);
int		ObjectInstance_cmp (MN, MN);

PE		ObjectInstanceList_encode (MNListVal*);
MNListVal*	ObjectInstanceList_decode (PE);
MNListVal*	ObjectInstanceList_parse (char*);
int		ObjectInstanceList_print (PS, MNListVal*, int);
int		ObjectInstanceList_free (MNListVal*);
int		ObjectInstanceListVal_cmp (MNListVal*, MNListVal*);
MN		ObjectInstanceList_getelem (MNListVal*);
MNListVal**	ObjectInstanceList_getnext (MNListVal*);

PE		CMISScope_encode (CMISScope*);
CMISScope*	CMISScope_decode (PE);
int		CMISScope_print (PS, CMISScope*, int);

PE		CMISFilter_encode (CMISFilter*);
CMISFilter*	CMISFilter_decode (PE);
int		CMISFilter_print (PS, CMISFilter*, int);

PE		OIDList_encode (ObjIdListVal*);
ObjIdListVal*	OIDList_decode (PE);
ObjIdListVal*	OIDList_parse (char*);
int		OIDList_print (PS, ObjIdListVal*, int);
int		OIDList_free (ObjIdListVal*);
int		OIDListVal_cmp (ObjIdListVal*, ObjIdListVal*);
OID		OIDList_getelem (ObjIdListVal*);
ObjIdListVal**	OIDList_getnext (ObjIdListVal*);

PE		ObservedValue_encode (ObservedValue*);
ObservedValue*  ObservedValue_decode (PE);
ObservedValue*  ObservedValue_parse (char*);
int		ObservedValue_print (PS, ObservedValue*, int);
int		ObservedValue_free (ObservedValue*);
ObservedValue*	ObservedValue_cpy (ObservedValue*);
int		ObservedValue_cmp (ObservedValue*, ObservedValue*);

PE			CounterThreshold_encode (CounterThresholdVal*);
CounterThresholdVal*	CounterThreshold_decode (PE);
CounterThresholdVal*	CounterThreshold_parse (PE);
int			CounterThreshold_print (PS, CounterThresholdVal*, int);
int			CounterThreshold_free (CounterThresholdVal*);
int			CounterThresholdVal_cmp (CounterThresholdVal*,
						CounterThresholdVal*);
CounterThresholdVal*	CounterThreshold_getelem (CounterThresholdVal*);
CounterThresholdVal**	CounterThreshold_getnext (CounterThresholdVal*);

PE			GaugeThreshold_encode (GaugeThresholdVal*);
GaugeThresholdVal*	GaugeThreshold_decode (PE);
GaugeThresholdVal*	GaugeThreshold_parse (char*);
int			GaugeThreshold_print (PS, GaugeThresholdVal*, int);
int			GaugeThreshold_free (GaugeThresholdVal*);
int			GaugeThresholdVal_cmp (GaugeThresholdVal*,
						GaugeThresholdVal*);
GaugeThresholdVal*	GaugeThreshold_getelem (GaugeThresholdVal*);
GaugeThresholdVal**	GaugeThreshold_getnext (GaugeThresholdVal*);

PE		TideMark_encode (TideMarkVal*);
TideMarkVal*	TideMark_decode (PE);
int		TideMark_print (PS, TideMarkVal*, int);
int		TideMark_free (TideMarkVal*);
int		TideMark_cmp (TideMarkVal*, TideMarkVal*);

PE		AdministrativeState_encode (AdminStateVal*);
AdminStateVal*	AdministrativeState_decode (PE);
AdminStateVal*	AdministrativeState_parse (char*);
int		AdministrativeState_print (PS, AdminStateVal*, int);
int		AdministrativeState_free (AdminStateVal*);
int		AdministrativeState_cmp (AdminStateVal*, AdminStateVal*);

PE		OperationalState_encode (OperStateVal*);
OperStateVal*	OperationalState_decode (PE);
OperStateVal*	OperationalState_parse (char*);
int		OperationalState_print (PS, OperStateVal*, int);
int		OperationalState_free (OperStateVal*);
int		OperationalState_cmp (OperStateVal*, OperStateVal*);

PE		UsageState_encode (type_SMI_UsageState*);
type_SMI_UsageState*	UsageState_decode (PE);
int		UsageState_print (PS, type_SMI_UsageState*, int);
int		UsageState_free (type_SMI_UsageState*);
int		UsageState_cmp (type_SMI_UsageState*, type_SMI_UsageState*);

PE		DestAddress_encode (DestAddressVal*);
DestAddressVal*	DestAddress_decode (PE);
int		DestAddress_print (PS, DestAddressVal*, int);
int		DestAddress_free (DestAddressVal*);

PE ObjectInfo_encode (struct type_SMI_ObjectInfo*);
struct type_SMI_ObjectInfo* ObjectInfo_decode (PE);
void ObjectInfo_free (struct type_SMI_ObjectInfo*);
void ObjectInfo_free_aux (struct type_SMI_ObjectInfo*);
int ObjectInfo_print (PS, struct type_SMI_ObjectInfo*, int);

struct type_SMI_AttrValChangeInfo* AttrValChangeInfo_alloc ();
PE AttrValChangeInfo_encode (struct type_SMI_AttrValChangeInfo*);
struct type_SMI_AttrValChangeInfo* AttrValChangeInfo_decode (PE);
void AttrValChangeInfo_free (struct type_SMI_AttrValChangeInfo*);
void AttrValChangeInfo_free_aux (struct type_SMI_AttrValChangeInfo*);
int AttrValChangeInfo_print (PS, struct type_SMI_AttrValChangeInfo*, int);

struct type_SMI_AlarmInfo* AlarmInfo_alloc ();
struct type_SMI_AlarmInfo* ThresholdAlarmInfo_alloc ();
struct type_SMI_AlarmInfo* ThresholdAlarmInfo_ProbsAttrs_alloc
			(struct type_SMI_AlarmInfo*, int, int);
PE AlarmInfo_encode (struct type_SMI_AlarmInfo*);
struct type_SMI_AlarmInfo* AlarmInfo_decode (PE);
void AlarmInfo_free (struct type_SMI_AlarmInfo*);
void AlarmInfo_free_aux (struct type_SMI_AlarmInfo*);
int AlarmInfo_print (PS, struct type_SMI_AlarmInfo*, int);

PE                 encodeLogFullAction(LogFullActionValue *);
LogFullActionValue *decodeLogFullAction(PE);
LogFullActionValue *parseLogFullAction(char *);
char               *LogFullAction2Str(LogFullActionValue *);
int                printLogFullAction(PS, LogFullActionValue *, int);
void               freeLogFullAction(LogFullActionValue *); /* == free() */
LogFullActionValue *copyLogFullAction(LogFullActionValue *);
int                compareLogFullAction(LogFullActionValue *,
					LogFullActionValue *);

PE                      encodeAvailabilityStatus(AvailabilityStatusValue *);
AvailabilityStatusValue *decodeAvailabilityStatus(PE);
AvailabilityStatusValue *parseAvailabilityStatus(char *);
char                    *AvailabilityStatus2Str(AvailabilityStatusValue *);
int                     printAvailabilityStatus(PS, AvailabilityStatusValue *,
						int);
void                    freeAvailabilityStatus(AvailabilityStatusValue *);
AvailabilityStatusValue *copyAvailabilityStatus(AvailabilityStatusValue *);
int                     compareAvailabilityStatus(AvailabilityStatusValue *,
						  AvailabilityStatusValue *);


#else	/* Old-style C function prototypes */

PE	r_octenc ();
struct qbuf*	r_octsdec ();
int	r_octprint ();
int	strprint ();
char*	hexprint ();
int	pstrcmp ();

int	oidprint ();

PE	time_enc ();
char*	time_dec ();
int	time_cmp ();

PE	dn_enc ();
DN	dn_dec ();

OID	dup_prim2oid ();

PE	bool_enc ();
Bool*	bool_dec ();
Bool*	bool_parse ();
int	bool_print ();

PE	int_enc ();
int*	int_dec ();
int*	int_parse ();
int	int_print ();
int     int_free ();
int*	int_cpy ();
int     int_cmp ();

PE	IntegerList_encode ();
IntegerListVal*	IntegerList_decode ();
int	IntegerList_print ();
int	IntegerList_free ();
int	IntegerListVal_cmp ();
int*	IntegerList_getelem ();
IntegerListVal**	IntegerList_getnext ();

PE	real_enc ();
double*	real_dec ();
double* real_parse ();
int	real_print ();
int     real_free ();
double*	real_cpy ();
int     real_cmp ();

PE	RealList_encode ();
RealListVal*	RealList_decode ();
int	RealList_print ();
int	RealList_free ();
int	RealListVal_cmp ();
double*	RealList_getelem ();
RealListVal**	RealList_getnext ();

StringListVal*	StringList_parse ();
int	StringList_print ();
int	StringList_free ();
int	StringListVal_cmp ();
char*	StringList_getelem ();
StringListVal**	StringList_getnext ();

PE	GraphicStringList_encode ();
StringListVal*	GraphicStringList_decode ();

PE	IA5StringList_encode ();
StringListVal*	IA5StringList_decode ();

PE	PrintableStringList_encode ();
StringListVal*	PrintableStringList_decode ();

PE	SimpleNameType_encode ();
SimpleNameTypeVal*	SimpleNameType_decode ();
SimpleNameTypeVal*	SimpleNameType_parse ();
int	SimpleNameType_print ();
int	SimpleNameType_free ();
int	SimpleNameType_cmp ();

MID		MID_parse ();
int		MID_print ();
int		MID_free ();
int		MID_cmp ();

MIDListVal*     MIDList_parse ();
int             MIDList_print ();
int             MIDList_free ();
int             MIDListVal_cmp ();
MID		MIDList_getelem ();
MIDListVal**    MIDList_getnext ();

PE		ObjectClass_encode ();
MID		ObjectClass_decode ();

PE		ObjectClassList_encode ();
MIDListVal*	ObjectClassList_decode ();

PE		AttributeId_encode ();
MID		AttributeId_decode ();

PE		AttributeIdList_encode ();
MIDListVal*	AttributeIdList_decode ();

PE		EventType_encode ();
MID		EventType_decode ();

PE		ObjectInstance_encode ();
MN		ObjectInstance_decode ();
MN		ObjectInstance_parse ();
int		ObjectInstance_print ();
int		ObjectInstance_free ();
int		ObjectInstance_cmp ();

PE		ObjectInstanceList_encode ();
MNListVal*	ObjectInstanceList_decode ();
MNListVal*	ObjectInstanceList_parse ();
int		ObjectInstanceList_print ();
int		ObjectInstanceList_free ();
int		ObjectInstanceListVal_cmp ();
MN		ObjectInstanceList_getelem ();
MNListVal**	ObjectInstanceList_getnext ();

PE		CMISScope_encode ();
CMISScope*	CMISScope_decode ();
int		CMISScope_print ();

PE		CMISFilter_encode ();
CMISFilter*	CMISFilter_decode ();
int		CMISFilter_print ();

PE		OIDList_encode ();
ObjIdListVal*	OIDList_decode ();
ObjIdListVal*   OIDList_parse ();
int		OIDList_print ();
int		OIDList_free ();
int		OIDListVal_cmp ();
OID		OIDList_getelem ();
ObjIdListVal**	OIDList_getnext ();

PE		ObservedValue_encode ();
ObservedValue*  ObservedValue_decode ();
ObservedValue*  ObservedValue_parse ();
int		ObservedValue_print ();
int		ObservedValue_free ();
ObservedValue*	ObservedValue_cpy ();
int		ObservedValue_cmp ();

PE			CounterThreshold_encode ();
CounterThresholdVal*	CounterThreshold_decode ();
CounterThresholdVal*	CounterThreshold_parse ();
int			CounterThreshold_print ();
int			CounterThreshold_free ();
int			CounterThresholdVal_cmp ();
CounterThresholdVal*	CounterThreshold_getelem ();
CounterThresholdVal**	CounterThreshold_getnext ();

PE			GaugeThreshold_encode ();
GaugeThresholdVal*	GaugeThreshold_decode ();
GaugeThresholdVal*	GaugeThreshold_parse ();
int			GaugeThreshold_print ();
int			GaugeThreshold_free ();
int			GaugeThresholdVal_cmp ();
GaugeThresholdVal*	GaugeThreshold_getelem ();
GaugeThresholdVal**	GaugeThreshold_getnext ();

PE		TideMark_encode ();
TideMarkVal*	TideMark_decode ();
int		TideMark_print ();
int		TideMark_free ();
int		TideMark_cmp ();

PE		AdministrativeState_encode ();
AdminStateVal*	AdministrativeState_decode ();
AdminStateVal*	AdministrativeState_parse ();
int		AdministrativeState_print ();
int		AdministrativeState_free ();
int		AdministrativeState_cmp ();

PE		OperationalState_encode ();
OperStateVal*	OperationalState_decode ();
OperStateVal*	OperationalState_parse ();
int		OperationalState_print ();
int		OperationalState_free ();
int		OperationalState_cmp ();

PE		UsageState_encode ();
struct type_SMI_UsageState*	UsageState_decode ();
int		UsageState_print ();
int		UsageState_free ();
int		UsageState_cmp ();

PE		DestAddress_encode ();
DestAddressVal*	DestAddress_decode ();
int		DestAddress_print ();
int		DestAddress_free ();

PE ObjectInfo_encode ();
struct type_SMI_ObjectInfo* ObjectInfo_decode ();
void ObjectInfo_free();
void ObjectInfo_free_aux();
int ObjectInfo_print();

struct type_SMI_AttrValChangeInfo* AttrValChangeInfo_alloc ();
PE AttrValChangeInfo_encode ();
struct type_SMI_AttrValChangeInfo* AttrValChangeInfo_decode ();
void AttrValChangeInfo_free();
void AttrValChangeInfo_free_aux();
int AttrValChangeInfo_print();

struct type_SMI_AlarmInfo* AlarmInfo_alloc ();
struct type_SMI_AlarmInfo* ThresholdAlarmInfo_alloc ();
struct type_SMI_AlarmInfo* ThresholdAlarmInfo_ProbsAttrs_alloc ();
PE AlarmInfo_encode ();
struct type_SMI_AlarmInfo* AlarmInfo_decode ();
void AlarmInfo_free ();
void AlarmInfo_free_aux ();
int AlarmInfo_print ();

PE                 encodeLogFullAction();
LogFullActionValue *decodeLogFullAction();
LogFullActionValue *parseLogFullAction();
char               *LogFullAction2Str();
int                printLogFullAction();
void               freeLogFullAction(); /* == free() */
LogFullActionValue *copyLogFullAction();
int                compareLogFullAction();

PE                      encodeAvailabilityStatus();
AvailabilityStatusValue *decodeAvailabilityStatus();
AvailabilityStatusValue *parseAvailabilityStatus();
char                    *AvailabilityStatus2Str();
int                     printAvailabilityStatus();
void                    freeAvailabilityStatus();
AvailabilityStatusValue *copyAvailabilityStatus();
int                     compareAvailabilityStatus();

#endif	/* FunctionPrototypes */


void  gms_name_syntaxes ();
void  smi_syntaxes ();


#ifdef	__cplusplus
}
#endif

#endif
