//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Top.h - top managed object class declaration

// By George Pavlou, June 1991


#ifndef _TOP_
#define _TOP_


#include "SMI-types.h"
#include "GenericMO.h"

class ObjectClass;
class ObjectClassList;
class ObjId;
class ObjIdList;


class Top : public MO
{
friend class MO;

private:
    static  int	  _source;
    static  int	  _classLevel;
    static  int	  _attrId;
    static  void* _oldAttrValue;
    static  PE	  _oldAttrValuePE;
    static  char* _addText;
    static  type_SMI_AttrValChangeInfo* _attrValChangeInfo;
    static  type_SMI_ObjectInfo* _objectInfo;
    static  CMISParamListVal*    _objectInfoListCurPos;

    static  MO*	  _testEventRecord;

    static MOClassInfo	_top;

    int	    _objectMgmtMask;		// mask for object/state mgmt SMFs

    int		initialiseSpecialAttrs ();
    static int	processNotification (MO*, OID, int, int,
		OID = NULLOID, MN = NULLMN, char* = NULLCP, PE = NULLPE);

    Top ();				// for the MIB root

protected:

#define TOP_LEVEL	0

#define OBJECTCLASS	0	/* ObjectClass */
#define NAMEBINDING	1	/* NameBinding */
#define PACKAGES	2	/* ObjIdList */
#define ALLOMORPHS	3	/* ObjectClassList */
#define TOPNATTRS	4

#define	ALLOMPACKAGE	0
#define	TOPNPACKAGES	1

#define MO_objectCreation		0
#define MO_objectDeletion		1
#define MO_attributeValueChange		2
#define MO_stateChange			3

    Attr*	_attrs [TOPNATTRS];

    int		addAllomorphicClass (OID);
    int		buildReport (int, int, void*&, Bool&);
    type_SMI_ObjectInfo*
		getObjectInfo ();	// gets attribute OID and PE values


    Top (RDN, MO*);
    Top (RDN, MO*, int, CMISParam*, CMISErrors*, CMISErrorInfo*);
    ~Top ();

public:
    // object/state management notifications

    OID		getObjectManagementEvent (int);

    int		triggerObjectCreation
			(int = int_SMI_SourceIndicator_resourceOperation,
			 char* = NULLCP);
    int		triggerObjectDeletion
			(int = int_SMI_SourceIndicator_resourceOperation,
			 char* = NULLCP);
    int		triggerAttributeValueChange (int, int, void* = NULLVD,
			int = int_SMI_SourceIndicator_resourceOperation,
			char* = NULLCP);
    int		triggerAttributeValueChange (char*, void* = NULLVD,
			int = int_SMI_SourceIndicator_resourceOperation,
			char* = NULLCP);
    int		triggerStateChange (int, int, void* = NULLVD,
			int = int_SMI_SourceIndicator_resourceOperation,
			char* = NULLCP);
    int		triggerStateChange (char*, void* = NULLVD,
			int = int_SMI_SourceIndicator_resourceOperation,
			char* = NULLCP);

    // general notifications

    int		triggerEvent (OID);
    int		triggerEvent (char*);
    int		triggerEvent (int, int);

    static int	relayEventReport (OID, MN, char*, OID, PE);
    static void	checkEFDsOnDisconnect (int);

    static int	initialiseClass ();
    static MOClassInfo* getClassInfo ();

    ObjectClass*	objectClass ();
    ObjId*		nameBinding ();
    ObjIdList*		packages ();
    ObjectClassList*	allomorphs ();
};


inline MOClassInfo* Top::getClassInfo ()
{ return &_top; }


inline ObjectClass*  Top::objectClass ()
{ return (ObjectClass*) _attrs[OBJECTCLASS]; }


inline ObjId*  Top::nameBinding ()
{ return (ObjId*) _attrs[NAMEBINDING]; }


inline ObjIdList*  Top::packages ()
{ return (ObjIdList*) _attrs[PACKAGES]; }


inline ObjectClassList*	Top::allomorphs ()
{ return (ObjectClassList*) _attrs[ALLOMORPHS]; }


inline int  Top::triggerEvent (OID eventType)
{ int eventId, classLevel; findEvent(eventType, &eventId, &classLevel);
  return eventId != NOTOK ?
  processNotification(this, eventType, eventId, classLevel) : NOTOK; }


inline int  Top::triggerEvent (char* eventName)
{ OID eventType = findEvent(eventName);
  return eventType ? triggerEvent(eventType) : NOTOK; }


inline int  Top::triggerEvent (int eventId, int classLevel)
{ OID eventType = findEvent(eventId, classLevel);
  return eventType ?
  processNotification(this, eventType, eventId, classLevel) : NOTOK; }


inline int  Top::relayEventReport (OID moClass, MN moInst,
				char* eventTime, OID eventType, PE eventInfo)
{ return processNotification(NULLMO, eventType, -1, -1,
				moClass, moInst, eventTime, eventInfo); }


inline int  Top::triggerObjectCreation (int source, char* addText)
{ _source = source; _addText = addText;
  OID oid = getObjectManagementEvent(MO_objectCreation);
  return oid ? processNotification(this, oid,
			MO_objectCreation, TOP_LEVEL) : NOTOK; }


inline int  Top::triggerObjectDeletion (int source, char* addText)
{ _source = source; _addText = addText;
  OID oid = getObjectManagementEvent(MO_objectDeletion);
  return oid ? processNotification(this, oid,
			MO_objectDeletion, TOP_LEVEL) : NOTOK; }


inline int  Top::triggerAttributeValueChange (int attrId, int classLevel,
				void* oldValue, int source, char* addText)
{ _classLevel = classLevel; _attrId = attrId; _oldAttrValue = oldValue;
  _source = source; _addText = addText;
  OID oid = getObjectManagementEvent(MO_attributeValueChange);
  return oid ? processNotification(this, oid,
			MO_attributeValueChange, TOP_LEVEL) : NOTOK; }


inline int  Top::triggerAttributeValueChange (char* attrName,
				void* oldValue, int source, char* addText)
{ int attrId = -1, classLevel; findAttr(attrName, &attrId, &classLevel);
  return attrId >= 0 ?
  triggerAttributeValueChange(attrId, classLevel, oldValue, source, addText)
  : NOTOK; }


inline int  Top::triggerStateChange (int attrId, int classLevel,
				void* oldValue, int source, char* addText)
{ _classLevel = classLevel; _attrId = attrId; _oldAttrValue = oldValue;
  _source = source; _addText = addText;
  OID oid = getObjectManagementEvent(MO_stateChange);
  return oid ? processNotification(this, oid, MO_stateChange, TOP_LEVEL)
  : NOTOK; }


inline int  Top::triggerStateChange (char* attrName,
				void* oldValue, int source, char* addText)
{ int attrId = -1, classLevel; findAttr(attrName, &attrId, &classLevel);
  return attrId >= 0 ?
  triggerStateChange(attrId, classLevel, oldValue, source, addText) : NOTOK; }


#endif
