/*
 * Copyright (c) 1994 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
** Pretty-Print MSAP structures
*/

/*
** Saleem N. Bhatti
** March 1992, updated April 1994
*/

#if !defined(_msap_pp_h_)
#define _msap_pp_h_

#if !defined(_a_)
#if defined(__STDC__) || defined(__cplusplus)
#define _a_(_args_) _args_
#else
#define _a_(_args_) ()
#endif
#endif

#include "msap.h"

#if defined(__cplusplus)
extern "C" {
#endif

/* Unused as yet ... */
extern unsigned long G_MSAP_PP_flags; /* used by msap */

#define MSAP_PP_ALL            (unsigned long) 0xffffffff
#define MSAP_PP_GetArg         (unsigned long) 0x00000001
#define MSAP_PP_SetArg         (unsigned long) 0x00000002
#define MSAP_PP_CreateArg      (unsigned long) 0x00000004
#define MSAP_PP_DeleteArg      (unsigned long) 0x00000008
#define MSAP_PP_ActionArg      (unsigned long) 0x00000010
#define MSAP_PP_CancelGet      (unsigned long) 0x00000020
#define MSAP_PP_GetRes         (unsigned long) 0x00000040
#define MSAP_PP_SetRes         (unsigned long) 0x00000080
#define MSAP_PP_CreateRes      (unsigned long) 0x00000100
#define MSAP_PP_DeleteRes      (unsigned long) 0x00000200
#define MSAP_PP_ActionRes      (unsigned long) 0x00000400
#define MSAP_PP_EventReportArg (unsigned long) 0x00000800
#define MSAP_PP_EventReportRes (unsigned long) 0x00001000
#define MSAP_PP_Indication     (unsigned long) 0x80000000

/*
** The following variables are all NULL terminated arrays of strings. Each
** array contains "nice" names for the various data values. The data-values
** are used as indexes for the arrays. The names of the data-types or data-type
** members whose values are used are given in the comments.
*/
extern
#if defined(__STDC__) || defined(__cplusplus)
const
#endif
char **CMISModifyOpStr,  /* CMISModifyOp */
     **CMISSyncStr,      /* CMISSync */
     **CMISErrorsStr,    /* CMISErrors */
     **MID_TypeStr,      /* MIDentifier.mid_type */
     **MN_TypeStr,       /* MName.mn_type */
     **SC_TypeStr,       /* CMISScope.sc_type */
     **FI_SubTypeStr,    /* CMISFilterSubstrings.fi_sub_type */
     **FI_TypeStr,       /* CMISFilterItem.fi_type */
     **FLT_TypeStr,      /* CMISFilter.flt_type */
     **CA_InstTypeStr,   /* CMISCreateArg.ca_inst_type */
     **LR_TypeStr,       /* CMISLinkedReplyArg.lr_type */
     **MA_SourceStr,     /* MSAPAbort.ma_source */
     **MA_ReasonStr,     /* MSAPAbort.ma_reason */
     **MI_TypeStr;       /* MSAPIndication.mi_type */

int ppSetIndentSize _a_((int)); /* returns previous indent size */
FILE *ppSetOutputFile _a_((FILE *)); /* returns previous FILE */
void ppCMISModifyOp _a_((CMISModifyOp));
void ppCMISSync _a_((CMISSync));
void ppCMISErrors _a_((CMISErrors));
void ppMIDentifier _a_((struct MIDentifier *));
/* void ppMIDListVal _a_((MIDListVal *)); */
void ppMName _a_((MName *));
void ppCMISParam _a_((CMISParam *));
void ppCMISGetAttr _a_((CMISGetAttr *));
void ppCMISSetAttr _a_((CMISSetAttr *));
void ppCMISScope _a_((CMISScope *));
void ppCMISFilterSubstrings _a_((CMISFilterSubstrings *));
void ppCMISFilterItem _a_((CMISFilterItem *));
void ppCMISFilter _a_((CMISFilter *));
void ppCMISEventReportArg _a_((CMISEventReportArg *));
void ppCMISEventReportRes _a_((CMISEventReportRes *));
void ppCMISGetArg _a_((CMISGetArg *));
void ppCMISGetRes _a_((CMISGetRes *));
void ppCMISSetArg _a_((CMISSetArg *));
void ppCMISSetRes _a_((CMISSetRes *));
void ppCMISActionArg _a_((CMISActionArg *));
void ppCMISActionRes _a_((CMISActionRes *));
void ppCMISCreateArg _a_((CMISCreateArg *));
void ppCMISCreateRes _a_((CMISCreateRes *));
void ppCMISDeleteArg _a_((CMISDeleteArg *));
void ppCMISDeleteRes _a_((CMISDeleteRes *));
void ppCMISCancelGetArg _a_((CMISCancelGetArg *));
void ppCMISNoSuchArgument _a_((CMISNoSuchArgument *));
void ppCMISInvalidArgumentValue _a_((CMISInvalidArgumentValue *));
void ppCMISBaseManagedObjectId _a_((CMISBaseManagedObjectId *));
void ppCMISProcessingFailure _a_((CMISProcessingFailure *));
void ppCMISComplextityLimitation _a_((CMISComplexityLimitation *));
void ppCMISActionError _a_((CMISActionError *));
void ppCMISDeleteError _a_((CMISDeleteError *));
void ppCMISLinkedReplyArg _a_((CMISLinkedReplyArg *));
void ppCMISErrorInfo _a_((CMISErrors, CMISErrorInfo *));
void ppAEInfo _a_((AEInfo *));
void ppMSAPAssocInfo _a_((MSAPAssocInfo *));
void ppMSAPConnect _a_((MSAPConnect *));
void ppMSAPStart _a_((MSAPStart *));
void ppMSAPTerm _a_((MSAPTerm *));
void ppMSAPUReject _a_((MSAPUReject *));
void ppMSAPPReject _a_((MSAPPReject *));
void ppMSAPAbort _a_((MSAPAbort *));
void ppMSAPError _a_((CMISErrors, MSAPError *));
void ppMSAPEventReportInv _a_((MSAPEventReportInv *));
void ppMSAPEventReportRes _a_((MSAPEventReportRes *));
void ppMSAPLinkedReplyInv _a_((MSAPLinkedReplyInv *));
void ppMSAPGetInv _a_((MSAPGetInv *));
void ppMSAPGetRes _a_((MSAPGetRes *));
void ppMSAPSetInv _a_((MSAPSetInv *));
void ppMSAPSetRes _a_((MSAPSetRes *));
void ppMSAPActionInv _a_((MSAPActionInv *));
void ppMSAPActionRes _a_((MSAPActionRes *));
void ppMSAPCreateInv _a_((MSAPCreateInv *));
void ppMSAPCreateRes _a_((MSAPCreateRes *));
void ppMSAPDeleteInv _a_((MSAPDeleteInv *));
void ppMSAPDeleteRes _a_((MSAPDeleteRes *));
void ppMSAPCancelGetInv _a_((MSAPCancelGetInv *));
void ppMSAPEmptyRes _a_((MSAPEmptyRes *));
void ppMSAPIndication _a_((MSAPIndication *));

#if defined(__cplusplus)
}
#endif

#endif /* !_msap_pp_h_ */
