//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// AVA.h - Attribute Value Assertion (+ a bit more...)

// By George Pavlou, July 1993


#ifndef _AVA_
#define _AVA_


#ifndef _GENERIC_ATTR_
#include "GenericAttr.h"
#endif

#ifndef _GEN_ARRAY_
#include "GenArray.h"
#endif


class AVA
{
    static Bool	 _createError;

    char*	 _name;
    OID		 _oid;
    Attr*	 _value;
    CMISErrors	 _error;
    CMISModifyOp _modify;

    void	 nullify ();

public:
    static Bool	 createError ();

    char*	 getName ();
    OID		 getOid ();
    Attr*	 getValue ();
    CMISErrors	 getError ();
    char*	 getErrorStr ();
    CMISModifyOp getModifyOp ();
    char*	 print ();		// result should be free'd
    void	 clear ();
    int		 adjustActionOid ();	// remove the Info/Reply suffix

    AVA (char*, void*, CMISModifyOp = m_noModifyOp, CMISErrors = m_noError);
    AVA (char*, char*, CMISModifyOp = m_noModifyOp, CMISErrors = m_noError);
    AVA (char*, Attr*, CMISModifyOp = m_noModifyOp, CMISErrors = m_noError);
    AVA (CMISErrors, char*, Attr* = NULLATTR);
    AVA (OID, Attr*, CMISModifyOp = m_noModifyOp, CMISErrors = m_noError);
    AVA (CMISErrors, OID, Attr* = NULLATTR);
    AVA (OID, PE, CMISErrors = m_noError, CMISModifyOp = m_noModifyOp);
    AVA (CMISErrors = m_noError);
    ~AVA ();
};

#if	defined(NULLAVA)
#undef	NULLAVA
#endif
#define NULLAVA	((AVA*) 0)

extern	AVA*& nullAvaRef;

#define NULLAVAREF (nullAvaRef)
#define DISCARDAVA NULLAVAREF


class AVAArray : public Array
{
protected:
    void	deleteElem (void*);

public:
    AVAArray (int);
};

#define NULLAVAARRAY	((AVAArray*) 0)




// AVA class inline methods

inline Bool  AVA::createError ()
{ return _createError; }

inline char*  AVA::getName ()
{ return _name ? _name : sprintoid(_oid); }

inline OID  AVA::getOid ()
{ return _oid; }

inline Attr*  AVA::getValue ()
{ return _value; }

inline CMISErrors  AVA::getError ()
{ return _error; }

inline char*  AVA::getErrorStr ()
{ return MErrString(_error); }

inline CMISModifyOp  AVA::getModifyOp ()
{ return _modify; }

inline int  AVA::adjustActionOid ()
{ if (_oid) --_oid->oid_nelem; return _oid ? OK : NOTOK; }

inline void  AVA::nullify ()
{ _createError = False; _name = NULLCP; _oid = NULLOID; _value = NULLATTR;
  _error = m_noError; _modify = m_noModifyOp; }

inline  AVA::AVA (OID oid, Attr* value, CMISModifyOp modify, CMISErrors error)
{ nullify (); _modify = modify; _oid = oid; _value = value; _error = error; }

inline  AVA::AVA (CMISErrors error, OID oid, Attr* value)
{ nullify (); _oid = oid; _value = value; _error = error; }

inline  AVA::AVA (CMISErrors error)
{ nullify (); _error = error; }

inline  AVA::~AVA ()
{ clear(); }



// AVAArray class inline methods

inline void  AVAArray::deleteElem (void* v)
{ delete (AVA*) v; }

inline AVAArray::AVAArray (int n) : Array (n, True)
{}


#endif
