//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// AnyType.h - generic abstract syntax type using oidtables

// By George Pavlou, December 1991


#ifndef _ANY_TYPE_
#define _ANY_TYPE_


#ifndef _GENERIC_ATTR_
#include "GenericAttr.h"
#endif


class AnyType : public Attr
{
    static Bool	_createError;
    sntx_table* _syntax;

    int		setup (short, void*);
    int		setup (short, char*);
    int		setup (short, PE);

protected:
    PE		_encode ();
    void*	_decode (PE);
    void	_free ();
    void*	_parse (char*);
    char*	_print ();
    int		_compare (void*, void*);
    void*	_getelem (void*);
    void**	_getnext (void*);

public:
    static Bool	createError ();
    char*	getSyntax ();

    AnyType (OID, PE);
    AnyType (char*, void*);
    AnyType (short, void*);
    AnyType (char*, char*);
    AnyType (short, char*);
    AnyType ();
};


// inline methods

inline PE  AnyType::_encode ()
{ ENCODE_FNX enc_fnx = _syntax->s_encode;
  return enc_fnx ? (PE) enc_fnx(getval()) : NULLPE; }

inline void*  AnyType::_decode (PE pe)
{ DECODE_FNX dec_fnx = _syntax->s_decode;
  return dec_fnx ? (void*) dec_fnx(pe) : NULLVD; }

inline void  AnyType::_free ()
{ FREE_FNX free_fnx = _syntax->s_free;
  if (free_fnx) free_fnx(getval()); }

inline void*  AnyType::_parse (char* strval)
{ PARSE_FNX parse_fnx = _syntax->s_parse;
  return parse_fnx ? (void*) parse_fnx(strval) : NULLVD; }

inline char*  AnyType::_print ()
{ PRINT_FNX prnt_fnx = _syntax->s_print;
  return prnt_fnx ? attrv2str(getval(),prnt_fnx) : NULLCP; }

inline int  AnyType::_compare (void* v1, void* v2)
{ COMPARE_FNX cmp_fnx = _syntax->s_compare;
  return cmp_fnx ? cmp_fnx(v1, v2) : Attr::_compare(v1, v2); }

inline void*  AnyType::_getelem (void* v)
{ GETELEM_FNX getelem_fnx = _syntax->s_pe_print;
  return getelem_fnx ? getelem_fnx(v) : NULLVD; }

inline void**  AnyType::_getnext (void* v)
{ GETNEXT_FNX getnext_fnx = _syntax->s_approx;
  return getnext_fnx ? getnext_fnx(v) : NULL; }

inline Bool  AnyType::createError ()
{ return _createError; }

inline char*  AnyType::getSyntax ()
{ return _syntax ? _syntax->s_sntx : NULLCP; }

inline  AnyType::AnyType (short sntxIndex, void* attrVal)
{ _createError = False;
  if (setup(sntxIndex, attrVal) == NOTOK) _createError = True; }

inline  AnyType::AnyType (short sntxIndex, char* attrVal)
{ _createError = False;
  if (setup(sntxIndex, attrVal) == NOTOK) _createError = True; }

inline  AnyType::AnyType ()
{ _createError = False; _syntax = NULL; }


#endif
