//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// GenericAttr.h - generic attribute declaration

// By George Pavlou, December 1991


#ifndef _GENERIC_ATTR_
#define _GENERIC_ATTR_


#ifndef _BOOLEAN_
#include "Boolean.h"
#endif

#if	!defined(NULLVD)
#define	NULLVD	((void*) 0)
#endif


class Attr
{
    void*	_val;
    PE		_pe;

    void		init ();
    void*		replvaltmp (void*);
    void*		replvalback (void*);

protected:
    // methods related to the ASN.1 syntax to be provided in derived classes

    virtual PE		_encode ();
    virtual void*	_decode (PE);
    virtual void	_free ();
    virtual void*	_parse (char*);
    virtual char*	_print ();
    virtual void*	_copy ();
    virtual int		_compare (void*, void*);
    virtual void*	_getelem (void*);
    virtual void**	_getnext (void*);

    // method that finds an element in a set-valued attribute

    void**		_find (void*, void**);

    // protected constructor (abstract class)

    Attr ();

public:
    // methods to print, free, copy, compare and manipulate ASN.1 values

    PE			encode ();
    PE			encode (void*);
    void*		decode (PE);
    void*		parse (char*);
    void		ffree ();
    void		ffree (void*);
    char*		print ();
    char*		print (void*);
    void*		copy ();
    void*		copy (void*);
    int			compare (void*);
    int			compare (void*, void*);
    void*		find (void*);
    void*		find (void*, void*);
    void*		getElem (void*);
    void*		getNext (void*);
    char*		asnMemoryDump (int*);
    char*		asnMemoryDump (void*, int*);
    int			asnFileDump (FILE*);
    int			asnFileDump (void*, FILE*);

    Bool		isMultiValued ();
    virtual char*	getSyntax ();	// returns the syntax name


    // methods that enable to manipulate the contained attribute value

    void*		getval ();
    void		setval (void*);
    void		replval (void*);
    int			setstr (char*);		// uses _parse

    void		replace (void*);	// the same as replval -
						// for backwards compatibility
    void		freePE ();		// may be used to save space
    PE			getPE ();		// not to be used!
    void		setPE (PE);		//  ..

    // methods that may be redefined in derived classes to associate behaviour

    virtual void*	get ();
    virtual int		set (void*);
    virtual int		setDefault (void*);
    virtual int		add (void*);
    virtual int		remove (void*);

    // methods to respond to CMIS operations

    CMISErrors		get (PE*);
    CMISErrors		set (PE);

    CMISErrors		replace (PE);		// the same as set -
						// for backwards compatibility
    virtual Bool	filter (int, void*);
    virtual Bool	filterSubstring (int, void*, Bool);
    Bool		filter (CMISFilter*);

    // destructing methods

    void		clear ();		// frees all contained data
    virtual ~Attr ();
};

#if	defined(NULLATTR)
#undef	NULLATTR
#endif
#define	NULLATTR	((Attr*) 0)


// inline methods

inline void*  Attr::replvaltmp (void* tmp)
{ void* prev = _val; _val = tmp; return prev; }

inline void*  Attr::replvalback (void* prev)
{ void* tmp = _val; _val = prev; return tmp; }

inline PE  Attr::_encode ()
{ return NULLPE; }

inline void*  Attr::_decode (PE)
{ return NULLVD; }

inline void  Attr::_free ()
{}

inline void*  Attr::_parse (char*)
{ return NULLVD; }

inline char*  Attr::_print ()
{ return NULLCP; }

inline int  Attr::_compare (void*, void*)
{ return 2; }

inline void*  Attr::_getelem (void*)
{ return NULL; }

inline void**  Attr::_getnext (void*)
{ return NULL; }

inline void  Attr::init ()
{ _val = NULLVD; _pe = NULLPE; }

inline void* Attr::getval ()
{ if (!_val && _pe) _val = _decode(_pe); return _val; }

inline void  Attr::setval (void* newval)
{ clear(); _val = newval; }

inline void  Attr::replval (void* newval)
{ if (_pe) { pe_free(_pe); _pe = NULLPE; } _val = newval; }

inline void*  Attr::_copy ()
{ encode(); return _pe ? _decode(_pe) : NULLVD; }

inline PE  Attr::encode (void* val)
{ void* prev = replvaltmp(val); PE pe = _encode(); replvalback(prev);
  return pe; }

inline void* Attr::decode (PE val)
{ return _decode(val); }

inline void* Attr::parse (char* strval)
{ return _parse(strval); }

inline void  Attr::ffree ()
{ clear(); }

inline void  Attr::ffree (void* val)
{ if (val) { void* prev = _val; _val = val; _free(); _val = prev; } }

inline char*  Attr::print ()
{ return _print(); }

inline char*  Attr::print (void* val)
{ void* prev = replvaltmp(val); char* cp = print(); replvalback(prev);
  return cp; }

inline void*  Attr::copy ()
{ return _copy(); }

inline void*  Attr::copy (void* val)
{ void* prev = replvaltmp(val); void* cpy = copy(); replvalback(prev);
  return cpy; }

inline int  Attr::compare (void* v)
{ return _compare(v, getval()); }

inline int  Attr::compare (void* v1, void* v2)
{ void* prev = replvaltmp(v2); int cmp = compare(v1); replvalback(prev);
  return cmp; }

inline void*  Attr::find (void* elem)
{ void** v = _find(elem, &_val); return v ? *v : NULLVD; }

inline void*  Attr::find (void* elem, void* list)
{ void *v, *prev = replvaltmp(list); v = find(elem); replvalback(prev);
  return v; }

inline void*  Attr::getElem (void* elem)
{ return _getelem(elem); }

inline void*  Attr::getNext (void* elem)
{ void** v = _getnext(elem); return v ? *v : NULLVD; }

inline char*  Attr::asnMemoryDump (int* len)
{ char* s; int r = pe2mem(encode(), &s, len); return r == OK ? s : NULLCP; }

inline char*  Attr::asnMemoryDump (void* val, int* len)
{ void* prev = replvaltmp(val); char* cp = asnMemoryDump(len);
  replvalback(prev); return cp; }

inline int  Attr::asnFileDump (FILE* file)
{ return pe2stream(encode(), file); }

inline int  Attr::asnFileDump (void* val, FILE* file)
{ void* prev = replvaltmp(val); int ret = asnFileDump(file);
  replvalback(prev); return ret; }

inline Bool  Attr::isMultiValued ()
{ return (_val == NULL || _getnext(_val) != NULL) ? True : False; }

inline void*  Attr::get ()
{ return getval(); }

inline int  Attr::set (void* newval)
{ setval(newval); return OK; }

inline void  Attr::replace (void* newval)
{ replval(newval); }

inline int  Attr::setDefault (void* dfltval)
{ set(dfltval); return OK; }

inline PE  Attr::getPE ()
{ return _pe; }

inline void  Attr::setPE (PE pe)
{ clear(); _pe = pe; }

inline void  Attr::freePE ()
{ if (_pe) pe_free(_pe); _pe = NULLPE; }

inline char*  Attr::getSyntax ()
{ return ""; }

inline CMISErrors  Attr::get (PE* attrval)
{ get(); return (*attrval = encode()) ? m_noError : m_processingFailure; }

inline CMISErrors  Attr::replace (PE attrval)
{ return set(attrval); }

inline Bool  Attr::filterSubstring (int, void*, Bool)
{ return False; }

inline  Attr::Attr ()
{ init(); }


#endif
