//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation, 
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// TkCoord.cc - Tk coordinator realisation

// By George Pavlou, Fernando Goncalves - June 1993


#include <stdlib.h>
#include <signal.h>
#include "TkCoord.h"


TkCoordinator::TkCoordinator ()
{
    _timerSet = False;
    setAlarmMode (False);	// default: Tk controls timers
}


int  TkCoordinator::commEndpointMaskChange (int fd, int mode)
{
    switch (mode) {
    case MASK_ADD:
	Tk_CreateFileHandler (fd, TK_READABLE,
			TkCoordinator::TkReadCommEndpoint, (ClientData) fd);
	break;

    case MASK_REMOVE:
	Tk_DeleteFileHandler (fd);
	break;
    }

    return OK;
}


int  TkCoordinator::nextWakeUpIntvlChange (long intvl)
{
    if (_timerSet) {
	_timerSet = False;
	Tk_DeleteTimerHandler (_token);
    }

    if (intvl == 0)
	return OK;	// no more OSIMIS alarms (for the moment)

    _timerSet = True;
    _token = Tk_CreateTimerHandler ((int) intvl*1000,
					TkCoordinator::TkWakeUp, NULL);
    return OK;
}


void  TkCoordinator::TkReadCommEndpoint (ClientData fd, int)
{
    int oldmask = sigblock (0x7fffffff);	// block signals
    ((TkCoordinator*) getInstance ())
		-> _ks[(int) fd] -> readCommEndpoint((int) fd);
    sigsetmask (oldmask);			// restore them
}


void  TkCoordinator::TkWakeUp (ClientData)
{
    Coordinator::serviceAlarm (OK);
}
