/* Copyright (c) 1994 Institute of Computer Science, Foundation of      */
/* Research and Technology - Hellas (ICS-FORTH)                         */ 
/* All rights reserved.                                                 */      

/* Redistribution and use in source and binary forms are permitted      */
/* provided that the above copyright notice and this paragraph are      */
/* duplicated in all such forms and that any documentation, advertising */
/* materials, and other materials related to such distribution and use  */
/* acknowledge that the software was developed by the Institute of      */
/* Computer Science, Foundation of Research and Technology - Hellas.    */
/* The name of the Institute or the Foundation may not be used to       */
/* endorse or promote products derived from this software without       */
/* specific prior written permission.                                   */
/* THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR       */
/* IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED       */
/* WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.  */

/* initialise the environment to use location transparency */
/* also, bind to the Directory 				   */

/* Costas Stathopoulos July 1993 */

#include "isode/quipu/util.h"
#include "isode/quipu/common.h"
#include "isode/quipu/entry.h"
#ifndef ICR1
#include "isode/tailor.h"
#else
#include "isode/compat/tailor.h"
#endif


#ifndef _LOCATION_TRANSPARENCY_
#include "lt.h"
#endif

/* various definitions */
#define OSIMISETCPATH           "OSIMISETCPATH"
#define DEFAULT_dsaptailor      "dsaptailor"

/* The DN of the user to bind to the directory and his password */
char           dir_user[STRINGLEN];
char           dir_passwd[16];
/* The part of the local DIT where we register the management processes */
char           lt_local_DIT[STRINGLEN];

extern char     * dsa_address, * myname;

char	initialised = FALSE;

int lt_init()
{
  char Read_in_Stuff[STRINGLEN];
  char osimis_dsaptailor[STRINGLEN];
  FILE *config_file;
  int nameFlag = 0, passwdFlag = 0, localDITFlag = 0;
  char          *p,
                *TidyString(),
                *part1,
                *part2,
		*osimisetcpath;

  if (initialised == TRUE) return (0);
  quipu_syntaxes();

  /* the oid table is already loaded from OSIMIS */

  /* read dsaptailor file in the OSIMIS area to get address */

  osimisetcpath = malloc (STRINGLEN);
  if ((osimisetcpath = getenv(OSIMISETCPATH)) == NULL) {
    fprintf ( stderr, "lt_init: Cannot get value for the environment variable %s\n", OSIMISETCPATH);
    return(1);
    }
  sprintf(osimis_dsaptailor, "%s/%s", osimisetcpath, DEFAULT_dsaptailor);
  dsa_address = NULLCP;

  if((config_file = fopen(osimis_dsaptailor, "r")) == (FILE *) NULL) {
    fprintf ( stderr, "lt_init: Cannot open dsaptailor file %s... \n", osimis_dsaptailor);
    return(1);
  }
  else {
    while(fgets(Read_in_Stuff, sizeof(Read_in_Stuff), config_file) != NULLCP)
      if ((*Read_in_Stuff != '#') && (*Read_in_Stuff != '\n') )
        (void) tai_string (Read_in_Stuff);
      (void) fclose(config_file);
   }

   /* Now try to get the Directory user, his passwd and the local  	*/
   /* Management Processes DN in the DIT from the dsaptailor 		*/

   if ((config_file = fopen(osimis_dsaptailor, "r")) == (FILE *) NULL) {
     /* This is already checked but... */
     fprintf ( stderr, "lt_init: Cannot open dsaptailor file %s... \n", osimis_dsaptailor);
     return(1);
   }
   while(fgets(Read_in_Stuff, (sizeof(Read_in_Stuff)), config_file) != NULLCP) {
      p = SkipSpace (Read_in_Stuff);
      if (( *p == '#') || (*p == '\0'))
        continue;  /* ignore comments and blanks */

      part1 = malloc (32); /* enough for every dsaptailor parameter */
      if (sscanf(p, "%s", part1) != 1)  
	continue;

      part2 = SkipSpace (p + strlen(part1) + 1);

      /* clean the string */
      part2 = TidyString (part2);
    
      if (lexequ (part1, "lt_manager") == 0) {
        if ( str2dn(*part2 != '@' ? part2 : ++part2) == NULLDN) {
          fprintf(stderr, "lt_init:Invalid DN for username: %s\n",part2);
          return (1);
        }
	(void) strcpy (dir_user, part2);
        nameFlag = 1;
      }
      else if (lexequ (part1, "lt_passwd") == 0) {
        (void) strcpy (dir_passwd, part2);
        passwdFlag = 1;
      }
      else if (lexequ (part1, "lt_local_DIT") == 0) {
        if ( str2dn(*part2 != '@' ? part2 : ++part2) == NULLDN) {
          fprintf(stderr, "lt_init:Invalid DN for lt_local_DIT: %s\n",part2);
          return (1);
        }
        (void) strcpy (lt_local_DIT, part2);
        localDITFlag = 1;
      }
      if ( nameFlag && passwdFlag && localDITFlag )
        break;
    }
    if ( !( nameFlag && passwdFlag && localDITFlag ) ) {
      fprintf(stderr, "lt_init: Error while reading %s:\n", osimis_dsaptailor);
      if (!nameFlag) fprintf(stderr, "lt_init: lt_manager not found.\n");
      if (!passwdFlag) fprintf(stderr, "lt_init: lt_passwd not found.\n");
      if (!localDITFlag) fprintf(stderr, "lt_init: lt_local_DIT not found.\n");
      return(1);
    }
  initialised = TRUE;
  return(0);
}

char *lt_bind()
{

  struct ds_bind_arg bindarg;
  struct ds_bind_arg bindresult;
  struct ds_bind_error binderr;

  /* set dsa_address */
  if (dsa_address == NULLCP) {
    advise(NULLCP, "dsa_address is NULLCP...\n");
    dsa_address = myname;
  }

  if (dsa_address == NULLCP || *dsa_address == '\0') {
    return("Cannot find name of DSA to bind to!");
  }

  /* now bind */
  bindarg.dba_version = DBA_VERSION_V1988;
  if (dir_passwd[0] == 0) {
    bindarg.dba_passwd_len = 0;
    bindarg.dba_passwd [0] = '\0';
  } else {
    bindarg.dba_passwd_len = strlen(dir_passwd);
    (void) strcpy (bindarg.dba_passwd, dir_passwd);
  }

  bindarg.dba_dn = (dir_user[0] == 0 ? NULLDN: str2dn(dir_user));

  bindarg.dba_auth_type = DBA_AUTH_SIMPLE;

  if (ds_bind (&bindarg, &binderr, &bindresult) != DS_OK)
    return(binderr.dbe_type == DBE_TYPE_SECURITY?
           "Security error - Check name and pasword.\n":
           "Service error - Can't connect to directory!\n");
}
