/* Copyright (c) 1994 Institute of Computer Science, Foundation of      */
/* Research and Technology - Hellas (ICS-FORTH)                         */ 
/* All rights reserved.                                                 */      

/* Redistribution and use in source and binary forms are permitted      */
/* provided that the above copyright notice and this paragraph are      */
/* duplicated in all such forms and that any documentation, advertising */
/* materials, and other materials related to such distribution and use  */
/* acknowledge that the software was developed by the Institute of      */
/* Computer Science, Foundation of Research and Technology - Hellas.    */
/* The name of the Institute or the Foundation may not be used to       */
/* endorse or promote products derived from this software without       */
/* specific prior written permission.                                   */
/* THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR       */
/* IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED       */
/* WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.  */

/* This is how to start a location transparent management server (agent)*/
/* on the multiple stacks supported from the ISODE			*/

/* Written by Costas Stathopoulos, Aug 1993 */

#include <stdio.h>

#ifndef ICR1
#include "isode/tailor.h"
#include "isode/tsap.h"
#else
#include "isode/compat/tailor.h"
#include "isode/ll/tsap.h"
#endif

#ifdef  TCP
#ifndef _INTERNET_
#ifndef ICR1
#include "isode/internet.h"
#else
#include "isode/ll/internet.h"
#endif
#endif
#endif

#ifdef  X25
#ifndef _X25_
#ifndef ICR1
#include "isode/x25.h"
#else
#include "isode/ll/x25.h"
#endif
#endif
#define COTS
#endif

#ifdef  TP4
extern char * local_nsap;
#ifndef _TP4_
#ifndef ICR1
#include "isode/tp4.h"
#else
#include "isode/ll/tp4.h"
#endif
#endif
#if     !defined(CLTS) && !defined(COTS)
#define COTS
#endif
#endif

struct PSAPaddr *pa_mark;
/* the next array holds the list of presentation 
	addresses where the agent listens	*/
struct PSAPaddr listen_pa[FD_SETSIZE];

#ifdef	LT_HACK
extern int lt_port;
#else
#include <sys/time.h>

/* generate a random port number in the range BASEPORT - BASEPORT+1024 */
int get_tcpport()
{
#define BASEPORT        11000           /* use ports over 11000 */
#define PORTMASK        0x000003ff      /* ten bits mask (for 1024 ports) */

    int seed;
    int rand_port;
    struct timeval  tval;
    struct timezone tzone;

    gettimeofday(&tval, &tzone);
    seed = tval.tv_usec / 1000;

    srand(seed);
    rand_port = BASEPORT + (rand() & PORTMASK);
    fprintf(stderr, "The port number is %d\n", rand_port);
    return(rand_port);
}
#endif


int lt_server_init(argc, argv, pa, td)
int argc;
char ** argv;
struct PSAPaddr *pa; 
struct TSAPdisconnect *td;
{

    struct PSAPaddr *pa_tmp;

    struct PSAPaddr *listen_pa_tmp;
    int started = 0;

#ifndef	LT_HACK
    int lt_attempts = 0;
#endif

    bzero ((char *) listen_pa, sizeof listen_pa);
    listen_pa_tmp = listen_pa;

    for (pa_tmp = pa; pa_tmp < pa_mark; pa_tmp++) {
#ifdef	LT_HACK
      if (iserver_initAux (argc, argv, pa_tmp, NULL, NULL, 0, td) == NOTOK) {
        advise(NULLCP,"could not start on presentation address: %s\n", pa2str(pa_tmp));
        continue;
      }
      started++;
      if ( pa_tmp->pa_addr.sa_addr.ta_addrs[0].na_stack == NA_TCP )
        pa_tmp->pa_addr.sa_addr.ta_addrs[0].na_port = lt_port;
      fprintf(stderr, "listening on presentation address : %s\n", pa2str(pa_tmp));
      *listen_pa_tmp = *pa_tmp; /* struct copy */
      listen_pa_tmp++;
#else
      for ( lt_attempts = 0 ; 
            iserver_initAux (argc, argv, pa_tmp, NULL, NULL, 0, td) == NOTOK; 
	    lt_attempts++ ) {
        advise(NULLCP,"Could not start on presentation address: %s\n", pa2str(pa_tmp));
        if ( (td -> td_reason == DR_CONGEST) && (lt_attempts < 10) &&
	     (pa_tmp->pa_addr.sa_addr.ta_addrs[0].na_stack == NA_TCP) )
	  pa_tmp->pa_addr.sa_addr.ta_addrs[0].na_port = get_tcpport();
	else
	  break;
      }
      if (lt_attempts < 10) {
        started++;
        fprintf(stderr, "listening on presentation address : %s\n", pa2str(pa_tmp));
        *listen_pa_tmp = *pa_tmp; /* struct copy */
        listen_pa_tmp++;
      }
#endif
    }

    pa_mark = listen_pa_tmp; /* update the pa_mark pointer */

    if (!started) return NOTOK; else return 0;
}



int lt_server_psaps(service, pas)
char * service;
struct PSAPaddr *pas;
{
    char* myhost = getlocalhost();
    char* mydomain = malloc (NSAP_DOMAINLEN + 1);

    int stacks_found = 0;
    
    struct PSAPaddr *pa;

#define NTADDRS FD_SETSIZE

    struct TSAPaddr *tz;
    struct TSAPaddr  tas[NTADDRS];

#ifdef  TCP
    int     port;
    struct NSAPaddr *tcp_na;
#endif

#ifdef  X25
    struct NSAPaddr *x25_na;
#endif

    isodesetailor(NULLCP);

    bzero ((char *) pas, sizeof pas);
    pa = pas;

    bzero ((char *) tas, sizeof tas);
    tz = tas;

#ifdef  TP4
    if (!(ts_stacks & TS_TP4))
      ; /* no TP4 stack supported */
    else {
      tz -> ta_addrs -> na_stack = NA_NSAP;
      tz -> ta_addrs -> na_community = ts_comm_nsap_default;
      (void)strcpy(tz -> ta_addrs -> na_address, local_nsap);
      tz -> ta_addrs -> na_addrlen = strlen (local_nsap);
      bcopy (service, tz -> ta_selector, tz -> ta_selectlen = strlen(service));
      tz -> ta_naddr = 1;

      pa->pa_addr.sa_addr = *tz;
      pa->pa_addr.sa_selectlen = 0;
      pa->pa_selectlen = 0;

      pa++;
      tz++;
      stacks_found++;
    }
#endif

#ifdef  TCP
    if (!(ts_stacks & TS_TCP)) 
      ; /* no TCP stack supported */
    else {
      if (getdomainname(mydomain, NSAP_DOMAINLEN + 1) == NULL);
      tcp_na = tz -> ta_addrs;
      tcp_na -> na_stack = NA_TCP;
      tcp_na -> na_community = ts_comm_tcp_default;
      if (getdomainname(mydomain, NSAP_DOMAINLEN + 1) == NULL)
        sprintf(tcp_na -> na_domain, "%s.%s", myhost, mydomain);
      else
        tcp_na -> na_domain[0] = NULL;

#ifdef	LT_HACK
      tcp_na -> na_port = 0;      /* null port for location transparent agent */
#else
      tcp_na -> na_port = get_tcpport();
#endif

      tz -> ta_naddr = 1;
    
      pa->pa_addr.sa_addr = *tz;
      pa->pa_addr.sa_selectlen = 0;
      pa->pa_selectlen = 0;

      pa++;
      tz++;
      stacks_found++;
    }
#endif

#ifdef  X25
    if (!(ts_stacks & TS_X25))
      ; /* no X25 stack supported */
    else {
      char	pidstr[8];

      x25_na = tz -> ta_addrs;
      x25_na -> na_stack = NA_X25;
      x25_na -> na_community = ts_comm_x25_default;
      if (x25_local_dte && *x25_local_dte) {
        (void) strcpy (x25_na -> na_dte, x25_local_dte);
        x25_na -> na_dtelen = strlen (x25_na -> na_dte);
      }
      sprintf(pidstr, "%d", getpid()); /* get a unique PID (= process id) */
      x25_na -> na_pidlen = str2sel (pidstr, -1, x25_na -> na_pid, NPSIZE);
      tz -> ta_naddr = 1;

      pa->pa_addr.sa_addr = *tz;
      pa->pa_addr.sa_selectlen = 0;
      pa->pa_selectlen = 0;

      pa++;
      tz++;
      stacks_found++;
    }
#endif

    if (!stacks_found) {
      error(NULLCP, "No transport stacks found!\n");
      return(1);
    }
    pa_mark = pa;
    return(0);
}
