/* Copyright (c) 1994 Institute of Computer Science, Foundation of      */
/* Research and Technology - Hellas (ICS-FORTH)                         */
/* All rights reserved.                                                 */

/* Redistribution and use in source and binary forms are permitted      */
/* provided that the above copyright notice and this paragraph are      */
/* duplicated in all such forms and that any documentation, advertising */
/* materials, and other materials related to such distribution and use  */
/* acknowledge that the software was developed by the Institute of      */
/* Computer Science, Foundation of Research and Technology - Hellas.    */
/* The name of the Institute or the Foundation may not be used to       */
/* endorse or promote products derived from this software without       */
/* specific prior written permission.                                   */
/* THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR       */
/* IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED       */
/* WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.  */

/* implementation of the Directory update mechanism on start-up and on	*/
/* shut-down of management processes					*/

/* Costas Stathopoulos June 1993 */

#include "isode/quipu/util.h"
#include "isode/quipu/common.h"
#include "isode/quipu/entry.h"
#ifndef ICR1
#include "isode/tailor.h"
#else
#include "isode/compat/tailor.h"
#endif


#ifndef _LOCATION_TRANSPARENCY_
#include "lt.h"
#endif

extern short    oc_sntx;

extern struct  PSAPaddr * pa_mark;

static char     * context = "management";

char *	lt_now_modify();
char *	lt_add_agent();

extern char *	lt_bind();
extern int	lt_init();

extern char	lt_oper_status;


int lt_update(agentDN, pa, operationalStatus)
char * agentDN;		/* a character string format of the DN of the 	*/
			/* listening management process */
struct PSAPaddr  *pa;
char operationalStatus;
{
  char * bind_mess, * modify_mess, * add_mess;
  char * oper_str;

  char	globalName[STRINGLEN];

  strcpy (globalName, agentDN);

  if ( lt_init() != 0 ) return (1);

  if ( (bind_mess = lt_bind()) != NULLCP ) {
     fprintf ( stderr, "%s\n", bind_mess );
     return(1);
  }

  oper_str = malloc (4);
  oper_str = (operationalStatus == UP) ? UPSTR : DOWNSTR;

  modify_mess = malloc(STRINGLEN);
  add_mess = malloc(STRINGLEN);

  if ( (modify_mess = lt_now_modify(str2dn(globalName), pa, oper_str) )
								!= NULLCP )
    {
       if ( strcmp(modify_mess, "NO_SUCH_OBJECT") == 0 ) {
         if ( (add_mess = lt_add_agent (str2dn(globalName), pa, oper_str))
								!= NULLCP) {
	   fprintf ( stderr, "lt_update: %s", add_mess );
	   return(1);
	 }
       } else {
          fprintf ( stderr, "lt_update: %s", modify_mess );
          return(1);
        }
    }
  ds_unbind();
  return(0);
}


char * lt_now_modify(Name, pas, oper_str)
DN      Name;
struct  PSAPaddr  *pas;
char    *oper_str;
{

  struct  PSAPaddr      *pa = pas;

  struct ds_modifyentry_arg modarg;
  struct DSError moderr;

  AV_Sequence    oper_av, pa_av, pa_av_tmp;
  struct entrymod * em;

  /* set the service controls */
  if (get_default_service(&modarg.mea_common) != 0)
    return("Directory Service Error! Can't get default services\n");
  modarg.mea_common.ca_servicecontrol.svc_options = SVC_OPT_DONTUSECOPY;
  modarg.mea_common.ca_servicecontrol.svc_prio = SVC_PRIO_HIGH;
  modarg.mea_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
  modarg.mea_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;
  modarg.mea_common.ca_servicecontrol.svc_scopeofreferral = SVC_REFSCOPE_DMD;

  modarg.mea_object = Name;  /* the entry we want to modify! */

  /* remove the old operational status */
  modarg.mea_changes = em_alloc();
  modarg.mea_changes->em_type = EM_REMOVEATTRIBUTE;
  modarg.mea_changes->em_what = as_comp_new(str2AttrT(OPER_STATUS), NULLAV, NULLACL_INFO);
  modarg.mea_changes->em_what->attr_link = NULLATTR;

  /* add the new operational status */
  oper_av = avs_comp_new( str2AttrV(oper_str, str2syntax(OPER_STATUS_SNTX)));
  modarg.mea_changes->em_next = em_alloc();
  em = modarg.mea_changes->em_next;
  em->em_type = EM_ADDATTRIBUTE;
  em->em_what = as_comp_new(str2AttrT(OPER_STATUS), oper_av, NULLACL_INFO);
  em->em_what->attr_link = NULLATTR;

  if ( pa != NULLPA ) {
    /* remove the old presentation address */
    em->em_next = em_alloc();
    em = em->em_next;
    em->em_type = EM_REMOVEATTRIBUTE;
    em->em_what = as_comp_new(str2AttrT(PRES_ADD), NULLAV, NULLACL_INFO);
    em->em_what->attr_link = NULLATTR;
    pa_av = avs_comp_new( str2AttrV(pa2str(pa), str2syntax(PRES_ADD_SNTX)) );
    fprintf(stderr, "Presentation address : %s\n", pa2str(pa));
    pa++;
    for ( ; pa < pa_mark ; pa++) {
      /* add the new presentation address */
      fprintf(stderr, "Presentation address : %s\n", pa2str(pa));
      pa_av_tmp = avs_comp_new( str2AttrV(pa2str(pa), str2syntax(PRES_ADD_SNTX)));
      pa_av = avs_merge(pa_av, pa_av_tmp);
    }
      em->em_next = em_alloc();
      em = em->em_next;
      em->em_type = EM_ADDATTRIBUTE;
      em->em_what = as_comp_new(str2AttrT(PRES_ADD), pa_av, NULLACL_INFO);
      em->em_what->attr_link = NULLATTR;
  }

  em->em_next = NULLMOD;

  /* now do the modification */
  if (ds_modifyentry(&modarg, &moderr) != DS_OK)
    switch (moderr.dse_type) {
    case DSE_NAMEERROR: 
      switch (moderr.dse_un.dse_un_name.DSE_na_problem) {
      case DSE_NA_NOSUCHOBJECT:
        return ("NO_SUCH_OBJECT");
      case DSE_NA_ALIASPROBLEM:
      case DSE_NA_ALIASDEREFERENCE:
	 return ("Error! Alias problem.\n");
      case DSE_NA_INVALIDATTRIBUTESYNTAX:
	return ("Name Error! Invalid attribute syntax.\n");
      }
    case DSE_LOCALERROR:
      return(duaerror);

    case DSE_REMOTEERROR:
      return(localdsaerror);

    case DSE_ATTRIBUTEERROR:
      return(attributerror);

    case DSE_REFERRAL:
    case DSE_DSAREFERRAL:
      return(remotedsaerror);

    case DSE_SECURITYERROR:
      return(security);

    case DSE_SERVICEERROR:
      return(serviceerror);

    case DSE_UPDATEERROR:
      return(updaterror);

    default:
      return(localdsaerror);
    }

  return(NULL);
}


char * lt_add_agent(Name, pas, oper_str)
DN      Name;
struct PSAPaddr *pas;
char	*oper_str;
{

  struct  PSAPaddr      *pa = pas;

  struct ds_addentry_arg        addarg;
  struct DSError                adderr;

  register Attr_Sequence as = NULLATTR,
                         as_temp;
  char * buf;
  AV_Sequence   av_seq,
                pa_av, pa_av_tmp;
  short acl_syntax;

  fprintf(stderr, "Adding a new Management Process in the Directory!\n");
 
  /* First set the service controls */
  if (get_default_service(&addarg.ada_common) != 0)
    return("Directory Service Error! Can't get default services\n");
  addarg.ada_common.ca_servicecontrol.svc_options = SVC_OPT_DONTUSECOPY;
  addarg.ada_common.ca_servicecontrol.svc_prio = SVC_PRIO_HIGH;
  addarg.ada_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
  addarg.ada_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;
  addarg.ada_common.ca_servicecontrol.svc_scopeofreferral = SVC_REFSCOPE_DMD;

  addarg.ada_object = Name;

  /* Now form the entry that we are going to add to the Directory */

  /* put the objectClass */
  av_seq = avs_comp_alloc();
  av_seq->avseq_next = NULLAV;
  av_seq->avseq_av.av_syntax = oc_sntx;
  av_seq->avseq_av.av_struct = (caddr_t) str2oc(TOP_DSCLASS);
  add_oc_avs(str2oc(AE_DSCLASS), &av_seq);
  add_oc_avs(str2oc(QUIPU_OBJECT_DSCLASS), &av_seq);
  add_oc_avs(str2oc(ICM_AGENT_DSCLASS), &av_seq);
  as = as_comp_new(str2AttrT(DSCLASS), av_seq, NULLACL_INFO);

  /*  put the supported application context, operational status
                and presentation address attributes             */
  as_temp = as_comp_new(str2AttrT(SAC), avs_comp_new( str2AttrV(context, str2syntax(SAC_SNTX))), NULLACL_INFO);
  as = as_merge(as, as_temp);
  as_temp = as_comp_new(str2AttrT(OPER_STATUS), avs_comp_new( str2AttrV(oper_str, str2syntax(OPER_STATUS_SNTX))), NULLACL_INFO);
  as = as_merge(as, as_temp);

  pa_av = avs_comp_new( str2AttrV(pa2str(pa), str2syntax(PRES_ADD_SNTX)));
  fprintf(stderr, "Presentation address : %s\n", pa2str(pa));
  pa++;
  for ( ; pa < pa_mark ; pa++) {
    /* add the new presentation address */
    fprintf(stderr, "Presentation address : %s\n", pa2str(pa));
    pa_av_tmp = avs_comp_new( str2AttrV(pa2str(pa), str2syntax(PRES_ADD_SNTX)));
    pa_av = avs_merge(pa_av, pa_av_tmp);
  }

  as_temp = as_comp_new(str2AttrT(PRES_ADD), pa_av, NULLACL_INFO);
  as = as_merge(as, as_temp);

  /* put a null ACL value so that we get the default */
  av_seq = NULLAV;
  as_temp = as_comp_new(str2AttrT(ACL), av_seq, NULLACL_INFO);
  as = as_merge(as, as_temp);

  addarg.ada_entry = as;

  /* now add */
  if ( ds_addentry(&addarg, &adderr) != DS_OK )
    switch (adderr.dse_type) {
    case DSE_NAMEERROR:
      return(namerror);

    case DSE_LOCALERROR:
      return(duaerror);

    case DSE_REMOTEERROR:
      return(localdsaerror);

    case DSE_ATTRIBUTEERROR:
      return(attributerror);

    case DSE_REFERRAL:
    case DSE_DSAREFERRAL:
      return(remotedsaerror);

    case DSE_SECURITYERROR:
      return(security);

    case DSE_SERVICEERROR:
      return(serviceerror);

    case DSE_UPDATEERROR:
      return(updaterror);

    default:
      return(localdsaerror);
    }

    return(NULL);
}

