//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/***********************************************************/
/* 
	File:	Cmisiviews.cc

	Notes:

		Interviews interface to cmisbrowser 

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/***********************************************************/
/* Include files */


#include <signal.h>
#include "cmisiviews.h"
#include "ivutils.h"
#include "ivadisplay.h"
#include "ivmonitor.h"
#include "ivbrowser.h"
#include "ivhosts.h"

#include "ivutils.cc"
#include "ivhosts.cc"
#include "ivadisplay.cc"
#include "ivbrowser.cc"
#include "ivmonitor.cc"

/***********************************************************/

extern "C" {
void  start_malloc_trace (char*);
}

/***********************************************************/
/* Host window well known so that signal handler can find it */

Hosts_Window	*host_window;

/***********************************************************/
/* Signal handler */
/***********************************************************/

static	void	signal_alarm()
{

	host_window->signal_alarm();

} /* signal_alarm */

/***********************************************************/
/* Startup */
/***********************************************************/

void	init_iviews(int argc, char **argv, GString hosts_table_name)
{
World	*world = (NEW World("World", argc, argv));

	host_window = (NEW Hosts_Window(hosts_table_name));
	host_window->SetName("Cmis browser");
	world->Insert(host_window);

#ifdef  __GNUG__
	signal(SIGALRM, &signal_alarm);		// set up signal handler
#elif   __cplusplus && !AIXV3
	signal(SIGALRM, (SIG_PF)&signal_alarm);
#elif	__cplusplus && AIXV3
	signal(SIGALRM, (void (*)(int)) &signal_alarm);
#endif
//	start_malloc_trace ((char*) 0);

	host_window->Run();

	delete host_window;

//	This causes interviews to crash

//	delete world;

} /* init_iviews */

/***********************************************************/
/* End */
/***********************************************************/


