//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/******************************************************************/
/*

	File:	cmisresult.cc

	Notes:
		CMIS Client Result Class Methods

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/******************************************************************/

#include "cmis.h"
#include "cmisclient.h"
#include "cmisresult.h"

/******************************************************************/
/* Public methods */
/******************************************************************/

Result	Cmis_client_result:: identify_code(struct MSAPIndication *mi)

{
int	count;
typedef	Boolean	(*DECODE_FN_PTR)(int, struct MSAPIndication *);
static	struct	Decode_table_rec
	{
	int		type;
	Result		result;
	DECODE_FN_PTR	decode_invoke;
	}
	decode_table[] =
{
	{MI_ERROR, 	Error,	   Cmis_client_result::decode_error_id},
	{MI_CREATE_RES, Success,   Cmis_client_result::decode_create_id},
	{MI_GET_RES, 	Success,   Cmis_client_result::decode_get_id},
	{MI_SET_RES, 	Success,   Cmis_client_result::decode_set_id},
	{MI_ACTION_RES,	Success,   Cmis_client_result::decode_action_id},
	{MI_DELETE_RES,	Success,   Cmis_client_result::decode_delete_id},
	{MI_LINKED_REPLY,Continue, Cmis_client_result::decode_linked_id},
	{MI_EMPTY_RES,  Finish,    Cmis_client_result::decode_empty_id}
};
static  int	lgth = sizeof(decode_table)/sizeof(struct  Decode_table_rec);
/*
printf("\nFound error code %d", mi->mi_errtype);
*/
for (count=0; count<lgth; count++)
	if (mi->mi_type == decode_table[count].type)
	{
	DECODE_FN_PTR	fn_ptr = decode_table[count].decode_invoke;
/*
printf("\nFound code %d", decode_table[count].type); 
*/
		if ((*fn_ptr)(_invoke_id, mi) == True)
			return(decode_table[count].result);
		}

	return(Notfound);	/* Not found */

} /* identify_code */

/******************************************************************/
/* Decode error id */

Boolean	Cmis_client_result::decode_error_id(int invoke_id,
					    struct MSAPIndication *mi)
{
	if (mi->mi_error.id == invoke_id)
		return(True);

	return(False);

} /* decode_error_id */

/******************************************************************/
/* Decode create id */

Boolean	Cmis_client_result::decode_create_id(int invoke_id,
					    struct MSAPIndication *mi)
{
	if (mi->mi_createres.id == invoke_id)
		return(True);

	return(False);

} /* decode_create_id */

/******************************************************************/
/* Decode get id */

Boolean	Cmis_client_result::decode_get_id(int invoke_id,
					    struct MSAPIndication *mi)
{
	if (mi->mi_getres.id == invoke_id)
		return(True);

	return(False);

} /* decode_get_id */

/******************************************************************/
/* Decode set id */

Boolean	Cmis_client_result::decode_set_id(int invoke_id,
					    struct MSAPIndication *mi)
{
	if (mi->mi_setres.id == invoke_id)
		return(True);

	return(False);

} /* decode_set_id */

/******************************************************************/
/* Decode action id */

Boolean	Cmis_client_result::decode_action_id(int invoke_id,
					    struct MSAPIndication *mi)
{
	if (mi->mi_actionres.id == invoke_id)
		return(True);

	return(False);

} /* decode_action_id */

/******************************************************************/
/* Decode delete id */

Boolean	Cmis_client_result::decode_delete_id(int invoke_id,
					    struct MSAPIndication *mi)
{
	if (mi->mi_deleteres.id == invoke_id)
		return(True);

	return(False);

} /* decode_delete_id */

/******************************************************************/
/* Decode linked id */

Boolean	Cmis_client_result::decode_linked_id(int invoke_id,
					    struct MSAPIndication *mi)
{
	if (mi->mi_linkedreplyinv.args.lr_linkid == invoke_id)
		return(True);

	return(False);

} /* decode_linked_id */

/******************************************************************/
/* Decode empty id */

Boolean Cmis_client_result::decode_empty_id(int invoke_id,
                                            struct MSAPIndication *mi)
{
        if (mi->mi_emptyres.id == invoke_id)
                return(True);

        return(False);

} /* decode_empty_id */

/******************************************************************/
/* Decode class and instance names */

Boolean	Cmis_client_result::decode_class_instance(
		int class_type, MID class_id, int instance_type, 
		MN instance, GString *class_string, GString *instance_string)
{
char	buffer[MAX_TEMP_BUFFER];

	if (class_type == MID_GLOBAL)
	{
		*class_string = sprintoid(class_id->mid_global);
	}
	else
	{
		sprintf(buffer, "%d", class_id->mid_local);
		*class_string = buffer;
	}

	if (instance_type == MN_DN || instance_type == MN_LOCALDN)
	{
	char	*inst_buffer;

/*	Not needed in Isode 7.0

		dn_decode(instance->mn_dn);
*/
		inst_buffer = dn2str(instance->mn_dn);
		*instance_string = inst_buffer;
		free(inst_buffer);
	}
	else
		*instance_string = instance->mn_nonspec;

	if (strlen(*class_string) > 0)
		return(True);

	return(False);

} /* decode_class_instance */

/******************************************************************/
/* Decode attributes */

Boolean	Cmis_client_result::decode_attribute(CMISParam *attr,
					     Cmis_Object *object)
{

	if (attr->mp_id.mid_type != MID_GLOBAL)
	{
/*
		fprintf(stderr, "\nCmisresult.cc: MID_GLOBAL expected");
*/
		return(False);
	}
	object->add(attr->mp_id.mid_global, attr->mp_val);

	return(True);

} /* decode_attribute */

/******************************************************************/
/* Decode error */

char	*Cmis_client_result::decode_error(struct MSAPIndication *mi,
					  CMISErrors *error_code)
{
	switch (*error_code=mi->mi_errtype)
	{
	case m_setListError:
	{
		CMISSetRes      *set = &mi->mi_setres.result;
		int		count;

		for (count=0; count<set->sr_nattrs; count++)
		{
		CMISErrors	error = set->sr_attrs[count].sa_error;

			if (error)
			{
				*error_code = error;
/*
printf("\nSET LIST ERROR %d %s", error, MErrString(error));
*/
				return(MErrString(error));
			}
		}
	}

	case m_getListError:
	{
		CMISGetRes      *get = &mi->mi_getres.result;
		int		count;

		for (count=0; count<get->gr_nattrs; count++)
		{
		CMISErrors	error = get->gr_attrs[count].ga_error;

			if (error)
			{
				*error_code = error;
				return(MErrString(error));
			}
		}
	}
	default:
		break;
	}

	return(MErrString(*error_code));

} /* decode_error */

/******************************************************************/
/* Derived result class methods */
/******************************************************************/

Result	Create_result::process(struct MSAPIndication *mi)
{

	return(identify_code(mi));

} /* process */

/******************************************************************/

Result	Action_result::process(struct MSAPIndication *mi)
{

	return(identify_code(mi));

} /* process */

/******************************************************************/

Result	Delete_result::process(struct MSAPIndication *mi)
{

	return(identify_code(mi));

} /* process */

/******************************************************************/

Result	Get_result::process(struct MSAPIndication *mi)
{
Result		result;
CMISGetRes      *get;
GString		class_name, instance;
Cmis_Object	*object;
CMISErrors	error_code;
char		*error_string;

	switch (result=identify_code(mi))
	{
 	case Notfound:
	case Finish:
		return(result);

	case Error:
		error_string = decode_error(mi, &error_code);
/*
printf("\nError code %d GString %s", error_code, error_string);
*/
		result_list->set_error(error_string, error_code);
		return(result);

	case Success:
		get = &mi->mi_getres.result;
		break;

	case Continue:
                if (mi->mi_linkedreplyinv.args.lr_type == LR_GET_RES)
                        get = &mi->mi_linkedreplyinv.args.lr_getres;
                else
                {
                        error_code = (CMISErrors) mi->mi_linkedreplyinv.args.lr_type;
                        error_string = MErrString(error_code);
                        result_list->set_error(error_string, error_code);
                        return(Error);
                }
		break;
	}

	if (decode_class_instance(get->gr_class.mid_type, &(get->gr_class),
				  get->gr_inst.mn_type, &(get->gr_inst),
				  &class_name, &instance) == True)
	{
	int	count;

		object = result_list->add(class_name, instance);

                for (count=0; count<get->gr_nattrs; count++)
                        if (!decode_attribute(&(get->gr_attrs[count].ga_ava),
					      object))
				return(Error);

	}

	return(result);

} /* process */

/******************************************************************/

Result	Set_result::process(struct MSAPIndication *mi)
{
Result		result;
CMISErrors	error_code;
char		*error_string;
CMISSetRes      *set;
GString		class_name, instance;
Cmis_Object	*object;

	switch (result=identify_code(mi))
	{
 	case Notfound:
	case Finish:
		return(result);

	case Error:
		error_string = decode_error(mi, &error_code);
/*
printf("\nError code %d GString %s", error_code, error_string);
*/
		result_list->set_error(error_string, error_code);
		return(result);

	case Success:
		set = &mi->mi_setres.result;
		break;

	case Continue:
                if (mi->mi_linkedreplyinv.args.lr_type == LR_GET_RES)
                        set = &mi->mi_linkedreplyinv.args.lr_setres;
                else
                {
                        error_code = (CMISErrors) mi->mi_linkedreplyinv.args.lr_type;
                        error_string = MErrString(error_code);
                        result_list->set_error(error_string, error_code);
                        return(Error);
                }
		break;
	}

	if (decode_class_instance(set->sr_class.mid_type, &(set->sr_class),
				  set->sr_inst.mn_type, &(set->sr_inst),
				  &class_name, &instance) == True)
	{
	int	count;

		object = result_list->add(class_name, instance);

                for (count=0; count<set->sr_nattrs; count++)
                        if (!decode_attribute(&(set->sr_attrs[count].sa_ava),
					      object))
				return(Error);
	}

	return(result);

} /* process */

/******************************************************************/
/* End of module */
/******************************************************************/




