//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/******************************************************************/
/*
	File:	cmisresult.h

	Notes:	CMIS Result Class

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/******************************************************************/

#include "cmisx500.h"

/******************************************************************/
/* Attribute class */

class	Cmis_Attribute
{
	X500_attribute	_x500_attr;
public:
	Cmis_Attribute(OID oid, PE pe)
	{

		_x500_attr.decode(oid, pe);

#ifdef	DEBUG_CLIENT
		CLIENT_TRACE("\n\tAttribute %s Value %s", 
			     _x500_attr.get_oid_name(), 
			     _x500_attr.get_value());
#endif
	}

	char		*get_id()	{return (_x500_attr.get_oid_name());}
	char		*print_id()	{return (_x500_attr.get_oid_name());}
	char		*get_val()	{return (_x500_attr.get_value());}
	char		*print_val()	{return (_x500_attr.get_value());}
};

/******************************************************************/

class Attribute_list : public Ptrlist
{
public:

        Attribute_list()	{}	// Stop Sun C++ warning
        ~Attribute_list()
        {
		while (vdelete(0))
			;
        }

	int	vdelete(int index)
        {
        Cmis_Attribute  *attr;

                if (attr=(Cmis_Attribute *) get(index))
                {
                        delete attr;
			remove(index);
			return(TRUE);
                }
		else
			return(FALSE);
	}
};

/******************************************************************/
/* Result list class */
/******************************************************************/
/* Object descriptor */

class	Cmis_Object
{
	Attribute_list	_attribute_list;
	GString		_class_name;
	GString		_instance;
	GString		_class_string;
public:
	Cmis_Object(GString class_name, GString instance)
	{
/*
	GString	quoted_class;
*/
		_class_name = class_name; 
		_instance = instance;
		_class_string = oid2name(str2oid(class_name), OIDPART);
/*
		quoted_class = oid2name(str2oid(class_name), OIDPART);
*/
		// oid strings are enclosed in quotes
		// strip them off

/*		ISODE 7.0 strips off the quotes 

                quoted_class = quoted_class.after("\"");
                _class_string = quoted_class.before("\"");
*/

#ifdef  DEBUG_CLIENT
        	CLIENT_TRACE("\nClass: Oid %s Name %s Instance %s",
	                     (char *) _class_string, (char *) _class_name,
			     (char *) _instance);
#endif

	}

	void	add(OID oid, PE pe)
	{
	Cmis_Attribute	*attr = (NEW Cmis_Attribute(oid, pe));

		_attribute_list.add(attr);
	}

        Cmis_Attribute	*get(int i)
	{
		return((Cmis_Attribute *) _attribute_list[i]);
	}

	int		attr_count()		
	{
		return(_attribute_list.count());
	}
	char		*print_class() 		{return(_class_string);}
	char		*get_class()		{return(_class_name);}
	char		*print_instance()	{return(_instance);}
};

/******************************************************************/

class Object_list : public Ptrlist
{
public:

        Object_list()	{}	// Stop Sun C++ warning
        ~Object_list()
        {
		while (vdelete(0))
			;
        }

	int	vdelete(int index)
        {
        Cmis_Object	*object;

                if (object=(Cmis_Object *) get(index))
                {
                        delete object;
			remove(index);
			return(TRUE);
                }
		else
			return(FALSE);
	}
};
/******************************************************************/
/* Result list class */

class	Cmis_result
{
	Object_list	_object_list;
	CMISErrors	error_code;
	GString		error_string;
public:
	Cmis_result()
	{
		error_code = m_noError;
		error_string = "";
	}

	Cmis_Object	*add(GString class_name, GString instance)
	{
	Cmis_Object *new_object = (NEW Cmis_Object(class_name, instance));

		_object_list.add(new_object);
		return(new_object);
	}

        Cmis_Object	*get(int i)
	{
		return((Cmis_Object *)_object_list[i]);
	}

	char		*get_error()		{return((char *)error_string);}
	CMISErrors	get_error_code()	{return(error_code);}
	void		set_error(char *msg, CMISErrors code)
	{
		error_string = msg; error_code = code;
	}
};

/******************************************************************/
/* Cmis client result classes */
/******************************************************************/
/* Abstract class */

class	Cmis_client_result
{
	int	_invoke_id;

	static	Boolean	decode_error_id (int, struct MSAPIndication *);
	static	Boolean	decode_create_id (int, struct MSAPIndication *);
	static	Boolean	decode_get_id (int, struct MSAPIndication *);
	static	Boolean	decode_set_id (int, struct MSAPIndication *);
	static	Boolean	decode_action_id (int, struct MSAPIndication *);
	static	Boolean	decode_delete_id (int, struct MSAPIndication *);
	static	Boolean	decode_linked_id (int, struct MSAPIndication *);
        static  Boolean decode_empty_id (int, struct MSAPIndication *);
public:

	Cmis_client_result(int invoke_id)
	{
		_invoke_id = invoke_id;
	}

	virtual	~Cmis_client_result()
	{
	}

	Result	identify_code(struct MSAPIndication *);

	virtual	Result	process(struct MSAPIndication *)
				{return(Error);};

	Boolean	decode_class_instance(int, MID, int, MN, GString *,
				      GString *);

	Boolean	decode_attribute(CMISParam *, Cmis_Object *);

	char	*decode_error(struct MSAPIndication *, CMISErrors *);
	void	set_error(char *, CMISErrors);
};

/******************************************************************/
/* Create result */

class	Create_result : public Cmis_client_result
{
public:
	Create_result(int invoke) : Cmis_client_result(invoke) {}

	Result	process(struct MSAPIndication *);
};

/******************************************************************/
/* Action result */

class	Action_result : public Cmis_client_result
{
public:
	Action_result(int invoke) : Cmis_client_result(invoke) {}

	Result	process(struct MSAPIndication *);
};

/******************************************************************/
/* Delete result */

class	Delete_result : public Cmis_client_result
{
public:
	Delete_result(int invoke) : Cmis_client_result(invoke) {}

	Result	process(struct MSAPIndication *);
};

/******************************************************************/
/* Get result */

class	Get_result : public Cmis_client_result
{
	Cmis_result	*result_list;
public:
	Get_result(int invoke, Cmis_result *list)
	: Cmis_client_result(invoke) {result_list = list;}

	Result	process(struct MSAPIndication *);
};

/******************************************************************/
/* Set result */

class	Set_result : public Cmis_client_result
{
	Cmis_result	*result_list;
public:
	Set_result(int invoke, Cmis_result *list)
	: Cmis_client_result(invoke) {result_list = list;}

	Result	process(struct MSAPIndication *);
};

/******************************************************************/
/* End of header file */
/******************************************************************/

