//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/***********************************************************/
/* 
	File:	ivadisplay.cc

	Notes:	Cmis Interviews attribute display class methods

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/***********************************************************/
/* Attribute Display methods */
/***********************************************************/

Boolean	Attr_Display::display_attributes(StringBrowser *attr_browser,
					 Cmis_result *result)
{
Cmis_Object     *object;

	if (object=result->get(0))
	{
/*
		printf("\n\nclass %s Instance %s",
			object->print_class(), object->print_instance());
*/
		if (object->attr_count() == attr_browser->Count())
		{
			if (!changed(object))
				return(False);
		}

		save_attributes(object);

		write_attributes(attr_browser);
	}

	return(True);

} /* display_attributes */

/***********************************************************/

void	Attr_Display::save_attributes(Cmis_Object *object)
{
Cmis_Attribute  *attr;
int             count = object->attr_count();

	delete_attributes();

	_names = (NEW GStringPtr[count]);
	_values = (NEW GStringPtr[count]);

	for (_rows=0; attr=object->get(_rows); _rows++)
	{
		_names[_rows] = (NEW GString(attr->print_id()));
		_values[_rows] = (NEW GString(attr->print_val()));
	}

} /* save_attributes */

/***********************************************************/

int	Attr_Display::calc_max_name_lgth()
{
int     count, maxLgth = 0;

	for (count=0; count<_rows; count++)
	{
	char    *name = *_names[count];
	int     lgth = strlen(name);
/*
printf("\nName %d %s", count, name);
*/
		if (lgth > maxLgth)
			maxLgth = lgth;
	}
	return(maxLgth);

} /* calc_max_name_lgth */

/***********************************************************/

int	Attr_Display::calc_max_value_lgth()
{
int     count, maxLgth = 0;

	for (count=0; count<_rows; count++)
	{
	char    *value = *_values[count];
	int     lgth = strlen(value);
/*
printf("\nValue %d %s", count, value);
*/
		if (lgth > maxLgth)
			maxLgth = lgth;
	}
	return(maxLgth);

} /* calc_max_value_lgth */

/***********************************************************/

void	Attr_Display::write_attributes(StringBrowser *attr_browser)
{
int     count;
int	max_name_lgth = calc_max_name_lgth() + 5,
	max_value_width = calc_max_value_lgth();

	// clear browser 

	attr_browser->Clear();

	for (count=0; count<_rows; count++)
	{
	char	*name = *_names[count];
	int	name_lgth = strlen(name);
	char	*value = *_values[count];
	GString	attr_str = name;
/*
printf("\nCount %d Name %s Value %s", count, name, value);
*/
		while (name_lgth++ < max_name_lgth)
			attr_str += ' ';

		attr_str += value;

		attr_browser->Append((char *) attr_str);
	}

	attr_browser->Flush();

	_cols = max_name_lgth + max_value_width;

} /* write_attributes */

/***********************************************************/

void	Attr_Display::update_attributes(StringBrowser *attr_browser,
					 Cmis_result *result)
{
Cmis_Object     *object;
Boolean		changed = False;

	/* Get the attributes */

	if (object=result->get(0))
	{
	Cmis_Attribute  *attr;
	int	max_value_width = 0;
	int	count;

		for (count=0; attr=object->get(count); count++)
		{
		char	*name = attr->print_id();
		char	*value = attr->print_val();

			// Find attribute in current display

			for (int row=0; row<_rows; row++)
			{
				if (!strcmp((char *) *_names[row],
					    name) &&
				    strcmp((char *) *_values[row],
					   value))
				{
					changed = True;
					*_values[row] = value;
/*
printf("\nNew value %d %s", row, (char *)*_values[row]);
*/
				}
			}
			
		}
	}
	/* Redisplay attributes */

	if (changed)
		write_attributes(attr_browser);

} /* update_attributes */

/***********************************************************/

Boolean	Attr_Display::changed(Cmis_Object *object)
{
Cmis_Attribute  *attr;
int	rows;

	for (rows=0; attr=object->get(rows); rows++)
	{
	char	*value = *_values[rows];

		if (strcmp(value, attr->print_val()))
			return(True);
	}

	return(False);

} /* changed */

/***********************************************************/
/* End */
/***********************************************************/


