//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/***********************************************************/
/* 
	File:	ivbrowser.cc

	Notes:

		Interviews browser methods

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)

*/
/***********************************************************/
/* Browser Methods */
/***********************************************************/
/* Initialise */

void	Browser_Window::initialise(GString host, GString service, GString top,
				   GString top_dn, Boolean &result)

{
int	space = round(1*inch);

	result = True;
	_this_interactor = 0;

	_cmis_client = (NEW Cmis_client);

	if (_cmis_client->initialise(host, service) == Error)
		result = False;
	else
	{
		input = (NEW Sensor);

		input->Catch(DownEvent);
		input->Catch(UpEvent);

		_this_interactor = interior();
		Insert(_this_interactor);
		if (display_class(top, top_dn) == True)
		{
		Leaf	*leaf = (NEW Leaf(top, top_dn));
/*
			printf("\nNew Leaf %ld class %s instance %s",
				leaf, leaf->get_class_name(),
				leaf->get_instance());
*/		
			_hierarchy.add(leaf);
		}
	}

} /* initialise */

/***********************************************************/

void Browser_Window::Xcallback(int option)
{
        switch (option)
        {
        case XCALL_BCHANGE:
                modify();
                break;
        case XCALL_BSETDEFAULT:
                set_default();
                break;
        case XCALL_BPOLLMONITOR:
                monitor();
                break;
        case XCALL_BUP:
                move_up();
                break;
        case XCALL_BDOWN:
                move_down();
                break;
        case XCALL_BREFRESH:
                refresh();
                break;
        case XCALL_BQUIT:
                quit();
                break;
        }

} /* Xcallback */

/***********************************************************/

void	Browser_Window::move_up()		
{
int	current = _hierarchy.count() - 1;
/*
	printf("\nMove Up %d", current);
*/
	_msg_line->update("");

	if (!current)
	{
	Leaf	*leaf = (Leaf *)_hierarchy.first();

		if (display_class(leaf->get_class_name(),
				  leaf->get_instance()) == True)

			_msg_line->error_msg("At top of tree");
		else
		{
			_msg_line->error_msg("SMA not responding");
			quit();
		}
	}
	else
	{
		do
		{
		Leaf	*leaf;

			_hierarchy.vdelete(current);
			leaf = (Leaf *)_hierarchy[--current];
/*
			printf("\nMove up leaf %ld to class %s instance %s",
				leaf, leaf->get_class_name(),
				leaf->get_instance());
*/
			if (display_class(leaf->get_class_name(),
					  leaf->get_instance()) == True)
				break;
			else
			{
				if (leaf == _hierarchy.first())
				{
				GString	msg = "SMA not responding";

					_msg_line->error_msg(msg);
					quit();
					break;
				}
				else
				{
				GString	msg = "Object no longer exists:";
			
					msg += leaf->get_instance();
					_msg_line->error_msg(msg);
				}
			}
		}
		while (current);

		_msg_line->update("");
	}

} /* move_up */

/***********************************************************/

void	Browser_Window::move_down()		
{
Subordinate	*subordinate;
/*
	printf("\nMove down");
*/
	_msg_line->update("");

	if (subordinate=select_subordinate())
	{
	GString class_name, instance;

		class_name = subordinate->get_class();
		instance = subordinate->get_instance();
/*
		printf("\nSelected class %s instance %s",
			(char *) class_name, (char *) instance);
*/
		if (display_class(class_name, instance) == True)
		{
		Leaf	*leaf = (NEW Leaf(class_name, instance));
/*
			printf("\nDown:Leaf %ld class %s instance %s",
				leaf, leaf->get_class_name(),
				leaf->get_instance());
*/		
			_hierarchy.add(leaf);
		}
		else
		{
		GString	msg = "Cannot move to " + instance;

			_msg_line->error_msg(msg);

			/* Redisplay current class */

			refresh();
		}
	}

} /* move_down */

/***********************************************************/

void	Browser_Window::modify()		
{
char	*attribute;
/*
	printf("\nMove down");
*/
	_msg_line->update("");

	if (attribute=select_attribute())
	{
	GString		selected, oid_name;
	X500_attribute	x500;

		selected = attribute;
		oid_name = selected.before(" ");
		if (x500.is_encodeable(oid_name))
		{
		GString	str_value;

			str_value = selected.after(" ");
			str_value = str_value.after(RXwhite);

			// Change value

			_modify_editor->Read_Input(this,
						  (Menu_Valid_Fn)&Browser_Window::set_attribute,
						  (char *) oid_name, (char *) str_value);
/*
printf("\nOid name is %s Address %ld, this %ld", 
	(char *) oid_name, (char *) oid_name, this);
*/
		}
		else
		{
		GString	msg;
		char	*syntax;

			msg = oid_name + " cannot be modified. Syntax: ";
			if (syntax=x500.get_syntax(oid_name))
				msg += syntax;
			else
				msg += "unknown";

			_msg_line->error_msg(msg);
		}
	}

	_msg_line->update("");

} /* modify */

/***********************************************************/
/* Change attribute */

Boolean         Browser_Window::set_attribute(char *oid_name, char *svalue)
{
OID     type;
PE      value;
X500_attribute  x500;
Leaf	*leaf = (Leaf *)_hierarchy.last();
GString	class_name = leaf->get_class_name();
GString	instance = leaf->get_instance();
Boolean	result = True;
GString	tmp(svalue), tmp1(svalue);	// parse modifies string
					// so save (nasty fix!)

	_hosts_window->set_alarm_off(); // switch off any alarms
/*
printf("\nEncode x500 %s Address %ld %s", oid_name, oid_name, svalue);
*/
	if (x500.encode(oid_name, (char *)tmp, &type, &value))
	{
	Cmis_result	cmis_result;

		if (_cmis_client->set(class_name, instance, oid_name,
				      (char *)tmp1, &cmis_result))
#ifdef	SETREFRESH
			refresh();
#else
			_attr_display.update_attributes(_attribute_browser,
							&cmis_result);
#endif
		else
		{
			_msg_line->error_msg(cmis_result.get_error());
			result = False;
		}
	}
	else
	{
		_msg_line->error_msg("Cannot encode value");
/*
printf("\nx500 %s failure %s", oid_name, svalue);
*/
		result = False;
	}

	_hosts_window->set_alarm_on();
	return(result);

} /* set_attribute */

/***********************************************************/

void	Browser_Window::set_default()		
{
char	*attribute;
Leaf	*leaf = (Leaf *)_hierarchy.last();
GString	class_name = leaf->get_class_name();
GString	instance = leaf->get_instance();

	_msg_line->update("");

	if (attribute=select_attribute())
	{
	GString		selected, oid_name;
	Cmis_result	result;

		selected = attribute;
		oid_name = selected.before(" ");

		_hosts_window->set_alarm_off(); // switch off any alarms

		if (_cmis_client->set(class_name, instance, oid_name, &result))
#ifdef	SETREFRESH
			refresh();
#else
			_attr_display.update_attributes(_attribute_browser,
							&result);
#endif
		else
			_msg_line->error_msg(result.get_error());

		_hosts_window->set_alarm_on();
	}

	_msg_line->update("");

} /* set_default */

/***********************************************************/

void	Browser_Window::monitor()		
{
Monitor_Window	*monitor;
Leaf	*leaf = (Leaf *)_hierarchy.last();
int	max_cols;
int	cols = _attr_display.get_cols();
int	max_rows = _attr_display.get_rows();

	if ((max_cols=strlen((char *) leaf->get_instance())+6) < cols)
		max_cols = cols;

	monitor = (NEW Monitor_Window(leaf->get_class_name(),
				     leaf->get_instance(),
				     max_rows, max_cols, this));

	_monitor_list.add(monitor);

	GetWorld()->Insert(monitor);

	if (!strlen((char *) leaf->get_instance()))
		monitor->SetName((char *) leaf->get_class_name());
	else
		monitor->SetName((char *) leaf->get_instance());

} /* monitor */

/***********************************************************/

int	Browser_Window::refresh_monitor(long cur_time)		
{
Monitor_Window	*monitor;
int		count;
long		next_alarm = 0;

	for (count=0; monitor=(Monitor_Window *)_monitor_list[count]; count++)
	{
	long	alarm;

		alarm = monitor->refresh(cur_time);
		if (!next_alarm)
			next_alarm = alarm;
		else
			if (alarm && (alarm < next_alarm))
				next_alarm = alarm;
	}

	_next_alarm = next_alarm;
	return(next_alarm);

} /* refresh_monitor */

/***********************************************************/
/* Monitor is requesting new time */

void	Browser_Window::set_signal_alarm(int secs)
{
Monitor_Window	*monitor;
int		count;
long		cur_time = time((long *) 0);
long		next_alarm = 0;

	// see if this is first request from a monitor 

	if (!_monitor_list.count())
		next_alarm = cur_time + secs;
	else
	{
		// check if monitor is requesting a faster alarm than current

		for (count=0; monitor=(Monitor_Window *)_monitor_list[count];
		     count++)
		{
		long	alarm;

			alarm = monitor->get_next_alarm();
/*
	printf("\nBrowser monitor alarm %ld", alarm)
*/;
			if (!next_alarm)
				next_alarm = alarm;
			else
				if (alarm && (alarm < next_alarm))
					next_alarm = alarm;
		}
	}

	if (next_alarm <= cur_time)
		next_alarm = cur_time + 1;

	_next_alarm = next_alarm;
/*
printf("\nBrowser set signal alarm %ld", _next_alarm);
*/
	_hosts_window->set_alarm_on();

} /* set_signal_alarm */

/***********************************************************/

void	Browser_Window::refresh()		
{
Leaf	*leaf = (Leaf *)_hierarchy.last();
/*
	printf("\nRefresh: %ld class %s instance %s",
		leaf, leaf->get_class_name(),
		leaf->get_instance());
*/
	if (display_class(leaf->get_class_name(),
			  leaf->get_instance()) == False)
	{
	GString	msg = "Object no longer exists:";
	
		msg += leaf->get_instance();
		_msg_line->error_msg(msg);
		move_up();
	}

	_msg_line->update("");

} /* refresh */

/***********************************************************/

void	Browser_Window::quit()		
{
	// Ask hosts window to delete this window

	_hosts_window->delete_browser(this);

	// delete this;

} /* quit */

/***********************************************************/

Subordinate	*Browser_Window::select_subordinate()
{
Subordinate	*result = (Subordinate *) 0;

	if (!_subordinate_browser->Count())
	{
		_msg_line->error_msg("No subordinates");
		return((Subordinate *) 0);
	}

	if (_subordinate_browser->Count() == 1)
	{
		return ((Subordinate *)_subordinate_list.get(0));
	}

	_msg_line->update("Select Subordinate");

	_subordinate_browser->UnselectAll();
	_subordinate_browser->Browse();
/*
	printf("\nAfter Select %d", _subordinate_browser->Selections());
*/
	if (_subordinate_browser->Selections())
	{
	int	count, total = _subordinate_browser->Count(); 

		for (count=0; count<total; count++)
		{
			if (_subordinate_browser->Selected(count) == true)
			{
				result = (Subordinate *)
					 _subordinate_list.get(count);
			}
		}
	}

	_subordinate_browser->UnselectAll();
	_msg_line->update("");

	return(result);

} /* select_subordinate */

/***********************************************************/

char	*Browser_Window::select_attribute()
{
char	*result = (char *) 0;

	if (!_attribute_browser->Count())
	{
		_msg_line->error_msg("No attributes");
		return((char *) 0);
	}

	_msg_line->update("Select Attribute");

	_attribute_browser->UnselectAll();
	_attribute_browser->Browse();
/*
	printf("\nAfter Select %d", _attribute_browser->Selections());
*/
	if (_attribute_browser->Selections())
	{
	int	count, total = _attribute_browser->Count(); 

		for (count=0; count<total; count++)
		{
			if (_attribute_browser->Selected(count) == true)
				result = _attribute_browser->String(count);
		}
	}

	_attribute_browser->UnselectAll();
	_msg_line->update("");

	return(result);

} /* select_attribute */

/***********************************************************/

Interactor	*Browser_Window::interior()
{
int     space = round(0.12*inch);

	_msg_line = (NEW Msg_line);

	return (NEW VBox(
		new HBox(commands(), new HGlue),
		new HBorder,
		new VGlue(space, space, 0),
		new MarginFrame(
		    new VBox(
			new HBox(
			    current_object_display()
			    ),
                	new VGlue(space, space, 0)
			),
            	    space, 0
		),
		new MarginFrame(
		    new VBox(
			new HBox(
			    attributes_display()
			    ),
                	new VGlue(space, space, 0)
			),
            	    space, 0
		),
		new MarginFrame(
		    new VBox(
			new HBox(
			    subordinates_display()
			    ),
                	new VGlue(space, space, 0)
			),
            	    space, 0
		),
		_msg_line
        ));

} /* interior */

/***********************************************************/

Interactor	*Browser_Window::commands()
{
static struct   PullDownOptions modify_menu[] =
{
        { "Change",     XCALL_BCHANGE},
        { "Set Default",XCALL_BSETDEFAULT},
        { 0,            0}

};
static struct   Menu_Options browser_menu[] =
{
        { "Up",         XCALL_BUP,              0},
        { "Down",       XCALL_BDOWN,            0},
        { "Modify",     0, (struct PullDownOptions *) &modify_menu},
        { "Monitor",    XCALL_BPOLLMONITOR,	0},
        { "Refresh",    XCALL_BREFRESH,         0},
        { "Quit",       XCALL_BQUIT,            0},
        { 0,            0,                      0}
};

	_menu = menu_initialise(browser_menu);

	return(_menu);

} /* commands */

/***********************************************************/

Interactor	*Browser_Window::current_object_display()
{
ButtonState	*bstate = (NEW ButtonState);

	_current_object = (NEW StringBrowser(bstate, 1, 25, false));

	return (NEW VBox(
		new HBox(
		    new Message("Current Object"),
		    new HGlue
		),
		new Frame(_current_object)
    	));

} /* current_object_display */

/***********************************************************/

Interactor	*Browser_Window::attributes_display()
{
ButtonState	*bstate = (NEW ButtonState);

	_attribute_browser = (NEW StringBrowser(bstate, 25, 50));

	return (NEW VBox(
		new HBox(
		    new Message("Attributes"),
		    new HGlue
		),
		new Frame(
		    add_vscroller(_attribute_browser)
		)
    	));

} /* attributes_display */

/***********************************************************/

Interactor	*Browser_Window::subordinates_display()
{
ButtonState	*bstate = (NEW ButtonState);

	_subordinate_browser = (NEW StringBrowser(bstate, 15, 40));

	return (NEW VBox(
		new HBox(
		    new Message("Subordinate Objects"),
		    new HGlue
		),
		new Frame(
		    add_vscroller(_subordinate_browser)
		)
    	));

} /* subordinates_display */

/***********************************************************/

Interactor	*Browser_Window::add_vscroller(Interactor* i)
{
        return (NEW HBox(
                new MarginFrame(i, 2),
                new VBorder,
                new VBox(
                    new UpMover(i, 1),
                    new HBorder,
                    new VScroller(i),
                    new HBorder,
                    new DownMover(i, 1)
                )
        ));

} /* add_vscroller */

/***********************************************************/

Interactor	*Browser_Window::add_hscroller(Interactor* i)
{
        return (NEW VBox(
                new MarginFrame(i, 0, 2),
                new HBorder,
                new HBox(
                    new LeftMover(i, 1),
                    new VBorder,
                    new HScroller(i),
                    new VBorder,
                    new RightMover(i, 1)
                )
        ));

} /* add_hscroller */

/***********************************************************/

Boolean	Browser_Window::display_class(GString class_name, GString instance)
{
Cmis_result	result, subordinate_result;

/*
printf("\nCLASS %s INSTANCE %s", (char *) class_name, (char *) instance);
*/
        if (cmis_read(class_name, instance, GET_ALLATTRS, 0, &result) == False)
		return(False);

        GString  class_instance = class_name + " " + instance;
/*
printf("\nAFTER CLASS %s INSTANCE %s", (char *) class_name, (char *) instance);
*/
	_current_object->Clear();
	_current_object->Append((char *) class_instance);
	_attr_display.display_attributes(_attribute_browser, &result);

	_subordinate_list.clear();	// lear subordinate list

	if (cmis_read(class_name, instance, GET_NOATTRS, 1,
		      &subordinate_result) == True)
		display_subordinates(&subordinate_result);
	else
	{
		_subordinate_browser->Clear(); 
		_msg_line->error_msg("Cannot retrieve subordinates");
	}

	return(True);

} /* display_class */

/***********************************************************/

Boolean Browser_Window::cmis_read(GString class_name, GString instance,
				  int nattrs, int level, Cmis_result *result)
{
Boolean	success;

	_hosts_window->set_alarm_off(); // switch off any alarms

	success = _cmis_client->get(class_name, instance, nattrs, 
				    level, result);

	_hosts_window->set_alarm_on();

	return(success);

} /* cmis_read */

/***********************************************************/

void	Browser_Window::display_subordinates(Cmis_result *result)
{
int             obj_count;
Cmis_Object     *object;
int		max_lgth = 0;

	_subordinate_browser->Clear(); 

	for (obj_count=0; object=result->get(obj_count); obj_count++)
	{
	int	lgth;

		if ((lgth=strlen(object->print_class())) > max_lgth)
			max_lgth = lgth;
	}

	max_lgth += 5;	/* column separator */

	for (obj_count=0; object=result->get(obj_count); obj_count++)
	{
	GString	subordinate_display = object->print_class();
	int	lgth = strlen(object->print_class());
	char	*rdn = find_rdn(object->print_instance());
	Subordinate *subordinate = (NEW Subordinate(object->print_class(),
						    object->print_instance()));

		while (lgth++ < max_lgth)
			subordinate_display += ' ';

		subordinate_display += rdn;
		_subordinate_list.add(subordinate);

		_subordinate_browser->Append((char *) subordinate_display); 
/*
		printf("\n\nObject %d class %s Instance %s",
			obj_count, object->print_class(),
			object->print_instance());
*/
	}

} /* display_subordinates */

/***********************************************************/

char	*Browser_Window::find_rdn(char *dn)
{
char	*rdn;

	for (rdn=dn+(strlen(dn)-1); rdn != dn; rdn--)
		if (*rdn == '@')
			return(rdn+1);

	return (dn);
	
} /* find_rdn */

/***********************************************************/
/* End */
/***********************************************************/


