//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/***********************************************************/
/* 
	File:	ivbrowser.h

	Notes:	Interviews browser class header file

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/***********************************************************/
/* Define X Callback offsets */

#define XCALL_BCHANGE           1
#define XCALL_BSETDEFAULT       2
#define XCALL_BPOLLMONITOR      3
#define XCALL_BUP               4
#define XCALL_BDOWN             5
#define XCALL_BREFRESH          6
#define XCALL_BQUIT             7

/***********************************************************/

class	Leaf
{
	GString	_class_name;
	GString	_instance;
public:
	Leaf(GString class_name, GString instance)
	{
		_class_name = class_name;
		_instance = instance;
	}

	char	*get_class_name()	{return((char *) _class_name);}
	char	*get_instance()		{return((char *) _instance);}
};

/***********************************************************/

class	Hierarchy : public Ptrlist
{
public:
        Hierarchy()	{}
        ~Hierarchy()
        {
                while (vdelete(0))
                        ;
        }

        int     vdelete(int index)
        {
        Leaf	*leaf;

                if (leaf=(Leaf *)get(index))
                {
                        delete leaf;
                        remove(index);
                        return(TRUE);
                }
		else
                	return(FALSE);
        }
};

/***********************************************************/

class	Subordinate
{
	GString _class_name, _instance;
public:
	Subordinate(char *class_name, char *instance)
	{
		_class_name = class_name;
		_instance = instance;
	}
	char	*get_class()
	{
		return ((char *) _class_name);
	} 
	char	*get_instance()
	{
		return ((char *) _instance);
	} 
};

class	SubordinateList : public Ptrlist
{
public:
        SubordinateList()	{}
        ~SubordinateList()
        {
                clear();
        }

	void	clear()
	{
		while (vdelete(0))
			;
	}

        int     vdelete(int index)
        {
        Subordinate	*subordinate;

                if (subordinate=(Subordinate *)get(index))
                {
                        delete subordinate;
                        remove(index);
                        return(TRUE);
                }
		else
                	return(FALSE);
        }
};
/***********************************************************/
/* Browser Window Class */
/***********************************************************/

class	Browser_Window : public Menu_Window
{
	Cmis_client	*_cmis_client;
	Hierarchy	_hierarchy;
	Monitor_list	_monitor_list;
	SubordinateList _subordinate_list;
	MenuBar		*_menu;
	StringBrowser	*_attribute_browser;
	Attr_Display	_attr_display;
	StringBrowser	*_subordinate_browser;
	StringBrowser	*_current_object;
	Msg_line	*_msg_line;
	Interactor	*_this_interactor;
	Hosts_Window	*_hosts_window;
	StringDialog	*_modify_editor;
	long		_next_alarm;

	Interactor	*interior();
	Interactor	*commands();
	Interactor	*current_object_display();
	Interactor	*attributes_display();
	Interactor	*subordinates_display();
	Interactor	*add_vscroller(Interactor *);
	Interactor	*add_hscroller(Interactor *);

	char		*select_attribute();
	Subordinate	*select_subordinate();
	Boolean		display_class(GString, GString);
	void		display_subordinates(Cmis_result *);
	char		*find_rdn(char *);
	void		initialise(GString, GString, GString, GString, Boolean &);
	Boolean		set_attribute(char *, char *);
public:
	Browser_Window(Hosts_Window *hosts_window, GString host, 
		       GString service, GString top,
		       GString top_dn, Boolean &result)
	{
		_hosts_window = hosts_window;
		_next_alarm = 0;
		initialise(host, service, top, top_dn, result);
		_modify_editor = (NEW StringDialog("Modify Attribute", 
						  round(3*inches)));
	}
	~Browser_Window()
	{
		_cmis_client->terminate();
		delete _cmis_client;
		delete _modify_editor;
		if (_this_interactor)
			GetWorld()->Remove(_this_interactor);
	}

	void	Xcallback(int);
	void	move_up();
	void	move_down();
	void	modify();
	void	set_default();
	void	monitor();
	void	refresh();
	void	quit();
        int    refresh_monitor(long);

        void    delete_monitor(Monitor_Window *monitor)
        {
                _monitor_list.delete_monitor(monitor);
		if (_monitor_list.count() == 0)
			set_signal_alarm(0);
        }
	Boolean	cmis_read(GString, GString, int, int, Cmis_result *);

	void	set_signal_alarm(int); 
	long	get_signal_alarm()	{return(_next_alarm);}
};

/***********************************************************/
/* End */
/***********************************************************/


