//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/***********************************************************/
/* 
	File:	ivhosts.cc

	Notes:

		Interviews host window methods

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)

*/
/***********************************************************/
/* Hosts window */
/***********************************************************/

#ifdef AIXV3
extern "C" { strcasecmp (const char*, const char*); }
#endif

/***********************************************************/

void	Hosts_Window::initialise(GString hosts_table_name)

{
int	space = round(1*inch);

	read_hosts(hosts_table_name);

	input = (NEW Sensor);

	input->Catch(DownEvent);
	input->Catch(UpEvent);

	_this_interactor = interior();
	Insert(_this_interactor);

} /* initialise */

/***********************************************************/

void	Hosts_Window::read_hosts(GString file_name)
{
FILE    *fp;
char    buf[256];

#ifdef USE_X500

DNlist  Agentlist;
extern  char    *lt_local_DIT; /* read from the dsaptailor file */
char    success = FALSE;

fprintf(stderr, "\nSearching the Directory for Management Processes (only the running ones)...\n");

if ( (Agentlist = lt_get_mp(lt_local_DIT)) != NULL) {
        DNlist ptr;

        success = TRUE;

        for ( ptr = Agentlist; ptr != NULL; ptr = ptr->next ) {
        char    *host,
                *agent;
        Host_Agent      *host_agent;

        agent = "@";
        host = strdup(ptr->name);
        host_agent = (NEW Host_Agent(host, agent));

        _hosts_list.add(host_agent);
        }

        if (!_hosts_list.count()) {
          fprintf(stderr, "\nNo information found in the Directory!?\n");
          success = FALSE;
        }
} else
  fprintf(stderr, "no runnable process found.\n");

if ( !success ) {
        fprintf(stderr, "\nReading hosts file %s", (char *) file_name);

        if ((fp=fopen(file_name, "r")) == 0)
        {
                fprintf(stderr, "\nCannot open %s", (char *) file_name);
                exit(1);
        }

        while (fgets(buf, sizeof(buf), fp) != (char *) 0)
        {
                if (buf[0] != '#' && buf[0] != '\n')
                {
                char    host[50], agent[50];

                        if (sscanf(buf, "%s %s", host, agent) == 2)
                        {
                        Host_Agent      *host_agent =
                                        (NEW Host_Agent(host, agent));

                                if (strcasecmp(host, "default") == 0 ||
                                    host[0] == '"' ||
                                    strcasecmp(agent, "default") == 0 ||
                                    agent[0] == '"')
                                        continue;

                                _hosts_list.add(host_agent);
                        }
                        else
                        {
                                /*
                                printf("\nError reading %s Line: %s",
                                        (char *) file_name, buf);
                                exit(1);
                                */
                                continue;
                        }
                }
        }

        if (!_hosts_list.count())
        {
                fprintf(stderr, "\nNo information in %s",
                        (char *) file_name);
                exit(1);
        }

        fclose(fp);
}

#else /* in case USE_X500 is undefined use the isoentities as a back up */

        fprintf(stderr, "\nReading hosts file %s", (char *) file_name);

        if ((fp=fopen(file_name, "r")) == 0)
        {
                fprintf(stderr, "\nCannot open %s", (char *) file_name);
                exit(1);
        }

        while (fgets(buf, sizeof(buf), fp) != (char *) 0)
        {
                if (buf[0] != '#' && buf[0] != '\n')
                {
                char    host[50], agent[50];

                        if (sscanf(buf, "%s %s", host, agent) == 2)
                        {
                        Host_Agent      *host_agent =
                                        (NEW Host_Agent(host, agent));

                                if (strcasecmp(host, "default") == 0 ||
                                    host[0] == '"' ||
                                    strcasecmp(agent, "default") == 0 ||
                                    agent[0] == '"')
                                        continue;

                                _hosts_list.add(host_agent);
                        }
                        else
                        {
                                /*
                                printf("\nError reading %s Line: %s",
                                        (char *) file_name, buf);
                                exit(1);
                                */
                                continue;
                        }
                }
        }

        if (!_hosts_list.count())
        {
                fprintf(stderr, "\nNo information in %s",
                        (char *) file_name);
                exit(1);
        }

        fclose(fp);

#endif /* USE_X500 */

} /* read_hosts */

/***********************************************************/

void Hosts_Window::Xcallback(int option)
{
        switch (option)
        {
        case XCALL_HCONNECT:
                connect();
                break;
        case XCALL_HQUIT:
                quit();
                break;
        }

} /* xcallback */

/***********************************************************/
/* Connect to host */

void	Hosts_Window::connect()		
{
Browser_Window	*browser;
Host_Agent      *host_agent;

        if (select_host(&host_agent))
	{
	Boolean	result;
        GString  temp = host_agent->get_host();
#ifdef USE_X500
        GString  str;

        if (strcmp (host_agent->get_agent(), "@") == 0 )
          str = temp;
        else
          str = temp + "-" + host_agent->get_agent();
#else
        GString  str = temp + "-" + host_agent->get_agent();
#endif
        GString  msg = "Trying to connect to " + str;

		_msg_line->update((char *) msg);

                browser = (NEW Browser_Window(this, host_agent->get_host(),
                                             host_agent->get_agent(),
                                             "", TOP_DN, result));

		if (result == True)
		{
			_msg_line->update("");
			_browser_list.add(browser);
			GetWorld()->Insert(browser);
			browser->SetName((char *)str);
		}
		else
		{
		GString	msg = "Cannot connect to ";

			msg += str;
			_msg_line->error_msg((char *) msg);
			delete browser;
		}
	}

} /* connect */

/***********************************************************/
/* Quit */

void	Hosts_Window::quit()
{
/*
	printf("\nQuit");
*/
	set_alarm_off();
	_done = True;
/*
	cmisbrowser_exit(0);
*/
} /* quit */

/***********************************************************/

Boolean	Hosts_Window::select_host(Host_Agent **host_agent)
{
Boolean	result = False;

	_msg_line->update("Select Host:Agent");
	_host_browser->UnselectAll();
	_host_browser->Browse();
/*
	printf("\nAfter Select %d", _host_browser->Selections());
*/
	if (_host_browser->Selections())
	{
	int	count, total;

		for (count=0,total=_host_browser->Count(); count<total; count++)
		{
			if (_host_browser->Selected(count) == true)
			{
			Host_Agent	*selected;

				selected = (Host_Agent *)_hosts_list[count];
				*host_agent = selected;
				result = True;
				break;
			}
		}
	}

	_host_browser->UnselectAll();
	_msg_line->update("");
	return(result);

} /* select_host */

/***********************************************************/

Interactor	*Hosts_Window::interior()
{
int     space = round(0.12*inch);

	_msg_line = (NEW Msg_line);

	return (NEW VBox(
		new HBox(commands(), new HGlue),
		new HBorder,
		new MarginFrame(
		    new VBox(
			new HBox(
			    browser()
			    ),
                	new VGlue(space, space, 0)
			),
            	space, space, 0
		),
		_msg_line
        ));

} /* interior */

/***********************************************************/

Interactor	*Hosts_Window::commands()
{
static struct   Menu_Options host_menu[] =
{
        { "Connect",    XCALL_HCONNECT, 0},
        { "Quit",       XCALL_HQUIT, 0},
        { 0,            0,      0}
};

	_menu = menu_initialise(host_menu);

	return(_menu);

} /* commands */ 

/***********************************************************/

Interactor	*Hosts_Window::browser()
{
ButtonState	*bstate = (NEW ButtonState);
int		count, total = _hosts_list.count();

	_host_browser = (NEW StringBrowser(bstate, 15, 25));

	for (count=0; count<total; count++)
        {
        Host_Agent *host_agent = (Host_Agent *)_hosts_list[count];
        GString  temp = host_agent->get_host();
#ifdef USE_X500
        GString  str;
        char    * rdnvalue = rindex(temp, '=');

        if (strcmp (host_agent->get_agent(), "@") == 0 ) {
          rdnvalue++;
          str = rdnvalue;
        } else
          str = temp + "-" + host_agent->get_agent();
#else
        GString  str = temp + "-" + host_agent->get_agent();
#endif

		_host_browser->Append(str);
	}

	return (NEW VBox(
		new HBox(
		    new Message("Hosts:Agents"),
		    new HGlue
		),
		new Frame(
		    add_vscroller(_host_browser)
		)
    	));

} /* browser */

/***********************************************************/

Interactor	*Hosts_Window::add_vscroller(Interactor* i)
{
	return new HBox(
		new MarginFrame(i, 2),
		new VBorder,
		new VBox(
		    new UpMover(i, 1),
		    new HBorder,
		    new VScroller(i),
		    new HBorder,
		    new DownMover(i, 1)
		)
	);

} /* add_vscroller */

/***********************************************************/

void	Hosts_Window::Run()
{

	_done = False;

	do
	{
	Event	e;

		Read(e);
	/*
	printf("\nHosts Run");
		if (e.eventType == UpEvent)
			printf("\nUpevent");
		if (e.eventType == DownEvent)
			printf("\nDownevent");
	*/
		e.target->Handle(e);

	}
	while (_done == False);	

} /* Run */

/***********************************************************/

void	Hosts_Window::Handle(Event& /* event */)
{
/*
fprintf(stderr, "\nReceived event");
	if (event.eventType == DownEvent)
		invert();
	else
		if (event.eventType == UpEvent)
			QuitRunning(event);
*/
} /* Handle */

/***********************************************************/

void	Hosts_Window::signal_alarm()
{
Browser_Window	*browser;
int		count;
int		alarm_interval;
long		cur_time = time((long *) 0);
long		new_alarm = 0;
/*
printf("\nIn signal_alarm");
*/
	for (count=0; browser=(Browser_Window *)_browser_list[count]; count++)
	{
	int	alarm;

		alarm = browser->refresh_monitor(cur_time);

		if (!new_alarm)
			 new_alarm = alarm;
		else
		{
			if (alarm && (alarm < new_alarm))
			 	new_alarm = alarm;
		}
	}
/*
printf("\nOut of signal_alarm");
*/
	if (!new_alarm)
		alarm_interval = 0;
	else
	{
		if (new_alarm <= cur_time)
			alarm_interval = 1;
		else
			alarm_interval = (int)(new_alarm - cur_time);
	}
/*
	printf("\nHosts: Reset alarm %d time %ld alarm time %ld",
		alarm_interval, cur_time, new_alarm);
*/
	alarm(alarm_interval);		// reset alarm

} /* signal_alarm */

/***********************************************************/

void	Hosts_Window::set_alarm_on()
{
Browser_Window	*browser;
int		count;
int		alarm_interval;
long		new_alarm = 0;
long		cur_time = time((long *) 0);

	for (count=0; browser=(Browser_Window *)_browser_list[count]; count++)
	{
	long	alarm;

		alarm = browser->get_signal_alarm();

		if (!new_alarm)
			 new_alarm = alarm;
		else
		{
			if (alarm && (alarm < new_alarm))
			 	new_alarm = alarm;
		}
	}

	if (!new_alarm)
		alarm_interval = 0;
	else
	{
		if (new_alarm <= cur_time)
			alarm_interval = 1;
		else
			alarm_interval = (int)(new_alarm - cur_time);
	}
/*
	printf("\nHosts: Set signal alarm %d time %ld alarm time %ld",
		alarm_interval, cur_time, new_alarm);
*/
	alarm(alarm_interval);

} /* set_alarm_on */

/***********************************************************/

void	Hosts_Window::set_alarm_off()
{

	alarm(0);

} /* set_alarm_off */

/***********************************************************/
/* End */
/***********************************************************/


