//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/***********************************************************/
/* 
	File:	ivmonitor.cc

	Notes:	Cmis Interviews monitor class methods

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/

/***********************************************************/

Interactor	*Monitor_Window::interior(int rows, int cols)
{
int     space = round(0.12*inch);

	_msg_line = (NEW Msg_line);

	return (NEW VBox(
		new HBox(commands(), new HGlue),
		new HBorder,
		new VGlue(space, space, 0),
		new MarginFrame(
		    new VBox(
			new HBox(
			    attributes_display(rows, cols)
			    ),
                	new VGlue(space, space, 0)
			),
            	    space, 0
		),
		_msg_line
	));

} /* interior */
/***********************************************************/

void Monitor_Window::Xcallback(int option)
{
        switch (option)
        {
        case XCALL_MINTERVAL:
                interval();
                break;
        case XCALL_MQUIT:
                quit();
                break;
        }

} /* Xcallback */

/***********************************************************/

Interactor	*Monitor_Window::commands()
{
static struct   Menu_Options monitor_menu[] =
{
        { "Interval",   XCALL_MINTERVAL,        0},
        { "Quit",       XCALL_MQUIT,            0},
        { 0,            0,                      0}
};

        _menu = menu_initialise(monitor_menu);

	return(_menu);

} /* commands */

/***********************************************************/

Interactor	*Monitor_Window::attributes_display(int rows, int cols)
{
ButtonState     *bstate = (NEW ButtonState);

        _attribute_browser = new StringBrowser(bstate, rows, cols);

        return (NEW VBox(
                new HBox(
                    new Message("Attributes"),
                    new HGlue
                ),
                new Frame( _attribute_browser)
        ));

} /* attributes_display */

/***********************************************************/

void	Monitor_Window::quit()
{
        _browser->delete_monitor(this);

} /* quit */

/***********************************************************/

void	Monitor_Window::interval()
{
char    num_str[10];

	sprintf(num_str, "%d", _alarm_interval);

	_interval_editor->Read_Input(this, (Menu_Valid_Fn) &Monitor_Window::check_interval,
				    (char *) _instance, num_str);
} /* interval */

/***********************************************************/

Boolean	Monitor_Window::check_interval(char *, char *new_value)
{
int	new_interval = atoi(new_value);

	set_signal_alarm(new_interval);

	return(True);

} /* check_interval */
/***********************************************************/

void	Monitor_Window::refresh()
{
Cmis_result	result;

	_msg_line->update("Polling");

	if (_browser->cmis_read(_class_name, _instance, GET_ALLATTRS, 
				0, &result))
	{
		if (!_attr_display.display_attributes(_attribute_browser, 
						      &result))
		{
			_msg_line->update("No Change");
			return;
		}
	}
	else
	{
		_msg_line->error_msg("Object no longer exists");
		_browser->delete_monitor(this);
	}
	_msg_line->update("");
 
} /* refresh */

/***********************************************************/

void    Monitor_Window::set_signal_alarm(int secs)
{

	_next_alarm_time = time((long *) 0) + secs;
	_alarm_interval = secs;
/*
printf("\nMonitor set alarm %d next alarm %ld", 
	_alarm_interval, _next_alarm_time);
*/

	_browser->set_signal_alarm(secs);

} /* set_signal_alarm */

/***********************************************************/
/* End */
/***********************************************************/


