//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/******************************************************************/
/*
	File:	ptrlist.c

	Notes:
		implements a generic single linked list

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/*****************************************************************/


#ifdef  __GNUG__
#include <std.h>
#elif   __cplusplus
#include <stdlib.h>
#endif

#include "ptrlist.h"

/*****************************************************************/
/* Ptrlist functions */
/*****************************************************************/
/* Check list is empty */
/* This functtion used to clear a list, but you cannot
   delete objects that are being addressed generically 
   (ie via void *)
*/

void Ptrlist::check_empty()
{
	if (_list_count > 0)
	{
		printf("\nPtrlist not clear");
		exit(1);
	}

} /* check_empty */

#ifdef	OLD_VERSION

void Ptrlist::clear()
{
Ptrlink	*list_ptr, *next;

	for (list_ptr=_start; list_ptr; list_ptr=next)
	{
		next = list_ptr->_next;
		delete list_ptr;
	}

	_start = _current = _end = 0;
	_list_count = 0;

} /* clear */

#endif

/*****************************************************************/
/* Add new record */

void	Ptrlist::add(PTR new_value)
{
	if (_start)
		_end = _end->_next = (NEW Ptrlink(new_value, 0));
	else
		_start = _end = (NEW Ptrlink(new_value, 0));

	_current = _end;
	_list_count++;

} /* add */

/*****************************************************************/
/* Remove a particular record */

int     Ptrlist::remove(PTR value)
{
Ptrlink *ptr, *prev;

        for (prev=ptr=_start; ptr; prev = ptr, ptr = ptr->_next)

                if (ptr->_value == value)
                        break;

        if (!ptr)
                return(FALSE);

        if (ptr == _start)
                _start = _start->_next;
        else
        {
                if (ptr == _end)
                {
                        _end = prev;
                        prev->_next = 0;
                }
                else
                        prev->_next = ptr->_next;
        }

        delete ptr;
        _list_count--;
        return(TRUE);

} /* remove */

/*****************************************************************/
/* Remove a particular record */

int	Ptrlist::remove(int index)
{
Ptrlink	*ptr, *prev;
int	count;

	if (index >= _list_count)
		return(FALSE);

	for (prev=ptr=_start, count=0; count<index; count++)
	{
		prev = ptr;
		ptr = ptr->_next;
	}

	if (ptr == _start)
		_start = _start->_next;
	else
	{
		if (ptr == _end)
		{
			_end = prev;
			prev->_next = 0;
		}
		else
			prev->_next = ptr->_next;
	}

	delete ptr;
	_list_count--;
	return(TRUE);

} /* remove */

/*****************************************************************/
/* Get first record */

PTR	Ptrlist::first()
{
	if (!_start)
		return(0);

	_current = _start;

	return(_start->_value);

} /* first */

/*****************************************************************/
/* Get last record */

PTR	Ptrlist::last()
{
	if (!_start)
		return(0);

	_current = _end;

	return(_end->_value);

} /* last */

/*****************************************************************/
/* Get next record */

PTR	Ptrlist::next()
{

	if (!_current)
		return(0);

	if (!_current->_next)
		return(0);

	_current = _current->_next;

	return(_current->_value);

} /* next */

/*****************************************************************/
/* Extremely inefficient array subscripting */

PTR	Ptrlist::get(int index)
{
int	count;
Ptrlink	*ptr;

	if (!_list_count || index >= _list_count)
		return(0);

	for (ptr=_start, count=0; count<index; count++, ptr=ptr->_next)
		;

	return(ptr->_value);

} /* get */

/*****************************************************************/
/* Find an object in a list */

int	Ptrlist::find(PTR object)
{
int	count;
Ptrlink	*ptr;

	for (ptr=_start, count=0; count<_list_count; count++, ptr=ptr->_next)
		if (ptr->_value == object)
			return(count);

	return(-1);

} /* find */

/*****************************************************************/
/* End of module */
/*****************************************************************/


