//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/******************************************************************/
/*	

	File:		ptrlist.h

	Header file for ptrlist.cc

	2 classes are defined:
		Ptrlink class: link in a list
		Ptrlist class: list of links

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/*****************************************************************/

#ifdef AIXV3
#include <stdio.h>
#endif
#include "gen.h"

/*****************************************************************/
/* Class link */

class	Ptrlink
{
	friend	class	Ptrlist;

	Ptrlink	*_next;
	PTR	_value;
public:

	/* constructor */

	Ptrlink(PTR value, Ptrlink *next) 
	{
		_value = value;
		_next = next;
	}
};

/*****************************************************************/
/* Class list */

class	Ptrlist
{
	Ptrlink	*_start, *_end, *_current;
	int	_list_count;
	void	check_empty();
protected:
	int	remove(int);		/* remove element from list */
	int	remove(PTR);		/* remove element from list */
public:
	/* constructors */

	Ptrlist()
	{
		_list_count=0; _start=_current=_end=0;
	}

	/* destructor */

	virtual ~Ptrlist()	
	{
		check_empty();
	}

	void	add(PTR);		/* add new element at end of list */
	virtual	int	vdelete(int)	/* delete element */
	{
		printf("\nVdelete cannot be used in base class");
		exit(1);	
		return(0);
	}
	int	count()
	{
		return(_list_count);
	}
	PTR 	first();		/* get first record */
	PTR 	last();			/* get last record */
	PTR 	next();			/* get next record */
	PTR	get(int);		/* get */
	PTR	operator[](int index)	/* array subscript operator */
	{
		return(get(index));
	}
	int	find(PTR);
};

/*****************************************************************/
/* End of module */
/*****************************************************************/


