//
// COPYRIGHT (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// MibDump.cc - enables to read managed objects/attributes from a remote MIB

// By George Pavlou, February 1991


#include <stdlib.h>
#include <string.h>
#include <signal.h>

#ifdef	__GNUG__
#ifndef __HPUX__
#define SIGHANDLER	SignalHandler
#else
#define SIGHANDLER      void
#endif
#include <std.h>
extern "C" { int  strncasecmp (const char*, const char*, size_t); }
#elif	__cplusplus
#ifndef AIXV3
#define SIGHANDLER	SIG_PF
#else	/* AIXV3 */
#define SIGHANDLER	void
#endif
#endif

#include "msap.h"
#include "Error.h"
#include "Syntaxes.h"
#include <memory.h>
#if defined(SECURITY)
#include "mauth.h"
#include "mintegrity.h"
#endif

#if defined(DEBUG)
#include "msap-pp.h"
FILE *pduDump = (FILE *) 0;
#endif

#define MAXATTRS	50

int	msd,
        request_id = 0,     /* Sent by mibdump */
        request_invoke = 0, /* For use in Cancel Get */
        invoke = 0;         /* Received from agent */
Bool cancel_get = False;
MSAPIndication mis;
Bool	timing = False;

int   parseFilterArg (int, char**, char**, int*);
static int   do_get (int, OID, MN, CMISScope*, CMISFilter*, CMISSync,
			int, MID, MSAPIndication*);
static int   do_wait (int, MSAPIndication*);

static void  do_cancelget (int);

main (int argc, char** argv)
{
    char* application = NULLCP;
    char* host = NULLCP;
    char* strObjClass = NULLCP;
    char* strObjInst = NULLCP;
    char* strScope = NULLCP;
    char* strSync = NULLCP;
    char* strFilter = NULLCP;
    char* strAttrs[MAXATTRS];
    int   nattrs = GET_ALLATTRS;	// default
    int   i, j, k, l, state = 0;

    OID objClass = NULLOID;
    MN  objInst  = NULLMN;
    CMISScope* scope = NULLMSCOPE;	// default
    CMISSync sync = s_bestEffort;	//  ..
    CMISFilter* filter = NULLMFILTER;	//  ..
    struct MIDentifier attrs[MAXATTRS];

#ifdef SECURITY
    int IFlag = 0;
    char *peer = (char *) 0,
         *dsa = (char *) 0;
#endif

    // parse command line arguments

    if (argc < 3)
	goto usage;

    if (strcmp(argv[1], "-S") == 0)
	application = argv[2];

    if (argc > 3) {
	for (i = 3; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'c':
			state = 1;
			break;
		    case 'i':
			state = 2;
			break;
		    case 's':
			state = 3;
			break;
		    case 'f':
			state = 4;
			break;
		    case 'a':
			state = 5;
			break;
		    case 't':
			timing = True;
			break;
#ifdef SECURITY
                    case 'P':
                        state = 6;
                        break;
                    case 'D':
                        state = 7;
                        break;
                    case 'I':
                        state = 0;
                        IFlag = 1;
                        break;
#endif
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		strObjClass = argv[i];
		state = 0;
		break;

	    case 2:
		strObjInst = argv[i];
		state = 0;
		break;

	    case 3:
		strScope = argv[i];
		if (i < argc-1 && *argv[i+1] != '-')
		    strSync = argv[++i];
		state = 0;
		break;

	    case 4:
		if (parseFilterArg(argc-i, &argv[i], &strFilter, &j) == NOTOK)
		    goto usage;
		i += j-1;
		state = 0;
		break;

	    case 5:
	    {
		Bool more = False;
		for (j = i; ; j++)
		    if (*argv[j] == '-' && strlen(argv[j]) == 2) {
			more = True;
			break;
		    }
		    else
		    if (j == argc-1)
			break;

		nattrs = more ? j-i : j-i+1;
		if (nattrs > MAXATTRS)
		    error(NULLCP, "too many attrs - max number %s", MAXATTRS);

		for (k = i, l = 0; k < j+1; k++, l++)
		    strAttrs[l] = argv[k];
		state = 0;
		i = more ? j-1 : j;
	    }
		break;
#ifdef SECURITY
            case 6:
                peer = argv[i];
                state = 0;
                break;
            case 7:
                dsa = argv[i];
                state = 0;
                break;
#endif
	    }
    }


    // load syntaxes

    if (initialiseSyntaxes(NULLCP, attrSyntaxes) == NOTOK)
	error(NULLCP, "mibdump: cannot initialise syntaxes");

#ifdef SECURITY
    initialiseOsisec(argc, argv);
    initialiseMd5KeySpace();
    initialiseIdSpace();
#endif

    // convert the string arguments to CMIS parameters

    if (strObjClass && (objClass = name2oid(strObjClass)) == NULLOID)
	error(NULLCP, "invalid object class: %s", strObjClass);

    if (strObjInst && (objInst = ObjectInstance_parse(strObjInst)) == NULLMN)
        error(NULLCP, "invalid object instance %s", strObjInst);

    if (strScope && (scope = str2scope(strScope)) == NULLMSCOPE)
	error(NULLCP, "invalid scope %s", strScope);

    if (strSync && (sync = str2sync(strSync)) == s_invalid)
	error(NULLCP, "invalid sync %s", strSync);
	
    if (strFilter && strFilter[0] != '\0'  &&
		(filter = str2mfilter(strFilter)) == NULLMFILTER)
	error(NULLCP, "invalid filter %s", strFilter);

    if (nattrs > 0 && strcasecmp(strAttrs[0], "none") == 0)
	nattrs = GET_NOATTRS;
    else
	for (i = 0; i < nattrs; i++) {
	    if ((attrs[i].mid_global = name2oid(strAttrs[i])) == NULLOID)
		error(NULLCP, "invalid attribute name %s", strAttrs[i]);
	    attrs[i].mid_type = MID_GLOBAL;
	}


    // now try to connect to the remote agent

    MInitOption *miop[2]; /* Connect options */

#ifdef SECURITY
    MInitOption mio;
    AuthAssocIntegrityInfo *aaii;
    aaii = AuthAssocIntegrityInfo_make();

    if (peer != (char *) 0)
        aaii->peer = strdup(peer);
    if (dsa != (char *) 0)
        aaii->dsa = strdup(dsa);
    if (IFlag)
        memcpy((void *) aaii->sessionKey, (void *) makeMd5Key(), 8);

    mio.optionId = MIO_AuthAssocIntegrityInfo;
    mio.optionInfo = (void *) aaii;

    miop[0] = &mio;
    miop[1] = (MInitOption *) 0;

#else

   miop[0] = miop[1] = (MInitOption *) 0;

#endif

    if (timing)
	do_timing(NULLCP);

    host = application ? NULLCP : argv[2];
    application = application ? application : argv[1]; 

    if ((msd = m_initialise(application, host, miop)) < 0)
	error(NULLCP, "cannot establish association to %s at %s",
			argv[1], argv[2]);
    if (timing)
	do_timing("established association");

#if defined(SECURITY)
    if (aaii != (AuthAssocIntegrityInfo *) 0)
        AuthAssocIntegrityInfo_free(aaii);
#endif

    // perform the M-GET request

    if (!timing)
        printf("\n\n");

    do_get(msd, objClass, objInst, scope, filter, sync, nattrs, attrs, &mis);

    // set signal handlers

#ifdef	__GNUG__
    signal(SIGINT,  do_cancelget);
    signal(SIGQUIT, do_cancelget);
    signal(SIGTERM, do_cancelget);
#else
    signal(SIGINT,  (SIGHANDLER) do_cancelget);
    signal(SIGQUIT, (SIGHANDLER) do_cancelget);
    signal(SIGTERM, (SIGHANDLER) do_cancelget);
#endif

    // wait for the replies
#if defined(DEBUG2)
    if (!timing && (pduDump = fopen("mibdump.rx.log", "w")) != (FILE *) 0)
        ppSetOutputFile(pduDump);
#endif

    while (do_wait(msd, &mis) != DONE);

    request_id = 0;	// not to CancelGet anymore
    if (timing)
	do_timing("retrieved objects");

    // terminate

    m_terminate(msd);	// release association
    if (timing)
	do_timing("released association");

#if defined(DEBUG2)
    if (pduDump != (FILE *) 0)
        fclose(pduDump);
#endif

    exit (OK);

usage:
    fprintf(stderr, "usage: mibdump { <agent> <host> | -S <agent> }\n\
\t[-c <class>] [-i <instance>]\n\
\t[-s <scope> [<sync>]] [-f filter] [-a attr ...] ]\n");
#if defined(SECURITY)
    fprintf(stderr, "\t[-P <peer DN>] [-D <dsa name/addr>] [-I]\n");
#endif
    exit(0);
}

static int  do_get (int msd, OID objClass, MN objInst,
			CMISScope* scope, CMISFilter* filter, CMISSync sync,
			int nattrs, MID attrs, MSAPIndication* mi)
{
    struct MIDentifier obj_class;
    int		status;

    obj_class.mid_type = MID_GLOBAL;
    obj_class.mid_global = objClass;

    if (!objInst) {
        objInst = new  MName;
        objInst -> mn_type = MN_LOCALDN;
        objInst -> mn_dn = dn_comp_new(NULLRDN);
    }

#if defined(SECURITY)
    request_id = makeInitiatorId(msd, request_id);
#else
    request_id = 1;
#endif

    status = M_Get(msd, request_id, &obj_class, objInst, scope, filter,
			NULLMACCESS, sync, nattrs, attrs, mi);
    request_invoke = mi->mi_invoke;

    // release memory of arguments (for the sake of it...)

    mid_free(&obj_class);
    mn_free(objInst);
    delete objInst;
    if (scope)
	free((char*) scope);
    if (filter)
	mfilter_free(filter);
    for (int i = 0; i < nattrs; i++)
	mid_free(&attrs[i]);

    return status;
}

static int  do_wait (int msd, MSAPIndication* mi)
{
    int status = OK;
    static int nobjects = 0;
    PE pdu;

    M_WaitReqAux(msd, NOTOK, mi, &pdu);

    int mask = sigblock(0x7fffffff);

#if defined(DEBUG)
    if (pduDump != (FILE *) 0)
        ppMSAPIndication(mi);
#endif

#if defined(SECURITY)
    invoke = makeResponderId(msd, invoke);

    MD5Value md5;

    memset((void *) &md5, 0, sizeof(MD5Value));

    if (mi->mi_invoke != 0) {

        if (pdu == (PE) 0) {
            /*
            ** Empty Result.
            **
            ** The last of a series of linked replies or a Get
            ** that returned nothing.
            */
            if (mi->mi_type == MI_EMPTY_RES) {
                printf("Empty Result: ");
                pdu = int2prim(request_id);
            }

            /*
            ** Error.
            **
            ** Some error cases do not have any info and pdu will be NULL.
            */
            else if (mi->mi_type == MI_ERROR)
                pdu = int2prim(nobjects == 0 ? request_id : invoke);

            else
                printf("Unexpected MSAPIndication type: %d\n", mi->mi_type);
        }

        if (makeMd5Value(msd, pdu, &md5) != 0)
            printf("Could not make integrity check for PDU\n");

        register
        int chk1 = ENCODE_ID(invoke, md5.md5_words[0]),
            chk2 = ENCODE_ID(request_id, md5.md5_words[0]);

        register
        int i = mi->mi_invoke == chk1 ? invoke :
            (mi->mi_invoke == chk2 ? request_id : 0);

        printf("check %s - invokeId=%#08x id=%#08x chk=%#08x\n",
               i == 0 ? "failed" : "ok",
               (u_int) mi->mi_invoke, (u_int) i, md5.md5_words[0]);

        if (i == 0)
            printf("chk1=%#08x chk2=%#08x invoke=%#08x request_id=%#08x\n",
                   (u_int) chk1, (u_int) chk2,
                   (u_int) invoke, (u_int) request_id);
    }
#endif

    sigsetmask(mask);

    if (pdu != (PE) 0)
        pe_free(pdu);

    if (mi->mi_type == MI_GET_RES ||
        (mi->mi_type == MI_ERROR && mi->mi_errtype == m_getListError) ||
        mi->mi_type == MI_LINKED_REPLY) {

        struct CMISGetRes * g;
	int	id;

        switch (mi->mi_type) {
        case MI_GET_RES:
        case MI_ERROR:
	    g = &mi -> mi_getres.result;
	    id = mi -> mi_getres.id;
	    if (!timing)
		printf("Get Result - id %#08x, ", (u_int) id);
	    if (!cancel_get)
		status = DONE;	// only one object in remote MIB
	    break;

        case MI_LINKED_REPLY:
	    id = mi -> mi_linkedreplyinv.args.lr_linkid;
	    switch (mi -> mi_linkedreplyinv.args.lr_type) {
	    case LR_GET_RES:
	    case LR_GET_LISTERR:
		g = &mi -> mi_linkedreplyinv.args.lr_getres;
		break;
		
	    case LR_PROC_FAILURE:
		// should not really happen
		printf("Linked Reply Processing Failure - id %#08x\n",
                       (u_int) id);
		return status;
	    }
	    if (!timing)
		printf("Linked Reply Get - id %#08x, ", (u_int) id);
	    break;
        }

	++nobjects;

	if (!timing) {
            if (g->gr_class.mid_global)
                printf("class %s, ", oid2name(g->gr_class.mid_global, OIDPART));
            if (g->gr_inst.mn_dn) {
	        char * inst;
	        inst = dn2str(g->gr_inst.mn_dn);
	        printf("instance {%s}", inst);
	        free(inst);
            }
            if (g->gr_time)
                printf(", time %s\n", g->gr_time);
            else
	        printf("\n");
	}
        for (int i = 0; i < g->gr_nattrs; i++)
	    if (!g->gr_attrs[i].ga_error) {
		char *attrid, *attrval;
		ava2str(g->gr_attrs[i].ga_ava.mp_id.mid_global,
			    g->gr_attrs[i].ga_ava.mp_val,
			    &attrid, &attrval);
		if (!timing)
		    fprintf(stdout, "%s:  %s\n", attrid, attrval);
		free(attrid); free(attrval);
	    }
	    else
	    if (!timing)
		fprintf(stdout, "%s: Error: %s\n",
			    oid2name(g->gr_attrs[i].ga_ava.mp_id.mid_global,
				    OIDPART),
			    MErrString(g->gr_attrs[i].ga_error));
	if (!timing)
            printf("\n\n");
    }
    else
    if (mi->mi_type == MI_EMPTY_RES) {
	if (mi->mi_emptyres.id ==
#if defined(SECURITY)
            ENCODE_ID(request_id, md5.md5_words[0])
#else
            request_id
#endif
            )
            status = DONE;
	else
	    if (mi->mi_emptyres.id ==
#if defined(SECURITY)
                ENCODE_ID(invoke, md5.md5_words[0])
#else
                invoke
#endif
                )
		status = DONE;
    }
    else
    if (mi->mi_type == MI_PREJECT || mi->mi_type == MI_ABORT) {
	fprintf(stderr, "mibdump: received %s, exits\n",
                mi->mi_type == MI_PREJECT ?
                "provider reject" : "association abort");
#if defined(DEBUG2)
        if (pduDump != (FILE *) 0)
            fclose(pduDump);
#endif
        exit(NOTOK);
    }
    else
    if (mi->mi_type == MI_ERROR) {
	fprintf(stderr, "Error: %s\n", MErrString(mi->mi_errtype));
	if (!cancel_get || mi->mi_errtype == m_noSuchInvokeId)
	    status = DONE;
    }
    else {
	printf("Unexpected mindication type: %d\n", mi->mi_type);
	status = NOTOK;
    }

    if (status == DONE && !timing)
	printf("%d replies received\n", nobjects);
    mi_free(mi);
    return status;
}


static void  do_cancelget (int)
{
    if (!cancel_get) {
#if defined(SECURITY)
        request_id = makeInitiatorId(msd, request_id);
#else
        ++request_id;
#endif
	M_CancelGet(msd, request_id, request_invoke, &mis);
        cancel_get = True;
    }
}
