//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// EvSink.cc - a primitive event sink that receives and prints
//	       event reports from a remote MIB

// By George Pavlou, February 1991


#include <stdlib.h>
#include "msap.h"
#include "Error.h"
#include "Coordinator.h"
#include "RMIBAgent.h"
#include "EventMgr.h"
#include "Syntaxes.h"



main (int argc, char** argv)
{
    char* application = NULLCP;
    char* host = NULLCP;

    if (argc < 3)
	error(NULLCP, "\nusage: evsink { <agent> <host> | -S <agent> } \
[ <eventype> ... ]");

#ifdef USE_X500
    if (strcmp(argv[1], "-S") == 0)
        application = argv[2];
#endif

    // load syntaxes
    initialiseSyntaxes(NULLCP, evrepSyntaxes);

    // instantiate the coordinator object (always needed with the RMIB)
    Coordinator  coordinator;

    // instantiate one remote MIB agent (we deal with one remote MIB)
    RMIBAgent	remoteAgent;

    // initialise the remote MIB agent
    host = application ? NULLCP : argv[2];
    application = application ? application : argv[1];

    if (remoteAgent.connect(application, host) == NOTOK)
	error(NULLCP, "cannot establish association to %s at %s",
			argv[1], argv[2]);

    // instantiate the event manager
    EventMgr eventMgr(&remoteAgent);

    // tell it to listen for the events specified
    if (argc > 3 && strcmp(argv[3], "all") != 0) {
	for(int i = 3; i < argc; i++)
	    if (eventMgr.receiveEvent(NULLCP, NULLCP, argv[i]) == NOTOK)
		error(NULLCP, "evsink: %s: unknown event type \"%s\"", argv[i]);
    }
    else
	eventMgr.receiveEvent(NULLCP, NULLCP, NULLCP);	// listen for all events

    // start listening...
    coordinator.listen();
}
