//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// MAction.cc - enables to perform an action on managed objects in a remote MIB

// By George Pavlou, June 1993


#include <stdlib.h>
#include <string.h>
#include "msap.h"
#include "Error.h"
#include "RMIBAgent.h"
#include "Coordinator.h"
#include "Syntaxes.h"
#include "AnyType.h"

int   parseAvaArg (int, char**, char**, char**, int*);
int  parseFilterArg (int, char**, char**, int*);

Bool timing = False;

// the remote MIB agent

RMIBAgent	agent;

// Note: the coordinator is not needed as there are no asynchronous
//       operations for the RMIBAgent within this program


main (int argc, char** argv)
{
    char* application = NULLCP;
    char* host = NULLCP;
    char* strObjClass = NULLCP;
    char* strObjInst = NULLCP;
    char* strScope = NULLCP;
    char* strFilter = NULLCP;
    char* strSync = NULLCP;
    char* strActionType = NULLCP;
    char* strActionValue = NULLCP;
    int   i, j, state = 0;
    CMISScope* scope = NULLMSCOPE;	// default
    CMISFilter* filter = NULLMFILTER;	//  ..
    CMISSync sync = NULLMSYNC;		//  ..
    AnyType* actionInfo = NULL;
    AVA* actionArg = NULLAVA;
    CMISObjectList* objectList;


    // parse command line arguments

    if (argc < 5)
	goto usage;

    if (strcmp(argv[1], "-S") == 0)
        application = argv[2];

    for (i = 3; i < argc; i++)
	switch (state) {
	case 0:
	    if (*argv[i] == '-' && strlen(argv[i]) == 2)
		switch (*(argv[i]+1)) {
		case 'c':
		    state = 1;
		    break;
		case 'i':
		    state = 2;
		    break;
		case 's':
		    state = 3;
		    break;
		case 'f':
		    state = 4;
		    break;
		case 'a':
		    state = 5;
		    break;
		case 't':
		    timing = True;
		    break;
		default:
		    goto usage;
		    }
	    else
		goto usage;
	    break;

	case 1:
	    strObjClass = argv[i];
	    state = 0;
	    break;

	case 2:
	    strObjInst = argv[i];
	    state = 0;
	    break;

	case 3:
	    strScope = argv[i];
	    if (i < argc-1 && *argv[i+1] != '-')
		strSync = argv[++i];
	    state = 0;
	    break;

	case 4:
	    if (parseFilterArg(argc-i, &argv[i], &strFilter, &j) == NOTOK)
		goto usage;
	    i += j-1;
	    state = 0;
	    break;

	case 5:
	{
	    Bool more = False;
	    for (j = i; j < argc; j++)
	        if (*argv[j] == '-' && strlen(argv[j]) == 2) {
		    more = True;
		    break;
		}

	    if (parseAvaArg(argc-i, &argv[i],
			&strActionType, &strActionValue, &j) == NOTOK)
		goto usage;
	    i = more ? i+j-1 : i+j;
	}
	    state = 0;
	    break;
	}

    if (!strActionType)
	goto usage;

    // load syntaxes

    if (initialiseSyntaxes(NULLCP, actionSyntaxes) == NOTOK)
	error(NULLCP, "maction: cannot initialise syntaxes");


    // convert the string arguments to CMIS parameters where necessary

    if (strScope && (scope = str2scope(strScope)) == NULLMSCOPE)
	error(NULLCP, "invalid scope %s", strScope);

    if (strSync && (sync = str2sync(strSync)) == s_invalid)
        error(NULLCP, "invalid sync %s", strSync);


    if (strActionValue) {
	char strActionTypeSuffix[128];
	sprintf(strActionTypeSuffix, "%sInfo", strActionType);
	actionInfo = new AnyType(strActionTypeSuffix, strActionValue);
    }
    actionArg = new AVA(strActionType, actionInfo);


    // initialise the remote MIB agent

    if (timing)
	do_timing(NULLCP);

    host = application ? NULLCP : argv[2];
    application = application ? application : argv[1];

    if (agent.connect(application, host) != OK)
	error(NULLCP, "cannot establish association to %s at %s",
			argv[1], argv[2]);

    if (timing)
	do_timing("established association");


    // perform the action request
    agent.Action(strObjClass, strObjInst,
		scope ? scope->sc_type : Sc_BaseObject,
		scope ? scope->sc_level : 0, strFilter,
		NULLMACCESS, sync, actionArg, objectList);

    if (timing)
        do_timing("performed action");

     if (!timing)
	objectList->print();
    delete actionArg;
    delete objectList;

    agent.disconnect();

    if (timing)
        do_timing("released association");

    exit(OK);

usage:
    error(NULLCP, "usage: maction { <agent> <host> | -S <agent> }\n\
\t[-c <class>] [-i <instance>]\n\
\t[-s <scope> [<sync>]] [-f filter] -a <actionType>[=<actionValue]\n");
}
