//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// MibDump.cc - enables to read managed objects/attributes from a remote MIB

// By George Pavlou / Thurain Tin, June 1993


#include <stdlib.h>
#include <string.h>
#ifdef	__GNUG__
extern "C" {
    int strcasecmp (const char*, const char*);
    int strncasecmp (const char*, const char*, size_t);
}
#endif
#include "msap.h"
#include "Error.h"
#include "RMIBAgent.h"
#include "Coordinator.h"
#include "Syntaxes.h"

int  parseFilterArg (int, char**, char**, int*);

// the remote MIB agent

RMIBAgent	agent;

// Note: the coordinator is not needed as there are no asynchronous
//       operations for the RMIBAgent within this program


main (int argc, char** argv)
{
    char* application = NULLCP;
    char* host = NULLCP;
    char* strObjClass = NULLCP;
    char* strObjInst = NULLCP;
    char* strScope = NULLCP;
    char* strSync = NULLCP;
    char* strFilter = NULLCP;		// default
    char* strAttrs [MAXATTRS+1];
    int nattrs = GET_ALLATTRS;		//  ..
    int   i, j, k, l, state = 0;
    CMISScope* scope = NULLMSCOPE;	//  ..
    CMISSync sync = NULLMSYNC;		//  ..
    CMISObjectList* objectList;


    // parse command line arguments

    if (argc < 3)
	goto usage;

#ifdef USE_X500
    if (strcmp(argv[1], "-S") == 0)
        application = argv[2];
#endif

    if (argc > 3)
	for (i = 3; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'c':
			state = 1;
			break;
		    case 'i':
			state = 2;
			break;
		    case 's':
			state = 3;
			break;
		    case 'f':
			state = 4;
			break;
		    case 'a':
			state = 5;
			break;
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		strObjClass = argv[i];
		state = 0;
		break;

	    case 2:
		strObjInst = argv[i];
		state = 0;
		break;

	    case 3:
		strScope = argv[i];
		if (i < argc-1 && *argv[i+1] != '-')
		    strSync = argv[++i];
		state = 0;
		break;

	    case 4:
		if (parseFilterArg(argc-i, &argv[i], &strFilter, &j) == NOTOK)
		    goto usage;
		i += j-1;
		state = 0;
		break;

	    case 5:
	    {
		Bool more = False;
		for (j = i; ; j++)
		    if (*argv[j] == '-' && strlen(argv[j]) == 2) {
			more = True;
			break;
		    }
		    else
		    if (j == argc-1)
			break;

		nattrs = more ? j-i : j-i+1;
		if (nattrs > MAXATTRS)
		    error(NULLCP, "too many attrs - max number %s", MAXATTRS);

		for (k = i, l = 0; k < j+1; k++, l++)
		    strAttrs[l] = argv[k];
		state = 0;
		i = more ? j-1 : j;
	    }
		break;
	    }


    // load syntaxes

    if (initialiseSyntaxes(NULLCP, attrSyntaxes) == NOTOK)
	error(NULLCP, "mibdump: cannot initialise syntaxes");


    // convert the string arguments to CMIS parameters where necessary

    if (strScope && (scope = str2scope(strScope)) == NULLMSCOPE)
	error(NULLCP, "invalid scope %s", strScope);

    if (strSync && (sync = str2sync(strSync)) == s_invalid)
        error(NULLCP, "invalid sync %s", strSync);

    if (nattrs > 0) {
	if (strcasecmp(strAttrs[0], "none") == 0) {
	    strAttrs[0] = "none";
	    strAttrs[1] = NULLCP;
	}
	else
	    strAttrs[nattrs] = NULLCP;
    }
    else {
	strAttrs[0] = "all";
	strAttrs[1] = NULLCP;
    }


    // initialise the remote MIB agent

    host = application ? NULLCP : argv[2];
    application = application ? application : argv[1];

    if (agent.connect(application, host) != OK)
	error(NULLCP, "cannot establish association to %s at %s",
			argv[1], argv[2]);


    // perform the Get request
    if (agent.Get(strObjClass, strObjInst,
		scope ? scope->sc_type : Sc_BaseObject,
		scope ? scope->sc_level : 0,
		strFilter, NULLMACCESS, sync, strAttrs, objectList)
	    == OK) {
	objectList->print();
	delete objectList;
    }
    else
	advise(NULLCP, "Get error:  %s", objectList->getErrorStr());

    agent.disconnect();
    exit(OK);

usage:
    error(NULLCP, "usage: mibdump { <agent> <host> | -S <agent> }\n\
\t[-c <class>] [-i <instance>]\n\
\t[-s <scope> [<sync>]] [-f filter] [-a attr ...]");
}

