// RMIBAgent.cc - the Remote MIB Agent realisation

// By George Pavlou, December 1991 / Thurain Tin, April 1993


#ifdef __GNUG__
#include <std.h>
extern "C" { int  strncasecmp (const char*, const char*, size_t); }
#elif  __cplusplus
#include <stdlib.h>
#endif

#include "msap.h"
#include "Error.h"
#include "SmiSntx.h"
#include "GenericAttr.h"
#include "ListIterator.h"
#include "AnyType.h"
#include "Time.h"
#include "RMIBresult.h"
#include "RMIBManager.h"
#include "RMIBAgent.h"


static char* strTime;
int RMIBAgent::_invoke;


static char* now ()
{
    static UTCtime utct;
    long sect = time((long*) 0);
    tm2ut(localtime(&sect), &utct);
    if (strTime) free(strTime);
    return strTime = strdup(utct2str(&utct));
}


int RMIBAgent::setTitle (char* title)
{
    if (!getisobjectbyname(title)) {
	advise(NULLCP, "\nRMIBAgent::setTitle: '%s' not registerd in ETC/isobjects!",
			title);
	advise(NULLCP, "event reporting will use the 'thisAssociation' \
convention at present");
	advise(NULLCP, "but you should register the missing title to ETC/isobjects\n");
	return NOTOK;
    }
    _title = strdup(title);
    return OK;
}


void RMIBAgent::clear (Bool del)
{
    _msd = UNCONNECTED;
    _interval = DFLT_TIMEOUT;

    if (del) {
	if (_agent) {
	    free(_agent);
	    _agent = NULLCP;
	}
	if (_host) {
	    free(_host);
	    _host = NULLCP;
	}
    }

    if (_inst_mn) {
	mn_free(_inst_mn);
	delete _inst_mn;
	_inst_mn = NULLMN;
    }

    _listening = _blocking = False;
    _hasFilter = False;

    if (_nFlt || _nAll) {
	_evList.clear();
	_evIterator.setList(&_evList);
	_nFlt = _nAll = 0;
    }

    if (_efd_mid) {
	mid_free(_efd_mid);
	delete _efd_mid;
	_efd_mid = NULLMID;
    }

    if (_efd_mn) {
	mn_free(_efd_mn);
	delete _efd_mn;
	_efd_mn = NULLMN;
    }

    if (_efd_name) {
	free(_efd_name);
	_efd_name = NULLCP;
    }

    if (_title) {
	free(_title);
	_title = NULLCP;
    }

    _callbackList.clear();
    _callbackIterator.setList(&_callbackList);
    _obj = NULLCMISOBJECT;
    _objList = NULLCMISOBJECTLIST;
    _report = (CMISEventReportArg*) 0;
}


int RMIBAgent::startListening ()
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not yet in association with agent!");
	return NOTOK;
    }

    if (startListen(_msd) == NOTOK) {
	advise(NULLCP, "RMIBAgent: comm endpoint registration failed!");
	return NOTOK;
    }
    _listening = True;
    return OK;
}


inline int RMIBAgent::checkScope (int type, int level, CMISScope* scope)
{
    scope->sc_type  = type;
    scope->sc_level = type > Sc_WholeSubtree ? level : 0;
    return (type < Sc_BaseObject || type > Sc_BaseToNthLevel || level < 0)
	? NOTOK : OK;
} 


NotifyEvent* RMIBAgent::getEvent (char* objClass, char* objInst,
	char* eventType, char* eventTime, RMIBManager* mgr)
{
    NotifyEvent* e;
    _evIterator.reset();
    while (e = (NotifyEvent*) _evIterator.getNext()) {
	if (e->mgr != mgr || e->sfilter || !e->mfilter)
	    continue;
	if (strcmp1(e->objClass, objClass) == OK &&
		strcmp1(e->objInst, objInst) == OK &&
		strcmp1(e->eventType, eventType) == OK &&
		strcmp1(e->eventTime, eventTime) == OK)
	    return e;
    }
    return NULLNOTIFYEVENT;
}


NotifyEvent* RMIBAgent::getEvent (char* flt, RMIBManager* mgr)
{
    NotifyEvent* e;
    _evIterator.reset();
    while (e = (NotifyEvent*) _evIterator.getNext()) {
	if (e->mgr != mgr || !e->sfilter || !e->mfilter ||
		e->objClass || e->objInst || e->eventType || e->eventTime)
	    continue;
	if (strcmp1(e->sfilter, flt) == OK) return e;
    }
    return NULLNOTIFYEVENT;
}


NotifyEvent* RMIBAgent::getEvent (CMISFilter* mf, RMIBManager* mgr)
{
    NotifyEvent* e;
    _evIterator.reset();
    while (e = (NotifyEvent*) _evIterator.getNext()) {
	if (e->mgr != mgr || e->sfilter || !e->mfilter ||
		e->objClass || e->objInst || e->eventType || e->eventTime)
	    continue;
	if (!strcmp(sprintmfilter(e->mfilter), sprintmfilter(mf))) return e;
    }
    return NULLNOTIFYEVENT;
}


inline Bool RMIBAgent::checkAllEvent (RMIBManager* mgr)
{
    if (!_nAll || _hasFilter) return False;

    NotifyEvent* e;
    _evIterator.reset();
    while (e = (NotifyEvent*) _evIterator.getNext()) {
	if (e->mgr == mgr && e->isAllEvent() == True) {
	    advise(NULLCP, "RMIBAgent: already registered to receive all!");
	    return True;
	}
    }
    return False;
}


int RMIBAgent::receiveAllEvent (RMIBManager* mgr)
{
    if (_hasFilter) {
	advise(NULLCP, "RMIBAgent: filter expression assertion exist,");
	advise(NULLCP, "unable to register to receive all event reports!");
	return NOTOK;
    }

    NotifyEvent* e;
    _evIterator.reset();
    while (e = (NotifyEvent*) _evIterator.getNext()) {
	if (e->mgr == mgr && removeEvent(e) == OK) {
	    --_nFlt;
	    _evIterator.reset();
	}
    }

    e = new NotifyEvent(mgr);
    _evList.append(e);
    ++_nAll;
    advise(NULLCP, "RMIBAgent: manager registered to receive all reports");
    return OK;
}


int RMIBAgent::stopReceiveAllEvent (RMIBManager* mgr)
{
    if (!_nAll || _hasFilter)
	return NOTOK;

    NotifyEvent* e;
    _evIterator.reset();
    while (e = (NotifyEvent*) _evIterator.getNext()) {
	if (e->mgr == mgr && e->isAllEvent() == True &&
		removeEvent(e) == OK) {
	    advise(NULLCP, "RMIBAgent: manager deregistered from receiving all reports");
	    --_nAll;
	    return OK;
	}
    }
    return NOTOK;
}


int RMIBAgent::removeEvent (NotifyEvent* ev)
{
    NotifyEvent* e;
    if (e = (NotifyEvent*) _evList.remove(ev)) {
	delete e;
	return OK;
    }
    advise(NULLCP, "RMIBAgent: unable to remove event entry from list!");
    return NOTOK;
}


int RMIBAgent::createDiscrObj (PE pe, MID& mid, MN& mn)
{
    CMISParam attrs[2];
    int nattrs = 0;
    char buf[256], *myhost;

    mid = new MIDentifier;
    mid->mid_type = MID_GLOBAL;
    mid->mid_global = name2oid("eventForwardingDiscriminator");
/*
    if (_title) {
        Ava* ava;
#ifndef LOCALHOST
        myhost = getlocalhost();
#else
        myhost = "localhost";
#endif
        sprintf(buf, "%s-%s", _title, myhost);
        if ((ava = str2ava("destination", buf)) == NULL)
            error(NULLCP, "invalid destination: %s", buf);
        attrs[nattrs].mp_id.mid_type = MID_GLOBAL;
        attrs[nattrs].mp_id.mid_global = ava->ava_oid;
        attrs[nattrs].mp_val = ava->ava_value;
        free ((char*) ava);
	++nattrs;
    }
*/
    if (pe) {
	attrs[nattrs].mp_id.mid_type = MID_GLOBAL;
	attrs[nattrs].mp_id.mid_global = name2oid("discriminatorConstruct");
	attrs[nattrs].mp_val = pe;
	++nattrs;
    }

    MSAPIndication mi;
    register int i;
    MN efd_mn = _efd_name ? ObjectInstance_parse(_efd_name) : NULLMN;

    if (M_Create(_msd, get_invoke(), mid, efd_mn, 0, NULLMN,
	    		NULLMACCESS, nattrs, attrs, &mi) == NOTOK) {
	advise(NULLCP, "RMIBAgent: EFD creation failed, provider rejected %s",
		mi.mi_preject.mpr_data);
	for (i = 0; i < nattrs; i++)
	    mparm_free(&attrs[i]);
	if (efd_mn) {
	    mn_free(efd_mn);
	    free((char*) efd_mn);
	}
	mi_free(&mi);
	return NOTOK;
    }

    if (efd_mn) {
	mn_free(efd_mn);
	free((char*) efd_mn);
    }

    for (i = 0; i < nattrs; i++)
	mparm_free(&attrs[i]);

    while ((i = M_WaitReq(_msd, _interval, &mi)) == OK) {
	switch (mi.mi_type) {
	case MI_CREATE_RES:
	    break;

	case MI_EVENT_REP:
	    //eventNotification(&mi.mi_eventrepinv.args);
	    mi_free(&mi);
	    continue;

	default:
	    advise(NULLCP, "RMIBAgent: unexpected indication %d", mi.mi_type);
            mi_free(&mi);
	    return NOTOK;
	}

	if (mi.mi_type == MI_CREATE_RES)
	    break;
    }

    if (i != OK)
	return NOTOK;

    mn = new MName;
    *mn = mi.mi_createres.result.cr_inst;
    mi.mi_createres.result.cr_inst.mn_dn = NULLDN;

    char* cp = mn->mn_dn ? dn2str(mn->mn_dn) : NULLCP;
    advise(NULLCP, "RMIBAgent: EFD object created with name {%s}", cp);

    if (cp) free(cp);
    mi_free(&mi);

    return mn->mn_dn ? OK : NOTOK;
}


int RMIBAgent::deleteDiscrObj (MN mn)
{
    if (mn) {
	MSAPIndication mi;
	M_Delete(_msd, get_invoke(), NULLMID, mn, NULLMSCOPE,
		NULLMFILTER, NULLMACCESS, NULLMSYNC, &mi);

	while (M_WaitReq(_msd, _interval, &mi) == OK) {
	    char* cp;

	    switch (mi.mi_type) {
	    case MI_DELETE_RES:
		advise(NULLCP, "RMIBAgent: EFD object deleted, name {%s}",
                    cp = dn2str(mi.mi_deleteres.result.dr_inst.mn_dn));
                if (cp) free(cp);
		mi_free(&mi);
	        return OK;

	    case MI_EVENT_REP:
		//eventNotification(&mi.mi_eventrepinv.args);
		mi_free(&mi);
		continue;

	    default:
		// this could be the objectDeletion report for our dicriminator
		// or something else, but at this moment we don't care
		mi_free(&mi);
		continue;
	    }
	}
    }
    return NOTOK;
}


int RMIBAgent::adjustReporting ()
{
    if (_evList.getCount() == 0 && !_nAll && !_nFlt) {
	if (_efd_mn) {
	    deleteDiscrObj(_efd_mn);
	    mn_free(_efd_mn);
	    delete _efd_mn;
	    _efd_mn = NULLMN;
	}

	if (_efd_mid) {
	    mid_free(_efd_mid);
	    delete _efd_mid;
	    _efd_mid = NULLMID;
	}

	if (_efd_name) {
	    free(_efd_name);
	    _efd_name = NULLCP;
	}
	return OK;
    }

    MSAPIndication mi;
    PE filter_pe = NULLPE;

    if (!_nAll && _nFlt) {
	CMISFilter *filter, *prevFilter = NULLMFILTER;
	NotifyEvent* e;
	_evIterator.reset();
	while (e = (NotifyEvent*) _evIterator.getNext()) {
	    if (!e->mfilter)
		continue;
	    e->mfilter->flt_next = prevFilter;
	    prevFilter = e->mfilter;
	}

	filter = new CMISFilter;
	filter->flt_type = Filter_Or;
	filter->flt_filter = prevFilter;
	filter_pe = CMISFilter_encode(filter);
	delete filter;
    }

    if (!_efd_mid) {
	if (createDiscrObj(filter_pe, _efd_mid, _efd_mn) != OK) {
	    mid_free(_efd_mid);
	    delete _efd_mid;
	    _efd_mid = NULLMID;
	    return NOTOK;
	}
    }
    else {
	CMISSetAttr setattr;
	setattr.sa_id.mid_type = MID_GLOBAL;
	setattr.sa_id.mid_global = name2oid("discriminatorConstruct");
	if (filter_pe) {
	    setattr.sa_modify = m_replace;
	    setattr.sa_val = filter_pe;
	}
	else {
	    setattr.sa_modify = m_setToDefault;
	    setattr.sa_val = NULLPE;
	}

	M_Set(_msd, get_invoke(), _efd_mid, _efd_mn, NULLMSCOPE,
		NULLMFILTER, NULLMACCESS, NULLMSYNC, 1, &setattr, &mi);
	mparm_free(&setattr.sa_ava);
    }
    return OK;
}


int RMIBAgent::eventNotification (CMISEventReportArg* arg)
{
    _report = arg;
    char* objClass  = oid2name(arg->ea_class.mid_global, OIDPART);
    char* objInst   = dn2str(arg->ea_inst.mn_dn);
    char* eventType = oid2name(arg->ea_type.mid_global, OIDPART);

    Attr* attr;
    register int nmgr = 0;
    RMIBManager* mgr[MAXMGRS];
    NotifyEvent* e;
    _evIterator.reset();

    while (e = (NotifyEvent*) _evIterator.getNext()) {
	if (!e->mgr)
	    continue;

	if (e->isAllEvent() == True) {
	    *(mgr+nmgr++) = e->mgr;
	    continue;
	}

	if ((e->objClass && strcmp(e->objClass, objClass) != 0) ||
		(e->objInst && strcmp(e->objInst, objInst) != 0) ||
		(e->eventType && strcmp(e->eventType, eventType) != 0))
	    continue;

	if (e->eventTime) {
	    if (!arg->ea_time) {
		e->mgr->noTimeAssertion(this);
		continue;
	    }

	    attr = new Time;
	    attr->set((void*) arg->ea_time);
	    CMISFilter* f = str2mfilter(e->eventTime);
	    Bool b = attr->filter(f);
	    mfilter_free(f);
	    attr->clear();
	    delete attr;
	    if (b != True) continue;
	}
	*(mgr+nmgr++) = e->mgr;
    }

    attr = new AnyType(arg->ea_type.mid_global, arg->ea_info);

    if (attr && attr->get() == NULLVD &&
	    arg->ea_info->pe_errno != PE_ERR_NONE) {
	attr->clear();
	delete attr;
	attr = NULLATTR;
    }

    if (attr) {
        for (register int i = 0; i < nmgr; i++)
            (*(mgr+i))->eventNotification(objClass, objInst,
                                eventType, arg->ea_time, attr, this);
        attr->clear();
        delete attr;
    }
 
    if (objInst)
	free(objInst);

    _report = (CMISEventReportArg*) 0;
    arg->ea_info = NULLPE;
    return OK;
}


inline RMIBCallback* RMIBAgent::getCallback (int id)
{
    RMIBCallback* c;
    _callbackIterator.reset();
    while (c = (RMIBCallback*) _callbackIterator.getNext()) {
	if (c->id == id) return c;
    }
    return NULLRMIBCALLBACK;
}


int RMIBAgent::removeCallback (RMIBCallback* c)
{
    RMIBCallback* rc;
    if (rc = (RMIBCallback*) _callbackList.remove(c)) {
	delete rc;
	return OK;
    }
    advise(NULLCP, "RMIBAgent: unable to remove callback entry from list!");
    return NOTOK;
}


void RMIBAgent::brokenAssociation ()
{
    RMIBCallback* c;
    _callbackIterator.reset();
    while (c = (RMIBCallback*) _callbackIterator.getNext()) {
	if (c->id == -1 && c->mgr)
	    c->mgr->brokenAssociation(this);
    }
}


int RMIBAgent::readMsapIndication ()
{
    int id = 0, ind = OK;
    CMISObject* obj = NULLCMISOBJECT;
    MSAPIndication mi;
    M_WaitReq(_msd, _interval, &mi);

    switch (mi.mi_type) {
    case MI_GET_RES:
	obj = process_getres(id = mi.mi_getres.id,
			&mi.mi_getres.result);
	adjustResult(id, obj);
	reportResult(id);
	break;

    case MI_SET_RES:
	obj = process_setres(id = mi.mi_setres.id,
			&mi.mi_setres.result);
	adjustResult(id, obj);
	reportResult(id);
	break;

    case MI_ACTION_RES:
	obj = process_actionres(id = mi.mi_actionres.id,
			&mi.mi_actionres.result);
	adjustResult(id, obj);
	reportResult(id);
	break;

    case MI_CREATE_RES:
	obj = process_createres(id = mi.mi_createres.id,
			&mi.mi_createres.result);
	reportResult(id, obj);
	break;

    case MI_DELETE_RES:
	obj = process_deleteres(id = mi.mi_deleteres.id,
			&mi.mi_deleteres.result);
	adjustResult(id, obj);
	reportResult(id);
	break;

    case MI_ERROR:
	switch (mi.mi_errtype) {
	case m_getListError:
	    obj = process_getres(id = mi.mi_getres.id,
			&mi.mi_getres.result, m_getListError);
	    adjustResult(id, obj);
	    reportResult(id);
	    break;

	case m_setListError:
	    obj = process_setres(id = mi.mi_setres.id,
			&mi.mi_setres.result, m_setListError);
	    adjustResult(id, obj);
	    reportResult(id);
	    break;

	default:
	    obj = process_errinfo(id = mi.mi_error.id,
			&mi.mi_error.info, mi.mi_errtype);
	    adjustResult(id, obj);
	    reportResult(id, obj);
	    ind = NOTOK;
	    break;
	}
	break;

    case MI_LINKED_REPLY:
	id = mi.mi_linkedreplyinv.args.lr_linkid;
	switch (mi.mi_linkedreplyinv.args.lr_type) {

	case LR_GET_RES:
	case LR_GET_LISTERR:
	    obj = process_getres(id,
			&mi.mi_linkedreplyinv.args.lr_getres, mi.mi_errtype);
	    break;

	case LR_SET_RES:
	case LR_SET_LISTERR:
	    obj = process_setres(id,
			&mi.mi_linkedreplyinv.args.lr_setres, mi.mi_errtype);
	    break;

	case LR_ACTION_RES:
	    obj = process_actionres(id,
			&mi.mi_linkedreplyinv.args.lr_actionres);
	    break;

	case LR_DELETE_RES:
	    obj = process_deleteres(id,
			&mi.mi_linkedreplyinv.args.lr_deleteres);
	    break;

	case LR_PROC_FAILURE:
	    obj = process_procfailure(id,
			&mi.mi_linkedreplyinv.args.lr_procfailure);
	    break;

	case LR_ACTION_ERR:
	    obj = process_actionerr(id,
			&mi.mi_linkedreplyinv.args.lr_actionerr);
	    break;

	case LR_DELETE_ERR:
	    obj = process_deleteerr(id,
			&mi.mi_linkedreplyinv.args.lr_deleteerr);
	    break;
	}
	adjustResult(id, obj);
	break;

    case MI_EMPTY_RES:
	reportResult(mi.mi_emptyres.id);
	break;

    case MI_EVENT_REP:
	eventNotification(&mi.mi_eventrepinv.args);
	break;

    case MI_TERM:
	advise(NULLCP, "RMIBAgent: %s-%s terminated association, reason: %d!!!",
		_agent, _host, mi.mi_term.mt_reason);
	if (_listening && stopListen(_msd) == NOTOK)
	    advise(NULLCP, "RMIBAgent: stopListen failed, msd %d", _msd);
	brokenAssociation();
	clear(True);
	ind = NOTOK;
	break;

    case MI_PREJECT:	
    case MI_ABORT:
	advise(NULLCP, "RMIBAgent: %s from %s-%s received!!!",
	    mi.mi_type == MI_PREJECT ? "provider reject" : "association abort",
	    _agent, _host);
	advise(NULLCP, "data: %s", mi.mi_type == MI_PREJECT
	    ? mi.mi_preject.mpr_data : mi.mi_abort.ma_data);
	if (_listening && stopListen(_msd) == NOTOK)
	    advise(NULLCP, "RMIBAgent: stopListen failed, msd %d", _msd);
	brokenAssociation();
	clear();
	ind = NOTOK;
	break;

    case MI_UREJECT:
	advise(NULLCP, "RMIBAgent: user reject from %s-%s received!!!",
	    _agent, _host);
	advise(NULLCP, "id: %d, reason: %d, data: %s",
	    id = mi.mi_ureject.mur_id, mi.mi_ureject.mur_reason,
	    mi.mi_ureject.mur_data);
	if (_listening && stopListen(_msd) == NOTOK)
	    advise(NULLCP, "RMIBAgent: stopListen failed, msd %d", _msd);
	brokenAssociation();
	clear();
	ind = NOTOK;
	break;

    default:
	advise(NULLCP, "RMIBAgent: unknown msap indication!!");
	ind = NOTOK;
	break;
    }
    mi_free(&mi);
    return ind;
}


int RMIBAgent::adjustResult (int id, CMISObject* obj)
{
    if (!obj)
	return NOTOK;

    if (_blocking && get_curId() == id) {
	if (_objList) {
	    if (!obj->getOperation())
		obj->setOperation(_objList->getOperation());
	    _objList->append(obj);
	    if (obj->getError() && obj->getError() != m_getListError &&
			obj->getError() != m_setListError)
		_objList->setErrObj(obj);
	}
	else {
	    if (!obj->getOperation())
		obj->setOperation(M_CREATE);
	}
	return OK;
    }

    RMIBCallback* c = getCallback(id);
    if (c) {
	if (!obj->getOperation())
	    obj->setOperation(c->operation);
	if (c->objList) {
	    c->objList->append(obj);
	    if (obj->getError() && obj->getError() != m_getListError &&
			obj->getError() != m_setListError)
		c->objList->setErrObj(obj);
	    return OK;
	}
	if (c->operation == M_GET && c->mgr)
	    c->mgr->singleGetResult(obj, this);
	else
	if (c->operation != M_CREATE)
	    delete obj;
	return OK;
    }

    delete obj;
    return NOTOK;
}


int RMIBAgent::reportResult (int id, CMISObject* obj)
{
    if (_blocking && get_curId() == id) {
	if (obj)
	    _obj = obj;
	_blocking = False;
	return OK;
    }

    RMIBCallback* c = getCallback(id);
    RMIBManager* mgr = c ? c->mgr : NULLRMIBMANAGER;

    if (mgr) {
	if (c->objList)
	    mgr->generalResult(c->objList, this);
	if (obj && c->operation == M_CREATE)
	    mgr->createResult(obj, this);
    }

    if (c)
	removeCallback(c);
    return OK;
}


int RMIBAgent::connect (char* agent, char* host)
{
    if (_msd != UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: association exists to %s-%s!!",
		_agent, _host);
	return NOTOK;
    }

    if (host && *host != '\0' && !_host)
	_host = strdup(host);

    if (agent && *agent != '\0' && !_agent)
	_agent = strdup(agent);

    if (!_agent
#ifndef USE_X500				/* if no LT hooks */
		|| !_host
#endif
	) {
	advise(NULLCP, "RMIBAgent: invalid agent %s or host %s parameter",
		_agent, _host);
	clear(True);
	return NOTOK;
    }

    MInitOption* miop[2];
    MInitOption mio_aet;
    char buf[256], *myhost;
    miop[0] = NULL;

/*
 * the following code that passes the calling AET ACSE parameter
 * has been removed as it affects interoperability with other platforms

    if (_title) {
        miop[0] = &mio_aet;
        miop[1] = NULL;
        mio_aet.optionId = MIO_CallingAET;
        mio_aet.optionInfo = buf;
#ifndef LOCALHOST
        myhost = getlocalhost();
#else
        myhost = "localhost";
#endif
        sprintf(buf, "%s-%s", _title, myhost);
    }
 */

    if ((_msd = m_initialise(_agent, _host, _title ? miop : NULL)) == NOTOK) {
	advise(NULLCP, "RMIBAgent: association failed to %s-%s",
		_agent, _host);
	clear();
	return NOTOK;
    }

    advise(NULLCP, "RMIBAgent: association established to %s-%s with msd %d",
		_agent, _host, _msd);
    return OK;
}


int RMIBAgent::disconnect ()
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in connection!!");
	return NOTOK;
    }

    if (_efd_mn)
	deleteDiscrObj(_efd_mn);

    if (_listening && stopListen(_msd) == NOTOK)
	advise(NULLCP, "RMIBAgent: stopListen failed, msd %d", _msd);

    if (m_terminate(_msd) == NOTOK) {
	clear(True);
	return NOTOK;
    }
    advise(NULLCP, "\nRMIBAgent: connection released from %s-%s",
		_agent, _host);
    clear(True);
    return OK;
}


int RMIBAgent::reset (char* agent, char* host)
{
    if (_msd == UNCONNECTED) {
	clear(True);
	_host = host ? strdup(host) : NULLCP;
	_agent = agent ? strdup(agent) : NULLCP;
	return OK;
    }
    return NOTOK;
}


int RMIBAgent::notifyBrokenAssociation (RMIBManager* mgr)
{
    if (!mgr)
	return NOTOK;

    RMIBCallback* c = new RMIBCallback(-1, mgr, NULLCMISOBJECTLIST, -1);
    _callbackList.append(c);
    advise(NULLCP, "RMIBAgent: manager registered for broken assoc notification");
    return OK;
}


int RMIBAgent::receiveEvent (char* objClass, char* objInst,
	char* eventType, char* eventTime, RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in connection!!!");
	return NOTOK;
    }

    if (!mgr || (!_listening && startListening() == NOTOK) ||
	    checkAllEvent(mgr) == True)
	return NOTOK;

    if (!objClass && !objInst && !eventType && !eventTime) {
	if (receiveAllEvent(mgr) == NOTOK)
	    return NOTOK;
    }
    else {
	if (getEvent(objClass, objInst, eventType,
		    	eventTime, mgr) != NULLNOTIFYEVENT)
	    return NOTOK;

	CMISFilter*
	mf = buildFilter(objClass, objInst, eventType, eventTime);

	if (!mf) {
	    advise(NULLCP, "RMIBAgent: invalid filter produced for class \
%s, inst %s, type %s, time %s",
		    objClass, objInst, eventType, eventTime);
	    return NOTOK;
	}

	NotifyEvent* e = new NotifyEvent(objClass, objInst,
				eventType, eventTime, NULLCP, mf, mgr);
	_evList.append(e);
	++_nFlt;

        advise(NULLCP, "RMIBAgent: manager registered to receive events \
with class %s, inst %s, type %s, time %s",
                objClass, objInst, eventType, eventTime);
#ifdef DEBUG
        advise(NULLCP, "filter construct: %s", sprintmfilter(e->mfilter));
#endif
    }
    return adjustReporting();
}


int RMIBAgent::stopReceiveEvent (char* objClass, char* objInst,
	char* eventType, char* eventTime, RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (!mgr || !_listening ||
	    (_hasFilter && !objClass && !objInst && !eventType && !eventTime))
	return NOTOK;

    if (!objClass && !objInst && !eventType && !eventTime) {
	if (stopReceiveAllEvent(mgr) == NOTOK)
	    return NOTOK;
    }
    else {
	NotifyEvent* e;
	if (e = getEvent(objClass, objInst, eventType, eventTime, mgr)) {
#ifdef DEBUG
            advise(NULLCP, "filter construct: %s", sprintmfilter(e->mfilter));
#endif
	    if (removeEvent(e) == NOTOK)
		return NOTOK;

            advise(NULLCP, "RMIBAgent: manager de-registered from receiving \
events with class %s, inst %s, type %s, time %s",
                    objClass, objInst, eventType, eventTime);
	    --_nFlt;
	    _evIterator.reset();
	}
	else {
	    return NOTOK;
	}
    }
    return adjustReporting();
}


int RMIBAgent::receiveEvent (char* flt, RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (!mgr || _hasFilter || (_nAll && flt) ||
	    (!_listening && startListening() == NOTOK) ||
	    checkAllEvent(mgr) == True)
	return NOTOK;

    if (!flt) {
	if (receiveAllEvent(mgr) == NOTOK)
	    return NOTOK;
    }
    else {
	if (getEvent(flt, mgr) != NULLNOTIFYEVENT)
	    return NOTOK;

	CMISFilter* mf = str2mfilter(flt);
	if (!mf) {
	    advise(NULLCP, "RMIBAgent: invalid filter produced for %s", flt);
	    return NOTOK;
	}

	NotifyEvent* e = new NotifyEvent(NULLCP, NULLCP,
					NULLCP, NULLCP, flt, mf, mgr);
	_evList.append(e);
	++_nFlt;
	_hasFilter = True;

        advise(NULLCP, "RMIBAgent: manager registered to receive events \
using filter, %s",
		flt);
    }
    return adjustReporting();
}


int RMIBAgent::stopReceiveEvent (char* flt, RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (!mgr || !_listening || (_hasFilter && !flt) ||
	    (!_hasFilter && flt) || (_nAll && flt))
	return NOTOK;

    if (!flt) {
	if (stopReceiveAllEvent(mgr) == NOTOK)
	    return NOTOK;
    }
    else {
	NotifyEvent* e;
	if ((e = getEvent(flt, mgr)) && removeEvent(e) == OK) {
            advise(NULLCP, "RMIBAgent: manager de-registered from receiving \
events using filter, %s",
			flt);
	    --_nFlt;
	    _hasFilter = False;
	}
	else {
	    return NOTOK;
	}
    }
    return adjustReporting();
}


int RMIBAgent::receiveEvent (CMISFilter* mflt, RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (!mgr || _hasFilter || (_nAll && mflt) ||
	    (!_listening && startListening() == NOTOK) ||
	    checkAllEvent(mgr) == True)
	return NOTOK;

    if (!mflt) {
	if (receiveAllEvent(mgr) == NOTOK)
	    return NOTOK;
    }
    else {
	if (getEvent(mflt, mgr) != NULLNOTIFYEVENT)
	    return NOTOK;

	NotifyEvent* e = new NotifyEvent(NULLCP, NULLCP, NULLCP,
				NULLCP, NULLCP, mflt, mgr, False);
	_evList.append(e);
	++_nFlt;
	_hasFilter = True;

        advise(NULLCP, "RMIBAgent: manager registered to receive events \
using filter, %s",
		sprintmfilter(mflt));
    }
    return adjustReporting();
}


int RMIBAgent::stopReceiveEvent (CMISFilter* mflt, RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (!mgr || !_listening || (_hasFilter && !mflt) ||
	    (!_hasFilter && mflt) || (_nAll && mflt))
	return NOTOK;

    if (!mflt) {
	if (stopReceiveAllEvent(mgr) == NOTOK)
	    return NOTOK;
    }
    else {
	NotifyEvent* e;
	if ((e = getEvent(mflt, mgr)) && removeEvent(e) == OK) {
            advise(NULLCP, "RMIBAgent: manager de-registered from receiving \
events using filter, %s",
		sprintmfilter(mflt));
	    --_nFlt;
	    _hasFilter = False;
	}
	else {
	    return NOTOK;
	}
    }
    return adjustReporting();
}


int RMIBAgent::cancelTimeAssertion (RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (!mgr)
	return NOTOK;

    NotifyEvent* e;
    _evIterator.reset();
    while (e = (NotifyEvent*) _evIterator.getNext()) {
	if (e->eventTime && e->mgr == mgr && removeEvent(e) == OK) {
	    --_nFlt;
	    _evIterator.reset();
	}
    }
    return adjustReporting();
}


int RMIBAgent::CmisGet (OID oid, MN mn, CMISScope* scope, CMISFilter* filter,
	External* access, CMISSync sync, int nattrs, OID oid_attrs[],
	CMISObjectList*& objList, RMIBManager* mgr, Bool perPdu)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    register int i;
    struct MIDentifier obj_class, attrs[MAXATTRS];
    oid2mid(oid, &obj_class);

    if (!mn && !_inst_mn) {
	_inst_mn = new MName;
	_inst_mn->mn_type = MN_LOCALDN;
	_inst_mn->mn_dn = dn_comp_new(NULLRDN);
    }

    for (i = 0; i < nattrs; i++)
	oid2mid(oid_attrs[i], &attrs[i]);

    if (M_Get(_msd, i = get_invoke(), &obj_class, mn ? mn : _inst_mn,
	    scope, filter, access, sync, nattrs, attrs, &_mi) == NOTOK) {
#ifdef DEBUG
	advise(NULLCP, "RMIBAgent: M_Get failed, provider rejected, %s",
		_mi.mi_preject.mpr_data);
#endif
	mi_free(&_mi);
	return NOTOK;
    }

    _objList = (mgr && perPdu) ? NULLCMISOBJECTLIST
			: new CMISObjectList(i, M_GET);

    if (!mgr) {
	_blocking = True;

	do {
	    if ((i = readMsapIndication()) == NOTOK)
		_blocking = False;
	} while (_blocking);

	objList = _objList;
	_objList = NULLCMISOBJECTLIST;
	return i;
    }

    RMIBCallback* c = new RMIBCallback(i, mgr, _objList, M_GET);
    _callbackList.append(c);
    return i;
}


int RMIBAgent::CmisSet (OID oid, MN mn, CMISScope* scope, CMISFilter* filter,
	External* access, CMISSync sync, AVAArray* ava_attrs,
	CMISObjectList*& objList, RMIBManager* mgr, Bool noConf)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    register int i;
    struct MIDentifier obj_class;
    oid2mid(oid, &obj_class);
    CMISSetAttr attrs[MAXATTRS];
    int nattrs = ava_attrs ? ava_attrs->getCount() : 0;

    if (!mn && !_inst_mn) {
	_inst_mn = new MName;
	_inst_mn->mn_type = MN_LOCALDN;
	_inst_mn->mn_dn = dn_comp_new(NULLRDN);
    }

    for (i = 0; i < nattrs; i++) {
	AVA* ava = (AVA*) ava_attrs->get(i);
	attrs[i].sa_id.mid_type = MID_GLOBAL;
	attrs[i].sa_id.mid_global = ava ? ava->getOid() : NULLOID;
	Attr* value = ava ? ava->getValue() : NULLATTR;
	attrs[i].sa_val = value ? value->encode() : NULLPE;
	attrs[i].sa_modify = ava ? ava->getModifyOp() : m_noModifyOp;
    }

    if (noConf) {
	if (M_Set(_msd, get_invoke(), &obj_class, mn ? mn : _inst_mn,
		scope, filter, access, sync, nattrs, attrs, &_mi) == NOTOK) {
#ifdef DEBUG
	    advise(NULLCP, "RMIBAgent: M_Set failed, provider rejected, %s",
		_mi.mi_preject.mpr_data);
#endif
	    mi_free(&_mi);
	    return NOTOK;
	}
	return OK;
    }

    if (M_SetConf(_msd, i = get_invoke(), &obj_class, mn ? mn : _inst_mn,
	    scope, filter, access, sync, nattrs, attrs, &_mi) == NOTOK) {
#ifdef DEBUG
	advise(NULLCP, "RMIBAgent: M_SetConf failed, provider rejected, %s",
		_mi.mi_preject.mpr_data);
#endif
	mi_free(&_mi);
	return NOTOK;
    }

    _objList = new CMISObjectList(i, M_SET);

    if (!mgr) {
	_blocking = True;

	do {
	    if ((i = readMsapIndication()) == NOTOK)
		_blocking = False;
	} while (_blocking);

	objList = _objList;
	_objList = NULLCMISOBJECTLIST;
	return i;
    }

    RMIBCallback* c = new RMIBCallback(i, mgr, _objList, M_SET);
    _callbackList.append(c);
    return i;
}


int RMIBAgent::CmisAction (OID oid, MN mn, CMISScope* scope, CMISFilter* filter,
	External* access, CMISSync sync, AVA* action, CMISObjectList*& objList,
	RMIBManager* mgr, Bool noConf)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    register int i;
    struct MIDentifier obj_class, action_type;
    oid2mid(oid, &obj_class);

    if (!mn && !_inst_mn) {
	_inst_mn = new MName;
	_inst_mn->mn_type = MN_LOCALDN;
	_inst_mn->mn_dn = dn_comp_new(NULLRDN);
    }

    action_type.mid_type = MID_GLOBAL;
    action_type.mid_global = action ? action->getOid() : NULLOID;
    Attr* action_value = action ? action->getValue() : NULLATTR;

    if (noConf) {
	if (M_Action(_msd, get_invoke(), &obj_class, mn ? mn : _inst_mn,
		scope, filter, access, sync, &action_type,
		action_value ? action_value->encode() : NULLPE, &_mi)
	    == NOTOK) {
#ifdef DEBUG
	    advise(NULLCP, "RMIBAgent: M_Action failed, provider rejected, %s",
			_mi.mi_preject.mpr_data);
#endif
		mi_free(&_mi);
		return NOTOK;
	}
	return OK;
    }

    if (M_ActionConf(_msd, i = get_invoke(), &obj_class,
	    mn ? mn : _inst_mn, scope, filter, access, sync, &action_type,
	    action_value ? action_value->encode() : NULLPE, &_mi) == NOTOK) {
#ifdef DEBUG
	advise(NULLCP, "RMIBAgent: M_ActionConf failed, provider rejected, %s",
		_mi.mi_preject.mpr_data);
#endif
	mi_free(&_mi);
	return NOTOK;
    }

    _objList = new CMISObjectList(i, M_ACTION);

    if (!mgr) {
	_blocking = True;

	do {
	    if ((i = readMsapIndication()) == NOTOK)
		_blocking = False;
	} while (_blocking);

	objList = _objList;
	_objList = NULLCMISOBJECTLIST;
	return i;
    }

    RMIBCallback* c = new RMIBCallback(i, mgr, _objList, M_ACTION);
    _callbackList.append(c);
    return i;
}


int RMIBAgent::CmisCreate (OID oid, MN mn, int inst_type, MN ref_inst_mn,
	External* access, AVAArray* ava_attrs, CMISObject*& obj,
	RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    register int i;
    struct MIDentifier obj_class;
    oid2mid(oid, &obj_class);
    CMISParam attrs[MAXATTRS];
    int nattrs = ava_attrs ? ava_attrs->getCount() : 0;

    for (i = 0; i < nattrs; i++) {
	AVA* ava = (AVA*) ava_attrs->get(i);
	attrs[i].mp_id.mid_type = MID_GLOBAL;
	attrs[i].mp_id.mid_global = ava ? ava->getOid() : NULLOID;
	Attr* value = ava ? ava->getValue() : NULLATTR;
	attrs[i].mp_val = value ? value->encode() : NULLPE;
    }

    if (M_Create(_msd, i = get_invoke(), &obj_class, mn, inst_type,
	    ref_inst_mn, access, nattrs, attrs, &_mi) == NOTOK) {
#ifdef DEBUG
	advise(NULLCP, "RMIBAgent: M_Create failed, provider rejected, %s",
		_mi.mi_preject.mpr_data);
#endif
	mi_free(&_mi);
	return NOTOK;
    }

    if (!mgr) {
	_blocking = True;

	do {
	    if ((i = readMsapIndication()) == NOTOK)
		_blocking = False;
	} while (_blocking);

	obj = _obj;
	_obj = NULLCMISOBJECT;
	return i;
    }

    RMIBCallback* c = new RMIBCallback(i, mgr, NULLCMISOBJECTLIST, M_CREATE);
    _callbackList.append(c);
    return i;
}


int RMIBAgent::CmisDelete (OID oid, MN mn, CMISScope* scope, CMISFilter* filter,
	External* access, CMISSync sync, CMISObjectList*& objList,
	RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    register int i;
    struct MIDentifier obj_class;
    oid2mid(oid, &obj_class);

    if (!mn && !_inst_mn) {
	_inst_mn = new MName;
	_inst_mn->mn_type = MN_LOCALDN;
	_inst_mn->mn_dn = dn_comp_new(NULLRDN);
    }

    if (M_Delete(_msd, i = get_invoke(), &obj_class, mn ? mn : _inst_mn,
	    scope, filter, access, sync, &_mi) == NOTOK) {
#ifdef DEBUG
	advise(NULLCP, "RMIBAgent: M_Delete failed, provider rejected, %s",
		_mi.mi_preject.mpr_data);
#endif
	mi_free(&_mi);
	return NOTOK;
    }

    _objList = new CMISObjectList(i, M_DELETE);

    if (!mgr) {
	_blocking = True;

	do {
	    if ((i = readMsapIndication()) == NOTOK)
		_blocking = False;
	} while (_blocking);

	objList = _objList;
	_objList = NULLCMISOBJECTLIST;
	return i;
    }

    RMIBCallback* c = new RMIBCallback(i, mgr, _objList, M_DELETE);
    _callbackList.append(c);
    return i;
}


int RMIBAgent::Get (char* objClass, char* objInst, int scopeType,
	int scopeLevel, char* filter, External* access, CMISSync sync,
	char* attrs[], CMISObjectList*& objList, RMIBManager* mgr, Bool perPdu)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    CMISScope scope;
    if (checkScope(scopeType, scopeLevel, &scope) == NOTOK ||
	    (sync != s_bestEffort && sync != s_atomic))
	return NOTOK;

    OID oid = objClass ? name2oid(objClass) : NULLOID;
    MN inst_mn = objInst ? ObjectInstance_parse(objInst) : NULLMN;

    if ((objClass && !oid) || (objInst && !inst_mn)) {
#ifdef DEBUG
	advise(NULLCP, "Get : base object translation failed for class \
%s, inst %s",
	    objClass, objInst);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    CMISFilter* flt = encodeFilter(filter);
    if (filter && !flt) {
#ifdef DEBUG
	advise(NULLCP, "Get : invalid filter expr %s", filter);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    register int n = GET_ALLATTRS;
    OID oid_attrs[MAXATTRS];

    if (attrs) {
	if (!attrs[0] || strncasecmp(attrs[0], "none", 4) == 0)
	    n = GET_NOATTRS;
	else
	if (strncasecmp(attrs[0], "all", 3) == 0)
	    n = GET_ALLATTRS;
	else
	for (n = 0; attrs[n]; n++)
	    if (!(oid_attrs[n] = name2oid(attrs[n]))) {
#ifdef DEBUG
		advise(NULLCP, "Get : unknown attribute type, %s", attrs[n]);
#endif
		if (oid) oid_free(oid);
		if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
		if (flt) mfilter_free(flt);
		return NOTOK;
	    }
    }

    int result = CmisGet(oid, inst_mn, &scope, flt, access,
		    (scopeType != Sc_BaseObject) ? sync : NULLMSYNC,
		    n, oid_attrs, objList, mgr, perPdu);

    if (oid) oid_free(oid);
    if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
    if (flt) mfilter_free(flt);

    while (--n >= 0)
	oid_free(oid_attrs[n]);

    return result;
}


int RMIBAgent::CancelGet (int getId)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    MSAPIndication mi;
    if (M_CancelGet(_msd, get_invoke(), getId, &mi) == NOTOK) {
#ifdef DEBUG
	advise(NULLCP, "RMIBAgent: M_CancelGet failed, provider rejected %s",
		mi.mi_preject.mpr_data);
#endif
	mi_free(&mi);
	return NOTOK;
    }

    RMIBCallback* c = getCallback(getId);
    if (c) {
	delete c->objList;
	removeCallback(c);
    }
    return OK;
}


int RMIBAgent::Set (char* objClass, char* objInst, int scopeType,
	int scopeLevel, char* filter, External* access, CMISSync sync,
	AVAArray* attrs, CMISObjectList*& objList, RMIBManager* mgr,
	Bool noConf)
{
    if (!attrs || !attrs->getCount())
	return NOTOK;

    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    CMISScope scope;
    if (checkScope(scopeType, scopeLevel, &scope) == NOTOK ||
	    (sync != s_bestEffort && sync != s_atomic))
	return NOTOK;

    OID oid = objClass ? name2oid(objClass) : NULLOID;
    MN inst_mn = objInst ? ObjectInstance_parse(objInst) : NULLMN;

    if ((objClass && !oid) || (objInst && !inst_mn)) {
#ifdef DEBUG
	advise(NULLCP, "Set : base object translation failed for class \
%s, inst %s",
	    objClass, objInst);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    CMISFilter* flt = encodeFilter(filter);
    if (filter && !flt) {
#ifdef DEBUG
	advise(NULLCP, "Set : invalid filter expr %s", filter);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    int result = CmisSet(oid, inst_mn, &scope, flt, access, 
		    (scopeType != Sc_BaseObject) ? sync : NULLMSYNC,
		    attrs, objList, mgr, noConf);

    if (oid) oid_free(oid);
    if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
    if (flt) { mfilter_free(flt); }

    return result;
}


int RMIBAgent::Action (char* objClass, char* objInst, int scopeType,
	int scopeLevel, char* filter, External* access, CMISSync sync,
	AVA* action, CMISObjectList*& objList, RMIBManager* mgr, Bool noConf)
{
    if (!action)
	return NOTOK;

    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    CMISScope scope;
    if (checkScope(scopeType, scopeLevel, &scope) == NOTOK ||
	    (sync != s_bestEffort && sync != s_atomic))
	return NOTOK;

    OID oid = objClass ? name2oid(objClass) : NULLOID;
    MN inst_mn = objInst ? ObjectInstance_parse(objInst) : NULLMN;

    if ((objClass && !oid) || (objInst && !inst_mn)) {
#ifdef DEBUG
	advise(NULLCP, "Action : base object translation failed for \
class %s, inst %s",
	    objClass, objInst);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    CMISFilter* flt = encodeFilter(filter);
    if (filter && !flt) {
#ifdef DEBUG
	advise(NULLCP, "Action : invalid filter expr %s", filter);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    int result = CmisAction(oid, inst_mn, &scope, flt, access,
		    (scopeType != Sc_BaseObject) ? sync : NULLMSYNC,
		    action, objList, mgr, noConf);

    if (oid) oid_free(oid);
    if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
    if (flt) mfilter_free(flt);

    return result;
}


int RMIBAgent::Create (char* objClass, char* objInst, int instType,
	char* refInst, External* access, AVAArray* attrs, CMISObject*& obj,
	RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    OID oid = objClass ? name2oid(objClass) : NULLOID;
    MN inst_mn = objInst ? ObjectInstance_parse(objInst) : NULLMN;

    if ((objClass && !oid) || (objInst && !inst_mn)) {
#ifdef DEBUG
	advise(NULLCP, "Create : base object translation failed for \
class %s, inst %s",
		objClass, objInst);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    MN ref_inst_mn = refInst ? ObjectInstance_parse(refInst) : NULLMN;
    if (refInst && !ref_inst_mn) {
#ifdef DEBUG
	advise(NULLCP, "Create : invalid reference inst translated %s",
		refInst);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    int result = CmisCreate(oid, inst_mn, instType,
			ref_inst_mn, access, attrs, obj, mgr);

    if (oid) oid_free(oid);
    if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
    if (ref_inst_mn) { mn_free(ref_inst_mn); free((char*) ref_inst_mn); }

    return result;
}


int RMIBAgent::Delete (char* objClass, char* objInst, int scopeType,
	int scopeLevel, char* filter, External* access, CMISSync sync,
	CMISObjectList*& objList, RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    if (mgr && !_listening && startListening() == NOTOK)
	return NOTOK;

    CMISScope scope;
    if (checkScope(scopeType, scopeLevel, &scope) == NOTOK ||
	    (sync != s_bestEffort && sync != s_atomic))
	return NOTOK;

    OID oid = objClass ? name2oid(objClass) : NULLOID;
    MN inst_mn = objInst ? ObjectInstance_parse(objInst) : NULLMN;

    if ((objClass && !oid) || (objInst && !inst_mn)) {
#ifdef DEBUG
	advise(NULLCP, "Delete : base object translation failed for \
class %s, inst %s",
		objClass, objInst);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    CMISFilter* flt = encodeFilter(filter);
    if (filter && !flt) {
#ifdef DEBUG
	advise(NULLCP, "Delete : invalid filter expr %s", filter);
#endif
	if (oid) oid_free(oid);
	if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
	return NOTOK;
    }

    int result = CmisDelete(oid, inst_mn, &scope, flt, access,
		    (scopeType != Sc_BaseObject) ? sync : NULLMSYNC,
		    objList, mgr);

    if (oid) oid_free(oid);
    if (inst_mn) { mn_free(inst_mn); free((char*) inst_mn); }
    if (flt) mfilter_free(flt);

    return result;
}


int RMIBAgent::cancelCallbacks (RMIBManager* mgr)
{
    if (_msd == UNCONNECTED) {
	advise(NULLCP, "RMIBAgent: not in association!!!");
	return NOTOK;
    }

    NotifyEvent* e;
    _evIterator.reset();
    while (e = (NotifyEvent*) _evIterator.getNext()) {
	if (e->mgr != mgr) continue;

	if (e->isAllEvent()) {
	    if (removeEvent(e) == OK) {
		--_nAll;
		_evIterator.reset();
	    }
	}
	else {
	    if (removeEvent(e) == OK) {
		--_nFlt;
		_evIterator.reset();
	    }
	}
    }

    RMIBCallback* c;
    _callbackIterator.reset();
    while (c = (RMIBCallback*) _callbackIterator.getNext()) {
	if (c->mgr == mgr && removeCallback(c) == OK)
	    _callbackIterator.reset();
    }
    return adjustReporting();
}


void RMIBAgent::print ()
{
    fprintf(stdout, "\n############### dumping rmib agent ###############\n\n");
    fprintf(stdout, "agent name              =     %s\n", _agent);
    fprintf(stdout, "host name               =     %s\n", _host);
    fprintf(stdout, "service descriptor      =     %d\n", _msd);
    fprintf(stdout, "timeout interval        =     %d seconds\n", _interval);
    fprintf(stdout, "current local time      =     %s\n", now());
    char* cp = _efd_mn ? dn2str(_efd_mn->mn_dn) : NULLCP;
    if (cp) {
	fprintf(stdout, "efd object instance     =     {%s}\n", cp);
	free(cp);
    }
    fprintf(stdout, "filtered entries        =     %d\n", _nFlt);
    fprintf(stdout, "all event entries       =     %d\n", _nAll);
    fprintf(stdout, "total event entries     =     %d\n", _evList.getCount());
    NotifyEvent* e;
    _evIterator.reset();
    while (e = (NotifyEvent*) _evIterator.getNext()) {
	fprintf(stdout, "\n\tclass       =  %s\n", e->objClass);
	fprintf(stdout, "\tinstance    =  %s\n", e->objInst);
	fprintf(stdout, "\tevent type  =  %s\n", e->eventType);
	fprintf(stdout, "\tevent time  =  %s\n", e->eventTime);
	fprintf(stdout, "\tsfilter     =  %s\n", e->sfilter);
	fprintf(stdout, "\tmfilter     =  %s\n", sprintmfilter(e->mfilter));
	fprintf(stdout, "\tmanager     =  %d\n", e->mgr);
    }
    fprintf(stdout, "\n##################################################\n");
}


RMIBAgent::RMIBAgent (char* agent, char* host)
{
    init_invoke();
    _msd = UNCONNECTED;
    _interval = DFLT_TIMEOUT;
    _agent = agent ? strdup(agent) : NULLCP;
    _host = host ? strdup(host) : NULLCP;
    _inst_mn = NULLMN;
    _listening = _blocking = False;
    _hasFilter = False;
    _nFlt = _nAll = 0;
    _efd_mid = NULLMID;
    _efd_mn = NULLMN;
    _efd_name = NULLCP;
    _evList.clear();
    _evIterator.setList(&_evList);
    _callbackList.clear();
    _callbackIterator.setList(&_callbackList);
    _obj = NULLCMISOBJECT;
    _objList = NULLCMISOBJECTLIST;
    _report = (CMISEventReportArg*) 0;
    _title = NULLCP;
}
