// RMIBresult.h - the Remote MIB result class methods

// By George Pavlou / Thurain Tin, October 1993


#ifndef _RMIB_RESULT_
#define _RMIB_RESULT_

#ifndef _MSAP_
#include "msap.h"
#endif

#ifndef _AVA_
#include "AVA.h"
#endif

#ifndef _GENERIC_LIST_
#include "GenList.h"
#endif


class CMISObject
{
    friend class RMIBAgent;

    OID		_oid;
    char*	_strClass;
    DN		_dn;
    char*	_strInst;
    char*	_strTime;	
    AVAArray*	_attrs;		// Get, Set
    AVA*	_ava;		// Action
    AVA*	_error;		// error code / info
    int		_invoke;
    int		_operation;

    void	init (OID, DN, char*, AVAArray*, AVA*, int, int);

protected:

    void	setError (CMISErrors);
    void	setError (AVA*);

public:

    char*	getClass ();
    OID		getClassOid ();
    char*	getInstance ();
    DN		getInstanceDn ();
    char*	getTime ();
    AVAArray*	getAttrs ();
    AVA*	getActionRes ();
    CMISErrors	getError ();
    char*	getErrorStr ();
    AVA*	getErrorInfo ();
    int		getInvoke ();
    int		getOperation ();
    void	print ();

    CMISObject (OID, DN, char*, AVAArray*, AVA*,
		CMISErrors, int = 0, int = 0);
    CMISObject (OID, DN, char*, AVAArray*, AVA*,
		AVA*, int = 0, int = 0);
    ~CMISObject ();
};

#define NULLCMISOBJECT ((CMISObject*) 0)


// inline methods

inline void CMISObject::setError (CMISErrors e)
{ if (_error) delete _error; _error = new AVA(e); }

inline void CMISObject::setError (AVA* e)
{ if (_error) delete _error; _error = e; }

inline char* CMISObject::getClass ()		{ return _strClass; }
inline OID CMISObject::getClassOid ()		{ return _oid; }
inline char* CMISObject::getInstance ()		{ return _strInst; }
inline DN CMISObject::getInstanceDn ()		{ return _dn; }
inline char* CMISObject::getTime ()		{ return _strTime; }
inline AVAArray* CMISObject::getAttrs ()	{ return _attrs; }
inline AVA* CMISObject::getActionRes ()		{ return _ava; }
inline CMISErrors CMISObject::getError ()	{ return _error->getError(); }
inline char* CMISObject::getErrorStr ()		{ return _error->getErrorStr();}
inline AVA* CMISObject::getErrorInfo ()		{ return _error; }
inline int CMISObject::getInvoke ()		{ return _invoke; }
inline int CMISObject::getOperation ()		{ return _operation; }


class CMISObjectList : public List
{
    friend class RMIBAgent;

    int		_invoke;
    int		_operation;

protected:

    void	deleteElem (void*);

public:

    CMISErrors	getError ();
    char*	getErrorStr ();
    int		getInvoke ();
    int		getOperation ();
    void	print ();

    CMISObjectList (int = 0, int = 0);
    ~CMISObjectList ();
};

#define NULLCMISOBJECTLIST ((CMISObjectList*) 0)


// inline methods

inline void CMISObjectList::deleteElem (void* v)
{ if (v) delete (CMISObject*) v; }

inline CMISErrors CMISObjectList::getError ()
{ return ((CMISObject*) first())->getError(); }

inline char* CMISObjectList::getErrorStr ()
{ return ((CMISObject*) first())->getErrorStr();}

inline int CMISObjectList::getInvoke ()		{ return _invoke; }
inline int CMISObjectList::getOperation ()	{ return _operation; }
inline CMISObjectList::CMISObjectList (int id, int op) : List ()
{ _invoke = id; _operation = op; }
inline CMISObjectList::~CMISObjectList ()	{ clear(); }


#endif /* _RMIB_RESULT_ */
