// RMIButil.cc - the Remote MIB util realisation

// By George Pavlou / Thurain Tin, October 1993


#include <stdlib.h>
#include <string.h>
#include "msap.h"
#include "RMIButil.h"


int strcmp1 (char* s1, char* s2)
{
    if (!s1 || !s2)
	return (!s1 && !s2) ? OK : NOTOK;

    return strcmp(s1, s2) == 0 ? OK : NOTOK;
}


CMISFilter* buildFilter (char* objClass,
	char* objInst, char* eventType, char* eventTime)
{
    CMISFilter* f;
    char buf[256], *cp = buf;

    sprintf(cp++, " ");

    if (objClass) {
	sprintf(cp, "(managedObjectClass=%s)", objClass);
	cp += strlen(cp);
    }

    if (objInst) {
	if (objClass) {
	    sprintf(cp++, "&");
	    buf[0] = '(';
	}
	sprintf(cp, "(managedObjectInstance=%s)", objInst);
	cp += strlen(cp);
    }

    if (eventType) {
	if (objClass || objInst) {
	    sprintf(cp++, "&");
	    buf[0] = '(';
	}
	sprintf(cp, "(eventType=%s)", eventType);
	cp += strlen(cp);
    }

    if (eventTime) {
	if (objClass || objInst || eventType) {
	    sprintf(cp++, "&");
	    buf[0] = '(';
	}
	sprintf(cp, "%s", eventTime);
	cp += strlen(cp);
    }

    if (buf[0] == '(')
	sprintf(cp, ")");

    return (f = str2mfilter(buf)) ? f : NULLMFILTER;
}


NotifyEvent::NotifyEvent (RMIBManager* manager)
{
    objClass = objInst = eventType = eventTime = NULLCP;
    sfilter = NULLCP; mfilter = NULLMFILTER;
    mgr = manager;
    del = False;
}


NotifyEvent::NotifyEvent (char* obj_class, char* obj_inst,
	char* event_type, char* event_time, char* s_filter,
	CMISFilter* m_filter, RMIBManager* manager, Bool delMode)
{
    objClass = obj_class ? strdup(obj_class) : NULLCP;
    objInst = obj_inst ? strdup(obj_inst) : NULLCP;
    eventType = event_type ? strdup(event_type) : NULLCP;
    eventTime = event_time ? strdup(event_time) : NULLCP;
    sfilter = s_filter ? strdup(s_filter) : NULLCP;
    mfilter = m_filter ? m_filter : NULLMFILTER;
    mgr = manager;
    del = delMode;
}


NotifyEvent::~NotifyEvent ()
{
    if (objClass)
	free(objClass);
    objClass = NULLCP;

    if (objInst)
	free(objInst);
    objInst = NULLCP;

    if (eventType)
	free(eventType);
    eventType = NULLCP;

    if (eventTime)
	free(eventTime);
    eventTime = NULLCP;

    if (sfilter)
	free(sfilter);
    sfilter = NULLCP;

    if (mfilter && del)
	mfilter_free(mfilter);
    mfilter = NULLMFILTER;
}


CMISObject* process_getres (int id, CMISGetRes* gr, CMISErrors error)
{
    if (!gr)
	return NULLCMISOBJECT;

    AVAArray*
    attrs = gr->gr_nattrs ? new AVAArray(gr->gr_nattrs) : NULLAVAARRAY;

    for (register int i = 0; i < gr->gr_nattrs; i++) {
	struct CMISParam* g_ava = &gr->gr_attrs[i].ga_ava;
	AVA* ava = new AVA(g_ava->mp_id.mid_global,
				g_ava->mp_val ? g_ava->mp_val : NULLPE,
				gr->gr_attrs[i].ga_error);

	if (ava->createError()) {
	    fprintf(stderr, "AVA creation failed for %s\n",
				sprintoid(g_ava->mp_id.mid_global));
	    delete ava;
	    ava = NULLAVA;
	}
	else {
	    g_ava->mp_id.mid_global = NULLOID;
	    if (g_ava->mp_val) g_ava->mp_val = NULLPE;
	}
	attrs->add(ava);
    }

    CMISObject* obj = new CMISObject(gr->gr_class.mid_global,
				gr->gr_inst.mn_dn, gr->gr_time,
				attrs, NULLAVA, error, id, M_GET);

    gr->gr_class.mid_global = NULLOID;
    gr->gr_inst.mn_dn = NULLDN;
    gr->gr_time = NULLCP;

    return obj;
}


CMISObject* process_setres (int id, CMISSetRes* sr, CMISErrors error)
{
    if (!sr)
	return NULLCMISOBJECT;

    AVAArray*
    attrs = sr->sr_nattrs ? new AVAArray(sr->sr_nattrs) : NULLAVAARRAY;

    for (register int i = 0; i < sr->sr_nattrs; i++) {
	struct CMISParam* s_ava = &sr->sr_attrs[i].sa_ava;
	AVA* ava = new AVA(s_ava->mp_id.mid_global,
				s_ava->mp_val ? s_ava->mp_val : NULLPE,
				sr->sr_attrs[i].sa_error);

	if (ava->createError()) {
	    fprintf(stderr, "AVA creation failed for %s\n",
				sprintoid(s_ava->mp_id.mid_global));
	    delete ava;
	    ava = NULLAVA;
	}
	else {
	    s_ava->mp_id.mid_global = NULLOID;
	    if (s_ava->mp_val) s_ava->mp_val = NULLPE;
	}
	attrs->add(ava);
    }

    CMISObject* obj = new CMISObject(sr->sr_class.mid_global,
				sr->sr_inst.mn_dn, sr->sr_time,
				attrs, NULLAVA, error, id, M_SET);

    sr->sr_class.mid_global = NULLOID;
    sr->sr_inst.mn_dn = NULLDN;
    sr->sr_time = NULLCP;

    return obj;
}


CMISObject* process_actionres (int id, CMISActionRes* ar)
{
    if (!ar)
	return NULLCMISOBJECT;

    struct CMISParam* ar_reply = &ar->ar_reply;
    AVA* ava = NULLAVA;

    if (ar_reply->mp_id.mid_global && ar_reply->mp_val) {
	// add the "Reply" suffix to the action OID
	massage_actionReplyOID(ar_reply->mp_id.mid_global);

	ava = new AVA(ar_reply->mp_id.mid_global, ar_reply->mp_val);

	if (ava->createError()) {
	    fprintf(stderr, "AVA creation failed for %s\n",
				sprintoid(ar_reply->mp_id.mid_global));
	    delete ava;
	    ava = NULLAVA;
	}
	else {
	    ar_reply->mp_id.mid_global = NULLOID;
	    ar_reply->mp_val = NULLPE;
	}
    }

    CMISObject* obj = new CMISObject(ar->ar_class.mid_global,
				ar->ar_inst.mn_dn, ar->ar_time,
				NULLAVAARRAY, ava, NULLMERROR, id, M_ACTION);

    ar->ar_class.mid_global = NULLOID;
    ar->ar_inst.mn_dn = NULLDN;
    ar->ar_time = NULLCP;

    return obj;
}


CMISObject* process_createres (int id, CMISCreateRes* cr)
{
    if (!cr)
	return NULLCMISOBJECT;

    AVAArray*
    attrs = cr->cr_nattrs ? new AVAArray(cr->cr_nattrs) : NULLAVAARRAY;

    for (register int i = 0; i < cr->cr_nattrs; i++) {
	struct CMISParam* c_ava = &cr->cr_attrs[i];

	AVA* ava = new AVA(c_ava->mp_id.mid_global, c_ava->mp_val);

	if (ava->createError()) {
	    fprintf(stderr, "AVA creation failed for %s\n",
				sprintoid(c_ava->mp_id.mid_global));
	    delete ava;
	    ava = NULLAVA;
	}
	else {
	    c_ava->mp_id.mid_global = NULLOID;
	    c_ava->mp_val = NULLPE;
	}
	attrs->add(ava);
    }

    CMISObject* obj = new CMISObject(cr->cr_class.mid_global,
				cr->cr_inst.mn_dn, cr->cr_time, attrs,
				NULLAVA, NULLMERROR, id, M_CREATE);

    cr->cr_class.mid_global = NULLOID;
    cr->cr_inst.mn_dn = NULLDN;
    cr->cr_time = NULLCP;

    return obj;
}


CMISObject* process_deleteres (int id, CMISDeleteRes* dr)
{
    if (!dr)
	return NULLCMISOBJECT;

    CMISObject* obj = new CMISObject(dr->dr_class.mid_global,
				dr->dr_inst.mn_dn, dr->dr_time, NULLAVAARRAY,
				NULLAVA, NULLMERROR, id, M_DELETE);

    dr->dr_class.mid_global = NULLOID;
    dr->dr_inst.mn_dn = NULLDN;
    dr->dr_time = NULLCP;

    return obj;
}


CMISObject* process_procfailure (int id, CMISProcessingFailure* pf)
{
    if (!pf)
	return NULLCMISOBJECT;

    AVA* errorInfo = new AVA(pf->pf_error.mp_id.mid_global,
				pf->pf_error.mp_val, m_processingFailure);

    if (errorInfo->createError()) {
	fprintf(stderr, "AVA creation failed for %s\n",
				sprintoid(pf->pf_error.mp_id.mid_global));
	delete errorInfo;
	errorInfo = new AVA(m_processingFailure);
    }
    else {
	pf->pf_error.mp_id.mid_global = NULLOID;
	pf->pf_error.mp_val = NULLPE;
    }

    CMISObject* obj = new CMISObject(pf->pf_class.mid_global, pf->pf_inst.mn_dn,
				NULLCP, NULLAVAARRAY, NULLAVA, errorInfo, id);

    pf->pf_class.mid_global = NULLOID;
    pf->pf_inst.mn_dn = NULLDN;

    return obj;
}


CMISObject* process_actionerr (int id, CMISActionError* ae)
{
    if (!ae)
	return NULLCMISOBJECT;

    CMISObject* obj = new CMISObject(ae->ae_class.mid_global,
				ae->ae_inst.mn_dn, ae->ae_time, NULLAVAARRAY,
				NULLAVA, ae->ae_error, id, M_ACTION);

    ae->ae_class.mid_global = NULLOID;
    ae->ae_inst.mn_dn = NULLDN;
    ae->ae_time = NULLCP;

    return obj;
}


CMISObject* process_deleteerr (int id, CMISDeleteError* de)
{
    if (!de)
	return NULLCMISOBJECT;

    CMISObject* obj = new CMISObject(de->de_class.mid_global,
				de->de_inst.mn_dn, de->de_time, NULLAVAARRAY,
				NULLAVA, de->de_error, id, M_DELETE);

    de->de_class.mid_global = NULLOID;
    de->de_inst.mn_dn = NULLDN;
    de->de_time = NULLCP;

    return obj;
}


CMISObject* process_errinfo (int id, CMISErrorInfo* info,
		CMISErrors error)
{
    if (!info)
	return NULLCMISOBJECT;

    OID oid = NULLOID;
    DN dn = NULLDN;
    AVA* errorInfo = NULLAVA;

    switch (error) {
    case m_noSuchAttribute:
	errorInfo = new AVA(m_noSuchAttribute,
			info->ei_noSuchAttribute.mid_global);
	info->ei_noSuchAttribute.mid_global = NULLOID;
	break;

    case m_missingAttributeValue:
	errorInfo = new AVA(m_missingAttributeValue,
			info->ei_missingAttributeValue->midl_mid.mid_global);
	info->ei_missingAttributeValue->midl_mid.mid_global = NULLOID;
	break;

    case m_invalidAttributeValue:
	errorInfo = new AVA(info->ei_invalidAttributeValue.mp_id.mid_global,
			info->ei_invalidAttributeValue.mp_val,
			m_invalidAttributeValue);
	if (errorInfo->createError()) {
	    fprintf(stderr, "AVA creation failed for %s\n",
		sprintoid(info->ei_invalidAttributeValue.mp_id.mid_global));
	    delete errorInfo;
	    errorInfo = new AVA(m_invalidAttributeValue);
	}
	else {
	    info->ei_invalidAttributeValue.mp_id.mid_global = NULLOID;
	    info->ei_invalidAttributeValue.mp_val = NULLPE;
	}
	break;

    case m_processingFailure:
	errorInfo = new AVA(
			info->ei_processingFailure.pf_error.mp_id.mid_global,
			info->ei_processingFailure.pf_error.mp_val,
			m_processingFailure);
	if (errorInfo->createError()) {
	    fprintf(stderr, "AVA creation failed for %s\n",
	       sprintoid(info->ei_processingFailure.pf_error.mp_id.mid_global));
	    delete errorInfo;
	    errorInfo = new AVA(m_processingFailure);
	}
	else {
	    info->ei_processingFailure.pf_error.mp_id.mid_global = NULLOID;
	    info->ei_processingFailure.pf_error.mp_val = NULLPE;
	}
	oid = info->ei_processingFailure.pf_class.mid_global;
	dn  = info->ei_processingFailure.pf_inst.mn_dn;
	info->ei_processingFailure.pf_class.mid_global = NULLOID;
	info->ei_processingFailure.pf_inst.mn_dn = NULLDN;
	break;

    default:
	break;
    }

    if (!errorInfo)
	errorInfo = new AVA(error);

    CMISObject* obj = new CMISObject(oid, dn, NULLCP, NULLAVAARRAY,
				NULLAVA, errorInfo, id);

    return obj;
}
