// RMIButil.h - the Remote MIB util methods

// By George Pavlou / Thurain Tin, October 1993


#ifndef _RMIB_UTIL_
#define _RMIB_UTIL_

#ifndef _RMIB_MANAGER_
#include "RMIBManager.h"
#endif


int strcmp1 (char*, char*);
CMISFilter* buildFilter (char*, char*, char*, char*);


// process msap results and errors

CMISObject* process_getres (int, CMISGetRes*, CMISErrors = m_noError);
CMISObject* process_setres (int, CMISSetRes*, CMISErrors = m_noError);
CMISObject* process_actionres (int, CMISActionRes*);
CMISObject* process_createres (int, CMISCreateRes*);
CMISObject* process_deleteres (int, CMISDeleteRes*);
CMISObject* process_procfailure (int, CMISProcessingFailure*);
CMISObject* process_actionerr (int, CMISActionError*);
CMISObject* process_deleteerr (int, CMISDeleteError*);
CMISObject* process_errinfo (int, CMISErrorInfo*, CMISErrors);


class NotifyEvent
{
public:

    char*	    objClass;
    char*	    objInst;
    char*	    eventType;
    char*	    eventTime;
    char*	    sfilter;
    CMISFilter*	    mfilter;
    RMIBManager*    mgr;
    Bool	    del;

    static int compare (NotifyEvent* e1, NotifyEvent* e2)
    {
	return (e1->mgr == e2->mgr &&
		strcmp1(e1->objClass, e2->objClass) == OK &&
		strcmp1(e1->objInst, e2->objInst) == OK &&
		strcmp1(e1->eventType, e2->eventType) == OK &&
		strcmp1(e1->eventTime, e2->eventTime) == OK &&
		strcmp1(e1->sfilter, e2->sfilter) == OK) ? OK : NOTOK;
    }

    Bool isAllEvent ()
    {
	return (objClass || objInst || eventType ||
		eventTime || sfilter || mfilter) ? False : True;
    }

    NotifyEvent (RMIBManager* m)
    {
	objClass = objInst = eventType = eventTime = NULLCP;
	sfilter = NULLCP;
	mfilter = NULLMFILTER;
	mgr = m; del = False;
    }

    NotifyEvent (char* cl, char* in, char* ty, char* tm, char* sf,
	CMISFilter* mf, RMIBManager* m, Bool de = True)
    {
	objClass = cl ? strdup(cl) : NULLCP;
	objInst = in ? strdup(in) : NULLCP;
	eventType = ty ? strdup(ty) : NULLCP;
	eventTime = tm ? strdup(tm) : NULLCP;
	sfilter = sf ? strdup(sf) : NULLCP;
	mfilter = mf ? mf : NULLMFILTER;
	mgr = m; del = de;
    }

    ~NotifyEvent ()
    {
	if (objClass)
	    free(objClass);
	objClass = NULLCP;

	if (objInst)
	    free(objInst);
	objInst = NULLCP;

	if (eventType)
	    free(eventType);
	eventType = NULLCP;

	if (eventTime)
	    free(eventTime);
	eventTime = NULLCP;

	if (sfilter)
	    free(sfilter);
	sfilter = NULLCP;

	if (mfilter && del)
	    mfilter_free(mfilter);
	mfilter = NULLMFILTER;
    }
};

#define NULLNOTIFYEVENT ((NotifyEvent*) 0)


class NotifyEventList : public List
{
protected:

    void deleteElem (void* e)
    { if (e) delete (NotifyEvent*) e; }

    int compare (void* e1, void* e2)
    { return NotifyEvent::compare((NotifyEvent*) e1, (NotifyEvent*) e2); }
};

#define NULLNOTIFYEVENTLIST ((NotifyEventList*) 0)


class RMIBCallback
{
public:

    int			id, operation;
    RMIBManager*	mgr;
    CMISObjectList*	objList;

    static int compare (RMIBCallback* c1, RMIBCallback* c2)
    { return (c1->id == c2->id) ? OK : NOTOK; }

    RMIBCallback (int i, RMIBManager* m, CMISObjectList* l, int o)
    { id = i; mgr = m; objList = l; operation = o; }

    ~RMIBCallback ()
    { id = operation = 0; }
};

#define NULLRMIBCALLBACK ((RMIBCallback*) 0)


class RMIBCallbackList : public List
{
protected:

    void deleteElem (void* c)
    { if (c) delete (RMIBCallback*) c; }

    int compare (void* c1, void* c2)
    { return RMIBCallback::compare((RMIBCallback*) c1, (RMIBCallback*) c2); }
};

#define NULLRMIBCALLBACKLIST ((RMIBCallbackList*) 0)


#endif /* _RMIB_UTIL_ */
