//
// Copyright (c) 1994 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// ActionNoConf.cc - connect to an SMA agent on a given host (or otherwise
//	using the location transparency service), do a non-confirmed Action
//	on the {uxObjId=test} instance, print the result and disconnect.

// By Thurain Tin, December 1994


#include <stdlib.h>
#include "msap.h"
#include "Error.h"
#include "RMIBAgent.h"
#include "Syntaxes.h"


main (int argc, char** argv)
{
    if (argc < 3)
	error(NULLCP, "usage: mactionnoconf <agent> <host>");

    RMIBAgent	agent;	// instantiate one Remote MIB agent
    // Note that the coordinator is only needed for asynchronous services,
    // this is why it is not instantiated here

    // load the syntaxes, see also initialiseSyntaxes(3n)
    if (initialiseSyntaxes(NULLCP, attrSyntaxes) == NOTOK)
	error(NULLCP, "unable to initialise syntaxes");

    // attempt connection to the remote agent
    if (agent.connect(argv[1], argv[2]) == NOTOK)
	error(NULLCP, "unable to establish association to %s at %s",
		argv[1], argv[2]);

    // specify the action to perform
    AVA* action = new AVA("getUserNames", NULLCP);

    // perform the non-confirmed Action operation
    if (agent.Action(NULLCP, "uxObjId=test", Sc_BaseObject, 0, NULLCP,
		NULLMACCESS, s_bestEffort, action, NULLCMISOBJECTLISTREF,
		NULLRMIBMANAGER, True) == OK)
	fprintf(stdout, "non-confirmed Action request sent\n");

    delete action;

    // disconnect from the remote agent
    agent.disconnect();
}
