//
// Copyright (c) 1991 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// TestMgr.cc - realisation

// By Thurain Tin, June 1994


#ifdef __GNUG__
#include <std.h>
#elif __cplusplus
#include <stdlib.h>
#endif

#include "msap.h"
#include "Error.h"
#include "RMIBAgent.h"
#include "TestMgr.h"


int TestMgr::wakeUp (char* token)
{
    // what to do at the wake up
    return (strcmp(token, "do_operation")) ? NOTOK : do_operation();
}


int TestMgr::readCommEndpoint (int)
{
    return OK;
}


int TestMgr::eventNotification (char* objClass, char* objInst,
	char* eventType, char* eventTime, Attr* report, RMIBAgent* agent)
{
    fprintf(stdout, "\n\nEvent Report - %s from %s-%s",
	eventType, agent->getAgentName(), agent->getHostName());

    char* cp;
    fprintf(stdout, "\nclass: %s\ninst: {%s}\ntime: %s\nreport: %s\n",
	objClass, objInst, eventTime, cp = report->print());

    if (cp) free(cp);
    return OK;
}


int TestMgr::generalResult (CMISObjectList* objList, RMIBAgent* agent)
{
    // general result callback for operations which
    // are applied to a number of managed objects
    // i.e. Get, Set, Action, Delete
    if (!objList)
	return NOTOK;

    fprintf(stdout, "\ngeneralResult callback from %s-%s",
	agent->getAgentName(), agent->getHostName());

    switch(objList->getOperation())
    {
    case M_GET:
	objList->print();
	delete objList;
	return OK;

    default:
	objList->print();
	delete objList;
	return NOTOK;
    }
}


int TestMgr::singleGetResult (CMISObject* obj, RMIBAgent* agent)
{
    // receive Get results one by one without any assembly
    // to enable this callback change the Get in "do_operation"
    // below to include the boolean value True as the last parameter
    fprintf(stdout, "\nsingleGetResult callback from %s-%s",
	agent->getAgentName(), agent->getHostName());

    obj->print();
    delete obj;
    return OK;
}


int TestMgr::createResult (CMISObject* obj, RMIBAgent* agent)
{
    // the Create result callback
    fprintf(stdout, "\ncreateResult callback from %s-%s",
	agent->getAgentName(), agent->getHostName());

    obj->print();
    delete obj;
    return OK;
}


int TestMgr::noTimeAssertion (RMIBAgent* agent)
{
    // what to do if the remote agent does not support the
    // time parameter in event reporting, and there exist a
    // number of event registrations with time discrimination
    return agent->cancelTimeAssertion(this);
}


int TestMgr::brokenAssociation (RMIBAgent* agent)
{
    // action to take if the remote agent terminates/aborts
    cancelWakeUps("do_operation");
    exit(NOTOK);
    return OK;
}


int TestMgr::do_operation ()
{
    CMISObjectList* list;

    // perform a Get operation on the {uxObjId=test} instance
    return _agent->Get(NULLCP, "uxObjId=test", Sc_BaseObject, 0,
	"(wiseSaying := *Hello*)", NULLMACCESS, s_bestEffort,
	NULL, list, this /*, True */);
}


TestMgr::TestMgr (RMIBAgent* agent)
{
    // initialise the given remote agent
    _agent = agent;

    // register with the Coordinator for waking up
    scheduleWakeUps(30, "do_operation");

    // tell the agent to send all event reports that
    // evaluate True to the given filter expression
    _agent->receiveEvent(NULLCP, NULLCP, NULLCP,
	"((eventTime>=940101000000Z) & (eventTime<=941231115959Z))", this);

    // tell the agent to notify if the association gets broken
    _agent->notifyBrokenAssociation(this);    
}


TestMgr::~TestMgr ()
{
    // cancel the wakeup registration
    cancelWakeUps("do_operation");

    // cancel every callbacks registered with the agent
    _agent->cancelCallbacks(this);
}
