#
# Copyright (c) 1995 University College London
# All rights reserved.
#
# Redistribution and use in source and binary forms are permitted
# provided that the above copyright notice and this paragraph are
# duplicated in all such forms and that any documentation,
# advertising materials, and other materials related to such
# distribution and use acknowledge that the software was developed
# by the Department of Computer Science, University College London.
# The name of the University may not be used to
# endorse or promote products derived from this software without
# specific prior written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# 

# result.tcl

# By Thurain Tin, February 1995


#
# mrproc -
#   An example procedure to process the synchronous
#   and asynchronous managed object results list.
#

proc mrproc {molist} {
    puts "\n************************************************************\n"
    set objId [lindex $molist 0]
    puts "object id: \t\t$objId"
    set replId [lindex $molist 1]
    puts "reply id: \t\t$replId"
    set appl [lindex $molist 2]
    puts "application: \t\t$appl"
    set host [lindex $molist 3]
    puts "host name: \t\t$host"
    set operT [lindex $molist 4]
    puts "operation: \t\t$operT"
    set fErr [lindex $molist 5]
    puts "first error: \t\t$fErr"
    set numbMO [expr [llength $molist]-6]
    puts "number of MOs: \t\t$numbMO"
    if {$numbMO > 0} {
        for {set i 0} {$i < $numbMO} {incr i 1} {
            puts "\n"
            set mo [lindex $molist [expr $i+6]]
            set class [lindex $mo 1]
            puts "      class: \t\t\t$class"
            set inst [lindex $mo 2]
            puts "      instance: \t\t{$inst}"
            set time [lindex $mo 3]
            puts "      operation time: \t\t$time"
            set error [lindex $mo 0]
            if {$error == "noError" || $error == "getListError" ||
                    $error == "setListError"} {
                if {$operT == "M_GET" || $operT == "M_SET" ||
                        $operT == "M_CREATE" || $operT == "M_ACTION"} {
                    set numbAva [expr [llength $mo]-4]
                    if {$numbAva > 0} {
                        for {set j 0} {$j < $numbAva} {incr j 1} {
                            set ava [lindex $mo [expr $j+4]]
                            set type [lindex $ava 0]
                            set value [lindex $ava 1]
                            # no attribute level error checking
                            puts "      $type: \t\t$value"
                        }
                    }
                }
            } elseif {$error == "noSuchAttribute" ||
                      $error == "missingAttributeValue" ||
                      $error == "invalidAttributeValue" ||
                      $error == "processingFailure"} {
                set errInfo [lindex $mo 4]
                set type [lindex $errInfo 0]
                if {$error == "noSuchAttribute" ||
                          $error == "missingAttributeValue"} {
                    puts "      $error: \t\t$type"
                } else {
                    set info [lindex $errInfo 1]
                    puts "      $type: \t\t$info"
                }
            } else {
                # error is self-explanatory!
                puts "      error code: \t\t$error" 
            }
        }
    }
    puts "\n************************************************************\n"
}
