//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// For pretty-printing logs created by EventRecord

// Saleem N. Bhatti
// June 1992

#if !defined(_ReadLog_cc_)
#define _ReadLog_cc_

#if defined(__GNUG__)
#include <std.h>
#elif defined(__cplusplus)
#include <stdlib.h>
#endif
#include "ucbext.h"

#include "msap.h"
#include "Error.h"
#include "dbm-lock.h"
#include "Syntaxes.h"

/*
** A rather inelegant solution to a cross dependncy between libdsap.a and
** libosisec.a. The fix is actually to bracket libosisec.a between libdsap.a
** on your link line but I can see peoples puzzled expressions as they see
** libdsap.a twice on the link line (three if you use LT!), so ...
*/
#if defined(SECURITY)
extern "C" {
void certificate_pair_syntax();
}

VoidFunc readLogSyntaxes[] = {
#ifdef SECURITY
    certificate_pair_syntax,
#endif
    smi_syntaxes,
#ifdef  MONMET
    monitor_syntaxes,
#endif
#ifdef  IQA
    iqa_attr_syntaxes,
#endif
    isode_syntaxes,
#ifdef  OIM
    addOimAttributeSyntax,
    addOimReportSyntax,
#endif
    NULL
};
#else
#define readLogSyntaxes attrSyntaxes
#endif

main(int argc, char **argv)
{
    if (argc != 2)
        error((char *) 0, "usage : readlog <dbm-log-file>");

    if (initialiseSyntaxes((char *) 0, readLogSyntaxes) != OK)
        error((char *) 0, "Cannot intialise syntaxes");

    register
    DbmFile *db;

    if ((db = dbmOpen(argv[1])) == (DbmFile *) 0)
        error((char *) 0, "Could not open %s", argv[1]);

    register
    datum *key, *entry;
    register
    int n;

    for(n = 0, key = dbmGetKey(db, DBM_GetFirstKey); key != (datum *) 0;
	key = dbmGetKey(db), ++n) {

        if ((entry = dbmFetch(db, key)) == (datum *) 0)
            continue;

        char data[DBM_MaxRecordSize + 1], name[80], value[4096];

        bcopy(entry->dptr, data, entry->dsize);
        data[entry->dsize] = '\0';

        register
        char *cp0 = data, *cp1;
        register
        int i, nAttrs;

        printf("\n");
        sscanf(cp0, "%[^:]", value); // MO class

        OID oid, event = (OID) 0;

        oid = name2oid(value);

        if (oid == (OID) 0) {
            printf("Could not find OID for %s\n", value);
            printf("Possibly bad file/record format\n");
            break;
        }
        else oid_free(oid);

        printf("Managed Object class:%s\n", value);

        cp1 = strchr(cp0, ':'); cp0 = cp1 + 1;
        sscanf(cp0, "%[^:]", value); // DN
        DN dn = str2dn(value);

        if (dn == (DN) 0) {
            dn_free(dn);
            printf("Could not get DN from %s\n", value);
            printf("Possibly bad file/record format\n");
            break;
        }
        dn_free(dn);
        printf("Managed Object instance:{%s}\n", value);
        printf("\n");

        cp1 = strchr(cp0, ':'); cp0 = cp1 + 1;
        sscanf(cp0, "%[^:]", value); // nAttrs
        nAttrs = atoi(value);

        for(i = 0; i < nAttrs; ++i) {

            cp1 = strchr(cp0, ':'); cp0 = cp1 + 1;
            if (sscanf(cp0, "%[^=]=%[^:]", name, value) != 2) {
                printf("Bad file format\n");
                continue;
            }

            printf("%s:", name);

            if (value[0] == '-') { // no PE value was written
                printf("<No value stored>\n");
                continue;
            }

            oid = name2oid(name);
            if (oid == (OID) 0) {
                printf("<Could not find OID>\n");
                continue;
            }

            char memory[2048], *vcp0, *vcp1;
            register
            int l;
            int v;

            /* Covert hex-byte to binary */
            for(l = 0, vcp0 = value, vcp1 = memory; *vcp0 != '\0';
                ++l, vcp0 += 2, ++vcp1) {
                sscanf(vcp0, "%02x", &v);
                *vcp1 = (char) v;
            }

            PE pe;
            if (mem2pe(&pe, memory, l) == NOTOK || pe == (PE) 0) {
                printf("<Could not convert value>\n");
                if (oid != (OID) 0)
                    oid_free(oid);
                continue;
            }

            /*
            ** Hard wired knowledge about the EventRecord class
            ** for the additionalText attribute.
            */
            if (strcmp(name, "eventType") == 0) {
                MID mid = EventType_decode(pe);
                event = oid_cpy(mid->mid_global);
                mid_free(mid);
            }

            /* Convert the hexdumped BER back to a PE */
            if (strcmp(name, "additionalText") == 0) {

                char *type = (char *) 0, *report = (char *) 0;

                if (ava2str(oid, pe, &type, &report) == NOTOK ||
                    pe == (PE) 0) {
                    printf("<Could not pretty-print additionalText value>\n");
                    if (oid != (OID) 0)
                        oid_free(oid);
                    pe_free(pe);
                    continue;
                }

                /* Covert hex-byte to binary */
                for(l = 0, vcp0 = report, vcp1 = memory; *vcp0 != '\0';
                    ++l, vcp0 += 2, ++vcp1) {

                    if (*vcp0 == '\\' || *vcp0 == '\n') {
                        --l;
                        --vcp0;
                        --vcp1;
                    } else {
                        sscanf(vcp0, "%02x", &v);
                        *vcp1 = (char) (v & 0xff);
                    }
                }

                free(type);
                free(report);
                pe_free(pe);

                if (mem2pe(&pe, memory, l) == NOTOK || pe == (PE) 0) {
                    printf("<Could not convert additionalText value>\n");
                    if (oid != (OID) 0)
                        oid_free(oid);
                    if (event != (OID) 0) {
                        oid_free(event);
                        event = (OID) 0;
                    }
                    continue;
                }

                if (oid != (OID) 0)
                    oid_free(oid);
                oid = event;
            }

            if (ava2str(oid, pe, &vcp0, &vcp1) == NOTOK) {
                printf("<Could not pretty-print value>\n");
                if (oid != (OID) 0)
                    oid_free(oid);
                if (event != (OID) 0) {
                    oid_free(oid);
                    event = (OID) 0;
                }
                pe_free(pe);
            }
            else {
                printf("%s\n", vcp1);
                free(vcp0);
                free(vcp1);
            }
        }

        fprintf(stderr, "\n--- Press RETURN ---\n");
        while(getchar() != '\n') ;

        if (event != (OID) 0) {
            oid_free(event);
            event = (OID) 0;
        }
    }

    dbmClose(db);
    printf("%d entries in %s\n", n, argv[1]);
}

#endif /* !_ReadLog_cc_ */
