/*
 * Copyright (c) 1994 Technical Research Centre of Finland
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* acserver2.c - static server dispatch for xom/xmp */

/*
 * By Matti Hallivuori, December 1994
 */

#ifndef	lint
static char *rcsid = "$Header: acserver2.c,v 1.3 94/12/19 09:02:03 mha Exp $";
#endif


/* LINTLIBRARY */
#include "msap.h"
#include "mparm.h"
#include "xmsap.h"
#include <signal.h>
extern IamAgent;
extern IamSun;
extern int sesCnt;
extern struct PSAPaddr * gmyAddr;
extern AEI gmyTitle;
int fd_defined = 0;
static	int	is_nfds;
long int oldSigMask = 0;
extern xom_head xom_heads[NR_HEADS];

int  xmplose (td, reason, ap) 
register struct TSAPdisconnect *td;
int     reason;
char * ap;
{
  if (td) {
     bzero ((char *) td, sizeof *td);
     td -> td_reason = reason;
  }
  fprintf(stderr,"xmplose: %s\n", ap);
  return NOTOK;
}


typedef struct TSAPdisconnect * TSAPd; 

int iserver_init (argc, argv, aei, initfnx, td)
int argc;
char **argv;
AEI aei;
INIT_FN	initfnx;
TSAPd td;  /* typedef struct TSAPdisconnect * TSAPd; in isoaddrs.h */
{
  struct PSAPaddr *pa = NULLPA;
  IamAgent = 1;
#if defined(sun)
  IamSun = 1;
#endif    
  fprintf(stderr, "iserver_init\n");
  if (argc <= 1 && (pa = aei2addr (aei)) == NULLPA)
    return xmplose (td, DR_ADDRESS, "address translation failed");
  if (argc > 1) {
    return xmplose (td, DR_NETWORK, "dynamic responder not supported");
  }
  if (!oldSigMask) {
    oldSigMask = sigblock(sigMask);
    sigsetmask(oldSigMask);
  }
  return iserver_initAux (pa,aei,td);
}

#if defined(sun)
/* Current (8.0) revision of sun xmp stack do not provide the fd. The fd
 * is assumed to be the last pipe opened by this program. The same pipe
 * is used in this version for both incoming and outgoing trafic.
 */

#if defined(SOLARIS)
#include "sys/stat.h"
#endif

int last_fd_in_use()
{
struct stat buf;
int i,fd,time=0;
  if (fd_defined) {
    fd = xom_heads[0].fd;
  } else {
    fd_defined = 1;
    for (i=0;i<20;i++) {
      if ((!fstat(i,&buf)) && (buf.st_mode & S_IFIFO) &&
         (buf.st_ctim.tv_sec > time)) {
         time = buf.st_ctim.tv_sec;
	 fd = i; 
       }
    }
    xom_heads[0].fd = fd;
  }
  /*printf("chosen fd = %d\n", fd);*/
  return fd;
}
#else

int xom_filedesc(session)
OM_private_object session;
{
  OM_return_code    om_return;
  OM_value_position total_number;
  OM_public_object  file_descriptor_object;
  static OM_type    included_types[] = {MP_FILE_DESCRIPTOR,OM_NO_MORE_TYPES};
  int fd;

   om_return = om_get(session,
                     OM_EXCLUDE_ALL_BUT_THESE_TYPES, included_types,
                     OM_FALSE, OM_ALL_VALUES,OM_ALL_VALUES,
                     &file_descriptor_object, &total_number);
  if ((om_return != OM_SUCCESS) || (total_number <1)) {
    return NOTOK;
  }
  fd = file_descriptor_object[0].value.integer;
  (void) om_delete(file_descriptor_object);
  return fd;
}
#endif

int	iserver_initAux (pa, aei, td)
struct PSAPaddr *pa;
AEI    aei;
struct TSAPdisconnect *td;
{
  int	fd;
  is_nfds = 0;
/* open the xmp session and bind it   */
  if (TNetListenXMP(pa, aei, td) == NOTOK)
    return NOTOK;
  fd = xom_filedesc(xom_heads[0].session);
  if (fd == NOTOK) {
    return NOTOK;
  }
  is_nfds = fd;
  xom_heads[0].new_session=1;
  return OK;
}

/*  */

int iserver_wait (initfnx, workfnx, losefnx, nfds, rfds, wfds, efds, secs, td)
INIT_FN	initfnx;  /* IFP */
WORK_FN workfnx;
LOSE_FN	losefnx;
int	nfds;
fd_set *rfds,  /* wfds and efds are dummies */
       *wfds,
       *efds;
int	secs;
TSAPd  td;
{
  int  fd, n, vecp, msd,ret;
  fd_set  ifds;
  char   *vec[4];
  ifds = (*rfds);

  if (workfnx == NULLIFP) {
    return  xmplose(td,DR_OPERATION,"no workfunction in iserver_wait");
  }

/* map osimis msd to fd */
  if (!xom_heads[0].session) {
    TNetListenXMP(gmyAddr,gmyTitle,td);
  }
  for (n=0;n <= sesCnt;n++) {
    fd = xom_heads[n].fd;
    FD_CLR(n,&ifds); /* mask std in,  out, stderr  etc */
    FD_SET(fd,&ifds);
  }
  if (is_nfds > nfds) {
    nfds = is_nfds;
  }
  if ((n =TNetAccept_xmp(nfds,&ifds, secs, td))  == NOTOK) {
    return NOTOK;
  }
  if ((1 == xom_heads[0].new_session) && (n || 
      ( FD_ISSET (xom_heads[0].fd, &ifds)))) {
    xom_heads[0].new_session = 0;
    vecp = 1;
    vec[0] = (char *) &"xom/xmp listen";
    if ((msd = (*initfnx) (vecp, vec)) != NOTOK) ;
    /* The new session is given firts time slot here to make sure
     * the simulated (create_ind) session is used once */
    (*workfnx) (msd);
    FD_CLR (xom_heads[msd].fd, &ifds);
    goto exit;
  }

  /* All agent sessions currently devide the same fd. This may change
   * therefore the code below is planned to handle several fds.
   * 0 no sessions, wait information
   * 1 first session
   * 2 second session  etc
   */

  for (msd = 1; (msd<=sesCnt) && (n>0); msd++) {
    fd = xom_heads[msd].fd;
    if (FD_ISSET (fd, &ifds)) {
      n--;
      FD_CLR (xom_heads[msd].fd, &ifds);
      if ((*workfnx) (msd) == NOTOK) {
        ret =  xmplose(td,DR_OPERATION,"routine for fd is not ok");
	(void) TNetCloseXMP (msd, td);
      }
    }
  }
exit:
  if (rfds)
    *rfds = ifds;
  return OK;
}














