/*
 * Copyright (c) 1988 University College London and VTT
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this parasraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Management data structures freeing routines */


/*
 * By George Pavlou, April 1990
 * changes by Matti Hallivuori, December 1994
 */
#include <stdio.h>
#include "msap.h"


/* LINTLIBRARY */

/* NOTICE, that in many cases the life time memory management frees all
 * memory as one group. Therefore individual freeing is only needed for
 * those rare cases the memory is not allected from life time memory pool.
 */
void  mc_free (mc)
MSAPConnect * mc;
{
}


void  ms_free (ms)
MSAPStart * ms;
{
}

void  mn_free (mn)
MN mn;
{
    if ((mn -> mn_type == MN_DN || mn -> mn_type == MN_LOCALDN)
		&& mn -> mn_dn) {
	dn_free(mn -> mn_dn);
	mn -> mn_dn = NULLDN;
    }
    else
    if (mn -> mn_type == MN_NONSPEC) {
	free(mn -> mn_nonspec);
	mn -> mn_nonspec = NULLCP;
    }
}


void  mparm_free (mparm)
CMISParam* mparm;
{
}


void  ava_free (pava)
Ava* pava;
/* pava and NOT ava as the latter is a struct in <isode/quipu/name.h> */
{
    if (!pava)
	return;
    if (pava->ava_oid)
	oid_free(pava->ava_oid);
    if (pava->ava_value)
	pe_free(pava->ava_value);
}


static void  mfitem_free (item)
CMISFilterItem * item;
{
    switch (item -> fi_type) {
    case FI_Equality:
    case FI_GreaterOrEqual:
    case FI_LessOrEqual:
    case FI_SubsetOf:
    case FI_SupersetOf:
    case FI_NonNullSetIntersect:
	mparm_free(&item -> fi_attr);
	break;

    case FI_Present:
	mid_free(&item -> fi_attr.mp_id);
	break;

    case FI_Substrings:
    {
	register CMISFilterSubstrings *substring = item->fi_substrings, *cur;
	for (; substring;) {
	    substring = (cur = substring) -> fi_sub_next;
	    mid_free(&cur -> fi_sub_attrid);
	    if (cur -> fi_sub_string)
		pe_free(cur -> fi_sub_string);
	    free ((char *) cur);
	}
    }
	break;

    default:
	break;
    }
    free ((char *) item);
}


static void  mfitemdata_free (item)
CMISFilterItem * item;
{
    switch (item -> fi_type) {
    case FI_Equality:
    case FI_GreaterOrEqual:
    case FI_LessOrEqual:
    case FI_SubsetOf:
    case FI_SupersetOf:
    case FI_NonNullSetIntersect:
	mparm_free(&item -> fi_attr); /* NOTE possible memory leak */
	break;

    case FI_Present:
	mid_free(&item -> fi_attr.mp_id);
	break;

    case FI_Substrings:
    {
	register CMISFilterSubstrings *cur;
	for (cur = item->fi_substrings; cur; cur = cur -> fi_sub_next) {
	    mid_free(&cur -> fi_sub_attrid);
	    if (cur -> fi_sub_string)
		pe_free(cur -> fi_sub_string);
	}
    }
	break;

    default:
	break;
    }
}


void  mfilter_free (filter)
CMISFilter * filter;
{
    if (filter == NULLMFILTER)
	return;
    switch (filter -> flt_type) {
    case Filter_Item:
	mfitem_free (filter -> flt_item);
	break;

    case Filter_And:
    case Filter_Or:
    {
	register CMISFilter *cur, *next;
	for (cur = filter -> flt_filter; cur; cur = next) {
	    next = cur -> flt_next;
	    mfilter_free(cur);
	}
    }
	break;

    case Filter_Not:
	mfilter_free(filter -> flt_filter);
	break;

    default:
	break;
    }
    free((char*) filter);
}


void  mfilterdata_free (filter)
CMISFilter * filter;
{
    if (filter == NULLMFILTER)
	return;
    switch (filter -> flt_type) {
    case Filter_Item:
	mfitemdata_free (filter -> flt_item);
	break;

    case Filter_And:
    case Filter_Or:
    {
	register CMISFilter *cur;
	for (cur = filter -> flt_filter; cur; cur = cur -> flt_next)
	    mfilterdata_free(cur);
    }
	break;

    case Filter_Not:
	mfilterdata_free(filter -> flt_filter);
	break;

    default:
	break;
    }
}






