/*
 * Copyright (c) 1988 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* oid.c - various OID related functions */

/*
 * By George Pavlou, June 1991
 */

typedef struct OIDentifier {
    int     oid_nelem;          /* number of sub-identifiers */

    unsigned int *oid_elements; /* the (ordered) list of sub-identifiers */
}       OIDentifier;
typedef OIDentifier *OID;
#define NULLOID ((OID) 0)

/* LINTLIBRARY */

static int  elem_cmp_loc (p, q, nelem)
register unsigned int *p, *q;
register int nelem;
{
    while (--nelem > -1)
	if (p[nelem] != q[nelem])
	    return 0;
    return 1;
}

/* oid_equal: compares two OIDs for equality only (more efficient than oid_cmp)
 *            returns 1 (TRUE) if equal, 0 (FALSE) otherwise
 */

int  oid_equal (p, q)
register OID  p, q;
{
    if (p == NULLOID || q == NULLOID) {
	if (p == NULLOID && q == NULLOID)
	    return 1;
	return 0;
    }

    if (p->oid_nelem != q->oid_nelem)
	return 0;

    return elem_cmp_loc (p->oid_elements,q->oid_elements, p->oid_nelem);
}





