/*
 * Copyright (c) 1990 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* */


/* Saleem N. Bhatti */

/* */
/*****************************************************************************
Initialising Syntaxes



Saleem N. Bhatti
July 1992
*****************************************************************************/

#include <msap.h>
/*#include <isode/quipu/name.h>*/

#define OSIMISETCPATH    "OSIMISETCPATH"
#define OSIMIS_name      "osimis"
#define DEFAULT_tailor   "tailor"
#define DEFAULT_oidtable "oidtable"

static char *envName= OSIMISETCPATH;
static char *myName= OSIMIS_name;
char *etcdir; /* The path of the etc directory */
char *tailor; /* The full pathname of the tailor */
extern char *isodetcpath; /* This is in the ISODE library */
char globalNameDomain[256];	/* hopefully enough */


void  initialiseGlobalNameDomain ()
{
    FILE *fp;
    char buffer[256];
    DN dn;
    int found = 0;

    if ((fp = fopen(tailor, "r")) == (FILE *) 0)
        error((char *) 0, "initialiseGlobalNameDomain: could not fopen() \
tailor %s", tailor);

    while(fgets(buffer, (sizeof(buffer) - 1), fp) && (*buffer != EOF)) {
        if (!found && (sscanf(buffer, "globalNameDomain : %255s", globalNameDomain) == 1))
            found = 1;
        if (!found && (sscanf(buffer, "gmsGlobalNameDomain : %255s", globalNameDomain) == 1))
            found = 1;
    }
    fclose(fp);

    if (!found)
	error(NULLCP, "initialiseGlobalNameDomain: Could not read \
globalNameDomain from %s", tailor);
    else
	advise(NULLCP,"globalNameDomain: %s", globalNameDomain);

    /* check if globalNameDomain is a valid DN */
    if ((dn = str2dn(globalNameDomain)) == NULLDN)
         error(NULLCP, "initialiseGlobalNameDomain: cannot form DN for \
globalNameDomain '%s'", globalNameDomain);
    dn_free(dn);
}

int setEnvVarName(str)
char *str;
{
  if((str) && (strlen(str))) {
    envName = strdup(str);
    return OK;
  } else {
    envName = strdup(OSIMISETCPATH);
    return NOTOK;
  }
}

int setMyName(str)
char *str;
{
  if((str) && (strlen(str))) {
    myName = strdup(str);
    return OK;
  } else {
    myName = strdup(OSIMIS_name);
    return NOTOK;
  }
}


int
#if defined(__STDC__) || defined(__cplusplus)
initialiseSyntaxes(const char *tailorFile, VoidFunc *syntaxes)
#else
initialiseSyntaxes(tailorFile, syntaxes)
char     *tailorFile;
VoidFunc *syntaxes;
#endif
{
    static
    int done = 0;

    register
    FILE *fp;
    register
    char *cp, *tailorName;
    VoidFunc *pfv;
    char buffer[256]; 

    if (done) return(OK);

    /* Tailor file */
    cp = getenv(envName);
    if (cp == NULL) {
        fprintf(stderr, "Error: No value for the environment variable %s\n",
        envName);
        return(NOTOK);
    }
    sprintf(buffer, "%s/", cp);

    /* The following is used to direct ISODE to the OSIMIS etc directory */
    isodetcpath = etcdir = strdup(buffer);

    /* tailoring */
    tailorName = (char *) tailor;
    if ((tailorName == NULL) || (*tailorName == '\0'))
	sprintf(buffer, "%s%s%s", etcdir, myName, DEFAULT_tailor);
    else
	sprintf(buffer, "%s%s", etcdir, tailorName);


    /* check if the tailor file exists */
    if ((fp = fopen(buffer, "r")) != NULL) {
	close(fp);
	isodesetailor(tailorName);
        tailor = strdup(buffer);
	fprintf(stderr, "tailor : %s\n", tailor);
    }
    else {
	fprintf(stderr, "Warning: Cannot open tailor file : %s\n", buffer);
        tailor = NULL;
    }

    /* Syntax info */
    sprintf(buffer, "%s%s", etcdir, DEFAULT_oidtable);
    fprintf(stderr, "oidtable : %s\n", buffer);

    for(pfv = syntaxes; *pfv != NULL; ++pfv) (*pfv)();

    load_oid_table(buffer);

    done = 1;

    initialiseGlobalNameDomain();

    return(OK);
}






