//
// Copyright (c) 1991 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// CMISService.h - provides a virtual CMIS (MSAP) API

// By George Pavlou / Thurain Tin, July 1994


#ifndef _CMIS_SERVICE_
#define _CMIS_SERVICE_


#ifndef _MSAP_
#include "msap.h"
#endif

#ifndef _ISODE_AGENT_
#include "IsodeAgent.h"
#endif


#undef M_EventRep;
#undef M_EventRepConf;
#undef M_Set;
#undef M_SetConf;
#undef M_Action;
#undef M_ActionConf;
#undef M_WaitReq;


// CMISService

class CMISService
{
    static CMISService* _instance;

public:

    virtual int M_EventRep (int, int, MID, MN, MID, char*,
		    PE, MSAPIndication*);

    virtual int M_EventRepConf (int, int, MID, MN, MID, char*,
		    PE, MSAPIndication*);

    virtual int M_EventRepRes (int, int, MID, MN, char*, CMISParam*,
		    CMISErrors, CMISErrorInfo*, MSAPIndication*);

    virtual int M_Get (int, int, MID, MN, CMISScope*, CMISFilter*,
		    External*, CMISSync, int, MID, MSAPIndication*);

    virtual int M_GetRes (int, int, int, MID, MN, char*, int, CMISGetAttr*,
		    CMISErrors, CMISErrorInfo*, MSAPIndication*);

    virtual int M_Set (int, int, MID, MN, CMISScope*, CMISFilter*,
		    External*, CMISSync, int, CMISSetAttr*, MSAPIndication*);

    virtual int M_SetConf (int, int, MID, MN, CMISScope*, CMISFilter*,
		    External*, CMISSync, int, CMISSetAttr*, MSAPIndication*);

    virtual int M_SetRes (int, int, int, MID, MN, char*, int, CMISSetAttr*,
		    CMISErrors, CMISErrorInfo*, MSAPIndication*);

    virtual int M_Action (int, int, MID, MN, CMISScope*, CMISFilter*,
		    External*, CMISSync, MID, PE, MSAPIndication*);

    virtual int M_ActionConf (int, int, MID, MN, CMISScope*, CMISFilter*,
		    External*, CMISSync, MID, PE, MSAPIndication*);

    virtual int M_ActionRes (int, int, int, MID, MN, char*, CMISParam*,
		    CMISErrors, CMISErrorInfo*, MSAPIndication*);

    virtual int M_Create (int, int, MID, MN, int, MN, External*,
		    int, CMISParam*, MSAPIndication*);

    virtual int M_CreateRes (int, int, MID, MN, char*, int, CMISParam*,
		    CMISErrors, CMISErrorInfo*, MSAPIndication*);

    virtual int M_Delete (int, int, MID, MN, CMISScope*, CMISFilter*,
		    External*, CMISSync, MSAPIndication*);

    virtual int M_DeleteRes (int, int, int, MID, MN, char*,
		    CMISErrors, CMISErrorInfo*, MSAPIndication*);

    virtual int M_CancelGet (int, int, int, MSAPIndication*);

    virtual int M_CancelGetRes (int, int, CMISErrors,
		    CMISErrorInfo*, MSAPIndication*);

    virtual int M_WaitReqAux (int, int, MSAPIndication*, PE*);

    virtual int M_WaitReq (int, int, MSAPIndication*);

    virtual int M_InitialiseReq (AEI, struct PSAPaddr*, AEI,
		    struct PSAPaddr*, char*, int, int, External*,
		    External*, MSAPConnect*, MSAPIndication*);

    virtual int M_Init (int, char**, MSAPStart*, MSAPIndication*);

    virtual int M_InitialiseResp (MSAPStart*, int, AEI, struct PSAPaddr*,
		    char*, int, int, External*, External*, MSAPIndication*);

    virtual int M_TerminateReq (int, int, PE, MSAPIndication*);

    virtual int M_TerminateResp (int, int, int, PE, MSAPIndication*);

    virtual int M_AbortReq (int, External*, MSAPIndication*);

    virtual int m_initialise (char*, char*, MInitOption**);

    virtual int m_terminate (int);

    static CMISService* getInstance ();
    virtual int registerServiceUser (AEI, ISODEAgent*, Bool = False);
    virtual int deregisterServiceUser (AEI, ISODEAgent*);

    CMISService ();
};


// CMISService inline methods

inline int CMISService::M_EventRep (int msd, int invoke, MID obj_class,
	MN obj_inst, MID event_type, char* event_time, PE event_info,
	MSAPIndication* mi)
{
    return ::M_EventRepAux(msd, M_EVENT_REP, invoke, obj_class, obj_inst,
		event_type, event_time, event_info, mi);
}

inline int CMISService::M_EventRepConf (int msd, int invoke, MID obj_class,
	MN obj_inst, MID event_type, char* event_time, PE event_info,
	MSAPIndication* mi)
{
    return ::M_EventRepAux(msd, M_EVENT_REPC, invoke, obj_class, obj_inst,
		event_type, event_time, event_info, mi);
}

inline int CMISService::M_EventRepRes (int msd, int invoke, MID obj_class,
	MN obj_inst, char* cur_time, CMISParam* event_reply, CMISErrors error,
	CMISErrorInfo* error_info, MSAPIndication* mi)
{
    return ::M_EventRepRes(msd, invoke, obj_class, obj_inst, cur_time,
		event_reply, error, error_info, mi);
}

inline int CMISService::M_Get (int msd, int invoke, MID obj_class,
	MN obj_inst, CMISScope* scope, CMISFilter* filter, External* access,
	CMISSync sync, int nattrs, MID attrs, MSAPIndication* mi)
{
    return ::M_Get(msd, invoke, obj_class, obj_inst, scope, filter, access,
		sync, nattrs, attrs, mi);
}

inline int CMISService::M_GetRes (int msd, int invoke, int linked,
	MID obj_class, MN obj_inst, char* cur_time, int nattrs,
	CMISGetAttr* attrs, CMISErrors error, CMISErrorInfo* error_info,
	MSAPIndication* mi)
{
    return ::M_GetRes(msd, invoke, linked, obj_class, obj_inst, cur_time,
		nattrs, attrs, error, error_info, mi);
}

inline int CMISService::M_Set (int msd, int invoke, MID obj_class,
	MN obj_inst, CMISScope* scope, CMISFilter* filter, External* access,
	CMISSync sync, int nattrs, CMISSetAttr* attrs, MSAPIndication* mi)
{
    return ::M_SetAux(msd, M_SET, invoke, obj_class, obj_inst, scope,
		filter, access, sync, nattrs, attrs, mi);
}

inline int CMISService::M_SetConf (int msd, int invoke, MID obj_class,
	MN obj_inst, CMISScope* scope, CMISFilter* filter, External* access,
	CMISSync sync, int nattrs, CMISSetAttr* attrs, MSAPIndication* mi)
{
    return ::M_SetAux(msd, M_SETC, invoke, obj_class, obj_inst, scope,
		filter, access, sync, nattrs, attrs, mi);
}

inline int CMISService::M_SetRes (int msd, int invoke, int linked,
	MID obj_class, MN obj_inst, char* cur_time, int nattrs,
	CMISSetAttr* attrs, CMISErrors error, CMISErrorInfo* error_info,
	MSAPIndication* mi)
{
    return ::M_SetRes(msd, invoke, linked, obj_class, obj_inst, cur_time,
		nattrs, attrs, error, error_info, mi);
}

inline int CMISService::M_Action (int msd, int invoke, MID obj_class,
	MN obj_inst, CMISScope* scope, CMISFilter* filter, External* access,
	CMISSync sync, MID action_type, PE action_info, MSAPIndication* mi)
{
    return ::M_ActionAux(msd, M_ACTION, invoke, obj_class, obj_inst, scope,
		filter, access, sync, action_type, action_info, mi);
}

inline int CMISService::M_ActionConf (int msd, int invoke, MID obj_class,
	MN obj_inst, CMISScope* scope, CMISFilter* filter, External* access,
	CMISSync sync, MID action_type, PE action_info, MSAPIndication* mi)
{
    return ::M_ActionAux(msd, M_ACTIONC, invoke, obj_class, obj_inst, scope,
		filter, access, sync, action_type, action_info, mi);
}

inline int CMISService::M_ActionRes (int msd, int invoke, int linked,
	MID obj_class, MN obj_inst, char* cur_time, CMISParam* action_reply,
	CMISErrors error, CMISErrorInfo* error_info, MSAPIndication* mi)
{
    return ::M_ActionRes(msd, invoke, linked, obj_class, obj_inst,
		cur_time, action_reply, error, error_info, mi);
}

inline int CMISService::M_Create (int msd, int invoke, MID obj_class,
	MN obj_inst, int instance_type, MN reference_inst, External* access,
	int nattrs, CMISParam* attrs, MSAPIndication* mi)
{
    return ::M_Create(msd, invoke, obj_class, obj_inst, instance_type,
		reference_inst, access, nattrs, attrs, mi);
}

inline int CMISService::M_CreateRes (int msd, int invoke, MID obj_class,
	MN obj_inst, char* cur_time, int nattrs, CMISParam* attrs,
	CMISErrors error, CMISErrorInfo* error_info, MSAPIndication* mi)
{
    return ::M_CreateRes(msd, invoke, obj_class, obj_inst, cur_time,
		nattrs, attrs, error, error_info, mi);
}

inline int CMISService::M_Delete (int msd, int invoke, MID obj_class,
	MN obj_inst, CMISScope* scope, CMISFilter* filter, External* access,
	CMISSync sync, MSAPIndication* mi)
{
    return ::M_Delete(msd, invoke, obj_class, obj_inst, scope, filter,
		access, sync, mi);
}

inline int CMISService::M_DeleteRes (int msd, int invoke, int linked,
	MID obj_class, MN obj_inst, char* cur_time, CMISErrors error,
	CMISErrorInfo* error_info, MSAPIndication* mi)
{
    return ::M_DeleteRes(msd, invoke, linked, obj_class, obj_inst,
		cur_time, error, error_info, mi);
}

inline int CMISService::M_CancelGet (int msd, int invoke, int get_invoke,
	MSAPIndication* mi)
{
    return ::M_CancelGet(msd, invoke, get_invoke, mi);
}

inline int CMISService::M_CancelGetRes (int msd, int invoke,
	CMISErrors error, CMISErrorInfo* error_info, MSAPIndication* mi)
{
    return ::M_CancelGetRes(msd, invoke, error, error_info, mi);
}

inline int CMISService::M_WaitReqAux (int msd, int secs,
	MSAPIndication* mi, PE* pdu)
{
    return ::M_WaitReqAux(msd, secs, mi, pdu);
}

inline int CMISService::M_WaitReq (int msd, int secs, MSAPIndication* mi)
{
    return ::M_WaitReqAux(msd, secs, mi, (PE*) 0);
}

inline int CMISService::M_InitialiseReq (AEI callingTitle,
	struct PSAPaddr* callingAddr, AEI calledTitle,
	struct PSAPaddr* calledAddr, char* context, int protvrsn, int fununits,
	External* access, External* info, MSAPConnect* mc, MSAPIndication* mi)
{
    return ::M_InitialiseReq(callingTitle, callingAddr, calledTitle,
		calledAddr, context, protvrsn, fununits, access, info, mc, mi);
}

inline int CMISService::M_Init (int vecp, char** vec,
	MSAPStart* ms, MSAPIndication* mi)
{
    return ::M_Init(vecp, vec, ms, mi);
}

inline int CMISService::M_InitialiseResp (MSAPStart* ms, int status,
	AEI respTitle, struct PSAPaddr* respAddr, char* context, int protvrsn,
	int fununits, External* access, External* info, MSAPIndication* mi)
{
    return ::M_InitialiseResp(ms, status, respTitle, respAddr, context,
		protvrsn, fununits, access, info, mi);
}

inline int CMISService::M_TerminateReq (int msd, int reason,
	PE info, MSAPIndication* mi)
{
    return ::M_TerminateReq(msd, reason, info, mi);
}

inline int CMISService::M_TerminateResp (int msd, int status,
	int reason, PE info, MSAPIndication* mi)
{
    return ::M_TerminateResp(msd, status, reason, info, mi);
}

inline int CMISService::M_AbortReq (int msd, External* info,
	MSAPIndication* mi)
{
    return ::M_AbortReq(msd, info, mi);
}

inline int CMISService::m_initialise (char* agent, char* rhost,
	MInitOption** options)
{
    return ::m_initialise(agent, rhost, options);
}

inline int CMISService::m_terminate (int msd)
{
    return ::m_terminate(msd);
}

inline CMISService* CMISService::getInstance ()
{ return _instance; }

inline int CMISService::registerServiceUser (AEI, ISODEAgent*, Bool)
{ return OK; }

inline int CMISService::deregisterServiceUser (AEI, ISODEAgent*)
{ return OK; }

inline CMISService::CMISService ()
{ _instance = this; }


#endif /* _CMIS_SERVICE_ */
