/*
 * Copyright (c) 1988 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* attribute value related routines */

/*
 * By George Pavlou / Saleem N. Bhatti, April 1990
 */

#include <stdio.h>
#include "msap.h"

int  empty_asn1_type;


/* LINTLIBRARY */

/* returns a string representation of an attribute value, using
 * the print function for that syntax as loaded in the oid_table
 */


char* attrv2str (attrv, print_fnx)
caddr_t attrv;
IFP	print_fnx;
{
    PS ps;
    char* buf;

    if (((ps = ps_alloc(str_open)) == NULLPS) ||
		(str_setup(ps, NULLCP, BUFSIZ, 0) == NOTOK))
	return NULLCP;

    print_fnx(ps, attrv, EDBOUT);
    *ps->ps_ptr = NULL;
    buf = ps->ps_base;
    ps->ps_base = NULLCP;	/* not to be free'd */
    ps_free(ps);
    return buf;
}


/*
 * different version of the above routine using a static PS for efficiency
 * it does not work though for recursive calls as needed by mfilter_print
 *

static PS ps = NULLPS;

char* attrv2str (attrv, print_fnx)
caddr_t attrv;
IFP	print_fnx;
{
    char* buf;

    if (ps == NULL &&
		((ps = ps_alloc(str_open)) == NULLPS) ||
		(str_setup(ps, NULLCP, BUFSIZ, 0) == NOTOK)) {
	if (ps) {
	    ps_free (ps); ps = NULLPS;
	}
	return NULLCP;
    }

    print_fnx(ps, attrv, EDBOUT);
    *ps->ps_ptr = NULL;

    buf = ps->ps_base;

    ps -> ps_base = NULL; ps -> ps_cnt = 0;
    ps -> ps_ptr = NULL; ps -> ps_bufsiz = 0;

    return buf;
}
 */


/* turns an attribute type (oid) and value (pe) to strings
 * using the oidtable registered syntax
 */

int  ava2str (attr_type, attr_val, str_type, str_val)
OID    attr_type;
PE     attr_val;
char** str_type;
char** str_val;
{
    oid_table_attr* at;
    sntx_table* table; 
    caddr_t value;

    if ((at = oid2attr(attr_type)) == NULL) {
	*str_type = strdup(sprintoid(attr_type));
	goto failed;
    }

    if (str_type)
	*str_type = strdup(at->oa_ot.ot_name);
    if (!str_val)
	return OK;
    if (!attr_val) {
	*str_val = strdup("");
	return OK;
    }

    table = get_syntax_table(at->oa_syntax);
    if (table -> s_decode == NULL ||
		table -> s_print == NULL ||
		table -> s_free == NULL ||
		(value = (caddr_t) table->s_decode(attr_val)) == NULL &&
		attr_val->pe_errno != PE_ERR_NONE)
	goto failed;

    *str_val = attrv2str(value, table->s_print);
    table->s_free(value);

    if (*str_val == NULLCP)
	goto failed;

    return OK;

failed:
    *str_val = strdup("");
    return NOTOK;
}


Ava *
#if defined(__STDC__) || defined(__cplusplus)
str2ava(const char *name, const char *value)
#else
str2ava(name, value)
char *name;
char *value;
#endif
{
    register
    oid_table_attr *type;
    register
    sntx_table *syntax;
    register
    void *v;
    OID oid;
    PE pe;
    Ava *attr;

    if ((oid = name2oid((char *) name)) == NULL) {
#if defined(DEBUG)
	printf("str2ava() : unknown name \"%s\"\n", name);
#endif
	return(NULL);
    }

    if (!value || strlen(value) == 0) {
	pe = NULLPE;
	goto ready;
    }

    if ((type = oid2attr(oid)) == NULL) {
#if defined(DEBUG)
	printf("str2ava() : unknown attribute \"%s\"\n", name);
#endif
	oid_free(oid);
	return(NULL);
    }

    if ((syntax = get_syntax_table(type->oa_syntax)) == NULL) {
#if defined(DEBUG)
	printf("str2ava() : unknown syntax for \"%s\"\n", name);
#endif
	oid_free(oid);
	return(NULL);
    }

    if (syntax->s_parse == NULL) {
#if defined(DEBUG)
	printf("str2ava() : no parser for \"%s\"\n", name);
#endif
	oid_free(oid);
	return(NULL);
    }

    if (syntax->s_free == NULL) {
#if defined(DEBUG)
	printf("str2ava() : no freer \"%s\"\n", name);
#endif
	oid_free(oid);
	return(NULL);
    }

    if (syntax->s_encode == NULL) {
#if defined(DEBUG)
	printf("str2ava() : no encoder for \"%s\"\n", name);
#endif
	oid_free(oid);
	return(NULL);
    }

    empty_asn1_type = 0;
    if ((v = (void *) syntax->s_parse((char *) value)) == NULL &&
		!empty_asn1_type) {
#if defined(DEBUG)
	printf("str2ava() : parse failed for - \"%s\", \"%s\"\n", name, value);
#endif
	oid_free(oid);
	return(NULL);
    }

    if ((pe = (PE) syntax->s_encode(v)) == NULL) {
#if defined(DEBUG)
	printf("str2ava() : encode failed for - \"%s\", \"%s\"\n", name, value);
#endif
	syntax->s_free(v);
	oid_free(oid);
	return(NULL);
    }

    syntax->s_free(v);

ready:
    attr = (Ava *) malloc(sizeof(Ava));
    attr->ava_oid = oid;
    attr->ava_value = pe;

    return(attr);
}


int  massage_actionOID_aux (oid, type)
OID  oid;
int  type;
{
    oid->oid_elements = (unsigned int*) realloc(
		(char*) oid->oid_elements,
		(oid->oid_nelem+1) * sizeof(unsigned int) );
    switch (type) {
    case ACTION_INFO:
	oid->oid_elements[oid->oid_nelem] = 1;
	break;
    case ACTION_REPLY:
	oid->oid_elements[oid->oid_nelem] = 2;
	break;
    default:
	return NOTOK;
    }
    ++oid->oid_nelem;
    return OK;
}


int  ava2str_aux (attr_type, attr_val, str_type, str_val, type)
OID    attr_type;
PE     attr_val;
char** str_type;
char** str_val;
int type;
{
    if (!attr_val)					/* no fiddling to do */
	return ava2str(attr_type, attr_val, str_type, str_val);

    massage_actionOID_aux(attr_type, type);

    if (ava2str(attr_type, attr_val, str_type, str_val) == NOTOK) {
	--attr_type->oid_nelem;
	return NOTOK;
    }
    --attr_type->oid_nelem;
    return OK;
}


Ava*  str2ava_aux (name, value, type)
char *name;
char *value;
int type;
{
    char buf[128];
    Ava* a;	/* for some strange reason cc complains if this is ava */

    if (!value)
	return str2ava(name, value);	/* no fiddling to do... */

    switch (type) {
    case ACTION_INFO:
	sprintf(buf, "%sInfo", name);
	break;
    case ACTION_REPLY:
	sprintf(buf, "%sReply", name);
	break;
    default:
	return NULL;
    }

    if (a = str2ava(buf, value))
	--a->ava_oid->oid_nelem;	/* remove the last element */
    return a;
}

