/*
 * Copyright (c) 1988 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this parasraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Management parameter checking routines */

/*
 * By George Pavlou, April 1990
 */

#include <stdio.h>
#include "msap.h"


/* LINTLIBRARY */

int  check_scope (scope)
CMISScope* scope;
{
    switch (scope -> sc_type) {
    case Sc_BaseObject:
    case Sc_FirstLevel:
    case Sc_WholeSubtree:
	return OK;

    case Sc_IndividualLevel:
    case Sc_BaseToNthLevel:
	if (scope -> sc_level < 0)
	    return NOTOK;
	return OK;

    default:
	return NOTOK;
    }
}


int  check_filter (filter)
CMISFilter* filter;
{
    switch (filter->flt_type) {
    case Filter_Item:
	if (!filter->flt_item)
	    return NOTOK;

	if (filter->flt_item->fi_type == FI_Substrings) {
	    /*
	     * do not allow for different attribute ids in the substring list
	     *  and allow for only one initial and one final substring
	     */
	    register CMISFilterSubstrings* substr =
					filter->flt_item->fi_substrings;
	    OID firstAttrId;
	    int initial = 0, final = 0;

	    if (!substr || substr->fi_sub_attrid.mid_type != MID_GLOBAL)
		return NOTOK;

	    firstAttrId = substr->fi_sub_attrid.mid_global;

	    switch (substr->fi_sub_type) {
	    case FI_Substring_Initial:
		++initial;
		break;
	    case FI_Substring_Final:
		++final;
		break;
	    }

	    for (substr = substr->fi_sub_next; substr;
					substr = substr->fi_sub_next) {
		if (substr->fi_sub_attrid.mid_type != MID_GLOBAL ||
		    !oid_equal(substr->fi_sub_attrid.mid_global, firstAttrId))
		    return NOTOK;

		switch (substr->fi_sub_type) {
		case FI_Substring_Initial:
		    if (++initial > 1)
			return NOTOK;
		    break;
		case FI_Substring_Final:
		    if (++final > 1)
			return NOTOK;
		    break;
	        }
	    }
	}
	else
	if (filter->flt_item->fi_attr.mp_id.mid_type != MID_GLOBAL)
	    return NOTOK;
	return OK;

    case Filter_Not:
	if (!filter -> flt_filter)
	    return NOTOK;
	return check_filter(filter -> flt_filter);

    case Filter_And:
    case Filter_Or:
    {
	register CMISFilter* cur;
	for (cur = filter -> flt_filter; cur; cur = cur -> flt_next)
	    if (check_filter(cur) == NOTOK)
		return NOTOK;
    }
	return OK;

    default:
	return NOTOK;
    }    
}

