/*
 * Copyright (c) 1988 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Routines to realise the M-XXX.indication primitives */

/*
 * By George Pavlou, April 1990
 */

/*
** SAVE_PDU() added by Saleem
** March/April 1994
**
** M_WaitReq() changed to M_WaitReqAux() by Saleem
** March/April 1994
** M_WaitReq() is now a macro in msap.h
*/

#include <stdio.h>
#include "msap.h"

#define NULLOP	((struct RoSAPinvoke *) 0)
#define NULLRES	((struct RoSAPresult *) 0)
#define NULLERR	((struct RoSAPerror *) 0)

static int  event_rep_ind(), linked_reply_ind(), get_ind(), set_ind(),
	    action_ind(), create_ind(), delete_ind(),
	    event_rep_res(), get_res(), set_res(), action_res(),
	    create_res(), delete_res(), merror(), parse_fail(),
            cancel_get_ind();
static void check_actual_class();
static void check_noattrs();


#ifndef	SECURITY
#define SAVE_PDU(_pep, _pe) \
{ if ((_pep) != (PE *) 0) { *(_pep) = NULLPE; if (_pe) pe_free(_pe); } }
#else
#define SAVE_PDU(_pep, _pe) \
{ if ((_pep) != (PE *) 0) { *(_pep) = (_pe); (_pe) = (PE) 0; } }
#endif

/* LINTLIBRARY */

int  M_WaitReqAux (msd, secs, mi, pdu)
int msd, secs;
MSAPIndication * mi;
PE *pdu;
{
    struct RoSAPindication rois;
    (void) RoWaitRequest(msd, secs, &rois);

    if (pdu != (PE *) 0)
        *pdu = (PE) 0;

    mi->mi_invoke = 0;
    switch (rois.roi_type) {
    case ROI_INVOKE:
	if (rois.roi_invoke.rox_args == NULLPE)
	    return parse_fail(msd, "OperationArgument missing",
				&rois.roi_invoke, NULLRES, NULLERR, mi);

	if (rois.roi_invoke.rox_nolinked) {
	    if (rois.roi_invoke.rox_op == M_LINKED_REPLY)
	    	return parse_fail(msd, "LinkedReply but no linked id!",
				&rois.roi_invoke, NULLRES, NULLERR, mi);
	}
	else	/* linked remote operation */
	    if (rois.roi_invoke.rox_op != M_LINKED_REPLY)
	    	return parse_fail(msd, "Linked operation other than \
LinkedReply!",
				&rois.roi_invoke, NULLRES, NULLERR, mi);

	(void) bzero((char*) mi, sizeof(*mi));

        mi->mi_invoke = rois.roi_invoke.rox_id;
	switch (rois.roi_invoke.rox_op) {
	case M_EVENT_REP:
	    mi->mi_type = MI_EVENT_REP;
	    return event_rep_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_EVENT_REPC:
	    mi->mi_type = MI_EVENT_REPC;
	    return event_rep_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_LINKED_REPLY:
	    mi->mi_type = MI_LINKED_REPLY;
	    return linked_reply_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_GET:
	    mi->mi_type = MI_GET;
	    return get_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_SET:
	    mi->mi_type = MI_SET;
	    return set_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_SETC:
	    mi->mi_type = MI_SETC;
	    return set_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_ACTION:
	    mi->mi_type = MI_ACTION;
	    return action_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_ACTIONC:
	    mi->mi_type = MI_ACTIONC;
	    return action_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_CREATE:
	    mi->mi_type = MI_CREATE;
	    return create_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_DELETE:
	    mi->mi_type = MI_DELETE;
	    return delete_ind(msd, &rois.roi_invoke, mi, pdu);

	case M_CANCEL_GET:
	    mi->mi_type = MI_CANCEL_GET;
	    return cancel_get_ind(msd, &rois.roi_invoke, mi, pdu);

	default:
	    mi->mi_type = MI_PREJECT;
	    mi->mi_preject.mpr_reason = ROS_IP_UNRECOG;
	    mi->mi_preject.mpr_id = rois.roi_invoke.rox_id;
	    (void) sprintf(mi->mi_preject.mpr_data, "Unknown Operation %d",
				rois.roi_invoke.rox_op);
	    (void) RoURejectRequest(msd, &rois.roi_invoke.rox_id,
				ROS_IP_UNRECOG, ROS_NOPRIO, &rois);
	    ROXFREE(&rois.roi_invoke);
	    return NOTOK;
	}

    case ROI_RESULT:
	(void) bzero((char*) mi, sizeof(*mi));

        mi->mi_invoke = rois.roi_result.ror_id;
	if (rois.roi_result.ror_result == NULLPE) {
	    mi->mi_type = MI_EMPTY_RES;
	    mi->mi_emptyres.id = rois.roi_result.ror_id;
	    return OK;
	}

	switch (rois.roi_result.ror_op) {
	case M_EVENT_REPC:
	    mi->mi_type = MI_EVENT_REP_RES;
	    return event_rep_res(msd, &rois.roi_result, mi, pdu);

	case M_GET:
	    mi->mi_type = MI_GET_RES;
	    return get_res(msd, &rois.roi_result, mi, pdu);

	case M_SETC:
	    mi->mi_type = MI_SET_RES;
	    return set_res(msd, &rois.roi_result, mi, pdu);

	case M_ACTIONC:
	    mi->mi_type = MI_ACTION_RES;
	    return action_res(msd, &rois.roi_result, mi, pdu);

	case M_CREATE:
	    mi->mi_type = MI_CREATE_RES;
	    return create_res(msd, &rois.roi_result, mi, pdu);

	case M_DELETE:
	    mi->mi_type = MI_DELETE_RES;
	    return delete_res(msd, &rois.roi_result, mi, pdu);

	default:
	    mi->mi_type = MI_PREJECT;
	    mi->mi_preject.mpr_reason = ROS_RRP_UNRECOG;
	    mi->mi_preject.mpr_id = rois.roi_result.ror_id;
	    (void) sprintf(mi->mi_preject.mpr_data, "Unknown Result %d",
				rois.roi_result.ror_op);
	    (void) RoURejectRequest(msd, &rois.roi_result.ror_id,
				ROS_RRP_UNRECOG, ROS_NOPRIO, &rois);
	    RORFREE(&rois.roi_result);
	    return NOTOK;
	}

    case ROI_ERROR:
	mi->mi_type = MI_ERROR;
        mi->mi_invoke = rois.roi_error.roe_id;
	rois.roi_error.roe_error += CMIS_ERROR_OFFSET;
	if (rois.roi_error.roe_param == NULLPE &&
		/* error codes with no associated error info */
		rois.roi_error.roe_error != (int) m_accessDenied &&
		rois.roi_error.roe_error != (int) m_mistypedOperation &&
		rois.roi_error.roe_error != (int) m_operationCancelled &&

		/* error codes with optional error info */
		rois.roi_error.roe_error != (int) m_complexityLimitation &&
		rois.roi_error.roe_error != (int) m_processingFailure)
	    return parse_fail(msd, "ErrorArgument", NULLOP, NULLRES,
						&rois.roi_error, mi);
	return merror(msd, &rois.roi_error, mi, pdu);

    case ROI_FINISH:
	mi->mi_type = MI_TERM;
	mi->mi_errtype = m_noError;
	mi->mi_term.mt_reason = rois.roi_finish.acf_reason;
	if (rois.roi_finish.acf_ninfo) {
	    mi->mi_term.mt_info = rois.roi_finish.acf_info[0];
	    rois.roi_finish.acf_info[0] = NULLPE;
	}
	else
	    mi->mi_term.mt_info = NULLPE;
	ACFFREE(&rois.roi_finish);
	return OK;

    case ROI_PREJECT:
	if (rois.roi_preject.rop_reason == ROS_ABORTED) {
	    mi->mi_type = MI_ABORT;
	    mi->mi_abort.ma_source = MA_USER;
	    mi->mi_abort.ma_reason = MA_ABORTED;

	    /* it should parse the CMIPAbortInfo here but ISODE does not
	     * support data parameters for the A-ABORT.INDICATION primitive
	     */
	    mi->mi_abort.ma_info = NULLEXTERN;

	    (void) strncpy(mi->mi_abort.ma_data, "Peer aborted association",
				MA_SIZE);
	    return NOTOK;
	}
	mi->mi_type = MI_PREJECT;
	mi->mi_preject.mpr_reason = rois.roi_preject.rop_reason;
	mi->mi_invoke = mi->mi_preject.mpr_id = 0;
	(void) strncpy(mi->mi_preject.mpr_data, rois.roi_preject.rop_data,
						MPR_SIZE);
	mi->mi_errtype = m_noError;
	return NOTOK;

    case ROI_UREJECT:
	mi->mi_type = MI_UREJECT;
	mi->mi_invoke = mi->mi_ureject.mur_id = rois.roi_ureject.rou_id;

	switch (mi->mi_ureject.mur_reason = rois.roi_ureject.rou_reason) {
	    case ROS_IP_DUP:
		mi->mi_errtype = m_duplicateInvocation;
		break;
	    case ROS_IP_UNRECOG:
		mi->mi_errtype = m_unrecognisedOperation;
		break;
	    case ROS_IP_MISTYPED:
		mi->mi_errtype = m_mistypedArgument;
		break;
	    case ROS_IP_LIMIT:
		mi->mi_errtype = m_resourceLimitation;
		break;
	    default:
		mi->mi_errtype = m_noError;
		break;
	}
	(void) sprintf(mi->mi_preject.mpr_data, "MPDU rejected by remote user");
	return NOTOK;

    default:
	mi->mi_type = MI_PREJECT;
	mi->mi_errtype = m_noError;
	mi->mi_preject.mpr_reason = ROS_IP_UNRECOG;
	mi->mi_invoke = mi->mi_preject.mpr_id = 0;
	(void) sprintf(mi->mi_preject.mpr_data, "Unknown indication %d \
returned by RoWaitReq", rois.roi_type);
	return NOTOK;
    }
}


static int  event_rep_ind (msd, op, mi, pdu)
int msd;
struct RoSAPinvoke* op;
MSAPIndication* mi;
PE *pdu;
{
    CMISEventReportArg* ea;

    if (parse_CMIP_EventReportArgument(op->rox_args, 1, NULLIP, NULLVP, &ea)
		== NOTOK)
	return parse_fail(msd, "EventReportArgument", op, NULLRES, NULLERR, mi);
    mi->mi_eventrepinv.id = op->rox_id;
    mi->mi_eventrepinv.args = *ea;	/* struct cpy */

    SAVE_PDU(pdu, op->rox_args);
    ROXFREE(op);
    free((char*) ea);
    return OK;
}


static int  linked_reply_ind (msd, op, mi, pdu)
int msd;
struct RoSAPinvoke* op;
MSAPIndication* mi;
PE *pdu;
{
    CMISLinkedReplyArg* lr;

    if (parse_CMIP_LinkedReplyArgument(op->rox_args, 1, NULLIP, NULLVP, &lr)
		== NOTOK)
	return parse_fail(msd, "LinkedReplyArgument", op, NULLRES, NULLERR, mi);
    mi->mi_linkedreplyinv.id = op->rox_id;
    mi->mi_linkedreplyinv.args = *lr;	/* struct cpy */
    mi->mi_linkedreplyinv.args.lr_linkid = op->rox_linkid;

    SAVE_PDU(pdu, op->rox_args);
    ROXFREE(op);
    free((char*) lr);
    return OK;
}


static int  get_ind (msd, op, mi, pdu)
int msd;
struct RoSAPinvoke* op;
MSAPIndication* mi;
PE *pdu;
{
    CMISGetArg* ga;
/*
    char buff[1024];
    char* cp = buff; int len; register int i;
    pe2mem(op->rox_args, &cp, &len);
    for (i = 0; i < len; i++)
	printf("%02x", (unsigned int) ((unsigned char) cp[i]));
    printf("\n");
*/

    if (parse_CMIP_GetArgument(op->rox_args, 1, NULLIP, NULLVP, &ga)
		== NOTOK)
	return parse_fail(msd, "GetArgument", op, NULLRES, NULLERR, mi);
    mi->mi_getinv.id = op->rox_id;
    mi->mi_getinv.args = *ga;	/* struct cpy */
    check_actual_class(&mi->mi_getinv.args.ga_class);
    check_noattrs(&mi->mi_getinv.args);

    SAVE_PDU(pdu, op->rox_args);
    ROXFREE(op);
    free((char*) ga);
    return OK;
}


static int  set_ind (msd, op, mi, pdu)
int msd;
struct RoSAPinvoke * op;
MSAPIndication * mi;
PE *pdu;
{
    CMISSetArg* sa;

    if (parse_CMIP_SetArgument(op->rox_args, 1, NULLIP, NULLVP, &sa)
		== NOTOK)
	return parse_fail(msd, "SetArgument", op, NULLRES, NULLERR, mi);
    mi->mi_setinv.id = op->rox_id;
    mi->mi_setinv.args = *sa;	/* struct cpy */
    check_actual_class(&mi->mi_setinv.args.sa_class);

    SAVE_PDU(pdu, op->rox_args);
    ROXFREE(op);
    free((char*) sa);
    return OK;
}


static int  action_ind (msd, op, mi, pdu)
int msd;
struct RoSAPinvoke * op;
MSAPIndication * mi;
PE *pdu;
{
    CMISActionArg* aa;

    if (parse_CMIP_ActionArgument(op->rox_args, 1, NULLIP, NULLVP, &aa)
		== NOTOK)
	return parse_fail(msd, "ActionArgument", op, NULLRES, NULLERR, mi);
    mi->mi_actioninv.id = op->rox_id;
    mi->mi_actioninv.args = *aa;	/* struct cpy */
    check_actual_class(&mi->mi_actioninv.args.aa_class);

    SAVE_PDU(pdu, op->rox_args);
    ROXFREE(op);
    free((char*) aa);
    return OK;
}


static int  create_ind (msd, op, mi, pdu)
int msd;
struct RoSAPinvoke * op;
MSAPIndication * mi;
PE *pdu;
{
    CMISCreateArg* ca;

    if (parse_CMIP_CreateArgument(op->rox_args, 1, NULLIP, NULLVP, &ca)
		== NOTOK)
	return parse_fail(msd, "CreateArgument", op, NULLRES, NULLERR, mi);
    mi->mi_createinv.id = op->rox_id;
    mi->mi_createinv.args = *ca;	/* struct cpy */

    SAVE_PDU(pdu, op->rox_args);
    ROXFREE(op);
    free((char*) ca);
    return OK;
}


static int  delete_ind (msd, op, mi, pdu)
int msd;
struct RoSAPinvoke * op;
MSAPIndication * mi;
PE *pdu;
{
    CMISDeleteArg* da;

    if (parse_CMIP_DeleteArgument(op->rox_args, 1, NULLIP, NULLVP, &da)
		== NOTOK)
	return parse_fail(msd, "DeleteArgument", op, NULLRES, NULLERR, mi);
    mi->mi_deleteinv.id = op->rox_id;
    mi->mi_deleteinv.args = *da;	/* struct cpy */
    check_actual_class(&mi->mi_deleteinv.args.da_class);

    SAVE_PDU(pdu, op->rox_args);
    ROXFREE(op);
    free((char*) da);
    return OK;
}


static int  cancel_get_ind (msd, op, mi, pdu)
int msd;
struct RoSAPinvoke * op;
MSAPIndication * mi;
PE *pdu;
{
    if ((mi->mi_cancelgetinv.args.cga_getinvoke = prim2num(op->rox_args))
		== NOTOK && op->rox_args->pe_errno != PE_ERR_NONE)
	return parse_fail(msd, "CancelGetArgument", op, NULLRES, NULLERR, mi);
    mi->mi_cancelgetinv.id = op->rox_id;
    SAVE_PDU(pdu, op->rox_args);
    ROXFREE(op);
    return OK;
}


static int  event_rep_res (msd, res, mi, pdu)
int msd;
struct RoSAPresult * res;
MSAPIndication * mi;
PE *pdu;
{
    CMISEventReportRes* er;

    if (res->ror_result &&
	parse_CMIP_EventReportResult(res->ror_result, 1, NULLIP, NULLVP, &er)
		== NOTOK)
	return parse_fail(msd, "EventReportResult", NULLOP, res, NULLERR, mi);
    mi->mi_eventrepres.id = res->ror_id;
    mi->mi_eventrepres.result = *er;	/* struct cpy */

    SAVE_PDU(pdu, res->ror_result);
    RORFREE(res);
    free((char*) er);
    return OK;
}


static int  get_res (msd, res, mi, pdu)
int msd;
struct RoSAPresult * res;
MSAPIndication * mi;
PE *pdu;
{
    CMISGetRes* gr;
/*
    char buff[1024];
    char* cp = buff; int len; register int i;
    pe2mem(res->ror_result, &cp, &len);
    for (i = 0; i < len; i++)
	printf("%02x", (unsigned int) ((unsigned char) cp[i]));
    printf("\n");
*/

    if (res->ror_result &&
	parse_CMIP_GetResult(res->ror_result, 1, NULLIP, NULLVP, &gr)
		== NOTOK)
	return parse_fail(msd, "GetResult", NULLOP, res, NULLERR, mi);
    mi->mi_getres.id = res->ror_id;
    mi->mi_getres.result = *gr;	/* struct cpy */

    SAVE_PDU(pdu, res->ror_result);
    RORFREE(res);
    free((char*) gr);
    return OK;
}


static int  set_res (msd, res, mi, pdu)
int msd;
struct RoSAPresult * res;
MSAPIndication * mi;
PE *pdu;
{
    CMISSetRes* sr;

    if (res->ror_result &&
	parse_CMIP_SetResult(res->ror_result, 1, NULLIP, NULLVP, &sr)
		== NOTOK)
	return parse_fail(msd, "SetResult", NULLOP, res, NULLERR, mi);
    mi->mi_setres.id = res->ror_id;
    mi->mi_setres.result = *sr;	/* struct cpy */

    SAVE_PDU(pdu, res->ror_result);
    RORFREE(res);
    free((char*) sr);
    return OK;
}


static int  action_res (msd, res, mi, pdu)
int msd;
struct RoSAPresult * res;
MSAPIndication * mi;
PE *pdu;
{
    CMISActionRes* ar;

    if (res->ror_result &&
	parse_CMIP_ActionResult(res->ror_result, 1, NULLIP, NULLVP, &ar)
		== NOTOK)
	return parse_fail(msd, "ActionResult", NULLOP, res, NULLERR, mi);
    mi->mi_actionres.id = res->ror_id;
    mi->mi_actionres.result = *ar;	/* struct cpy */

    SAVE_PDU(pdu, res->ror_result);
    RORFREE(res);
    free((char*) ar);
    return OK;
}


static int  create_res (msd, res, mi, pdu)
int msd;
struct RoSAPresult * res;
MSAPIndication * mi;
PE *pdu;
{
    CMISCreateRes* cr;

    if (res->ror_result &&
	parse_CMIP_CreateResult(res->ror_result, 1, NULLIP, NULLVP, &cr)
		== NOTOK)
	return parse_fail(msd, "CreateResult", NULLOP, res, NULLERR, mi);
    mi->mi_createres.id = res->ror_id;
    mi->mi_createres.result = *cr;	/* struct cpy */

    SAVE_PDU(pdu, res->ror_result);
    RORFREE(res);
    free((char*) cr);
    return OK;
}


static int  delete_res (msd, res, mi, pdu)
int msd;
struct RoSAPresult * res;
MSAPIndication * mi;
PE *pdu;
{
    CMISDeleteRes* dr;

    if (res->ror_result &&
	parse_CMIP_DeleteResult(res->ror_result, 1, NULLIP, NULLVP, &dr)
		== NOTOK)
	return parse_fail(msd, "DeleteResult", NULLOP, res, NULLERR, mi);
    mi->mi_deleteres.id = res->ror_id;
    mi->mi_deleteres.result = *dr;	/* struct cpy */

    SAVE_PDU(pdu, res->ror_result);
    RORFREE(res);
    free((char*) dr);
    return OK;
}


static int  merror (msd, err, mi, pdu)
int msd;
struct RoSAPerror * err;
struct MSAPIndication * mi;
PE *pdu;
{
    struct RoSAPindication rois;
    caddr_t errinfo = NULL;

    (void) bzero((char *) &mi->mi_error.info, sizeof(mi->mi_error.info));

    switch (mi->mi_errtype = (CMISErrors) (err->roe_error)) {
    case m_noSuchObjectClass:
	if (parse_CMIP_ObjectClass(err->roe_param, 1, NULLIP, NULLVP,
			(MID*) &errinfo) == NOTOK)
	    return parse_fail(msd, "ObjectClass", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_noSuchObjectClass = *(MID) errinfo;
							/* struct cpy*/
	break;

    case m_noSuchObjectInstance:
    case m_duplicateManagedObjectInstance:
    case m_noSuchReferenceObject:
    case m_invalidObjectInstance:
	if (parse_CMIP_ObjectInstance(err->roe_param, 1, NULLIP, NULLVP,
			(MN*) &errinfo) == NOTOK)
	    return parse_fail(msd, "ObjectInstance", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_noSuchObjectInstance = *(MN) errinfo;
							/* struct cpy*/
	break;

    case m_accessDenied:
    case m_mistypedOperation:
    case m_operationCancelled:
	mi->mi_error.id = err->roe_id;
	break;

    case m_syncNotSupported:
	if (parse_CMIP_CMISSync(err->roe_param, 1, NULLIP, NULLVP,
		(struct type_CMIP_CMISSync**) &errinfo) == NOTOK)
	    return parse_fail(msd, "CMISSync", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_syncNotSupported = *(CMISSync*) errinfo;
	break;

    case m_invalidFilter:
	if (parse_CMIP_CMISFilter(err->roe_param, 1, NULLIP, NULLVP,
			&mi->mi_error.info.ei_invalidFilter) == NOTOK)
	    return parse_fail(msd, "CMISFilter", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	break;

    case m_noSuchAttribute:
	if (parse_CMIP_AttributeId(err->roe_param, 1, NULLIP, NULLVP,
			(MID*) &errinfo) == NOTOK)
	    return parse_fail(msd, "AttributeId", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_noSuchAttribute = *(MID) errinfo;
							/* struct cpy*/
	break;

    case m_invalidAttributeValue:
	if (parse_CMIP_Attribute(err->roe_param, 1, NULLIP, NULLVP,
			(CMISParam**) &errinfo) == NOTOK)
	    return parse_fail(msd, "Attribute", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_invalidAttributeValue = *(CMISParam*) errinfo;
							/* struct cpy*/
	break;

    case m_getListError:
	if (parse_CMIP_GetListError(err->roe_param, 1, NULLIP, NULLVP,
			(CMISGetRes**) &errinfo) == NOTOK)
	    return parse_fail(msd, "GetListError", NULLOP, NULLRES, err, mi);
	mi->mi_getres.id = err->roe_id;
	mi->mi_getres.result = *(CMISGetRes*) errinfo;	/* struct cpy*/
	break;

    case m_setListError:
	if (parse_CMIP_SetListError(err->roe_param, 1, NULLIP, NULLVP,
			(CMISSetRes**) &errinfo) == NOTOK)
	    return parse_fail(msd, "SetListError", NULLOP, NULLRES, err, mi);
	mi->mi_setres.id = err->roe_id;
	mi->mi_setres.result = *(CMISSetRes*) errinfo;	/* struct cpy*/
	break;

    case m_noSuchAction:
	if (parse_CMIP_NoSuchAction(err->roe_param, 1, NULLIP, NULLVP,
			(CMISNoSuchArgument**) &errinfo) == NOTOK)
	    return parse_fail(msd, "NoSuchAction", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_noSuchAction = *(CMISNoSuchArgument*) errinfo;
							/* struct cpy*/
	break;

    case m_processingFailure:
	if (err->roe_param &&
	    parse_CMIP_ProcessingFailure(err->roe_param, 1, NULLIP, NULLVP,
			(CMISProcessingFailure**) &errinfo) == NOTOK)
	    return parse_fail(msd, "ProcessingFailure",NULLOP,NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_processingFailure =
			*(CMISProcessingFailure*) errinfo;	/* struct cpy*/
	break;

    case m_noSuchEventType:
	if (parse_CMIP_NoSuchEventType(err->roe_param, 1, NULLIP, NULLVP,
			(CMISNoSuchArgument**) &errinfo) == NOTOK)
	    return parse_fail(msd, "NoSuchEventType", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_noSuchEventType = *(CMISNoSuchArgument*) errinfo;
							/* struct cpy*/
	break;

    case m_noSuchArgument:
	if (parse_CMIP_NoSuchArgument(err->roe_param, 1, NULLIP, NULLVP,
			(CMISNoSuchArgument**) &errinfo) == NOTOK)
	    return parse_fail(msd, "NoSuchArgument", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_noSuchArgument = *(CMISNoSuchArgument*) errinfo;
							/* struct cpy*/
	break;

    case m_invalidArgumentValue:
	if (parse_CMIP_InvalidArgumentValue(err->roe_param, 1, NULLIP, NULLVP,
			(CMISInvalidArgumentValue**) &errinfo) == NOTOK)
	    return parse_fail(msd,"InvalidArgumentValue",NULLOP,NULLRES,err,mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_invalidArgumentValue =
			*(CMISInvalidArgumentValue*) errinfo;	/* struct cpy*/
	break;

    case m_invalidScope:
	if (parse_CMIP_Scope(err->roe_param, 1, NULLIP, NULLVP,
			(CMISScope**) &errinfo) == NOTOK)
	    return parse_fail(msd, "Scope", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_invalidScope = *(CMISScope*) errinfo;
							/* struct cpy*/
	break;

    case m_missingAttributeValue:
	if (parse_CMIP_AttributeIdList(err->roe_param, 1, NULLIP, NULLVP,
			&mi->mi_error.info.ei_missingAttributeValue) == NOTOK)
	    return parse_fail(msd, "AttributeIdList", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	break;

    case m_classInstanceConflict:
	if (parse_CMIP_BaseManagedObjectId(err->roe_param, 1, NULLIP, NULLVP,
			(CMISBaseManagedObjectId**) &errinfo) == NOTOK)
	    return parse_fail(msd,"BaseManagedObjectId",NULLOP,NULLRES,err,mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_classInstanceConflict =
			*(CMISBaseManagedObjectId*) errinfo;	/* struct cpy*/
	break;
	
    case m_complexityLimitation:
	if (err->roe_param &&
	    parse_CMIP_ComplexityLimitation(err->roe_param, 1, NULLIP, NULLVP,
			(CMISComplexityLimitation**) &errinfo) == NOTOK)
	    return parse_fail(msd,"ComplexityLimitation",NULLOP,NULLRES,err,mi);
	mi->mi_error.id = err->roe_id;
	mi->mi_error.info.ei_complexityLimitation =
			*(CMISComplexityLimitation*) errinfo;	/* struct cpy*/
	break;
	
    case m_noSuchInvokeId:
	if (prim2num(err->roe_param) == NOTOK &&
			err->roe_param->pe_errno != PE_ERR_NONE)
	    return parse_fail(msd, "NoSuchInvokeId", NULLOP, NULLRES, err, mi);
	mi->mi_error.id = err->roe_id;
	break;

    default:
	mi->mi_type = MI_PREJECT;
	mi->mi_preject.mpr_reason = ROS_REP_RECERR;
	mi->mi_preject.mpr_id = err->roe_id;
	(void) sprintf(mi->mi_preject.mpr_data, "Unknown Error %d",
        	                        err->roe_error);
	(void) RoURejectRequest(msd, &err->roe_id,
				ROS_REP_RECERR, ROS_NOPRIO, &rois);
	return NOTOK;
    }

    SAVE_PDU(pdu, err->roe_param);
    ROEFREE(err);
    if (errinfo)
	free(errinfo);		/* not the contents */
    return OK;
}


static int  parse_fail (msd, msg, op, res, err, mi)
int msd;
char * msg;
struct RoSAPinvoke * op;
struct RoSAPresult * res;
struct RoSAPerror * err;
MSAPIndication * mi;
{
    int *invoke;
    struct RoSAPindication rois;

    mi->mi_type = MI_PREJECT;
    (void) sprintf(mi->mi_preject.mpr_data,
			"%.*s: Badly formed APDU received", MPR_SIZE-35, msg);
    if (op) {
	mi->mi_preject.mpr_reason = ROS_IP_MISTYPED;
	mi->mi_preject.mpr_id = op->rox_id;
	invoke = &op->rox_id;
	ROXFREE(op)
    }
    else
    if (res) {
	mi->mi_preject.mpr_reason = ROS_RRP_MISTYPED;
	mi->mi_preject.mpr_id = res->ror_id;
	invoke = &res->ror_id;
	RORFREE(res)
    }
    else
    if (err) {
	mi->mi_preject.mpr_reason = ROS_REP_MISTYPED;
	mi->mi_preject.mpr_id = err->roe_id;
	invoke = &err->roe_id;
	ROEFREE(err);
    }
    (void) RoURejectRequest(msd, invoke, mi->mi_preject.mpr_reason,
							ROS_NOPRIO, &rois);
    return NOTOK;
}


static void  check_actual_class (obj_class)
MID  obj_class;
{
    if (obj_class->mid_type != MID_GLOBAL ||
		!oid_equal(obj_class->mid_global, ACTUALCLASS_ASN_OID))
	return;
    oid_free(obj_class->mid_global);
    obj_class->mid_global = NULLOID;
}


static void  check_noattrs (get_arg)
CMISGetArg*  get_arg;
{
    if (get_arg->ga_nattrs == 1 &&
		get_arg->ga_attrs[0].mid_type == MID_LOCAL &&
		get_arg->ga_attrs[0].mid_local == -1) {
	free((char*) get_arg->ga_attrs);
	get_arg->ga_nattrs = 0;
    }
    return;
}
