/*
 * Copyright (c) 1988 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * moper.h - auxilliary routines and integrity checks for
 *	     management operations
 */

/*
 * By George Pavlou, April 1990
 */

/*
** Integrity check stuff added by Saleem
** March/April 1994
*/

#ifndef _MOPER_
#define _MOPER_


#define missing_midP(p)		(!(p) || !(p)->mid_global)
#define missing_mnP(p)		(!(p) || !(p)->mn_dn)
#define null_midP(p)		(!(p)->mid_global)
#define null_mnP(p)		(!(p)->mn_dn)
#define null_mparmP(p)		(!(p)->mp_id.mid_global || !(p)->mp_val)


int    build_fail(), ros_fail(), arg_fail(), error_fail(), M_Error();
char*  op2str();


#if defined(SECURITY)
#include "mintegrity.h"

FILE *G_PRINT_CHECK_file = stderr; /* Set to 0 to disable check messages */

/*
** Use macros to avoid the overhead of another function call.
*/

#define PRINT_CHECK(_v, _i, _c, _m) \
    if (G_PRINT_CHECK_file != (FILE *) 0) \
        fprintf(G_PRINT_CHECK_file, "%s: invokeId=%08x id=%08x chk=%08x\n", \
                _m, (u_int) (_v), (u_int) (_i), (u_int) (_c))

#define INTEGRITY_CHECK(_msd, _id, _pdu, _msapId, _msg) \
{ \
    MD5Value md5_; \
    register int i_; \
    PE pe_; \
    if ((pe_ = _pdu) == (PE) 0) /* Empty Result or no info */ \
        pe_ = int2prim(_id); \
    if (makeMd5Value(_msd, pe_, &md5_) != 0) { \
        fprintf(stderr, "makeMd5Value(): failed - %s\n", _msg); \
        pe_free(pe_); \
        return NOTOK; \
    } \
    i_ = ENCODE_ID(_id, md5_.md5_words[0]); \
    PRINT_CHECK(i_, _id, md5_.md5_words[0], _msg); \
    _msapId = _id = i_; \
    if (_pdu == (PE) 0) \
        pe_free(pe_); \
}
#else
#define INTEGRITY_CHECK(_msd, _id, _pdu, _msapId, _msg) { _msapId = _id; }
#endif


#endif
