/*
 * Copyright (c) 1988 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* moperagent.c - perform agent management operations */

/*
 * By George Pavlou, April 1990
 */

/*
** Integrity check stuff added by Saleem
** March/April 1994
*/

#include <stdio.h>
#include "msap.h"
#include "moper.h"

static int    M_LinkedReplyGet(), M_LinkedReplySet(),
	      M_LinkedReplyAction(), M_LinkedReplyDelete();


/* LINTLIBRARY */

int  M_EventRepAux (msd, op, invoke, obj_class, obj_inst,
			event_type, event_time, event_info, mi)
int  msd, op, invoke;
MID  obj_class, event_type;
MN   obj_inst;
char* event_time;
PE   event_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISEventReportArg parm;
    PE pe;

    mi->mi_invoke = 0;
    if (missing_midP(obj_class))
	return arg_fail(invoke, op, 0, "Object Class", mi);
    if (missing_mnP(obj_inst))
	return arg_fail(invoke, op, 0, "Object Instance", mi);
    if (missing_midP(event_type))
	return arg_fail(invoke, op, 0, "Event Type", mi);

    parm.ea_class = *obj_class;		/* struct copy */
    parm.ea_inst  = *obj_inst;		/* struct copy */
    parm.ea_type  = *event_type;	/* struct copy */
    parm.ea_time  = event_time;
    parm.ea_info  = event_info;

    if (build_CMIP_EventReportArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Event Report Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Event Report Argument");

    if (RoInvokeRequest(msd, op, ROS_ASYNC, pe, invoke,
				NULLIP, ROS_NOPRIO, &rois) == NOTOK) {
	pe_free(pe);
	return ros_fail(&rois, mi);
    }

    pe_free(pe);
    return OK;
}


int  M_GetRes (msd, invoke, linked, obj_class, obj_inst, cur_time,
			nattrs, attrs, error, error_info, mi)
int  msd, invoke, linked;
MID  obj_class;
MN   obj_inst;
char* cur_time;
int  nattrs;
CMISGetAttr* attrs;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISGetRes parm;
    PE pe = NULLPE;
    int i, result = 0;

    mi->mi_invoke = 0;
    if (linked)
	return M_LinkedReplyGet(msd, invoke, linked, obj_class, obj_inst,
				cur_time, nattrs, attrs, error, error_info, mi);

    switch (error) {
    case m_noError:
    case m_getListError:
	break;
    case m_duplicateInvocation:
    case m_unrecognisedOperation:
    case m_mistypedArgument:
    case m_resourceLimitation:
    case m_noSuchObjectClass:
    case m_noSuchObjectInstance:
    case m_accessDenied:
    case m_invalidScope:
    case m_invalidFilter:
    case m_syncNotSupported:
    case m_classInstanceConflict:
    case m_complexityLimitation:
    case m_operationCancelled:
    case m_processingFailure:
	return M_Error(msd, M_GET, invoke, error, error_info, mi);
    default:
	return error_fail(invoke, M_GET, error, mi);
    }

    for (i = 0; i < nattrs; i++) {
	if (null_midP(&attrs[i].ga_id))
	    return arg_fail(invoke, M_GET, 1, "Attribute Id", mi);
	if (attrs[i].ga_val == NULLPE && !attrs[i].ga_error)
	    return arg_fail(invoke, M_GET, 1, "Attribute Value", mi);
    }

    if (obj_class) {
	parm.gr_class = *obj_class;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.gr_class, sizeof(parm.gr_class));

    if (obj_inst) {
	parm.gr_inst  = *obj_inst;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.gr_inst, sizeof(parm.gr_inst));

    if (parm.gr_time = cur_time)
	result = 1;

    if (parm.gr_nattrs = nattrs)
	result = 1;
    parm.gr_attrs = attrs;

    if (result)
	if (!error) {
	    if (build_CMIP_GetResult(&pe, 1, 0, NULLCP, &parm) == NOTOK)
		return build_fail(invoke, "Get Result", mi);
	}
	else
	    if (build_CMIP_GetListError(&pe, 1, 0, NULLCP, &parm) == NOTOK)
		return build_fail(invoke, "GetListError", mi);

    INTEGRITY_CHECK(msd, invoke, pe, invoke, error ? "GetListError" :
                    "Get Result");

    result = error == m_noError ?
	RoResultRequest(msd, invoke, M_GET, pe, ROS_NOPRIO, &rois) :
	RoErrorRequest(msd, invoke, (int) error - CMIS_ERROR_OFFSET, pe,
				ROS_NOPRIO, &rois);
/*
    {
    char buff[1024];
    char* cp = buff; int len;

    pe2mem(pe, &cp, &len);
    for (i = 0; i < len; i++)
	printf("%02x", (unsigned int) ((unsigned char) cp[i]));
    printf("\n");
    }
*/
    if (pe)
	pe_free(pe);

    if (result == NOTOK)
	return ros_fail(&rois, mi);
    return OK;
}


static int  M_LinkedReplyGet (msd, invoke, linked, obj_class, obj_inst,
				cur_time, nattrs, attrs, error, error_info, mi)
int  msd, invoke, linked;
MID  obj_class;
MN   obj_inst;
char* cur_time;
int  nattrs;
CMISGetAttr* attrs;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    struct CMISLinkedReplyArg parm;
    PE pe;
    int i;

    mi->mi_invoke = 0;
    switch (error) {
    case m_noError:
    case m_getListError:
	if (missing_midP(obj_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Class", mi);
	if (missing_mnP(obj_inst))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Instance", mi);
	parm.lr_getres.gr_nattrs = nattrs;

	for (i = 0; i < nattrs; i++) {
	    if (null_midP(&attrs[i].ga_id))
		return arg_fail(invoke, M_LINKED_REPLY, 0, "Attribute Id", mi);
	    if (attrs[i].ga_val == NULLPE && !attrs[i].ga_error)
		return arg_fail(invoke, M_LINKED_REPLY, 0, "Attribute value", mi);
	}

	parm.lr_getres.gr_class = *obj_class;	/* struct copy */
	parm.lr_getres.gr_inst  = *obj_inst;	/* struct copy */
	parm.lr_getres.gr_time  = cur_time;
	parm.lr_getres.gr_attrs = attrs;
        parm.lr_type		= error ? LR_GET_LISTERR : LR_GET_RES;
	break;
		
    case m_processingFailure:
	if (!error_info)
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - ErrorInfo", mi);
	if (null_midP(&error_info->ei_processingFailure.pf_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - Object Class", mi);
	if (null_mparmP(&error_info->ei_processingFailure.pf_error))
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - SpecificErrorInfo", mi);
	bzero((char*) &parm.lr_procfailure.pf_inst,
				sizeof(parm.lr_procfailure.pf_inst));
	parm.lr_procfailure = error_info->ei_processingFailure;	/* struct copy*/
	parm.lr_type = LR_PROC_FAILURE;
	break;

    default:
	return error_fail(invoke, M_LINKED_REPLY, error, mi);
    }

    if (build_CMIP_LinkedReplyArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Linked Reply Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Get Linked Reply Argument");

    if (RoInvokeRequest(msd, M_LINKED_REPLY, ROS_ASYNC, pe, invoke,
				&linked, ROS_NOPRIO, &rois) == NOTOK)
	return ros_fail(&rois, mi);

    pe_free(pe);
    return OK;
}


int  M_SetRes (msd, invoke, linked, obj_class, obj_inst, cur_time,
			nattrs, attrs, error, error_info, mi)
int  msd, invoke, linked;
MID  obj_class;
MN   obj_inst;
char* cur_time;
int  nattrs;
CMISSetAttr* attrs;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISSetRes parm;
    PE pe = NULLPE;
    int i, result = 0;

    mi->mi_invoke = 0;
    if (linked)
	return M_LinkedReplySet(msd, invoke, linked, obj_class, obj_inst,
				cur_time, nattrs, attrs, error, error_info, mi);

    switch (error) {
    case m_noError:
    case m_setListError:
	break;
    case m_duplicateInvocation:
    case m_unrecognisedOperation:
    case m_mistypedArgument:
    case m_resourceLimitation:
    case m_noSuchObjectClass:
    case m_noSuchObjectInstance:
    case m_accessDenied:
    case m_invalidScope:
    case m_invalidFilter:
    case m_syncNotSupported:
    case m_classInstanceConflict:
    case m_complexityLimitation:
    case m_processingFailure:
	return M_Error(msd, M_SETC, invoke, error, error_info, mi);
    default:
	return error_fail(invoke, M_SETC, error, mi);
    }

    for (i = 0; i < nattrs; i++) {
	if (null_midP(&attrs[i].sa_id))
	    return arg_fail(invoke, M_SETC, 1, "Attribute Id", mi);
        if (attrs[i].sa_val == NULLPE && !attrs[i].sa_error)
	    return arg_fail(invoke, M_SETC, 1, "Attribute Value", mi);
    }

    if (obj_class) {
	parm.sr_class = *obj_class;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.sr_class, sizeof(parm.sr_class));

    if (obj_inst) {
	parm.sr_inst  = *obj_inst;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.sr_inst, sizeof(parm.sr_inst));

    if (parm.sr_time = cur_time)
	result = 1;

    if (parm.sr_nattrs = nattrs)
	result = 1;
    parm.sr_attrs = attrs;

    if (result)
	if (!error) {
	    if (build_CMIP_SetResult(&pe, 1, 0, NULLCP, &parm) == NOTOK)
		return build_fail(invoke, "Set Result", mi);
	}
	else
	    if (build_CMIP_SetListError(&pe, 1, 0, NULLCP, &parm) == NOTOK)
		return build_fail(invoke, "SetListError", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, error ? "SetListError" :
                    "Set Result");

    result = error == m_noError ?
	RoResultRequest(msd, invoke, M_SETC, pe, ROS_NOPRIO, &rois) :
	RoErrorRequest(msd, invoke, (int) error - CMIS_ERROR_OFFSET, pe,
				ROS_NOPRIO, &rois);

    if (pe)
	pe_free(pe);

    if (result == NOTOK)
	return ros_fail(&rois, mi);
    return OK;
}


static int  M_LinkedReplySet (msd, invoke, linked, obj_class, obj_inst,
				cur_time, nattrs, attrs, error, error_info, mi)
int  msd, invoke, linked;
MID  obj_class;
MN   obj_inst;
char* cur_time;
int  nattrs;
CMISSetAttr* attrs;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    struct CMISLinkedReplyArg parm;
    PE pe;
    int i;

    mi->mi_invoke = 0;
    switch (error) {
    case m_noError:
    case m_setListError:
	if (missing_midP(obj_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Class", mi);
	if (missing_mnP(obj_inst))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Instance", mi);
	if ((parm.lr_setres.sr_nattrs = nattrs) == 0)
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Attribute List", mi);

	for (i = 0; i < nattrs; i++) {
	    if (null_midP(&attrs[i].sa_id))
		return arg_fail(invoke, M_LINKED_REPLY, 0, "Attribute Id", mi);
	    if (attrs[i].sa_val == NULLPE)
		return arg_fail(invoke, M_LINKED_REPLY, 0, "Attribute Value",
				mi);
        }

	parm.lr_setres.sr_class = *obj_class;	/* struct copy */
	parm.lr_setres.sr_inst  = *obj_inst;	/* struct copy */
	parm.lr_setres.sr_time  = cur_time;
	parm.lr_setres.sr_attrs = attrs;
	parm.lr_type		= error ? LR_SET_LISTERR : LR_SET_RES;
	break;
		
    case m_processingFailure:
	if (!error_info)
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - ErrorInfo", mi);
	if (null_midP(&error_info->ei_processingFailure.pf_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - Object Class", mi);
	if (null_mparmP(&error_info->ei_processingFailure.pf_error))
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - SpecificErrorInfo", mi);
	bzero((char*) &parm.lr_procfailure.pf_inst,
				sizeof(parm.lr_procfailure.pf_inst));
	parm.lr_procfailure = error_info->ei_processingFailure;	/* struct copy*/
	parm.lr_type = LR_PROC_FAILURE;
	break;

    default:
	return error_fail(invoke, M_LINKED_REPLY, error, mi);
    }

    if (build_CMIP_LinkedReplyArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Linked Reply Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Set Linked Reply Argument");

    if (RoInvokeRequest(msd, M_LINKED_REPLY, ROS_ASYNC, pe, invoke,
				&linked, ROS_NOPRIO, &rois) == NOTOK)
	return ros_fail(&rois, mi);

    pe_free(pe);
    return OK;
}


int  M_ActionRes (msd, invoke, linked, obj_class, obj_inst, cur_time,
				action_reply, error, error_info, mi)
int  msd, invoke, linked;
MID  obj_class;
MN   obj_inst;
char* cur_time;
CMISParam* action_reply;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISActionRes parm;
    PE pe = NULLPE;
    int result = 0;

    mi->mi_invoke = 0;
    if (linked)
	return M_LinkedReplyAction(msd, invoke, linked, obj_class, obj_inst,
				cur_time, action_reply, error, error_info, mi);

    switch (error) {
    case m_noError:
	break;
    case m_duplicateInvocation:
    case m_unrecognisedOperation:
    case m_mistypedArgument:
    case m_resourceLimitation:
    case m_noSuchObjectClass:
    case m_noSuchObjectInstance:
    case m_accessDenied:
    case m_invalidScope:
    case m_invalidFilter:
    case m_syncNotSupported:
    case m_classInstanceConflict:
    case m_complexityLimitation:
    case m_noSuchAction:
    case m_noSuchArgument:
    case m_invalidArgumentValue:
    case m_processingFailure:
	return M_Error(msd, M_ACTIONC, invoke, error, error_info, mi);
    default:
	return error_fail(invoke, M_ACTIONC, error, mi);
    }

    if (obj_class) {
	parm.ar_class = *obj_class;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.ar_class, sizeof(parm.ar_class));

    if (obj_inst) {
	parm.ar_inst  = *obj_inst;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.ar_inst, sizeof(parm.ar_inst));

    if (parm.ar_time = cur_time)
	result = 1;

    if (action_reply) {
	if (null_midP(&action_reply->mp_id))
	    return arg_fail(invoke, M_ACTIONC, 1, "Action Reply Id", mi);
	if (action_reply->mp_val == NULLPE)
	    return arg_fail(invoke, M_ACTIONC, 1, "Action Reply Info", mi);
	parm.ar_reply = *action_reply;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.ar_reply, sizeof(parm.ar_reply));

    if (result && build_CMIP_ActionResult(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Action Result", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Action Result");

    if (RoResultRequest(msd, invoke, M_ACTIONC, pe, ROS_NOPRIO, &rois)
		== NOTOK) {
	if (pe)
	    pe_free(pe);
	return ros_fail(&rois, mi);
    }

    if (pe)
	pe_free(pe);
    return OK;
}


static int  M_LinkedReplyAction (msd, invoke, linked, obj_class, obj_inst,
				cur_time, action_reply, error, error_info, mi)
int  msd, invoke, linked;
MID  obj_class;
MN   obj_inst;
char* cur_time;
CMISParam* action_reply;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    struct CMISLinkedReplyArg parm;
    PE pe;

    mi->mi_invoke = 0;
    switch (error) {
    case m_noError:
	if (missing_midP(obj_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Class", mi);
	if (missing_mnP(obj_inst))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Instance", mi);

	parm.lr_actionres.ar_class = *obj_class;	/* struct copy */
	parm.lr_actionres.ar_inst  = *obj_inst;		/* struct copy */
	parm.lr_actionres.ar_time  = cur_time;
	if (action_reply)
	    parm.lr_actionres.ar_reply = *action_reply;	/* struct copy */
	else
	    bzero((char*) &parm.lr_actionres.ar_reply,
				sizeof(parm.lr_actionres.ar_reply));
	parm.lr_type = LR_ACTION_RES;
	break;
		
    case m_accessDenied:
	if (missing_midP(obj_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Class", mi);
	if (missing_mnP(obj_inst))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Instance", mi);
	if (!error_info ||
		null_midP(&error_info->ei_linkedActionAccessDenied))
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
				"Error (m_accessDenied) - ActionTypeId", mi);

	parm.lr_actionerr.ae_class = *obj_class;	/* struct copy */
	parm.lr_actionerr.ae_inst  = *obj_inst;		/* struct copy */
	parm.lr_actionerr.ae_time  = cur_time;
	parm.lr_actionerr.ae_error = m_accessDenied;
	parm.lr_actionerr.ae_accessDenied =
	    error_info->ei_linkedActionAccessDenied;	/* struct copy */
	parm.lr_type = LR_ACTION_ERR;
	break;

    case m_noSuchAction:
    case m_noSuchArgument:
	if (missing_midP(obj_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Class", mi);
	if (missing_mnP(obj_inst))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Instance", mi);
	if (!error_info ||
		null_midP(&error_info->ei_noSuchArgument.nsa_type))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "ActionTypeId", mi);

	parm.lr_actionerr.ae_class = *obj_class;	/* struct copy */
	parm.lr_actionerr.ae_inst  = *obj_inst;		/* struct copy */
	parm.lr_actionerr.ae_time  = cur_time;
	parm.lr_actionerr.ae_error = error;
	if (error == m_noSuchAction)
	    parm.lr_actionerr.ae_noSuchAction =
			error_info->ei_noSuchAction.nsa_type;
							/* struct copy */
	else
	    parm.lr_actionerr.ae_noSuchArgument =
			error_info->ei_noSuchArgument;	/* struct copy */
	parm.lr_type = LR_ACTION_ERR;
	break;

    case m_invalidArgumentValue:
	if (missing_midP(obj_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Class", mi);
	if (missing_mnP(obj_inst))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Instance", mi);
	if (!error_info ||
		null_mparmP(&error_info->ei_invalidArgumentValue))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "ActionReplyId", mi);

	parm.lr_actionerr.ae_class = *obj_class;	/* struct copy */
	parm.lr_actionerr.ae_inst  = *obj_inst;		/* struct copy */
	parm.lr_actionerr.ae_time  = cur_time;
	parm.lr_actionerr.ae_error = m_invalidArgumentValue;
	parm.lr_actionerr.ae_invalidArgumentValue =
		error_info->ei_invalidArgumentValue;	/* struct copy */
	parm.lr_type = LR_ACTION_ERR;
	break;
	
    case m_processingFailure:
	if (!error_info)
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - ErrorInfo", mi);
	if (null_midP(&error_info->ei_processingFailure.pf_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - Object Class", mi);
	if (null_mparmP(&error_info->ei_processingFailure.pf_error))
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - SpecificErrorInfo", mi);
	bzero((char*) &parm.lr_procfailure.pf_inst,
				sizeof(parm.lr_procfailure.pf_inst));
	parm.lr_procfailure = error_info->ei_processingFailure;	/* struct copy*/
	parm.lr_type = LR_PROC_FAILURE;
	break;

    default:
	return error_fail(invoke, M_LINKED_REPLY, error, mi);
    }

    if (build_CMIP_LinkedReplyArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Linked Reply Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Action Linked Reply Argument");

    if (RoInvokeRequest(msd, M_LINKED_REPLY, ROS_ASYNC, pe, invoke,
				&linked, ROS_NOPRIO, &rois) == NOTOK)
	return ros_fail(&rois, mi);

    pe_free(pe);
    return OK;
}


int  M_CreateRes (msd, invoke, obj_class, obj_inst, cur_time, nattrs, attrs,
				error, error_info, mi)
int  msd, invoke;
MID  obj_class;
MN   obj_inst;
char* cur_time;
int  nattrs;
CMISParam* attrs;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISCreateRes parm;
    PE pe = NULLPE;
    int i, result = 0;

    mi->mi_invoke = 0;
    switch (error) {
    case m_noError:
	break;
    case m_duplicateInvocation:
    case m_unrecognisedOperation:
    case m_mistypedArgument:
    case m_resourceLimitation:
    case m_noSuchObjectClass:
    case m_duplicateManagedObjectInstance:
    case m_noSuchReferenceObject:
    case m_accessDenied:
    case m_noSuchAttribute:
    case m_noSuchObjectInstance:
    case m_classInstanceConflict:
    case m_invalidObjectInstance:
    case m_missingAttributeValue:
    case m_invalidAttributeValue:
    case m_processingFailure:
	return M_Error(msd, M_CREATE, invoke, error, error_info, mi);
    default:
	return error_fail(invoke, M_CREATE, error, mi);
    }

    for (i = 0; i < nattrs; i++) {
	if (null_midP(&attrs[i].mp_id))
	    return arg_fail(invoke, M_CREATE, 1, "Attribute Id", mi);
	if (attrs[i].mp_val == NULLPE)
	    return arg_fail(invoke, M_CREATE, 1, "Attribute Value", mi);
    }

    if (obj_class) {
	parm.cr_class = *obj_class;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.cr_class, sizeof(parm.cr_class));

    if (obj_inst) {
	parm.cr_inst  = *obj_inst;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.cr_inst, sizeof(parm.cr_inst));

    if (parm.cr_time = cur_time)
	result = 1;

    if (parm.cr_nattrs = nattrs)
	result = 1;
    parm.cr_attrs = attrs;

    if (result && build_CMIP_CreateResult(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Create Result", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Create Result");

    if (RoResultRequest(msd, invoke, M_CREATE, pe, ROS_NOPRIO, &rois)
		== NOTOK) {
	if (pe)
	    pe_free(pe);
	return ros_fail(&rois, mi);
    }

    if (pe)
	pe_free(pe);
    return OK;
}


int  M_DeleteRes (msd, invoke, linked, obj_class, obj_inst, cur_time,
				error, error_info, mi)
int  msd, invoke, linked;
MID  obj_class;
MN   obj_inst;
char* cur_time;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISDeleteRes parm;
    PE pe = NULLPE;
    int result = 0;

    mi->mi_invoke = 0;
    if (linked)
	return M_LinkedReplyDelete(msd, invoke, linked, obj_class, obj_inst,
					cur_time, error, error_info, mi);

    switch (error) {
    case m_noError:
	break;
    case m_duplicateInvocation:
    case m_unrecognisedOperation:
    case m_mistypedArgument:
    case m_resourceLimitation:
    case m_noSuchObjectClass:
    case m_noSuchObjectInstance:
    case m_accessDenied:
    case m_invalidScope:
    case m_invalidFilter:
    case m_syncNotSupported:
    case m_classInstanceConflict:
    case m_complexityLimitation:
    case m_processingFailure:
	return M_Error(msd, M_DELETE, invoke, error, error_info, mi);
    default:
	return error_fail(invoke, M_DELETE, error, mi);
    }

    if (obj_class) {
	parm.dr_class = *obj_class;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.dr_class, sizeof(parm.dr_class));

    if (obj_inst) {
	parm.dr_inst  = *obj_inst;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.dr_inst, sizeof(parm.dr_inst));

    if (parm.dr_time = cur_time)
	result = 1;

    if (result && build_CMIP_DeleteResult(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Delete Result", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Delete Result");

    if (RoResultRequest(msd, invoke, M_DELETE, pe, ROS_NOPRIO, &rois)
		== NOTOK) {
	if (pe)
	    pe_free(pe);
	return ros_fail(&rois, mi);
    }

    if (pe)
	pe_free(pe);
    return OK;
}


static int  M_LinkedReplyDelete (msd, invoke, linked, obj_class, obj_inst,
					cur_time, error, error_info, mi)
int  msd, invoke, linked;
MID  obj_class;
MN   obj_inst;
char* cur_time;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    struct CMISLinkedReplyArg parm;
    PE pe;

    mi->mi_invoke = 0;
    switch (error) {
    case m_noError:
	if (missing_midP(obj_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Class", mi);
	if (missing_mnP(obj_inst))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Instance", mi);

	parm.lr_deleteres.dr_class = *obj_class;	/* struct copy */
	parm.lr_deleteres.dr_inst  = *obj_inst;		/* struct copy */
	parm.lr_deleteres.dr_time  = cur_time;
	parm.lr_type = LR_DELETE_RES;
	break;
		
    case m_accessDenied	:
	if (missing_midP(obj_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Class", mi);
	if (missing_mnP(obj_inst))
	    return arg_fail(invoke, M_LINKED_REPLY, 0, "Object Instance", mi);

	parm.lr_deleteerr.de_class = *obj_class;	/* struct copy */
	parm.lr_deleteerr.de_inst  = *obj_inst;		/* struct copy */
	parm.lr_deleteerr.de_time  = cur_time;
	parm.lr_deleteerr.de_error = m_accessDenied;
	parm.lr_type = LR_DELETE_ERR;
	break;
	
    case m_processingFailure:
	if (!error_info)
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - ErrorInfo", mi);
	if (null_midP(&error_info->ei_processingFailure.pf_class))
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - Object Class", mi);
	if (null_mparmP(&error_info->ei_processingFailure.pf_error))
	    return arg_fail(invoke, M_LINKED_REPLY, 0,
			"Error (processingFailure) - SpecificErrorInfo", mi);
	bzero((char*) &parm.lr_procfailure.pf_inst,
				sizeof(parm.lr_procfailure.pf_inst));
	parm.lr_procfailure = error_info->ei_processingFailure;	/* struct copy*/
	parm.lr_type = LR_PROC_FAILURE;
	break;

    default:
	return error_fail(invoke, M_LINKED_REPLY, error, mi);
    }

    if (build_CMIP_LinkedReplyArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Linked Reply Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Delete Linked Reply Argument");

    if (RoInvokeRequest(msd, M_LINKED_REPLY, ROS_ASYNC, pe, invoke,
				&linked, ROS_NOPRIO, &rois) == NOTOK)
	return ros_fail(&rois, mi);

    pe_free(pe);
    return OK;
}


int  M_CancelGetRes (msd, invoke, error, error_info, mi)
int  msd, invoke;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    PE pe; /* for integrity checks */

    mi->mi_invoke = 0;
    switch (error) {
    case m_noError:
	break;
    case m_duplicateInvocation:
    case m_unrecognisedOperation:
    case m_resourceLimitation:
    case m_noSuchInvokeId:
    case m_mistypedOperation:
    case m_processingFailure:
	return M_Error(msd, M_CANCEL_GET, invoke, error, error_info, mi);
    default:
	return error_fail(invoke, M_CANCEL_GET, error, mi);
    }

    /* A wee workaround ... */
    pe = int2prim(invoke);
    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "CancelGet Result");
    pe_free(pe);

    if (RoResultRequest(msd, invoke, M_CANCEL_GET, NULLPE, ROS_NOPRIO, &rois)
		== NOTOK)
	return ros_fail(&rois, mi);
    return OK;
}

