/*
 * Copyright (c) 1988 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* moperaux.c - auxilliary routines used by management operation primitives */

/*
 * By George Pavlou, April 1990
 */

#include <stdio.h>
#include "msap.h"
#include "moper.h"

extern int moperation_id;	/* in msap.py */


int  M_Error (msd, op, invoke, error, error_info, mi)
int  msd, op, invoke;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    PE pe = (PE) 0;
    int result, reason = 0;

    if (!error_info && !(error == m_duplicateInvocation ||
			 error == m_unrecognisedOperation ||
			 error == m_mistypedArgument ||
			 error == m_resourceLimitation ||
			 error == m_accessDenied ||
			 error == m_mistypedOperation ||
			 error == m_operationCancelled ||
			 error == m_complexityLimitation ||
			 error == m_processingFailure)) {
	mi->mi_type = MI_PREJECT;
	mi->mi_preject.mpr_reason = MA_BADARGS;
	mi->mi_invoke = mi->mi_preject.mpr_id = invoke;
	(void) sprintf(mi->mi_preject.mpr_data,
			"%s Result: no error info for error %s",
			op2str(op), MErrString(error));
	(void) fprintf(stderr, "%s!\n", mi->mi_preject.mpr_data);
	return NOTOK;
    }

    switch (error) {
    case m_duplicateInvocation:
	reason = ROS_IP_DUP;
    case m_unrecognisedOperation:
	if (!reason)
	    reason = ROS_IP_UNRECOG;
    case m_mistypedArgument:
	if (!reason)
	    reason = ROS_IP_MISTYPED;
    case m_resourceLimitation:
	if (!reason)
	    reason = ROS_IP_LIMIT;

	if (RoURejectRequest(msd, &invoke, reason, ROS_NOPRIO, &rois) == NOTOK)
	    return ros_fail(&rois, mi);
	return OK;

    case m_noSuchObjectClass:
	if (null_midP(&error_info->ei_noSuchObjectClass))
	    return arg_fail(invoke, op, 1, "Error (noSuchObjectClass) - \
Object Class", mi);
	if (build_CMIP_ObjectClass(&pe, 1, 0, NULLCP,
			&error_info->ei_noSuchObjectClass) == NOTOK)
	    return build_fail(invoke, "ObjectClass", mi);
	break;

    case m_noSuchObjectInstance:
    case m_duplicateManagedObjectInstance:
    case m_noSuchReferenceObject:
    case m_invalidObjectInstance:
	if (null_mnP(&error_info->ei_noSuchObjectInstance))
	    return arg_fail(invoke, op, 1, "Error - Object Instance", mi);
	if (build_CMIP_ObjectInstance(&pe, 1, 0, NULLCP,
			&error_info->ei_noSuchObjectInstance) == NOTOK)
	    return build_fail(invoke, "ObjectInstance", mi);
	break;

    case m_accessDenied:
    case m_mistypedOperation:
    case m_operationCancelled:
	pe = NULLPE;
	break;

    case m_syncNotSupported:
	if (build_CMIP_CMISSync(&pe, 1, 0, NULLCP,
			(int *) &error_info->ei_syncNotSupported) == NOTOK)
	    return build_fail(invoke, "CMISSync", mi);
	break;

    case m_invalidFilter:
	if (error_info->ei_invalidFilter == NULLMFILTER)
	    return arg_fail(invoke, op, 1, "Error (invalidFilter) - \
CMIS Filter", mi);
	if (build_CMIP_CMISFilter(&pe, 1, 0, NULLCP,
			error_info->ei_invalidFilter) == NOTOK)
	    return build_fail(invoke, "CMISFilter", mi);
	break;

    case m_noSuchAttribute:
	if (null_midP(&error_info->ei_noSuchAttribute))
	    return arg_fail(invoke, op, 1, "Error - AttributeId", mi);
	if (build_CMIP_AttributeId(&pe, 1, 0, NULLCP,
			&error_info->ei_noSuchAttribute) == NOTOK)
	    return build_fail(invoke, "AttributeId", mi);
	break;

    case m_invalidAttributeValue:
	if (null_midP(&error_info->ei_invalidAttributeValue.mp_id))
	    return arg_fail(invoke, op, 1, "Error (invalidAttributeValue) - \
Attribute Id", mi);
	if (error_info->ei_invalidAttributeValue.mp_val == NULLPE)
	    return arg_fail(invoke, op, 1, "Error (invalidAttributeValue) - \
Attribute Value", mi);
	if (build_CMIP_Attribute(&pe, 1, 0, NULLCP,
			&error_info->ei_invalidAttributeValue) == NOTOK)
	    return build_fail(invoke, "Attribute", mi);
	break;

    case m_noSuchAction:
	if (null_midP(&error_info->ei_noSuchAction.nsa_class))
	    return arg_fail(invoke, op, 1, "Error (noSuchAction) - \
Object Class", mi);
	if (null_midP(&error_info->ei_noSuchAction.nsa_type))
	    return arg_fail(invoke, op, 1, "Error (noSuchAction) - \
ActionTypeId", mi);
	if (build_CMIP_NoSuchAction(&pe, 1, 0, NULLCP,
			&error_info->ei_noSuchAction) == NOTOK)
	    return build_fail(invoke, "NoSuchAction", mi);
	break;

    case m_processingFailure:
	/* check in case no info has been supplied - error parameter optional */
	if (!error_info) {
	    pe = NULLPE;
	    break;
	}
	/* if info has been supplied check its completeness */
	if (null_midP(&error_info->ei_processingFailure.pf_class))
	    return arg_fail(invoke, op, 1, "Error (processingFailure) - \
Object Class", mi);
	if (null_midP(&error_info->ei_processingFailure.pf_error.mp_id))
	    return arg_fail(invoke, op, 1, "Error (processingFailure) - \
ErrorInfoId", mi);
	if (error_info->ei_processingFailure.pf_error.mp_val == NULLPE)
	    return arg_fail(invoke, op, 1, "Error (processingFailure) - \
ErrorInfo", mi);

	if (build_CMIP_ProcessingFailure(&pe, 1, 0, NULLCP,
			&error_info->ei_processingFailure) == NOTOK)
	    return build_fail(invoke, "ProcessingFailure", mi);
	break;

    case m_noSuchEventType:
	if (null_midP(&error_info->ei_noSuchEventType.nsa_class))
	    return arg_fail(invoke, op, 1, "Error (noSuchEventType) - \
Object Class", mi);
	if (null_midP(&error_info->ei_noSuchAction.nsa_type))
	    return arg_fail(invoke, op, 1, "Error (noSuchEventType) - \
EventTypeId", mi);
	if (build_CMIP_NoSuchEventType(&pe, 1, 0, NULLCP,
			&error_info->ei_noSuchEventType) == NOTOK)
	    return build_fail(invoke, "NoSuchEventType", mi);
	break;

    case m_noSuchArgument:
	if (null_midP(&error_info->ei_noSuchArgument.nsa_class))
	    return arg_fail(invoke, op, 1, "Error (noSuchArgument) - \
Object Class", mi);
	if (null_midP(&error_info->ei_noSuchArgument.nsa_type))
	    return arg_fail(invoke, op, 1, "Error (noSuchArgument) - \
Action/Event TypeId", mi);
	moperation_id = op;	/* for pepy encoder */
	if (build_CMIP_NoSuchArgument(&pe, 1, 0, NULLCP,
			&error_info->ei_noSuchArgument) == NOTOK)
	    return build_fail(invoke, "NoSuchArgument", mi);
	break;

    case m_invalidArgumentValue:
	if (null_mparmP(&error_info->ei_invalidArgumentValue))
	    return arg_fail(invoke, op, 1, "Error (invalidArgumentValue) - \
Action/EventType Id", mi);
	moperation_id = op;	/* for pepy encoder */
	if (build_CMIP_InvalidArgumentValue(&pe, 1, 0, NULLCP,
			&error_info->ei_invalidArgumentValue) == NOTOK)
	    return build_fail(invoke, "InvalidArgumentValue", mi);
	break;

    case m_invalidScope:
	if (build_CMIP_Scope(&pe, 1, 0, NULLCP,
			&error_info->ei_invalidScope) == NOTOK)
	    return build_fail(invoke, "Scope", mi);
	break;

    case m_missingAttributeValue:
	if (error_info->ei_missingAttributeValue == NULL)
	    return arg_fail(invoke, op, 1, "Error (missingAttributeValue) - \
AttributeValue", mi);
	if (build_CMIP_AttributeIdList(&pe, 1, 0, NULLCP,
			error_info->ei_missingAttributeValue) == NOTOK)
	    return build_fail(invoke, "AttributeIdList", mi);
	break;

    case m_classInstanceConflict:
	if (null_midP(&error_info->ei_classInstanceConflict.bmo_class))
	    return arg_fail(invoke, op, 1, "Error (classInstanceConflict) - \
Object Class", mi);
	if (null_mnP(&error_info->ei_classInstanceConflict.bmo_inst))
	    return arg_fail(invoke, op, 1, "Error (classInstanceConflict) - \
Object Instance", mi);
	if (build_CMIP_BaseManagedObjectId(&pe, 1, 0, NULLCP,
			&error_info->ei_classInstanceConflict) == NOTOK)
	    return build_fail(invoke, "BaseManagedObjectId", mi);
	break;
	
    case m_complexityLimitation:
	/* check in case no info has been supplied - error parameter optional */
	if (!error_info) {
	    pe = NULLPE;
	    break;
	}
	if (build_CMIP_ComplexityLimitation(&pe, 1, 0, NULLCP,
			&error_info->ei_complexityLimitation) == NOTOK)
	    return build_fail(invoke, "ComplexityLimitation", mi);
	break;
	
    case m_noSuchInvokeId:
	if ((pe = int2prim(error_info->ei_noSuchInvokeId)) == NULLPE)
	    return build_fail(invoke, "Integer (NoSuchInvokeId)", mi);
	break;

    default:
	(void) fprintf(stderr, "Invalid Error Type: %d\n", error);
	return NOTOK;
    }

    /*
    ** needs checking ###
    */
    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, MErrString(error));

    result = RoErrorRequest(msd, invoke, (int) error - CMIS_ERROR_OFFSET, pe,
			ROS_NOPRIO, &rois);
    if (pe)
	pe_free(pe);

    if (result == NOTOK)
	return ros_fail(&rois, mi);
    return OK;
}


int  error_fail (invoke, op, error, mi)
int  invoke, op;
CMISErrors error;
MSAPIndication* mi;
{
    mi->mi_type = MI_PREJECT;
    mi->mi_preject.mpr_reason = MA_BADARGS;
    mi->mi_invoke = mi->mi_preject.mpr_id = invoke;
    (void) sprintf(mi->mi_preject.mpr_data,
		"%s Result: invalid error %s", op2str(op), MErrString(error));
    (void) fprintf(stderr, "%s!\n", mi->mi_preject.mpr_data);
    return NOTOK;
}


int  ros_fail (roi, mi)
struct RoSAPindication* roi;
MSAPIndication* mi;
{
    mi->mi_type = MI_PREJECT;
    mi->mi_preject.mpr_reason = roi->roi_preject.rop_reason;
    mi->mi_invoke = mi->mi_preject.mpr_id = roi->roi_preject.rop_id;
    (void) strncpy(mi->mi_preject.mpr_data,roi->roi_preject.rop_data, MPR_SIZE);
    (void) fprintf(stderr, "%s!\n", mi->mi_preject.mpr_data);
    ROPFREE(&roi->roi_preject);
    return NOTOK;
}


int  arg_fail (invoke, op, res, arg, mi)
int  invoke, op, res;
char* arg;
MSAPIndication* mi;
{
    mi->mi_type = MI_PREJECT;
    mi->mi_preject.mpr_reason = MA_BADARGS;
    mi->mi_invoke = mi->mi_preject.mpr_id = invoke;
    if (res) {
        (void) sprintf(mi->mi_preject.mpr_data,
		"%s Result: argument \"%s\" is mandatory", op2str(op), arg);
	(void) fprintf(stderr, "%s!\n", mi->mi_preject.mpr_data);
    }
    else {
        (void) sprintf(mi->mi_preject.mpr_data,
		"%s: argument \"%s\" is mandatory", op2str(op), arg);
	(void) fprintf(stderr, "%s!\n", mi->mi_preject.mpr_data);
    }
    return NOTOK;
}


int  build_fail (invoke, s, mi)
int invoke;
char* s;
MSAPIndication* mi;
{
    mi->mi_type = MI_PREJECT;
    mi->mi_preject.mpr_reason = MA_BADARGS;
    mi->mi_invoke = mi->mi_preject.mpr_id = invoke;
    (void) sprintf(mi->mi_preject.mpr_data, "Build %.*s: Incorrect Arguments",
		MPR_SIZE-30, s);
    (void) fprintf(stderr, "%s!\n", mi->mi_preject.mpr_data);
    return NOTOK;
}


char*  op2str (op)
int op;
{
    static char buf[16];
    switch (op) {
    case M_EVENT_REP:
	(void) strcpy(buf, "M_EventRep");
	break;
    case M_EVENT_REPC:
	(void) strcpy(buf, "M_EventRepC");
	break;
    case M_LINKED_REPLY:
	(void) strcpy(buf, "M_LinkedReply");
	break;
    case M_GET:
	(void) strcpy(buf, "M_Get");
	break;
    case M_SET:
	(void) strcpy(buf, "M_Set");
	break;
    case M_SETC:
	(void) strcpy(buf, "M_SetC");
	break;
    case M_ACTION:
	(void) strcpy(buf, "M_Action");
	break;
    case M_ACTIONC:
	(void) strcpy(buf, "M_ActionC");
	break;
    case M_CREATE:
	(void) strcpy(buf, "M_Create");
	break;
    case M_DELETE:
	(void) strcpy(buf, "M_Delete");
	break;
    case M_CANCEL_GET:
	(void) strcpy(buf, "M_CancelGet");
	break;
    default:
	(void) fprintf(stderr, "op2str: unknown operation code %d\n", op);
	buf[0] = '\0';
	break;
    }
    return buf;
}


void  do_timing (msg)
char* msg;
{
    static struct timeval prevTime;
    struct timeval now;
    long diffsec, diffusec;

    if (!msg) {
	gettimeofday(&prevTime, NULL);
	return;
    }
	
    gettimeofday(&now, NULL);

    diffsec = now.tv_sec - prevTime.tv_sec;
    if ((diffusec = now.tv_usec - prevTime.tv_usec) < 0)
        --diffsec, diffusec += 1000000;

    prevTime = now;	/* struct copy */

    printf("%s in %d.%06d seconds\n\n", msg, diffsec, diffusec);
}
