/*
 * Copyright (c) 1988 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* mopermgr.c - perform manager operations */

/*
 * By George Pavlou, April 1990
 */

/*
** Integrity check stuff added by Saleem
** March/April 1994
*/

#include <stdio.h>
#include "msap.h"
#include "moper.h"


/* LINTLIBRARY */

int  M_EventRepRes (msd, invoke, obj_class, obj_inst, cur_time, event_reply,
				error, error_info, mi)
int  msd, invoke;
MID  obj_class;
MN   obj_inst;
char* cur_time;
CMISParam* event_reply;
CMISErrors error;
CMISErrorInfo* error_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISEventReportRes parm;
    PE pe = NULLPE;
    int result = 0;

    mi->mi_invoke = 0;
    switch (error) {
    case m_noError:
	break;
    case m_duplicateInvocation:
    case m_unrecognisedOperation:
    case m_mistypedArgument:
    case m_resourceLimitation:
    case m_noSuchObjectClass:
    case m_noSuchObjectInstance:
    case m_noSuchEventType:
    case m_noSuchArgument:
    case m_invalidArgumentValue:
    case m_processingFailure:
	return M_Error(msd, M_EVENT_REPC, invoke, error, error_info, mi);
    default:
	return error_fail(invoke, M_EVENT_REPC, error, mi);
    }

    if (obj_class) {
	parm.er_class = *obj_class;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.er_class, sizeof(parm.er_class));

    if (obj_inst) {
	parm.er_inst  = *obj_inst;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.er_inst, sizeof(parm.er_inst));

    if (parm.er_time = cur_time)
	result = 1;

    if (event_reply) {
	parm.er_reply = *event_reply;	/* struct copy */
	result = 1;
    }
    else
	bzero((char*) &parm.er_reply, sizeof(parm.er_reply));

    if (result && build_CMIP_EventReportResult(&pe, 1, 0, NULLCP, &parm)
			== NOTOK)
	return build_fail(invoke, "Event Report Result", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Event Report Result");

    if (RoResultRequest(msd, invoke, M_EVENT_REPC, pe, ROS_NOPRIO, &rois)
		== NOTOK) {
	if (pe)
	    pe_free(pe);
	return ros_fail(&rois, mi);
    }

    if (pe)
	pe_free(pe);
    return OK;
}


int  M_Get (msd, invoke, obj_class, obj_inst, scope, filter,
			access, sync, nattrs, attrs, mi)
int  msd, invoke;
MID  obj_class;
MN   obj_inst;
CMISScope* scope;
CMISFilter* filter;
External* access;
CMISSync sync;
int  nattrs;
MID  attrs;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISGetArg parm;
    PE pe;
    int i;
    MIDentifier dummyAttr;	/* because pepsy does not like GET_NOATTRS */

    mi->mi_invoke = 0;
    if (missing_mnP(obj_inst))
	return arg_fail(invoke, M_GET, 0, "Object Instance", mi);
    if (nattrs > 0 && !attrs)
	return arg_fail(invoke, M_GET, 0, "AttributeIdList", mi);
    if (missing_midP(obj_class)) {
	parm.ga_class.mid_type   = MID_GLOBAL;
	parm.ga_class.mid_global = ACTUALCLASS_ASN_OID;
    }
    else
	parm.ga_class = *obj_class;	/* struct copy */

    if (nattrs > 0) {
	for (i = 0; i < nattrs; i++)
	    if (null_midP(&attrs[i]))
		return arg_fail(invoke, M_GET, 0, "AttributeId", mi);
    }
    else
    if (nattrs == GET_NOATTRS) {
	/*
	 * this is because there is no way to convince pepsy to tell you
	 * if a SET OF is empty i.e. {} as opposed to non-existent (optional),
	 * so we send a dummy attribute across -
	 * this should be fixed in order to interoperate !!!
	 */
	dummyAttr.mid_type = MID_LOCAL;
	dummyAttr.mid_local = -1;
	nattrs = 1;
	attrs = &dummyAttr;
    }
    parm.ga_nattrs = nattrs;
    parm.ga_attrs  = attrs;

    if (scope)
	parm.ga_scope = *scope;		/* struct copy */
    else
	parm.ga_scope.sc_type = Sc_BaseObject;

    parm.ga_inst   = *obj_inst;		/* struct copy */
    parm.ga_filter = filter;
    parm.ga_access = access;
    parm.ga_sync   = sync;

    if (build_CMIP_GetArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Get Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Get Argument");
/*
    {
    char buff[1024];
    char* cp = buff; int len;

    pe2mem(pe, &cp, &len);
    for (i = 0; i < len; i++)
	printf("%02x", (unsigned int) ((unsigned char) cp[i]));
    printf("\n");
    }
*/

    if (RoInvokeRequest(msd, M_GET, ROS_ASYNC, pe, invoke,
				NULLIP, ROS_NOPRIO, &rois) == NOTOK) {
	pe_free(pe);
	return ros_fail(&rois, mi);
    }

    pe_free(pe);
    return OK;
}


int  M_SetAux (msd, op, invoke, obj_class, obj_inst, scope, filter,
			access, sync, nattrs, attrs, mi)
int  msd, op, invoke;
MID  obj_class;
MN   obj_inst;
CMISScope* scope;
CMISFilter* filter;
External* access;
CMISSync sync;
int  nattrs;
CMISSetAttr* attrs;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISSetArg parm;
    PE pe;
    int i;

    mi->mi_invoke = 0;
    if (missing_mnP(obj_inst))
	return arg_fail(invoke, op, 0, "Object Instance", mi);
    if (!nattrs || !attrs)
	return arg_fail(invoke, op, 0, "Attribute List", mi);
    if (missing_midP(obj_class)) {
	parm.sa_class.mid_type   = MID_GLOBAL;
	parm.sa_class.mid_global = ACTUALCLASS_ASN_OID;
    }
    else
	parm.sa_class = *obj_class;	/* struct copy */

    for (i = 0; i < nattrs; i++) {
	if (null_midP(&attrs->sa_id))
	    return arg_fail(invoke, M_SET, 0, "AttributeId", mi);	
	if (attrs -> sa_val == NULLPE && attrs -> sa_modify != m_setToDefault)
	    return arg_fail(invoke, op, 0, "Attribute Value", mi);
    }
    parm.sa_nattrs = nattrs;
    parm.sa_attrs  = attrs;

    if (scope)
	parm.sa_scope = *scope;		/* struct copy */
    else
	parm.sa_scope.sc_type = Sc_BaseObject;

    parm.sa_inst  = *obj_inst;		/* struct copy */
    parm.sa_filter = filter;
    parm.sa_access = access;
    parm.sa_sync   = sync;

    if (build_CMIP_SetArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Set Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Set Argument");

    if (RoInvokeRequest(msd, op, ROS_ASYNC, pe, invoke,
				NULLIP, ROS_NOPRIO, &rois) == NOTOK) {
	pe_free(pe);
	return ros_fail(&rois, mi);
    }

    pe_free(pe);
    return (OK);
}


int  M_ActionAux (msd, op, invoke, obj_class, obj_inst, scope, filter,
			access, sync, action_type, action_info, mi)
int  msd, op, invoke;
MID  obj_class;
MN   obj_inst;
CMISScope* scope;
CMISFilter* filter;
External* access;
CMISSync sync;
MID  action_type;
PE   action_info;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISActionArg parm;
    PE pe;

    mi->mi_invoke = 0;
    if (missing_mnP(obj_inst))
	return arg_fail(invoke, op, 0, "Object Instance", mi);
    if (missing_midP(action_type))
	return arg_fail(invoke, op, 0, "Action Type", mi);
    if (missing_midP(obj_class)) {
	parm.aa_class.mid_type   = MID_GLOBAL;
	parm.aa_class.mid_global = ACTUALCLASS_ASN_OID;
    }
    else
	parm.aa_class = *obj_class;	/* struct copy */

    if (scope)
	parm.aa_scope = *scope;		/* struct copy */
    else
	parm.aa_scope.sc_type = Sc_BaseObject;

    parm.aa_inst  = *obj_inst;		/* struct copy */
    parm.aa_info.mp_id = *action_type;	/* struct copy */
    parm.aa_info.mp_val = action_info;
    parm.aa_filter = filter;
    parm.aa_access = access;
    parm.aa_sync   = sync;

    if (build_CMIP_ActionArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Action Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Action Argument");

    if (RoInvokeRequest(msd, op, ROS_ASYNC, pe, invoke,
				NULLIP, ROS_NOPRIO, &rois) == NOTOK) {
	pe_free(pe);
	return ros_fail(&rois, mi);
    }

    pe_free(pe);
    return OK;
}


int  M_Create (msd, invoke, obj_class, obj_inst, instance_type, reference_inst,
			access, nattrs, attrs, mi)
int  msd, invoke;
MID  obj_class;
MN   obj_inst, reference_inst;
int  instance_type;
External* access;
int  nattrs;
CMISParam* attrs;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISCreateArg parm;
    PE pe;
    int i;

    mi->mi_invoke = 0;
    if (missing_midP(obj_class))
	return arg_fail(invoke, M_CREATE, 0, "Object Class", mi);

    if (obj_inst) {
	if (instance_type != CA_OBJECT_INST &&
			instance_type != CA_SUPERIOR_INST)
	    return arg_fail(invoke, M_CREATE, 0, "Instance Type", mi);
	parm.ca_inst_type = instance_type;
	parm.ca_inst      = *obj_inst;	/* struct copy */
    }
    else
	bzero((char*) &parm.ca_inst, sizeof(parm.ca_inst));

    for (i = 0; i < nattrs; i++) {
	if (null_midP(&attrs[i].mp_id))
	    return arg_fail(invoke, M_CREATE, 0, "Attribute Id", mi);
	if (attrs[i].mp_val == NULLPE)
	    return arg_fail(invoke, M_CREATE, 0, "Attribute Value", mi);
    }

    if (reference_inst)
	parm.ca_reference_inst = *reference_inst;	/* struct copy */
    else
	bzero((char*) &parm.ca_reference_inst, sizeof(parm.ca_reference_inst));

    parm.ca_class = *obj_class;		/* struct copy */
    parm.ca_access = access;
    parm.ca_nattrs = nattrs;
    parm.ca_attrs  = attrs;

    if (build_CMIP_CreateArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Create Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Create Argument");

    if (RoInvokeRequest(msd, M_CREATE, ROS_ASYNC, pe, invoke,
				NULLIP, ROS_NOPRIO, &rois) == NOTOK) {
	pe_free(pe);
	return ros_fail(&rois, mi);
    }

    pe_free(pe);
    return OK;
}


int  M_Delete (msd, invoke, obj_class, obj_inst, scope, filter,
			access, sync, mi)
int  msd, invoke;
MID  obj_class;
MN   obj_inst;
CMISScope* scope;
CMISFilter* filter;
External* access;
CMISSync sync;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    CMISDeleteArg parm;
    PE pe;

    mi->mi_invoke = 0;
    if (missing_mnP(obj_inst))
	return arg_fail(invoke, M_DELETE, 0, "Object Instance", mi);
    if (missing_midP(obj_class)) {
	parm.da_class.mid_type   = MID_GLOBAL;
	parm.da_class.mid_global = ACTUALCLASS_ASN_OID;
    }
    else
	parm.da_class = *obj_class;	/* struct copy */

    if (scope)
	parm.da_scope = *scope;		/* struct copy */
    else
	parm.da_scope.sc_type = Sc_BaseObject;

    parm.da_inst   = *obj_inst;		/* struct copy */
    parm.da_filter = filter;
    parm.da_access = access;
    parm.da_sync   = sync;

    if (build_CMIP_DeleteArgument(&pe, 1, 0, NULLCP, &parm) == NOTOK)
	return build_fail(invoke, "Delete Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "Delete Argument");

    if (RoInvokeRequest(msd, M_DELETE, ROS_ASYNC, pe, invoke,
				NULLIP, ROS_NOPRIO, &rois) == NOTOK) {
	pe_free(pe);
	return ros_fail(&rois, mi);
    }

    pe_free(pe);
    return OK;
}


int  M_CancelGet (msd, invoke, get_invoke, mi)
int  msd, invoke, get_invoke;
MSAPIndication* mi;
{
    struct RoSAPindication rois;
    PE pe;

    mi->mi_invoke = 0;
    if ((pe = int2prim(get_invoke)) == NULLPE)
	return build_fail(invoke, "Cancel Get Argument", mi);

    INTEGRITY_CHECK(msd, invoke, pe, mi->mi_invoke, "CancelGet Argument");

    if (RoInvokeRequest(msd, M_CANCEL_GET, ROS_ASYNC, pe, invoke,
				NULLIP, ROS_NOPRIO, &rois) == NOTOK) {
	pe_free(pe);
	return ros_fail(&rois, mi);
    }

    pe_free(pe);
    return OK;
}

