--
-- Copyright (c) 1988 University College London
-- All rights reserved.
--
-- Redistribution and use in source and binary forms are permitted
-- provided that the above copyright notice and this paragraph are
-- duplicated in all such forms and that any documentation,
-- advertising materials, and other materials related to such
-- distribution and use acknowledge that the software was developed
-- by the Department of Computer Science, University College London.
-- The name of the University may not be used to
-- endorse or promote products derived from this software without
-- specific prior written permission.
-- THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
-- IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
-- WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--

-- msap.py - pepy source for CMIP

-- By George Pavlou, June 1991


CMIP2

DEFINITIONS ::=


%{
#include <stdio.h>
#include "msap.h"

/* LINTLIBRARY */
%}


PREFIXES build parse print

BEGIN


-- Directory Service Definitions

IMPORTS
	DistinguishedName, RDNSequence FROM IF
	ObjectClass, ObjectInstance, CMISSync, Scope, CMISFilter,
	AttributeId, Attribute FROM CMIP
	;



ENCODER build

GetArgument [[P CMISGetArg *]] %{ MID attr; int i; %} ::= SEQUENCE
	{
	baseManagedObjectClass	ObjectClass [[p &parm -> ga_class]],
	baseManagedObjectInstance ObjectInstance [[p &parm -> ga_inst]],
	accessControl		[5] EXTERNAL [[p parm -> ga_access]]
				OPTIONAL << parm -> ga_access >>,
	synchronization		[6] IMPLICIT CMISSync
				[[p (int *) &parm -> ga_sync]]
				DEFAULT bestEffort
				<< parm -> ga_sync != s_bestEffort >>,
	scope			[7] Scope [[p &parm -> ga_scope]]
				DEFAULT baseObject
				<< parm -> ga_scope.sc_type != Sc_BaseObject >>,
	filter			CMISFilter [[p parm -> ga_filter]]
				DEFAULT and {}
				<< parm -> ga_filter >>,
	attributeIdList		[12] IMPLICIT SET OF
				<< i = 0, attr = parm->ga_attrs;
				   i < parm->ga_nattrs; i++, attr++ >>
				AttributeId [[p attr]]
				OPTIONAL << parm -> ga_nattrs != GET_ALLATTRS >>
	}

CreateArgument [[P CMISCreateArg *]] %{ CMISParam* attr; int i; %}
	::= SEQUENCE
	{
	managedObjectClass	ObjectClass [[p &parm -> ca_class]],
	instance CHOICE
		<< parm -> ca_inst_type >>
		{
		managedObjectInstance	ObjectInstance [[p &parm -> ca_inst]],
	 	superiorObjectInstance	[8] ObjectInstance
						[[p &parm -> ca_inst]]
		} OPTIONAL
		<< parm -> ca_inst.mn_dn >>,
	accessControl		[5] EXTERNAL [[p parm -> ca_access]]
				OPTIONAL << parm -> ca_access >>,
	referenceObjectInstance [6] ObjectInstance
					[[p &parm -> ca_reference_inst]]
				OPTIONAL
				<< parm->ca_reference_inst.mn_dn >>,
	attributeList		[7] IMPLICIT SET OF
				<< i = 0, attr = parm->ca_attrs;
				   i < parm->ca_nattrs; i++, attr++ >>
				Attribute [[p attr]]
				OPTIONAL << parm -> ca_nattrs >>
	}
	

DECODER parse

GetArgument [[P CMISGetArg *]]
    %{ MID attr, prev_block; int nblocks = 0; %}
    ::= SEQUENCE
    %{ parm -> ga_nattrs = GET_ALLATTRS; /* no attributeIdList */ %}
    {
    baseManagedObjectClass	CHOICE
	{
	globalForm   [0] IMPLICIT OBJECT IDENTIFIER
			 [[O parm->ga_class.mid_global]]
			 %{ parm->ga_class.mid_type = MID_GLOBAL; %},
	localForm    [1] IMPLICIT INTEGER
			 [[i parm->ga_class.mid_local]]
			 %{ parm->ga_class.mid_type = MID_LOCAL; %}
	},

    baseManagedObjectInstance	CHOICE
	{
	distinguishedName	[2] IMPLICIT DistinguishedName
				[[p &parm->ga_inst.mn_dn]]
				%{ parm->ga_inst.mn_type = MN_DN; %},
	nonSpecificForm         [3] IMPLICIT OCTET STRING
				[[s parm->ga_inst.mn_nonspec]]
				%{ parm->ga_inst.mn_type = MN_NONSPEC; %},
	localDistinguishedName  [4] IMPLICIT RDNSequence
				[[p &parm->ga_inst.mn_dn]]
				%{ parm->ga_inst.mn_type = MN_LOCALDN; %},
	},

    accessControl		[5] EXTERNAL [[p &parm -> ga_access]]
				OPTIONAL,
    synchronization		[6] IMPLICIT CMISSync
				[[p (int *) &parm -> ga_sync]]
				DEFAULT bestEffort,
    scope			[7] CHOICE
        {
        namedNumbers	    INTEGER [[i parm->ga_scope.sc_type]]
			    { baseObject (0),
                  	      firstLevelOnly (1),
                  	      wholeSubtree (2) }
			    %{ parm->ga_scope.sc_level = 0; %},
        individualLevels    [1] IMPLICIT INTEGER [[i parm->ga_scope.sc_level]]
			    %{ parm->ga_scope.sc_type = Sc_IndividualLevel; %},
        baseToNthLevel	    [2] IMPLICIT INTEGER [[i parm->ga_scope.sc_level]]
			    %{ parm->ga_scope.sc_type = Sc_BaseToNthLevel; %}
        }
	DEFAULT baseObject,

    filter			CMISFilter [[p &parm -> ga_filter]]
				DEFAULT and {},
    attributeIdList	    %{ parm->ga_nattrs = GET_NOATTRS;
			       /* empty attributeIdList */ %}
			    [12] IMPLICIT SET OF
			    %{ if (nblocks == 0) {
			           if ((prev_block = attr = parm->ga_attrs =
				           (MIDentifier*) malloc(M_ATTRBLOCK *
					   sizeof(MIDentifier))) == NULL)
				       return NOTOK;
			           nblocks = 1;
			           parm->ga_nattrs = 0;
			       }
			    %}
			    AttributeId [[p (MID) ga_pattrs]]
			    %{ *attr = *(MID) ga_pattrs; /* struct cpy */
			       free((char*) ga_pattrs);
			       if (++parm->ga_nattrs % M_ATTRBLOCK == 0 &&
					    (parm->ga_attrs = (MIDentifier*)
					    realloc((char*) parm->ga_attrs,
					    (unsigned) ++nblocks*M_ATTRBLOCK*
					    sizeof(MIDentifier))) == NULL)
				   return NOTOK;
	   		       if (parm->ga_attrs != prev_block) {
			           attr = parm->ga_attrs +
						(nblocks-1)*M_ATTRBLOCK;
			           prev_block = parm->ga_attrs; }
	   		       else
				   attr++;
			    %}
			    OPTIONAL
	}


CreateArgument [[P CMISCreateArg *]]
    %{ CMISParam *attr, *prev_block; int nblocks = 1; %}
    ::= SEQUENCE
    %{ if ((prev_block = attr = parm->ca_attrs = (CMISParam*)
			malloc(M_ATTRBLOCK * sizeof(CMISParam))) == NULL)
           return NOTOK;
       parm -> ca_nattrs = 0;
    %}
    {
    managedObjectClass	CHOICE
	{
	globalForm   [0] IMPLICIT OBJECT IDENTIFIER
			 [[O parm->ca_class.mid_global]]
			 %{ parm->ca_class.mid_type = MID_GLOBAL; %},
	localForm    [1] IMPLICIT INTEGER
			 [[i parm->ca_class.mid_local]]
			 %{ parm->ca_class.mid_type = MID_LOCAL; %},
	},

    instance CHOICE
	{
        managedObjectInstance	CHOICE
	    {
	    distinguishedName	[2] IMPLICIT DistinguishedName
				[[p &parm->ca_inst.mn_dn]]
				%{ parm->ca_inst.mn_type = MN_DN; %},
	    nonSpecificForm     [3] IMPLICIT OCTET STRING
				[[s parm->ca_inst.mn_nonspec]]
				%{ parm->ca_inst.mn_type = MN_NONSPEC; %},
	    localDistinguishedName  [4] IMPLICIT RDNSequence
				[[p &parm->ca_inst.mn_dn]]
				%{ parm->ca_inst.mn_type = MN_LOCALDN; %},
	    }
	    %{ parm -> ca_inst_type = CA_OBJECT_INST; %},

	superiorObjectInstance	[8] CHOICE
	    {
	    distinguishedName	[2] IMPLICIT DistinguishedName
				[[p &parm->ca_inst.mn_dn]]
				%{ parm->ca_inst.mn_type = MN_DN; %},
	    nonSpecificForm     [3] IMPLICIT OCTET STRING
				[[s parm->ca_inst.mn_nonspec]]
				%{ parm->ca_inst.mn_type = MN_NONSPEC; %},
	    localDistinguishedName  [4] IMPLICIT RDNSequence
				[[p &parm->ca_inst.mn_dn]]
				%{ parm->ca_inst.mn_type = MN_LOCALDN; %},
	    }
	    %{ parm -> ca_inst_type = CA_SUPERIOR_INST; %}
	} OPTIONAL,

    accessControl		[5] EXTERNAL [[p &parm -> ca_access]]
				OPTIONAL,
    referenceObjectInstance	[6] CHOICE
	{
	distinguishedName	[2] IMPLICIT DistinguishedName
				[[p &parm->ca_reference_inst.mn_dn]]
			%{ parm->ca_reference_inst.mn_type = MN_DN; %},
	nonSpecificForm         [3] IMPLICIT OCTET STRING
				[[s parm->ca_reference_inst.mn_nonspec]]
			%{ parm->ca_reference_inst.mn_type = MN_NONSPEC; %},
	localDistinguishedName  [4] IMPLICIT RDNSequence
				[[p &parm->ca_reference_inst.mn_dn]]
			%{ parm->ca_reference_inst.mn_type = MN_LOCALDN; %},
	} OPTIONAL,

    attributeList		[7] IMPLICIT SET OF Attribute
				[[p (CMISParam*) ga_pattrs]]
			        %{ *attr = *(CMISParam*) ga_pattrs;
			           free((char*) ga_pattrs);
				%{ if (++parm->ca_nattrs % M_ATTRBLOCK == 0 &&
					    (parm->ca_attrs = (CMISParam*)
					    realloc((char*) parm->ca_attrs,
					    (unsigned) ++nblocks*M_ATTRBLOCK*
					    sizeof(CMISParam))) == NULL)
				       return NOTOK;
	   			   if (parm->ca_attrs != prev_block) {
				       attr = parm->ca_attrs +
						(nblocks-1)*M_ATTRBLOCK;
				       prev_block = parm->ca_attrs; }
	   			   else
				       attr++;
				%}
				OPTIONAL
	}

%}
