/*
 * Copyright (c) 1994 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
** mtime - lightweight GeneralizedTime
*/

/*
** Saleem N. Bhatti
** May 1994
*/

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include "mtime.h"

#define ERROR(_s) \
    fprintf(stderr, "%s: errno %d - %s\n", _s, errno, strerror(errno))

int
#if defined(__STDC__) || defined(__cplusplus)
mTime2str(const MTime *mt, char *str)
#else
mTime2str(mt, str)
MTime *mt;
char *str;
#endif
{
    static
    char buffer[128];

    register
    char *bp;
    register
    struct tm *t;
    register
    int i;

    struct timeval tv;
    struct timezone tz;
    MTime m;

    if (mt == (MTime *) 0) {
        if (gettimeofday(&tv, &tz) != 0) {
#if defined(DEBUG)
            ERROR("mTime2str(): gettimeofday() failed");
#endif
            return -1;
        }

        t = localtime(&tv.tv_sec);

        m.year = t->tm_year;
        /*
        ** Hardwired hack ... struct tm has only a 2 digit year!
        */
        if (m.year < 100) {
            buffer[5];
            if (strftime(buffer, 5, "%Y", t) != 0)
                m.year += 1900;
            else
                sscanf(buffer, "%04d", &m.year);
        }
        m.month = t->tm_mon + 1;
        m.day = t->tm_mday;
        m.hours = t->tm_hour;
        m.minutes = t->tm_min;
        m.seconds = t->tm_sec;
        m.microseconds = tv.tv_usec;
#if	defined(SOLARIS) || defined(__HPUX__) || defined(__linux__)
	m.zone = timezone;	/* external variable in time.h */
#else
        m.zone = t->tm_gmtoff;
#endif
    }
    else
        memcpy((void *) &m, (void *) mt, sizeof(MTime));
/*
    sprintf(buffer, "%04d%02d%02d%02d%02d%02d.%06d%s%02d%02d",
            m.year, m.month, m.day,
            m.hours, m.minutes, m.seconds, m.microseconds,
#if	defined(SOLARIS) || defined(__HPUX__) || defined(__linux__)
            m.zone > 0 ? "-" : "+", abs(m.zone) / 3600, abs(m.zone) % 60);
#else
            m.zone < 0 ? "-" : "+", m.zone / 3600, m.zone % 60);
#endif
*/
    /* the format above is not understood by some other platforms */
    sprintf(buffer, "%04d%02d%02d%02d%02d%02d",
	    m.year, m.month, m.day,
	    m.hours, m.minutes, m.seconds);


#if 0
    bp = buffer + strlen(buffer);
    sprintf(bp, "%s%02d%02d",
            m.zone < 0 ? "-" : "+", m.zone / 3600, m.zone % 60);
#endif

    i = strlen(buffer);
    memcpy((void *) str, (void *) buffer, i);
    str[i] = '\0';

    return 0;
}

int
#if defined(__STDC__) || defined(__cplusplus)
str2mTime(const char *str, MTime *mt)
#else
str2mTime(str, mt)
char *str;
MTime *mt;
#endif
{
    int year, month, day, hours, minutes, seconds, microseconds = 0,
        z1 = -1, z2 = -1,
        i;
    register
    char *bp;

    if (mt == (MTime *) 0)
        return 0;

    memset((void *) mt, 0, sizeof(MTime));

    if (sscanf(str, "%04d%02d%02d%02d%02d%02d",
               &year, &month, &day, &hours, &minutes, &seconds) != 6)
        return -1;

    /* microseconds */
    if ((bp = strchr(str, '.')) != (char *) 0) {
        ++bp;
        if (sscanf(bp, "%06d", &microseconds) != 1)
            return -1;
    }

    /* time zone */
    if ((bp = strchr(str, '+')) != (char *) 0 ||
        (bp = strchr(str, '-')) != (char *) 0) {
        i = *bp == '+' ? 1 : -1;
        ++bp;
        if (sscanf(bp, "%02d%02d", &z1, &z2) != 2)
            return -1;
    }

    mt->year = year;
    mt->month = month;
    mt->day = day;
    mt->hours = hours;
    mt->minutes = minutes;
    mt->seconds = seconds;
    mt->microseconds = microseconds;
    if (z1 != -1 && z2 != -1)
        mt->zone = i * ((z1 * 3600) + (z2 * 60));
    else
        mt->zone = 0;

    return 0;
}

char *
mtime()
{
    static
    char buffer[80];

    return mTime2str((MTime *) 0, buffer) == 0 ? buffer : (char *) 0;
}
