//
// Copyright (c) 1994 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// K.M.T.McCarthy

#include <stdlib.h>
#include <signal.h>
#include "msap.h"
#include "Error.h"
#include "Coordinator.h"
#include "CmisAgent.h"
#include "Pma.h"
#include "SnmpImageMOClassInfo.h"
#include "SmiSntx.h"
#include "IQASntx.h"

extern  int watch ; // the SnmpV1 libraries packet dump flag

extern	int syntaxesBeingUsedByProxy ;  // Defined in IQASntx.c
					// Needed to indicate whether generic RDNs should be
					// printed in "dot" format or not.

char*		 service = "IQA";	// the logical application name

VoidFunc IQASyntaxes[] =
{
    smi_syntaxes,
    iqa_all_syntaxes,
    NULL
};


main (int argc, char** argv)
{
    // initialise the ASN.1 syntaxes
    if (initialiseSyntaxes(NULLCP, IQASyntaxes ) == NOTOK)
	error(NULLCP, "cannot initialise syntaxes");

    // instantiate the coordinator object
    Coordinator coordinator;

    // instantiate and initialise the CMIS agent object
    CMISAgent agent(1, argv, service);

    // Should we enable Snmp packet screen dumping ?
    if( argc > 1 )
    	++watch ;

    // parse the MO class information file..
    if( SnmpImageMOClassInfo::initialiseSnmpClassInfo( IQA_INPUT_FILE ) == NOTOK )
    	error(NULLCP, "snmpInputFile initialisation failed" ) ;

    // create the initial managed objects through the ETC/MIB_INIT_FILE
    if (MO::initialiseMIB(MIB_INIT_FILE) == NOTOK)
	error(NULLCP, "MIB initialisation failed");

    // perform the required addClass() calls for the SNMP MIB hierarchy...
    MIBInfo::setUpAddClassInfo() ;
    
    // Print generic RDNs in "dot" format..
    syntaxesBeingUsedByProxy = TRUE ;

    // start listening...
#ifdef DEBUG
    advise(NULLCP, "listening...");
#endif
    coordinator.listen();
}
