#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>

#ifdef  __GNUG__
#include <std.h>
#include <netdb-gnu.h>
#include <netinet/in-gnu.h>
typedef  char * malloc_t ;
#elif	__cplusplus
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#endif


#include "msap.h"
#include "Error.h"
#include "mtime.h"
#include "cmipsnmpProxy.h"
#include "cmipsnmpProxyAgent.h"
#include "ProxyKS.h"
#include "SnmpRMIBif.h"
#include "IQASntx.h"
#include "SnmpImageMOClassInfo.h"
#include "ucbext.h"
#include <sys/time.h>
#if !defined(SOLARIS) || defined(SUNCC4)
extern "C" { int gettimeofday(struct timeval*, struct timezone*); }
#endif


ProxyAlarmKS* ProxyAlarmKS::_instance  = NULL ;
static	cmipsnmpProxy* _cmipsnmpProxy  = NULL ;
static	OID	_cmipsnmpProxyAgentOID = NULLOID ;
static	OID	_iimcAutoObjAndAttrOID = NULLOID ;

void ProxyAlarmKS::closePorts( void )
{ 
struct PortDetails * portp ;

	for( int i=0; (portp = (struct PortDetails*)_portArrayp->get(i)) != NULLVD ; i++ ) {

	    stopListen(  portp->sdPort ) ;
	    (void)close( portp->sdPort ) ;
	    portp->sdPort = -1 ;
	}
}

int  ProxyAlarmKS::initialisePort( int trapUdpPort )
{
    struct sockaddr_in sin;
    int	sdPort ;

    // Are we already listening for this port ?
    struct PortDetails portDetails ;
    portDetails.sdPort = -1, portDetails.udpPort = trapUdpPort ;
    if( _portArrayp->find( &portDetails /*udp*/) != -1 ) // purely a udp port search
    	return OK ;

    bzero ((char *)&sin, sizeof sin );

    if ((sdPort = socket( AF_INET, SOCK_DGRAM, 0 )) < 0) {
	perror("ProxyAlarmKS.cc 1"); return NOTOK;
    }
    
    sin.sin_family = AF_INET;
    sin.sin_port   = trapUdpPort ;
    sin.sin_addr.s_addr = INADDR_ANY;

    if( bind( sdPort, (struct sockaddr*) &sin, sizeof(sin)) < 0) {
	perror("ProxyAlarmKS.cc 2"); return NOTOK;
    }

    listen( sdPort, 20 );

    startListen( sdPort );

    struct PortDetails* pdp = new PortDetails ;
    pdp->sdPort  = sdPort ;
    pdp->udpPort = trapUdpPort ;
    _portArrayp->add( (void*)pdp ) ;

    return OK;
}


int  ProxyAlarmKS::registerMO( MO* mo, char* moclass )
{
    if (++_nobjs > 1)
	return _nobjs;

    // first object
    if (strcmp(moclass, "cmipsnmpProxy") == 0)
	_cmipsnmpProxy = (cmipsnmpProxy*) mo;
    else {
	advise(NULLCP, "ProxyAlarmKS::registerMO: unknown class %s", moclass);
	return NOTOK;	// we do not expect anything else to register...
    }

    if( _cmipsnmpProxyAgentOID == NULLOID &&
    	(_cmipsnmpProxyAgentOID = name2oid( "cmipsnmpProxyAgent" )) == NULLOID ) {
	advise(NULLCP, "ProxyAlarmKS::registerMO: unknown OID for class %s", "cmipsnmpProxyAgent" );
	return NOTOK;
    }

    if( _iimcAutoObjAndAttrOID == NULLOID &&
    	(_iimcAutoObjAndAttrOID = name2oid( "iimcAutoObjAndAttr" )) == NULLOID ) {
	advise(NULLCP, "ProxyAlarmKS::registerMO: unknown OID for \"iimcAutoObjAndAttr\"" );
	return NOTOK ;
    }

    return 1 ;
}


int  ProxyAlarmKS::deregisterMO (MO*)
{
    if (_nobjs == 0) {
	advise(NULLCP, "ProxyAlarmKS::deregisterMO: no objects registered!");
	return NOTOK;
    }

    if (--_nobjs > 0)
	return _nobjs;

    // last object gone...
    closePorts();

    return 0;
}


int  ProxyAlarmKS::readCommEndpoint( int port )
{
    	// Read in the incoming trap..
	char 	trapbuffer[PROXYALARMTRAPPSLEN] ; 
	cmipsnmpProxyAgent * matchedPA = NULL ;
	struct type_PrxIIMCSntx_TransportAddresses * matchedAgentTA ;
	Bool 	multipleMatches = False ;

    	int nread = read( port, (char*)trapbuffer, PROXYALARMTRAPPSLEN );

	if( nread > 0 ) {
	
	    PE	   trappe = NULLPE ;
	    int	   portArrayIndex  ;
	    struct PortDetails portDetails, *pdp ; 
	    int	   recvUdpTrapPort ;

	    portDetails.sdPort = port, portDetails.udpPort = -1 ; // purely a socket port search

	    if( (portArrayIndex = _portArrayp->find( (void*)&portDetails /*sd*/ )) == NOTOK ||
  	    	(pdp = (struct PortDetails*)_portArrayp->get( portArrayIndex )) == NULL ||
		(recvUdpTrapPort = pdp->udpPort) == -1 ) {

		advise(NULLCP, "ProxyAlarmKS::readCommEndpoint(%d): call for unknown sd port", port );

	    } else if( mem2pe( &trappe, trapbuffer, nread ) == OK && trappe != NULLPE ) {

		struct type_SNMP_Snmp1157Msg* msg ;
		
		if( (msg = SnmpRMIBdecodeTrap( trappe )) != (struct type_SNMP_Snmp1157Msg*)0) {

		    struct type_SNMP_Trap__PDU * mdt = msg->data->un.trap ;

		    /* First check the agent trap community strings... */
		    MO * subordMO ;
		    OID  snmpUDPDomainOID = name2oid( "snmpUDPDomain" ) ;

		    for( subordMO = _cmipsnmpProxy->getSubordinate() ;
		    	 subordMO != NULLMO && multipleMatches == False ;
			 subordMO = subordMO->getPeer() ) {

			if( oid_equal( subordMO->getClass(), _cmipsnmpProxyAgentOID ) ) {

			    struct qbuf* recvNetAddressqb = mdt->agent__addr ;
			    struct type_PrxIIMCSntx_TransportAddresses * agentTA ;
	
			    /* Run through the transport addresses checking for the
			       one that emitted this trap.. */
			    for( agentTA =  (struct type_PrxIIMCSntx_TransportAddresses *)subordMO->getAttrVal( "transportAddresses" ) ;
			    	 agentTA != NULL && multipleMatches == False ;
				 agentTA =  agentTA->next ) {

				if( agentTA->member_PrxIIMCSntx_1->domain != NULLOID &&
				    oid_equal( agentTA->member_PrxIIMCSntx_1->domain, snmpUDPDomainOID ) &&
				    /* Comparing the network address component.. */
				    agentTA->member_PrxIIMCSntx_1->address != NULL &&
				    IpAddress_equal( agentTA->member_PrxIIMCSntx_1->address,
				    		     recvNetAddressqb                       ) ) {
				    
/* What about the port of the sender to get the transport address component ??????????????? */

				    /* Check the trap community string.. */
				    ProxyAgentTrapDetailsArray  * patdap ;
				    patdap = ((cmipsnmpProxyAgent*)subordMO)->getTrapDetails() ;

				    /* Confirm that the community string is correct.. */
				    struct ProxyAgentTrapDetails * trapDetailsp ;

				    for( int patdIndex=0 ; (trapDetailsp = (struct ProxyAgentTrapDetails *)patdap->get(patdIndex)) != NULLVD ; patdIndex++ ) {
				    
					if( trapDetailsp->udpTrapPort == recvUdpTrapPort &&
					    qb_cmp( msg->community, trapDetailsp->trapCommunityStringqb ) == 0 ) {

					    /* Found a full match ! */
					    if( matchedPA == NULL ) {
					
						matchedPA = (cmipsnmpProxyAgent*)subordMO ;
						matchedAgentTA = agentTA ;
						break ;

					    } else {
						/* > 1 proxy agent with the transport address
					           and trap community string details ! */
						multipleMatches = True ;
					    	break ;
					    }
					}
				    }
				}
			    }
			}
 		    }

		    if( matchedPA == NULL ) {
			char *cp ;
			advise( NULLCP, "WARNING: ProxyKS.cc: Incoming Trap could not be matched to a proxy agent..." );
			
			if( (cp = qb2IpAddress( mdt->agent__addr )) != NULLCP )
			    advise( NULLCP, "Trap IP Address %s", cp ) ;
			    
			if( (cp = qb2str(msg->community)) != NULLCP ) {
			    advise( NULLCP, "Trap Community String %s", cp ) ; (void)free(cp) ;
			}
		    } else {

			// ...at least one match

			OID eventType = formEventType( mdt ) ;
			PE  eventInfo = formEventInfo( mdt, eventType, (struct qbuf*)msg->community, 
						(multipleMatches==False) ? matchedPA : NULL, matchedAgentTA ) ;
			struct MName moMN ;
			moMN.mn_type = MN_DN ;

			if( multipleMatches == True ) {

			    /* Event report is to be emitted from the "cmipsnmpProxy" object.. */
			    moMN.mn_dn   = _cmipsnmpProxy->getDN() ;
			    Top::relayEventReport( _cmipsnmpProxy->getClass(), &moMN,
								mtime(), eventType, eventInfo ) ;
			} else {

			    /* Event report is to be emitted from the respective "cmipsnmpProxyAgent"
			       "internetSystem" object.. */
			    moMN.mn_dn = matchedPA->getDN() ;
			    Top::relayEventReport( matchedPA->getClass(), &moMN,
								mtime(), eventType, eventInfo ) ;
			}

			if( eventInfo != NULLPE )
			    pe_free( eventInfo ) ;

			if( eventType != NULLOID )
			    oid_free( eventType ) ;
			    
	    	    } /* else discard this trap */

		    if( snmpUDPDomainOID != NULLOID )
		    	oid_free( snmpUDPDomainOID ) ;

		    (void)free_SNMP_Snmp1157Msg( msg ) ;
		}
	    }
	}

	return OK;
}


OID
ProxyAlarmKS::formEventType( struct type_SNMP_Trap__PDU * mdt )
{
OID	eventTypeOID = NULLOID ;
int	gtrap ;

	switch( gtrap = mdt->generic__trap ) {

		case int_SNMP_generic__trap_coldStart:
			eventTypeOID = name2oid( "internetAlarmColdStart" ) ;
			break ;
		case int_SNMP_generic__trap_warmStart:
			eventTypeOID = name2oid( "internetAlarmWarmStart" ) ;
			break ;
		case int_SNMP_generic__trap_linkDown:
			eventTypeOID = name2oid( "internetAlarmLinkDown"  ) ;
			break ;
		case int_SNMP_generic__trap_linkUp:
			eventTypeOID = name2oid( "internetAlarmLinkUp"    ) ;
			break ;
		case int_SNMP_generic__trap_authenticationFailure:
			eventTypeOID = name2oid( "internetAlarmAuthenticationFailure" ) ;
			break ;
		case int_SNMP_generic__trap_egpNeighborLoss:
			eventTypeOID = name2oid( "internetAlarmEgpNeighbourLoss" ) ;
			break ;
		case int_SNMP_generic__trap_enterpriseSpecific:
			eventTypeOID = oid_cpy( mdt->enterprise ) ;
			eventTypeOID->oid_elements = (unsigned int*)realloc( (malloc_t)eventTypeOID->oid_elements,
							sizeof( unsigned int ) * (eventTypeOID->oid_nelem + 2 )) ;

			if( eventTypeOID->oid_elements == NULL ) {
			    oid_free( eventTypeOID ), eventTypeOID = NULLOID ;
			} else {
			    // IIMCIMIBTRANS 2.1.2.2 also RFC1452...
			    eventTypeOID->oid_elements[ (eventTypeOID->oid_nelem)++ ] = 0 ;
			    eventTypeOID->oid_elements[ (eventTypeOID->oid_nelem)++ ] = mdt->specific__trap ;
			}
			break ;
		default:
			advise( NULLCP, "ProxyAlarmKS.cc::formEventType - unknown generic trap number %d", gtrap ) ;
			break ;
	}
			
	return eventTypeOID ;
}


OID
ProxyAlarmKS::convertInternetOIDtoIIMCOID( OID internetOID )
{
OID	iimcOID = NULLOID ;

#ifdef DEBUG
	advise( NULLCP, "Converting internetOID %s", sprintoid( internetOID ) ) ;
#endif

	if( internetOID != NULLOID && _iimcAutoObjAndAttrOID != NULLOID ) {

	    if( (iimcOID = (OIDentifier*)calloc( 1, sizeof( OIDentifier ))) != NULLOID &&
	    	(iimcOID->oid_elements = (unsigned int*)malloc(
			sizeof( unsigned int ) * (internetOID->oid_nelem + _iimcAutoObjAndAttrOID->oid_nelem))) != NULL) {

		bcopy( (char*)_iimcAutoObjAndAttrOID->oid_elements,
		       (char*)iimcOID->oid_elements,
		       _iimcAutoObjAndAttrOID->oid_nelem * sizeof( unsigned int )) ;

	    	bcopy( (char*)internetOID->oid_elements,
		       (char*)&(iimcOID->oid_elements[ _iimcAutoObjAndAttrOID->oid_nelem ]),
		       internetOID->oid_nelem  * sizeof( unsigned int )) ;
		       
		iimcOID->oid_nelem = _iimcAutoObjAndAttrOID->oid_nelem +
							internetOID->oid_nelem ;
#ifdef DEBUG
		advise( NULLCP, "to %s", sprintoid( iimcOID ) ) ;
#endif
	    } else if( iimcOID != NULLOID ) {
	    	(void)free((char*)iimcOID );
		iimcOID = NULLOID ;
	    }
	}

	return iimcOID ;

}



//
// Note: the 'iimcVarBindNameOID' include appended SMI instance sub-identifiers - 
// 	 if an associated MO instance is found then we have the information needed
//	 to remove them, else they are left.
//
DN
ProxyAlarmKS::determineObjectInstanceDN(
	cmipsnmpProxyAgent* proxyAgentMO, OID iimcVarBindNameOID )
{
DN	objectInstanceDN = NULLDN ;
MO*	remoteSystemMO ;

	if( proxyAgentMO && (remoteSystemMO = proxyAgentMO->getRemoteSystem())) {

	    // Search the subtree for the IIMC attribute Id...

	    struct  ObjIdListVal * firstSupportedMIBlistp ;
	    SnmpImageMOClassInfo* imageInfop ;
	    Array 	imageClassArray(False) ;

	    if( (firstSupportedMIBlistp = proxyAgentMO->getMIBOIDs()) &&
		(imageInfop = MIBInfo::findSnmpImageMOClassInfoFromAttrOID( firstSupportedMIBlistp, 
					iimcVarBindNameOID, imageClassArray )) != NULLIMAGEINFO ) {

	 	DN instanceDN = dn_cpy( remoteSystemMO->getDN() ) ;
	    	Bool	useGenericRDNs = imageInfop->getMIBInfo()->getUseGenericRDNs() ;

		// Array is reverse ordered...
		for( int i=imageClassArray.getCount() - 1 ; 
			i >= 0 && (imageInfop = (SnmpImageMOClassInfo*)imageClassArray.get(i)) != NULLIMAGEINFO ;
 			--i ) {
#ifdef DEBUG		    
advise( NULLCP, "***%s", imageInfop->getClassName() ) ;
#endif
		    Bool multiInstanced = imageInfop->getIsMultiInstanced() ;
		    OID  namingAttrOID  = imageInfop->getAttrOid( 0 ) ;
		    char rdnBuffer1[2048], rdnBuffer2[2048] ;

		    if( useGenericRDNs == False ) {
		    
			if( multiInstanced == False ) {
		    	    strcpy( rdnBuffer2, "NULL" ) ;

			    // Lose the instance sub-identifier...
			    if( iimcVarBindNameOID -> oid_elements[ iimcVarBindNameOID -> oid_nelem - 1 ] == 0 )
			    	--iimcVarBindNameOID -> oid_nelem ; 
				
			} else {

			    rdnBuffer2[0] = '\0' ;
		    
			    // +1 to exclude the attribute specific sub-identifier...
			    for( int i=namingAttrOID->oid_nelem +1 ; i<iimcVarBindNameOID->oid_nelem; i++ ) {

				sprintf( rdnBuffer2, "%s%d.", rdnBuffer2, iimcVarBindNameOID->oid_elements[i] ) ;
			    }

			    if( rdnBuffer2[0] != '\0' )
				rdnBuffer2[ strlen(rdnBuffer2) - 1 ] = '\0' ; // remove the trailing "."

			    // Lose the instance sub-identifiers...
			    iimcVarBindNameOID -> oid_nelem -= 
			    	(iimcVarBindNameOID -> oid_nelem - (namingAttrOID->oid_nelem +1)) ;
			}

			sprintf( rdnBuffer1, "%sId=%s", imageInfop->getClassName(), rdnBuffer2 ) ;

			DN	objRdnDN ;
		    
			if( (objRdnDN = str2dn( rdnBuffer1 )) == NULLDN ) {
#ifdef DEBUG
			    advise( NULLCP, "ProxyAlarmKS::determineObjectInstanceDN() failed str2dn for %s", rdnBuffer1 ) ;
#endif
		    	    if( instanceDN != NULLDN ) {
				dn_free( instanceDN ), instanceDN = NULLDN ;
			    }

		    	    break ;
	
		    	} else {

		    	    if( instanceDN == NULLDN )
			    	instanceDN = objRdnDN ;
		 	    else
			    	dn_append( instanceDN, objRdnDN ) ;
			}
			
		    } else {

			RDN objRDN   = NULLRDN ;
			DN  objRdnDN = NULLDN ;
			char * rdncp = NULLCP, * nextrdncp = NULLCP,
			     * mocNamingAttrcp = imageInfop->getAttrName(0) ;
			
			if( multiInstanced == False ) {
				
			    if( (objRDN  = makeNullGenericRDN( mocNamingAttrcp )) != NULLRDN )
			    	objRdnDN = dn_comp_new( objRDN ) ;
			    
			    // Lose the instance sub-identifier...
			    if( iimcVarBindNameOID -> oid_elements[ iimcVarBindNameOID -> oid_nelem - 1 ] == 0 )
			    	--iimcVarBindNameOID -> oid_nelem ; 

			} else {

			    rdnBuffer2[0] = '\0' ;

			    // +1 to exclude the attribute specific sub-identifier...
			    for( int i=namingAttrOID->oid_nelem + 1; i<iimcVarBindNameOID->oid_nelem; i++ ) {

				sprintf( rdnBuffer2, "%s%d.", rdnBuffer2, iimcVarBindNameOID->oid_elements[i] ) ;
			    }

			    if( rdnBuffer2[0] != '\0' )
				rdnBuffer2[ strlen(rdnBuffer2) - 1 ] = '\0' ; // remove the trailing "."
			    
			    rdncp = rdnBuffer2 ;


			    short * attrIndexSyntaxes	   = imageInfop->getAttrIndexSyntaxList() ;
            		    int   * attrIndexFixedOctStrLenList = imageInfop->getAttrIndexFixedOctStrLenList() ;
			    int	    index=0, numIndices    = imageInfop->getAttrIndexNum() ;
			    struct  type_PrxSV1Sntx_SnmpGenericRDN *SGRDNp = NULL ;
	    
			    for( index=0 ; rdncp != NULLCP && strlen( rdncp ) > 0 &&
			    	    attrIndexSyntaxes != NULL && index<numIndices ; 
				    	++index, rdncp = nextrdncp ) {

				short attrIndexSyntax  = attrIndexSyntaxes[ index ] ;

			    	int attrIndexFixedOctStrLen =
					(attrIndexFixedOctStrLenList != (int*)NULL) ? 
					    attrIndexFixedOctStrLenList[ index ] : -1 ;

			    	if( formSnmpV1RDNcomponent( rdncp, &nextrdncp, &SGRDNp,
						attrIndexSyntax, attrIndexFixedOctStrLen ) == NOTOK ) {

				    break ;
				}
			    }

			    if( index != numIndices ) {
				advise( NULLCP, "ProxyKS.cc::determineObjectInstanceDN(): Trap instance sub-identifiers could not be parsed" ) ;
			    } else if( SGRDNp == NULL ||
			    	       (objRDN = makeGenericRDN( mocNamingAttrcp, SGRDNp )) == NULLRDN ) {
				advise( NULLCP, "ProxyKS.cc::determineObjectInstanceDN(): Failed to create generic Trap RDN" ) ;
			    } else if( objRDN != NULLRDN ) {
			    	objRdnDN = dn_comp_new( objRDN ) ;
			    }
			    
			    if( objRDN == NULLRDN && SGRDNp != NULL )
			    	SnmpGenericRDN_free( SGRDNp ) ; // Free a partially constructed RDN

			    // Lose the instance sub-identifiers...
			    iimcVarBindNameOID -> oid_nelem -= 
			    	(iimcVarBindNameOID -> oid_nelem - (namingAttrOID->oid_nelem +1)) ;
			}

			if( objRdnDN == NULLDN ) {
#ifdef DEBUG
			    advise( NULLCP, "ProxyAlarmKS::determineObjectInstanceDN() failed str2dn for %s", rdnBuffer1 ) ;
#endif
		    	    if( instanceDN != NULLDN ) {
				dn_free( instanceDN ), instanceDN = NULLDN ;
			    }

		    	    break ;

			} else {
		    	    if( instanceDN == NULLDN )
			    	instanceDN = objRdnDN ;
		 	    else
			    	dn_append( instanceDN, objRdnDN ) ;
			}
		    }
		}

		objectInstanceDN = instanceDN ; // May be NULL
	    }
	}
	
	return objectInstanceDN ;
}

PE
ProxyAlarmKS::formEventInfo( struct type_SNMP_Trap__PDU * mdt, OID eventTypeOID,
				struct qbuf* trapCommStringqb,
				cmipsnmpProxyAgent * proxyAgentMO,
 				struct type_PrxIIMCSntx_TransportAddresses * agentTAp )
{
PE	eventInfoPE = NULLPE ;

struct	type_PrxAlarmSntx_InternetAlarmInfo * IAIp ;

	if( (IAIp = (struct type_PrxAlarmSntx_InternetAlarmInfo*)calloc( 1, 
			sizeof( struct type_PrxAlarmSntx_InternetAlarmInfo ))) == NULL ||
	    (IAIp->probableCause = (struct type_PrxAlarmSntx_ProbableCause*)calloc( 1, sizeof(struct type_PrxAlarmSntx_ProbableCause))) == NULL ||
	    (IAIp->probableCause->un.globalValue = (OID)oid_cpy( eventTypeOID )) == NULLOID ) {

	    InternetAlarmInfo_free( IAIp ) ;
	    return NULLPE ;

	} else IAIp->probableCause->offset = type_PrxAlarmSntx_ProbableCause_globalValue ;


	/* Run through the variable-bindings list.. */
	struct type_SNMP_VarBindList * varBindListp ;
	struct type_SNMP_VarBind     * varBindp ;
	OID	varBindNameOID ;
	PE	varBindValuePE ;
	
	struct type_PrxAlarmSntx_AttributeIdentifierList ** currAttrIdListpp =
					&(IAIp->attributeIdList = NULL) ;
	struct type_PrxAlarmSntx_ObjectInstanceList ** currObjInstListpp =
					&(IAIp->objectInstanceList = NULL) ;
	struct type_PrxAlarmSntx_UnknownVarBindList ** currUnknownVBListpp = 
					&(IAIp->unknownVarBindList = NULL) ;
	struct type_PrxAlarmSntx_InternetTrapInfo ** currIntTrapInfopp =
					&(IAIp->internetTrapInfo = NULL) ;

	for( varBindListp = mdt->variable__bindings ; varBindListp != NULL ;
					varBindListp = varBindListp->next ) {

	    if( (varBindp = varBindListp->VarBind) != NULL &&
	    	(varBindNameOID = varBindp->name)  != NULL &&
		(varBindValuePE = varBindp->value) != NULLPE ) {

	    	/* Process the var bind name... */
		OID iimcVarBindNameOID ;
		if( (iimcVarBindNameOID = convertInternetOIDtoIIMCOID( varBindNameOID )) != NULLOID ) {

		    //
 		    // NOTE: iimcVarBindNameOID has the instance sub-identifiers appended e.g. '.0'
		    //

		    DN objectInstanceDN = NULLDN ;

		    currAttrIdListpp =
			appendAttributeIdListAttrId( currAttrIdListpp, iimcVarBindNameOID ) ;
			

		    /* Can we determine the corresponding object instance ?... */
		    if( proxyAgentMO == NULL ||
			    (objectInstanceDN = determineObjectInstanceDN( proxyAgentMO, iimcVarBindNameOID )) == NULLDN ) {

			// ...could not determine the respective object instances (4.1.4 IIMCIMIBTRANS.)

			// NOTE: iimcVarBindNameOID STILL has the instance sub-identifiers appended e.g. '.0',
			// since we do not have the object instance data to remove the identifiers !
	    
		 	// add an entry to the "UnknownVarBindList"...
			currUnknownVBListpp =
			    appendUnknownVarBindListVarBind( currUnknownVBListpp, iimcVarBindNameOID, varBindp->value ) ;		    		    

			// The InternetTrap details have an the object instance
			// as an OPTIONAL field...
			currIntTrapInfopp =
			    appendIntTrapInfo( currIntTrapInfopp, iimcVarBindNameOID, varBindp->value, objectInstanceDN ) ;

		    } else {

			// NOTE: iimcVarBindNameOID NO LONGER has the instance sub-identifiers appended !

			currObjInstListpp =
			    appendObjectInstList( currObjInstListpp, objectInstanceDN ) ;

			currIntTrapInfopp =
			    appendIntTrapInfo( currIntTrapInfopp, iimcVarBindNameOID, varBindp->value, objectInstanceDN ) ;

			dn_free( objectInstanceDN ), objectInstanceDN = NULLDN ;
		    }			

		    oid_free( iimcVarBindNameOID ) ;
		}
	    }
	}


	if( (IAIp->perceivedSeverity = (struct type_PrxAlarmSntx_PerceivedSeverity*)calloc(1,
			sizeof( struct type_PrxAlarmSntx_PerceivedSeverity ))) == NULL ) {

	    InternetAlarmInfo_free( IAIp ) ;
	    return NULLPE ;
	} else {

	    // Indeterminate - at least for now...
	    IAIp->perceivedSeverity->parm = int_PrxAlarmSntx_PerceivedSeverity_indeterminate ;
	}

	
	struct timeval  tp ;
	struct timezone tzp ;
	if( gettimeofday( &tp, &tzp ) == 0 ) {

	    if( (IAIp->notificationId = (struct type_PrxAlarmSntx_NotificationIdentifier*)calloc(1,
			sizeof( struct type_PrxAlarmSntx_NotificationIdentifier ))) == NULL ) {
		InternetAlarmInfo_free( IAIp ) ;
	    	return NULLPE ;

	    } else IAIp->notificationId->parm = tp.tv_sec ;

	} else IAIp->notificationId = NULL ; 


	// Not supported...
	IAIp->correlatedNot = (struct type_PrxAlarmSntx_CorrelatedNotifications*)NULL ;


	if( agentTAp != NULL ) {

	    if( (IAIp->transportDomain  = oid_cpy( agentTAp->member_PrxIIMCSntx_1->domain )) == NULLOID ||
	        (IAIp->transportAddress = qb_cpy(  agentTAp->member_PrxIIMCSntx_1->address )) == NULL ) {

		InternetAlarmInfo_free( IAIp ) ;
		return NULLPE ;
	    }

	} else {
	    IAIp->transportDomain  = NULLOID ;
	    IAIp->transportAddress = NULL ;
	}
	

	if( (IAIp->accessControlInfo = (struct type_PrxIIMCSntx_AccessControlInfo*)calloc(1,
						sizeof( struct type_PrxIIMCSntx_AccessControlInfo))) == NULL ||
	    (IAIp->accessControlInfo->un.communityString = qb_cpy( trapCommStringqb )) == NULL ) {

	    InternetAlarmInfo_free( IAIp ) ;
	    return NULLPE ;

	} else IAIp->accessControlInfo->offset = 
			type_PrxIIMCSntx_AccessControlInfo_communityString ;


	// No additional information (yet!)...
	IAIp->additionalInformation = (struct type_PrxAlarmSntx_AdditionalInformation*)NULL ;


	if( IAIp != NULL ) {
	    eventInfoPE = build_InternetAlarmInfo( IAIp ) ;
	    InternetAlarmInfo_free( IAIp ) ;
	}
	    
	return eventInfoPE ;
}
