//
// Copyright (c) 1994 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

/////////////////////////////////////////////////////////////////////////////////////////
// SnmpImageAccess.cc
/////////////////////////////////////////////////////////////////////////////////////////
//
// Modifications:
//
//	19-APR-94	[KMTM] In both refreshImageSubordinate and refreshImageSubordinates
//			the error response from a new SnmpImageMO must be modified
//			to be a processing failure to ensure that only legal Get errors
//			are reported.
//	19-APR-94	[KMTM] In refreshImageSubordinates if the correct number
//			of indices can not be retrieved then the default MO must
//			be deleted (instead of being carried over to the next loop.)
//
//	03-FEB-95	[KMTM] Caching inserted. Attribute value change inserted.
//
/////////////////////////////////////////////////////////////////////////////////////////

// K.M.T.McCarthy Nov-93
//

#include <stdlib.h>

#include "msap.h"
#include "Error.h"
#include "SnmpImageAccess.h"
#include "SnmpImageMO.h"
#include "SnmpImageMOClassInfo.h"
#include "IQASntx.h"

extern "C" {
	char* syntax2str( short ) ;
}

CMISErrors
SnmpImageAccess::refreshImageSubordinate( SnmpImageMOClassInfo * subordinateClassInfop, RDN subordinateRDN, AVA*& errorInfo )
{
CMISErrors		refreshMSAPerror = m_noError ;
cmipsnmpProxyAgent   * 	cmipsnmpProxyAgentp = getCmipsnmpProxyAgent() ;
SnmpRMIBAgent	     * 	snmpRMIBagentp = getSnmpRMIBAgent() ;
SnmpRMIBAgentResults	subordinateResults ;
SnmpRMIBerrorEnum	snmpRMIBerror = SnmpRMIBnoError ;
Bool	isMultiInstanced = subordinateClassInfop->getIsMultiInstanced() ;
	
	// The class may have just a naming attibute...
	if( isMultiInstanced == False && subordinateClassInfop->getNAttrs() == 1 ) {

	    // See if we already have this MO..
	    SnmpImageMO * subordinateMO;
	    for( subordinateMO = ((SnmpImageMO*)this)->getClassFirst( subordinateClassInfop ) ;
	    	 subordinateMO != NULL && rdn_cmp( subordinateMO->getRDN(), subordinateRDN ) != 0 ;
		 subordinateMO = (SnmpImageMO*)subordinateMO -> getClassNext( subordinateClassInfop ) )
		; // Empty loop body
	
	    if( subordinateMO == NULL ) {

		subordinateMO = new SnmpImageMO( rdn_cpy( subordinateRDN ), this, subordinateClassInfop, NULL, NULL,
			    					0, NULL, &refreshMSAPerror ) ;
		    					
		if( refreshMSAPerror != m_noError ) {
		    refreshMSAPerror = m_processingFailure ;
		    errorInfo = new AVA( m_processingFailure ) ;
		    delete subordinateMO ; // Serious error !
		}
	    }

	    return refreshMSAPerror ;
	}

	    
	// See if we already have this MO..
	SnmpImageMO * subordinateMO;
	for( subordinateMO = ((SnmpImageMO*)this)->getClassFirst( subordinateClassInfop ) ;
	     subordinateMO != NULL && rdn_cmp( subordinateMO->getRDN(), subordinateRDN ) != 0 ;
	     subordinateMO = (SnmpImageMO*)subordinateMO -> getClassNext( subordinateClassInfop ) )
	    ; // Empty loop body

	if( subordinateMO != NULL ) {
	    // Check if we should utilise caching...
	    timeval now;
	    if( gettimeofday(&now, NULL) == 0 ) {
	    	if( (now.tv_sec  - subordinateMO->_lastCachedTimeVal.tv_sec + 
	    	    (now.tv_usec - subordinateMO->_lastCachedTimeVal.tv_usec ) / 1000000) < 
		    		subordinateMO->_staleS )
		    return m_noError ;

	    	subordinateMO->_lastCachedTimeVal = now; // struct copy
	    }
	}
	

	if( isMultiInstanced == False )
	    snmpRMIBerror = snmpRMIBagentp->snmpGetSingleInstancedMO( 
					    	subordinateClassInfop, subordinateResults ) ;
	else 
	    snmpRMIBerror = snmpRMIBagentp->snmpGetOneMultiInstancedMO (
		    				subordinateClassInfop, subordinateResults, subordinateRDN ) ;
	
	if( snmpRMIBerror != SnmpRMIBnoError ) {

	    refreshMSAPerror = m_processingFailure ;

	    if( snmpRMIBerror == SnmpRMIBmultipleTimeouts )
		errorInfo = new AVA( m_processingFailure, "noResponseIIMC" ) ;
	    else
	    	errorInfo = new AVA( m_processingFailure ) ;

	} else {
		
	    if( subordinateMO == NULL ) {

	    	subordinateMO = new SnmpImageMO( rdn_cpy( subordinateRDN ), this, subordinateClassInfop, NULL, NULL,
			    					0, NULL, &refreshMSAPerror ) ;
		    					
		if( refreshMSAPerror != m_noError ) {
		    refreshMSAPerror = m_processingFailure ;
		    errorInfo = new AVA( m_processingFailure ) ;
		    delete subordinateMO ; // Serious error !
		}
	    }

	    // Skipping the naming attribute..
	    int nRetrievedAttrs = 0 ;

	    if( refreshMSAPerror == m_noError ) {
	    
	    	int	nAttrs = subordinateClassInfop->getNAttrs() ;
	

	    	subordinateResults.getFirst() ;
	    
		AttrNotifType * attrNotifTypeList = subordinateClassInfop->getAttrNotifList() ;

		List* triggerAVchangeList = new List;

	    	for( int i=1 ; i<nAttrs && refreshMSAPerror == m_noError ; i++ ) {
	 
	 	    PE resultPE = subordinateResults.snmpExtractResult( i, subordinateClassInfop, snmpRMIBagentp,
									NULL /*&refreshMSAPerror*/ ) ;
	
		    char * attrNamecp = subordinateClassInfop->getAttrName( i ) ;
		
		    Attr * attr = subordinateMO->getAttr( attrNamecp ) ;
	
		    Bool avChanged = False ;

		    if( attr ) {

		    	if( resultPE != NULLPE ) {

			    if( attrNotifTypeList != (AttrNotifType*)NULL &&
			    	attrNotifTypeList[ i ] == AttrValChangeNotif ) {

				PE oldPE = attr->getPE() ;

				if( oldPE != NULLPE && 
					pe_cmp( oldPE, resultPE ) != 0 ) {

				    avChanged = True ;
				}
			    }

			    attr->setPE( resultPE ) ;

			    if( avChanged == True ) {
				/* Prepase for an attribute value change report... */
				AttrValue* av = new AttrValue(i, 1);
				triggerAVchangeList->append(av) ;
			    }

			    ++nRetrievedAttrs ;

		    	} else {
		    	    attr->setstr( GENERICIQADEFAULTATTRVALUE ) ;
		    	}
		    }
	    	}

		// Do we have any attribute value changes to be reported ?
		if( triggerAVchangeList->getCount() > 0 ) {
		     ((Top*) subordinateMO)->triggerAttributeValueChange(
					triggerAVchangeList,
					int_SMI_SourceIndicator_resourceOperation);
		} else delete triggerAVchangeList ;
		
	    }

//	    if( refreshMSAPerror == m_noError && nRetrievedAttrs == 0 ) {
//	        refreshMSAPerror = m_processingFailure ;
//		errorInfo = new AVA( m_processingFailure ) ;
//	    }
	
	    if( refreshMSAPerror == m_noError && nRetrievedAttrs == 0 ) {

		if( subordinateMO )
		    delete subordinateMO ;

	    	// WE DO NOT SET UP AN MSAP ERROR AS THE GMS WILL DETERMINE THAT
		// THE OBJECT DOES NOT EXIST AND WILL GENERATE THE REQUIRED ERROR
	    }

	}

	subordinateResults.clear() ;
	
	return refreshMSAPerror ;
}

//
// Refresh the immediate subordinate(s).
//
// We may be refreshing single or multi-instanced MOs, with string or generic RDNs.
//
CMISErrors
SnmpImageAccess::refreshImageSubordinates( SnmpImageMOClassInfo * subordinateClassInfop, AVA*& errorInfo )
{
CMISErrors		refreshMSAPerror = m_noError ;
cmipsnmpProxyAgent   * 	cmipsnmpProxyAgentp = getCmipsnmpProxyAgent() ;
SnmpRMIBAgent	     * 	snmpRMIBagentp = getSnmpRMIBAgent() ;
SnmpRMIBAgentResults	subordinateResults ;
SnmpRMIBerrorEnum	snmpRMIBerror = SnmpRMIBnoError ;
Bool 			useGenericRDNs ;
char *			mocNamingAttrcp ;
Bool	isMultiInstanced = subordinateClassInfop->getIsMultiInstanced() ;

	useGenericRDNs   = MIBInfo::getUseGenericRDNs() ;

	mocNamingAttrcp  = subordinateClassInfop->getAttrName( 0 ) ;

	// The class may have just a naming attibute...
	if( isMultiInstanced == False && subordinateClassInfop->getNAttrs() == 1 ) {

	    RDN    	subordinateRDN 	= NULLRDN ;
	    char	rdnBuffer[1024] ;

	    if( useGenericRDNs == False ) {
		// "Xid=NULL"...
		sprintf( rdnBuffer, "%s=NULL", mocNamingAttrcp ) ;
		subordinateRDN = str2rdn( rdnBuffer ) ;

	    } else {
		subordinateRDN = makeNullGenericRDN( mocNamingAttrcp ) ;
	    }

	    if( subordinateRDN != NULLRDN ) {

		// See if we already have this MO..
		SnmpImageMO * subordinateMO;
		for( subordinateMO = ((SnmpImageMO*)this)->getClassFirst( subordinateClassInfop ) ;
	    	     subordinateMO != NULL && rdn_cmp( subordinateMO->getRDN(), subordinateRDN ) != 0 ;
		     subordinateMO = (SnmpImageMO*)subordinateMO -> getClassNext( subordinateClassInfop ) )
		    ; // Empty loop body
	
		if( subordinateMO == NULL ) {

		    subordinateMO = new SnmpImageMO( subordinateRDN, this, subordinateClassInfop, NULL, NULL,
			    					0, NULL, &refreshMSAPerror ) ;
		    					
		    if( refreshMSAPerror != m_noError ) {
			refreshMSAPerror = m_processingFailure ;
			errorInfo = new AVA( m_processingFailure ) ;
			delete subordinateMO ; // Serious error !
		    }

		} else rdn_free( subordinateRDN ) ;
	    }

	    return refreshMSAPerror ;
	}

	
	if( isMultiInstanced == False )
	    snmpRMIBerror = snmpRMIBagentp->snmpGetSingleInstancedMO( 
						subordinateClassInfop, subordinateResults ) ;
	else 
	    snmpRMIBerror = snmpRMIBagentp->snmpGetMultiInstancedMOs (
						subordinateClassInfop, this, subordinateResults ) ;

	if( snmpRMIBerror != SnmpRMIBnoError ) {

	    refreshMSAPerror = m_processingFailure ;

	    if( snmpRMIBerror == SnmpRMIBmultipleTimeouts )
		errorInfo = new AVA( m_processingFailure, "noResponseIIMC" ) ;
	    else
	    	errorInfo = new AVA( m_processingFailure ) ;

	} else {

	    SnmpImageMO * subordinateMO = getClassFirst( subordinateClassInfop );

	    char 	rdnBuffer[1024], *resetPtInRdnBuffercp ;

	    // Buffer the naming attributes across..
	    sprintf( rdnBuffer, "%s=", mocNamingAttrcp ) ;
	    resetPtInRdnBuffercp = (strchr( rdnBuffer, '=' ) + 1) ;

	
	    short * attrIndexSyntaxes	   = subordinateClassInfop->getAttrIndexSyntaxList() ;
	    int   * attrIndexFixedOctStrLenList = subordinateClassInfop->getAttrIndexFixedOctStrLenList() ;
	    int	    attrIndexNum           = subordinateClassInfop->getAttrIndexNum() ;

	    for( subordinateResults.getFirst() ;
	    	 subordinateResults.getNresultAttrs() > 0 && refreshMSAPerror == m_noError ;
		 subordinateResults.nextMOCinstanceResults() ) {

		RDN    subordinateRDN = NULLRDN ;

		// Prepare those RDN..
		if( isMultiInstanced == True ) {

		    // "Xid=a.b.c.d"...
		    if( subordinateResults.appendRDNstrToBuffer(subordinateClassInfop, rdnBuffer) == NOTOK ) {

			advise( NULLCP, "Failed to retrieve the RDN for the.." ) ;
			advise( NULLCP, "..class %s", subordinateClassInfop->getClassName() );
			refreshMSAPerror = m_processingFailure ;
			errorInfo = new AVA( m_processingFailure ) ;
			break ; // Serious error !
		    }

		    if( useGenericRDNs == True ) {

			// Extract the sub-identifiers off index attribute by index attribute...

			struct  type_PrxSV1Sntx_SnmpGenericRDN *SGRDNp = NULL ;
			char * rdncp = strchr( rdnBuffer, '=' ), * nextrdncp = NULLCP ;
			
			if( rdncp ) ++rdncp ; // skip the '='

			int	attrIndex = 0 ;
			
			for( attrIndex=0 ; rdncp != NULLCP && strlen( rdncp ) > 0 &&
				attrIndexSyntaxes != NULL  && attrIndex < attrIndexNum ; 
				    	++attrIndex, rdncp = nextrdncp ) {

			    short attrIndexSyntax  = attrIndexSyntaxes[ attrIndex ] ;

			    int attrIndexFixedOctStrLen = 
			    	    (attrIndexFixedOctStrLenList != (int*)NULL) ? 
					attrIndexFixedOctStrLenList[ attrIndex ] : -1 ;

			    if( formSnmpV1RDNcomponent( rdncp, &nextrdncp, &SGRDNp,
							attrIndexSyntax,
 							attrIndexFixedOctStrLen ) == NOTOK ) {
				refreshMSAPerror = m_processingFailure ;
				errorInfo = new AVA( m_processingFailure ) ;
				break ; // Serious error !
			    }
			}

			if( attrIndex != attrIndexNum ) {

			    advise( NULLCP, "SnmpImageAccess.cc::refreshImageSubordinates() instance sub-identifiers could not be parsed for class %s", 
			    				subordinateClassInfop->getClassName() ) ;
			    refreshMSAPerror = m_processingFailure ;
			    errorInfo = new AVA( m_processingFailure ) ;

			} else if( SGRDNp == NULL ||
			    	   (subordinateRDN = makeGenericRDN( mocNamingAttrcp, SGRDNp )) == NULLRDN ) {

			    advise( NULLCP, "SnmpImageAccess.cc::refreshImageSubordinates() Failed to create generic RDN" ) ;
			    refreshMSAPerror = m_processingFailure ;
			    errorInfo = new AVA( m_processingFailure ) ;
			}

			if( subordinateRDN == NULLRDN && SGRDNp != NULL )
			    SnmpGenericRDN_free( SGRDNp ) ; // Free a partially constructed RDN

			if( refreshMSAPerror != m_noError )
			    break ; // Serious error !

		    } else {

		    	// string based RDN..
			subordinateRDN = str2rdn( rdnBuffer ) ;
		    }

		} else {
		
		    // SINGLE INSTANCED

		    if( useGenericRDNs == True ) {
		    
		    	if( subordinateMO == NULL )
		            subordinateRDN = makeNullGenericRDN( mocNamingAttrcp ) ;

		    } else {

		    	// "Xid=NULL"...
			if( subordinateMO == NULL ) {
			    strcat( rdnBuffer, "NULL" ) ;
			    subordinateRDN = str2rdn( rdnBuffer ) ;
			}
		    }
		}

		// Do we need to create a new MO ?
		if( subordinateMO == NULL ) {

		    subordinateMO  = new SnmpImageMO( subordinateRDN, this, subordinateClassInfop, NULL, NULL, 
		    					0, NULL, &refreshMSAPerror ) ;
		    subordinateRDN = NULLRDN ;

		    if( refreshMSAPerror != m_noError ) {
		    	refreshMSAPerror = m_processingFailure ;
			errorInfo = new AVA( m_processingFailure ) ;
		    	delete subordinateMO ; // Serious error !
			subordinateMO = NULL ;
			break ; // Serious error !
		    }
		}


		int	nAttrs = subordinateClassInfop->getNAttrs() ;

		// Skipping the naming attribute index..
		int	nRetrievedAttrs = 0 ;
		struct  type_PrxSV1Sntx_SnmpGenericRDN *SGRDNp = NULL ;

		AttrNotifType * attrNotifTypeList = subordinateClassInfop->getAttrNotifList() ;

		List* triggerAVchangeList = new List;

		for( int i=1 ; i<nAttrs && refreshMSAPerror == m_noError ; i++ ) {

		    PE resultPE = subordinateResults.snmpExtractResult( i, subordinateClassInfop, snmpRMIBagentp,
									NULL /*&refreshMSAPerror*/ ) ;

		    char * attrNamecp = subordinateClassInfop->getAttrName( i ) ;
				
		    Attr * attr = subordinateMO->getAttr( attrNamecp ) ;

		    Bool avChanged = False ;

		    if( attr ) {
		    
		    	if( resultPE != NULLPE ) {

			    if( attrNotifTypeList != (AttrNotifType*)NULL &&
			 	    attrNotifTypeList[ i ] == AttrValChangeNotif ) {

			    	PE oldPE = attr->getPE() ;

			    	if( oldPE != NULLPE && 
				    	pe_cmp( oldPE, resultPE ) != 0 ) {

				    avChanged = True ;
			    	}
			    }

			    attr->setPE( resultPE ), resultPE = NULLPE ;

			    if( avChanged == True ) {
				/* Prepase for an attribute value change report... */
				AttrValue* av = new AttrValue(i, 1);
				triggerAVchangeList->append(av) ;
			    }

			    ++nRetrievedAttrs ;

			} else attr->setstr( GENERICIQADEFAULTATTRVALUE ) ;

		    } else {

			advise( NULLCP, "WARNING SnmpImageAccess.cc: Failed to find the attribute %s", attrNamecp ) ;
			if( resultPE != NULLPE ) {
			    pe_free( resultPE ), resultPE = NULLPE ;
			}
		    }
		}

		// Do we have any attribute value changes to be reported ?
		if( triggerAVchangeList->getCount() > 0 ) {
		     ((Top*) subordinateMO)->triggerAttributeValueChange(
					triggerAVchangeList,
					int_SMI_SourceIndicator_resourceOperation);
		} else delete triggerAVchangeList ;


	    	if( refreshMSAPerror == m_noError && nRetrievedAttrs == 0 ) {

		    advise( NULLCP, "WARNING SnmpImageAccess.cc: Failed to retrieve ANY attributes for MOC %s",
		    				subordinateClassInfop->getClassName() ) ;
//	            refreshMSAPerror =  m_processingFailure ;
//		    errorInfo = new AVA( m_processingFailure ) ;
		}

		if( refreshMSAPerror == m_noError && isMultiInstanced == True ) {

		    if( subordinateRDN != NULLRDN &&
			    rdn_cmp( subordinateMO->getRDN(), subordinateRDN ) ) {

			subordinateMO->setRDN( subordinateRDN );
			subordinateRDN = NULLRDN ;
		    }
		}

	    	if( subordinateRDN != NULLRDN )
		    rdn_free( subordinateRDN ) ;

		*resetPtInRdnBuffercp   = '\0' ; // reset the buffer

		if( subordinateMO != NULL )
		    subordinateMO = (SnmpImageMO*)subordinateMO->getClassNext( subordinateClassInfop ) ;
	    }

	    if( isMultiInstanced == True && refreshMSAPerror == m_noError ) {

		// Tidy up any surplus MOs..
		for( SnmpImageMO *currMO = subordinateMO, *nextMO ;
		     			currMO != NULL ; currMO = nextMO ) {

		    nextMO = (SnmpImageMO*)currMO->getClassNext( subordinateClassInfop ) ;
		    delete currMO ;
		}
	    }
	}

	subordinateResults.clear() ;

	return refreshMSAPerror ;
}


//
// "this" points to the superior of the first object to be retrieved..
//
SnmpImageMO *
SnmpImageAccess::getClassFirst( SnmpImageMOClassInfo* MOCimageInfop )
{
SnmpImageMO * firstMOforMOCinThisMIB = NULL, * currMO ;
OID	classOID ;

	if( MOCimageInfop != NULLIMAGEINFO ) {

	    classOID = MOCimageInfop->getClass() ;
	    
	    for( currMO =  (SnmpImageMO*)this->getSubordinate() ;
		 currMO != NULL ;
		 currMO =  (SnmpImageMO*)currMO->getPeer() )
	    {
	    	SnmpImageMOClassInfo* imageInfop = currMO->getClassInfo() ;

		if( imageInfop == NULLIMAGEINFO ) {

		    break ; // serious error exit
		    
		} else if( oid_equal( imageInfop->getClass(), classOID ) ) {

		    firstMOforMOCinThisMIB = currMO ;
		    break ;
		}
	    }
	}

	return firstMOforMOCinThisMIB ;
}


//
//
//
SnmpImageMO *
SnmpImageAccess::getClassNext( SnmpImageMOClassInfo* MOCimageInfop )
{
SnmpImageMO * nextMOforMOCinThisMIB = NULL, * currMO ;
OID	classOID ;

	if( MOCimageInfop != NULLIMAGEINFO ) {

	    classOID = MOCimageInfop->getClass() ;
	    
	    for( currMO =  (SnmpImageMO*)this->getPeer() ;
		 currMO != NULL ;
		 currMO =  (SnmpImageMO*)currMO->getPeer() )
	    {
	    	SnmpImageMOClassInfo* imageInfop = currMO->getClassInfo() ;

		if( imageInfop == NULLIMAGEINFO ) {

		    break ; // serious error exit
		    
		} else if( oid_equal( imageInfop->getClass(), classOID ) ) {

		    nextMOforMOCinThisMIB = currMO ;
		    break ;
		}
	    }
	}

	return nextMOforMOCinThisMIB ;
}


// 
// Check that the remoteAgent is not currently locked...
//
Bool	
SnmpImageAccess::isRemoteAgentLocked( int& status, CMISErrors& msapError )
{
	MO*  cspaMO = getCmipsnmpProxyAgent() ;
	AdminStateVal *lockingAttrValp =  (AdminStateVal*)cspaMO->getAttrVal( "administrativeState" ),
 		       lockedAttrVal = as_locked ;
	Bool locked = False ;

	if(lockingAttrValp && !AdministrativeState_cmp( lockingAttrValp, &lockedAttrVal ) ) {

	    status	= NOTOK ; 
	    msapError 	= m_processingFailure ;
	    locked   	= True ;   
	}

	return locked ;
}

Bool	
SnmpImageAccess::isRemoteAgentLocked( int& status, AVA*& errorInfo )
{
	MO*  cspaMO = getCmipsnmpProxyAgent() ;
	AdminStateVal *lockingAttrValp =  (AdminStateVal*)cspaMO->getAttrVal( "administrativeState" ),
 		       lockedAttrVal = as_locked ;
	Bool locked = False ;

	if(lockingAttrValp && !AdministrativeState_cmp( lockingAttrValp, &lockedAttrVal ) ) {

	    status	= NOTOK ; 
	    errorInfo 	= new AVA( m_processingFailure, "adminStateLockedIIMC" ) ;
	    locked   	= True ;   
	}

	return locked ;
}


//
// Delete any subordinate objects..
//
SnmpImageAccess::~SnmpImageAccess()
{
MO	*deleteMO, *subordMO ;

	for( subordMO = this->getSubordinate() ; subordMO != NULLMO ; ) {

	    deleteMO = subordMO ;
	    subordMO = subordMO->getPeer() ;
	    
	    delete deleteMO ;
	}
}
	
