//
// Copyright (c) 1994 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

#if !defined(_SnmpImageMO_cc_)
#define _SnmpImageMO_cc_

#if defined(__GNUG__)
#include <std.h>
typedef	char * malloc_t ;
#elif defined(__cplusplus)
#include <stdlib.h>
#endif

#include "msap.h"
#include "SnmpImageMO.h"
#include "SnmpImageAccess.h"
#include "Error.h"
#include "IQASntx.h"

extern "C" {
	char * syntax2str( short ) ;
}


SnmpImageMO::SnmpImageMO( RDN rdn, MO *superior, SnmpImageMOClassInfo* imageInfop, 
				MO *ref, Bool * refSuppliedAttrIds, 
				int nAttrs, CMISParam *attrs, 
				CMISErrors *err, CMISErrorInfo *errInfo, OID * eventId )
: SnmpImageAccess( rdn, superior )
{
	/* Initialise Class */

	int 	nAttrIds = imageInfop->getNAttrs(), anyTypeAttrId ;
	short * Syntaxes = imageInfop->getAttrSyntaxes();
	CMISErrors localError = m_noError ;

	// Create an AnyType for each MOC attribute..
	for( anyTypeAttrId=0 ; anyTypeAttrId<nAttrIds ; anyTypeAttrId++ ) {

	    if( anyTypeAttrId == 0 ) {
	    	if( MIBInfo::getUseGenericRDNs() == FALSE )
		    _attrs[anyTypeAttrId] = new AnyType( Syntaxes[anyTypeAttrId], 
		    						(char*)rdn->rdn_av.av_struct );
		else
		    _attrs[anyTypeAttrId] = new AnyType( Syntaxes[anyTypeAttrId], 
		    	(void*)SnmpGenericRDN_cpy( (struct type_PrxSV1Sntx_SnmpGenericRDN*)rdn->rdn_av.av_struct ));
	    } else {
		_attrs[anyTypeAttrId] = new AnyType( Syntaxes[anyTypeAttrId], NULLVD ) ;
		_attrs[anyTypeAttrId] -> setstr( GENERICIQADEFAULTATTRVALUE ) ;
	    }
	}

	if( ref != NULLMO && refSuppliedAttrIds != NULL ) {

	    // Run through the reference object supplied attributes..

	    for( int anyTypeAttrId=1 ; anyTypeAttrId<nAttrIds ; anyTypeAttrId++ ) {
	    
	    	if( refSuppliedAttrIds[anyTypeAttrId] == True ) {
		    Attr* refAttrp = ref->getAttr( imageInfop->getAttrName( anyTypeAttrId )) ;

		    if( refAttrp != NULLATTR )
		    	_attrs[ anyTypeAttrId ]->set( refAttrp->copy() ) ;
		}
	    }
	}
	
	// Run through the supplied attributes..
	for (int i=0 ; i<nAttrs ; i++) {

	    localError = m_noError ;

	    if( attrs[i].mp_id.mid_type != MID_GLOBAL )
	    	localError = m_noSuchAttribute;

	    else {

		anyTypeAttrId = imageInfop->getAttrId( attrs[i].mp_id.mid_global ) ;
    	    
		if( anyTypeAttrId < 0 || anyTypeAttrId >= nAttrIds ) {
		    localError = m_noSuchAttribute;

		} else if( anyTypeAttrId != 0 ) {
 		    // ..skipping the RDN case

		    // Buffer the new value..
		    localError = _attrs[anyTypeAttrId]->set( attrs[i].mp_val ) ;
		}
	    }
	    
    	    // Tidy up or report error information..
	    if( err != NULL)
	    	*err = localError ;
	    
	    switch ((int)localError) {

            	case (int) m_noError :
		    mid_free(&attrs[ i ].mp_id);
		    break;

		case (int) m_processingFailure :
	      	case (int) m_noSuchAttribute :
		    if( errInfo )
		    	errInfo->ei_noSuchAttribute = attrs[ i ].mp_id;
		    return; 

		case (int) m_invalidAttributeValue:
		    if( errInfo ) {
           	    	errInfo -> ei_invalidAttributeValue.mp_id  = attrs[ i ].mp_id ;
            	    	errInfo -> ei_invalidAttributeValue.mp_val = attrs[ i ].mp_val ;
		    }
            	    return;
	    }    
        }


	// Buffer the three frequently used class pointers..
	_imageInfop = imageInfop ;
	setCmipsnmpProxyAgent( ((SnmpImageMO*)superior)->getCmipsnmpProxyAgent() ) ;
	setSnmpRMIBAgent( ((SnmpImageMO*)superior)->getSnmpRMIBAgent() ) ;

	registerClass(nAttrIds, _attrs, imageInfop, True);

	if (checkNameBinding() == NOTOK) {
#if defined(DEBUG)
	    advise( NULLCP, "SnmpImageMO::SnmpImageMO() : checkNameBinding() failed\n");
#endif
	    if (err != NULL) *err = m_invalidObjectInstance;

	    return;
	}

	return ;
}

int
SnmpImageMO::createRR( AVA*& crErr, void* crTypep, int )
{
int	status	= OK ;

	return status ;
}

int
SnmpImageMO::deleteRR( AVA*& err, void* info, Bool checkOnly, int )
{
int	status	= OK ;

	return status ;
}

MOClassInfo*
SnmpImageMO::getMOClassInfo( void )
{
static	OID	    	       buffClassOID   = NULLOID ;
static	SnmpImageMOClassInfo * buffClassInfop = NULLIMAGEINFO ;
OID	operClassOID = MO::getOperationClass() ;

	if( operClassOID != NULLOID ) {
	
	    // Do we have the image info pointer buffered already ?
	    if( (buffClassOID == NULLOID) || !oid_equal( buffClassOID, operClassOID ) ) {
	    	
		if( (buffClassInfop = MIBInfo::getSnmpImageMOClassInfo( operClassOID )) == NULLIMAGEINFO ) {

		    advise( NULLCP, "SnmpImageMO.cc::getMOClassInfo() - Failed to find Image Class Info for the class %s",
sprintoid( operClassOID )) ;
		    
		} else {
		
		    if( buffClassOID != NULLOID )
		    	oid_free( buffClassOID ) ;

		    buffClassOID = oid_cpy( operClassOID ) ;
		}
	    }

	} else {
	    /* else should NEVER occur !! */
	    advise( NULLCP, "FATAL error : SnmpImageMO::getMOClassInfo() could not access the current operation's class OID" ) ;
	}

	return (MOClassInfo*)buffClassInfop ;
}


//
// Note: duplicate MOs will have already be checked for by the GMS.
//
MO *
SnmpImageMO::cmisCreate(RDN rdn, MO *superior,
            MO *ref, int nAttrs, CMISParam *attrs,
            CMISErrors *err, CMISErrorInfo *errInfo, OID *event )
{
int	status ;
MO *	newMO    = NULLMO ;
OID 	classOID = SnmpImageMO::getOperationClass() ;
struct 	ObjIdListVal * firstSupportedMIBlistp ;
cmipsnmpProxyAgent   * cmipsnmpProxyAgentp = ((SnmpImageAccess*)superior)->getCmipsnmpProxyAgent() ;
CMISErrors createMSAPerror = m_noError ;
SnmpImageMOClassInfo* imageInfop = NULLIMAGEINFO ;

	*err = m_processingFailure ;

	// Check that the remoteAgent is not currently locked...
	if( ((SnmpImageAccess*)superior)->isRemoteAgentLocked( status, createMSAPerror ) == True ) {
	    if( err ) *err = createMSAPerror ;
	    return NULLMO ;
	}

	if( cmipsnmpProxyAgentp &&
	    (firstSupportedMIBlistp = cmipsnmpProxyAgentp->getMIBOIDs()) ) {

	    if( (imageInfop == NULLIMAGEINFO) /* we do not have the info already */ && 
	    	(imageInfop = MIBInfo::getSnmpImageMOClassInfo( firstSupportedMIBlistp, classOID )) == NULLIMAGEINFO )

		advise( NULLCP, "Failed to find Image Class Info for the class %s", sprintoid( classOID )) ;

	    else {

		int	MOCnAttrs = imageInfop->getNAttrs() ;
		Bool *	refSuppliedAttrIds = (Bool*)NULL ;

		if( ref != NULLMO ) {

		    refSuppliedAttrIds = new Bool[ MOCnAttrs ] ;

		    refSuppliedAttrIds[0] = False ; // The naming attribute

		    for( int i=1 ; i<MOCnAttrs ; i++ )
			refSuppliedAttrIds[i] = True ;

		    // Check which attributes have been supplied as CMIS parameters..
		    for( i=1 ; i<nAttrs ; i++ ) {
			refSuppliedAttrIds[imageInfop->
					getAttrId( attrs[i].mp_id.mid_global )] = False ;
		    }
		}

		// Form an array of attribute Ids and PEs that are to be applied..
		int	nextIndex=0 ;
		int * 	attrIdp   = (int*)malloc( sizeof(int) * MOCnAttrs ) ;
		PE  *	createPEp = (PE*)malloc( sizeof(PE) * MOCnAttrs ) ;
			
		for( int i=1 ; i<MOCnAttrs ; i++ ) {
		     // ..skipping the naming attribute

		    if( ref != NULLMO && refSuppliedAttrIds[i] == True ) {
			// required PE must be obtained from the reference MO..
			attrIdp[ nextIndex ] = i ;

			SnmpImageMOClassInfo * imageInfop = ((SnmpImageMO*)ref)->getClassInfo() ;
			Attr* refAttrp = ref->getAttr( imageInfop->getAttrName(i) ) ;
			createPEp[  nextIndex ] = refAttrp->getPE() ; // DO NOT FREE
			++nextIndex ;
			
		    } else {
			    
			// find the cmisCreate parameter attribute for this Id..
			for( int j=0; j<nAttrs ; j++ ) {
			    
			    if( imageInfop->getAttrId(attrs[j].mp_id.mid_global) == i ) {

				attrIdp[ nextIndex ] = i ;
				createPEp[ nextIndex ] = attrs[j].mp_val ; // DO NOT FREE
				++nextIndex ;
				break ;
			    }
			}
		    }
		}
			
		// Attempt to create the new object at the remote SNMP agent..
		SnmpRMIBAgent * snmpRMIBagentp = ((SnmpImageMO*)superior)->getSnmpRMIBAgent() ;
		SnmpRMIBAgentResults	createResults ;
			
		if( snmpRMIBagentp != NULL && nextIndex > 0 &&
			    snmpRMIBagentp->snmpCreateEntry( rdn, attrIdp, createPEp, nextIndex,
							imageInfop, createResults ) == SnmpRMIBnoError ) {
			
		    /* Make the MO */
		    newMO = new SnmpImageMO(rdn, superior, imageInfop, ref, refSuppliedAttrIds, 
				  	  			nAttrs, attrs, err, errInfo, event );

		    if( newMO != NULLMO )
			*err = m_noError ;
			
		    Bool isMultiInstanced = imageInfop->getIsMultiInstanced() ;
			
		    if( newMO != NULLMO && isMultiInstanced == False )
			imageInfop->setFirst( newMO );
		}

		createResults.clear() ;
		(void)free((char*) createPEp ) ;
		(void)free((char*) attrIdp ) ;
	    }
	}
    
	return( (MO *) newMO );
}


CMISErrors
SnmpImageMO::cmisDelete()
{ 
CMISErrors	deleteMSAPerror = m_processingFailure ;
int	status ;
int*	deleteInfop = _imageInfop->getAttrDeleteInfo() ;
DeleteType	deleteMask = DeleteContainedObjects ;

	if( isRemoteAgentLocked( status, deleteMSAPerror ) == True ) {
	    return deleteMSAPerror ;
	}

	if( deleteInfop &&
	    ((*deleteInfop & deleteMask) != 0 )) {

	    // Request the deletion of the remote SNMP object..
	    int    deleteAttrId = _imageInfop->getAttrDeleteAttrId() ;
	    Attr * deleteAttr   = _imageInfop->getAttrDeleteAttr() ;

	    if( deleteAttr != NULL && deleteAttrId > 0 && deleteAttrId < _imageInfop->getNAttrs() ) {

	    	SnmpRMIBAgent * snmpRMIBagentp = getSnmpRMIBAgent() ;
		SnmpRMIBAgentResults	deleteResults ;

		PE deletePE = deleteAttr->getPE() ; // DO NOT FREE THIS PE !

		if( snmpRMIBagentp && deletePE != NULLPE &&
		    snmpRMIBagentp->snmpSetMOattributes( this, &deleteAttrId, &deletePE, 1,
						_imageInfop, deleteResults ) == SnmpRMIBnoError ) {

 		    deleteResults.getFirst() ;

		    if( deleteResults.getNresultAttrs() > 0 ) {

	    		// delete request was successful...
		    	delete this ; // SnmpImageMO is no more !

		    	deleteMSAPerror = m_noError ;
		    }
		}

		deleteResults.clear() ;
	    }
	}

	return( deleteMSAPerror );
}


int
SnmpImageMO::set( CMISModifyOp setOperMode, int attrId, int classLevel,
					void* setValuep, AVA*& errorInfo, Bool, int )
{
int		status = NOTOK ;
CMISErrors      setMSAPerror = m_processingFailure ;
Bool		deleteAttempted = False ;
static		int  numAttrsBuffered = 0, maxArraySize = 0 ;
static		int* attrIdp = NULL ;
static		PE*  setPEp  = NULL ;

	if( isRemoteAgentLocked( status, errorInfo ) == True ) {
	    return NOTOK ;
	}
	
	if( maxArraySize == 0 ) {

	    // initialise buffer storage..
	    maxArraySize = 20 ;

	    attrIdp = (int*)malloc( sizeof(int) * maxArraySize ) ;
	    setPEp  = (PE*)malloc(  sizeof(PE)  * maxArraySize ) ;
	}

	if( attrId == MO_NOMOREATTRS && classLevel == MO_ALLEVELS ) {

	    // We must flush any pending remote agent sets..
	
	    if( numAttrsBuffered > 0 ) {

		// We can now perform the remote request...
	    	SnmpRMIBAgent * snmpRMIBagentp = getSnmpRMIBAgent() ;
		SnmpRMIBAgentResults	setResults ;
		    
		if( snmpRMIBagentp && attrIdp != NULL && setPEp != NULL &&
		    snmpRMIBagentp->snmpSetMOattributes( this, attrIdp, setPEp, numAttrsBuffered,
							_imageInfop, setResults ) == SnmpRMIBnoError ) {

		    setResults.getFirst() ;

		    Bool setSucceeded = True ;
		    PE   retrievedPE ;
		    
	    	    // Check that all required attributes have been set successfully..
		    for( int i=0 ; i < numAttrsBuffered ; i++ ) {

		    	if( (retrievedPE = setResults.snmpExtractResult( attrIdp[i], _imageInfop, snmpRMIBagentp,
							NULL /*&refreshMSAPerror*/ )) == NULLPE ) {

			    setSucceeded = False ;
			    break ;

			} else pe_free( retrievedPE ) ;
		    }

		    if( setSucceeded == True ) {
			setMSAPerror = m_noError ;
		    }
		    
		} else setMSAPerror = m_processingFailure ;

		setResults.clear() ;

		// Free static storage PEs..
		for( int i=0 ; i < numAttrsBuffered ; i++ ) {
		    pe_free( setPEp[i] ), setPEp[i] = NULLPE ;
		}
	    }

	    numAttrsBuffered = 0 ;
	
	} else if( setOperMode != m_addValue && setOperMode != m_removeValue && setOperMode != m_setToDefault ) {

	    // ...m_replace assumed

	    GetReplaceType replaceMask = Replace ;

	    int * getReplaceBitsp ;

	    // Check that the attribute has the required replace property..
	    if( (getReplaceBitsp = _imageInfop->getAttrReplaceInfo()) != NULL &&
	    	((getReplaceBitsp[attrId] & replaceMask) != 0) ) {

		char *  attrNamecp = _imageInfop->getAttrName( attrId ) ;
		Attr *  setAttr    = getAttr( attrNamecp ) ;

		// check that we are not attempting to delete an entry..
		if( _imageInfop->getAttrDeleteAttrId() == attrId ) {

		    // Are we about to request for a delete value to be applied..
		    Attr * deleteAttr = _imageInfop->getAttrDeleteAttr() ;

		    if( deleteAttr != NULL ) {

			if( !deleteAttr->compare( setValuep ) ) {

			    // ..attempt to delete other than by cmisDelete..
			    setMSAPerror = m_invalidOperation ; // IIMC requirement
			    
			    deleteAttempted = True ;
			}
		    }
		}

		if( deleteAttempted == False ) {

		    PE setPE ;

		    if( attrIdp != NULL && setPEp != NULL && 
		    		(setPE = getAttrSetPE()) != NULLPE ) {

		    	if( numAttrsBuffered == maxArraySize ) {

			    // extend the static storage buffers..
		    	    maxArraySize+= 10 ;
			    attrIdp = (int*)realloc( (malloc_t)attrIdp, sizeof(int) * maxArraySize ) ;
			    setPEp  = (PE*)realloc(  (malloc_t)setPEp , sizeof(PE)  * maxArraySize ) ;
		        }

		    	if( attrIdp != NULL && setPEp != NULL ) {

		    	    attrIdp[numAttrsBuffered] = attrId ;
		    	    setPEp[numAttrsBuffered]  = pe_cpy( setPE ) ; 
			
			    ++numAttrsBuffered ;
		    	    setMSAPerror = m_noError ;
			}
		    }
		}
	    }
	    
	} else setMSAPerror = m_invalidOperation ; // IIMC requirement

	if( setMSAPerror != m_noError )
	    errorInfo = new AVA( setMSAPerror ) ;
	else
	    status = OK ;
	
	return status ;
}

//
//
//
int
SnmpImageMO::refreshSubordinate( RDN subordinateRDN, AVA*& errorInfo, int )
{
CMISErrors	refreshMSAPerror = m_noError ;
int		status = OK ;

	if( isRemoteAgentLocked( status, errorInfo ) == True )
	    return status ;

	char * subordinateRdnAtcp = AttrT2str( subordinateRDN->rdn_at ) ;

	if( subordinateRdnAtcp != NULLCP ) {

	    /* Extract off the trailing "Id"... */
	    *(subordinateRdnAtcp + strlen( subordinateRdnAtcp ) - 2) = '\0' ;
	    
	    OID subordinateClassOID = name2oid( subordinateRdnAtcp ) ;
		
	    cmipsnmpProxyAgent   * cmipsnmpProxyAgentp = getCmipsnmpProxyAgent() ;
	    struct  ObjIdListVal * firstSupportedMIBlistp ;

	    if( subordinateClassOID != NULLOID && cmipsnmpProxyAgentp && _imageInfop &&
			(firstSupportedMIBlistp = cmipsnmpProxyAgentp->getMIBOIDs())) {

		SnmpImageMOClassTraversal peerTraversalState ;
		SnmpImageMOClassInfo* subordinateClassInfop 	 = NULLIMAGEINFO ;
		SnmpImageMOClassInfo* firstSubordinateClassInfop = NULLIMAGEINFO ; 

		if( (firstSubordinateClassInfop = _imageInfop->getFirstSubordImageInfo() ) == NULLIMAGEINFO ) {
		    // ...no subordinates for this MIB

		    /* Check for subordinates in alternative MIBs... */
		    firstSubordinateClassInfop = SnmpImageMOClassInfo::getNextPeerImageInfo(
							NULLIMAGEINFO, _imageInfop->_mibInfo,
		     					firstSupportedMIBlistp, peerTraversalState, getClass() ) ;
		}

		for( subordinateClassInfop =  firstSubordinateClassInfop /* may still be NULL */ ;
		     subordinateClassInfop != NULLIMAGEINFO ;
		     subordinateClassInfop =  SnmpImageMOClassInfo::getNextPeerImageInfo(
						subordinateClassInfop->_peer,
						subordinateClassInfop->_mibInfo,
		     				firstSupportedMIBlistp, peerTraversalState, getClass() ) ) {

		    if( oid_equal( subordinateClassInfop -> getClass(), subordinateClassOID ) )
		    	break ;
		}
			 
		if( subordinateClassInfop != NULLIMAGEINFO ) {
			     
		    refreshMSAPerror = refreshImageSubordinate( subordinateClassInfop, subordinateRDN, errorInfo ) ;
		}
	    }

	    if( subordinateClassOID != NULLOID )
	    	oid_free( subordinateClassOID ) ;

	    (void)free( subordinateRdnAtcp ) ;
	}

	if( refreshMSAPerror != m_noError ) {
	    status    = NOTOK ;
	}

	return status ;
}

int
SnmpImageMO::refreshSubordinates( AVA*& errorInfo, int )
{
CMISErrors	refreshMSAPerror = m_noError ;
cmipsnmpProxyAgent   * cmipsnmpProxyAgentp = getCmipsnmpProxyAgent() ;
SnmpRMIBAgent	     * snmpRMIBagentp = getSnmpRMIBAgent() ;
struct  ObjIdListVal * firstSupportedMIBlistp ;
int	status = OK ;

	if( isRemoteAgentLocked( status, errorInfo ) == True ) {
	    return status ;
	}

	if( cmipsnmpProxyAgentp && snmpRMIBagentp && _imageInfop &&
		(firstSupportedMIBlistp = cmipsnmpProxyAgentp->getMIBOIDs()) ) {

	    SnmpImageMOClassTraversal peerTraversalState ;
	    SnmpImageMOClassInfo * subordinateClassInfop      = NULLIMAGEINFO ;
	    SnmpImageMOClassInfo * firstSubordinateClassInfop = NULLIMAGEINFO ; 

	    if( (firstSubordinateClassInfop = _imageInfop->getFirstSubordImageInfo() ) == NULLIMAGEINFO ) {
		// ...no subordinates for this MIB

		/* Check for subordinates in alternative MIBs... */
		firstSubordinateClassInfop = SnmpImageMOClassInfo::getNextPeerImageInfo(
							NULLIMAGEINFO, _imageInfop->_mibInfo,
		     					firstSupportedMIBlistp, peerTraversalState, getClass() ) ;
	    }

	    for( subordinateClassInfop =  firstSubordinateClassInfop /* may still be NULL */ ;
		 subordinateClassInfop != NULLIMAGEINFO ;
		 subordinateClassInfop =  SnmpImageMOClassInfo::getNextPeerImageInfo( 
					    subordinateClassInfop->_peer,
					    subordinateClassInfop->_mibInfo,
				 	    firstSupportedMIBlistp, peerTraversalState, getClass() ) ) {
			     
		if( (refreshMSAPerror = refreshImageSubordinates( subordinateClassInfop, errorInfo )) != m_noError )
		    break ;
	    }
	}

	if( refreshMSAPerror != m_noError ) {
	    status    = NOTOK ;
	}

	return status ;
}

#endif /* !_SnmpImageMO_cc_ */
