//////////////////////////////////////////////////////////////////////////////////////////
// SnmpImageMOClassInfo, MIBInfo and SnmpRMIBAgentClassInfo class methods
//////////////////////////////////////////////////////////////////////////////////////////
// Created 		02-Jun-93
//
// Modifications	DD-MMM-YY - ""
//			11-Oct-93 - "Code to set up new state item _isMultiInstanced."
//			09-Nov-93 - "Code to handle "tooBig" data."
//			10-Nov-93 - "Corrections to the initialisation of
//				     the static variables _attrByAgentIncStorageByNagents
//				     and _rootMIBInfo"
//////////////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#ifdef  __GNUG__
typedef	char * malloc_t ;
#endif

#include "msap.h"
#include "Error.h"
#include "CmisAgent.h"
#include "SnmpImageMOClassInfo.h"
#include "SnmpImageMO.h"

extern "C" {
	char * syntax2str( short ) ;
}


// static variable storage
OID SnmpImageMOClassInfo::_iimcAutoObjAndAttrOID = NULLOID ;


const int SnmpRMIBAgentClassInfo::_attrByAgentIncStorageByNagents = 3;


MIBInfo * MIBInfo::_rootMIBInfo    = NULLMIBINFO ;
char    * MIBInfo::_internetSystemClassName = NULLCP ;
char    * MIBInfo::_internetSystemIdName = NULLCP ;
Bool	  MIBInfo::_useGenericRDNs = False ;


// MIBInfo class constructor

MIBInfo::MIBInfo( OID mibOID )
{
	_MIBidentifier    = mibOID ;

	_MIBsuperiorClass = NULLOID ;
	
	_peerMIBInfo      = NULLMIBINFO ;

	_subordinateImageClass = NULLIMAGEINFO ;

	linkInMIB() ;

	return ;
}

MIBInfo::~MIBInfo()
{
	if( _MIBidentifier != NULLOID )
	    oid_free( _MIBidentifier ) ;

	if( _MIBsuperiorClass != NULLOID )
	    oid_free( _MIBsuperiorClass ) ;

	if( _internetSystemClassName != NULLCP )
	    (void)free( _internetSystemClassName ) ;

	if( _internetSystemIdName != NULLCP )
	    (void)free( _internetSystemIdName ) ;
}


//
// Link in this MIB's MIBInfo object into the current linked list
//
void	
MIBInfo::linkInMIB( void )
{
MIBInfo *	currMIBInfo ;

	if( _rootMIBInfo == NULLMIBINFO )
	{
	    _rootMIBInfo = this ;

	} else {

	    for( currMIBInfo = _rootMIBInfo ; currMIBInfo -> _peerMIBInfo != NULLMIBINFO ; 
		    					currMIBInfo = currMIBInfo -> _peerMIBInfo );

	    currMIBInfo -> _peerMIBInfo = this ;
	}

	return ;
}

//
// Initiate the construction for the MIBInfo subordinate SnmpImageMOClassInfo objects
//
int	MIBInfo::constructContainment( SnmpImageMOClassInfo * imageClassInfo )
{
int	status = OK ;

	status = imageClassInfo -> constructContainment() ;

	return( status ) ;
}

//
// Test routine to recursively traverse (depth first) the subordinate and 
// peer SnmpImageMOClassInfo objects
//
void 
MIBInfo::recursiveDisplaySubordinates( SnmpImageMOClassInfo *currInfo )
{
int	i, nAttrs ;
char *	cp ;

	if( currInfo == NULLIMAGEINFO ) 
	    return ;

	nAttrs = currInfo -> getNAttrs();

	advise( NULLCP, "Class name %s", currInfo -> getClassName() ) ;
	    
	if( currInfo -> _superior != NULLIMAGEINFO )
	    advise( NULLCP, "superior %s", currInfo -> _superior -> getClassName());

	if( currInfo -> _peer != NULLIMAGEINFO )
	    advise( NULLCP, "peer %s", currInfo -> _peer -> getClassName());

	if( currInfo -> _subordinate != NULLIMAGEINFO )
	    advise( NULLCP, "subordinate %s", currInfo -> _subordinate -> getClassName());

	// Display input file information...
	if( currInfo -> _attrSyntaxes && currInfo -> _attrGetReplaceInfo && 
							currInfo -> _attrMatchesForInfo )
	    for( i=0 ; i<nAttrs ; i++ ) {
	    	advise( NULLCP, "Attribute Id %d\tSyntax %s\tGetReplace %d\tMatches For %d", i,
			(cp = syntax2str( currInfo -> _attrSyntaxes[i] )) == NULLCP ? "UNKNOWN SYNTAX" : cp ,
			currInfo -> _attrGetReplaceInfo[i], currInfo -> _attrMatchesForInfo[i] );

		if( currInfo -> _attrFixedOctStrLenList && currInfo -> _attrFixedOctStrLenList[ i ] != -1 )
		    	advise( NULLCP, "attribute octet string syntax fixed length %d", currInfo -> _attrFixedOctStrLenList[ i ] ) ;
	    }
		

	if( currInfo -> _attrCreateInfo && currInfo -> _attrDeleteInfo )
	    advise( NULLCP, "Create %d\tDelete %d",
	    		currInfo -> _attrCreateInfo[0], currInfo -> _attrDeleteInfo[0] ) ;

	advise( NULLCP, "MOC is multi-instanced %s", 
		currInfo -> _isMultiInstanced == True ? "TRUE" : "FALSE" ) ;

	if( currInfo -> _attrIndexSyntaxList )
	    for( i=0 ; i < currInfo -> _nAttrIndexNum ; i++ )
	    	advise( NULLCP, "INDEX attribute syntax %d", currInfo -> _attrIndexSyntaxList[ i ] ) ;

	if( currInfo -> _attrDeleteAttrId != -1 )
	    advise( NULLCP, "DELETEATT Attribute Id %d value %s",
	    		currInfo -> _attrDeleteAttrId, currInfo -> _attrDeleteAttrValuecp ) ;

	advise( NULLCP, "--------------------" ) ;

	if( currInfo -> _subordinate != NULLIMAGEINFO )
	    recursiveDisplaySubordinates( currInfo -> _subordinate ) ;

	if( currInfo -> _peer != NULLIMAGEINFO )
	    recursiveDisplaySubordinates( currInfo -> _peer ) ;
	
	return ;
}
		
//
// Find the MIB containment hierarchy for the specified MIB identifier
//

MIBInfo * 
MIBInfo::findMIBInfo( OID requiredMIBidentifier )
{
MIBInfo * currMIBInfo, * foundMIBInfo = NULLMIBINFO ;

	for( currMIBInfo = _rootMIBInfo ; currMIBInfo != NULLMIBINFO ; currMIBInfo = currMIBInfo -> _peerMIBInfo )
	{
	    if( oid_equal( currMIBInfo -> _MIBidentifier, requiredMIBidentifier ) ) {

	    	foundMIBInfo = currMIBInfo ;
		break ;
	    }
	}

	return( foundMIBInfo ) ;
}


//
// Test method for dumping the multi-MIB containment hierarchy.
//
void
MIBInfo::printContainment( void )
{
MIBInfo	* mibInfo, * foundMIBInfo ;
SnmpImageMOClassInfo * currInfo ;
OID	testMIBidentifier = NULLOID ;

	for( mibInfo = this ; mibInfo != NULLMIBINFO ; mibInfo = mibInfo -> _peerMIBInfo )
	{
	    if( mibInfo -> _MIBidentifier != NULLOID )
		advise( NULLCP, "New MIB %s", oid2name( mibInfo -> _MIBidentifier, OIDPART ) ) ;
	    else
		advise( NULLCP, "ERROR : MIB identifier undefined" ) ;

	    if( mibInfo -> _MIBsuperiorClass != NULLOID )
		advise( NULLCP, "Superior MIB class %s", oid2name( mibInfo -> _MIBsuperiorClass, OIDPART ) ) ;
	    else
		advise( NULLCP, "ERROR : Superior MIB class identifier undefined" ) ;

	    currInfo = mibInfo -> getSubordinateImageClass() ;

	    recursiveDisplaySubordinates( currInfo ) ;

	    advise( NULLCP, "ENDOFMIBENDOFMIBENDOFMIBENDOFMIB" ) ;

	    testMIBidentifier = mibInfo -> _MIBidentifier ;
	}

	// perform a test on the findMIBInfo() method..
	if( (foundMIBInfo = findMIBInfo( testMIBidentifier )) != NULLMIBINFO )
	    advise( NULLCP, "Found MIBidentifier %s", oid2name( foundMIBInfo -> _MIBidentifier, OIDPART )) ;

	return ;
}


//
// Must be called after MO:initialiseMIB to ensure that the getClassArray()
// pointer has been initialised.
//
void
MIBInfo::recursiveSetUpAddClassInfo( SnmpImageMOClassInfo *currInfo )
{
	if( currInfo == NULLIMAGEINFO ) 
	    return ;

	// Add the new class can be added to the global list..
	MOClassArray* mocArrayClassp = MOClassInfo::getClassArray();

#ifdef	DEBUG
advise( NULLCP, "addClass %s", currInfo->getClassName() ) ;
#endif

	mocArrayClassp->addClass( 
	    		currInfo -> getClassName(),
			SnmpImageMO::getMOClassInfo ) ;

	if( currInfo -> _subordinate != NULLIMAGEINFO )
	    recursiveSetUpAddClassInfo( currInfo -> _subordinate ) ;

	if( currInfo -> _peer != NULLIMAGEINFO )
	    recursiveSetUpAddClassInfo( currInfo -> _peer ) ;
	
	return ;
}

void
MIBInfo::setUpAddClassInfo( void )
{
MIBInfo * mibInfo ;
SnmpImageMOClassInfo * currInfo ;

	for( mibInfo = MIBInfo::getRootMIBInfo() ; mibInfo != NULLMIBINFO ; mibInfo = mibInfo -> _peerMIBInfo )
	{
	    if( mibInfo -> _MIBidentifier == NULLOID )
		advise( NULLCP, "ERROR : MIB identifier undefined" ) ;

	    if( mibInfo -> _MIBsuperiorClass == NULLOID )
		advise( NULLCP, "ERROR : Superior MIB class identifier undefined" ) ;

	    currInfo = mibInfo -> getSubordinateImageClass() ;

	    recursiveSetUpAddClassInfo( currInfo ) ;
	}

	return ;
}

//
// We only require those MIBInfo nodes that have the required _MIBsuperiorClass.
//
MIBInfo * 
MIBInfo::getFirstMIBInfo( ObjIdListVal * firstSupportedMIBlistp, 
				int& nthSupportedMIBchecked, OID superiorClass )
{
MIBInfo * currMIBInfop = NULLMIBINFO, *foundMIBInfop = NULLMIBINFO ;

	nthSupportedMIBchecked = -1 ;

	for( ObjIdListVal * currSupportedMIBp = firstSupportedMIBlistp ;
	     currSupportedMIBp != (ObjIdListVal*)0 ;
	     currSupportedMIBp = currSupportedMIBp->oidl_next ) {

	     ++nthSupportedMIBchecked ;

	     if( (currMIBInfop = findMIBInfo( currSupportedMIBp->oidl_oid )) != NULLMIBINFO &&
	     	oid_equal( currMIBInfop->_MIBsuperiorClass, superiorClass ) ) {

		foundMIBInfop = currMIBInfop ;
		break ;
	     }
	}

	return foundMIBInfop ;
}


//
// We only require thos MIBInfo nodes that have the required _MIBsuperiorClass.
//
MIBInfo * 
MIBInfo::getNextPeerMIBInfo( ObjIdListVal * firstSupportedMIBlistp,
			     int& nthSupportedMIBchecked, OID superiorClass )
{
MIBInfo * currMIBInfop = NULLMIBINFO, *foundMIBInfop = NULLMIBINFO ;
ObjIdListVal * currSupportedMIBp = firstSupportedMIBlistp ;

	// Move to the first supportedMIB that has not be checked already..
	for( int i = -1 ; i<nthSupportedMIBchecked && currSupportedMIBp!= (ObjIdListVal*)0 ; i++ )
	    currSupportedMIBp = currSupportedMIBp->oidl_next ;

	for( ;
	     currSupportedMIBp != (ObjIdListVal*)0 ;
	     currSupportedMIBp = currSupportedMIBp->oidl_next ) {

	     ++nthSupportedMIBchecked ;

	     if( (currMIBInfop = findMIBInfo( currSupportedMIBp->oidl_oid )) != NULLMIBINFO &&
		 oid_equal( currMIBInfop->_MIBsuperiorClass, superiorClass ) ) {
		foundMIBInfop = currMIBInfop ;
		break ;
	     }
	}

	return foundMIBInfop ;
}


//
// The following method traverses the MIBInfo and SnmpImageMOClassInfo
// containment hierarchies searching for the particular SnmpImageMOClassInfo
// node that represents the given MIB and class details.
//

SnmpImageMOClassInfo * 
MIBInfo::getSnmpImageMOClassInfo( OID MIBoid, OID classOID, Bool mustFindInfo )
{
SnmpImageMOClassInfo * foundImageInfo = NULLIMAGEINFO ;
MIBInfo	* foundMIBInfo = NULLMIBINFO ;

	if( (foundMIBInfo = findMIBInfo( MIBoid )) == NULLMIBINFO )
	    advise( NULLCP, "Error: Failed to find Image Info node for MIB  %s", oid2name( MIBoid, OIDPART )) ;

	else if( (foundImageInfo = foundMIBInfo -> getSubordinateImageClass ()
						-> findImageInfo( classOID ) ) == NULLIMAGEINFO &&
		 mustFindInfo == True )
	    advise( NULLCP, "Error: Failed to find Image Info node for class %s", oid2name( classOID, OIDPART )) ;

	return( foundImageInfo ) ;
}

//
// The following method traverses supplied supported MIBs mib OIDs and then
// requests its name sake to search individual MIBs for the required class OID.
//

SnmpImageMOClassInfo * 
MIBInfo::getSnmpImageMOClassInfo( struct ObjIdListVal* firstSupportedMIBlistp, OID classOID )
{
SnmpImageMOClassInfo * foundImageInfo = NULLIMAGEINFO ;
MIBInfo	* foundMIBInfo = NULLMIBINFO ;

	if( firstSupportedMIBlistp != (ObjIdListVal*)0 && classOID != NULLOID ) {

	    ObjIdListVal* currMIBlistp ;

	    // run through the supportedMIBs linked list..
	    for( currMIBlistp = firstSupportedMIBlistp ;
		 currMIBlistp != (ObjIdListVal*)0 && foundImageInfo == NULLIMAGEINFO ;
		 currMIBlistp = currMIBlistp->oidl_next ) {

		if( currMIBlistp->oidl_oid != NULLOID )
		    foundImageInfo = getSnmpImageMOClassInfo( currMIBlistp->oidl_oid, classOID, False ) ;
	    }

	    if( foundImageInfo == NULLIMAGEINFO )
		advise( NULLCP, "Error: Failed to find Image Info node for class %s", oid2name( classOID, OIDPART )) ;
	}

	return( foundImageInfo ) ;
}

//
// The following method traverses all MIBs and then
// requests its name sake to search individual MIBs for the required class OID.
//

SnmpImageMOClassInfo * 
MIBInfo::getSnmpImageMOClassInfo( OID classOID )
{
SnmpImageMOClassInfo * foundImageInfo = NULLIMAGEINFO ;
OID	mibOID = NULLOID ;

	if( classOID != NULLOID ) {

	    for( MIBInfo * mibInfop = MIBInfo::getRootMIBInfo() ;
	    	 mibInfop != NULLMIBINFO && foundImageInfo == NULLIMAGEINFO ;
		 mibInfop  = mibInfop -> _peerMIBInfo ) {

		if( (mibOID = mibInfop->getMIBidentifier()) != NULLOID )
		    foundImageInfo = getSnmpImageMOClassInfo( mibOID, classOID, False ) ;

	    }

	    if( foundImageInfo == NULLIMAGEINFO )
		advise( NULLCP, "Error: Failed to find Image Info node for class %s", oid2name( classOID, OIDPART )) ;
	}

	return( foundImageInfo ) ;
}


//
// Attempt to find an the specified attribute Id within the classes for
// the specified MIB...
//
SnmpImageMOClassInfo * 
MIBInfo::findSnmpImageMOClassInfoFromAttrOID( struct ObjIdListVal* firstSupportedMIBlistp,
						OID MIBoid, OID iimcAttrOID, Array& imageClassArray )
{
SnmpImageMOClassInfo * foundImageInfo = NULLIMAGEINFO ;
MIBInfo	* foundMIBInfo = NULLMIBINFO ;

	if( (foundMIBInfo = findMIBInfo( MIBoid )) == NULLMIBINFO )
	    advise( NULLCP, "Failed to find Image Info node for MIB  %s", oid2name( MIBoid, OIDPART )) ;

	else {

	    int attrOidIndex = -1 ;

	    foundImageInfo = foundMIBInfo -> getSubordinateImageClass ()
				-> findImageInfoForAttrOID( firstSupportedMIBlistp, 
						iimcAttrOID, imageClassArray, attrOidIndex ) ;
	}

	return( foundImageInfo ) ;
}

//
// Attempt to find an the specified attribute Id within the classes for
// the specified MIBs...
//
SnmpImageMOClassInfo * 
MIBInfo::findSnmpImageMOClassInfoFromAttrOID( struct ObjIdListVal* firstSupportedMIBlistp, OID iimcAttrOID, Array& imageClassArray )
{
SnmpImageMOClassInfo * foundImageInfo = NULLIMAGEINFO ;
MIBInfo	* foundMIBInfo = NULLMIBINFO ;

	if( firstSupportedMIBlistp != (ObjIdListVal*)0 && iimcAttrOID != NULLOID ) {

	    ObjIdListVal* currMIBlistp ;

	    // run through the supportedMIBs linked list..
	    for( currMIBlistp = firstSupportedMIBlistp ;
		 currMIBlistp != (ObjIdListVal*)0 && foundImageInfo == NULLIMAGEINFO ;
		 currMIBlistp = currMIBlistp->oidl_next ) {

		if( currMIBlistp->oidl_oid != NULLOID )
		    foundImageInfo = findSnmpImageMOClassInfoFromAttrOID( firstSupportedMIBlistp, currMIBlistp->oidl_oid, iimcAttrOID, imageClassArray ) ;
	    }
	}

	return( foundImageInfo ) ;
}

////////////////////////////////////////////////////////////////////////////////

// static variable storage

SnmpImageMOClassInfo * SnmpImageMOClassInfo::_firstImageClass = NULLIMAGEINFO ;


//
// SnmpImageMOClassInfo constructor
//
SnmpImageMOClassInfo::SnmpImageMOClassInfo( Bool startNewMIBClassInfoList, MIBInfo * mibInfo )
{
	if( _iimcAutoObjAndAttrOID == NULLOID )
	    _iimcAutoObjAndAttrOID =  name2oid( "iimcAutoObjAndAttr" ) ;

	_attrSyntaxes = (short *)0 ;

	_attrGetReplaceInfo    = _attrMatchesForInfo = _attrFixedOctStrLenList =
					_attrCreateInfo = _attrDeleteInfo = ((int *)0) ;

	_isMultiInstanced      = False ;
	_nAttrIndexNum         = 0 ;
	_attrIndexSyntaxList   = (short *)0 ;
	_attrIndexFixedOctStrLenList = (int *)0 ;
	_attrDeleteAttrId      = -1 ;
	_attrDeleteAttrValuecp = (char *)0 ;
	_attrDeleteAttr        = NULL ;
	
	_mibInfo = mibInfo ;

	_peer = _subordinate = _superior = _nextImageClass = NULLIMAGEINFO ;

	linkInImageClassNode( startNewMIBClassInfoList ) ;
	
	return ;
}


void
SnmpImageMOClassInfo::clear()
{
int	nattrs = getNAttrs() ;

	if( nattrs ) {

	    (void)free( (char*)_attrSyntaxes ) ;
	    _attrSyntaxes = (short *)0 ;
	    
	    delete[] _attrSyntaxes   ;
	    delete[] _attrGetReplaceInfo ;
	    delete[] _attrMatchesForInfo ;
	    delete[] _attrFixedOctStrLenList ;
	    delete[] _attrCreateInfo ;
	    delete[] _attrDeleteInfo ;

	    delete[] _attrIndexSyntaxList ;
	    delete[] _attrIndexFixedOctStrLenList ;
	    delete[] _attrDeleteAttrValuecp ;
	    delete   _attrDeleteAttr ;

	    if( _attrNotifTypeList )
	    	delete[] _attrNotifTypeList ;
	}	

	return ;
}


//
// Range check (where applicable) and buffer the attribute data to be
// buffered to SnmpImageMOClassInfo state, as opposed to MOClassInfo state.
//
int
SnmpImageMOClassInfo::setAttrImage( int nattrs, short* attrSyntaxes, 
				int* matchesForFields, int* getReplaceFields,
				int* attrFixedOctStrLenList,
				int nindices, short* attrIndexSyntaxList, int* attrIndexFixedOctStrLenList,
				int deleteAttrId, char * deleteAttrValuecp, 
				Attr* deleteAttr, char** attrNameList,
 				AttrNotifType* attrNotifTypeList )
{
int	status=OK, i ;

	setAgentClassInfoNAttrs( nattrs ) ; // Finalise the initialisation of the
					    // parent SnmpRMIBAgentClassInfo.

	for( i=0 ; i < nattrs ; i++ )
	{
	    if( matchesForFields[ i ]  < 0 || matchesForFields[ i ] > 31 )
	    {
	    	advise( NULLCP, "Class%s: attr %s has an invalid MATCHES FOR field %d",
			getClassName(), attrNameList[ i ], matchesForFields[ i ] ) ;
	    	status = NOTOK ;
	    }

	    if( getReplaceFields[ i ]  < 1 || getReplaceFields[ i ] > 3 )
	    {
	    	advise( NULLCP, "Class%s: attr %s has an invalid GET/REPLACE field %d",
			getClassName(), attrNameList[ i ], getReplaceFields[ i ] ) ;
	    	status = NOTOK ;
	    }
	}

	if( attrNotifTypeList != (AttrNotifType*)NULL ) {
	    for( i=0 ; i< nattrs ; i++ ) {
	    	if( attrNotifTypeList[ i ] != NoAttrNotif &&
		    attrNotifTypeList[ i ] != AttrValChangeNotif ) {

		    advise( NULLCP, "Class%s: attr %s has an invalid notifications field %d",
			getClassName(), attrNameList[ i ], attrNotifTypeList[ i ] ) ;
		    status = NOTOK ;
		}
	    }
	}

	if( status == OK )
	{
	    _attrSyntaxes	    = attrSyntaxes ;
	    _attrGetReplaceInfo     = getReplaceFields ;
	    _attrMatchesForInfo     = matchesForFields ;
	    _attrFixedOctStrLenList = attrFixedOctStrLenList ;

	    _nAttrIndexNum	= nindices ;
	    _attrIndexSyntaxList= attrIndexSyntaxList ;
	    _attrIndexFixedOctStrLenList = attrIndexFixedOctStrLenList ;
	    
	    if( _nAttrIndexNum > 0 )
	    	_isMultiInstanced = True ;

	    _attrDeleteAttrId	  = deleteAttrId ;
	    _attrDeleteAttrValuecp= deleteAttrValuecp ;
	    _attrDeleteAttr       = deleteAttr ;

	    _attrNotifTypeList 	  = attrNotifTypeList ;

//	    _snmpRMIBAgentClassInfo = new SnmpRMIBAgentClassInfo( nattrs ) ;
	}

	return( status ) ;
}


//
// Range check and buffer the create and delete field data to private state
//
int
SnmpImageMOClassInfo::setNameBindingImage( int* createFields, int* deleteFields )
{
int	status=OK ;

	for( register int i=0 ; i < 1 /* nbindings */ ; i++)
	{
	    if( createFields[ i ]  < 0 || createFields[ i ] > 7 )
	    {
	    	advise( NULLCP, "Class%s: attr %s has invalid CREATE field %d",
			getClassName(), getAttrName( i ), createFields[ i ] ) ;
	    	status = NOTOK ; break ;
	    }

	    if( deleteFields[ i ]  < 0 || deleteFields[ i ] > 7 )
	    {
	    	advise( NULLCP, "Class%s: attr %s has invalid DELETE field %d",
			getClassName(), getAttrName( i ), deleteFields[ i ] ) ;
	    	status = NOTOK ; break ;
	    }
	}

	if( status == OK )
	{
	    _attrCreateInfo = createFields ;
	    _attrDeleteInfo = deleteFields ;
	}

	return( status ) ;
}


//
// If we are handling the first SnmpImageMOClassInfo object of a new
// MIB then store a pointer to it in "_firstImageClass" else
// link it in at the end of the list of these objects for the current MIB.
//
void
SnmpImageMOClassInfo::linkInImageClassNode( Bool startNewMIBClassInfoList )
{
SnmpImageMOClassInfo *	currInfo ;
static	Bool firstMIBhandled = False ;

	if( startNewMIBClassInfoList == True )
	{
	    if( firstMIBhandled == False )
	    {
#ifdef DEBUG
		if( _firstImageClass != NULLIMAGEINFO )
		    advise( NULLCP, "Warning new SnmpImageMOClassInfo list commenced when _firstImageClass non-Null" ) ;
#endif
		firstMIBhandled = True ;
	    }
	    
	    _firstImageClass =  this ;

	} else {

	    for( currInfo = _firstImageClass ; currInfo -> _nextImageClass != NULLIMAGEINFO ;
					currInfo = currInfo -> _nextImageClass );

	    currInfo -> _nextImageClass = this ;
	}

	return ;
}


//
// Return the object identifier for this object's containment hierarchy superior.
//
OID
SnmpImageMOClassInfo::getImageSuperior( void ) 
{
OID	superiorOID = NULLOID ;
NameBinding** currNBs ;
int	nbindings ;

	if( (currNBs = getNameBindings( &nbindings )) == ((NameBinding**) 0))
	    advise( NULLCP, "NULL Name bindings returned for \"%s\" from \"getNameBindings()\"", getClassName() ) ;

	else if( (superiorOID = currNBs[0] -> getSuperiorClass() ) == NULLOID )
	    advise( NULLCP, "NULL superior class returned for \"%s\" from \"getSuperiorClass()\"", getClassName() ) ;

	return( superiorOID ) ;
}


//
// Follow the peer pointers to obtain the next peer MOC within the currently supported MIBs.
// 
// If no more peer MOCs are available for the current MIB then check if there
// are any contained MIBs within the list of other supported MIBs. If so continue
// the peer search amongst these.
// Note: We consider the MOCs in the the other supported MIBs to be peers only if they 
//       have the same superior class.
//
SnmpImageMOClassInfo * 
SnmpImageMOClassInfo::getNextPeerImageInfo( SnmpImageMOClassInfo * 	peer,   // may be NULL
					    MIBInfo *			mibInfo,
					    ObjIdListVal * 		firstSupportedMIBlistp, 
					    SnmpImageMOClassTraversal&  peerTraversalState,
 					    OID 			superiorClass )
{
SnmpImageMOClassInfo * peerImageInfop = peer ;

	if( peerImageInfop == NULLIMAGEINFO ) {

	    // If the _peer is NULL indicating that all of the subordinate SnmpImageMOClassInfo
	    // classes have been traversed below the current SnmpImageMOClassInfo's MOC, then
	    // check that none of the other supported MIBs are "contained" within the current
	    // MIB's MIT immediately below the current SnmpImageMOClassInfo's MOC...

	    MIBInfo * initiatingMIBinfop = peerTraversalState._initiatingMIBinfop ;

	    if( initiatingMIBinfop == NULLMIBINFO ) {

		// We are initiating the search for contained MOCs here..
		initiatingMIBinfop = peerTraversalState._initiatingMIBinfop = mibInfo ;

	    } else {

	    	// This supported MIB has no more peer contained MOCs...
	    	++peerTraversalState._nthSuppMIBcheckedForPeerMOCs ; 
	    }

	    ObjIdListVal* currMIBlistp ;
	    int		MIBsChecked = 0 ;

	    // Try to find an alternative supported MIB that provides a "contained" MIT sub-tree..
	    for( currMIBlistp =  firstSupportedMIBlistp ;
		 currMIBlistp != (ObjIdListVal*)0 && peerImageInfop == NULLIMAGEINFO ;  ) {

		// 
		// Can this MIB yield extra peer class nodes ?
		//
		if( MIBsChecked <= peerTraversalState._nthSuppMIBcheckedForPeerMOCs ||
		    // ..has this supported MIB been tried already ?
		    oid_equal( initiatingMIBinfop->_MIBidentifier,
				   		currMIBlistp -> oidl_oid )            )
		    // ..is this the MIB that initiated the peer search in the 1st place ?
		{
		    // YES prepare to check the next supported MIB..
		    currMIBlistp = currMIBlistp->oidl_next ;
		    ++MIBsChecked ;

		} else {

		     // Find the MIBInfo that corresponds to the current supported MIB..

		     MIBInfo * findMIBInfop;
		     for( findMIBInfop = MIBInfo::getRootMIBInfo() ;
			  findMIBInfop != NULLMIBINFO &&
			  	!oid_equal( currMIBlistp->oidl_oid, findMIBInfop->_MIBidentifier ) ;
			  findMIBInfop = findMIBInfop->_peerMIBInfo )
 			; // DUMMY BODY

		    if( findMIBInfop == NULLMIBINFO ) {

			advise( NULLCP, "Warning : supportedMIB OID %s does not have a MIBInfo object",
			    					sprintoid( currMIBlistp->oidl_oid ) ) ;
			// Try the next supported MIB..
			currMIBlistp = currMIBlistp->oidl_next ;
			++MIBsChecked ;

		    } else {
			
			// ...we have found the corresponding MIBInfo

			// Is this supported MIBs sub-tree contained immediately 
			// below the required superior class... ?
			if( !oid_equal( findMIBInfop->_MIBsuperiorClass, superiorClass ) ||
			    (peerImageInfop = findMIBInfop->_subordinateImageClass) == NULLIMAGEINFO ) {

			    // Try the next supported MIB..
			    currMIBlistp = currMIBlistp->oidl_next ;
			    ++MIBsChecked ;
			}
		    }
	    	}
	    }
	}

	return peerImageInfop ;
}


//
// Link the SnmpImageMOClassInfo object into the list of peer objects
// with respect to the containment hierarchy. Also set up this object's
// superior pointer and the subordinate pointer of its superior object.
// 
void
SnmpImageMOClassInfo::insertImageInPeerListSetUpSuperAndSubord( 
			Bool superiorIsMIBImage, SnmpImageMOClassInfo * superiorInfo, OID MIBsuperiorClass )
{
SnmpImageMOClassInfo *	currPeerImage = NULLIMAGEINFO ;
MIBInfo * mibInfo = NULLMIBINFO ;

	mibInfo = getMIBInfo() ;

	if( superiorIsMIBImage == True )
	{
	    _superior = NULLIMAGEINFO ;

	    if( MIBsuperiorClass != NULLOID && mibInfo -> _MIBsuperiorClass == NULLOID )
		mibInfo -> _MIBsuperiorClass = oid_cpy( MIBsuperiorClass ) ;
	
	    if( mibInfo -> _subordinateImageClass == NULLIMAGEINFO )
	    	mibInfo -> _subordinateImageClass =  this ;

	    if( mibInfo -> getSubordinateImageClass() != this )
	    {
	    	for( currPeerImage = mibInfo -> getSubordinateImageClass() ; 
		    currPeerImage -> _peer != NULLIMAGEINFO ; currPeerImage = currPeerImage -> _peer ) ;

		// Append object to the linked list of peers..
	        currPeerImage -> _peer = this ;
	    }

	} else {
	
	    // the superior object is also an SnmpImageMOClassInfo object ..
	    
	    _superior = superiorInfo ;

	    if( superiorInfo -> _subordinate == NULLIMAGEINFO )
		superiorInfo -> _subordinate = this ;	// First subordinate object
	    else
	    {
	    	// ..The superior object already has at least one subordinate

		for( currPeerImage = superiorInfo -> _subordinate ; 
		     currPeerImage -> _peer != NULLIMAGEINFO ; currPeerImage = currPeerImage -> _peer ) ;

		// Append object onto the linked list of peers..
		currPeerImage -> _peer = this ;
	    }
	}

	return ;
}


//
// Run through the SnmpImageMOClassInfo objects causing the
// setting up of the _superior, _peer and _subordinate pointers.
//
// Note: it can not be assumed that "remoteSystem" will be the MIBs superior
// object in terms of containment e.g. consider the IP Forwarding Table MIB.
//
int
SnmpImageMOClassInfo::constructContainment( void )
{
int	status = OK ;
SnmpImageMOClassInfo * currInfo = _firstImageClass, * checkInfo = NULLIMAGEINFO ;
OID	currNBsuperiorOID, checkClassOID ;
static	OID systemOID = NULLOID ;
Bool	placedInContainment = False ;

	if( systemOID == NULLOID )
	    systemOID =  name2oid( "remoteSystem" ) ;	// [10165-2] system object

	for( currInfo = _firstImageClass ; currInfo != NULLIMAGEINFO ;
		currInfo = currInfo -> _nextImageClass, placedInContainment = False )
	{
	    if( (currNBsuperiorOID = currInfo -> getImageSuperior()) == NULLOID )
	    {
	    	advise( NULLCP, "Error: Superior Image OID is NULLOID" ) ;
		status = NOTOK ; break ;
	    }

	    if( oid_equal( currNBsuperiorOID, systemOID ) )
	    {
	    	// superior is the [10165-2] system object
		
		currInfo -> insertImageInPeerListSetUpSuperAndSubord( True, NULLIMAGEINFO, systemOID ) ;

		placedInContainment = True ;

	    } else {

		// Find the superior for this class..

		for( checkInfo = _firstImageClass ; checkInfo != NULLIMAGEINFO ; checkInfo = checkInfo -> _nextImageClass )
		{
		    if( (checkClassOID = checkInfo -> getClass()) == NULLOID )
		    {
		    	advise( NULLCP, "Error: Image OID found to equal NULLOID" ) ;
			status = NOTOK ; break ;
		    }

		    if( oid_equal( currNBsuperiorOID, checkClassOID ) )
		    {
		    	// ..found superior

			currInfo -> insertImageInPeerListSetUpSuperAndSubord( False, checkInfo, NULLOID ) ;
			    
			placedInContainment = True ; break ;
		    }
		}
	    }

	    if( placedInContainment == False )
	    {
	    	if( getMIBInfo() -> getSubordinateImageClass() == NULLIMAGEINFO )
		{
		    // Assume the current MIB does not have the "system" object as its superior object...
		
		    currInfo -> insertImageInPeerListSetUpSuperAndSubord( True, NULLIMAGEINFO, currNBsuperiorOID ) ;
		
		    advise( NULLCP, "Warning : Class \"%s\" assumed to be below non-\"system\" MIB Superior class\n", 
				currInfo->getClassName() ) ;
		} else {

		    advise( NULLCP, "Error : Class \"%s\" superior class not found when\n\
superior class for \"%s\" has been taken as the MIB superior class.\n\
Could the definition for the superior to either of these classes be missing ?", 
			currInfo->getClassName(), getMIBInfo()->getSubordinateImageClass()->getClassName() ) ;

		    status = NOTOK ;
		}
	    }
	}

	return( status ) ;
}


//
// Recursively traverse (depth first) the subordinate and 
// peer SnmpImageMOClassInfo objects until the required class node is found.
//
SnmpImageMOClassInfo *
SnmpImageMOClassInfo::findImageInfo( OID classOID )
{
SnmpImageMOClassInfo * foundImageInfo = NULLIMAGEINFO ;

	if( oid_equal( getClass(), classOID ) )
	    foundImageInfo = this ;

	if( foundImageInfo == NULLIMAGEINFO && _subordinate != NULLIMAGEINFO )
	    foundImageInfo =  _subordinate -> findImageInfo( classOID ) ;

	if( foundImageInfo == NULLIMAGEINFO && _peer != NULLIMAGEINFO )
	    foundImageInfo =  _peer -> findImageInfo( classOID ) ;
	
	return( foundImageInfo ) ;
}


SnmpImageMOClassInfo *
SnmpImageMOClassInfo::findImageInfoForAttrOIDinPeer( struct ObjIdListVal* firstSupportedMIBlistp, 
				OID iimcAttrOID, Array& imageClassArray, int& attrOidIndex )
{
SnmpImageMOClassInfo * foundImageInfo = NULLIMAGEINFO ;
SnmpImageMOClassTraversal peerTraversalState ;
SnmpImageMOClassInfo * peerClassInfop = NULLIMAGEINFO ;
SnmpImageMOClassInfo * firstPeerClassInfop = NULLIMAGEINFO ; 
int	currAttrOidIndex = attrOidIndex ;

	if( (firstPeerClassInfop = _peer ) == NULLIMAGEINFO ) {
	    // ...no subordinates for the current MIB

	    /* Check for subordinates in alternative MIBs... */
	    firstPeerClassInfop = SnmpImageMOClassInfo::getNextPeerImageInfo(
						NULLIMAGEINFO, _mibInfo,
		     				firstSupportedMIBlistp, peerTraversalState, getClass() ) ;
	}

	for( peerClassInfop =  firstPeerClassInfop /* may still be NULL */ ;
	     peerClassInfop != NULLIMAGEINFO ;
	     peerClassInfop =  SnmpImageMOClassInfo::getNextPeerImageInfo( 
					    peerClassInfop->_peer,
					    peerClassInfop->_mibInfo,
				 	    firstSupportedMIBlistp, peerTraversalState, getClass() ) ) {
			     
	    if( (foundImageInfo = peerClassInfop->findImageInfoForAttrOID( firstSupportedMIBlistp,
	    				iimcAttrOID, imageClassArray,attrOidIndex)) != NULLIMAGEINFO ||
		(currAttrOidIndex != attrOidIndex) /* ...a partial match was found
						      so abort other searches */) {
	    	break ;
	    }
	}

	return foundImageInfo ;
}

SnmpImageMOClassInfo *
SnmpImageMOClassInfo::findImageInfoForAttrOIDinSubord( struct ObjIdListVal* firstSupportedMIBlistp, OID iimcAttrOID, Array& imageClassArray, int& attrOidIndex )
{
SnmpImageMOClassInfo * foundImageInfo = NULLIMAGEINFO ;
SnmpImageMOClassTraversal peerTraversalState ;
SnmpImageMOClassInfo * subordinateClassInfop = NULLIMAGEINFO ;
SnmpImageMOClassInfo * firstSubordinateClassInfop = NULLIMAGEINFO ; 
int	currAttrOidIndex = attrOidIndex ;

	if( (firstSubordinateClassInfop = getFirstSubordImageInfo() ) == NULLIMAGEINFO ) {
	    // ...no subordinates for the cuurent MIB

	    /* Check for subordinates in alternative MIBs... */
	    firstSubordinateClassInfop = SnmpImageMOClassInfo::getNextPeerImageInfo(
						NULLIMAGEINFO, _mibInfo,
		     				firstSupportedMIBlistp, peerTraversalState, getClass() ) ;
	}

	for( subordinateClassInfop =  firstSubordinateClassInfop /* may still be NULL */ ;
	     subordinateClassInfop != NULLIMAGEINFO ;
	     subordinateClassInfop =  SnmpImageMOClassInfo::getNextPeerImageInfo( 
					    subordinateClassInfop->_peer,
					    subordinateClassInfop->_mibInfo,
				 	    firstSupportedMIBlistp, peerTraversalState, getClass() ) ) {
			     
	    if( (foundImageInfo = subordinateClassInfop->findImageInfoForAttrOID( firstSupportedMIBlistp,
	    				iimcAttrOID, imageClassArray,attrOidIndex)) != NULLIMAGEINFO ||
		(currAttrOidIndex != attrOidIndex) /* ...a partial match was found
						      so abort other searches */) {
	    	break ;
	    }
	}

	return foundImageInfo ;
}

//
// Given an IIMC attribute Oid construct an Array of SnmpImageMOClassInfo
// objects from the corresponding containment hierarchy. This is required
// by the Trap code to determine the object instance that corresponds to
// a Var Bind attribute.
//
SnmpImageMOClassInfo *
SnmpImageMOClassInfo::findImageInfoForAttrOID( struct ObjIdListVal* firstSupportedMIBlistp, 
		OID iimcAttrOID, Array& imageClassArray, int& attrOidIndex )
{
SnmpImageMOClassInfo * foundImageInfo = NULLIMAGEINFO ;
OID	iimcAutoObjAndAttrOID = NULLOID ;
OID 	attrOID ;
struct 	 OIDentifier oidEntifier ;
unsigned int oidElements[256] ;

	attrOID = getAttrOid(1) ; // non-naming attribute 

	iimcAutoObjAndAttrOID = getIimcAutoObjAndAttrOID() ;

	if( iimcAutoObjAndAttrOID == NULLOID )
	    return NULLIMAGEINFO ;


	if( attrOID == NULLOID ) { // Only naming attributes in the class !

	    //
 	    // Create an attribute OID from the "iimcAutoName" naming attribute...
	    //
	    register int i;
	    for( i = 0 ; i < iimcAutoObjAndAttrOID->oid_nelem ; i++ )
		oidElements[ i ] = iimcAutoObjAndAttrOID->oid_elements[i] ;

	    attrOID = getAttrOid(0) ; // naming attribute

	    for( i = iimcAutoObjAndAttrOID->oid_nelem ; i < attrOID -> oid_nelem ; i++ ) {
	    	oidElements[ i ] = attrOID->oid_elements[ i ] ;

		// NOTE: This assumes that the length of the "iimcAutoObjAndAttr" is the same as the
		//       length of the "iimcAutoName" OID !!
	    }
	    
	    oidEntifier.oid_nelem    = attrOID -> oid_nelem ;
	    oidEntifier.oid_elements = oidElements ;
	    attrOID = &oidEntifier ;
	}

#ifdef DEBUG1
fprintf( stderr, "trap  attr %s\n", sprintoid( iimcAttrOID )) ;
fprintf( stderr, "local attr %s\n", sprintoid( attrOID )) ;
#endif

#define	CONVLENTOINDEX(len) (len - 1)

	if( attrOidIndex == -1 ) {

	    /* Determine whether this MIB is a likely candidate */
	    
	    // Compare the OIDs - exclude the iimc component sub-identifiers,
	    // the object and attribute specific sub-identifers..

	    if( iimcAttrOID->oid_nelem <= attrOID->oid_nelem )
	    	return NULLIMAGEINFO ; // Not this MIB !

	    for( int i = CONVLENTOINDEX(attrOID->oid_nelem) - 1 - 1 ; 
	    		i >= iimcAutoObjAndAttrOID->oid_nelem ; --i  ) {

		if( attrOID->oid_elements[i] != iimcAttrOID->oid_elements[i] ) {
		    return NULLIMAGEINFO ; // Not this MIB !
		}
	    }

	    // Exclude the attribute specific identifier..
	    attrOidIndex = CONVLENTOINDEX(attrOID->oid_nelem) -1 ;
	}


	if( CONVLENTOINDEX(attrOID->oid_nelem)     < attrOidIndex ||
	    CONVLENTOINDEX(iimcAttrOID->oid_nelem) < attrOidIndex   )
	    return NULLIMAGEINFO ;

	if( attrOID->oid_elements[ attrOidIndex ] == 
			iimcAttrOID->oid_elements[ attrOidIndex ] ) {

	    // Run through this MOC's atributes...

	    for( int i=1 ; i<getNAttrs() ; i++ ) {

		OID thisAttrOID = getAttrOid( i ) ;

		if( thisAttrOID != NULLOID ) {

		    if( CONVLENTOINDEX(thisAttrOID->oid_nelem) > attrOidIndex &&
		    	iimcAttrOID->oid_nelem > thisAttrOID->oid_nelem         ) {

		    	Bool successfulMatch = True ;
		
		    	for( int j = CONVLENTOINDEX(thisAttrOID->oid_nelem) ;
				    		j >= attrOidIndex + 1 ; --j ) {

		    	    if( thisAttrOID->oid_elements[ j ] != iimcAttrOID->oid_elements[ j ] ) {
			    	successfulMatch = False ;
			    	break ;
			    }
		    	}

		    	if( successfulMatch == True ) {
		    	    // Found it !!!
		    	    imageClassArray.add( (void*)this ) ;
		    	    return this ;
		    	}
		    }
		}
	    }

	    
	    // We are still on the correct OID branch..
	    ++attrOidIndex ; // check the next sub-identifier

	    // find the OID branch for the next sub-identifier..
	    if( (foundImageInfo = findImageInfoForAttrOIDinSubord( firstSupportedMIBlistp, iimcAttrOID,
					imageClassArray, attrOidIndex )) != NULLIMAGEINFO ) {

		imageClassArray.add( (void*)this ) ;
	    }

	} else {

	    // Try a peer branch..
	    foundImageInfo = findImageInfoForAttrOIDinPeer( firstSupportedMIBlistp, iimcAttrOID,
						imageClassArray, attrOidIndex ) ;
	}
	
	return( foundImageInfo ) ;
}

////////////////////////////////////////////////////////////////////////////////

SnmpRMIBAgentClassInfo::SnmpRMIBAgentClassInfo( void )
{
	_nattrs = 0 ;

	_attrByAgentNumAgentsStorage		= 0 ;

	_attrByAgentRemoteIdp			= (int *)0 ;
//	_attrSupportedByAgentMOCStatep		= (SupportedMOCattrsType *)0 ;
	_attrSupportedByAgentFlagspp		= (ObjectSupportedType **)0 ;

	_attrTooBigByAgentMaxOidsPerPdup	= (int *)0 ;

	return ;
}


void 
SnmpRMIBAgentClassInfo::setAgentClassInfoNAttrs( int nattrs ) 
{ 
	_nattrs = nattrs ;

	(void)extendAttrByAgentStorage( 1, nattrs ) ;
}


void
SnmpRMIBAgentClassInfo::clear()
{
int	i ;

	if( _attrByAgentNumAgentsStorage > 0 ) {

	    if( _attrByAgentRemoteIdp != (int *)0 )
		(void)free( (char *)_attrByAgentRemoteIdp ) ;

//	    if( _attrSupportedByAgentMOCStatep != (SupportedMOCattrsType *)0 )
//	    	(void)free( (char *)_attrSupportedByAgentMOCStatep ) ;

	    if( _attrSupportedByAgentFlagspp != (ObjectSupportedType **)0 ) {

		for( i=0 ; i<_attrByAgentNumAgentsStorage ; i++ )
		    (void)free( (char *)_attrSupportedByAgentFlagspp[ i ] ) ;

		(void)free( (char *)_attrSupportedByAgentFlagspp ) ;
	    }

	    if( _attrTooBigByAgentMaxOidsPerPdup != (int *) 0 )
	    	(void)free( (char *)_attrTooBigByAgentMaxOidsPerPdup ) ;
	}

	return ;
}

//
// If the number of the agent to perform a remote operation is greater than
// the number provided with supported/tooBig attribute data storage then extend the
// storage.
//
int
SnmpRMIBAgentClassInfo::extendAttrByAgentStorage( int byNagents, int nattrs )
{
int	status=OK, reqNumAgentsStorage ;

	if( byNagents > 0 ) {

	    reqNumAgentsStorage = _attrByAgentNumAgentsStorage + byNagents ;

	    if( _attrByAgentRemoteIdp == (int *)0 )
	    	_attrByAgentRemoteIdp = (int *)malloc( sizeof( int ) * reqNumAgentsStorage ) ;
	    else	
	    	_attrByAgentRemoteIdp = (int *)realloc( (malloc_t)_attrByAgentRemoteIdp,
			    				sizeof( int ) * reqNumAgentsStorage ) ;

	    if( _attrByAgentRemoteIdp == (int *)0 ) {

		advise( NULLCP, "malloc/realloc failure for _attrByAgentRemoteIdp" ) ;
		status = NOTOK ;

	    } else {

		// Initialise the newly allocated indices remote agent Ids..
		register int i;
	    	for( i = _attrByAgentNumAgentsStorage ; i < reqNumAgentsStorage ; i++ )
		    _attrByAgentRemoteIdp[ i ] = -1 ;
		    
		if( _attrSupportedByAgentFlagspp == (ObjectSupportedType **)0 )
		    _attrSupportedByAgentFlagspp = (ObjectSupportedType **)malloc(
	    			sizeof( ObjectSupportedType * ) * reqNumAgentsStorage ) ;
		else
		    _attrSupportedByAgentFlagspp = (ObjectSupportedType **)realloc(
		    		(malloc_t)_attrSupportedByAgentFlagspp,
	    			sizeof( ObjectSupportedType * ) * reqNumAgentsStorage ) ;

		if( _attrSupportedByAgentFlagspp == (ObjectSupportedType **)0 ) {

		    advise( NULLCP, "malloc/realloc failure for _attrSupportedByAgentFlagspp[]" ) ;
		    status = NOTOK ;

		} else {

		    for( i = _attrByAgentNumAgentsStorage ; i< reqNumAgentsStorage ; i++ )
			if( (_attrSupportedByAgentFlagspp[ i ] = (ObjectSupportedType *)malloc(
			    sizeof(ObjectSupportedType) * nattrs )) == (ObjectSupportedType *)0) {
			    
			    advise( NULLCP, "malloc failure for _attrByAgentNumAgentsStorage[][]" ) ;
			    status = NOTOK ;
			    break ;

			} else for( register int j=0 ; j<nattrs ; j++ )
				_attrSupportedByAgentFlagspp[i][j] = notKnownIfSupported ;

		    
		    if( _attrTooBigByAgentMaxOidsPerPdup == (int*)0 )
			_attrTooBigByAgentMaxOidsPerPdup =  (int*)malloc( sizeof( int ) * reqNumAgentsStorage ) ;
		    else
			_attrTooBigByAgentMaxOidsPerPdup =  (int*)realloc( (malloc_t)_attrTooBigByAgentMaxOidsPerPdup,
			    					      sizeof( int ) * reqNumAgentsStorage ) ;

		    if( _attrTooBigByAgentMaxOidsPerPdup == (int*)0 ) {

			advise( NULLCP, "malloc/realloc failure for _attrTooBigByAgentMaxOidsPerPdup" ) ;
			status = NOTOK ;

		    } else {

			// Initialise the newly allocated index "tooBig" maximum number of OIDs per pdu..
	  	    	for( int i = _attrByAgentNumAgentsStorage ; i < reqNumAgentsStorage ; i++ )
			    _attrTooBigByAgentMaxOidsPerPdup[ i ] = nattrs ;
		    }
		}
	    }	    

	    if( status == OK )
		_attrByAgentNumAgentsStorage = reqNumAgentsStorage ;
	}

	return status ;
}


//
// Returns -1 on failure
//

int
SnmpRMIBAgentClassInfo::getAgentArrayIndex( int numberOfAgent )
{
int	agentArrayIndex, status = NOTOK ;

	// Check if this agent ID is known..
	for( agentArrayIndex=0 ; agentArrayIndex<_attrByAgentNumAgentsStorage ; agentArrayIndex++ ) {
	    if( _attrByAgentRemoteIdp[ agentArrayIndex ] == numberOfAgent )
	    	break ;
	}

	if( agentArrayIndex < _attrByAgentNumAgentsStorage ) {
	    // Requested agent already has assigned storage..
	    status = OK ;
	} else {

	    // ..Agent does not have assigned storage

	    // See if there is an index that has allocated storage
            // but has not been assigned to an agent..
	    for( agentArrayIndex=0 ; agentArrayIndex<_attrByAgentNumAgentsStorage ; agentArrayIndex++ )
		if( _attrByAgentRemoteIdp[ agentArrayIndex ] == -1 )
		    break ;

 	    if( agentArrayIndex < _attrByAgentNumAgentsStorage ) {

	        // Assign the storage to the agent..
	    	_attrByAgentRemoteIdp[ agentArrayIndex ] = numberOfAgent ;
		status = OK ;
	    
	    } else {

	    	// Need to extend the storage and then assign it to the agent..
		if( extendAttrByAgentStorage( _attrByAgentIncStorageByNagents, _nattrs ) == OK ) {

		    _attrByAgentRemoteIdp[ agentArrayIndex ] = numberOfAgent ;
		    status = OK ;
		}
	    }
	}

	if( status == NOTOK )
	    agentArrayIndex = -1 ;

	return agentArrayIndex ;
}


ObjectSupportedType * 
SnmpRMIBAgentClassInfo::getAttrSupportedByAgentFlags( int numberOfAgent )
{
ObjectSupportedType * attrSupportedByAgentp = (ObjectSupportedType *)0 ;
int	agentArrayIndex ;

	if( (agentArrayIndex = getAgentArrayIndex( numberOfAgent )) != -1 )
	    attrSupportedByAgentp = _attrSupportedByAgentFlagspp[ agentArrayIndex ] ;

	return attrSupportedByAgentp ;
}

//SupportedMOCattrsType
//SnmpRMIBAgentClassInfo::getMOCSupportedByAgentFlags( int numberOfAgent )
//{ 
//SupportedMOCattrsType MOCSupportedByAgent = supportStateOfAllMOCattrsUnknown ;
//int	agentArrayIndex ;
//
//	if( (agentArrayIndex = getAgentArrayIndex( numberOfAgent )) != -1 )
//	    MOCSupportedByAgent = _attrSupportedByAgentMOCStatep[ agentArrayIndex ] ;
//
//	return MOCSupportedByAgent ;
//}

//void
//SnmpRMIBAgentClassInfo::setMOCSupportedByAgentFlags( int numberOfAgent,
//						     SupportedMOCattrsType MOCSupportedByAgent )
//{
//int	agentArrayIndex ;
//
//	if( (agentArrayIndex = getAgentArrayIndex( numberOfAgent )) != -1 )
//	    _attrSupportedByAgentMOCStatep[ agentArrayIndex ] = MOCSupportedByAgent ;
//
//	return ;
//}

//
// Return the "tooBig" maximum number of OIDs per Pdu for this agent..
//
int
SnmpRMIBAgentClassInfo::getTooBigMaxOidsPerPdu( int numberOfAgent )
{
int	tooBigMaxOidsPerPdu = 1, agentArrayIndex ;

	if( (agentArrayIndex = getAgentArrayIndex( numberOfAgent )) != -1 )
	    tooBigMaxOidsPerPdu = _attrTooBigByAgentMaxOidsPerPdup[ agentArrayIndex ] ;
	
	return tooBigMaxOidsPerPdu ;
}

//
// Define the "tooBig" maximum number of OIDs per Pdu for this agent..
//
void
SnmpRMIBAgentClassInfo::setTooBigMaxOidsPerPdu( int numberOfAgent, int tooBigMaxOidsPerPdu )
{
int	agentArrayIndex ;

	if( (agentArrayIndex = getAgentArrayIndex( numberOfAgent )) != -1 )
	    _attrTooBigByAgentMaxOidsPerPdup[ agentArrayIndex ] = tooBigMaxOidsPerPdu ;
	
	return ;
}

///////////////////////////////////////////////////////////////////////////////////////
// Class SnmpImageMOClassTraversal...

SnmpImageMOClassTraversal::SnmpImageMOClassTraversal( void )
{
	_initiatingMIBinfop		= NULLMIBINFO ;
	_nthSuppMIBcheckedForPeerMOCs	= 0 ;
}
