////////////////////////////////////////////////////////////////////////////////
// SnmpImageMOClassInfo, MIBInfo and SnmpRMIBAgentClassInfo class specifications
////////////////////////////////////////////////////////////////////////////////
// Created 		02-Jun-93
//
// Modifications	DD-MMM-YY - ""
//			11-Oct-93 - "New state item _isMultiInstanced,
//				     new methods getFirstSubordImageInfo(),
//				                 getNextPeerImageInfo(),
//						 getIsMultiInstanced()"
//			21-Oct-93 - "SnmpImageMOClassInfo inherits from 
//				     SnmpRMIBAgentClassInfo"
//			09-Nov-93 - "Code to handle "tooBig" data."
//			10-Nov-93 - "Corrections to the initialisation of
//				     the static variables _attrByAgentIncStorageByNagents
//				     and _rootMIBInfo"
//			23-Nov-93 - "SnmpImageMOClassTraversal class added,
//				     getNextPeerImageInfo now supports spliced
//				     MIBs (one MIB contained within another.)"
//			14-Dec-93 - "getFirst( MO*superior) added to allow for
//				     the case when the same MOC appears in two
//				     MIB subtrees."
//			24-Nov-94 - "setInternetSystemClassName(), getInternetSystemClassName() and
//				     getInternetSystemIdName() added, to allow configuration of the 
//				     "internetSystem" class name.
//			02-Feb-94 - "updates for attribute value change notifications."
//
////////////////////////////////////////////////////////////////////////////////

#if defined(__cplusplus)
typedef  char * malloc_t ;
#endif

#if !defined(_SNMPIMAGEMOCLASSINFO_)
#define _SNMPIMAGEMOCLASSINFO_

#if !defined(_GENERIC_MO_)
#include "GenericMO.h"
#endif

#if !defined(_CREATE_)
#include "Create.h"
#endif

#if !defined(_MO_CLASS_INFO_)
#include "MOClassInfo.h"
#endif

#include "SmiAsn.h"

#ifndef _GEN_ARRAY_
#include "GenArray.h"
#endif


class	SnmpImageMOClassInfo ;


// In terms of containment, the MIBInfo linked lists represent
// the superior object for each MIB whether this is the [ISO10165-2] 
// DMI "system" object or not.

class	MIBInfo {

#define NULLMIBINFO	((MIBInfo *) 0)

friend	SnmpImageMOClassInfo ;

	OID			_MIBidentifier ;
	OID			_MIBsuperiorClass ; // usually the OID for "system"/"remoteSystem".

	static MIBInfo *	_rootMIBInfo ; // The first instantiated MIBInfo object
	MIBInfo *		_peerMIBInfo ; // The next MIBInfo object in the linked list
	
	SnmpImageMOClassInfo *	_subordinateImageClass ; // first subordinate object for this MIB

	static	char *		_internetSystemClassName ;
	static	char *		_internetSystemIdName ;
	static	Bool		_useGenericRDNs ;

	void	linkInMIB( void ) ;

	void	recursiveDisplaySubordinates( SnmpImageMOClassInfo * ) ;

	static	void	recursiveSetUpAddClassInfo(   SnmpImageMOClassInfo * ) ;

	int	constructContainment( SnmpImageMOClassInfo * ) ;

public:

	static	MIBInfo * findMIBInfo( OID ) ;

	OID     getMIBidentifier( void ) ;
	static	MIBInfo * getRootMIBInfo( void ) ;
	static	MIBInfo * getFirstMIBInfo( struct ObjIdListVal*, int&, OID ) ;
	MIBInfo * getNextPeerMIBInfo( struct ObjIdListVal*, int&, OID ) ;
	SnmpImageMOClassInfo *	getSubordinateImageClass( void ) ;

	void	printContainment( void ) ;

	static	void	setUpAddClassInfo( void ) ;

	static	SnmpImageMOClassInfo * getSnmpImageMOClassInfo( OID /*MIBoid*/,  OID /*classOID*/, Bool = False ) ; 
	static	SnmpImageMOClassInfo * getSnmpImageMOClassInfo( struct ObjIdListVal* /*MIBoids*/, OID /*classOID*/ ) ; 
	static	SnmpImageMOClassInfo * getSnmpImageMOClassInfo( OID /*classOID*/ ) ; 

	static	SnmpImageMOClassInfo * findSnmpImageMOClassInfoFromAttrOID( struct ObjIdListVal*,
						OID /*MIBoid*/, OID /*iimcAttrOID*/, Array& ) ;
	static	SnmpImageMOClassInfo * findSnmpImageMOClassInfoFromAttrOID( struct ObjIdListVal* /*MIBoids*/, 
						OID /*iimcAttrOID*/, Array& ) ;

	static	void  setInternetSystemClassName( char* ) ;
	static	char* getInternetSystemClassName( void  ) ;
	static	char* getInternetSystemIdName( void  ) ;

	static	void setUseGenericRDNs( Bool ) ;
	static	Bool getUseGenericRDNs( void ) ;

	MIBInfo( OID ) ;
	~MIBInfo() ;
} ;

inline	SnmpImageMOClassInfo * MIBInfo::getSubordinateImageClass( void )
				        { return _subordinateImageClass ; }
inline	OID	  MIBInfo::getMIBidentifier( void ){ return _MIBidentifier ; }
inline	MIBInfo * MIBInfo::getRootMIBInfo(   void ){ return _rootMIBInfo ; }
inline	void	  MIBInfo::setInternetSystemClassName( char* classcp ){ 
					if( _internetSystemClassName != NULLCP )
					    (void)free( _internetSystemClassName ) ;
					_internetSystemClassName = strdup(classcp) ;
				 	if( _internetSystemIdName != NULLCP )
					    (void)free( _internetSystemIdName ) ;
					_internetSystemIdName = (malloc_t)calloc( 1, strlen( classcp ) + 3 ) ;
					sprintf( _internetSystemIdName, "%sId", classcp ) ; }
inline	char *	  MIBInfo::getInternetSystemClassName( void  ){ return _internetSystemClassName; }
inline	char *	  MIBInfo::getInternetSystemIdName( void  ){ return _internetSystemIdName; }
inline	void 	  MIBInfo::setUseGenericRDNs( Bool flag ){   _useGenericRDNs = flag; }
inline	Bool 	  MIBInfo::getUseGenericRDNs( void ){ return _useGenericRDNs; }


///////////////////////////////////////////////////////////////////////////////////////
//
// Logs information pertaining to individual attribute
// support for the current remote SNMP agents.
//

typedef enum	{ notKnownIfSupported, notSupported, isSupported } 
		ObjectSupportedType ;


class	SnmpRMIBAgentClassInfo {

	int	_nattrs ;

	int	_attrByAgentNumAgentsStorage ;	
		// ...Storage provided for this number of agents so far,
		// for _attrByAgentRemoteIdp, _attrSupportedByForAgentIdp
		// and the first index of
		// _attrSupportedByAgentFlagspp, _attrTooBigByAgentMaxOidsPerPdup.

	static const
	int	_attrByAgentIncStorageByNagents ;
		// ... when the number of agents storage is exhausted realloc storage 
		// for this number of extra agents.

	int *	_attrByAgentRemoteIdp ;
		// ...holds the remote agent ID for each storage index.
		// E.g. if storage for three agents has been provided, this
		// array indicates the three remote agent identifiers for the
		// arrays that are indexed 0.._attrSupportedByAgentNumAgentsStorage-1.
		//
		// Note: Since the agent storage is increased by the number specified
		// 	 by _attrByAgentNumAgentsStorage, which may be greater than 1,
		//	 unused RemoteIds indices are indicated by "-1".
		//
		// Indexed : [0.._attrSupportedByAgentNumAgentsStorage-1]


	// 
	// Attribute supported by agent storage 
	//
	ObjectSupportedType ** _attrSupportedByAgentFlagspp ;
		// ... for a given remote agent index indicate whether the SNMP object 
		// (corresponding to an OSIMIS attribute ID) is supported.
		//
		// Indexed : [0.._attrByAgentNumAgentsStorage-1][0..MOClassInfo::_nattrs-1]


	//
	// Attribute "tooBig" agent storage
	//
	int *	_attrTooBigByAgentMaxOidsPerPdup ;
		// For the corresponding MOC indicate how the attributes are divided
		// between seperate retrieval requests to avoid "tooBig" errors.
		// E.g. if _attrTooBigByAgentMaxOidsPerPdup[0] = 10 then for the agent 
		// with Id 0, upto 10 OID retrieval requests will be packaged in the same Snmp PDU.
		//
		// Indexed : [0.._attrByAgentNumAgentsStorage-1]

	int	getAgentArrayIndex( int /*numberOfAgent*/ ) ;
	
	int	extendAttrByAgentStorage( int /*byNagents*/, int /*nattrs*/ ) ;

	
	void	clear( void ) ;		// releases allocated memory

public:
	
	ObjectSupportedType * getAttrSupportedByAgentFlags( int /*numberOfAgent*/ ) ;
//	SupportedMOCattrsType getMOCSupportedByAgentFlags(  int /*numberOfAgent*/ ) ;
//	void	setMOCSupportedByAgentFlags( int /*numberOfAgent*/, SupportedMOCattrsType ) ;
	int 	getTooBigMaxOidsPerPdu( int /*numberOfAgent*/ ) ;
	void 	setTooBigMaxOidsPerPdu( int /*numberOfAgent*/, int ) ;
	
	SnmpRMIBAgentClassInfo( void ) ;
	void	setAgentClassInfoNAttrs( int ) ;
	virtual	~SnmpRMIBAgentClassInfo() ;
} ;

inline 	SnmpRMIBAgentClassInfo::~SnmpRMIBAgentClassInfo( void ) { clear(); }


///////////////////////////////////////////////////////////////////////////////////////
// The simple input file bit definitions..

typedef enum    { Get=1, Replace=2, GetReplace=3 } 
		GetReplaceType;

typedef enum    { Equality=1, Ordering=2, SubStrings=4, SetComparison=8, SetInterSection=16 } 
		MatchesForType;

typedef enum	{ Create = 1, CreateWithReference = 2, CreateWithAutoInstanceNaming = 4 }
	        CreateType ;

typedef enum	{ Delete = 1, DeleteIfNoContained = 2, DeleteContainedObjects = 4 }
        	DeleteType ;

typedef enum	{ NoAttrNotif = 0, AttrValChangeNotif = 1 }
		AttrNotifType ;

class 	SnmpImageMO ;
class	SnmpImageMOClassTraversal ;


class	SnmpImageMOClassInfo : public MOClassInfo, public SnmpRMIBAgentClassInfo {

#define NULLIMAGEINFO   ((SnmpImageMOClassInfo *) 0)

friend	MIBInfo ;
friend	SnmpImageMO ;

	static	OID	_iimcAutoObjAndAttrOID ;
	
	// 0..(_nattrs-1)
	short *	_attrSyntaxes ;
	
	// 0..(_nattrs-1)
	int*	_attrGetReplaceInfo ;	// requires the "GetReplaceType" bits
	int*	_attrMatchesForInfo ;	// requires the "MatchesForType" bits
	int*	_attrFixedOctStrLenList;// NULL if there are no fixed length octet strings in this MOC
					// or [j] = -1 if the jth attribute is not a fixed length octet string 
					//          else the octet string length.

	// 0..(_nbindings-1)
	int*	_attrCreateInfo ;	// requires the "CreateType" bits
	int*	_attrDeleteInfo ;	// requires the "DeleteType" bits


	Bool    _isMultiInstanced ; 	// "True" if table entry MOC image
					// with INDEX information
	//	
	//	PARSE statements, just INDEX, DELETEATT and DELETEVALUE so far...
	//

	int	_nAttrIndexNum ;	// 0    if no INDEX IQA Snmp input file clause
        short*  _attrIndexSyntaxList ;  // NULL if no INDEX IQA Snmp input file clause
	int*	_attrIndexFixedOctStrLenList;// NULL if there are no fixed length octet strings amongst the indices
					// or [j] = -1 if the jth index is not a fixed length octet string 
					//          else the octet string length.

	int	_attrDeleteAttrId ;	// -1     if no DELETEATT   IQA Snmp input file clause
	char*	_attrDeleteAttrValuecp ;// NULLCP if no DELETEVALUE IQA Snmp input file clause
	Attr*	_attrDeleteAttr ; 	// NULL   if no DELETEVALUE IQA Snmp input file clause

	AttrNotifType* _attrNotifTypeList ;	// NULL	  if no notifications for this classes attributes

	MIBInfo	* _mibInfo ;
	
	SnmpImageMOClassInfo *	_superior ;	// immediate ancestor object
	SnmpImageMOClassInfo *	_peer ;		// next object at the same level
	SnmpImageMOClassInfo *	_subordinate ;	// first subordinate object


	static SnmpImageMOClassInfo *	_firstImageClass ; // Within the MIB currently being parsed
	SnmpImageMOClassInfo 	    *	_nextImageClass  ; // Linked list ptr for all image class
 							   // objects for the MIB being parsed.

	void	insertImageInPeerListSetUpSuperAndSubord( Bool, SnmpImageMOClassInfo *, OID ) ;
	OID	getImageSuperior( void ) ;

	void	linkInImageClassNode( Bool ) ;

	int	constructContainment( void ) ;
	
	SnmpImageMOClassInfo * findImageInfo( OID /*classOID*/ ) ;
	SnmpImageMOClassInfo * findImageInfoForAttrOIDinPeer( struct ObjIdListVal*, OID /*iimcAttrOID*/, 
								Array&, int& /*attrOidIndex*/ ) ;
	SnmpImageMOClassInfo * findImageInfoForAttrOIDinSubord( struct ObjIdListVal*, OID /*iimcAttrOID*/, 
								Array&, int& /*attrOidIndex*/ ) ;
	SnmpImageMOClassInfo * findImageInfoForAttrOID( struct ObjIdListVal*, OID /*iimcAttrOID*/, Array&, int& ) ;

	// Parse methods..
	static	int 	bufferNextLine( char*, int, FILE* ) ;
	static	char * 	removeAllSpaces( char* ) ;
	static	int	parseSingleMIBClassInfo( FILE*, MIBInfo* ) ;

	void	clear() ;			// releases allocated memory

public:

	static	int initialiseSnmpClassInfo( char* ) ;

	MIBInfo     * getMIBInfo( void ) ;

	OID	getIimcAutoObjAndAttrOID( void ) ;
	
	int	setAttrImage( int /*nattrs*/, short*, int*, int*, int*, int, 
				short*, int*, int, char*, Attr*, char**, AttrNotifType* ) ;
	int	setNameBindingImage( int*, int* ) ;

	// methods to traverse the image class containment hierarchy
	SnmpImageMOClassInfo * getFirstSubordImageInfo( void ) ;
	static SnmpImageMOClassInfo * getNextPeerImageInfo( SnmpImageMOClassInfo*, MIBInfo*,
				ObjIdListVal*, SnmpImageMOClassTraversal&, OID ) ;
	// ..when MIB splicing is required
	SnmpImageMOClassInfo * getNextPeerImageInfo( void ) ;
	// ..when MIB splicing is not required
	
	// inline methods to access attribute specific information
	short*	getAttrSyntaxes( void ) ;
	int* 	getAttrReplaceInfo( void ) ;
	int*	getAttrMatchesForInfo( void ) ;
	int*	getAttrFixedOctStrLenList( void ) ;
	int*	getAttrCreateInfo( void ) ;
	int*	getAttrDeleteInfo( void ) ;
	Bool	getIsMultiInstanced( void ) ;
	int	getAttrIndexNum( void ) ;
	short*	getAttrIndexSyntaxList( void ) ;
	int*	getAttrIndexFixedOctStrLenList( void ) ;
	int	getAttrDeleteAttrId(void) ;
	char*	getAttrDeleteAttrValue(void) ;
	Attr*	getAttrDeleteAttr(void) ;
	AttrNotifType* getAttrNotifList(void) ;

	SnmpImageMOClassInfo( Bool, MIBInfo * ) ;
	~SnmpImageMOClassInfo() ;
} ;

#define NULLIMAGEINFO	((SnmpImageMOClassInfo *) 0)

inline SnmpImageMOClassInfo::~SnmpImageMOClassInfo( void ) { clear(); }

inline OID    SnmpImageMOClassInfo::getIimcAutoObjAndAttrOID( void ){ return _iimcAutoObjAndAttrOID ;}

inline MIBInfo * SnmpImageMOClassInfo::getMIBInfo(  void ) { return _mibInfo ; }

inline SnmpImageMOClassInfo * SnmpImageMOClassInfo::getFirstSubordImageInfo( void )
							         { return _subordinate ; }
inline SnmpImageMOClassInfo * SnmpImageMOClassInfo::getNextPeerImageInfo( void )
								 { return _peer ; }

inline short* SnmpImageMOClassInfo::getAttrSyntaxes( void )	 { return _attrSyntaxes; }
inline int*   SnmpImageMOClassInfo::getAttrReplaceInfo(void)     { return _attrGetReplaceInfo; }
inline int*   SnmpImageMOClassInfo::getAttrMatchesForInfo(void)  { return _attrMatchesForInfo; }
inline int*   SnmpImageMOClassInfo::getAttrFixedOctStrLenList(void) { return _attrFixedOctStrLenList; }
inline int*   SnmpImageMOClassInfo::getAttrCreateInfo(void)      { return _attrCreateInfo; }
inline int*   SnmpImageMOClassInfo::getAttrDeleteInfo(void)      { return _attrDeleteInfo; }
inline Bool   SnmpImageMOClassInfo::getIsMultiInstanced( void )  { return _isMultiInstanced; }
inline int    SnmpImageMOClassInfo::getAttrIndexNum(void)	 { return _nAttrIndexNum; }
inline short* SnmpImageMOClassInfo::getAttrIndexSyntaxList(void) { return _attrIndexSyntaxList; }
inline int*   SnmpImageMOClassInfo::getAttrIndexFixedOctStrLenList(void) { return _attrIndexFixedOctStrLenList; }
inline int    SnmpImageMOClassInfo::getAttrDeleteAttrId(void)	 { return _attrDeleteAttrId; }
inline char*  SnmpImageMOClassInfo::getAttrDeleteAttrValue(void) { return _attrDeleteAttrValuecp; }
inline Attr*  SnmpImageMOClassInfo::getAttrDeleteAttr(void)	 { return _attrDeleteAttr; }
inline AttrNotifType* SnmpImageMOClassInfo::getAttrNotifList(void){return _attrNotifTypeList; }

///////////////////////////////////////////////////////////////////////////////////////
// Data class to keep track of the traversal of multiple MIBs, where one MIB
// is potentially "contained" at some point within the MIT of another supported MIB...

class 	SnmpImageMOClassTraversal {

friend	SnmpImageMOClassInfo ;

	MIBInfo * _initiatingMIBinfop ;
	int	  _nthSuppMIBcheckedForPeerMOCs ;

public:
	SnmpImageMOClassTraversal( void ) ;
} ;

#endif
