//
// Copyright (c) 1994 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

////////////////////////////////////////////////////////////////
// SnmpRMIBAgent class methods
////////////////////////////////////////////////////////////////
// Created              12-Jul-93
//
// Modifications        DD-MMM-YY - ""
//
////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <time.h>
#ifdef  __GNUG__
typedef	char * malloc_t ;
#endif

#include "ucbext.h"
#include "msap.h"
#include "SnmpRMIBAgent.h"
#include "SnmpImageAccess.h"


extern "C" {
void    advise(...) ;
/*
#ifdef  SYS5
void	srand() ;
#endif
*/
}

// static storage..
int	SnmpRMIBAgent::_snmpRMIBlastUsedAgentId = -1 ;

//
// Upon a "tooBig" error determine the number of OIDs
// that are to be used in the retry, based on the number
// of OIDs in the original request.
//
int
simpleTooBigReduceRetryNumber( int numOidsThatCausedTooBigError )
{
int	numOidsForTooBigRetry ;

	if( numOidsThatCausedTooBigError > 5 )
	    numOidsForTooBigRetry = (numOidsThatCausedTooBigError + 1) >> 1 ; // Halve the OIDs per Pdu
	else 
	    numOidsForTooBigRetry = -1 ; // Error condition

#ifdef DEBUG
	advise( NULLCP, "TOO BIG ERROR ON %d RETRY WITH %d", numOidsThatCausedTooBigError, numOidsForTooBigRetry ) ;
#endif

	return numOidsForTooBigRetry ;
}


//
// SnmpRMIBAgent constructor - initialise the private state
//
SnmpRMIBAgent::SnmpRMIBAgent( int magicNumber )
{
	snmpRMIBAgentInitialised = False ;

	_snmpRMIBAgentId = ++_snmpRMIBlastUsedAgentId ;

	version = SnmpVunknown ;
	iimcAutoObjAndAttrOID = NULLOID ;

	snmpRMIBdata.sd = -1 ;
	snmpRMIBdata.ps = NULLPS ;
	
	snmpRMIBdata.agentParty	  = NULLPARTY ;
	snmpRMIBdata.managerParty = NULLPARTY ;
	snmpRMIBdata.mgmtContext  = NULLCONTEXT ;

	snmpRMIBdata.maxRetries	= 3 ;
	snmpRMIBdata.timeoutS   = 4 ;
	
	snmpRMIBdata.agentOperationId = 0 ;

#ifdef  SYS5
    	(void) srand ((unsigned int) time ((long *) 0)) ;
	snmpRMIBdata.agentRequestId =  ((int) rand ()) & 0x7fffffff ;
#else
        (void) srandom ((int) time ((long *) 0));
	snmpRMIBdata.agentRequestId = ((int) random ()) & 0x7fffffff;
#endif

	snmpRMIBdata.outstandingRequestIDs = (int *)calloc( 1, sizeof( int ) * SNMPRMIBOUTSANDINGIDSBLOCKSIZE );
	snmpRMIBdata.pduTransmittedAtS     = (long*)calloc( 1, sizeof( long) * SNMPRMIBOUTSANDINGIDSBLOCKSIZE );
	snmpRMIBdata.numOutstandingRequestIDs = 0 ;

	snmpRMIBdata.numOfResponsesReceived = (long)0 ;

	snmpRMIBdata.tooBigReduceRetryFN = simpleTooBigReduceRetryNumber ; // Assign the retry procedure

	snmpRMIBdata.remoteAgentMayIgnoreTooBigMessages = ( magicNumber & 0x1 ) ? TRUE : FALSE ;
	remoteAgentIntermitObjSupport			= ( magicNumber & 0x2 ) ? TRUE : FALSE ;

	snmpRMIBdata.snmpRMIBoperationNodep = NULLSNMPRMIBOPERATIONNODE ;
	
	return ;
}

SnmpRMIBAgent::~SnmpRMIBAgent()
{
	// Free any private state storage..

	removeSNMPv1agentParty( &snmpRMIBdata.agentParty );
	removeSNMPv1managerParty( &snmpRMIBdata.managerParty );
	removeSNMPv1Context( &snmpRMIBdata.mgmtContext ) ;
	
	if( snmpRMIBdata.outstandingRequestIDs != NULL )
	    (void)free((char*) snmpRMIBdata.outstandingRequestIDs ) ;

	if( snmpRMIBdata.pduTransmittedAtS != NULL )
	    (void)free((char*) snmpRMIBdata.pduTransmittedAtS ) ;

	if( iimcAutoObjAndAttrOID != NULLOID ) {
	    oid_free( iimcAutoObjAndAttrOID ), iimcAutoObjAndAttrOID = NULLOID ;
	}

	snmpV1closeSession() ;
}


//
//
//
SnmpRMIBerrorEnum SnmpRMIBAgent::snmpV1openSession( struct qbuf * transportAddress,
							struct qbuf * communityString,
							int snmpMsgRetryLimit, int snmpMsgTimeoutS)
{
SnmpRMIBerrorEnum	snmpRMIBerror = SnmpRMIBnoError ;

	snmpRMIBdata.version = version = SnmpV1 ;
	
	if( transportAddress == NULL || communityString == NULL )
	    snmpRMIBerror =  SnmpRMIBnullParameter ;

	if( snmpRMIBerror == SnmpRMIBnoError ) {

	    if( (iimcAutoObjAndAttrOID = name2oid( "iimcAutoObjAndAttr" )) == NULLOID ) {

		advise( NULLCP, "Error: { iimcAutoObjAndAttr } undefined" ) ;
		snmpRMIBerror = SnmpRMIBundefOID ;
	    }
	    
	    // Set up the agent party..
    
	    if( (snmpRMIBerror = buildSNMPv1agentParty( transportAddress, communityString, 
	    						&snmpRMIBdata.agentParty ) ) == SnmpRMIBnoError )
	    {
	    	if( (snmpRMIBerror = buildSNMPv1managerParty( &snmpRMIBdata.managerParty ) ) == SnmpRMIBnoError )
		    snmpRMIBerror = buildSNMPv1Context( &snmpRMIBdata.mgmtContext ) ;
	    }

	    if( snmpRMIBerror == SnmpRMIBnoError )
	    {
	    	if( (snmpRMIBerror = SnmpRMIBpartyConnect( &snmpRMIBdata )) == SnmpRMIBnoError ) {

		    // Update the maximum retries and timeout parameters..
		    snmpRMIBdata.maxRetries = snmpMsgRetryLimit ;
		    snmpRMIBdata.timeoutS   = snmpMsgTimeoutS ;
			
		    snmpRMIBAgentInitialised = True ;
		}
#ifdef DEBUG
advise( NULLCP, "Party connect PS = %d SD = %d", snmpRMIBdata.ps , snmpRMIBdata.sd ) ;
#endif
	    }
	}

	return( snmpRMIBerror ) ;
}

//
//
//
SnmpRMIBerrorEnum SnmpRMIBAgent::snmpV1closeSession( void )
{
SnmpRMIBerrorEnum	snmpRMIBerror = SnmpRMIBnoError ;

	// Will eventually need to free party storage 
	// See f_quit/fin_party in snmpif.c

	SnmpRMIBpartyDisconnect( &snmpRMIBdata ) ;
#ifdef DEBUG
	advise( NULLCP, "snmpV1closeSession called" ) ;
#endif
	return snmpRMIBerror ;
}

//
//
//
SnmpRMIBerrorEnum SnmpRMIBAgent::convertOIDsIIMCtoInternet( OID * iimcOIDs, int numOIDs, OID ** internetOIDsp )
{
SnmpRMIBerrorEnum	snmpRMIBerror = SnmpRMIBnoError ;
const	int iimcAutoObjAndAttrSizeOIDelems = sizeof( iimcAutoObjAndAttrOID -> oid_elements[0] ) *
							(iimcAutoObjAndAttrOID -> oid_nelem) ;
unsigned char * attrIdIsSupportedByAgent = (unsigned char *)0 ;

	if( (*internetOIDsp = (OID *)malloc( numOIDs * sizeof( OID *) )) == ((OID *)0) )
	    snmpRMIBerror = SnmpRMIBmemAllocFailure ;

	else {

	    // Run through each OID removing the leading { iimcAutoObjAndAttr } sub-identifiers

	    for( int i = 0 ; i < numOIDs ; i++ ) {

		// Check that the { iimcAutoObjAndAttr } OID is present..
		if( iimcOIDs[ i ] -> oid_nelem <= iimcAutoObjAndAttrOID -> oid_nelem ||
		    bcmp( (char *)iimcAutoObjAndAttrOID -> oid_elements,
		    	  (char *)iimcOIDs[ i ] -> oid_elements,
			  iimcAutoObjAndAttrSizeOIDelems )          ) {
			
		    snmpRMIBerror = SnmpRMIBiimcAutoObjAndAttrNotInOID ;
		    break ;
		}

		if( ((*internetOIDsp)[i] = oid_cpy( iimcOIDs[ i ] )) == NULLOID ) {

		    snmpRMIBerror = SnmpRMIBmemAllocFailure ;

		    // Return any storage allocated so far..
		    (void)free((char *) *internetOIDsp );

		    for( int j = 0 ; j < i ; j++ )
			oid_free( iimcOIDs[ j ] ) ;
			
		    break ;
		}

		// Shift the sub-identifiers left to replace those
		// of the leading { iimcAutoObjAndAttr } sub-identifiers..

		bcopy( (char *)&((*internetOIDsp)[i]->oid_elements[ iimcAutoObjAndAttrOID->oid_nelem ]),
		       (char *)(*internetOIDsp)[i] -> oid_elements,
		       sizeof( iimcAutoObjAndAttrOID -> oid_elements[0] ) * 
			   	((*internetOIDsp)[i]->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ) );

		// and adjust the number of sub-identifiers remaining..
		(*internetOIDsp)[i]->oid_nelem -= iimcAutoObjAndAttrOID->oid_nelem ;
#ifdef DEBUG
fprintf( stderr, "new -> Internet oid %s newOIDlen %d\n", sprintoid( (*internetOIDsp)[i] ),
							 (*internetOIDsp)[i]->oid_nelem ) ;
#endif
	    }
	}

	return( snmpRMIBerror ) ;
}

// 
// Replace EXISTING OIDs with the IIMC equivalent
//
// SnmpRMIBundefOID
//
SnmpRMIBerrorEnum SnmpRMIBAgent::convertOIDsInternetToIIMC( OID * internetOIDs, int numOIDs )
{
SnmpRMIBerrorEnum	snmpRMIBerror = SnmpRMIBnoError ;

	if( internetOIDs == (OID *)0 )
	    snmpRMIBerror = SnmpRMIBnullParameter ;
	else {

	    // Run through each OID inserting the leading { iimcAutoObjAndAttr } sub-identifiers..
	    for( int i = 0 ; i < numOIDs ; i++ ) {

		if( internetOIDs[ i ] != NULLOID ) {

		    if((internetOIDs[ i ] -> oid_elements = 
			(unsigned int *)realloc( (malloc_t)internetOIDs[ i ] -> oid_elements,
			    (iimcAutoObjAndAttrOID->oid_nelem + internetOIDs[i]->oid_nelem) * 
					sizeof( iimcAutoObjAndAttrOID -> oid_elements[0] ))) == NULL ) {

			snmpRMIBerror = SnmpRMIBmemAllocFailure ;
			break ;

		    } else {

			// Shift the internet OIDs left to make space for the IIMC header..
			bcopy( (char *)internetOIDs[i]->oid_elements,
			       (char *)&(internetOIDs[i]->oid_elements[iimcAutoObjAndAttrOID->oid_nelem]),
			       sizeof( iimcAutoObjAndAttrOID -> oid_elements[0] ) * internetOIDs[i]->oid_nelem ) ;

		    	// Insert the IIMC header..
			bcopy( (char *)iimcAutoObjAndAttrOID -> oid_elements,
			       (char *)internetOIDs[i]  -> oid_elements,
			       sizeof( iimcAutoObjAndAttrOID -> oid_elements[0] ) * iimcAutoObjAndAttrOID->oid_nelem ) ;

			internetOIDs[ i ] -> oid_nelem += iimcAutoObjAndAttrOID -> oid_nelem ;
#ifdef DEBUG
fprintf( stderr, "new -> IIMC oid %s newOIDlen %d\n", sprintoid( internetOIDs[i] ), 
								 internetOIDs[i]->oid_nelem ) ;
#endif

		    }
		}
	    }
	}

	return( snmpRMIBerror ) ;
}


//
// Form an array of Internet OIDs for the specified attribute Ids with
// OID sub-identifiers that correspond to the rdnString appended.
// For single instanced objects we just need to append a trailing ".0".
//
// Note: the rdnString parameter will be changed by this method.
// Note: internetOIDs and attrIds must already point to arrays with "reqNAttrs" indices.
//
SnmpRMIBerrorEnum
SnmpRMIBAgent::createInternetOIDsandAppendRDNstr( SnmpImageMOClassInfo * snmpImageClassInfop,
					OID* internetOIDs, int reqNAttrs, int* attrIds, char * rdnString )
{
SnmpRMIBerrorEnum snmpRMIBerror = SnmpRMIBnoError ;

	if( snmpImageClassInfop == NULLIMAGEINFO || internetOIDs == (OID*)0 || 
			reqNAttrs <= 0 || attrIds == (int*)0 || rdnString == NULLCP ) 
	    snmpRMIBerror = SnmpRMIBnullParameter ;
	else {

	    char * baseRDNcp, * rdn1cp, *rdn2cp ;
	    
	    if( (baseRDNcp = strchr( rdnString, '=' )) != NULLCP ) {

	    	rdn2cp = rdn1cp = ++baseRDNcp ; // Skip the '='

		// Count the number of sub-identifiers within the rdn string..
		int numSubIds ;

		for( numSubIds = 1 ; (rdn2cp = strchr( rdn2cp, '.' )) != NULLCP ;
		     ++rdn2cp, ++numSubIds )
		    ; // EMPTY LOOP BODY

		for( int i=0 ; i<reqNAttrs ; i++ ) {

		    OID iimcToInternetOID = NULLOID ;
		    OID attrOid = snmpImageClassInfop->getAttrOid( attrIds[ i ] ) ;

		    if( attrOid != NULLOID ) {

			// Now create an OID of the correct size..
			iimcToInternetOID = (OID)malloc( sizeof( OIDentifier ) ) ;

			if( iimcToInternetOID == NULLOID ) 
			    snmpRMIBerror = SnmpRMIBmemAllocFailure ;
			else {
			
			    iimcToInternetOID->oid_nelem    = numSubIds + (attrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem); 
			    iimcToInternetOID->oid_elements = (unsigned int*)malloc( sizeof(unsigned int) * iimcToInternetOID->oid_nelem) ;

			    if( iimcToInternetOID->oid_elements == (unsigned int*)0 ) {
			    	(void)free((char*)iimcToInternetOID ) ;
				snmpRMIBerror = SnmpRMIBmemAllocFailure ;
			    } else {

				bcopy( (char*)&(attrOid->oid_elements[ iimcAutoObjAndAttrOID->oid_nelem ]),
				       (char*)iimcToInternetOID->oid_elements,
				       sizeof(unsigned int)*(attrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem) ) ;

				// Now buffer across the RDN sub-identifiers..
		 		int baseRDNindex = attrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ;

	  			for( int j=0 ; j<numSubIds ; j++ ) {

				    if( j != numSubIds - 1 ) {

					rdn2cp  = strchr( rdn1cp, '.' ) ;
					*rdn2cp = '\0' ;
					iimcToInternetOID->oid_elements[ baseRDNindex + j ] = atoi( rdn1cp ) ;
					*rdn2cp = '.'  ;    // ..replace the delimiter 
					rdn1cp = ++rdn2cp ; // ..skip the delimiter

				    } else iimcToInternetOID->oid_elements[ baseRDNindex+j ] = atoi( rdn1cp ) ;
				}

				rdn1cp = baseRDNcp ; // Reset the rdn pointer
			    }
			}
		    }

		    internetOIDs[ i ] = iimcToInternetOID ;
		}
	    }
	}

	return snmpRMIBerror ;
}


//
// For the moment - blocking only
//
// Note: will not attempt to retrieve attribute Id 0 (the naming attribute)
//
// Note: the OIDs within the returned bindings will not be converted
// to their IIMC equivalent as the search for results routine allows for
// this.
//
SnmpRMIBerrorEnum 
SnmpRMIBAgent::snmpGetSingleInstancedMO( SnmpImageMOClassInfo * snmpImageClassInfop,
					 SnmpRMIBAgentResults&  snmpRMIBAgentResultsr )
{
SnmpRMIBerrorEnum snmpRMIBerror = SnmpRMIBnoError ;
OID *	iimcOIDs = (OID *)0 ;
OID *	internetOIDs ;
int	operationId, maxNAttrs, reqNAttrs, attrIdOfNamingAttribute=-1, tooBigMaxOidsPerPdu=1 ;
int *	attrIds ;
struct	SnmpRMIBrxedResults  * currRXresultsp ;
int	i ;
ObjectSupportedType * attrIdIsSupportedByAgentp = (ObjectSupportedType *)0;

	if( snmpRMIBAgentInitialised == False )
	    snmpRMIBerror = SnmpRMIBagentNotInitialised ;

	else if( snmpImageClassInfop == NULLIMAGEINFO )
	    snmpRMIBerror = SnmpRMIBnullParameter ;

	else if(remoteAgentIntermitObjSupport == False && 
		(attrIdIsSupportedByAgentp =
		 	snmpImageClassInfop->getAttrSupportedByAgentFlags( _snmpRMIBAgentId ))
 									== (ObjectSupportedType *)0 )
	    snmpRMIBerror = SnmpRMIBunableToAccessSupportData ;

	else {

	    maxNAttrs = snmpImageClassInfop->getNAttrs() ;

	    reqNAttrs = maxNAttrs - 1 ; // excluding the naming attribute

	    if( remoteAgentIntermitObjSupport == False ) {

		// Determine the number of Internet objects to be retrieved
	    	for( i=1 ; i < maxNAttrs ; i++ )
		    if( attrIdIsSupportedByAgentp[ i ] == notSupported )
		    	reqNAttrs-- ;
	    }

	    if( reqNAttrs <= 0 ) {
#ifdef DEBUG
		advise( NULLCP, "Call to SnmpRMIBAgent::snmpGetSingleInstancedMO() with no attributes to be retrieved" ) ;
#endif
	    } else if( (attrIds = (int *)malloc( sizeof( int ) * reqNAttrs )) == (int *)0 ) {

	    	snmpRMIBerror = SnmpRMIBmemAllocFailure ;
		
	    } else {

		if( (iimcOIDs = (OID *)malloc( sizeof( OID ) * reqNAttrs )) == (OID *)0 )
		    snmpRMIBerror = SnmpRMIBmemAllocFailure ;

		else {

		    int nextUnusedIndex = 0 ;

		    // Set up the required attribute OIDs..
		    for( i=1 ; i < maxNAttrs ; i++ ) {

			if( remoteAgentIntermitObjSupport == True ||
				attrIdIsSupportedByAgentp[ i ] != notSupported ) {

			    iimcOIDs[ nextUnusedIndex ] = snmpImageClassInfop->getAttrOid( i ) ;
			    // ..Note: the MOClassInfo OIDs are NOT copied

			    attrIds[ nextUnusedIndex ] = i ;
			    ++nextUnusedIndex ;
			}
		    }

		    // Now convert these IIMC OIDs to their Internet equivalent..
		    if( (snmpRMIBerror = convertOIDsIIMCtoInternet( iimcOIDs, reqNAttrs, 
							&internetOIDs )) == SnmpRMIBnoError ) {

			tooBigMaxOidsPerPdu = snmpImageClassInfop->getTooBigMaxOidsPerPdu( _snmpRMIBAgentId ) ;					
			
			if( (snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata, GetNextOperator, 
					internetOIDs, reqNAttrs, (PE *)0, attrIds, &operationId, tooBigMaxOidsPerPdu )) == SnmpRMIBnoError ) {

			    if( (snmpRMIBerror = SnmpRMIBreceiveMessageFromAgent( &snmpRMIBdata,
		    						operationId, &currRXresultsp, &tooBigMaxOidsPerPdu )) == SnmpRMIBnoError ) {
			    	// Buffer the results for subsequent retrieval..
				snmpRMIBAgentResultsr.replace( currRXresultsp ) ;

				snmpImageClassInfop->setTooBigMaxOidsPerPdu( _snmpRMIBAgentId, tooBigMaxOidsPerPdu ) ;
			    }			
		    	}
		
		    	// Return the storage created during the OID conversion..
		    	for( i=0 ; i<reqNAttrs ; i++ )
			    oid_free( internetOIDs[ i ] ) ;

		    	(void)free( (char*)internetOIDs ) ;
	    	    }

		    (void)free((char *)iimcOIDs ) ;
	    	}

		(void)free( (char*)attrIds ) ;
	    }
	}

	return( snmpRMIBerror ) ;
}

//
// Form the SNMP Set parameters and then cause it to be transmitted. Finally
// buffer any successful responses.
//
// Note: the OIDs within the returned bindings will not be converted
// to their IIMC equivalent as the search for results routine allows for
// this.
//
SnmpRMIBerrorEnum 
SnmpRMIBAgent::snmpSetMOattributes( SnmpImageAccess* currMO, int *attrIds, 
				    PE* PEvalues, int reqNAttrs,
				    SnmpImageMOClassInfo * snmpImageClassInfop,
				    SnmpRMIBAgentResults&  snmpRMIBAgentResultsr )
{
SnmpRMIBerrorEnum snmpRMIBerror = SnmpRMIBnoError ;
OID *	internetOIDs = (OID*)NULL ;
int	operationId ;
struct	SnmpRMIBrxedResults  * currRXresultsp ;
char *	rdncp = NULLCP ;    

	if( snmpRMIBAgentInitialised == False )
	    snmpRMIBerror = SnmpRMIBagentNotInitialised ;

	else if( snmpImageClassInfop == NULLIMAGEINFO || reqNAttrs <= 0 )
	    snmpRMIBerror = SnmpRMIBnullParameter ;

	if( (internetOIDs = (OID *)malloc( sizeof( OID ) * reqNAttrs )) == (OID *)0 )
	    snmpRMIBerror = SnmpRMIBmemAllocFailure ;

	else {

	    //
	    // Prepare the Internet OIDs..
	    //
	    if( snmpImageClassInfop->getIsMultiInstanced() == True ) {
	    	rdncp = currMO -> getRDNString() ; 
	    } else
	    	rdncp = strdup( "rdnId=0" ) ; // require just a trailing ".0" for SNMP access

	    snmpRMIBerror = createInternetOIDsandAppendRDNstr( snmpImageClassInfop,
							internetOIDs, reqNAttrs, attrIds, rdncp ) ;

	    //
	    // Perform the SNMP Set operation..
	    //
	    if( snmpRMIBerror == SnmpRMIBnoError ) {

		// Note: any tooBig limitations must be ignored as Sets must remain atomic..

	    	if( (snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata,
					SetOperator, internetOIDs, reqNAttrs, PEvalues, attrIds, &operationId,
					reqNAttrs )) == SnmpRMIBnoError ) {

		    if( (snmpRMIBerror = SnmpRMIBreceiveMessageFromAgent( &snmpRMIBdata,
		    						operationId, &currRXresultsp, (int*)0 )) == SnmpRMIBnoError ) {

		    	// Buffer the results for subsequent retrieval..
		    	snmpRMIBAgentResultsr.replace( currRXresultsp ) ;
		    }
		}
	    }

	    if( rdncp != NULLCP )
	    	(void)free( rdncp ) ;
	
	    if( internetOIDs != (OID*)NULL ) {
	    	for( int i=0 ; i<reqNAttrs ; i++ )
	    	    if( internetOIDs[i] != NULLOID )
		    	oid_free( internetOIDs[i] ) ;
	    }
	}

	if( internetOIDs != (OID*)NULL )
	    (void)free( (char*)internetOIDs ) ;

	return( snmpRMIBerror ) ;
}


//
// If a "noSuchName" error is generated then one further Set is attempted
// that lists only those attributes that are writeable...
//
// Note: the order of PEvalues PEs may be altered by this procedure.
//
SnmpRMIBerrorEnum 
SnmpRMIBAgent::snmpCreateEntry( RDN rdn, int *attrIds, PE* PEvalues, int reqNAttrs,
				    SnmpImageMOClassInfo * snmpImageClassInfop,
				    SnmpRMIBAgentResults&  snmpRMIBAgentResultsr   )
{
SnmpRMIBerrorEnum snmpRMIBerror = SnmpRMIBnoError ;
OID *	internetOIDs = (OID*)NULL ;
int	operationId ;
struct	SnmpRMIBrxedResults  * currRXresultsp ;
char *	dotrdncp = NULLCP ;
int*	getReplaceInfop ;
Bool	successfulCreate = False ;

	if( snmpRMIBAgentInitialised == False )
	    snmpRMIBerror = SnmpRMIBagentNotInitialised ;

	else if( snmpImageClassInfop == NULLIMAGEINFO || reqNAttrs <= 0 || rdn == NULLRDN )
	    snmpRMIBerror = SnmpRMIBnullParameter ;

	if( (internetOIDs = (OID *)malloc( sizeof( OID ) * reqNAttrs )) == (OID *)0 )
	    snmpRMIBerror = SnmpRMIBmemAllocFailure ;

	else {

	    //
	    // Prepare the Internet OIDs..
	    //
	    if( snmpImageClassInfop->getIsMultiInstanced() == True )
	    	dotrdncp = rdn2str( rdn ) ;
	    else 
	    	dotrdncp = strdup( "rdnId=0" ) ; // require just a trailing ".0" for SNMP access

	    snmpRMIBerror = createInternetOIDsandAppendRDNstr( snmpImageClassInfop,
							internetOIDs, reqNAttrs, attrIds, dotrdncp ) ;

	    //
	    // Perform the SNMP Set operation..
	    //
	    if( snmpRMIBerror == SnmpRMIBnoError ) {

		// Note: any tooBig limitations must be ignored as Sets must remain atomic..

		Bool 	tryAllOids = True, tryWriteOnlyOids = False ;
		int	txNAttrs = reqNAttrs ;

		while( tryAllOids == True || tryWriteOnlyOids == True ) {

		    tryWriteOnlyOids = False ;

		    if( (snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata,
					SetOperator, internetOIDs, txNAttrs, PEvalues, attrIds, &operationId,
					reqNAttrs )) == SnmpRMIBnoError ) {

		        if( (snmpRMIBerror = SnmpRMIBreceiveMessageFromAgent( &snmpRMIBdata,
		    						operationId, &currRXresultsp, (int*)0 )) == SnmpRMIBnoError ) {

			    snmpRMIBAgentResultsr.replace( currRXresultsp ) ;

			    //
			    // Is a retry with write only objects necessary...
			    //
			    if( tryAllOids == True &&
				currRXresultsp != NULLSNMPRMIBRXEDRESULTS &&
			        currRXresultsp->OIDspecificSNMPerrorp    != NULL &&
				currRXresultsp->OIDspecificSNMPerrorp[0] == int_SNMP_error__status_noSuchName &&
  				(getReplaceInfop = snmpImageClassInfop->getAttrReplaceInfo()) != NULL ) {
			    	// ..check for "noSuchName" errors..
			    	
				//
			    	// Try an SNMP Set containing just the writeable attributes...
			    	//
			    	GetReplaceType replaceMask = Replace ;
			    	int writeableNattrs = 0, nonWriteableNattrs = 0 ;

			    	for( int i=0 ; writeableNattrs + nonWriteableNattrs != reqNAttrs ; ) {

				    if( (getReplaceInfop[attrIds[i]] & replaceMask) == 0 ) {

				    	++nonWriteableNattrs ;

			    	    	// shuffle down the other attributes..
				    	OID saveInternetOID = internetOIDs[i] ;
				    	PE  savePE		= PEvalues[i] ;
				    	int saveAttrId      = attrIds[i] ;

				    	for( int j=i ; j<reqNAttrs-nonWriteableNattrs ; j++ ) {
				    	    internetOIDs[j] = internetOIDs[j+1] ;
					    PEvalues[j]     = PEvalues[j+1] ;
					    attrIds[j]      = attrIds[j+1] ;
				    	}

				    	// To ensure the malloced storage is freed move 
				    	// it to the non-writeable end of the arrays...
				    	internetOIDs[ reqNAttrs - nonWriteableNattrs ] = saveInternetOID ;
				    	PEvalues[     reqNAttrs - nonWriteableNattrs ] = savePE ;
				    	attrIds[      reqNAttrs - nonWriteableNattrs ] = saveAttrId ;

			            } else {
				    	++writeableNattrs ;
				    	++i ;
				    }
			    	}

			    	if( writeableNattrs > 0 ) {
				    txNAttrs = writeableNattrs ;
				    tryWriteOnlyOids = True ;
				}

			    } else if( currRXresultsp-> numVarBindings > 0 )
			    	successfulCreate = True ;
			}
		    }

		    tryAllOids = False ;

		} // while loop
	    }

	    if( dotrdncp != NULLCP )
	    	(void)free( dotrdncp ) ;
	
	    if( internetOIDs != (OID*)NULL ) {
	    	for( int i=0 ; i<reqNAttrs ; i++ )
	    	    if( internetOIDs[i] != NULLOID )
		    	oid_free( internetOIDs[i] ) ;
	    }
	}

	if( internetOIDs != (OID*)NULL )
	    (void)free( (char*)internetOIDs ) ;

	if( snmpRMIBerror == SnmpRMIBnoError && successfulCreate == False )
	    snmpRMIBerror =  SnmpRMIBprocessingFailure ;

	return( snmpRMIBerror ) ;
}


Bool
SnmpRMIBAgent::isTableFullyTraversed( SnmpImageMOClassInfo* snmpImageClassInfop, 
				      int ncols, OID * rowOIDp, int* rowAttrIds,
 				      int numOperOids, int * OIDspecificSNMPerrorp )
{
Bool	isTableFullyTraversed = False ;
int	numResultsInCorrectColumn=0 ;
OID	iimcAttrOid ;

	if( rowOIDp == NULL || rowAttrIds == NULL || iimcAutoObjAndAttrOID == NULLOID || ncols == 0 ) {

	    isTableFullyTraversed = True ;

	} else {

	    //
	    // Are any of the result OIDs in the correct column...
	    //

	    for( int i=0 ; i<ncols ; i++ ) {

		// Check that the row sub-identifiers are for the correct column...

		iimcAttrOid  = snmpImageClassInfop->getAttrOid(rowAttrIds[i]) ;

		int nAttrSubIds    = iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ;
		int attrCmpCharLen = sizeof( unsigned int ) * nAttrSubIds ;

#ifdef DEBUG2
		fprintf( stderr, "TAB row OID  %s\n", sprintoid( rowOIDp[i]  )) ;
		fprintf( stderr, "TAB attr OID %s\n", sprintoid( iimcAttrOid )) ;
		fprintf( stderr, "TAB Attr cmp len %d\n", attrCmpCharLen ) ;
#endif
	    	if( rowOIDp[i]->oid_nelem > nAttrSubIds && 
			!bcmp( (char*)(rowOIDp[i]->oid_elements),
			       (char*)&(iimcAttrOid->oid_elements[ iimcAutoObjAndAttrOID->oid_nelem ]),
			       attrCmpCharLen ) ) {

		    numResultsInCorrectColumn++ ;
		}
	    }

	    if( numResultsInCorrectColumn == 0 )
	    	isTableFullyTraversed = True ;
	}

	return isTableFullyTraversed ;
}

//
// E.g. when just comparing "RDN" sub-identifiers...
//
static int 
iqaoid_cmp( OID p, OID q, int fromSubId )
{
	p -> oid_nelem    -= fromSubId ;
	p -> oid_elements += fromSubId ;
	q -> oid_nelem    -= fromSubId ;
	q -> oid_elements += fromSubId ;

	int match = oid_cmp( p, q ) ;

	p -> oid_nelem    += fromSubId ;
	p -> oid_elements -= fromSubId ;
	q -> oid_nelem    += fromSubId ;
	q -> oid_elements -= fromSubId ;

	return match ;
}


//
// Return a pointer to a static buffer that contains the lexicographic
// minimum OID, across all the result OIDs for this row...
//
OID
SnmpRMIBAgent::findLexicoMinRowOid( SnmpImageMOClassInfo* snmpImageClassInfop,
			OID * rowOIDp, int ncols, int * rowAttrIds, Bool& resynchRequired )
{
static	unsigned int minRowOidElements[256] ;
static	struct OIDentifier minRowOIDentifier ;
static	OID minRowOID = &minRowOIDentifier ;
OID 	iimcAttrOid  = snmpImageClassInfop->getAttrOid(1) ;
OID	iimcAutoObjAndAttrOID = snmpImageClassInfop->getIimcAutoObjAndAttrOID() ;

	resynchRequired = False ;
	
	minRowOID -> oid_nelem = 0 ;
	minRowOID -> oid_elements = minRowOidElements ;

	if( rowOIDp != (OID*)NULL && ncols > 0 && iimcAutoObjAndAttrOID != NULLOID && 
								iimcAttrOid != NULLOID ) {

#ifdef DEBUG2
	    fprintf( stderr, "iimcAttrOid %s\n", sprintoid( iimcAttrOid )) ;
#endif
	    Bool rdnSubIdsBuffered = False ;
	    int nAttrSubIds    = iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ;
	    int attrCmpCharLen = sizeof( unsigned int ) * nAttrSubIds ;

	    for( int i=0 ; i<ncols ; i++ ) {

		// Check that the row sub-identifiers are for the correct column...

		iimcAttrOid  = snmpImageClassInfop->getAttrOid(rowAttrIds[i]) ;

#ifdef DEBUG2
		fprintf( stderr, "row OID  %s\n", sprintoid( rowOIDp[i] )) ;
		fprintf( stderr, "attr OID %s\n", sprintoid( iimcAttrOid )) ;
		fprintf( stderr, "Attr cmp len %d\n", attrCmpCharLen ) ;
#endif
	    	if( rowOIDp[i]->oid_nelem > nAttrSubIds && 
			!bcmp( (char*)(rowOIDp[i]->oid_elements),
			       (char*)&(iimcAttrOid->oid_elements[ iimcAutoObjAndAttrOID->oid_nelem ]),
			       attrCmpCharLen ) ) {

		    // ...fine, now check the RDN sub-identifiers against the highest found so far...
		    minRowOID -> oid_elements[ nAttrSubIds-1 ] =     // Ensure any OID differences
		    	rowOIDp[i] -> oid_elements[ nAttrSubIds-1 ] ;// are in the RDN
#ifdef DEBUG2		    
		    fprintf( stderr, "Current min OID %s\n", sprintoid( minRowOID ) ) ;
#endif
		    if( rdnSubIdsBuffered == False || 
		    	iqaoid_cmp( minRowOID, rowOIDp[i], nAttrSubIds ) == 1 ) {
#ifdef DEBUG2
			if( rdnSubIdsBuffered == True ) {
			    fprintf( stderr, "RESYNCH required %s\n", sprintoid( rowOIDp[i] )) ;
			    fprintf( stderr, "rowOIDp[i] -> oid_nelem %d attrCmpCharLen %d\n",
			    		rowOIDp[i] -> oid_nelem, attrCmpCharLen ) ;
			}
#endif
			bcopy( 	(char*)rowOIDp[i] -> oid_elements,
				(char*)minRowOidElements,
			      	sizeof( unsigned int ) * rowOIDp[i] -> oid_nelem ) ;

			minRowOID -> oid_nelem = rowOIDp[i] -> oid_nelem ;

		    	if( rdnSubIdsBuffered == True )
			    resynchRequired = True ;
			
			rdnSubIdsBuffered = True ;
		    }
		}
	    }
	}
	
	return minRowOID ;
}

//
// If certain columnar objects are intermittently supported by
// a given SNMPv1 agent then ensure that all row OIDs have the
// same RDN sub-identifiers. Where the RDN is the lexicographic
// minimum across all the current row OIDs. E.g. for IfEntry rows
// CISCO agents only return ifDescr for inactive rows.
//

OID*
SnmpRMIBAgent::resynchRowOIDs( SnmpImageMOClassInfo* snmpImageClassInfop,
		OID * rowOIDp, int ncols, int * rowAttrIds, Bool& newOIDdata )
{
OID *	resynchRowOIDp  = rowOIDp ;
OID	minRowOID ;
Bool	resynchRequired = False ;

	newOIDdata = False ;

	if( (minRowOID = findLexicoMinRowOid( snmpImageClassInfop, rowOIDp, ncols, 
						rowAttrIds, resynchRequired )) != NULLOID ) {

	    if( resynchRequired == True ) {

	    	// Create new OID storage with the lexicographically 
		// minimum RDN sub-identifiers ALWAYS appended...
		newOIDdata = True ;

		OID iimcAttrOid = snmpImageClassInfop->getAttrOid(1) ;
		int nAttrSubIds = iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ;

		if( (resynchRowOIDp = (OID *)malloc( sizeof( OID ) * ncols )) != (OID *)0 ) {

		    for( int i=0 ; i<ncols ; i++ ) {

		    	resynchRowOIDp[i] = oid_cpy( minRowOID ) ; // will retain the rdn sub-ids
								   // from this OID.

#ifdef DEBUG2
			fprintf( stderr, "NEW RESYNCH OID    required %s\n", sprintoid( resynchRowOIDp[i] ) ) ;
			fprintf( stderr, "...OLD RESYNCH OID required %s\n", sprintoid( rowOIDp[i] ) ) ;
#endif
			resynchRowOIDp[i] -> oid_elements[ nAttrSubIds - 1 ] =
				rowOIDp[i] -> oid_elements[ nAttrSubIds - 1] ;

		    }
		}		
	    }
	}

	return resynchRowOIDp ;
}
									

SnmpRMIBerrorEnum
SnmpRMIBAgent::snmpRetrieveNewSubordinateMOattributes( SnmpImageMOClassInfo* snmpImageClassInfop,
						       SnmpRMIBAgentResults& snmpRMIBAgentResultsr,
						       OID* internetOIDs, int reqNAttrs, int* attrIds )
{
SnmpRMIBerrorEnum snmpRMIBerror = SnmpRMIBnoError ;
Bool	tableFullyTraversed = False, firstIteration = True ;
int	operationId, tooBigMaxOidsPerPdu ;
struct	SnmpRMIBrxedResults * currRXresultsp = NULLSNMPRMIBRXEDRESULTS ;
OID 	iimcAttrOid  = snmpImageClassInfop->getAttrOid(1) ;


	// Retrieve the table entry by entry..

	tooBigMaxOidsPerPdu = snmpImageClassInfop->getTooBigMaxOidsPerPdu( _snmpRMIBAgentId ) ;


	while( tableFullyTraversed == False && snmpRMIBerror == SnmpRMIBnoError ) {

	    if( firstIteration == True ) {
	    	snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata, GetNextOperator, 
					internetOIDs, reqNAttrs, (PE *)0, attrIds, &operationId, tooBigMaxOidsPerPdu ) ;
	    } else {
	    
		if( remoteAgentIntermitObjSupport == False ) {

		    snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata, GetNextOperator, 
					currRXresultsp->resOIDvec, currRXresultsp->numVarBindings, 
					(PE *)0, currRXresultsp->resAttrIds, &operationId, tooBigMaxOidsPerPdu ) ;
		} else {

		    Bool newOIDdata  = False ;
		    OID* resynchOIDs = resynchRowOIDs( snmpImageClassInfop, currRXresultsp->resOIDvec,
							currRXresultsp->numVarBindings,
							currRXresultsp->resAttrIds, newOIDdata ) ;

		    if( resynchOIDs == (OID*)NULL )
		    	snmpRMIBerror = SnmpRMIBmemAllocFailure ;
		    else {		 
	    	    	snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata, GetNextOperator, 
					resynchOIDs, currRXresultsp->numVarBindings, 
					(PE *)0, currRXresultsp->resAttrIds, &operationId, tooBigMaxOidsPerPdu ) ;

			if( newOIDdata == True ) {
		    	    for( int f=0 ; f<currRXresultsp->numVarBindings; f++ )
			    	oid_free( resynchOIDs[ f ] ) ;

		    	    (void)free( (char*)resynchOIDs ) ;
		    	}
		    }
		}
	    }

	    if( snmpRMIBerror == SnmpRMIBnoError ) {
	   
		if( (snmpRMIBerror = SnmpRMIBreceiveMessageFromAgent( &snmpRMIBdata,
		    			operationId, &currRXresultsp, &tooBigMaxOidsPerPdu )) == SnmpRMIBnoError ) {

		    if( SnmpRMIBnumResultNodeBindings( currRXresultsp ) == 0 ) {
		    
			tableFullyTraversed = True ;
			SnmpRMIBremoveReceivedResultsNode( &currRXresultsp ) ;
			
		    } else {

#ifdef DEBUG
			advise( NULLCP, "******************" ) ;
			for( int j = 0 ; j<currRXresultsp->numVarBindings;j++ )
			    if( currRXresultsp->resOIDvec[j] )
				advise( NULLCP, "Entry %s", sprintoid( currRXresultsp->resOIDvec[j] )) ;
#endif
			// Check if the table has been fully traversed..
			tableFullyTraversed = isTableFullyTraversed( snmpImageClassInfop,
							currRXresultsp->numVarBindings,
							currRXresultsp->resOIDvec,
	 						currRXresultsp->resAttrIds,
 							currRXresultsp->numOperOids,
							currRXresultsp->OIDspecificSNMPerrorp );

			if( tableFullyTraversed == False ) {
			    // buffer the results for subsequent retrieval..
			    Bool dummy;
			    OID  resOIDtoBeChecked = findLexicoMinRowOid( snmpImageClassInfop, 
			    						  currRXresultsp->resOIDvec,
					 	    			  currRXresultsp->numVarBindings, 
									  currRXresultsp->resAttrIds, dummy ) ;

			    snmpRMIBAgentResultsr.append( currRXresultsp, resOIDtoBeChecked,
					iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ) ;

			} else {
			    SnmpRMIBremoveReceivedResultsNode( &currRXresultsp ) ;
			    // ...free this storage
			}
		    }
		}
	    }

	    firstIteration = False ;

	} // while loop

	snmpImageClassInfop->setTooBigMaxOidsPerPdu( _snmpRMIBAgentId, tooBigMaxOidsPerPdu ) ;

	return snmpRMIBerror ;
}


// 0  uint1p sub-identifiers are lexicographically equal   to the uint2p sub-identifiers 
// 1  uint1p sub-identifiers are lexicographically greater than the uint2p sub-identifiers 
// -1 uint1p sub-identifiers are lexicographically less    than the uint2p sub-identifiers 
int
uint_cmp( unsigned int *uint1p, unsigned int *uint2p, int numuints )
{
int	match = 0 ;

	if( uint1p == NULL )
	    if( uint2p == NULL )
	    	match = 0 ;
	    else
	    	match = -1 ;
	else
	    if( uint2p == NULL )
	    	match = 1 ;
	    else {
	    	for( int i=0 ; i < numuints ; i++ ) {
	
		    if( uint1p[i] == uint2p[i] )
		    	continue ;
		    else if( uint1p[i] <  uint2p[i] ) {
		    	match = -1 ;
			break ;
		    } else {
		    	match = 1 ;
			break ;
		    }
		}
	    }

	return match ;
}

//
// Note: we can not assume that the first existing entry
// is still the first entry at the remote agent so the Get-Next
// that is used to retrieve the first entry doesnot specify any RDN
// sub-identifiers.
//
// ASSUMPTION : assumes the GMS entries use the same lexicographic ordering
// as the remote SNMP MIB as only the final thread is checked for having
// passed the end of the table.
//
// Reminder: Since Get-Nexts are used the retrieval based on an existing 
// entries RDN sub-identifiers will actually retrieve the NEXT entry.
//
// Note: The OIDs pointed to by internetOIDs[] will be freed by this method
//       (though not the array pointer itself.)
//
SnmpRMIBerrorEnum
SnmpRMIBAgent::snmpRetrieveRefreshSubordinateMOattributes( SnmpImageMOClassInfo* snmpImageClassInfop,
							   SnmpRMIBAgentResults& snmpRMIBAgentResultsr, 
							   OID* internetOIDs, int reqNAttrs, int* attrIds,
							   SnmpImageAccess* firstMO )
{
SnmpRMIBerrorEnum snmpRMIBerror = SnmpRMIBnoError ;
Bool	tableFullyTraversed = False, firstIteration = True ;
int	operationId ;
struct	SnmpRMIBrxedResults * currRXresultsp = NULLSNMPRMIBRXEDRESULTS ;
SnmpImageAccess* currMO = firstMO ;
int	tooBigMaxOidsPerPdu = 1 ;
int	bulkArrayMaxUsedThreadIndex = -1 ;
static	int   bulkArrayTotalSize = BULKARRAYINCSIZE ; 
static	unsigned int** bulkRdnSubIds = (unsigned int**)0 ;
static	int * bulkRdnNumSubIds   = (int*)0  ;
static	int * bulkOperationIds 	 = (int*)0  ;
static	Bool* bulkThreadOperationalp	 = (Bool*)0 ;
static 	Bool  bulkDynamicVarsInitialised = False ;

OID 	iimcAttrOid  = snmpImageClassInfop->getAttrOid(1) ;

	// Retrieve the table MULTIPLE entries at a time..

	/*
	 * Perform retrievals upon every Nth entry..
	 */

	if( bulkDynamicVarsInitialised == False ) {

	    if( (bulkRdnSubIds = ( unsigned int**)malloc(sizeof(unsigned int*)*bulkArrayTotalSize)) == (unsigned int**)0 )
	    	snmpRMIBerror = SnmpRMIBmemAllocFailure ;

	    else   if( (bulkRdnNumSubIds = (int*)malloc(sizeof(int)*bulkArrayTotalSize)) == (int*)0 ){

		(void)free((char*) bulkRdnSubIds ) ;
	    	snmpRMIBerror = SnmpRMIBmemAllocFailure ;

	    } else if( (bulkOperationIds = (int*)malloc(sizeof(int)*bulkArrayTotalSize)) == (int*)0 ){

		(void)free((char*) bulkRdnSubIds ) ;
		(void)free((char*) bulkRdnNumSubIds ) ;
	    	snmpRMIBerror = SnmpRMIBmemAllocFailure ;

	    } else if( (bulkThreadOperationalp = (Bool*)malloc(sizeof(Bool)*bulkArrayTotalSize)) == (Bool*)0 ){

		(void)free((char*) bulkRdnSubIds ) ;
		(void)free((char*) bulkRdnNumSubIds ) ;
		(void)free((char*) bulkOperationIds ) ;
	    	snmpRMIBerror = SnmpRMIBmemAllocFailure ;

	    } else bulkDynamicVarsInitialised = True ;
	}
 	
	if( bulkDynamicVarsInitialised == True ) {

	    tooBigMaxOidsPerPdu = snmpImageClassInfop->getTooBigMaxOidsPerPdu( _snmpRMIBAgentId ) ;
	    
#ifdef DEBUG
	    advise( NULLCP, "Refresh tables tooBigSize %d Nth entry retrieval for %d", tooBigMaxOidsPerPdu, BULKRETRIEVENTHENTRIES ) ;
#endif
	    // 
	    // Transmit the initial entry retrieval requests..
	    //
	    currMO = firstMO ;

	    for( int entryNumber=0 ; currMO != NULL ; ) {

		// For the first entry only we can utilise all of the method parameters else..
		if( entryNumber > 0 ) {

		    // We must prepare the retrieval request parameters from
		    // the existing entry MO's state..

	    	    // Set up the required attribute OIDs..
		    char *	rdncp = currMO -> getRDNString() ;
		    snmpRMIBerror = createInternetOIDsandAppendRDNstr( snmpImageClassInfop, 
								internetOIDs, reqNAttrs, attrIds, rdncp ) ;

		    if( rdncp != NULLCP )
			(void)free( rdncp ) ;

		    if( snmpRMIBerror != SnmpRMIBnoError )
			break ;
		}

	    	if( (snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata, GetNextOperator,
					internetOIDs, reqNAttrs, (PE *)0, attrIds, &operationId, tooBigMaxOidsPerPdu )) != SnmpRMIBnoError )
		    break ;
			
#ifdef DEBUG
		advise( NULLCP, "Initial refresh TX for %s", sprintoid( internetOIDs[0] ) ) ;
#endif
		// prepare to buffer the send parameters..
		if( ++bulkArrayMaxUsedThreadIndex >= bulkArrayTotalSize ) {

		    bulkArrayTotalSize += BULKARRAYINCSIZE ;

		    if( (bulkRdnSubIds    = (unsigned int**)realloc( (malloc_t)bulkRdnSubIds,
		    				sizeof(unsigned int*)*bulkArrayTotalSize)) == (unsigned int**)0 ||
			(bulkRdnNumSubIds = (int*)realloc( (malloc_t)bulkRdnNumSubIds,
		    				sizeof(int)*bulkArrayTotalSize)) == (int*)0    ||
		        (bulkOperationIds = (int*)realloc( (malloc_t)bulkOperationIds,
		    				sizeof(int)*bulkArrayTotalSize)) == (int*)0    ||
		        (bulkThreadOperationalp = (Bool*)realloc( (malloc_t)bulkThreadOperationalp,
		    				sizeof(Bool)*bulkArrayTotalSize)) == (Bool*)0      ) {

			// A realloc failure occurred..
			if( bulkRdnSubIds != (unsigned int**)0 ) {
	    		    for( int i=0 ; i<bulkArrayMaxUsedThreadIndex ; i++ )
				(void)free((char*)bulkRdnSubIds[i] ) ;

			    (void)free((char*)bulkRdnSubIds ) ;
			}

			if( bulkRdnNumSubIds != (int*)0 )
			    (void)free((char*)bulkRdnNumSubIds ) ;

			if( bulkOperationIds != (int*)0 )
			    (void)free((char*)bulkOperationIds ) ;

			if( bulkThreadOperationalp != (Bool*)0 )
			    (void)free((char*)bulkOperationIds ) ;

		        bulkArrayMaxUsedThreadIndex = -1 ;
			bulkDynamicVarsInitialised = False ;

			snmpRMIBerror = SnmpRMIBmemAllocFailure ;
			
			break ;
		    }
		}

		// Extract the RDN sub-identifiers off of any of this entries internet OIDs..
		if( entryNumber > 0 ) {

		    // ..we do not have the sub-identifiers for the first entries RDN..

		    int rdnNumSubIds = internetOIDs[0]->oid_nelem - (iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem) ;

		    if( (bulkRdnSubIds[ bulkArrayMaxUsedThreadIndex ] = (unsigned int*)malloc( sizeof(unsigned int) * rdnNumSubIds )) == (unsigned int*)0 ) {
		    	snmpRMIBerror = SnmpRMIBmemAllocFailure ;
		    	break ;
		    } else {

			// Block copy just the RDN sub-identifiers..
			bcopy( (char*)&(internetOIDs[0]->oid_elements[ iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ]),
			       (char*)bulkRdnSubIds[ bulkArrayMaxUsedThreadIndex ],
			       sizeof(unsigned int) * 
			           (internetOIDs[0]->oid_nelem - (iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem) ) ) ;

			bulkRdnNumSubIds[bulkArrayMaxUsedThreadIndex] = rdnNumSubIds ;
		    }
		}

		bulkOperationIds[ bulkArrayMaxUsedThreadIndex ] = operationId ;
		bulkThreadOperationalp[ bulkArrayMaxUsedThreadIndex ] = True ;

		// 
		// Prepare for the next loop iteration...
		//
		for( int i=0 ; i<reqNAttrs ; i++ )
		    if( internetOIDs[ i ] != NULLOID ) {
			oid_free( internetOIDs[ i ] ) ;
			internetOIDs[ i ] = NULLOID ;
		    }
		
	    	for( int j=0 ; j<BULKRETRIEVENTHENTRIES && currMO != NULL ; ++j ) {
		
		    currMO = (SnmpImageAccess*)currMO->getClassNext( snmpImageClassInfop ) ;
		    ++entryNumber ;
		}
		
	    
	    } // for loop


	    if( snmpRMIBerror != SnmpRMIBnoError ) {

		for( int i=0 ; i<reqNAttrs ; i++ )
		    if( internetOIDs[ i ] != NULLOID ) {
			oid_free( internetOIDs[ i ] ) ;
			internetOIDs[ i ] = NULLOID ;
		    }
	    
	    } else {

	    	// Now start receiving the retrieval responses..

		// Terminate any "threads" that that fall off the table
		// or return results that have already been received..
		int numThreadsRemaining = bulkArrayMaxUsedThreadIndex+1 ;

		while( numThreadsRemaining>0 && snmpRMIBerror == SnmpRMIBnoError ) {

		    struct SnmpRMIBrxedResults * currRXresultsp = NULLSNMPRMIBRXEDRESULTS ;

		    for( int currThreadNum=0 ; numThreadsRemaining>0 &&
		    			   snmpRMIBerror == SnmpRMIBnoError &&
		    			   currThreadNum<=bulkArrayMaxUsedThreadIndex ; currThreadNum++ ) {

			if( bulkThreadOperationalp[ currThreadNum ] == True &&
			    (snmpRMIBerror = SnmpRMIBreceiveMessageFromAgent( &snmpRMIBdata,
		    			bulkOperationIds[currThreadNum], &currRXresultsp, &tooBigMaxOidsPerPdu )) == SnmpRMIBnoError &&
					currRXresultsp->resOIDvec != NULL ) {
			
		    	    Bool continueThread = True ;

			    if( SnmpRMIBnumResultNodeBindings( currRXresultsp ) == 0 ) {

			    	continueThread =  False ;
				SnmpRMIBremoveReceivedResultsNode( &currRXresultsp ) ;
#ifdef DEBUG
				advise(NULLCP, "Thread %d terminated remote results invalid or empty" ) ;
#endif

			    } else if( currThreadNum == bulkArrayMaxUsedThreadIndex ) {

				// Last thread - have we gone off the end of the table ?
				Bool    tableFullyTraversed = False ;
				
				tableFullyTraversed = isTableFullyTraversed( snmpImageClassInfop,
								currRXresultsp->numVarBindings,
								currRXresultsp->resOIDvec,
		 						currRXresultsp->resAttrIds,
 								currRXresultsp->numOperOids,
								currRXresultsp->OIDspecificSNMPerrorp );

				if( tableFullyTraversed == False ) {
				    // buffer the results for subsequent retrieval..
				    Bool dummy ;
				    OID resOIDtoBeChecked = findLexicoMinRowOid( snmpImageClassInfop, currRXresultsp->resOIDvec,
 	    					currRXresultsp->numVarBindings, currRXresultsp->resAttrIds, dummy ) ;
				    snmpRMIBAgentResultsr.insert( currRXresultsp, resOIDtoBeChecked,
					iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ) ;

				} else
				    SnmpRMIBremoveReceivedResultsNode( &currRXresultsp ) ;
			    	    // free this storage

			    	if( tableFullyTraversed == True  )
				    continueThread = False ;

			    } else {
			    
			    	// Has this thread returned the RDN that commenced the next thread ?
				// Has this thread passed the next commencement RDN ?
				Bool dummy ;
				OID resOIDtoBeChecked = findLexicoMinRowOid( snmpImageClassInfop, currRXresultsp->resOIDvec,
 	    				currRXresultsp->numVarBindings, currRXresultsp->resAttrIds, dummy ) ;
			        int  rdnNumSubIds = resOIDtoBeChecked->oid_nelem - (iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem) ;
				int  matchResult  = 1 ;

				if( rdnNumSubIds != bulkRdnNumSubIds[ currThreadNum + 1 ] ||
				    (matchResult = uint_cmp( &(resOIDtoBeChecked->oid_elements[iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem]),
					                     bulkRdnSubIds[ currThreadNum + 1 ],
					                     rdnNumSubIds )) >= 0  ) {
				    continueThread = False ;
#ifdef DEBUG
				    advise(NULLCP, "Thread %d terminated HIT thread %d", currThreadNum, currThreadNum+1 ) ;
#endif
				}

				if( matchResult <= 0 ) {

				    snmpRMIBAgentResultsr.insert( currRXresultsp, resOIDtoBeChecked,
			    			iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ) ;
#ifdef DEBUG
				    advise( NULLCP, "inserting refresh entry %s", sprintoid( currRXresultsp->resOIDvec[0] )) ;
#endif
				} else {
#ifdef DEBUG
				    advise( NULLCP, "discarding refresh entry %s", sprintoid( currRXresultsp->resOIDvec[0] )) ;
#endif
				    SnmpRMIBremoveReceivedResultsNode( &currRXresultsp ) ;
				}   
			    }

			    if( continueThread == False ) {

			        bulkThreadOperationalp[ currThreadNum ] = False ;
			    	--numThreadsRemaining ;

			    } else {

			    	// Generate a new retrieval request..

				if( remoteAgentIntermitObjSupport == False ) {

				    snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata, GetNextOperator, 
					currRXresultsp->resOIDvec, currRXresultsp->numVarBindings, 
					(PE *)0, currRXresultsp->resAttrIds, &operationId, tooBigMaxOidsPerPdu ) ;
				} else {
				
				    Bool newOIDdata  = False ;
				    OID* resynchOIDs = resynchRowOIDs( snmpImageClassInfop, currRXresultsp->resOIDvec,
							currRXresultsp->numVarBindings,
							currRXresultsp->resAttrIds, newOIDdata ) ;

				    if( resynchOIDs == (OID*)NULL )
					snmpRMIBerror = SnmpRMIBmemAllocFailure ;
				    else {		 
			    	    	snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata, GetNextOperator, 
					    resynchOIDs, currRXresultsp->numVarBindings, 
					    (PE *)0, currRXresultsp->resAttrIds, &operationId, tooBigMaxOidsPerPdu ) ;

				    	if( newOIDdata == True ) {
				    	    for( int f=0 ; f<currRXresultsp->numVarBindings; f++ )
					    	oid_free( resynchOIDs[ f ] ) ;

				    	    (void)free( (char*)resynchOIDs ) ;
				    	}
				    }
				}
#ifdef DEBUG
				advise( NULLCP, "Refresh TX for %s for thread %d", sprintoid( currRXresultsp->resOIDvec[0] ),
 						currThreadNum ) ;
#endif
				if( snmpRMIBerror != SnmpRMIBnoError )
				    break ;
				else {

				    // Buffer the requests operationId..
				    bulkOperationIds[ currThreadNum ] = operationId ;
				}
			    }
			}		    
		    }	    
	        }
	    }

	    // Free the non-static storage (that was not passed as a method parameter)..
	    for( int i=1 ; i<=bulkArrayMaxUsedThreadIndex ; i++ )
		(void)free( (char*)bulkRdnSubIds[ i ] ) ;

	    snmpImageClassInfop->setTooBigMaxOidsPerPdu( _snmpRMIBAgentId, tooBigMaxOidsPerPdu ) ;
	}
	
	return snmpRMIBerror ;
}


//
// For the moment - blocking only
//
// Note: will not attempt to retrieve attribute Id 0 (the naming attribute)
//
// Note: the OIDs within the returned bindings will not be converted
// to their IIMC equivalent as the search for results routine allows for
// this.
//
SnmpRMIBerrorEnum 
SnmpRMIBAgent::snmpGetMultiInstancedMOs( SnmpImageMOClassInfo* snmpImageClassInfop,
					 SnmpImageAccess * superiorMO,
					 SnmpRMIBAgentResults& snmpRMIBAgentResultsr )
{
SnmpRMIBerrorEnum snmpRMIBerror = SnmpRMIBnoError ;
OID *	iimcOIDs = (OID *)0 ;
OID *	internetOIDs ;
int	maxNAttrs, reqNAttrs, attrIdOfNamingAttribute=-1 ;
int *	attrIds ;
ObjectSupportedType * attrIdIsSupportedByAgentp = (ObjectSupportedType *)0;
OID	MocClassOid = NULLOID ;
int	i ;

#ifdef DEBUG
	struct timeval  tvs, now;
	(void) gettimeofday (&tvs, (struct timezone *) 0);
#endif
	if( snmpRMIBAgentInitialised == False )
	    snmpRMIBerror = SnmpRMIBagentNotInitialised ;

	else if( snmpImageClassInfop == NULLIMAGEINFO )
	    snmpRMIBerror = SnmpRMIBnullParameter ;

	else if( remoteAgentIntermitObjSupport == False &&
		 (attrIdIsSupportedByAgentp = 
		 	snmpImageClassInfop->getAttrSupportedByAgentFlags(_snmpRMIBAgentId))==(ObjectSupportedType *)0)

	    snmpRMIBerror = SnmpRMIBunableToAccessSupportData ;

	else {

	    maxNAttrs = snmpImageClassInfop->getNAttrs() ;

	    reqNAttrs = maxNAttrs - 1 ; // excluding the naming attribute

	    if( remoteAgentIntermitObjSupport == False ) {

	    	// Determine the number of Internet objects to be retrieved
	    	for( i=1 ; i < maxNAttrs ; i++ )
		    if( attrIdIsSupportedByAgentp[ i ] == notSupported )
		    	reqNAttrs-- ;
	    }

	    attrIds = (int *)malloc( sizeof( int ) * reqNAttrs ) ;

	    if( reqNAttrs <= 0 ) {
#ifdef DEBUG
		advise( NULLCP, "Call to SnmpRMIBAgent::snmpGetMultiInstancedMOs() with no attributes to be retrieved" ) ;
#endif
	    } else if( attrIds == (int *)0 ) {

	    	snmpRMIBerror = SnmpRMIBmemAllocFailure ;
		
	    } else {

		if( (iimcOIDs = (OID *)malloc( sizeof( OID ) * reqNAttrs )) == (OID *)0 )
		    snmpRMIBerror = SnmpRMIBmemAllocFailure ;

		else {

		    int nextUnusedIndex = 0 ;

	    	    // Set up the required attribute OIDs..
		    for( i=1 ; i < maxNAttrs ; i++ ) {

			if( remoteAgentIntermitObjSupport == True ||
				attrIdIsSupportedByAgentp[ i ] != notSupported ) {

			    iimcOIDs[ nextUnusedIndex ] = snmpImageClassInfop->getAttrOid( i ) ;
			    // ..Note: the MOClassInfo OIDs are NOT copied

			    attrIds[ nextUnusedIndex ] = i ;
			    ++nextUnusedIndex ;
			}
		    }

		    // Now convert these IIMC OIDs to their Internet equivalent..
		    if( (snmpRMIBerror = convertOIDsIIMCtoInternet( iimcOIDs, reqNAttrs, 
							&internetOIDs )) == SnmpRMIBnoError ) {

			SnmpImageAccess * firstMO = (SnmpImageAccess*)superiorMO->getClassFirst( snmpImageClassInfop ), * currMO = firstMO ;

			// Count the number of existing MOs of this class...
			int  numMocMOs ;
			for( numMocMOs = 0 ;
				currMO != NULLMO ; currMO = (SnmpImageAccess*)currMO->getClassNext( snmpImageClassInfop ), ++numMocMOs )
			    ; // EMPTY BODY

			// Do we have any existing MOs..
			if( numMocMOs == 0 )
			    snmpRMIBerror = snmpRetrieveNewSubordinateMOattributes( snmpImageClassInfop,
				 	        snmpRMIBAgentResultsr, internetOIDs, reqNAttrs, attrIds ) ;
			else 
			    snmpRMIBerror = snmpRetrieveRefreshSubordinateMOattributes( snmpImageClassInfop,
					 	snmpRMIBAgentResultsr, internetOIDs, reqNAttrs, attrIds, firstMO ) ;
		
		    	// Return the storage created during the OID conversion..
		    	for( i=0 ; i<reqNAttrs ; i++ )
			    if( internetOIDs[ i ] != NULLOID )
				oid_free( internetOIDs[ i ] ) ;

		    	(void)free( (char*)internetOIDs ) ;
	    	    }

		    (void)free((char *)iimcOIDs ) ;
	    	}

		(void)free( (char*)attrIds ) ;
	    }
	}

#ifdef DEBUG
	(void) gettimeofday (&now, (struct timezone *) 0);
        now.tv_sec -= tvs.tv_sec;
	if ((now.tv_usec -= tvs.tv_usec) < 0)
            now.tv_sec--, now.tv_usec += 1000000;

	advise (NULLCP, "%d entries retrieved in %d.%06d seconds",
		snmpRMIBAgentResultsr.getCount(), now.tv_sec, now.tv_usec );
#endif	
	return( snmpRMIBerror ) ;
}


//
// For the moment - blocking only
//
// Note: will not attempt to retrieve attribute Id 0 (the naming attribute)
//
// Note: the OIDs within the returned bindings will not be converted
// to their IIMC equivalent as the search for results routine allows for
// this.
//
SnmpRMIBerrorEnum 
SnmpRMIBAgent::snmpGetOneMultiInstancedMO( SnmpImageMOClassInfo* snmpImageClassInfop,
					   SnmpRMIBAgentResults& snmpRMIBAgentResultsr,
 					   RDN rdn )
{
SnmpRMIBerrorEnum snmpRMIBerror = SnmpRMIBnoError ;
OID *	internetOIDs ;
int	operationId, maxNAttrs, reqNAttrs, attrIdOfNamingAttribute = -1 ;
int *	attrIds ;
ObjectSupportedType * attrIdIsSupportedByAgentp = (ObjectSupportedType *)0;
OID	MocClassOid = NULLOID ;
struct	SnmpRMIBrxedResults  * currRXresultsp ;
char *	dotrdncp ;
int	i ;

	if( snmpRMIBAgentInitialised == False )
	    snmpRMIBerror = SnmpRMIBagentNotInitialised ;

	else if( snmpImageClassInfop == NULLIMAGEINFO || rdn == NULLRDN )
	    snmpRMIBerror = SnmpRMIBnullParameter ;

	else if(remoteAgentIntermitObjSupport == False && 
		(attrIdIsSupportedByAgentp = 
		 	snmpImageClassInfop->getAttrSupportedByAgentFlags(_snmpRMIBAgentId))==(ObjectSupportedType *)0)

	    snmpRMIBerror = SnmpRMIBunableToAccessSupportData ;

	else if( (dotrdncp = rdn2str( rdn )) != NULLCP ) {

	    maxNAttrs = snmpImageClassInfop->getNAttrs() ;

	    reqNAttrs = maxNAttrs - 1 ; // excluding the naming attribute

	    if( remoteAgentIntermitObjSupport == False ) {

		// Determine the number of Internet objects to be retrieved
	    	for( i=1 ; i < maxNAttrs ; i++ )
		    if( attrIdIsSupportedByAgentp[ i ] == notSupported )
			reqNAttrs-- ;
	    }

	    attrIds = (int *)malloc( sizeof( int ) * reqNAttrs ) ;

	    if( reqNAttrs <= 0 ) {
#ifdef DEBUG
		advise( NULLCP, "Call to SnmpRMIBAgent::snmpGetOneMultiInstancedMO() with no attributes to be retrieved" ) ;
#endif
	    } else if( attrIds == (int *)0 ) {

	    	snmpRMIBerror = SnmpRMIBmemAllocFailure ;
		
	    } else {

		if( (internetOIDs = (OID *)malloc( sizeof( OID ) * reqNAttrs )) == (OID *)0 )
		    snmpRMIBerror = SnmpRMIBmemAllocFailure ;

		else {	    

		    int nextUnusedIndex = 0 ;
 
		    // Set up the required attribute OIDs..
		    for( i=1 ; i < maxNAttrs ; i++ ) {

			if( remoteAgentIntermitObjSupport == True ||
				attrIdIsSupportedByAgentp[ i ] != notSupported ) {

			    attrIds[ nextUnusedIndex ] = i ;
			    ++nextUnusedIndex ;
			}
		    }

		    // Now convert these IIMC OIDs to their Internet equivalent..
		    if( (snmpRMIBerror = createInternetOIDsandAppendRDNstr( snmpImageClassInfop,
		                                              internetOIDs, reqNAttrs, attrIds, dotrdncp )) == SnmpRMIBnoError ) {

			int tooBigMaxOidsPerPdu = snmpImageClassInfop->getTooBigMaxOidsPerPdu( _snmpRMIBAgentId ) ;

	    		if( (snmpRMIBerror = SnmpRMIBsendMessageToAgent( &snmpRMIBdata, GetOperator, 
					internetOIDs, reqNAttrs, (PE *)0, attrIds, &operationId, tooBigMaxOidsPerPdu )) == SnmpRMIBnoError ) {

			    if( (snmpRMIBerror = SnmpRMIBreceiveMessageFromAgent( &snmpRMIBdata,
		    			operationId, &currRXresultsp, &tooBigMaxOidsPerPdu )) == SnmpRMIBnoError ) {

			    	// Buffer the results for subsequent retrieval..
				snmpRMIBAgentResultsr.replace( currRXresultsp ) ;

	    		    	snmpImageClassInfop->setTooBigMaxOidsPerPdu( _snmpRMIBAgentId, tooBigMaxOidsPerPdu ) ;
			    }
			}

		    	// Return the storage created during the OID conversion..
		    	for( i=0 ; i<reqNAttrs ; i++ )
			    oid_free( internetOIDs[ i ] ) ;

		    	(void)free( (char*)internetOIDs ) ;
	    	    }
	    	}

		(void)free( (char*)attrIds ) ;
	    }

	    (void)free( dotrdncp ) ;
	}

	return( snmpRMIBerror ) ;
}


//
// Determine whether the current SNMP results contains a PE for the
// required attribute Id. If so return it in allocated storage.
//

PE
SnmpRMIBAgentResults::snmpExtractResult( int requiredAttrId,
				  SnmpImageMOClassInfo * snmpImageClassInfop,
				  SnmpRMIBAgent * snmpRMIBAgentp,
				  CMISErrors* cmisErrorp )
{
PE	extractedPE = NULLPE ;
int	resReqAttrIdIndex = -1, operReqAttrIdIndex = -1 ;
OID	attrOid = NULLOID ;
struct  SnmpRMIBrxedResults * snmpRMIBreceivedResultsp ;
ObjectSupportedType * attrIdIsSupportedByAgentp = (ObjectSupportedType *)0 ;
CMISErrors dummyError ;

	if( cmisErrorp != (CMISErrors *)0 )
	    *cmisErrorp = m_processingFailure ;
	else cmisErrorp = &dummyError ;
	
	// Check the requested attribute Id...
	if( snmpImageClassInfop && requiredAttrId < snmpImageClassInfop->getNAttrs() &&
    			(snmpRMIBreceivedResultsp = _currresp) != NULLSNMPRMIBRXEDRESULTS ) {

	    // check the requested AttrIds..
	    register int i;
	    for( i=0 ; i<snmpRMIBreceivedResultsp->numOperOids ; i++ ) {
		if( snmpRMIBreceivedResultsp->operAttrIds[ i ] == requiredAttrId ) {
		    operReqAttrIdIndex = i ;
		    break ;
		}
	    }

	    if( operReqAttrIdIndex != -1 ) {
		// check the response  AttrIds..
		for( i=0 ; i<snmpRMIBreceivedResultsp->numVarBindings ; i++ ) {
		    if( snmpRMIBreceivedResultsp->resAttrIds[ i ] == requiredAttrId ) {
			resReqAttrIdIndex = i ;
			break ;
		    }
		}

		if( snmpRMIBreceivedResultsp->numVarBindings > 0 &&
			resReqAttrIdIndex == -1 && 
			snmpRMIBreceivedResultsp->snmpOperation != GetNextOperator &&
			(attrIdIsSupportedByAgentp =
				snmpImageClassInfop->getAttrSupportedByAgentFlags( snmpRMIBAgentp->getAgentId() )) !=
					(ObjectSupportedType *)0 ){

		    if( snmpRMIBreceivedResultsp->OIDspecificSNMPerrorp[ operReqAttrIdIndex ] ==
									int_SNMP_error__status_noSuchName )
			attrIdIsSupportedByAgentp[ requiredAttrId ] = notSupported ;
		}
	    }

	    /* Check that a response for the required attribute Id has indeed been received.. */
	    if( resReqAttrIdIndex == -1 ) {

		if( snmpRMIBreceivedResultsp->snmpOperation == SetOperator )
		    *cmisErrorp = m_setListError ;
		else
		    *cmisErrorp = m_getListError ;
	    
	    } else {

	    	/*
		 * CHECK THE RESPONSE OID THAT SHOULD CORRESPOND TO THE
		 * REQUIRED ATTRIBUTE ID, TO ENSURE THAT THEY PAIR OFF..
		 */

		// Form the sub-identifiers that are to be used in this extraction...
		if( (attrOid = snmpImageClassInfop->getAttrOid( requiredAttrId )) != NULLOID ) {

		    OIDentifier checkOIDentifier ;
		    OID	iimcAutoObjAndAttrOID ;

		    iimcAutoObjAndAttrOID = snmpRMIBAgentp->getIimcAutoObjAndAttrOID() ;

		    // Exclude the { iimcAutoObjAndAttr } pre-pended sub-identifiers..
		    checkOIDentifier.oid_elements = &(attrOid->oid_elements[iimcAutoObjAndAttrOID->oid_nelem]) ;
		    checkOIDentifier.oid_nelem    = attrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem ;

		    Bool continueWithComparison = True, multiInstanced ;
		    int	resOIDvecOid_nelem = snmpRMIBreceivedResultsp->resOIDvec[ resReqAttrIdIndex ]->oid_nelem;
		    
		    if( (multiInstanced = snmpImageClassInfop->getIsMultiInstanced()) == True ) {

			if( checkOIDentifier.oid_nelem >= resOIDvecOid_nelem )
			    continueWithComparison = False ; // RDN sub-identifier(s) increase OID length

		    } else {
			// single instanced ( "+1" for the final ".0"..
			if( checkOIDentifier.oid_nelem + 1 != resOIDvecOid_nelem )
			    continueWithComparison = False ;
		    }
		    
		    if( continueWithComparison == True &&
			!bcmp( (char*)snmpRMIBreceivedResultsp->resOIDvec[resReqAttrIdIndex]->oid_elements,
		    	       (char*)checkOIDentifier.oid_elements,
			       sizeof(unsigned int) * checkOIDentifier.oid_nelem ) ) {

			if( multiInstanced == True || 
			    snmpRMIBreceivedResultsp->resOIDvec[ resReqAttrIdIndex ]->
			    				oid_elements[ resOIDvecOid_nelem-1 ] == 0 ) {
			    /* COPY THE RESPONSE PE.. */
			    if( multiInstanced == True &&
			    	snmpRMIBreceivedResultsp->resOIDforCompare != NULLOID &&
				snmpRMIBreceivedResultsp->resOIDforCompare -> oid_nelem > 0 ) {

				// Strictly only needed for a resynch case (e.g. a CISCO agent) !

				if( iqaoid_cmp( snmpRMIBreceivedResultsp->resOIDforCompare,
						snmpRMIBreceivedResultsp->resOIDvec[resReqAttrIdIndex],
						(attrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem)
					) == 0 ) { // ...RDN sub-identifiers match (correct entry !)

				    // PE did not come from an unsynched entry attribute...
				    extractedPE = pe_cpy( snmpRMIBreceivedResultsp->resPEvec[ resReqAttrIdIndex ] ) ;

				} // else NULLPE returned !

			    } else extractedPE = pe_cpy( snmpRMIBreceivedResultsp->resPEvec[ resReqAttrIdIndex ] ) ;

			    *cmisErrorp = m_noError ;
#ifdef DEBUG1
fprintf( stderr, "Extracted OID %s\n",
			    sprintoid( snmpRMIBreceivedResultsp->resOIDvec[ resReqAttrIdIndex ] ) ) ;
#endif		    	    			
			}
		    }

		    if( extractedPE == NULLPE ) {

			/* HANDLE ATTRIBUTE ID RETRIEVAL FAILURE */

			// For a Get-Next operation check for unsupported objects here...
			if( snmpRMIBreceivedResultsp->snmpOperation == GetNextOperator &&
			    (attrIdIsSupportedByAgentp =
			    	snmpImageClassInfop->getAttrSupportedByAgentFlags( snmpRMIBAgentp->getAgentId() )) !=
										(ObjectSupportedType *)0 ){

			    if( attrIdIsSupportedByAgentp != (ObjectSupportedType *)0 &&
		    	        snmpRMIBreceivedResultsp->OIDspecificSNMPerrorp[ operReqAttrIdIndex ] !=
										int_SNMP_error__status_timeOut )
				attrIdIsSupportedByAgentp[ requiredAttrId ] = notSupported ;
			}

			/* PREPARE THE ERROR RESPONSE.. */
			switch( snmpRMIBreceivedResultsp->OIDspecificSNMPerrorp[ operReqAttrIdIndex ] ) {

			    case int_SNMP_error__status_timeOut :
			    case int_SNMP_error__status_tooBig  :
			    case int_SNMP_error__status_genErr  :
				*cmisErrorp = m_processingFailure ;
			    break ;

			    case int_SNMP_error__status_noSuchName :
			    case int_SNMP_error__status_badValue   :
				if( snmpRMIBreceivedResultsp->snmpOperation == SetOperator )
				    *cmisErrorp = m_setListError ;
				else
				    *cmisErrorp = m_getListError ;
			    break ;

			    case int_SNMP_error__status_noError :
			    break ;

			    default : advise( NULLCP, "Warning unknown response error number %d in SnmpRMIBAgent::snmpExtractResult()", 
 					snmpRMIBreceivedResultsp->OIDspecificSNMPerrorp[ operReqAttrIdIndex ] );
			    break ;

			} // of switch
	    	    }
		}
	    }
	}
	
	return( extractedPE ) ;
}


////////////////////////////////////////////////////////////////////////////////
//
// SnmpRMIBAgentResults
//
////////////////////////////////////////////////////////////////////////////////


//
// Cause the RDN component sub-identifiers to be appended to the rdn buffer
// parameter in "dot" format...
//
int
SnmpRMIBAgentResults::appendRDNstrToBuffer( SnmpImageMOClassInfo* imageInfop, char * rdnBuffer )
{
int	status = NOTOK ;

	if( _currresp != NULLSNMPRMIBRXEDRESULTS && imageInfop != NULLIMAGEINFO &&
						    _currresp->resOIDvec != NULL ) {

	    // Ensure that we extract the RDN sub-identifiers from
	    // the lexicographic minimum...

	    Bool dummy ;
	    OID internetOid = SnmpRMIBAgent::findLexicoMinRowOid( imageInfop, _currresp->resOIDvec,
 	    				_currresp->numVarBindings, _currresp->resAttrIds, dummy ) ;

	    OID iimcAttrOid = imageInfop->getAttrOid( 1 ) ;
	    OID iimcAutoObjAndAttrOID = imageInfop->getIimcAutoObjAndAttrOID() ;

	    int rdnNumSubIds = internetOid->oid_nelem - (iimcAttrOid->oid_nelem - iimcAutoObjAndAttrOID->oid_nelem) ;

	    // Extract off the RDN component..
	    for( int i=0 ; i<rdnNumSubIds ; i++ ) {

	    	sprintf( (rdnBuffer + strlen(rdnBuffer)), "%s%d", i>0 ? "." : "",
				internetOid->oid_elements[iimcAttrOid->oid_nelem - 
						iimcAutoObjAndAttrOID->oid_nelem + i] ) ;
	    }

	    status = OK ;
	}

	return status ;
}
