//
// Copyright (c) 1994 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

////////////////////////////////////////////////////////////////
// SnmpRMIBAgent class specifications
////////////////////////////////////////////////////////////////
// Created              12-Jul-93
//
// Modifications        DD-MMM-YY - ""
//
////////////////////////////////////////////////////////////////

#ifndef _SNMP_RMIB_AGENT_H_
#define	_SNMP_RMIB_AGENT_H_

#ifndef _module_UNIV_defined_
#ifndef ICR1
#include <isode/pepy/UNIV-types.h>
#else
#include <isode/asn1/UNIV1-types.h>
#endif
#endif

#ifndef _LIST_ITERATOR_
#include "ListIterator.h"
#endif

#include "GenericKS.h"
#include "Boolean.h"
#include "SnmpRMIBerrors.h"
#include "SnmpRMIBif.h"
#include "SnmpImageMOClassInfo.h"

class SnmpImageAccess ;

class SnmpRMIBAgentResults ;


class SnmpRMIBAgent {

	static	int _snmpRMIBlastUsedAgentId ;
	int	_snmpRMIBAgentId ;
	
	SnmpVersionEnum	version ;
	OID	iimcAutoObjAndAttrOID ;

	Bool	snmpRMIBAgentInitialised ;
	
	struct  SnmpRMIBsnmpifData	snmpRMIBdata ;

	SnmpRMIBerrorEnum convertOIDsIIMCtoInternet( OID * /*iimcOIDs*/, int /*numOIDs*/, 
									OID ** /*internetOIDsp*/ ) ;
	SnmpRMIBerrorEnum convertOIDsInternetToIIMC( OID * /*internetOIDs*/, int /*numOIDs*/ ) ;

	SnmpRMIBerrorEnum createInternetOIDsandAppendRDNstr( SnmpImageMOClassInfo*, OID * /*internetOIDs*/, 
						int /*reqNAttrs*/, int* /*attrIds*/, char* /*rdnStringcp*/ ) ;
	
	SnmpRMIBerrorEnum snmpV1closeSession( void ) ;

	Bool isTableFullyTraversed( SnmpImageMOClassInfo*, int, OID*, int*, int, int* ) ;

	OID* resynchRowOIDs( SnmpImageMOClassInfo*, OID*, int , int*, Bool&) ;
	SnmpRMIBerrorEnum snmpRetrieveNewSubordinateMOattributes( SnmpImageMOClassInfo*,
							SnmpRMIBAgentResults&, OID*, int, int* ) ;
	SnmpRMIBerrorEnum snmpRetrieveRefreshSubordinateMOattributes( SnmpImageMOClassInfo*,
							SnmpRMIBAgentResults&,OID*, int, int*, SnmpImageAccess*) ;

public:
	Bool	remoteAgentIntermitObjSupport; // E.g. CISCO agent may support certain columnar objects
					       // only when the row is active.
					       // If True then getAttrSupportedByAgentFlags() should
					       // NOT be used.

	SnmpRMIBerrorEnum snmpV1openSession( struct qbuf * /*transportAddress*/, 
						struct qbuf * /*communityString*/,
						int /*snmpMsgRetryLimit*/ =3, int /*snmpMsgTimeoutS*/ =4 ) ; 

	OID     getIimcAutoObjAndAttrOID( void ) ;
	
	void	setMaxRetries( int ) ;
	void	setTimeoutS( int ) ;
	int	getMaxRetries( void ) ;
	int	getTimeoutS( void ) ;

	// Single instanced..
	SnmpRMIBerrorEnum snmpGetSingleInstancedMO( SnmpImageMOClassInfo*, SnmpRMIBAgentResults& ) ;

	// Multiply instanced..
	SnmpRMIBerrorEnum snmpGetOneMultiInstancedMO( SnmpImageMOClassInfo*, SnmpRMIBAgentResults&, RDN ) ;
	SnmpRMIBerrorEnum snmpGetMultiInstancedMOs(   SnmpImageMOClassInfo*, SnmpImageAccess*, SnmpRMIBAgentResults& ) ;
	SnmpRMIBerrorEnum snmpCreateEntry( RDN, int*, PE*, int, SnmpImageMOClassInfo*, SnmpRMIBAgentResults& ) ;

	// Either single or multi-instanced..
	SnmpRMIBerrorEnum snmpSetMOattributes( SnmpImageAccess*, int*, PE*, int, SnmpImageMOClassInfo*, SnmpRMIBAgentResults& ) ;

	
	int	getAgentId() ;
	static OID findLexicoMinRowOid( SnmpImageMOClassInfo*, OID*, int , int*, Bool&) ;

	SnmpRMIBAgent( int ) ;
	~SnmpRMIBAgent() ;
} ;

inline  OID     SnmpRMIBAgent::getIimcAutoObjAndAttrOID( void ){ return iimcAutoObjAndAttrOID ;}
inline	void 	SnmpRMIBAgent::setMaxRetries( int maxRetries ) { snmpRMIBdata.maxRetries = maxRetries ; }
inline	void	SnmpRMIBAgent::setTimeoutS( int timeoutS )     { snmpRMIBdata.timeoutS   = timeoutS   ; }
inline	int 	SnmpRMIBAgent::getMaxRetries( void )           { return snmpRMIBdata.maxRetries ; }
inline	int	SnmpRMIBAgent::getTimeoutS( void )	       { return snmpRMIBdata.timeoutS ; }
inline	int 	SnmpRMIBAgent::getAgentId() { return _snmpRMIBAgentId ; }

/////////////////////////////////////////////////////////////////////////////

class	SnmpRMIBAgentResults : public List {

	ListIterator _iterator ;

	struct 	SnmpRMIBrxedResults *	_currresp ;

protected:

	void deleteElem (void* resp ) ;
	
	int compare (void* res1p, void* res2p ) {
	    
	    return( (res1p == NULL || res2p == NULL || 
		    ((struct SnmpRMIBrxedResults*)res1p)->resOIDforCompare == NULL || 
		    ((struct SnmpRMIBrxedResults*)res2p)->resOIDforCompare == NULL   ) ?
	    	    NOTOK :
	    	    oid_cmp( ((struct SnmpRMIBrxedResults*)res1p)->resOIDforCompare, 
	    		     ((struct SnmpRMIBrxedResults*)res2p)->resOIDforCompare ) ) ;
	}


public:
	void replace( void* newresp ) {
	    clear() ;
	    List::append( newresp ) ;
	}

	void getFirst( void ) {
	    _currresp = (struct	SnmpRMIBrxedResults *)_iterator.getNext() ;
	}

	void nextMOCinstanceResults( void ) {
	    _currresp = (struct	SnmpRMIBrxedResults *)_iterator.getNext() ;
	}
	
	int getNresultAttrs( void ) {
	    return( (_currresp != NULL) ? _currresp->numVarBindings : 0 ) ; 
	}

	int appendRDNstrToBuffer( SnmpImageMOClassInfo*, char* ) ;
	
	PE	snmpExtractResult( int, SnmpImageMOClassInfo*, SnmpRMIBAgent *, CMISErrors* = ((CMISErrors*)0) ) ;

	int    append( struct  SnmpRMIBrxedResults* rxResultsp, OID compareOID, int attrSubIdIndex ) {
	    rxResultsp->resOIDforCompare = oid_cpy( compareOID ) ;
	    rxResultsp->resOIDforCompare->oid_elements[attrSubIdIndex-1] = 1 ; 
	    		// ...ensure any OID differences are in the RDN

	    List::append((void*) rxResultsp );
	    return OK;
	}

	int    insert( struct  SnmpRMIBrxedResults* rxResultsp, OID compareOID, int attrSubIdIndex ) {
	    rxResultsp->resOIDforCompare = oid_cpy( compareOID ) ;
	    rxResultsp->resOIDforCompare->oid_elements[attrSubIdIndex-1] = 1 ; 
	    		// ...ensure any OID differences are in the RDN

	    return List::insert((void*) rxResultsp );
	}

	SnmpRMIBAgentResults() : List() {
	    clear() ;
	    _iterator.setList( this ) ;
	    _currresp = NULLSNMPRMIBRXEDRESULTS ;
	}
} ;


inline void SnmpRMIBAgentResults::deleteElem( void* resp ) {
	struct SnmpRMIBrxedResults *resultp = (struct SnmpRMIBrxedResults *)resp ;
	if( resp ) SnmpRMIBremoveReceivedResultsList( &resultp ) ;
}


#define BULKARRAYINCSIZE 	50
#define BULKRETRIEVENTHENTRIES	3


#endif

