//
// Copyright (c) 1994 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

#include <stdlib.h>
#include "msap.h"
#include "Error.h"
#include "SimpleNameType.h"
#include "DName.h"
#include "OperationalState.h"
#include "String.h"
#include "UsageState.h"
#include "remoteSystem.h"
#include "SnmpImageMO.h"
#include "IQASntx.h"

MOClassInfo remoteSystem::_remoteSystem;
const int
remoteSystem::REMOTESYSTEMID	  = 0,
remoteSystem::REMOTESYSTEMTITLE	  = 1,
remoteSystem::REMOTESYSOPERSTATE  = 2,
remoteSystem::REMOTESYSUSAGESTATE = 3 ;

//
// May only be instantiated from the cmipsnmpProxyAgent class
//
remoteSystem::remoteSystem( AVA*& errorInfo, RDN rdn, MO *superior, cmipsnmpProxyAgent * cmipsnmpProxyAgentp,
							Bool&	remoteAgentIsReachible )
: SnmpImageAccess(rdn, superior)
{
int	status = OK ;

	remoteAgentIsReachible = False ;

	// initialise class information
	if (!_remoteSystem.initialised() && initialiseClass() == NOTOK)
	error(NULLCP, "remoteSystem class initialisation failed!");

	register
        char *name = rdn2str(rdn);
	fprintf(stderr, "remoteSystem::remoteSystem() for %s\n", name);
	(void)free( name ) ;

	// initialise attributes
        _attrs[REMOTESYSTEMID]		= new SimpleNameType ;
	_attrs[REMOTESYSTEMTITLE]	= new DName ;
        _attrs[REMOTESYSOPERSTATE]	= new OperationalState;
	_attrs[REMOTESYSUSAGESTATE]	= new UsageState;

fprintf(stderr, "attributes initialised ...\n");

	// register class
        registerClass(REMOTESYSTEMNATTRS, _attrs, &_remoteSystem, True);

	// check name binding
        if (checkNameBinding() == NOTOK) {
#if defined(DEBUG)
                advise(NULLCP, "remoteSystem::remoteSystem() checkNameBinding() failed\n");
#endif
//		if (err != NULL) *err = m_invalidObjectInstance;
//		return;
	}

	(DName *) _attrs[REMOTESYSTEMTITLE]->set((void*)NULL ) ;

	static OperStateVal operState  = os_disabled;
	static int  	    usageState = int_SMI_UsageState_idle ;

	// Buffer the proxy state..
	setCmipsnmpProxyAgent( cmipsnmpProxyAgentp ) ;
	setSnmpRMIBAgent( cmipsnmpProxyAgentp->getSnmpRMIBAgent() ) ;

	static 	OID internetSystemClassOid ;

	// Attempt to carry out an initial poll on the remote agent..
	char internetSystemRdnBuff[ 256 ] ;
	RDN  internetSystemRdn ;

	if( MIBInfo::getUseGenericRDNs() == False ) {
	
	    sprintf( internetSystemRdnBuff, "%s=NULL", MIBInfo::getInternetSystemIdName() ) ;
	    internetSystemRdn = str2rdn( internetSystemRdnBuff ) ;

	} else {

	    internetSystemRdn = makeNullGenericRDN( MIBInfo::getInternetSystemIdName() ) ;
	}

	if( internetSystemRdn != NULLRDN ) {

	    CMISErrors refreshMSAPerror = m_noError ;
		    
	    if( (status = refreshSubordinate( internetSystemRdn, errorInfo )) == OK ) {

	    	operState  		= os_enabled ;
		usageState 		= int_SMI_UsageState_active ;
		remoteAgentIsReachible 	= True ;
	    }

	    (void)rdn_free( internetSystemRdn ) ;
	}

	// Set up the state..
	(OperationalState *) _attrs[REMOTESYSOPERSTATE] ->set((void *) &operState  );
	(UsageState *) 	     _attrs[REMOTESYSUSAGESTATE]->set((void *) &usageState );

	return ;
}

int remoteSystem::initialiseClass ()
{	
	if (_remoteSystem.initialised())
		return OK;

	_remoteSystem.setParent(Top::getClassInfo()) ;
	
	if (_remoteSystem.setClass( "remoteSystem", 1, REMOTESYSTEMNATTRS, 0, 0, 0) == NOTOK ||

		_remoteSystem.setNameBinding("remoteSystem-systemNB", "System", "systemId") == NOTOK ||

		_remoteSystem.setAttr("systemId",   	  REMOTESYSTEMID      ) == NOTOK ||
		_remoteSystem.setAttr("systemTitle",      REMOTESYSTEMTITLE   ) == NOTOK ||
		_remoteSystem.setAttr("operationalState", REMOTESYSOPERSTATE  ) == NOTOK ||
		_remoteSystem.setAttr("usageState",       REMOTESYSUSAGESTATE ) == NOTOK   )
		return NOTOK;

	return OK;
}


int
remoteSystem::refreshSubordinate( RDN subordinateRDN, AVA*& errorInfo, int )
{
int		status = OK ;
CMISErrors	refreshMSAPerror = m_noError ;
static	OID 	systemOID = NULLOID ;

	if( systemOID == NULLOID )
	    systemOID =  name2oid( "remoteSystem" ) ;	// [10165-2] system object

	if( isRemoteAgentLocked( status, errorInfo ) == True )
	    return status ;

	char * subordinateRdnAtcp = AttrT2str( subordinateRDN->rdn_at ) ;

	if( subordinateRdnAtcp != NULLCP ) {

	    /* Extract off the trailing "Id"... */
	    *(subordinateRdnAtcp + strlen( subordinateRdnAtcp ) - 2) = '\0' ;

	    OID subordinateClassOID = name2oid( subordinateRdnAtcp ) ;
		
	    cmipsnmpProxyAgent   * cmipsnmpProxyAgentp = getCmipsnmpProxyAgent() ;
	    struct  ObjIdListVal * firstSupportedMIBlistp ;

	    if( subordinateClassOID != NULLOID && cmipsnmpProxyAgentp &&
		(firstSupportedMIBlistp = cmipsnmpProxyAgentp->getMIBOIDs()) ) {
	
		// run through the MIBInfo structures for the supportedMIBs that have
		// "system" as the superior class..
		int	nthSupportedMIBchecked ;
		SnmpImageMOClassInfo* subordinateClassInfop = NULLIMAGEINFO ;

		for( MIBInfo * currMIBinfop = MIBInfo::getFirstMIBInfo( firstSupportedMIBlistp, nthSupportedMIBchecked, systemOID ) ;
		     currMIBinfop != NULLMIBINFO && subordinateClassInfop == NULLIMAGEINFO ;
		     currMIBinfop = currMIBinfop->getNextPeerMIBInfo( firstSupportedMIBlistp, nthSupportedMIBchecked, systemOID ) ) {

		    // Now run through the immediate subordinate SnmpImageMOClassInfo objects
		    // until one is found that has the same class as the required RDN..
		    for( subordinateClassInfop = currMIBinfop->getSubordinateImageClass() ;
			 subordinateClassInfop != NULLIMAGEINFO ;
			 subordinateClassInfop = subordinateClassInfop->getNextPeerImageInfo() ) {

			if( oid_equal( subordinateClassInfop->getClass(), subordinateClassOID ) )
			    break ; // image Info node found !
		    }
		}

		if( subordinateClassInfop != NULLIMAGEINFO ) {
			     
		    refreshMSAPerror = refreshImageSubordinate( subordinateClassInfop, subordinateRDN, errorInfo ) ;
		}
	    }

	    if( subordinateClassOID != NULLOID )
		oid_free( subordinateClassOID ) ;

	    (void)free( subordinateRdnAtcp ) ;
	}

	if( refreshMSAPerror != m_noError ) {
	    status    = NOTOK ;
	}

	return status ;
}



int
remoteSystem::refreshSubordinates( AVA*&errorInfo, int )
{
int		status = OK ;
CMISErrors	refreshMSAPerror = m_noError ;
static	OID	systemOID = NULLOID ;

	if( systemOID == NULLOID )
	    systemOID =  name2oid( "remoteSystem" ) ;	// [10165-2] system object

	if( isRemoteAgentLocked( status, errorInfo ) == True )
	    return status ;

	cmipsnmpProxyAgent   * cmipsnmpProxyAgentp = getCmipsnmpProxyAgent() ;
	SnmpRMIBAgent	     * snmpRMIBagentp = getSnmpRMIBAgent() ;
	struct  ObjIdListVal * firstSupportedMIBlistp ;

	if( cmipsnmpProxyAgentp && snmpRMIBagentp &&
		(firstSupportedMIBlistp = cmipsnmpProxyAgentp->getMIBOIDs()) ) {

	    // run through the MIBInfo structures for the supportedMIBs that have
	    // "system" as the superior class..
	    int	nthSupportedMIBchecked ;
	    for( MIBInfo * currMIBinfop = MIBInfo::getFirstMIBInfo( firstSupportedMIBlistp, nthSupportedMIBchecked, systemOID ) ;
	    	 currMIBinfop != NULLMIBINFO ;
		 currMIBinfop = currMIBinfop->getNextPeerMIBInfo( firstSupportedMIBlistp, nthSupportedMIBchecked, systemOID ) ) {

		// Now run through the immediate subordinate SnmpImageMOClassInfo objects..
		for( SnmpImageMOClassInfo* subordinateClassInfop = currMIBinfop->getSubordinateImageClass() ;
		     subordinateClassInfop != NULLIMAGEINFO ;
		     subordinateClassInfop = subordinateClassInfop->getNextPeerImageInfo() ) {

		    if( (refreshMSAPerror = refreshImageSubordinates( subordinateClassInfop, errorInfo )) != m_noError )
		    	break ;
		}

		if( refreshMSAPerror != m_noError )
		    break ;
	    }
	}
	
	if( refreshMSAPerror != m_noError ) {
	    status    = NOTOK ;
	}

	return status ;
}
