//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// SimOS.h

// By George Pavlou, February 1994


#include "Create.h"
#include "SmiSntx.h"

#include "System.h"
#include "eventForwardingDiscriminator.h"
#include "eventLogRecord.h"
#ifdef	LOGGING
#include "Log.h"
#endif

#include "systemSim.h"



char*	service = "SMA";			// application name

#define	MIB_INIT_FILE	"mib.init.sim2"		/* MIB initialisation info */



// a table of all the MO classes this application supports -
// used to create the meta-class tree

MOClass moclasses[] =
{
//  SMF-related classes first and the MIT-top system

{ "top",
   Top::getClassInfo,
   Top::initialiseClass },

{ "system",
   System::getClassInfo,
   System::initialiseClass },

{ "discriminator",
   discriminator::getClassInfo,
   discriminator::initialiseClass },

{ "eventForwardingDiscriminator",
   eventForwardingDiscriminator::getClassInfo,
   eventForwardingDiscriminator::initialiseClass },

#ifdef	LOGGING
{ "log",
   Log::getClassInfo,
   Log::initialiseClass },
#endif

{ "logRecord",
   logRecord::getClassInfo,
   logRecord::initialiseClass },

{ "eventLogRecord",
   eventLogRecord::getClassInfo,
   eventLogRecord::initialiseClass },

};

int NCLASSES = sizeof moclasses / sizeof moclasses[0];



// a table mapping notifications to specific log records -
// used to process notifications 

NotificationMapping notificationMappingTable[] =
{
{"objectCreation", 		"eventLogRecord"},
{"objectDeletion",	 	"eventLogRecord"},
{"attributeValueChange",	"eventLogRecord"},
{"stateChange",			"eventLogRecord"},
};

int NNOTIFICATIONS = sizeof notificationMappingTable /
			    sizeof notificationMappingTable[0];



// a table of syntaxes this application uses

VoidFunc	syntaxes[] = { smi_syntaxes, NULL };

