/*
 * Copyright (c) 1994 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
** Autnetication dialogue exchange mechanism
*/

/*
** Saleem N. Bhatti
** March 1994
*/

#if !defined(_mauth_h_)
#define _mauth_h_

#include "msap.h"

#if !defined(_a_)
#if defined(__STDC__) || defined(__cplusplus)
#define _a_(_args_) _args_
#else
#define _a_(_args_) ()
#endif
#endif

#if defined(__cplusplus)
extern "C" {
#endif

/* This is to be used in the MInitOption to minitialise() */
#define MIO_AuthAssocIntegrityInfo 1

typedef struct AuthAssocIntegrityInfo_s {

    char *name;          /* String DN of user/application                */
                         /* If NULL then read "username" from ~/.authrc  */

    char *peer;          /* String DN of peer user/application           */
                         /* If NULL then read "username" from ~/.authrc  */

    char *ca;            /* String DN of Certification Authority         */
                         /* If NULL then read "authority" from ~/.authrc */

    char *dsa;           /* The name or ISODE string format presentation */
                         /* address of the DSA on which the CA node      */
                         /* resides                                      */
                         /* If NULL then read "dsa" from ~/.authrc       */

    char sessionKey[8];  /* MD5 Key - zeros if no key for this assoc     */

} AuthAssocIntegrityInfo;

AuthAssocIntegrityInfo *AuthAssocIntegrityInfo_make _a_(());
void AuthAssocIntegrityInfo_free _a_((AuthAssocIntegrityInfo *v));

/* These OIDs are in either ETC/oidtable.gen or ETC/isobjects */
/* joint-iso-ccitt(2) mgmt(37) ucl-mgmt(1) uclMiscellaneous(99) midas(1) */
#define CREDENTIAL_OID "2.37.1.99.1.1"
#define SESSIONKEY_OID "2.37.1.99.1.2"
/* iso(1) member-body(2) US(840) rsadsi(113549) ? ? ? */
#define MD2WITHRSA_OID "1.2.840.113549.1.1.2"

/* Return values from the following four functions are 0 or ... */
#define AUTH_error       (-1)  /* General error */
#define AUTH_privateInfo (-2)  /* Could not read user's private info */
#define AUTH_x500Error   (-3)  /* Trouble accessing directory */
#define AUTH_cryptError  (-4)  /* {en,de}cryption error */
/* getCredentials() and getAcseUserInfo() only */
#define AUTH_authFailed  (-5)  /* Authentication failed */
/* getAcseUserInfo() only */
#define AUTH_noUserInfo  (-6)  /* No ACSE user info */

/* Intermediate level API */
int makeCredentials _a_((AuthAssocIntegrityInfo *info, PE *pe));
int getCredentials _a_((PE pe, AuthAssocIntegrityInfo **info));

/* Higher level API */
int makeAcseUserInfo _a_((AuthAssocIntegrityInfo *info, External **external));
int getAcseUserInfo _a_((External *external, AuthAssocIntegrityInfo **info));

/*
** Above #define error values as arg to this function for error message.
** Do not free() return value
*/
char *authErrorMessage _a_((int e));

/* Initialisation */
void initialiseOsisec _a_((int argc, char **argv));

/* Reading ~/.authrc */
FILE *openAuthrc _a_(());
int rewindFile _a_((FILE *fp));
int readAuthrc _a_((FILE *fp, char *name, char *value));
void closeAuthrc _a_((FILE *fp));

#if defined(__cplusplus)
}
#endif

#endif /* !_mauth_h_ */
