/*
 * Copyright (c) 1994 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
** Managing invokeId values
*/

/*
** Saleem N. Bhatti
** April 1994
*/

#if defined(__GNUG__)
#include <std.h>
#endif
#include <stdlib.h>
#include <sys/types.h>
#include <stdio.h>
#include <malloc.h>
#include <memory.h>

#include "minvoke.h"

Invoke *
#if defined(__STDC__) || defined(__cplusplus)
Invoke_make(const int id, const int invokeId, Invoke *iList)
#else
Invoke_make(id, invokeId, iList)
int id;
int invokeId;
Invoke *iList;
#endif
{
    register
    Invoke *p, *i = (Invoke *) malloc(sizeof(Invoke));

    if (i == (Invoke *) 0) {
        fprintf(stderr, "Invoke_make(): malloc() failed\n");
        return (Invoke *) 0;
    }

    /* Fill in the new element */
    i->id = id;
    i->invokeId = invokeId;
    i->next = (Invoke *) 0;

    /* Get to end of list */
    for(p = iList; p != (Invoke *) 0 && p->next != (Invoke *) 0; p = p->next);

    /* Update pointers */
    if (p != (Invoke *) 0) {
        p->next = i;
        i->prev = p;
    }
    else i->prev = (Invoke *) 0;

    return (iList == (Invoke *) 0 ? i : iList);
}

void
#if defined(__STDC__) || defined(__cplusplus)
Invoke_free(Invoke *i)
#else
Invoke_free(i)
Invoke *i;
#endif
{
    if (i == (Invoke *) 0)
        return;

    /* Patch list ... */
    if (i->prev != (Invoke *) 0)
        i->prev->next = i->next;

    if (i->next != (Invoke *) 0)
        i->next->prev = i->prev;

    /* ... and delete */
    free(i);
}

Invoke *
#if defined(__STDC__) || defined(__cplusplus)
Invoke_search(Invoke *iList, const int v,
              const InvokeSearch s, const InvokeSearchType t)
#else
Invoke_search(iList, v, s, t)
Invoke *iList;
int v;
InvokeSearch s;
InvokeSearchType t;
#endif
{
    register
    Invoke *p;

    for(p = iList; p != (Invoke *) 0; p = p->next) {
        register
        int i;

        switch ((int) t) {

        case (int) InvokeSearchType_first: return iList;

        case (int) InvokeSearchType_search:
        case (int) InvokeSearchType_next:

            switch ((int) s) {
            case (int) InvokeSearch_id:       i = p->id;       break;
            case (int) InvokeSearch_invokeId: i = p->invokeId; break;
            }

            if (i == v) { /* found a match */
                if (t == InvokeSearchType_next)
                    p = p->next; /* walking the list */

                return p;
            }
            break;
        }
    }

    return (Invoke *) 0;
}

static
Invoke *S_lists[FD_SETSIZE];

#define GET_INVOKE(_fd) \
    ((_fd < 0 || _fd > FD_SETSIZE) ? (Invoke *) 0 : S_lists[_fd])

#define SET_INVOKE(_fd, _i) \
    if (_fd > 0 && _fd < FD_SETSIZE) S_lists[_fd] = _i

void
initialiseInvokeListSpace()
{
    static
    int done = 0;

    if (done)
        return;

    memset((void *) S_lists, 0, sizeof(Invoke *) * FD_SETSIZE);

    done = 1;
}

Invoke *
#if defined(__STDC__) || defined(__cplusplus)
getInvokeList(const int fd)
#else
getInvokeList(fd)
int fd;
#endif
{ return GET_INVOKE(fd); }

void
#if defined(__STDC__) || defined(__cplusplus)
setInvokeList(const int fd, Invoke *iList)
#else
setInvokeList(fd, iList)
int fd;
Invoke *iList;
#endif
{
    clearInvokeList(fd);
    SET_INVOKE(fd, iList);
}

void
#if defined(__STDC__) || defined(__cplusplus)
clearInvokeList(const int fd)
#else
clearInvokeList(fd)
int fd;
#endif
{
    register
    Invoke *n, *p;

    /* Delete existing list ... Invoke_free() not required */
    for(p = GET_INVOKE(fd); p != (Invoke *) 0; p = n) {
        n = p->next;
        free(p);
    }

    SET_INVOKE(fd, (Invoke *) 0);
}


void
#if defined(__STDC__) || defined(__cplusplus)
addInvoke(const int fd, const int id, const int invokeId)
#else
addInvoke(fd, id, invokeId)
int fd;
int id;
int invokeId;
#endif
{
    register
    Invoke *iList = GET_INVOKE(fd), *i;

    i = Invoke_make(id, invokeId, iList);

    if (iList == (Invoke *) 0)
        SET_INVOKE(fd, i);
}

void
#if defined(__STDC__) || defined(__cplusplus)
removeInvoke(const int fd, const int v, const InvokeSearch s)
#else
removeInvoke(fd, v, s)
int fd;
int v;
InvokeSearch s;
#endif
{
    register
    Invoke *iList = GET_INVOKE(fd);

    if (iList != (Invoke *) 0)
        Invoke_free(Invoke_search(iList, v, s));
}

int
#if defined(__STDC__) || defined(__cplusplus)
checkInvoke(const int fd, const int v,
            const InvokeSearch s, InvokeSearchType t)
#else
checkInvoke(fd, v, s, t)
int fd;
int v;
InvokeSearch s;
InvokeSearchType t;
#endif
{
    register
    Invoke *i, *iList = GET_INVOKE(fd);
    register
    int r = 0;

    if ((i = Invoke_search(iList, v, s, t)) != (Invoke *) 0)
        r = s == InvokeSearch_id ? i->invokeId : i->id;

    return r;
}

void
#if defined(__STDC__) || defined(__cplusplus)
clearInvoke(const int fd)
#else
clearInvoke(fd)
int fd;
#endif
{ clearInvokeList(fd); }

void
initialiseInvoke()
{ initialiseInvokeListSpace(); }

