/* attrv.c - Attribute Value routines */

#ifndef lint
static char *rcsid = "$Header: /xtel/isode/isode/dsap/common/RCS/attrv.c,v 9.0 1992/06/16 12:12:39 isode Rel $";
#endif

/*
 * $Header: /xtel/isode/isode/dsap/common/RCS/attrv.c,v 9.0 1992/06/16 12:12:39 isode Rel $
 *
 *
 * $Log: attrv.c,v $
 * Revision 9.0  1992/06/16  12:12:39  isode
 * Release 8.0
 *
 */

/*
 *                                NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */


/* LINTLIBRARY */

#include "util.h"
#include "entry.h"
#include "ds_error.h"
#include "malloc.h"
#ifndef	ICR1
#include "isode/cmd_srch.h"
#else
#include "isode/compat/cmd_srch.h"
#endif
#include <sys/stat.h>

extern int oidformat;
extern struct PSAPaddr * psap_cpy ();
extern str2asn (), pe_print() ;
extern LLog * log_dsap;
PE asn2pe();
int quipu_pe_cmp();
char t61_flag;
char crypt_flag;
extern char dsa_mode;
extern char chrcnv [];

/*
 * String syntax hash table...
 */

static	int hashlen ;

#define SBUCKETS	256

struct 	syntaxpair {

	int	s_entryNum ;
	struct	syntaxpair * s_chain ;
} ;

static struct syntaxpair *Sbuckets[SBUCKETS];

/*
 * Hashing : 0-255 key produced by taking account of the upper case first and
 * 	     second to last characters, plus the string length.
 */
#define		PBUCKETS	256
#define SHASH(nm) \
	( (hashlen = strlen(nm)) < 2  ? \
		((unsigned char)((chrcnv[((nm)[0])]))) : \
		  (hashlen & 0x01) == 1 ? \
		    ((unsigned char)((chrcnv[((nm)[0])] & 0xf) + \
    			((hashlen & 0xf) << 4) - (chrcnv[((nm)[hashlen-2])]))) : \
		    ((unsigned char)((chrcnv[((nm)[0])] & 0xf) + \
    			((hashlen & 0xf) << 4) + (chrcnv[((nm)[hashlen-2])]))))



#define SNTXTABLEBASESIZE	50
#define SNTXTABLEINCSIZE	25
static  sntx_table * syntax_table_basep = NULL ;
static 	short curr_syntax_table_size = 0 ;    /* Number of entries possible */
static 	short num_syntax = 0 ;		      /* Number of entries used */

static sntx_table syntax_table_asn1_template = {
	"ASN", 		/* ASN - default type */
	(IFP)pe_cpy,	/* default encode */ 
	NULLIFP,	/* no decoding needed */
	NULLIFP,	/* default parse */
	NULLIFP, 	/* default print */
	(IFP)pe_cpy,	/* default copy */
#ifdef STRICT_X500
	NULLIFP,	/* Not allowed if syntax unknown (X.501 9.6.2) */
#else
	quipu_pe_cmp,	/* default compare */
#endif
	pe_free,	/* default free */
	NULLCP,		/* no pe_printer */
	NULLIFP,	/* NO approx matching */
	TRUE,		/* one per line */
} ;
	

short	add_attribute_syntax (sntx,enc,dec,parse,print,cpy,cmp,sfree,print_pe,approx,multiline)
char *	sntx;
IFP	enc,dec,parse,print,cpy,cmp,sfree,approx;
char *  print_pe;
char	multiline;
{
int	   	hindex ;
register 	struct syntaxpair    *p;

	if( syntax_table_basep == NULL ) {

		if( (syntax_table_basep = (sntx_table*)malloc( sizeof(sntx_table) * SNTXTABLEBASESIZE)) == NULL ) {
			
			LLOG (log_dsap,LLOG_FATAL, ("attrv.c:add_attribute_syntax() syntax table malloc failure" ));
			advise( NULLCP, "attrv.c:add_attribute_syntax() syntax table malloc failure" ) ;
			return (-1); 
		} else {
			curr_syntax_table_size = SNTXTABLEBASESIZE ;
			num_syntax = 1 ;
			syntax_table_basep[0] = syntax_table_asn1_template ; /* struct copy */

			if( add_Sbucket( syntax_table_asn1_template.s_sntx, 0, 
						SHASH(syntax_table_asn1_template.s_sntx) ) == NOTOK )
			    return -1 ;
		}
	}

    	for (p = Sbuckets[ hindex = SHASH (sntx)];
		p && lexequ (((sntx_table*)&syntax_table_basep[p->s_entryNum])->s_sntx, sntx);
		p = p -> s_chain) 
	    ; /* NO-OP */

	if( p ) {
	    LLOG (log_dsap,LLOG_FATAL,("attrv.c:add_attribute_syntax() - duplicate syntax '%s'", sntx ));
#ifdef DEBUG
	    advise( NULLCP, "attrv.c:add_attribute_syntax() - duplicate syntax '%s'", sntx );
#endif
	    return p->s_entryNum ;
	}

	/* Insert the hash node... */
	if( add_Sbucket( sntx, num_syntax, hindex ) == NOTOK )
	    return -1 ;

	syntax_table_basep[num_syntax].s_sntx    = sntx;
	set_attribute_syntax( num_syntax,enc,dec,parse,print,cpy,cmp,sfree,print_pe,approx,multiline);

	/* Increment the syntax number (adjusting array sizes as required... */
	if( (num_syntax+1) >= curr_syntax_table_size ) {

		if( (syntax_table_basep = (sntx_table*)realloc( syntax_table_basep,
				sizeof(sntx_table) * (curr_syntax_table_size + SNTXTABLEINCSIZE))) == NULL ) {
			
			LLOG (log_dsap,LLOG_FATAL, ("attrv.c:add_attribute_syntax() syntax table realloc failure" ));
			advise( NULLCP, "attrv.c:add_attribute_syntax() syntax table realloc failure" ) ;
			return (-1); 
		} else {
			curr_syntax_table_size += SNTXTABLEINCSIZE ;
		}
	}

	return( num_syntax++ );
}

int
add_Sbucket( sntx, num_syntax, hindex )
char *	sntx ;
int	num_syntax, hindex ;
{
register 	struct syntaxpair    *p;

	if( (p = (struct syntaxpair*)calloc(1, sizeof(*p))) == NULL ) {
	    SLOG (log_dsap, LLOG_EXCEPTIONS, NULLCP,
	      		("attrv.c:add_attribute_syntax() - calloc of syntax hash structure failed"));
#ifdef DEBUG
	    advise( NULLCP, "attrv.c:add_attribute_syntax() - calloc of syntax hash structure failed" ) ;
#endif
	    return NOTOK ;
	}
	
	p -> s_entryNum = num_syntax ;
	p -> s_chain	= Sbuckets[ hindex ] ;
	Sbuckets[ hindex ] = p ;

	return OK ;
}

set_attribute_syntax (sntx,enc,dec,parse,print,cpy,cmp,sfree,print_pe,approx,multiline)
short sntx;
IFP	enc,dec,parse,print,cpy,cmp,sfree,approx;
char *  print_pe;
char	multiline;
{
        if (sntx > num_syntax)
	    return;

	syntax_table_basep[sntx].s_encode  = enc;
	syntax_table_basep[sntx].s_decode  = dec;
	syntax_table_basep[sntx].s_parse   = parse;
	syntax_table_basep[sntx].s_print   = print;
	syntax_table_basep[sntx].s_copy    = cpy;
	syntax_table_basep[sntx].s_compare = cmp;
	syntax_table_basep[sntx].s_free    = sfree;
	syntax_table_basep[sntx].s_pe_print= print_pe;
	syntax_table_basep[sntx].s_approx  = approx;
	syntax_table_basep[sntx].s_multiline  = multiline;
}

set_av_pe_print (sntx,print_pe)
short sntx;
char *  print_pe;
{
        if (sntx >= num_syntax)
	    return;

	syntax_table_basep[sntx].s_pe_print= print_pe;
}

set_av_printer (sntx,print)
short sntx;
IFP print;
{
        if (sntx >= num_syntax)
	    return;

	syntax_table_basep[sntx].s_print = print;
}

short modify_av_printer (at,sntx,print)
AttributeType at;
char * sntx;
IFP print;
{
short nstx;

	nstx  = add_attribute_syntax(sntx,
	    syntax_table_basep[at->oa_syntax].s_encode,
	    syntax_table_basep[at->oa_syntax].s_decode,
	    syntax_table_basep[at->oa_syntax].s_parse,
	    print,
	    syntax_table_basep[at->oa_syntax].s_copy,
	    syntax_table_basep[at->oa_syntax].s_compare,
	    syntax_table_basep[at->oa_syntax].s_free,
	    syntax_table_basep[at->oa_syntax].s_pe_print,
	    syntax_table_basep[at->oa_syntax].s_approx,
	    syntax_table_basep[at->oa_syntax].s_multiline);

	at->oa_syntax = nstx;
	return nstx;
}

split_attr (as)
Attr_Sequence as;
{
	if ((as->attr_type == NULLTABLE_ATTR)
	  ||(as->attr_type->oa_syntax >= AV_WRITE_FILE))
		return (TRUE);
	else 
		return (syntax_table_basep[as->attr_type->oa_syntax].s_multiline);
}

IFP approxfn (x)
short x;
{
        if (x >= num_syntax)
	    return NULLIFP;

	return (syntax_table_basep[x].s_approx);
}

char * syntax2str (sntx)
short sntx;
{
        if (sntx >= num_syntax)
	    return NULLCP;

	return (syntax_table_basep[sntx].s_sntx);
}

short str2syntax (str)
char * str;
{
register int    hindex ;
register struct syntaxpair    *p;

    	for (p = Sbuckets[ hindex = SHASH(str)];
		p && lexequ (((sntx_table*)&syntax_table_basep[p->s_entryNum])->s_sntx, str);
		p = p -> s_chain) 
	    ; /* NO-OP */

	if( p )
	    return( p->s_entryNum ) ;

	return (0);
}

sntx_table * get_syntax_table (x)
short x;
{
        if (x >= num_syntax)
	    return (sntx_table * )NULL;

	return ( &syntax_table_basep[x]);
}

AttrV_free (x)
AttributeValue x;
{
	AttrV_free_aux (x);
	free ((char *)  x);
}

AttrV_free_aux (x)
register AttributeValue x;
{
	if (x == NULLAttrV)
		return;

	if (( x->av_syntax < AV_WRITE_FILE ) 
		&& (syntax_table_basep[x->av_syntax].s_free != NULLIFP)
		&& (x->av_struct != NULL))
			(*syntax_table_basep[x->av_syntax].s_free) (x->av_struct);
}

PE      grab_pe(av)
AttributeValue  av;
{
PE ret_pe = NULLPE;

	if (syntax_table_basep[av->av_syntax].s_encode != NULLIFP)
		ret_pe = (PE)((*syntax_table_basep[av->av_syntax].s_encode) (av->av_struct));

	return (ret_pe);
}

AttrV_decode(x,y)
register AttributeType x;
register AttributeValue y;
{
int y_syntax;

	if (x == NULLAttrT)
		return (NOTOK);

	if ((y == NULLAttrV) || (y->av_syntax != 0))
		return (OK);

	if (y->av_struct == NULL)
		return (OK);

	y->av_syntax = x->oa_syntax;
	if ( (y_syntax = y->av_syntax) >= AV_WRITE_FILE ) 
		y_syntax = y->av_syntax - AV_WRITE_FILE;

	ATTRIBUTE_HEAP;

	if (syntax_table_basep[y_syntax].s_decode != NULLIFP) {
		PE oldpe = (PE)y->av_struct;
		if ((y->av_struct = (caddr_t)((*syntax_table_basep[y_syntax].s_decode) (oldpe))) == NULL) {
			RESTORE_HEAP;
			y->av_struct = (caddr_t)oldpe;
			y->av_syntax = 0;
			return (NOTOK);
		}
		pe_free (oldpe);
	} 

	RESTORE_HEAP;

	return (OK);
}

static strip_header(str)
char ** str;
{
register char * ptr, *save, val;
int syntax;
static CMD_TABLE cmd_syntax [] = {
        "FILE",		1,
	"ASN",		2,
	"T.61",		3,
	"CRYPT",	4,
        0,              0,
        } ;


	if (*str == NULLCP)
		return (0);

	ptr = SkipSpace (*str);
	if (*ptr == 0) {
		if ( ptr == (*str)+1) 	/* single space */
			return 5;
		return (0);
	}
	*str = ptr;

	t61_flag = FALSE;
	crypt_flag = FALSE;

	if (*ptr++ == '{') {
		/* look for syntax */	
		if (( *str = index (ptr,'}')) == 0) {
			parse_error ("syntax close bracket missing '%s'",--ptr);
			return (0);
		}
                save = *str;
                val = **str;

		*(*str)++ = 0;
		*str = SkipSpace (*str);

		if ((syntax = cmd_srch (ptr,cmd_syntax)) == 0) {
			parse_error ("unknown syntax '%s'",ptr);
			return (0);
		}
                *save = val;

		if ((syntax == 4) && ! dsa_mode) {
			parse_error ("{CRYPT} not allowed",ptr);
			return (0);
		}
		return (syntax);

	}
	return (5);
}

str_at2AttrV_aux (str,at,rav)
char * str;
AttributeType at;
AttributeValue rav;
{
	if (at == NULLAttrT)
		return (NOTOK);

	switch (strip_header (&str)) {
		case 0:	/* error */
			return (NOTOK);
		case 2: /* ASN */
			(void) str2AttrV_aux (str,0,rav);
			if (AttrV_decode (at,rav) == NOTOK) {
				parse_error ("ASN attribute decode failed",NULLCP);
				return (NOTOK);
			}
			return (OK);
		case 3: /* T.61 */
			t61_flag = TRUE;
			return (str2AttrV_aux (SkipSpace(str),at->oa_syntax,rav));
		case 4:
			crypt_flag = TRUE;
			/* fall */
		case 5:
			return (str2AttrV_aux (str,at->oa_syntax,rav));
	}
	return (NOTOK);
}


AttributeValue str_at2AttrV (str,at)
char * str;
AttributeType at;
{
AttributeValue av;

	if (at == NULLAttrT)
		return (NULLAttrV);

	switch (strip_header(&str)) {
		case 0:	/* error */
			return (NULLAttrV);
		case 2: /* ASN */
			av = str2AttrV (str,0);
			if (AttrV_decode (at,av) == NOTOK) {
				parse_error ("ASN attribute decode failed (2)",NULLCP);
				return (NULLAttrV);
			}
			return (av);
		case 3: /* T.61 */
			t61_flag = TRUE;
			return (str2AttrV (str,at->oa_syntax));
		case 4:
			crypt_flag = TRUE;
			/* fall */
		case 5:
			return (str2AttrV (str,at->oa_syntax));
	}
	return (NULLAttrV);
}

AttributeValue str2AttrV (str,syntax)
char * str;
short syntax;
{
AttributeValue x;
short ns;

	x = AttrV_alloc();

	if ( (ns = syntax) >= AV_WRITE_FILE ) 
		ns = syntax - AV_WRITE_FILE;
	
	if (str2AttrV_aux (str,ns,x) == OK)
		return (x);
	free ((char *)x);
	return NULLAttrV;
}

str2AttrV_aux (str,syntax,x)
char * str;
short syntax;
AttributeValue x;
{

	if (str == NULLCP)
		return (NOTOK);

	if ((x->av_syntax = syntax) > AV_WRITE_FILE)
		x->av_syntax = syntax - AV_WRITE_FILE;
	x->av_struct = NULL;

	if (syntax_table_basep[syntax].s_parse != NULLIFP) {
		if ((x->av_struct = (caddr_t)(*syntax_table_basep[syntax].s_parse) (str)) == NULL) 
			return (NOTOK);
		if (t61_flag) {
			parse_error ("invalid use of {T.61}",NULLCP);
			return (NOTOK);
		}
		if (crypt_flag) {
			parse_error ("invalid use of {CRYPT}",NULLCP);
			return (NOTOK);
		}
	} else if ((x->av_struct = (caddr_t) asn2pe (str)) == NULL) {
#ifdef DEBUG
		advise( NULLCP, "attrv.c::str2AttrV_aux invalid ASN.1 attribute str=%s sntx=%d", str==NULLCP ? "NULL" : str, syntax ) ;
#endif
		parse_error ("invalid ASN attribute",NULLCP);
		return NOTOK;
	}
	return (OK);
}

AttributeValue AttrV_cpy (x)
register AttributeValue x;
{
register AttributeValue y = NULLAttrV;

	y = AttrV_alloc();
	AttrV_cpy_aux (x,y);
	return (y);
}

AttrV_cpy_aux (x,y)
register AttributeValue x;
register AttributeValue y;
{
	y->av_syntax = x->av_syntax;

	if (x->av_struct == NULL) {
		y->av_struct = NULL;
		return;
	}

	if (syntax_table_basep[x->av_syntax].s_copy != NULLIFP) 
		y->av_struct = (caddr_t)(*syntax_table_basep[x->av_syntax].s_copy) (x->av_struct);
	else 
		y->av_struct = NULL;
}

AttrV_cpy_enc (x,y)
register AttributeValue x;
register AttributeValue y;
{
	/* Encode a copy of x into y */

	y->av_syntax = 0;

	if (x->av_struct == NULL) {
		y->av_struct = NULL;
		return;
	}

	if (syntax_table_basep[x->av_syntax].s_encode != NULLIFP) 
		y->av_struct = (caddr_t)(*syntax_table_basep[x->av_syntax].s_encode) (x->av_struct);
	else 
		y->av_struct = NULL;

	return;
}


rdn_cmp (a,b)
register RDN  a,b;
{
register int i;

	for (; (a != NULLRDN) && (b != NULLRDN) ; a = a->rdn_next, b = b->rdn_next) {
		if (a->rdn_at != b->rdn_at) 
			return ((a->rdn_at > b->rdn_at) ? 1 : -1);
		
		if (syntax_table_basep[a->rdn_av.av_syntax].s_compare == NULLIFP) 
			return (2); /* can't compare */
		else 
			if (( i = (*syntax_table_basep[a->rdn_av.av_syntax].s_compare) (a->rdn_av.av_struct,b->rdn_av.av_struct)) != 0)
				return i;
	}

	if ( (a == NULLRDN) && (b == NULLRDN) )  {
		return 0;
	} else {
		return ( a  ?  1  : -1);
	}

}

rdn_cmp_reverse (a,b)
register RDN  a,b;
{
register int i;

	for (; (a != NULLRDN) && (b != NULLRDN) ; a = a->rdn_next, b = b->rdn_next) {
		if (a->rdn_at != b->rdn_at) 
			return ((a->rdn_at > b->rdn_at) ? -1 : 1);
		
		if (syntax_table_basep[a->rdn_av.av_syntax].s_compare == NULLIFP) 
			return (2); /* can't compare */
		else 
			if (( i = (*syntax_table_basep[a->rdn_av.av_syntax].s_compare) (a->rdn_av.av_struct,b->rdn_av.av_struct)) != 0)
				return i;
	}

	if ( (a == NULLRDN) && (b == NULLRDN) )  {
		return 0;
	} else {
		return ( a  ?  1  : -1);
	}

}

AttrV_cmp (x,y)
register AttributeValue x,y;
{
	if (x->av_syntax != y->av_syntax) 
		return (-2);

	if (syntax_table_basep[x->av_syntax].s_compare != NULLIFP) 
		return ((*syntax_table_basep[x->av_syntax].s_compare) (x->av_struct,y->av_struct));
	else 
		return (2); /* can't compare */
}

IFP av_cmp_fn (syntax)
int syntax;
{
	if ( syntax >= AV_WRITE_FILE )
		return NULLIFP;

	if (syntax_table_basep[syntax].s_compare != NULLIFP) 
		return (syntax_table_basep[syntax].s_compare);
	else
		return NULLIFP;
}

AttrV_print (ps,x,format)
PS ps;
AttributeValue x;
int format;
{
extern int ps_printf ();

	if (format == RDNOUT)
	    format = EDBOUT;

	if (syntax_table_basep[x->av_syntax].s_print != NULLIFP) {
		if (x->av_struct != NULL)
			(*syntax_table_basep[x->av_syntax].s_print) (ps,x->av_struct,format);
	} else if (format == READOUT) {
                vpushquipu (ps);
                vunknown ((PE)x->av_struct);
                vpopquipu ();
	} else 
		pe_print (ps,(PE)x->av_struct,format);
}


#ifdef DEBUG
short oldstr2syntax (str)
char * str;
{
register sntx_table * ptr;
register int i;

	for (i=0, ptr = &syntax_table_basep[0] ; i<num_syntax; i++,ptr++) 
		if ( lexequ (ptr->s_sntx,str) == 0)
			return (i);

	return (0);
}

void
checkSBuckets()
{
int	b, i, cnt, totcnt=0, mymin=1000000, mymax=0 ;
struct  syntaxpair * p ;

	for(b=0;b<SBUCKETS;b++ ) {
	    
	    for( p = Sbuckets[b], cnt=0 ; p != NULL ; p=p->s_chain )
 	    	++cnt ;

	    fprintf( stderr, "SBucket %d=%d\n", b, cnt ) ;
	    totcnt += cnt ;

	    if( cnt < mymin ) mymin = cnt ;
	    if( cnt > mymax ) mymax = cnt ;

	    if( cnt > 5 ) {
	    	for( p = Sbuckets[b] ; p != NULL ; p=p->s_chain)
		    fprintf( stderr, "%s ", ((sntx_table*)&syntax_table_basep[p->s_entryNum])->s_sntx ) ;

		fprintf( stderr, "\n" ) ;
	    }
	}

	fprintf( stderr, "Buckets=%d TOTAL=%d MEAN=%d MIN=%d MAX=%d\n", SBUCKETS, totcnt, totcnt/SBUCKETS,
			mymin, mymax ) ;

	return ;
}

void
testSBuckets()
{
struct timeval  tvs, now;
int 	i,s  ;

	(void) gettimeofday (&tvs, (struct timezone *) 0);
	
	for( i=0 ; i<100 ; i++ )
	    s = oldstr2syntax( "IpNetToMediaEntryIdValue" ) ;

        (void) gettimeofday (&now, (struct timezone *) 0);
	now.tv_sec -= tvs.tv_sec;
	if ((now.tv_usec -= tvs.tv_usec) < 0)
	    now.tv_sec--, now.tv_usec += 1000000;
	advise (NULLCP,"100 OLD syntax retrieved in %d.%06d seconds", now.tv_sec, now.tv_usec);

	(void) gettimeofday (&tvs, (struct timezone *) 0);
	
	for( i=0 ; i<100 ; i++ )
	    s = str2syntax( "IpNetToMediaEntryIdValue" ) ;

        (void) gettimeofday (&now, (struct timezone *) 0);
	now.tv_sec -= tvs.tv_sec;
	if ((now.tv_usec -= tvs.tv_usec) < 0)
	    now.tv_sec--, now.tv_usec += 1000000;
	advise (NULLCP,"100 NEW syntax retrieved in %d.%06d seconds", now.tv_sec, now.tv_usec);

	(void) gettimeofday (&tvs, (struct timezone *) 0);
	
	for( i=0 ; i<100 ; i++ )
	    s = oldstr2syntax( "IpNetToMediaEntryIdValue" ) ;

        (void) gettimeofday (&now, (struct timezone *) 0);
	now.tv_sec -= tvs.tv_sec;
	if ((now.tv_usec -= tvs.tv_usec) < 0)
	    now.tv_sec--, now.tv_usec += 1000000;
	advise (NULLCP,"100 OLD syntax retrieved in %d.%06d seconds", now.tv_sec, now.tv_usec);

	(void) gettimeofday (&tvs, (struct timezone *) 0);
	
	for( i=0 ; i<100 ; i++ )
	    s = str2syntax( "IpNetToMediaEntryIdValue" ) ;

        (void) gettimeofday (&now, (struct timezone *) 0);
	now.tv_sec -= tvs.tv_sec;
	if ((now.tv_usec -= tvs.tv_usec) < 0)
	    now.tv_sec--, now.tv_usec += 1000000;
	advise (NULLCP,"100 NEW syntax retrieved in %d.%06d seconds", now.tv_sec, now.tv_usec);

	return ;
}

#endif
