/******************************************************************/
/*
	DebugMem.cc

	Memory debugging code that takes over new and delete
*/
/******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "DebugMem.h"

/******************************************************************/

#ifdef	DEBUG_MEMORY

/******************************************************************/

#define	MAXMEM	30000

/******************************************************************/

char	*curFile = "(unknown)";
int	curLine	= 0;

typedef	struct	MemTrace
{
	char		*file;
	int		line;
	void		*mem;
	MemTrace	*next_free;
};

static	MemTrace	*memTrace, *memFree;

static	int	newCount;

/******************************************************************/

void	init_mem_trace()
{

	if (!memTrace)
		memTrace = (MemTrace *)
			    calloc((size_t)MAXMEM, (size_t)sizeof(MemTrace));

	memFree	= &memTrace[0];

	for (int count=1; count<MAXMEM;	count++)

			memTrace[count-1].next_free = &memTrace[count];

} /* init_mem_trace */

/******************************************************************/

extern	void	*operator new(size_t new_size)
{
void	*res;

	res = (void *) malloc(new_size);

	// If before memory trace initialised

	if (!memTrace)
		return(res);

	if (curLine)
	{
		if (!memFree)
			printf("\nOut of Memory	Trace space");
		else
		{
			memFree->file =	curFile;
			memFree->line =	curLine;
			memFree->mem = res;
/*
			printf("\nNew Memory %ld File %s Line %d Alloc %d",
				memFree->mem, memFree->file, memFree->line,
				memFree	- &memTrace[0]);
*/
			memFree	= memFree->next_free;
		}
	}

	curFile	= "(unknown)";
	newCount++;
	return(res);

}

/******************************************************************/

extern	void	operator delete(void *mem)
{
	// If before memory trace initialised

	if (!memTrace)
	{
		if (mem)
			free((char *)mem);
		return;
	}
	for (int count=0; count<MAXMEM;	count++)
		if (memTrace[count].mem	== mem)
		{
/*
			printf("\nDelete Memory	%ld File %s Line %d Count %d",
				memTrace[count].mem, memTrace[count].file,
				memTrace[count].line, count);
*/
			memTrace[count].file = "";
			memTrace[count].line = 0;
			memTrace[count].mem = 0;
			memTrace[count].next_free = memFree;
			memFree	= &memTrace[count];
			newCount--;
			break;
		}

	if (mem)
		free((char *)mem);
}

/******************************************************************/

void	print_mem_trace()
{
	// If before memory trace initialised

	if (!memTrace)
		return;

	for (int count=0; count<MAXMEM;	count++)
		if (memTrace[count].mem)

		printf("\nMemory %ld not deleted File %s Line %d Count %d",
			memTrace[count].mem, memTrace[count].file,
			memTrace[count].line, count);

	printf("\nNumber of unfreed objects: %d\n", newCount);

} /* print_mem_trace */

/******************************************************************/

#endif

/******************************************************************/
/* End of program */
/******************************************************************/


