//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// GenArray.cc - generic array class
 
// By George Pavlou, February 1991


#include <stdlib.h>
#ifdef  __GNUG__
#include <std.h>
#endif
#include "GenArray.h"


int  Array::set (int pos, void* val)
{
    if (pos >= size)
	return NOTOK;

    if (cells[pos]) {
	deleteElem(cells[pos]);
	--count;
    }
    if (cells[pos] = val)
	++count;

    return OK;
}


int  Array::add (void* val)
{
    if (!val)
	return NOTOK;

    if (count == size)
	reset(size + incr);

    for (register int i = 0; i < size; i++)
	if (!cells[i])
	    break;

    ++count;
    cells[i] = val;
    return i;
}


void*  Array::remove (int pos)
{
    if (pos >= size)
	return NULLVD;

    void* elem = cells[pos];
    if (!elem)
	return NULLVD;

    cells[pos] = NULLVD;
    --count;
    return elem;
}


int  Array::remove (void* item)
{
    for (register int i = 0, j = 0; i < size; i++) {
	if (!cells[i])
	    continue;
	if (compare(item, cells[i]) == 0) {
	    deleteElem(cells[i]);
	    cells[i] = NULLVD;
	    return i;
	}
	if (++j == count)
	    return NOTOK;
    }

    return NOTOK;	// should not be reached
}


int  Array::find (void* item)
{
    for (register int i = 0, j = 0; i < size; i++) {
	if (!cells[i])
	    continue;
	if (compare(item, cells[i]) == 0)
	    return i;
	if (++j == count)
	    return NOTOK;
    }

    return NOTOK;	// should not be reached
}


void  Array::clear ()
{
    for (register int i = 0; i < size; i++)
	if (cells[i]) {
	    if (del)
		deleteElem(cells[i]);
	    cells[i] = NULLVD;
	}
    count = 0;
}


void  Array::compact ()
{
    int i, k = 0;
    register int j;

    for (i = 0; i < count; )
	if (cells[i])
	    i++;
	else {
	    k++;
	    for (j = i; j < size-k; j++)
		cells[j] = cells[j+1];
	    cells[j] = NULLVD;
	}
}


void  Array::reset (int sz)
{
    void** prev;
    int    cprev;

    if (sz == size)
	return;

    compact();
    prev  = cells;
    cprev = count;
    cells = new void*[sz];

    for (register int i = 0; i < sz; i++)
	if (sz < cprev)
	    cells[i] = prev[i];
	else
	    if (i < cprev)
		cells[i] = prev[i];
	    else
		cells[i] = NULLVD;

    if (sz < cprev) {
	for (i = sz; i < cprev; i++)
	    deleteElem(prev[i]);
	count = sz;
    }
    size = sz;

    if (prev)
	delete[] prev;
}


Array::Array (int sz, Bool d)
{
    del = d;
    size = sz;
    count = 0;
    incr = DEFAULT_INCR;
    cells = new void*[sz];
    for (register int i = 0; i < sz; i++)
	cells[i] = NULLVD;
}
