//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// GenArray.h - generic array

// By George Pavlou, June 1992


#ifndef _GEN_ARRAY_
#define _GEN_ARRAY_


#ifndef _BOOLEAN_
#include "Boolean.h"
#endif

#if !defined(OK) || !defined(NOTOK)
#define OK	0
#define NOTOK	-1
#endif

#if !defined(NULLVD)
#define	NULLVD	((void*) 0)
#endif


class Array
{
#define DEFAULT_INCR	20
    int	   incr;
    int	   size;
    int	   count;
    void** cells;

protected:
    Bool   del;

    virtual void deleteElem (void*);
    virtual int	 compare (void*, void*);

public:
    void*  get (int);
    int	   set (int, void*);
    int	   add (void*);
    void*  remove (int);
    int	   remove (void*);
    int	   find (void*);
    void   clear ();
    int	   getSize ();
    int	   getCount ();
    void   compact ();
    void   reset (int);
    void   setDeleteMode (Bool);
    void   setIncrement (int);

    Array (Bool = True);
    Array (int, Bool = True);
    virtual ~Array ();
};
#define NULLARRAY	((Array*) 0)



// inline methods

inline void  Array::deleteElem (void* elem)
{ if (del) delete elem; }

inline int   Array::compare (void*, void*)
{ return NOTOK; }

inline void*  Array::get (int pos)
{ return pos < size ? cells[pos] : NULLVD; }

inline int  Array::getSize ()
{ return size; }

inline int  Array::getCount ()
{ return count; }

inline void  Array::setDeleteMode (Bool d)
{ del = d; }

inline void  Array::setIncrement (int i)
{ incr = i; }

inline Array::Array (Bool d)
{ cells = 0; del = d; size = count = 0; incr = DEFAULT_INCR; }

inline Array::~Array ()
{ clear(); }


#endif
