//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// GenList.cc - generic linked list realisation

// By George Pavlou, February 1991


#include <stdlib.h>
#include "GenList.h"
#include "ListIterator.h"



void  List::append (void* item) 
{
    if (l)
	l = l -> next = new Link(item, l -> next);
    else {
	l = new Link(item, NULLINK);
	l -> next = l;
    }
    count++;
}


void  List::prepend (void* item)
{
    if (l)
	l -> next = new Link(item, l -> next);
    else {
	l = new Link(item, NULLINK);
	l -> next = l;
    }
    count++;
}


void*  List::get ()
{
    if (l == NULLINK)
	return NULLVD;
    Link* first = l -> next;
    void* item = first -> item;
    if (first == l)
	l = NULLINK;
    else
	l -> next = first -> next;
    delete first;
    count--;
    return item;
}


void*  List::getLast ()
{
    if (l == NULLINK)
	return NULLVD;

    void* item;

    if (l -> next == l) {	// list with one element
	item = l -> item;
	delete l;
	l = NULLINK;
    }
    else {
	Link* ll;
	for (ll = l -> next; ll -> next != l; ll = ll -> next);
	ll -> next = l -> next;

	ll = l;
	item = l -> item;
	l = l -> next;
	delete ll;
    }

    count--;
    return item;
}


int  List::insert (void* item)
{
    if (l == NULLINK || compare(item, l -> item) >= 0)
	append(item);
    else {
	Link* cur;
	for (cur = l; cur -> next != l; cur = cur -> next)
	    if (compare(item, cur->next->item) <= 0)
		break;
	cur -> next = new Link(item, cur -> next);
	count++;
    }

    return OK;
}


void*  List::remove (void* item)
{
    if (l == NULLINK)
	return NULLVD;	// empty list

    Link* dlink;
    void* rem_item;

    if (compare(item, l -> item) == 0) {		// remove last
	Link* cur;
	for (cur = l; cur -> next != l; cur = cur -> next);
	cur -> next = l -> next;
	dlink = l;
	if (l -> next == l)
	    l = NULLINK;
	else {
	    cur -> next = l -> next;
	    l = cur;
	}
    }
    else {
	Link* cur;
	for (cur = l; cur -> next != l; cur = cur -> next)
	    if (compare(item, cur->next->item) == 0)
		break;
	if (cur -> next == l)
	    return NULLVD;	// can't find it
	dlink = cur -> next;
	cur -> next = dlink -> next;
    }
    rem_item = dlink -> item;
    delete dlink;
    count--;
    return rem_item;
}


void*  List::find (void* item)
{
    if (l == NULLINK)
	return NULLVD;	// empty list

    if (compare(item, l -> item) == 0)
	return l->item;	// last

    for (Link* cur = l; cur -> next != l; cur = cur -> next)
	if (compare(item, cur->next->item) == 0)
	    return cur->next->item;
    return NULLVD;	// can't find it
}


void  List::clear ()
{
    if (l == NULLINK)
	return;
    Link* ll = l;
    do {
	Link* cur = ll;
	ll = ll -> next;
	if (del)
	    deleteElem(cur -> item);
	delete cur;
    } while (ll != l);
    l = NULLINK;
    count = 0;
}


void*  ListIterator::getNext ()
{
    if (last) {
	link = NULLINK;			// finished iterating
	last = False;
	return NULLVD;
    }

    if (link == NULLINK) {
	link = list -> l;
	if (link) {
	    link = link -> next;
	    if (link == list -> l)
		last = True;
	}
    }
    else {
	link = link -> next;
	if (link == list -> l)
	    last = True;
    }

    if (link)
	return link -> item;
    return NULLVD;
}
