//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// MAction.cc - enables to perform an action on managed objects in
//		a remote MIB

// By George Pavlou, January 1993

/*
** Security hooks added March/April 1994
** Saleem
*/

#include <stdlib.h>
#if	defined(__GNUG__) || (defined(__cplusplus) && defined(AIXV3))
extern "C" { int  strncasecmp(const char*, const char*, size_t); }
#endif

#include "msap.h"
#include "msap_u.h" /*XMP version calls*/

#include "Error.h"
#include "Syntaxes.h"
#include <memory.h>
#if defined(SECURITY)
#include "mauth.h"
#include "mintegrity.h"
#endif

#if defined(DEBUG)
#include "msap-pp.h"
FILE *pduDump = (FILE *) 0;
#endif

int request_id = 0,
    request_invoke = 0,
    invoke = 0;
MSAPIndication mis;

int   parseAvaArg (int, char**, char**, char**, int*);
int   parseFilterArg (int, char**, char**, int*);

static int   do_action (int, OID, MN, CMISScope*, CMISFilter*, CMISSync, Ava*,
			MSAPIndication*);
static int   do_wait (int, MSAPIndication*);

main (int argc, char** argv)
{
    char* strObjClass = NULLCP;
    char* strObjInst = NULLCP;
    char* strActionType = NULLCP;
    char* strActionValue = NULLCP;
    char* strScope = NULLCP;
    char* strSync = NULLCP;
    char* strFilter = NULLCP;
    int   i, j, state = 0;

    OID objClass = NULLOID;
    MN  objInst  = NULLMN;
    Ava* ava	 = NULLAVA;
    CMISScope* scope = NULLMSCOPE;	// default
    CMISSync sync = s_bestEffort;	//  ..
    CMISFilter* filter = NULLMFILTER;	//  ..
#ifdef SECURITY
    int IFlag = 0;
    char *peer = (char *) 0,
         *dsa = (char *) 0;
#endif

    if (argc < 5)
	goto usage;

    for (i = 3; i < argc; i++)
	switch (state) {
	case 0:
	    if (*argv[i] == '-' && strlen(argv[i]) == 2)
		switch (*(argv[i]+1)) {
		case 'c':
		    state = 1;
		    break;
		case 'i':
		    state = 2;
		    break;
		case 's':
		    state = 3;
		    break;
		case 'f':
		    state = 4;
		    break;
		case 'a':
		    state = 5;
		    break;
#ifdef SECURITY
                case 'P':
                    state = 6;
                    break;
                case 'D':
                    state = 7;
                    break;
                case 'I':
                    state = 0;
                    IFlag = 1;
                    break;
#endif
		default:
		    goto usage;
		    }
	    else
		goto usage;
	    break;

	case 1:
	    strObjClass = argv[i];
	    state = 0;
	    break;

	case 2:
	    strObjInst = argv[i];
	    state = 0;
	    break;

	case 3:
	    strScope = argv[i];
	    if (i < argc-1 && *argv[i+1] != '-')
		strSync = argv[++i];
	    state = 0;
	    break;

	case 4:
	    if (parseFilterArg(argc-i, &argv[i], &strFilter, &j) == NOTOK)
		goto usage;
	    i += j-1;
	    state = 0;
	    break;

	case 5:
	    if (parseAvaArg(argc-i, &argv[i],
			&strActionType, &strActionValue, &j) == NOTOK)
		goto usage;
	    i += j;
	    state = 0;
	    break;
#ifdef SECURITY
        case 6:
            peer = argv[i];
            state = 0;
            break;
        case 7:
            dsa = argv[i];
            state = 0;
            break;
#endif
	}

    if (!strActionType)
	goto usage;

    // load syntaxes

    if (initialiseSyntaxes(NULLCP, attrSyntaxes) == NOTOK)
	error(NULLCP, "maction: cannot initialise syntaxes");

#ifdef SECURITY
    initialiseOsisec(argc, argv);
    initialiseMd5KeySpace();
    initialiseIdSpace();
#endif

    // convert the string arguments to CMIS parameters

    if (strObjClass && (objClass = name2oid(strObjClass)) == NULLOID)
	error(NULLCP, "invalid object class: %s", strObjClass);

    if (strObjInst && (objInst = ObjectInstance_parse(strObjInst)) == NULLMN)
        error(NULLCP, "invalid object instance %s", strObjInst);

    if (!(ava = str2ava(strActionType, strActionValue)))
        error(NULLCP, "invalid action type \"%s\" or value \"%s\"",
			strActionType, strActionValue);

    if (strScope && (scope = str2scope(strScope)) == NULLMSCOPE)
	error(NULLCP, "invalid scope %s", strScope);

    if (strSync && (sync = str2sync(strSync)) == s_invalid)
	error(NULLCP, "invalid sync %s", strSync);
	
    if (strFilter && strFilter[0] != '\0'  &&
		(filter = str2mfilter(strFilter)) == NULLMFILTER)
	error(NULLCP, "invalid filter %s", strFilter);


    // now try to connect to the remote agent
    int msd;

    MInitOption *miop[2]; /* Connect options */

#ifdef SECURITY
    MInitOption mio;
    AuthAssocIntegrityInfo *aaii;
    aaii = AuthAssocIntegrityInfo_make();

    if (peer != (char *) 0)
        aaii->peer = strdup(peer);
    if (dsa != (char *) 0)
        aaii->dsa = strdup(dsa);
    if (IFlag)
        memcpy((void *) aaii->sessionKey, (void *) makeMd5Key(), 8);

    mio.optionId = MIO_AuthAssocIntegrityInfo;
    mio.optionInfo = (void *) aaii;

    miop[0] = &mio;
    miop[1] = (MInitOption *) 0;

#else

   miop[0] = miop[1] = (MInitOption *) 0;

#endif

    if ((msd = GMD_initialise_u(argv[1], argv[2], miop)) < 0)
	error(NULLCP, "cannot establish association to %s at %s",
			argv[1], argv[2]);

    // perform the M-ACTION request

    printf("\n\n");

    do_action(msd, objClass, objInst, scope, filter, sync, ava, &mis);

    // wait for the replies
#if defined(DEBUG)
    if ((pduDump = fopen("maction.rx.log", "w")) != (FILE *) 0)
        ppSetOutputFile(pduDump);
#endif

    // terminate

    m_terminate_u(msd);

#if defined(DEBUG)
    if (pduDump != (FILE *) 0)
        fclose(pduDump);
#endif

    exit(OK);

usage:
    fprintf(stderr, "usage: maction <agent> <host>\n\
\t-c <class> [-i <instance>]\n\
\t[-s <scope> [<sync>]] [-f filter] -a <actionType[=<actionValue>]\n");
#if defined(SECURITY)
    fprintf(stderr, "\t[-P <peer DN>] [-D <dsa name/addr>] [-I]\n");
#endif

    exit(0);
}


// performs an action
static int  do_action (int msd, OID objClass, MN objInst, CMISScope* scope,
			CMISFilter* filter, CMISSync sync, Ava* action,
			MSAPIndication* mi)
{
    struct MIDentifier	obj_class;
    struct MIDentifier	actionType;
    PE		actionValue;
    int		status;

    obj_class.mid_type = MID_GLOBAL;
    obj_class.mid_global = objClass;

    if (!objInst) {
        objInst = new  MName;
        objInst -> mn_type = MN_LOCALDN;
        objInst -> mn_dn = dn_comp_new(NULLRDN);
    }

    actionType.mid_type   = MID_GLOBAL;
    actionType.mid_global = action->ava_oid;

    actionValue		  = action->ava_value;

    request_id = 1;

    status = M_ActionAux_u_sync(msd, M_ACTIONC, request_id, &obj_class, 
			objInst, 
			scope,
			filter, NULLMACCESS, sync,
			&actionType, action->ava_value, mi);
    request_invoke = mi->mi_invoke;

    // release memory of arguments (for the sake of it...)

    mid_free(&obj_class);
    mn_free(objInst);
    delete objInst;
    if (scope)
	free((char*) scope);
    if (filter)
	mfilter_free(filter);
    ava_free(action);

    return status;
}

