//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// MCreate.cc - enables to create a managed object in a remote MIB

// By George Pavlou, January 1993


#include <stdlib.h>
#if	defined(__GNUG__) || (defined(__cplusplus) && defined(AIXV3))
extern "C" { int  strncasecmp(const char*, const char*, size_t); }
#endif

#include "msap.h"
#include "msap_u.h" /*XMP version calls*/

#include "Error.h"
#include "Syntaxes.h"
#include <memory.h>

#if defined(SECURITY)
#include "mauth.h"
#include "mintegrity.h"
#endif

#if defined(DEBUG)
#include "msap-pp.h"
FILE *pduDump = (FILE *) 0;
#endif

#define MAXATTRS	50

int request_id = 0,
    request_invoke = 0,
    invoke = 0;
MSAPIndication mis;

int   parseAvaArg (int, char**, char**, char**, int*);

static int   do_create (int, MID, MN, int, MN, int, CMISParam*,
			MSAPIndication*);
static int   do_wait (int, MSAPIndication*);

main (int argc, char** argv)
{
    char* strObjClass = NULLCP;
    char* strObjInst = NULLCP;
    char* strRefInst = NULLCP;
    char* strAttrType [MAXATTRS];
    char* strAttrValue [MAXATTRS];
    int nattrs = 0;
    int   i, j, state = 0;

    struct MIDentifier objClass;
    objClass.mid_type = MID_GLOBAL;
    objClass.mid_global = NULLOID;
    MN objInst = (MN) 0;
    MN refInst = (MN) 0;
    int instType;
    CMISParam attrs [MAXATTRS];
#ifdef SECURITY
    int IFlag = 0;
    char *peer = (char *) 0,
         *dsa = (char *) 0;
#endif

    if (argc < 5)
	goto usage;

    for (i = 3; i < argc; i++)
	switch (state) {
	case 0:
	    if (*argv[i] == '-' && strlen(argv[i]) == 2)
		switch (*(argv[i]+1)) {
		case 'c':
		    state = 1;
		    break;
		case 'i':
		    state = 2;
		    break;
		case 's':
		    state = 3;
		    break;
		case 'r':
		    state = 4;
		    break;
		case 'a':
		    state = 5;
		    break;
#ifdef SECURITY
                case 'P':
                        state = 6;
                        break;
                case 'D':
                        state = 7;
                        break;
                case 'I':
                        state = 0;
                        IFlag = 1;
                        break;
#endif
		default:
		    goto usage;
		}
	    else
		goto usage;
	    break;

	case 1:
	    strObjClass = argv[i];
	    state = 0;
	    break;

	case 2:
	    strObjInst = argv[i];
	    instType = CA_OBJECT_INST;
	    state = 0;
	    break;

	case 3:
	    strObjInst = argv[i];
	    instType = CA_SUPERIOR_INST;
	    state = 0;
	    break;

	case 4:
	    strRefInst = argv[i];
	    state = 0;
	    break;

	case 5:
	    if (parseAvaArg(argc-i, &argv[i], &strAttrType[nattrs],
			&strAttrValue[nattrs], &j) == NOTOK ||
			!strAttrValue[nattrs] ||
			strAttrValue[nattrs][0] == '\0')
		goto usage;
	    ++nattrs;
	    i += j - 1;
	    state = 0;
	    break;
#ifdef SECURITY
        case 6:
                peer = argv[i];
                state = 0;
                break;
        case 7:
                dsa = argv[i];
                state = 0;
                break;
#endif
	}

    if (!strObjClass)
	goto usage;

    // load syntaxes

    if (initialiseSyntaxes(NULLCP, attrSyntaxes) == NOTOK)
	error(NULLCP, "mcreate: cannot initialise syntaxes");

#ifdef SECURITY
    initialiseOsisec(argc, argv);
    initialiseMd5KeySpace();
    initialiseIdSpace();
#endif


    // convert the string arguments to CMIS parameters

    if ((objClass.mid_global = name2oid(strObjClass)) == NULLOID)
	error(NULLCP, "invalid object class: %s", strObjClass);

    if (strObjInst && (objInst = ObjectInstance_parse(strObjInst)) == NULLMN)
            error(NULLCP, "invalid object instance %s", strObjInst);

    if (strRefInst && (refInst = ObjectInstance_parse(strRefInst)) == NULLMN)
            error(NULLCP, "invalid reference instance %s", strRefInst);

    for (i = 0; i < nattrs; i++) {
	Ava* ava;
	if ((ava = str2ava(strAttrType[i], strAttrValue[i])) == NULLAVA)
	    error(NULLCP, "invalid attribute type \"%s\" or value \"%s\"",
			strAttrType[i], strAttrValue[i]);
	attrs[i].mp_id.mid_type = MID_GLOBAL;
	attrs[i].mp_id.mid_global = ava->ava_oid;
	attrs[i].mp_val = ava->ava_value;
	free ((char*) ava);
    }

    // now try to connect to the remote agent

    int msd;
    MInitOption *miop[2]; /* Connect options */

#ifdef SECURITY
    MInitOption mio;
    AuthAssocIntegrityInfo *aaii;
    aaii = AuthAssocIntegrityInfo_make();

    if (peer != (char *) 0)
        aaii->peer = strdup(peer);
    if (dsa != (char *) 0)
        aaii->dsa = strdup(dsa);
    if (IFlag)
        memcpy((void *) aaii->sessionKey, (void *) makeMd5Key(), 8);

    mio.optionId = MIO_AuthAssocIntegrityInfo;
    mio.optionInfo = (void *) aaii;

    miop[0] = &mio;
    miop[1] = (MInitOption *) 0;

#else

   miop[0] = miop[1] = (MInitOption *) 0;

#endif

    if ((msd = GMD_initialise_u(argv[1], argv[2], miop)) < 0)
        error(NULLCP, "cannot establish association to %s at %s",
                        argv[1], argv[2]);
#if defined(SECURITY)
    if (aaii != (AuthAssocIntegrityInfo *) 0)
        AuthAssocIntegrityInfo_free(aaii);
#endif

    // perform the M-CREATE request
    do_create(msd, &objClass, objInst, instType, refInst, nattrs, attrs, &mis);

#if defined(DEBUG)
    if ((pduDump = fopen("mcreate.rx.log", "w")) != (FILE *) 0)
        ppSetOutputFile(pduDump);
#endif

    // terminate

    m_terminate_u(msd);

#if defined(DEBUG)
    if (pduDump != (FILE *) 0)
        fclose(pduDump);
#endif

    exit(OK);

usage:
    fprintf(stderr, "usage: mcreate <agent> <host>\n\
\t-c <class>\n\
\t[-i <instance> | -s <superiorInst>] [-r referenceInst]\n\
\t[-a <attrType>=<attrVal>] ...\n");
#if defined(SECURITY)
    fprintf(stderr, "\t[-P <peer DN>] [-D <dsa name/addr>] [-I]\n");
#endif

    exit(0);
}

// perform a M-CREATE operation
static int  do_create (int msd, MID objClass, MN objInst, int instType,
		       MN refInst, int nattrs, CMISParam* attrs,
		       MSAPIndication* mi)
{
    int		status;

    request_id = 1;

    status = M_Create_u_sync(msd, request_id, objClass,
			objInst, instType, refInst,
			NULLMACCESS, nattrs, attrs, mi);
    request_invoke = mi->mi_invoke;

    // release memory of arguments (for the sake of it...)

    mid_free(objClass);
    if (objInst->mn_dn) {
	mn_free(objInst);
        delete objInst;
    }
    if (refInst->mn_dn) {
	mn_free(refInst);
        delete refInst;
    }
    for (register int i = 0; i < nattrs; i++)
	mparm_free(&attrs[i]);

    return status;
}

