//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// MSet.cc - enables to set management attributes in a remote MIB

// By George Pavlou, January 1993


#include <stdlib.h>
#if	defined(__GNUG__) || (defined(__cplusplus) && defined(AIXV3))
extern "C" { int  strncasecmp(const char*, const char*, size_t); }
#endif

#include "msap.h"
#include "msap_u.h" /*XMP version calls*/

#include "Error.h"
#include "Syntaxes.h"
#include <memory.h>
#if defined(SECURITY)
#include "mauth.h"
#include "mintegrity.h"
#endif

#if defined(DEBUG)
#include "msap-pp.h"
FILE *pduDump = (FILE *) 0;
#endif

#define MAXATTRS	50

int request_id = 0,
    request_invoke = 0,
    invoke = 0;
MSAPIndication mis;

int   parseAvaArg (int, char**, char**, char**, int*);
int   parseFilterArg (int, char**, char**, int*);

static int   do_set (int, OID, MN, CMISScope*, CMISFilter*, CMISSync,
                        int, CMISSetAttr*, MSAPIndication*);
static int   do_wait (int, MSAPIndication*);

main (int argc, char** argv)
{
    char* strObjClass = NULLCP;
    char* strObjInst = NULLCP;
    char* strAttrType [MAXATTRS];
    char* strAttrValue [MAXATTRS];
    int   nattrs = 0;
    char* strScope = NULLCP;
    char* strSync = NULLCP;
    char* strFilter = NULLCP;
    int   i, j, state = 0;

    OID objClass = NULLOID;
    MN  objInst  = NULLMN;
    CMISSetAttr attrs [MAXATTRS];
    CMISScope* scope = NULLMSCOPE;	// default
    CMISSync sync = s_bestEffort;	//  ..
    CMISFilter* filter = NULLMFILTER;	//  ..
#ifdef SECURITY
    int IFlag = 0;
    char *peer = (char *) 0,
         *dsa = (char *) 0;
#endif

    if (argc < 5)
	goto usage;

    for (i = 3; i < argc; i++)
	switch (state) {
	case 0:
	    if (*argv[i] == '-' && strlen(argv[i]) == 2)
		switch (*(argv[i]+1)) {
		case 'c':
		    state = 1;
		    break;
		case 'i':
		    state = 2;
		    break;
		case 's':
		    state = 3;
		    break;
		case 'f':
		    state = 4;
		    break;
		case 'w':
		    state = 5;
		    break;
		case 'a':
		    state = 6;
		    break;
		case 'r':
		    state = 7;
		    break;
		case 'd':
		    state = 8;
		    break;
#ifdef SECURITY
                case 'P':
                    state = 9;
                    break;
                case 'D':
                    state = 10;
                    break;
                case 'I':
                    state = 0;
                    IFlag = 1;
                    break;
#endif
		default:
		    goto usage;
		    }
	    else
		goto usage;
	    break;

	case 1:
	    strObjClass = argv[i];
	    state = 0;
	    break;

	case 2:
	    strObjInst = argv[i];
	    state = 0;
	    break;

	case 3:
	    strScope = argv[i];
	    if (i < argc-1 && *argv[i+1] != '-')
		strSync = argv[++i];
	    state = 0;
	    break;

	case 4:
	    if (parseFilterArg(argc-i, &argv[i], &strFilter, &j) == NOTOK)
		goto usage;
	    i += j-1;
	    state = 0;
	    break;

	case 5:
	case 6:
	case 7:
	case 8:
	    switch (state) {
	    case 5:
		attrs[nattrs].sa_modify = m_replace;
		break;
	    case 6:
		attrs[nattrs].sa_modify = m_addValue;
		break;
	    case 7:
		attrs[nattrs].sa_modify = m_removeValue;
		break;
	    case 8:
		attrs[nattrs].sa_modify = m_setToDefault;
		break;
	    }
	    if (parseAvaArg(argc-i, &argv[i],
		    &strAttrType[nattrs], &strAttrValue[nattrs], &j) == NOTOK)
		goto usage;
	    ++nattrs;
	    i += j;
	    state = 0;
	    break;
#ifdef SECURITY
        case 9:
            peer = argv[i];
            state = 0;
            break;
        case 10:
            dsa = argv[i];
            state = 0;
            break;
#endif
	}

    if (!nattrs)
	goto usage;

    // load syntaxes

    if (initialiseSyntaxes(NULLCP, attrSyntaxes) == NOTOK)
	error(NULLCP, "mset: cannot initialise syntaxes");

#ifdef SECURITY
    initialiseOsisec(argc, argv);
    initialiseMd5KeySpace();
    initialiseIdSpace();
#endif

    // convert the string arguments to CMIS parameters

    if (strObjClass && (objClass = name2oid(strObjClass)) == NULLOID)
	error(NULLCP, "invalid object class: %s", strObjClass);

    if (strObjInst && (objInst = ObjectInstance_parse(strObjInst)) == NULLMN)
        error(NULLCP, "invalid object instance %s", strObjInst);

    if (strScope && (scope = str2scope(strScope)) == NULLMSCOPE)
	error(NULLCP, "invalid scope %s", strScope);

    if (strSync && (sync = str2sync(strSync)) == s_invalid)
	error(NULLCP, "invalid sync %s", strSync);
	
    if (strFilter && strFilter[0] != '\0'  &&
		(filter = str2mfilter(strFilter)) == NULLMFILTER)
	error(NULLCP, "invalid filter %s", strFilter);

    for (i = 0; i < nattrs; i++) {
	if (attrs[i].sa_modify == m_setToDefault) {
	    if (strAttrValue[i])
		advise(NULLCP, "default value \"%s\" not used",
							strAttrValue[i]);
	}
	else	// replace, add or remove
	    if (!strAttrValue[i]) {
		advise(NULLCP, "missing attribute value for \"%s\"",
							strAttrType[i]);
		goto usage;
	    }
	Ava* ava;
	if ((ava = str2ava(strAttrType[i], strAttrValue[i])) == NULLAVA)
	    error(NULLCP, "invalid attribute type \"%s\" or value \"%s\"",
				strAttrType[i], strAttrValue[i]);
	attrs[i].sa_id.mid_type = MID_GLOBAL;
	attrs[i].sa_id.mid_global = ava->ava_oid;
	attrs[i].sa_val = ava->ava_value;
	free ((char*) ava);
    }

    // now try to connect to the remote agent
    int msd;

    MInitOption *miop[2]; /* Connect options */

#ifdef SECURITY
    MInitOption mio;
    AuthAssocIntegrityInfo *aaii;
    aaii = AuthAssocIntegrityInfo_make();

    if (peer != (char *) 0)
        aaii->peer = strdup(peer);
    if (dsa != (char *) 0)
        aaii->dsa = strdup(dsa);
    if (IFlag)
        memcpy((void *) aaii->sessionKey, (void *) makeMd5Key(), 8);

    mio.optionId = MIO_AuthAssocIntegrityInfo;
    mio.optionInfo = (void *) aaii;

    miop[0] = &mio;
    miop[1] = (MInitOption *) 0;

#else

   miop[0] = miop[1] = (MInitOption *) 0;

#endif

    if ((msd = m_initialise_u(argv[1], argv[2], miop)) < 0)
        error(NULLCP, "cannot establish association to %s at %s",
              argv[1], argv[2]);

#if defined(SECURITY)
    if (aaii != (AuthAssocIntegrityInfo *) 0)
        AuthAssocIntegrityInfo_free(aaii);
#endif

    // perform the M-SET request

    printf("\n\n");

    do_set(msd, objClass, objInst, scope, filter, sync, nattrs, attrs, &mis);

    // wait for the replies

#if defined(DEBUG)
    if ((pduDump = fopen("mset.rx.log", "w")) != (FILE *) 0)
        ppSetOutputFile(pduDump);
#endif

    while (do_wait(msd, &mis) != DONE);

    // terminate

    m_terminate_u(msd);

#if defined(DEBUG)
    if (pduDump != (FILE *) 0)
        fclose(pduDump);
#endif

    exit(OK);

usage:
    fprintf(stderr, "usage:mset <agent> <host>\n\
\t-c <class> [-i <instance>]\n\
\t[-s <scope> [<sync>]] [-f filter] -[w|a|r|d] <attrType[=<attrValue>] ...\n");
#if defined(SECURITY)
    fprintf(stderr, "\t[-P <peer DN>] [-D <dsa name/addr>] [-I]\n");
#endif

    exit(0);
}


// does the set
static int  do_set (int msd, OID objClass, MN objInst,
			CMISScope* scope, CMISFilter* filter, CMISSync sync,
			int nattrs, CMISSetAttr* attrs, MSAPIndication* mi)
{
    struct MIDentifier	obj_class;
    int		status;

    obj_class.mid_type = MID_GLOBAL;
    obj_class.mid_global = objClass;

    if (!objInst) {
        objInst = new  MName;
        objInst -> mn_type = MN_LOCALDN;
        objInst -> mn_dn = dn_comp_new(NULLRDN);
    }


    request_id = 1;

    status = M_SetAux_u(msd, M_SETC, request_id, &obj_class, objInst, scope,
                       filter, NULLMACCESS, sync,
                       nattrs, attrs, mi);
    request_invoke = mi->mi_invoke;

    // release memory of arguments (for the sake of it...)

    mid_free(&obj_class);
    mn_free(objInst);
    delete objInst;
    if (scope)
	free((char*) scope);
    if (filter)
	mfilter_free(filter);

    return status;
}

static int  do_wait (int msd, MSAPIndication* mi)
{
    int status = OK;
    static int nobjects = 0;
    Bool linked = False;

    M_WaitReq_u(msd, NOTOK, mi);

#if defined(DEBUG)
    if (pduDump != (FILE *) 0)
        ppMSAPIndication(mi);
#endif

    if (mi->mi_type == MI_SET_RES ||
        (mi->mi_type == MI_ERROR && mi->mi_errtype == m_setListError) ||
        mi->mi_type == MI_LINKED_REPLY) {
        struct CMISSetRes * s;
	int	id;

        switch (mi->mi_type) {
        case MI_SET_RES:
	case MI_ERROR:
	    s = &mi -> mi_setres.result;
	    id = mi -> mi_setres.id;
	    printf("Set Result - id %#08x, ", (u_int) id);
	    status = DONE;	// only one object in remote MIB
	    break;

        case MI_LINKED_REPLY:
	    id = mi -> mi_linkedreplyinv.args.lr_linkid;
            linked = True;
	    switch (mi -> mi_linkedreplyinv.args.lr_type) {
	    case LR_SET_RES:
	    case LR_SET_LISTERR:
		s = &mi -> mi_linkedreplyinv.args.lr_setres;
		break;
		
	    case LR_PROC_FAILURE:
		// should not really happen
		printf("Linked Reply Processing Failure - id %#08x\n",
                       (u_int) id);
		return status;
	    }
	    printf("Linked Reply Set - id %#08x, ", (u_int) id);
	    break;
        }

	if (id != (linked ? request_invoke :
                   request_id
                  )) {
	    // should not happen for a correctly behaving agent
	    printf("Set result with a wrong request id received!\n");
	    mi_free(mi);
	    return DONE;
	}

	++nobjects;

        if (s->sr_class.mid_global)
            printf("class %s, ", oid2name(s->sr_class.mid_global, OIDPART));
        if (s->sr_inst.mn_dn) {
	    char * inst;
	    inst = dn2str(s->sr_inst.mn_dn);
	    printf("instance {%s}", inst);
	    free(inst);
        }
        if (s->sr_time)
            printf(", time %s\n", s->sr_time);
        else
	    printf("\n");
        for (int i = 0; i < s->sr_nattrs; i++) {
	    char *attrid, *attrval;
	    ava2str(s->sr_attrs[i].sa_ava.mp_id.mid_global,
			s->sr_attrs[i].sa_ava.mp_val, &attrid, &attrval);
	    fprintf(stdout, "%s: %s  ", attrid, attrval);
	    free(attrid); free(attrval);
	    if (s->sr_attrs[i].sa_error)
	        fprintf(stdout, "Error: %s\n",
				MErrString(s->sr_attrs[i].sa_error));
	    else
		fprintf(stdout, "\n");
	}
	fprintf(stdout, "\n\n");
    }
    else
    if (mi->mi_type == MI_EMPTY_RES && mi->mi_emptyres.id ==
        request_id
        )
	status = DONE;
    else
    if (mi->mi_type == MI_PREJECT || mi->mi_type == MI_ABORT) {
	fprintf(stderr, "mibdump: received %s, exits\n",
                mi->mi_type == MI_PREJECT ?
                "provider reject" : "association abort");
#if defined(DEBUG)
        if (pduDump != (FILE *) 0)
            fclose(pduDump);
#endif

        exit(NOTOK);
    }
    else
    if (mi->mi_type == MI_ERROR || mi->mi_type == MI_UREJECT) {
	fprintf(stderr, "Error: %s\n", MErrString(mi->mi_errtype));
	status = DONE;
    }
    else {
	printf("Unexpected mindication type: %d\n", mi->mi_type);
	status = NOTOK;
    }

    if (status == DONE)
	printf("\n%d replies received\n", nobjects);
    mi_free(mi);
    return status;
}
