/* Copyright (c) 1993 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

/* Jonathan Ladipo, December 1993 */

 /*************************************************************************
 *
 * INTERFACE ROUTINE: M_CancelGet - cancel the result of a pending,
 * asynchronously-executing Get operation in an orderly manner.
 *
 *************************************************************************/

#include "xmp.h"
#include "msap.h"

#if !defined(_a_)
#if defined(__STDC__) || defined(__cplusplus)
#define _a_(_args_) _args_
#else
#define _a_(_args_) ()
#endif
#endif

extern MIDAS_CONTEXT context;
extern MIDAS_SESSION default_session;
extern MIDAS_WORKSPACE midas_workspace;
extern int xmp_initialised;
extern MIDAS_SESSION *session_get();
extern int arg_fail _a_((int, int, int, char*, MSAPIndication*));
extern char*  op2st _a_((int));
extern int debug_test _a_((char *));

#define missing_midP(p)         (!(p) || !(p)->mid_global)
#define missing_mnP(p)          (!(p) || !(p)->mn_dn)
#define null_midP(p)            (!(p)->mid_global)
#define null_mnP(p)             (!(p)->mn_dn)
#define null_mparmP(p)          (!(p)->mp_id.mid_global || !(p)->mp_val)
/*************************************************************************/

int
M_CancelGet_u(msd, invoke, get_invoke, mi)
    int  msd, invoke, get_invoke;
    MSAPIndication* mi; 

{
    CMISCancelGetArg 	argument;
    OM_sint32           invoke_id;
    MIDAS_STATUS_P      statp;
    MIDAS_SESSION	*sessp;

    /*check xmp has been initialised*/

    debug_test("In M_CancelGet_u");

    if (!xmp_initialised) {
        fprintf(stderr, "M-InitialiseReq: XMP not initialised\n");
        return(NOTOK);
    }

    if ((sessp = session_get(msd))
                == (MIDAS_SESSION *)NULL){
        /*assume UCL mode*/
        sessp = &default_session;
    }

    /*set up the context*/

    context.connection_id = msd;
    context.mode = MP_T_CONFIRMED;
    context.asynchronous = 1;

    /*set up the argument*/

/*
   NOTE
   The `invoke` argument must be passed to M_CancelGet called by 
   mp_cancel_get_req_M.
 */
    invoke_id = invoke;
    argument.cga_getinvoke = get_invoke; /* Invoke ID of GET operation
					    which is to be cancelled */

    /*Now do the cancel get operation*/

    statp = mp_cancel_get_req_M(sessp, &context, &argument,
                                   &invoke_id);

    if((statp == MIDAS_NO_WORKSPACE)||(statp == MIDAS_INVALID_SESSION)||
                (statp == MIDAS_INSUFFICIENT_RESOURCES)){

        fprintf(stderr, "M_CancelGet_u: local error\n");
        return(NOTOK);
    }

     else if(statp == MIDAS_SUCCESS){
        /*Since this was an async call there will be no result*/
	debug_test("M_CancelGet_u: @ 1\n");
        return(OK);
     }

     else{
        /*pointer to a valid MIDAS_STATUS structure*/
        /*copy status to caller's structure*/

         if(statp->MIDAS_parameter)
             *mi = *(statp->MIDAS_parameter);            /*struct copy*/
         else
             mi = (MSAPIndication*)NULL;

         debug_test("M_CancelGet_u: @ 2\n");
         return(NOTOK);
     }
}

