/* Copyright (c) 1993 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Jonathan Ladipo, November 1993 */

/*************************************************************************
 *
 * INTERFACE ROUTINE: M_Create_u - create a managed object
 *
 *************************************************************************/

#include "xmp.h"
#include "msap.h"
#include <stdio.h>

#if !defined(_a_)
#if defined(__STDC__) || defined(__cplusplus)
#define _a_(_args_) _args_
#else
#define _a_(_args_) ()
#endif
#endif

extern MIDAS_CONTEXT context;
extern MIDAS_SESSION default_session;
extern MIDAS_WORKSPACE midas_workspace;
extern int xmp_initialised;
extern MIDAS_SESSION *session_get();
extern int arg_fail _a_((int, int, int, char*, MSAPIndication*));
extern char*  op2str _a_((int));
extern int debug_test _a_((char *));
		
#define missing_midP(p)         (!(p) || !(p)->mid_global)
#define missing_mnP(p)          (!(p) || !(p)->mn_dn)
#define null_midP(p)            (!(p)->mid_global)
#define null_mnP(p)             (!(p)->mn_dn)
#define null_mparmP(p)          (!(p)->mp_id.mid_global || !(p)->mp_val)
/*************************************************************************/

int
M_Create_u(msd, invoke, obj_class, obj_inst, instance_type, reference_inst,
                        access, nattrs, attrs, mi)
    int  msd, invoke;
    MID  obj_class;
    MN   obj_inst;
    int instance_type;
    MN reference_inst;
    External* access;
    int nattrs;
    CMISParam* attrs;
    MSAPIndication* mi; {

    CMISCreateArg  argument;
    OM_sint32           invoke_id_return = (OM_sint32)invoke; /*NON
							    STANDARD XMP USAGE*/
    MIDAS_STATUS_P      statp;
    MIDAS_RESULT	*resultp;
    MIDAS_SESSION	*sessp;
    int 		i;

     debug_test("In M_Create_u @ 1");
    /*check xmp has been initialised*/

    if (!xmp_initialised) {
        fprintf(stderr, "M-InitialiseReq: XMP not initialised\n");
        return(NOTOK);
    }

    if ((sessp = session_get(msd))
                == (MIDAS_SESSION *)NULL){
        /*assume UCL mode*/
        sessp = &default_session;
    }

    /*set up the context*/

    context.connection_id = msd;
    context.mode = MP_T_CONFIRMED;
    context.asynchronous = 1;

    /*set up the argument*/

    if (nattrs > 0 && !attrs){
        arg_fail(invoke, M_CREATE, 0, "AttributeIdList", mi);
	exit(1);
    }

        argument.ca_class = *obj_class;     /* struct copy */

    if (nattrs > 0)
        for (i = 0; i < nattrs; i++)
            if (null_midP(&attrs[i].mp_id)){
                arg_fail(invoke, M_CREATE, 0, "AttributeId", mi);
		exit(1);
	    }
    argument.ca_nattrs = nattrs;
    argument.ca_attrs = attrs;          /*struct copy*/

    if(obj_inst)
	argument.ca_inst  = *obj_inst;      /*struct copy*/
    else
	argument.ca_inst.mn_type = MN_NONSPEC;

    if(reference_inst)
	argument.ca_reference_inst = *reference_inst;
    else
	argument.ca_reference_inst.mn_type = MN_NONSPEC;

    argument.ca_access = access;
    argument.ca_inst_type = instance_type;

    /*Now do the get*/

    statp = mp_create_req_M(sessp, &context, &argument,
                               &resultp, &invoke_id_return);

    if((statp == MIDAS_NO_WORKSPACE)||(statp == MIDAS_INVALID_SESSION)||
		(statp == MIDAS_INSUFFICIENT_RESOURCES)){

	fprintf(stderr, "M_Create: local error\n");
	return(NOTOK);
     }
    
     else if(statp == MIDAS_SUCCESS){
	/*Since this was an async call there will be no result*/
        debug_test("In M_Create_u @ 2");

     	return(OK);
     }

     else{
    	/*pointer to a valid MIDAS_STATUS structure*/
	/*copy status to caller's structure*/
	 debug_test("In M_Create_u @ 3");

         if(statp->MIDAS_parameter)
             *mi = *(statp->MIDAS_parameter);            /*struct copy*/
         else
             mi = (MSAPIndication*)NULL;

         return(NOTOK);
     }
    
}

