/* Copyright (c) 1993 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "xmp.h"
#include "midas_xmp.h"
#include "msap.h"

#if !defined(_a_)
#if defined(__STDC__) || defined(__cplusplus)
#define _a_(_args_) _args_
#else
#define _a_(_args_) ()
#endif
#endif

/*
 *#define DO_DEBUG
 *For debbugging purposes only. It is used to debug 
 *libXMP_use and libXMP_provide function calls.
 */

/*#define DO_DEBUG	Remove this line when debugging is complete*/


#define null_midP(p)            (!(p)->mid_global)
#define null_mnP(p)             (!(p)->mn_dn)
#define null_mparmP(p)          (!(p)->mp_id.mid_global || !(p)->mp_val)

extern void dn_print();
extern char* MErrString _a_((CMISErrors));

/*
 * print_status - print status of a get request
 * 
 */

int print_status(statp)
    MIDAS_STATUS *statp; {

    if (statp == (MIDAS_STATUS *)NULL){
	fprintf(stderr, "print_status: NULL statp\n");
	return(NOTOK);
    }

    print_ind(statp->MIDAS_parameter);
    return (OK);
}

/*
 * print_result - print result of a get request
 *
 */

int print_result(resultp)
    MIDAS_RESULT *resultp; {

    if (resultp == (MIDAS_RESULT *)NULL){
	fprintf(stderr, "print_result: NULL resultp\n");
	return(NOTOK);
    }

    print_ind(&resultp->MIDAS_result_info);
    debug_test("Leaving print_result @ 1");
    return (OK);

}

int print_ind(mi)
    MSAPIndication *mi;

{
    
    debug_test("In print_ind @ 1");

#ifdef DO_DEBUG
    printf("print_ind: mi->mi_type = %d\n", mi->mi_type);
    printf("print_ind: mi->mi_errtype = %s\n", MErrString(mi->mi_errtype));
#endif

    switch (mi->mi_type){
        /*Only one reply which was succesfully received*/
        case MI_GET_RES:
            return (print_get_ind(mi));
	case MI_SET_RES:
            return (print_set_ind(mi));
        case MI_CREATE_RES:
            return (print_create_ind(mi));
	case MI_ACTION_RES:
	    return (print_action_ind(mi));
        case MI_DELETE_RES:
            debug_test("In print_ind @ 2");
            return (print_delete_ind(mi));

        /*Get has linked replies one of which has been succesully received*/
        case MI_LINKED_REPLY:
            debug_test("In print_ind @ 3");
	    switch (mi-> mi_linkedreplyinv.args.lr_type) {

                case LR_GET_RES:
        	case LR_GET_LISTERR:
                /*Get has linked replies one of which
                  has beensuccesully received*/
                    return (print_get_ind(mi));

                case LR_SET_RES:
        	case LR_SET_LISTERR:
                /*Set has linked replies one of which
                  has beensuccesully received*/
                    return (print_set_ind(mi));

                case LR_ACTION_RES:
                /*Action has linked replies one of which
                  has beensuccesully received*/
                    return (print_action_ind(mi));

                case LR_DELETE_RES:
	            debug_test("In print_ind @ 4");
                /*Delete has linked replies one of which
                  has beensuccesully received*/
                    return (print_delete_ind(mi));

/*
   For LR_XXX_ERR`s we need to make a function similar to print_xx_ind
   which examines the contents of the MSAPIndication 'mi' and prints a more
   descriptive idea of what the problem is. If an error has occurred 'mi'
   should contain all the error info inside. So, all we need to do is examine
   it and let the user see what's going on.
 */

        	case LR_ACTION_ERR:
                    printf("Linked Reply Action Error - id %d\n",
                                mi -> mi_linkedreplyinv.args.lr_linkid);
                    return (NOTOK);

        	case LR_DELETE_ERR:
                    printf("Linked Reply Delete Error - id %d\n",
                                mi -> mi_linkedreplyinv.args.lr_linkid);
                    return (NOTOK);

        	case LR_PROC_FAILURE:
                    printf("Linked Reply Processing Failure - id %d\n", 
				mi -> mi_linkedreplyinv.args.lr_linkid);
                    return (NOTOK);

                default:
                /*All other possibilities. This should not really happen*/
                    printf("Unknown Linked Reply Type - id %d\n",
                                mi -> mi_linkedreplyinv.args.lr_linkid);
                    return (NOTOK);

	    }


            case MI_EMPTY_RES:
            /*This signifies the end of a sequence of linked replies*/
                return (DONE);

	    case MI_UREJECT:
		fprintf(stderr, " User reject received \n");
		return (NOTOK);

	    case MI_PREJECT:
		fprintf(stderr, " Provider reject received \n");
		return (NOTOK);
				
	    case MI_ABORT:
		fprintf(stderr, " Association abort  reject received \n");
		return (NOTOK);

            case MI_ERROR:
	        fprintf(stderr, "Error: %s\n", MErrString(mi->mi_errtype));
                return (NOTOK);

	    default:
	        printf("Unexpected mindication type: %d\n", mi->mi_type);
	        return (NOTOK);
    }


}


/* 
 * print_get_ind - print the indication from a get request
 *
 */

int print_get_ind(mi)
    MSAPIndication *mi;
{

    static int nobjects = 0;
    int i, id;
    struct CMISGetRes * g;
    char *inst, *attrid, *attrval;

	switch (mi->mi_type) {
        case MI_GET_RES:
	    g = &mi -> mi_getres.result;
	    id = mi -> mi_getres.id;
	    printf("Get Result - id %d, ", id);
	    break;

        case MI_LINKED_REPLY:
	    id = mi -> mi_linkedreplyinv.args.lr_linkid;
	    g = &mi -> mi_linkedreplyinv.args.lr_getres;
	    printf("Linked Reply Get - id %d, ", id);
	    break;
        }

#ifdef REQUEST_ID_CHECK
	if (id != request_id) {
	    /* should not happen for a correctly behaving agent*/
	    printf("Get result with a wrong request id received!\n");
	    mi_free(mi);
	    return DONE;
	}
#endif

	++nobjects;

        if (g->gr_class.mid_global)
            printf("class %s, ", oid2name(g->gr_class.mid_global, OIDPART));
        if (g->gr_inst.mn_dn) {
	    inst = dn2str(g->gr_inst.mn_dn);
	    printf("instance {%s}", inst);
	    free(inst);
        }
        if (g->gr_time)
            printf(", time %s\n", g->gr_time);
        else
	    printf("\n");
        for (i = 0; i < g->gr_nattrs; i++)
	    if (!g->gr_attrs[i].ga_error) {
		ava2str(g->gr_attrs[i].ga_ava.mp_id.mid_global,
			g->gr_attrs[i].ga_ava.mp_val,
			&attrid, &attrval);

		free(attrid); free(attrval);
	    }
	    else {
		fprintf(stdout, "%s: Error: %s\n",
			oid2name(g->gr_attrs[i].ga_ava.mp_id.mid_global,
				OIDPART),
			MErrString(g->gr_attrs[i].ga_error));
	    }


    printf("\n%d replies received\n", nobjects);
    mi_free(mi);
    return (DONE);
}


/*
 * print_set_ind - print the indication from a set request
 *
 */

int print_set_ind(mi)
    MSAPIndication *mi;
{

    static int nobjects = 0;
    int i, id;
    struct CMISSetRes * s;
    char *inst, *attrid, *attrval;


        switch (mi->mi_type) {
        case MI_SET_RES:
            s = &mi -> mi_setres.result;
            id = mi -> mi_setres.id;
            printf("Set Result - id %d,\n ", id);
            break;

        case MI_LINKED_REPLY:
            id = mi -> mi_linkedreplyinv.args.lr_linkid;
            s = &mi -> mi_linkedreplyinv.args.lr_setres;
            printf("Linked Reply Set - id %d, ", id);
            break;
        }

#ifdef REQUEST_ID_CHECK
        if (id != request_id) {
            /* should not happen for a correctly behaving agent */
            printf("Set result with a wrong request id received!\n");
            mi_free(mi);
            return DONE;
        }
#endif

        ++nobjects;

        if (s->sr_class.mid_global)
            printf("class %s, ", oid2name(s->sr_class.mid_global, OIDPART));
        if (s->sr_inst.mn_dn) {
            inst = dn2str(s->sr_inst.mn_dn);
            printf("instance {%s}", inst);
            free(inst);
        }
        if (s->sr_time)
            printf(", time %s\n", s->sr_time);
        else
            printf("\n");
        for (i = 0; i < s->sr_nattrs; i++)
	    if (!s->sr_attrs[i].sa_error){
            	ava2str(s->sr_attrs[i].sa_ava.mp_id.mid_global,
                        s->sr_attrs[i].sa_ava.mp_val, 
			&attrid, &attrval);

                free(attrid); free(attrval);
	    }
	    else{
                fprintf(stdout, "Error: %s\n",
                                MErrString(s->sr_attrs[i].sa_error));
                fprintf(stdout, "%s: Error: %s\n",
                        oid2name(s->sr_attrs[i].sa_ava.mp_id.mid_global,
                                OIDPART),
                        MErrString(s->sr_attrs[i].sa_error));
	    }

    printf("\n%d replies received\n", nobjects);
    mi_free(mi);
    return (DONE);

}

/*
 * print_create_ind - print the indication from a create request
 *
 */

int print_create_ind(mi)
MSAPIndication *mi;
{
    static int nobjects = 0;
    int i;
    CMISCreateRes* cr;
    char *inst, *attrid, *attrval;

        cr = &mi -> mi_createres.result;
        printf("Create Result - id %d, ", mi -> mi_createres.id);

#ifdef REQUEST_ID_CHECK
        if (mi->mi_createres.id != request_id) {
            /* should not happen for a correctly behaving agent */
            printf("Create result with wrong request id received!\n");
            mi_free(mi);
            return DONE;
        }
#endif

        ++nobjects;

        if (cr->cr_class.mid_global)
            printf("class %s ", oid2name(cr->cr_class.mid_global, OIDPART));

        if (cr->cr_inst.mn_dn) {
            inst = dn2str(cr->cr_inst.mn_dn);
            printf("instance {%s} ", inst);
            free(inst);
        }

        if (cr->cr_time)
            printf("time %s", cr->cr_time);
        printf("\n");

        for (i = 0; i < cr->cr_nattrs; i++) {
            ava2str(cr->cr_attrs[i].mp_id.mid_global, cr->cr_attrs[i].mp_val,
                        &attrid, &attrval);
            fprintf(stdout, "%s:  %s\n", attrid, attrval);
            free(attrid); free(attrval);
        }

    printf("\n%d replies received\n", nobjects);
    mi_free(mi);
    return (DONE);

}

/*
 * print_delete_ind - prints the indication from an delete request
 *
 */

int print_delete_ind(mi)
MSAPIndication *mi;
{
    int     id;
    MID     objClass;
    MN      objInst;
    char*   time = NULLCP;
    CMISErrors error = m_noError;
    int status = OK;
    static int nsucc = 0, nfail = 0;

    debug_test("In print_delete @ 1");

        switch (mi->mi_type) {
        case MI_DELETE_RES:
	    debug_test("In print_delete @ 2");
            id       = mi->mi_deleteres.id;
            objClass = &mi->mi_deleteres.result.dr_class;
            objInst  = &mi->mi_deleteres.result.dr_inst;
            time     = mi->mi_deleteres.result.dr_time;
            printf("Delete Result - id %d, ", id);
            ++nsucc;
            status = DONE;      /* only one deleted */
            break;

        case MI_LINKED_REPLY:
            debug_test("In print_delete @ 3");
            id = mi->mi_linkedreplyinv.args.lr_linkid;
            switch (mi->mi_linkedreplyinv.args.lr_type) {
            case LR_DELETE_RES:
                objClass = &mi->mi_linkedreplyinv.args.lr_deleteres.dr_class;
                objInst  = &mi->mi_linkedreplyinv.args.lr_deleteres.dr_inst;
                time     = mi->mi_linkedreplyinv.args.lr_deleteres.dr_time;
                ++nsucc;
                break;

            case LR_DELETE_ERR:
                objClass = &mi->mi_linkedreplyinv.args.lr_deleteerr.de_class;
                objInst  = &mi->mi_linkedreplyinv.args.lr_deleteerr.de_inst;
                time     = mi->mi_linkedreplyinv.args.lr_deleteerr.de_time;
                error    = mi->mi_linkedreplyinv.args.lr_deleteerr.de_error;
                ++nfail;
                break;

            case LR_PROC_FAILURE:
                /* this should not really happen for a well debugged agent */
                objClass = &mi->mi_linkedreplyinv.args.lr_procfailure.pf_class;
                objInst  = &mi->mi_linkedreplyinv.args.lr_procfailure.pf_inst;
                error    = m_processingFailure;
                ++nfail;
                break;
            }
            printf("Linked Reply Delete - id %d, ", id);
            break;
        }

/*
        if (id != request_id) {
            printf("Delete result with wrong request id received!\n");
            mi_free(mi);
            return DONE;
        }
 */
        if (objClass->mid_global)
            printf("class %s ", oid2name(objClass->mid_global, OIDPART));

        if (objInst->mn_dn) {
            char* inst = dn2str(objInst->mn_dn);
            printf("instance {%s} ", inst);
            free(inst);
        }

    if (time)
        printf("time %s ", time);

    if (error)
        printf("\nError: %s", MErrString(error));

    printf("\n\n");

    mi_free(mi);
    debug_test("Leaving print_delete");
    return (DONE);


}

/*
 * print_action_ind - print the indication from an action request
 *
 */

int print_action_ind(mi)
    MSAPIndication *mi;
{

    int status = OK;
    static int nsucc = 0, nfail = 0;
    int     id;
    MID     objClass;
    MN      objInst;
    char*   time = NULLCP;
    CMISParam* result = NULLMPARM;
    CMISErrors error = m_noError;

    debug_test("In print_action @ 1");

        switch (mi->mi_type) {
        case MI_ACTION_RES:
            debug_test("In print_action @ 2");
            id       = mi->mi_actionres.id;
            objClass = &mi->mi_actionres.result.ar_class;
            objInst  = &mi->mi_actionres.result.ar_inst;
            time     = mi->mi_actionres.result.ar_time;
            result   = &mi->mi_actionres.result.ar_reply;
            printf("Action Result - id %d, ", id);
            ++nsucc;
            status = DONE;      /* action on only one object */
            break;

        case MI_LINKED_REPLY:
            debug_test("In print_action @ 3");
            id = mi->mi_linkedreplyinv.args.lr_linkid;
            switch (mi->mi_linkedreplyinv.args.lr_type) {
            case LR_ACTION_RES:
                objClass = &mi->mi_linkedreplyinv.args.lr_actionres.ar_class;
                objInst  = &mi->mi_linkedreplyinv.args.lr_actionres.ar_inst;
                time     = mi->mi_linkedreplyinv.args.lr_actionres.ar_time;
                result   = &mi->mi_linkedreplyinv.args.lr_actionres.ar_reply;
                ++nsucc;
                break;

            case LR_ACTION_ERR:
                objClass = &mi->mi_linkedreplyinv.args.lr_actionerr.ae_class;
                objInst  = &mi->mi_linkedreplyinv.args.lr_actionerr.ae_inst;
                time     = mi->mi_linkedreplyinv.args.lr_actionerr.ae_time;
                error    = mi->mi_linkedreplyinv.args.lr_actionerr.ae_error;
                ++nfail;
                break;

            case LR_PROC_FAILURE:
                /* should not really happen */
                objClass = &mi->mi_linkedreplyinv.args.lr_procfailure.pf_class;
                objInst  = &mi->mi_linkedreplyinv.args.lr_procfailure.pf_inst;
                error    = m_processingFailure;
                ++nfail;
                break;
            }
            printf("Linked Reply Action - id %d, ", id);
            break;
        }

        if (objClass->mid_global)
            printf("class %s ", oid2name(objClass->mid_global, OIDPART));

        if (objInst->mn_dn) {
            char* inst = dn2str(objInst->mn_dn);
            printf("instance {%s} ", inst);
            free(inst);
        }

        if (time)
            printf("time %s ", time);

        if (error)
            printf("\nError: %s", MErrString(error));
        else
        if (result->mp_val) {
            char *actionval;
            ava2str(result->mp_id.mid_global, result->mp_val, NULL, &actionval);
            printf("\nResult: %s", actionval);
            free(actionval);
        }

    printf("\n");

    mi_free(mi);
    debug_test("Leaving print_action");
    return (DONE);
}
