/* advise() removed as defined in OSIMIS library also */


/* snmpif.c - SNMP proxy library */

#ifndef	lint
static char *rcsid = "$Header$";
#endif

/*
 * $Header$
 *
 * Originally contributed by NYSERNet Inc.  This work was partially supported
 * by the U.S. Defense Advanced Research Projects Agency and the Rome Air
 * Development Center of the U.S. Air Force Systems Command under contract
 * number F30602-88-C-0016.
 *
 *
 * $Log$
 */

/*
 *			  ISODE 8.0 NOTICE
 *
 *   Acquisition, use, and distribution of this module and related
 *   materials are subject to the restrictions of a license agreement.
 *   Consult the Preface in the User's Manual for the full terms of
 *   this agreement.
 *
 *
 *			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */

#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <varargs.h>
#include <malloc.h>
#include "SNMP-types.h"
#include "objects.h"
#include <sys/ioctl.h>
#include "tailor.h"

#include "dgram.h"
#ifdef	TCP
#include "internet.h"
#endif

/* party.h included within.. */
#include "SnmpRMIBif.h"


#define	TRUE	1
#define	FALSE	0

/*    DATA */

int	clocking = 0;
int	debug = 0;
static	int	lasterr;

static	int	retries = 3;
static	int	timeout = 10;

static	int	verbose = 0;
int	watch = 0;		/* 0 to keep it quiet */

static	char   *myname = "snmpProxy";

static	char  **op = NULLVP;

static	int	ontty;
static	int	armed;
static	jmp_buf	intrenv;
static	int	interrupted;

static	SFP	istat;

static  SFD	intrser ();

static	char   *defs = NULLCP;


static	OID	nullSpecific = NULLOID;
	PE	unSpecified = NULLPE;

static	PS	ps = NULL;

static	int	sd = NOTOK;

static	char   *pParty = "party";
static  struct snmp_party *agentParty = NULL;
static  struct snmp_party *managerParty = NULL;
static  struct snmp_context *mgmtContext = NULL;

static int	object_parse (), object_print ();
static struct snmp_party *find_party ();
static struct snmp_context *find_context ();

char   *snmp_error (), *snmpV2_exception ();
static struct type_SNMP_SnmpAuthMsg *new_message ();

void	adios (), advise ();

/* */

struct dispatch {
    char   *ds_name;		/* command name */
    IFP	    ds_fnx;		/* dispatch */

    char   *ds_help;		/* help string */
};

static struct dispatch *getds ();


static  int	f_audit (), f_bulk (), f_dump (), f_get (), f_get_next (), f_inform (),
    	f_set ();
static  int	f_compile (), f_help (), f_quit ();
static  int	f_adjust_clock (), f_secrets (), f_tune (), f_sync ();
static  int	f_agent (), f_manager (), f_print_all_parties ();
static  int	f_access (), f_context ();
static  int	f_status (), f_retries (), f_timeout (), f_watch ();

static struct dispatch dispatches[] = {
    "access", f_access, "print out access control policy",

    "adjustclock", f_adjust_clock, "adjust clocks",

    "agent", f_agent, "change agent (destination) party",

    "audit", f_audit, "audit traps",

    "bulk", f_bulk, "bulk retrieve columns from a table",

    "compile", f_compile, "write compiled objects file",

    "context", f_context, "change management context",

    "dump", f_dump, "dump a portion of the MIB",

    "get", f_get, "perform get operation",

    "help", f_help, "print help information",

    "inform", f_inform, "perform inform operation",
    
    "manager", f_manager, "change manager (source) party",

    "next", f_get_next, "perform powerful get-next operation",

    "pparties", f_print_all_parties, "print out all parties",

    "quit", f_quit, "terminate program",

    "retries", f_retries, "set the number retries for each operation",

    "secrets", f_secrets, "change secrets",

    "set", f_set, "perform set operation",

    "status", f_status, "report status",

    "sync", f_sync, "synchronize party information to \"stable-storage\"",

    "timeout", f_timeout, "set the number of seconds between retries",

    "tune", f_tune, "tune agent parameters",

    "watch", f_watch, "toggle watching on and off",

    NULL
};


static	int	helpwidth;


#ifndef	SYS5
long	random ();
#endif
long	time ();

/*   MAIN */

/* ARGSUSED */

OLDmain (argc, argv, envp)
int	argc;
char  **argv,
      **envp;
{
    int	    eof,
	    status;
    char    buffer[BUFSIZ],
	   *vec[NVEC + 1];

    arginit (argv);

    status = 0;
    if (op) {
	lasterr = -1;
	if (snmploop (op, NOTOK) == NOTOK)
	    status = 1;
	else
	    if (lasterr >= 0)
		status = lasterr;

	goto were_out_of_here;
    }

    istat = signal (SIGINT, intrser);
    
    eof = 0;
    for (interrupted = 0;; interrupted = 0) {
	if (getline ("%s> ", buffer) == NOTOK) {
	    if (eof)
		break;

	    eof = 1;
	    continue;
	}
	eof = 0;

	bzero ((char *) vec, sizeof vec);
	if (str2vec (buffer, vec) < 1)
	    continue;

	switch (snmploop (vec, OK)) {
	    case NOTOK:
	        status = 1;
		break;

	   case OK:
	   default:
		if (party_changed > 0 || context_changed || access_changed)
		    check_party ();
		continue;

	   case DONE:
		status = 0;
		break;
	}
	break;
    }

    (void) signal (SIGINT, istat);

were_out_of_here: ;
    (void) f_quit (NULLVP);
    (void) fin_party ();
    exit (status);		/* NOTREACHED */
}

/* */

static int  snmploop (vec, error)
char  **vec;
int	error;
{
    register struct dispatch *ds;

    if ((ds = getds (strcmp (*vec, "?") ? *vec : "help")) == NULL)
	return error;
    switch ((*ds -> ds_fnx) (vec)) {
	case NOTOK:
	    return error;

	case OK:
	default:
	    return OK;

	case DONE:
	    return DONE;
    }
}

/* */

static struct dispatch *getds (name)
char   *name;
{
    register int    longest,
                    nmatches;
    register char  *p,
                   *q;
    char    buffer[BUFSIZ];
    register struct dispatch   *ds,
                               *fs;

    longest = nmatches = 0;
    for (ds = dispatches; p = ds -> ds_name; ds++) {
	for (q = name; *q == *p++; q++)
	    if (*q == NULL)
		return ds;

	if (*q == NULL)
	    if (q - name > longest) {
		longest = q - name;
		nmatches = 1;
		fs = ds;
	    }
	    else
		if (q - name == longest)
		    nmatches++;
    }

    switch (nmatches) {
	case 0: 
	    advise (NULLCP, "unknown operation \"%s\"", name);
	    return NULL;

	case 1: 
	    return fs;

	default: 
	    for (ds = dispatches, p = buffer; q = ds -> ds_name; ds++)
		if (strncmp (q, name, longest) == 0) {
		    (void) sprintf (p, "%s \"%s\"", p != buffer ? "," : "", q);
		    p += strlen (p);
		}
	    advise (NULLCP, "ambiguous operation, it could be one of:%s",
			buffer);
	    return NULL;
    }
}

/*   OPERATIONS */

static int  f_audit (vec)
char  **vec;
{
    int	    follow,
	    forever,
	    i;
    long    longtimeago,
	    now;
    char   *cp,
	   *file;
    PE	    pe,
	    p;
    PS	    ps2;
    struct type_SNMP_Audit *au;
    FILE   *fp;

    file = "snmpd.traps";
    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("audit [-f | -N | +N] [file]\n");
	printf ("    audit trap sink\n");
	printf ("        -f:   endless loop\n");
	printf ("        -N:   last N traps\n");
	printf ("        +N:   first N traps\n");
	printf ("        file: trap file (default %s)\n", file);

	return OK;
    }

    follow = forever = 0;
    for (; cp = *vec; vec++)
	switch (*cp) {
	    case '-':
	        if (strcmp (cp, "-f") == 0)
		    forever++, follow = 0;
		else
		    follow = -atoi (++cp), forever = 0;
		break;

	    case '+':
		follow = atoi (++cp), forever = 0;
		break;

	    default:
		file = cp;
		break;
	}

    file = _isodefile (isodelogpath, file);
    if ((fp = fopen (file, "r")) == NULL) {
	advise (file, "unable to read");
	return OK;
    }

    pe = p = NULLPE, au = NULL;
    if ((ps2 = ps_alloc (std_open)) == NULLPS) {
	advise (NULLCP, "ps_alloc(std_open): you lose");
	goto out;
    }

    if (std_setup (ps2, fp) == NOTOK) {
	advise (NULLCP, "std_setup: %s", ps_error (ps2 -> ps_errno));
	goto out;
    }

    (void) time (&now);
    longtimeago = now - 6L * 30L * 24L * 60L * 60L;

    if (follow < 0) {
	register long    offset,
			*lp,
			*ep;
	long   *opp;

	follow = -follow;
	if ((opp = (long *) calloc ((unsigned) follow, sizeof *opp)) == NULL)
	    adios (NULLCP, "out of memory");
	offset = ftell (fp);
	for (ep = (lp = opp) + follow; lp < ep; lp++)
	    *lp = offset;
	lp = opp;

	for (;;) {
	    if ((pe = ps2pe (ps2)) == NULLPE)
		break;
	    if (decode_SNMP_Audit (pe, 1, NULLIP, NULLVP, &au) == NOTOK)
		goto bad_audit;

	    (void) fseek (fp, (long) au -> sizeOfEncodingWhichFollows, 1);

	    free_SNMP_Audit (au);

	    *lp++ = offset;
	    if (lp >= ep)
		lp = opp;
	    offset = ftell (fp);
	}

	(void) fseek (fp, *lp, 0);
	free ((char *) opp);
	follow = 0;
    }

    for (i = 1;; i++) {
	long	mtime;
	UTC	ut;

	if (follow > 0 && i > follow)
	    break;

	if ((pe = ps2pe (ps2)) == NULLPE) {
	    if (ps2 -> ps_errno)
		advise (NULLCP, "ps2pe: %s", ps_error (ps2 -> ps_errno));
	    else
		if (forever) {
		    clearerr (fp);
		    ps2 -> ps_errno = PS_ERR_NONE;
		    sleep (1);
		    continue;
		}
	    break;
	}
	if (decode_SNMP_Audit (pe, 1, NULLIP, NULLVP, &au) == NOTOK) {
bad_audit: ;
	    advise (NULLCP, "decode_SNMP_Audit: %s", PY_pepy);
	    break;
	}

	if ((cp = qb2str (au -> dateAndTime)) == NULL) {
no_mem: ;
	    advise (NULLCP, "qb2str: out of memory");
	    break;
	}
	ut = str2gent (cp, strlen (cp));
	free (cp);
	if (ut == NULL) {
	    advise (NULLCP, "str2gent: you lose");
	    break;
	}
	mtime = gtime (ut2tm (ut));
	cp = ctime (&mtime);
	if (forever)
	    (void) time (&now);
	if (mtime < longtimeago || mtime > now)
	    printf ("%-7.7s %-4.4s ", cp + 4, cp + 20);
	else
	    printf ("%-12.12s ", cp + 4);

	if ((cp = qb2str (au -> source)) == NULL)
	    goto no_mem;
	printf ("%s\n", cp);
	free (cp);

	if ((p = ps2pe (ps2)) == NULLPE) {
	    if (ps2 -> ps_errno)
		advise (NULLCP, "ps2pe: %s", ps_error (ps2 -> ps_errno));
	    break;
	}
	if (print_SNMP_SnmpAuthMsg (p, 1, NULLIP, NULLVP, NULLCP) == NOTOK)
	    printf ("\n");
	pe_free (p), p = NULL;

	free_SNMP_Audit (au), au = NULL;
    }

out: ;
    (void) fclose (fp);
    if (ps2)
	ps_free (ps2);
    if (pe)
	pe_free (pe);
    if (p)
	pe_free (p);
    if (au)
	free_SNMP_Audit (au);

    return OK;
}

/* */

static	int	max_repetitions = 0;
static	int	non_repeaters = 0;

int	bulk1 (), bulk2 (), bulk3 ();


static int  f_bulk (vec)
char  **vec;
{
    int	    mr = 1,
	    nr = -1,
	    rr = 0,
	    result,
	    silent = 0;
    IFP	    fnx = bulk1;
    register struct type_SNMP_VarBindList **vp;
    struct type_SNMP_VarBindList *vb;
    OT	    et = NULL;

    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("bulk [-alg1 | -alg2 | -alg3 | non-repeaters [max-repetitions]] columns...\n");
	printf ("    with non-repeaters, bulk retrieves objects\n");
	printf ("    otherwise, bulk retrieves columns from a table\n");

	return OK;
    }

    if (agentParty == NULL) {
	printf ("please use \"agent\" to set the destination party\n");
	return OK;
    }
    if (managerParty == NULL) {
	printf ("please use \"manager\" to set the source party\n");
	return OK;
    }
    if (agentParty -> party_tDomain != managerParty -> party_tDomain) {
	printf ("please use a source and destination that uses the same transport domain\n");
	return OK;
    }
    if (mgmtContext == NULL) {
	printf ("please use \"context\" to set the context\n");
	return OK;
    }

    while (*vec) {
	if (!rr && isdigit (**vec)) {
	    if (nr < 0)
		nr = atoi (*vec);
	    else
		mr = atoi (*vec), rr = 1;
	    vec++;
	    continue;
	}
	if (strcmp (*vec, "-alg1") == 0) {
	    fnx = bulk1;
	    vec++;
	    continue;
	}
	if (strcmp (*vec, "-alg2") == 0) {
	    fnx = bulk2;
	    vec++;
	    continue;
	}
	if (strcmp (*vec, "-alg3") == 0) {
	    fnx = bulk3;
	    vec++;
	    continue;
	}
	if (strcmp (*vec, "-silent") == 0) {
	    silent = 1;
	    vec++;
	    continue;
	}
	break;
    }
    if (*vec == NULL)
	return OK;

    if (nr >= 0) {
	struct type_SNMP_SnmpAuthMsg *authmsg;

	max_repetitions = mr, non_repeaters = nr;
	authmsg = new_message (type_SNMP_PDUs_get__bulk__request, --vec);
	max_repetitions = non_repeaters = 0;

	(void) process (authmsg, 0 );
	return OK;
    }

    vp = &vb, vb = NULL;
    for (result = NOTOK; *vec; vec++) {
	register struct type_SNMP_VarBindList *bind;
	register struct type_SNMP_VarBind *v;
	OT	ot;

	if ((ot = text2obj (*vec)) == NULL) {
	    advise (NULLCP, "unknown object \"%s\"", *vec);
	    goto out;
	}
	if (!ot -> ot_syntax) {
	    advise (NULLCP, "%s is not a column object", ot -> ot_text);
	    goto out;
	}

	if (et) {
	    register OID    eid = et -> ot_name,
			    oid = ot -> ot_name;

	    if (eid -> oid_nelem != oid -> oid_nelem - 1
		    || bcmp ((char *) eid -> oid_elements,
			     (char *) oid -> oid_elements,
			     eid -> oid_nelem
			         * sizeof *eid -> oid_elements) != 0) {
		advise (NULLCP, "%s not in same table as previous arguments",
			ot -> ot_text);
		goto out;
	    }
	}
	else {
	    ot -> ot_name -> oid_nelem--;
	    et = name2obj (ot -> ot_name);
	    ot -> ot_name -> oid_nelem++;

	    if (et == NULL) {
		advise (NULLCP, "unable to find row object for %s",
			ot -> ot_text);
		goto out;
	    }
	    if (et -> ot_syntax) {
		advise (NULLCP, "%s is not a column object", ot -> ot_text);
		goto out;
	    }
	}

	if ((bind = (struct type_SNMP_VarBindList *) calloc (1, sizeof *bind))
	    	    == NULL)
	    adios (NULLCP, "out of memory");
	*vp = bind, vp = &bind -> next;

	if ((v = (struct type_SNMP_VarBind *) calloc (1, sizeof *v)) == NULL)
	    adios (NULLCP, "out of memory");
	bind -> VarBind = v;

	if ((v -> name = oid_cpy (ot -> ot_name)) == NULLOID)
	    adios (NULLCP, "out of memory");
	(v -> value = unSpecified) -> pe_refcnt++;
    }

    (*fnx) (ps, sd, vb, managerParty, agentParty, mgmtContext, silent);
    result = OK;

out: ;
    free_SNMP_VarBindList (vb);
    return result;
}

/* */

static int  f_dump (vec)
char  **vec;
{
    int	    bulking = 0,
	    found,
	    gotone,
	    request_id,
	    vars,
	    timing = 0,
	    tries;
    char   *nvec[3];
    OID	    oid = NULLOID;
    PE	    pe = NULLPE;
    struct qbuf *qb = NULL;
    struct type_SNMP_SnmpAuthMsg *authmsg = NULL,
				 *newmsg = NULL;
    struct snmp_party *srcParty,
		      *dstParty;
    register struct type_SNMP_PDU *parm;
    register struct type_SNMP_VarBindList *vp;
    struct timeval  tvs,
		    now;

    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("dump [-bulk] [object]\n");
	printf ("    with no arguments, dump entire MIB\n");
	printf ("    with an argument, dump a portion of the MIB\n");

	return OK;
    }

    if (agentParty == NULL) {
	printf ("please use \"dst\" to set the destination party\n");
	return OK;
    }
    if (managerParty == NULL) {
	printf ("please use \"src\" to set the source party\n");
	return OK;
    }
    if (agentParty -> party_tDomain != managerParty -> party_tDomain) {
	printf ("please use a source and destination that uses the same transport domain\n");
	return OK;
    }
    if (mgmtContext == NULL) {
	printf ("please use \"context\" to set the context\n");
	return OK;
    }

    if (*vec && strcmp (*vec, "-bulk") == 0) {
	bulking++;
	max_repetitions = 1000, non_repeaters = 0;
	vec++;
    }

    if (*vec && strcmp (*vec, "-time") == 0) {
	timing++;
	(void) gettimeofday (&tvs, (struct timezone *) 0);
	vec++;
    }

    nvec[0] = "dump";
    nvec[1] = *vec ? *vec : "0.0";
    nvec[2] = NULL;

    authmsg = new_message (bulking ? type_SNMP_PDUs_get__bulk__request
			           : type_SNMP_PDUs_get__next__request, nvec);
    if (bulking)
	max_repetitions = non_repeaters = 0;

    if (authmsg == NULL)
	return OK;

    request_id = authmsg -> authData -> pdu -> un.response -> request__id
									= 0;

    if (*vec) {
	if ((oid = oid_cpy (authmsg -> authData -> pdu
					        -> un.get__next__request
						-> variable__bindings
						-> VarBind
						-> name)) == NULLOID)
	    adios (NULLCP, "out of memory");
    }
    else
	oid = NULLOID;
    vars = 0;

    gotone = 0, tries = retries;
again: ;

    if (agentParty -> party_tDomain == RFC1157DOMAIN) {
	if (send_v1_message (ps, (struct type_SNMP_Snmp1157Msg *) authmsg, &pe)
	        == NOTOK )
	    goto out;
    }
    else
	if (send_v2_message ( ps, authmsg, managerParty, agentParty,
			   mgmtContext, &pe) == NOTOK)

try_again: ;
    if (pe)
	pe_free (pe), pe = NULL;
    if (qb)
	qb_free (qb), qb = NULL;

    {
	fd_set	rfds;

	FD_ZERO (&rfds);
	FD_SET (sd, &rfds);

	switch (xselect (sd + 1, &rfds, NULLFD, NULLFD, timeout)) {
	    case NOTOK:
	        advise ("failed", "xselect");
		goto out;

	    default:
		if (FD_ISSET (sd, &rfds))
		    break;
		/* else fall... */
	    case OK:
		if (watch)
		    fprintf (stderr, "timeout...\n");
		if (tries-- > 0)
		    goto again;
		advise (NULLCP,
			"no %sresponse within %d retries of %s%d second%s each",
			gotone ? "acceptable " : "", retries + gotone,
			gotone ? "upto " : "", timeout,
			timeout != 1 ? "s" : "");
		goto out;
	}
    }

    if (agentParty -> party_tDomain == RFC1157DOMAIN) {

	struct type_SNMP_Snmp1157Msg *msg;

	if (recv_v1_message ( ps, (struct type_SNMP_Snmp1157Msg **) &newmsg, &pe)
	        == NOTOK)
	    goto out;
	gotone++;

	if ((msg = (struct type_SNMP_Snmp1157Msg *) newmsg) -> data -> offset
	        != type_SNMP_PDUs_response) {
	    if (watch)
		fprintf (stderr, "unexpected message type %d\n",
			 msg -> data -> offset);

	    goto out ;
	 }
    }
    else {

	if (recv_v2_message ( &newmsg, &srcParty, &dstParty, &pe, &qb)
	        == NOTOK)
	    goto out;
	gotone++;

	if (newmsg -> authData -> pdu -> offset != type_SNMP_PDUs_response) {
	    if (watch)
		fprintf (stderr, "unexpected message type %d\n",
			 newmsg -> authData -> pdu -> offset);
	    goto out ;
	}
    }

    if ((parm = newmsg -> authData -> pdu -> un.response) -> request__id
	    != request_id) {
	fprintf (stderr, "request-id mismatch (got %d, wanted %d)\n",
		 parm -> request__id, request_id);

	goto try_again;
    }
	
    if (parm -> error__status != int_SNMP_error__status_noError) {
	if (parm -> error__status != int_SNMP_error__status_noSuchName)
	    fprintf (stderr, "%s at position %d\n",
		     snmp_error (parm -> error__status), parm -> error__index);
	goto out;
    }

    found = 0;
    for (vp = parm -> variable__bindings; vp; vp = vp -> next) {
	caddr_t	 value;
	register OI	oi;
	register OS	os;
	register struct type_SNMP_VarBind *v = vp -> VarBind;

	vars++;
	if (oid
	        && (oid -> oid_nelem > v -> name -> oid_nelem
		        || bcmp ((char *) oid -> oid_elements,
				 (char *) v -> name -> oid_elements,
				 oid -> oid_nelem
				 	* sizeof *oid -> oid_elements)))
	    goto out;

	if (timing)
	    continue;

	if (v -> value -> pe_class == PE_CLASS_CONT) {
	    if (v -> value -> pe_id != 2) {
		printf ("%s=", oid2ode (vp -> VarBind -> name));
		printf ("%s\n", snmpV2_exception (v -> value -> pe_id));
	    }
	    continue;
	}
	printf ("%s=", oid2ode (vp -> VarBind -> name));
	if ((oi = name2inst (v -> name)) == NULL
	        || (os = oi -> oi_type -> ot_syntax) == NULL)
	    vunknown (v -> value);
	else
	    if ((*os -> os_decode) (&value, v -> value) == NOTOK)
		vunknown (v -> value);
	    else {
		(*os -> os_print) (value, os);
		printf ("\n");

		(*os -> os_free) (value);
	}
	found++;
    }
    if (found == 0)
	goto out;

    free_SNMP_SnmpAuthMsg (authmsg);
    authmsg = newmsg, newmsg = NULL;
    if (bulking) {
	register struct type_SNMP_VarBindList  *vb,
					      **vpp;

	authmsg -> authData -> pdu -> offset =
					    type_SNMP_PDUs_get__bulk__request;
	parm -> error__status = 0, parm -> error__index = 1000;

	vp = NULL;
	for (vpp = &parm -> variable__bindings;
	         vb = *vpp;
	         vp = vb -> next, vpp = &vb -> next)
	    if  (!vb -> next)
		break;
	if (*vpp == NULL)
	    goto out;
	*vpp = NULL;
	free_SNMP_VarBindList (parm -> variable__bindings);
	parm -> variable__bindings = vp;

	bulking++;
    }
    else
	authmsg -> authData -> pdu -> offset =
					    type_SNMP_PDUs_get__next__request;
    request_id = ++parm -> request__id;
    goto again;

out: ;
    if (oid)
	oid_free (oid);
    if (pe)
	pe_free (pe);
    if (qb)
	qb_free (qb);
    if (authmsg)
	free_SNMP_SnmpAuthMsg (authmsg);
    if (newmsg)
	free_SNMP_SnmpAuthMsg (newmsg);

    if (timing) {
	(void) gettimeofday (&now, (struct timezone *) 0);
	now.tv_sec -= tvs.tv_sec;
	if ((now.tv_usec -= tvs.tv_usec) < 0)
	    now.tv_sec--, now.tv_usec += 1000000;
	advise (NULLCP,
		"%d variable%s retrieved in %d.%06d seconds",
		vars, vars != 1 ? "s" : "", now.tv_sec, now.tv_usec);
    }
    if (bulking)
	advise (NULLCP, "%d bulk-interactions retrieving %d variable%s",
		bulking, vars, vars != 1 ? "s" : "");

    return OK;
}




/* */

static int  f_get (vec)
char  **vec;
{
    (void) process (new_message (type_SNMP_PDUs_get__request, vec), 1 );
    return OK;
}


static int  f_get_next (vec)
char  **vec;
{
    (void) process (new_message (type_SNMP_PDUs_get__next__request, vec), 1 );
    return OK;
}


static int  f_inform (vec)
char  **vec;
{
    (void) process (new_message (type_SNMP_PDUs_inform__request, vec), 0 );
    return OK;
}


static int  f_set (vec)
char  **vec;
{
    (void) process (new_message (type_SNMP_PDUs_set__request, vec), 1 );
    return OK;
}

/* */

static char *errors[] = {
    "noError",
    "tooBig",
    "noSuchName",
    "badValue",
    "readOnly",
    "genErr",
    "noAccess",
    "wrongType",
    "wrongLength",
    "wrongEncoding",
    "wrongValue",
    "noCreation",
    "inconsistentValue",
    "resourceUnavailable",
    "commitFailed",
    "undoFailed",
    "authorizationError",
    "notWritable",
    "inconsistentName"
};

char   *snmp_error (i)
integer	i;
{
    static char buffer[BUFSIZ];

    if (0 <= i && i < sizeof errors / sizeof errors[0])
	return errors[i];
    (void) sprintf (buffer, "error-status %d", i);

    return buffer;
}


static char *exceptions[] = {
    "noSuchObject",
    "noSuchInstance",
    "endOfMibView"
};


char   *snmpV2_exception (i)
PElementID	i;
{
    static char buffer[BUFSIZ];

    if (i < sizeof exceptions / sizeof exceptions[0])
	return exceptions[i];
    (void) sprintf (buffer, "exception %u", (unsigned int) i);

    return buffer;
}

/* */

static struct type_SNMP_SnmpAuthMsg *new_message (offset, vec)
int	offset;
char  **vec;
{
    struct type_SNMP_SnmpAuthMsg *authmsg;
    struct type_SNMP_AuthInformation *authinfo;
    struct type_SNMP_SnmpMgmtCom *mgmtmsg;
    struct type_SNMP_PDUs *pdu;
    struct type_SNMP_PDU *parm;
    struct type_SNMP_VarBindList **vp;

    if (agentParty == NULL) {
	printf ("please use \"dst\" to set the destination party\n");
	return NULL;
    }
    if (managerParty == NULL) {
	printf ("please use \"src\" to set the source party\n");
	return NULL;
    }
    if (agentParty -> party_tDomain != managerParty -> party_tDomain) {
	printf ("please use a source and destination that uses the same transport domain\n");
	return OK;
    }
    if (mgmtContext == NULL) {
	printf ("please use \"context\" to set the context\n");
	return OK;
    }

/*  authmsg = NULL;*/

    if (agentParty -> party_tDomain == RFC1157DOMAIN) {

	struct type_SNMP_Snmp1157Msg *msg;

	switch (offset) {
	    case type_SNMP_PDUs_get__bulk__request:
	        offset = type_SNMP_PDUs_get__next__request;
		break;

	    case type_SNMP_PDUs_inform__request:
		advise( NULLCP, "context held by SNMPv1 agent" ) ;

		return NULL;

	    default:
		break;
	}

	if ((msg = (struct type_SNMP_Snmp1157Msg *) calloc (1, sizeof *msg))
	        == NULL)
	    goto no_mem;
	    
	msg -> version = int_SNMP_version_version__1;

	if ((msg -> community = str2qb ((char *) agentParty -> party_AuthPrivate,
				    agentParty -> party_AuthPrivate_len, 1))
	        == NULL) {

	    free ((char *) msg);
	    goto no_mem;
	}
	
	if ((pdu = (struct type_SNMP_PDUs *) calloc (1, sizeof *pdu)) == NULL)
	    goto no_mem;

	msg -> data = pdu;
	pdu -> offset = offset;

	if ((parm = (struct type_SNMP_PDU *) calloc (1, sizeof *parm)) == NULL)
	    goto no_mem;

	authmsg = (struct type_SNMP_SnmpAuthMsg *) msg ;
    }
    else {

    	if ((authmsg = (struct type_SNMP_SnmpAuthMsg *)
					calloc (1, sizeof *authmsg)) == NULL) {
no_mem: ;
	    advise (NULLCP, "out of memory");
	    if (authmsg)
	    	free_SNMP_SnmpAuthMsg (authmsg);
	    return NULL;
    	}

    	if ((authinfo = (struct type_SNMP_AuthInformation *)
					calloc (1, sizeof *authinfo)) == NULL)
	    goto no_mem;

    	authmsg -> authInfo = authinfo;   

    	if ((mgmtmsg = (struct type_SNMP_SnmpMgmtCom *) 
					calloc (1, sizeof *mgmtmsg)) == NULL)
	    goto no_mem;

   	authmsg -> authData = mgmtmsg;

    	if ((pdu = (struct type_SNMP_PDUs *) calloc (1, sizeof *pdu)) == NULL)
	    goto no_mem;

    	mgmtmsg -> pdu = pdu;

/* for now, always a PDU... */
    	pdu -> offset = offset;

	if ((parm = (struct type_SNMP_PDU *) calloc (1, sizeof *parm)) == NULL)
	    goto no_mem;

	if ((pdu -> offset = offset) == type_SNMP_PDUs_get__bulk__request) {
	    parm -> error__status = non_repeaters;
	    parm -> error__index = max_repetitions;
	}
    }

    pdu -> un.get__request = parm;

#ifndef	SYS5
    parm -> request__id = ((int) random ()) & 0x7fffffff;
#else
    parm -> request__id = ((int) rand ()) & 0x7fffffff;
#endif

    vp = &parm -> variable__bindings;
    for (vec++; *vec; vec++) {
	register struct type_SNMP_VarBindList *bind;
	register struct type_SNMP_VarBind *v;

	if ((bind = (struct type_SNMP_VarBindList *) calloc (1, sizeof *bind))
	    	    == NULL)
	    adios (NULLCP, "out of memory");
	*vp = bind, vp = &bind -> next;

	if ((v = (struct type_SNMP_VarBind *) calloc (1, sizeof *v)) == NULL)
	    adios (NULLCP, "out of memory");
	bind -> VarBind = v;

	if (get_ava (v, *vec, offset) == NOTOK) {
	    free_SNMP_SnmpAuthMsg (authmsg);
	    return NULL;
	}
    }

    return authmsg;
}

/* */

static int  get_ava (v, ava, offset)
register struct type_SNMP_VarBind *v;
char   *ava;
int	offset;
{
    register char *cp;

    if (cp = index (ava, '=')) {
	if (offset != type_SNMP_PDUs_set__request
	        && offset != type_SNMP_PDUs_inform__request)
	    advise (NULLCP, "value unnecessary for operation");
	*cp++ = NULL;
    }
    else
	if (offset == type_SNMP_PDUs_set__request
	        || offset == type_SNMP_PDUs_inform__request) {
	    advise (NULLCP, "need variable=value for operation");
	    return NOTOK;
	}

    if (object_parse (&v -> name, ava, NULLOS) == NOTOK) {
unknown: ;
	advise (NULLCP, "unknown variable \"%s\"", ava);
	return NOTOK;
    }
    if (v -> name -> oid_nelem < 2) {
	advise (NULLCP, "OIDs require at least two sub-identifiers");
	return NOTOK;
    }

    if (cp == NULL)
	(v -> value = unSpecified) -> pe_refcnt++;
    else {
	int	result;
	register OS    os;
	register OT    ot;
	caddr_t value;

	if ((ot = name2obj (v -> name)) == NULLOT) {
	    oid_free (v -> name), v -> name = NULL;
	    goto unknown;
	}

	if ((os = ot -> ot_syntax) == NULL) {
	    advise (NULLCP, "no syntax defined for object \"%s\"", ava);
	    return NOTOK;
	}

	if ((*os -> os_parse) (&value, cp, os) == NOTOK) {
	    advise (NULLCP, "invalid value for variable \"%s\": \"%s\"",
		    ava, cp);
	    return NOTOK;
	}
	result = (*os -> os_encode) (value, &v -> value);
	(*os -> os_free) (value);

	if (result == NOTOK) {
	    advise (NULLCP, "encoding error for variable \"%s\"", ava);
	    return NOTOK;
	}	
    }

    return OK;
}


/***********************************************************************************/

/*
 * Depth first SnmpRMIBtxPduNode linked list traversal..
 * Note: fnToBeApplied have common parameters see :-
 *
 * 	SnmpRMIBcheckForPendingTxPdu()
 *	SnmpRMIBremoveTxPduNode()
 *	SnmpRMIBcheckForOperIdInTxPduNode().
 */
void
SnmpRMIBrecursiveTxPduNodeOperation( currTxPduNodep, operOrRequestId, fnCompletedp,
						operationIdp, thisTxPduNodepp, fnToBeApplied )
struct	SnmpRMIBtxPduNode * currTxPduNodep ;		  /* Parameters supplied	  */
int	operOrRequestId, * fnCompletedp, * operationIdp ; /* depend on the fnToBeApplied()*/
struct	SnmpRMIBtxPduNode ** thisTxPduNodepp ;		  /*				  */
int	(*fnToBeApplied)() ;
{
struct SnmpRMIBtxPduNode * nextSnmpRMIBsubordTxPduNodep,
			 * nextSnmpRMIBpeerTxPduNodep   ;

	if( *fnCompletedp == TRUE )
	    return ;

	nextSnmpRMIBsubordTxPduNodep = currTxPduNodep -> nextSnmpRMIBsubordTxPduNodep ;
	nextSnmpRMIBpeerTxPduNodep   = currTxPduNodep -> nextSnmpRMIBpeerTxPduNodep ;

	*fnCompletedp = fnToBeApplied( currTxPduNodep, operOrRequestId, operationIdp, thisTxPduNodepp ) ;
	/* ...Note: for delete node operations the "currTxPduNodep" node now no longer exists */

	if( *fnCompletedp == FALSE ) {

	    if( nextSnmpRMIBsubordTxPduNodep )
		SnmpRMIBrecursiveTxPduNodeOperation( nextSnmpRMIBsubordTxPduNodep,
	    				operOrRequestId, fnCompletedp, operationIdp, thisTxPduNodepp, fnToBeApplied ) ;
	    
	    if( *fnCompletedp == FALSE && nextSnmpRMIBpeerTxPduNodep )
		SnmpRMIBrecursiveTxPduNodeOperation( nextSnmpRMIBpeerTxPduNodep,
	    				operOrRequestId, fnCompletedp, operationIdp, thisTxPduNodepp, fnToBeApplied ) ;
	}
	
	return ;
}

/*
 *
 */
int
SnmpRMIBcheckForOperIdInTxPduNode( currTxPduNodep, requestId, operationIdp, thisTxPduNodepp )
struct	SnmpRMIBtxPduNode * currTxPduNodep ;
int 	requestId, * operationIdp ;
struct  SnmpRMIBtxPduNode ** thisTxPduNodepp ;
{
	/* Must ensure that we are only checking the requestId for the lowest level nodes */
	if( currTxPduNodep -> nextSnmpRMIBsubordTxPduNodep == NULLSNMPRMIBTXPDUNODE &&
	    currTxPduNodep -> requestId == requestId ) {

	    *thisTxPduNodepp = currTxPduNodep ;
	    *operationIdp = currTxPduNodep -> superiorSnmpRMIBoperNodep -> operationId ;

	    return TRUE ; /* Operation complete */
	}
	
	return FALSE ;
}

/*
 * Search the transmitted PDU nodes for the given requestId so
 * that the corresponding operationId can be returned.
 *
 * Result parameter(s) : operationIdp, thisOperNodepp.
 *
 * Error returns : SnmpRMIBnoError,
 *		   SnmpRMIBunknownRequestId
 */

SnmpRMIBerrorEnum
SnmpRMIBgetOperationIdforRequestId( snmpRMIBAgentDatap, requestId, operationIdp, thisTxPduNodepp )
struct  SnmpRMIBsnmpifData    * snmpRMIBAgentDatap ;
int	requestId, * operationIdp ;
struct	SnmpRMIBtxPduNode ** thisTxPduNodepp ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
struct  SnmpRMIBoperationNode * currOperNodep  = NULLSNMPRMIBOPERATIONNODE ;
int	fnCompleted = FALSE ;

	*operationIdp = -1 ;
	
	for( currOperNodep = snmpRMIBAgentDatap->snmpRMIBoperationNodep ;
	     *operationIdp == -1 && currOperNodep != NULLSNMPRMIBOPERATIONNODE ;
	     currOperNodep = currOperNodep->nextSnmpRMIBoperationNodep ) {

	    SnmpRMIBrecursiveTxPduNodeOperation( currOperNodep->snmpRMIBtxPduNodep, requestId,
  		&fnCompleted, operationIdp, thisTxPduNodepp, SnmpRMIBcheckForOperIdInTxPduNode ) ;
	}

	if( *operationIdp == -1 )
	    retVal = SnmpRMIBunknownRequestId ;

	return retVal ;
}


/*
 * The results node might just indicate errors - so we need
 * a check if result OIDs/PEs are available
 */
int
SnmpRMIBnumResultNodeBindings( rxResultsNodep )
struct	SnmpRMIBrxedResults * rxResultsNodep ;
{
int	numBindings = 0 ;

	if( rxResultsNodep != NULL && rxResultsNodep->resOIDvec != NULL && rxResultsNodep->resOIDvec[0] != NULLOID )
	    numBindings = rxResultsNodep->numVarBindings ;

	return numBindings ;
}

/*
 * Unlink and free the linled list of SnmpRMIBrxedResults structures.
 *
 * Input parameter(s) :
 *
 * Result parameter : NONE
 *
 * Error returns : NONE.
 */

void
SnmpRMIBremoveReceivedResultsNode( rxResultsNodepp )
struct	SnmpRMIBrxedResults ** rxResultsNodepp ;
{
int	i ;
struct	SnmpRMIBrxedResults *  rxResultsp ;

	if( rxResultsNodepp &&
	    (rxResultsp = *rxResultsNodepp) != NULLSNMPRMIBRXEDRESULTS ) {

	    /* Free the OIDs and PEs.. */
	    for( i=0 ; i < rxResultsp->numVarBindings ; i++ ) {
		if( rxResultsp->resOIDvec[i] != NULLOID )
	    	    oid_free( rxResultsp->resOIDvec[i] ) ;

		if( rxResultsp->resPEvec[i] != NULLPE )
		    pe_free( rxResultsp->resPEvec[i] ) ;
	    }

	    /* Free the OID and PE vector storage.. */
	    if( rxResultsp->resOIDvec != NULL ) {
	    	(void)free( (char *)rxResultsp->resOIDvec ) ;
		rxResultsp->resOIDvec = NULL ;
	    }
	    
	    if( rxResultsp->resPEvec != NULL ) {
	    	(void)free( (char *)rxResultsp->resPEvec  ) ;
		rxResultsp->resPEvec = NULL ;
	    }
	    
	    if( rxResultsp->resAttrIds != NULL ) {
	    	(void)free( (char *)rxResultsp->resAttrIds ) ;
		rxResultsp->resAttrIds = NULL ;
	    }

	    if( rxResultsp->OIDspecificSNMPerrorp != NULL ) {
	    	(void)free( (char *)rxResultsp->OIDspecificSNMPerrorp ) ;
		rxResultsp->OIDspecificSNMPerrorp = NULL ;
	    }

	    if( rxResultsp->operAttrIds != NULL ) {
	    	(void)free( (char *)rxResultsp->operAttrIds ) ;
		rxResultsp->operAttrIds = NULL ;
	    }
	    
	    /* Unlink the node.. */
	    *rxResultsNodepp = rxResultsp->nextSnmpRMIBrxedResultsNodep ;

	    /* Free the node storage.. */
	    (void)free( (char *)rxResultsp ) ;
	}

	return ;
}

void
SnmpRMIBremoveReceivedResultsList( rxResultsNodepp )
struct	SnmpRMIBrxedResults ** rxResultsNodepp ;
{
	while( rxResultsNodepp && *rxResultsNodepp ) {
	    SnmpRMIBremoveReceivedResultsNode( rxResultsNodepp ) ;
	}

	return ;
}

SnmpRMIBerrorEnum
SnmpRMIBcreateOrUpdateRxedResultsNode( thisOperNodep, thisTxPduNodep, varBindListp )
struct  SnmpRMIBoperationNode * thisOperNodep ;
struct  SnmpRMIBtxPduNode     * thisTxPduNodep ;
struct	type_SNMP_VarBindList * varBindListp  ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
int	possNumBindings=0, numOfNewBindings=0, useExistingResNode = FALSE ;
struct	SnmpRMIBrxedResults * rxResultsNodep ;
struct	type_SNMP_VarBindList   * vp ;

	/* Determine whether we already have any partial results for this requestId.. */

	if( (rxResultsNodep = thisOperNodep->snmpRMIBreceivedResultsNodep) == NULLSNMPRMIBRXEDRESULTS ) {
	
	    if( (rxResultsNodep = thisOperNodep->snmpRMIBreceivedResultsNodep =
	    		(struct SnmpRMIBrxedResults *)calloc( 1,
				sizeof( struct SnmpRMIBrxedResults ))) == NULLSNMPRMIBRXEDRESULTS )
		retVal = SnmpRMIBmemAllocFailure ;

	    else {

		rxResultsNodep->snmpOperation = thisOperNodep->snmpOperation ;
	    	rxResultsNodep->nextSnmpRMIBrxedResultsNodep = NULLSNMPRMIBRXEDRESULTS ;
		rxResultsNodep->numOperOids = thisOperNodep->numOperOids ;
	    	rxResultsNodep->OIDspecificSNMPerrorp = (int *)0 ;
		rxResultsNodep->operAttrIds = (int *)0 ;
	    }
					
	} else useExistingResNode = TRUE ;

	if( retVal == SnmpRMIBnoError ) {

	    /* Count the number of new bindings.. */
	    for( possNumBindings = 0, vp = varBindListp ; vp ; vp = vp -> next, possNumBindings++ )
		; /* Empty loop */

	    if( possNumBindings > 0 ) {

	    	/* OID storage.. */

	    	if( useExistingResNode == TRUE ) {

		    /* The existing storage must be extended.. */
		    if( (rxResultsNodep -> resOIDvec = (OID *)realloc((char*)rxResultsNodep -> resOIDvec,
				(size_t)(rxResultsNodep->numVarBindings + possNumBindings) * sizeof( OID *))) == ((OID *) 0) ) {

		    	(void)free( (char *)rxResultsNodep ) ;
		    	retVal = SnmpRMIBmemAllocFailure ;
		    }
		
	    	} else if( (rxResultsNodep -> resOIDvec= (OID *)malloc( possNumBindings * sizeof( OID *))) == ((OID *) 0) ) {

		    (void)free( (char *)rxResultsNodep ) ;
		    retVal = SnmpRMIBmemAllocFailure ;
	    	}
	    }
	}

	if( retVal == SnmpRMIBnoError && possNumBindings > 0 ) {

	    /* PE storage.. */

	    if( useExistingResNode == TRUE ) {

		/* The existing storage must be extended.. */
		if( (rxResultsNodep -> resPEvec = (PE *)realloc((char*)rxResultsNodep -> resPEvec,
		    	(size_t)(rxResultsNodep->numVarBindings + possNumBindings) * sizeof( PE *))) == ((PE *) 0) ) {	
			(void)free( (char *)rxResultsNodep ) ;
			(void)free( (char *)rxResultsNodep -> resOIDvec ) ;
		    retVal = SnmpRMIBmemAllocFailure ;
		}
		
	    } else if( (rxResultsNodep -> resPEvec= (PE *)malloc( possNumBindings * sizeof( PE *))) == ((PE *) 0) ) {

		(void)free( (char *)rxResultsNodep ) ;
		(void)free( (char *)rxResultsNodep -> resOIDvec ) ;
		retVal = SnmpRMIBmemAllocFailure ;
	    }
	}

	if( retVal == SnmpRMIBnoError && possNumBindings > 0 ) {

	    /* attrId storage.. */

	    if( useExistingResNode == TRUE ) {

		/* The existing storage must be extended.. */
		if( (rxResultsNodep -> resAttrIds = (int *)realloc((char*)rxResultsNodep -> resAttrIds,
		    	(size_t)(rxResultsNodep->numVarBindings + possNumBindings) * sizeof( int *))) == ((int *) 0) ) {	
		    (void)free( (char *)rxResultsNodep ) ;
		    (void)free( (char *)rxResultsNodep -> resOIDvec ) ;
		    (void)free( (char *)rxResultsNodep -> resPEvec ) ;
		    retVal = SnmpRMIBmemAllocFailure ;
		}
		
	    } else if( (rxResultsNodep -> resAttrIds= (int *)malloc( possNumBindings * sizeof( int *))) == ((int *) 0) ) {

		(void)free( (char *)rxResultsNodep ) ;
		(void)free( (char *)rxResultsNodep -> resOIDvec ) ;
		(void)free( (char *)rxResultsNodep -> resPEvec ) ;
		retVal = SnmpRMIBmemAllocFailure ;
	    }
	}

	if( retVal == SnmpRMIBnoError && possNumBindings > 0 ) {

	    /* Buffer the result bindings across */

	    for( numOfNewBindings = 0, vp = varBindListp ; vp ; vp = vp -> next ) {

/* REMOVED.. 	caddr_t value ;
		register OI	oi;
		register OS	os; ..REMOVED */
		register struct type_SNMP_VarBind *v = vp -> VarBind;

		if (v -> value -> pe_class == PE_CLASS_CONT) {
		    printf ("Received results error 1 : %s\n", snmpV2_exception (v -> value -> pe_id));
		    continue;
		}
/* REMOVED..	    if ((oi = name2inst (v -> name)) == NULL) {
			advise (NULLCP, "Received results error 2 unknown variable" ) ;
			vunknown (v -> value);
			continue;
		    }
		    if ((os = oi -> oi_type -> ot_syntax) == NULL) {
			advise (NULLCP, "Received results error 3 unknown syntax for object \"%s\"",
			    oi -> oi_type -> ot_text);
			goto no_dice;
		    }
		    if ((*os -> os_decode) (&value, v -> value) == NOTOK) {
			advise (NULLCP, "decode error for variable \"%s\"",
			    oid2ode (v -> name));
			goto no_dice;
		    }
..REMOVED */

		rxResultsNodep -> resOIDvec[ rxResultsNodep -> numVarBindings + numOfNewBindings ] = v->name ;
		rxResultsNodep -> resPEvec [ rxResultsNodep -> numVarBindings + numOfNewBindings ] = v->value ;
		v->name=NULLOID, v->value=NULLPE ; /* to prevent them being freed */

		rxResultsNodep -> resAttrIds[ rxResultsNodep -> numVarBindings + numOfNewBindings ] = 
			thisOperNodep->operAttrIds[ thisTxPduNodep->baseIndexWithinOperationOIDvec+numOfNewBindings  ] ;

		numOfNewBindings++ ;
	    }

	    rxResultsNodep -> numVarBindings += numOfNewBindings ;
	}

	return retVal ;
}


/* 
 * Assumes depth first linked list traversal..
 */
int
SnmpRMIBremoveTxPduNode( removeTxPduNodep, dummyint, dummyintp, dummyTxPduNodepp )
struct	SnmpRMIBtxPduNode * removeTxPduNodep ;
int	dummyint, *dummyintp ;
struct  SnmpRMIBtxPduNode ** dummyTxPduNodepp ;
{
int 	i ;

	if( removeTxPduNodep != NULLSNMPRMIBTXPDUNODE ) {
	    
	    if( removeTxPduNodep -> OIDvec ) {

		for( i=0 ; i<removeTxPduNodep->numOids ; i++ )
		    if( removeTxPduNodep->OIDvec[i] != NULLOID )
		    	oid_free( removeTxPduNodep->OIDvec[i] ) ;

		(void)free( (char *)removeTxPduNodep->OIDvec ) ; /* Free the array itself */
	    }

	    if( removeTxPduNodep -> PEvec ) {

		for( i=0 ; i<removeTxPduNodep->numOids ; i++ )
		    if( removeTxPduNodep->PEvec[i] != NULLPE )
		    	pe_free( removeTxPduNodep->PEvec[i] ) ;

		(void)free( (char *)removeTxPduNodep->PEvec ) ; /* Free the array itself */
	    }

	    if( removeTxPduNodep -> authMsg ) {
		if( removeTxPduNodep -> version == SnmpV1 )
		    free_SNMP_Snmp1157Msg((struct type_SNMP_Snmp1157Msg *)removeTxPduNodep -> authMsg );
		else
		    free_SNMP_SnmpAuthMsg( removeTxPduNodep -> authMsg);
	    }

	    /* Finally free the node storage.. */
	    (void)free( (char *)removeTxPduNodep ) ;
	}

	return FALSE ;
}

/*
 * Unlink and free the SnmpRMIBtxPduNode linked lists.
 *
 * Input parameter(s) : currOperNodep
 *
 * Result parameter : NONE
 *
 * Error returns : NONE.
 */
void
SnmpRMIBremoveTxPduNodeLists( currOperNodep )
struct	SnmpRMIBoperationNode * currOperNodep ;
{
int	fnCompleted = FALSE ;

	if( currOperNodep != NULLSNMPRMIBOPERATIONNODE ) {

	    SnmpRMIBrecursiveTxPduNodeOperation( currOperNodep -> snmpRMIBtxPduNodep, 0, 
	    	&fnCompleted, (int *)0, (struct SnmpRMIBtxPduNode**)0, SnmpRMIBremoveTxPduNode ) ;

	    currOperNodep -> snmpRMIBtxPduNodep = NULLSNMPRMIBTXPDUNODE ;
	}

	return ;
}


/*
 * Unlink and free the SnmpRMIBoperationNode structure.
 * Note : any subordinate SnmpRMIBtxPduNode linked lists will also be freed.
 *
 * Input parameter(s) : nextOperationNodepp - address of the RMIB agents snmpRMIBoperationNodep
 *	 				      or the previous nextSnmpRMIBoperationNodep.
 *
 * Result parameter : NONE
 *
 * Error returns : NONE.
 */
void 
SnmpRMIBremoveOperationNode( nextOperationNodepp )
struct 	SnmpRMIBoperationNode ** nextOperationNodepp ;
{
struct  SnmpRMIBoperationNode *  removeOperationNodep ;

	if( nextOperationNodepp != NULL &&
	    (removeOperationNodep = *nextOperationNodepp) != NULLSNMPRMIBOPERATIONNODE ) {

	    /* Remove any subordinate transmitted PDU nodes.. */
	    SnmpRMIBremoveTxPduNodeLists( removeOperationNodep ) ;

	    /* Remove any results storage */
	    SnmpRMIBremoveReceivedResultsList( &(removeOperationNodep->snmpRMIBreceivedResultsNodep) ) ;

	    if( removeOperationNodep->OIDspecificSNMPerrorp )
		(void)free( (char *)removeOperationNodep->OIDspecificSNMPerrorp ) ;

	    if( removeOperationNodep->operAttrIds )
		(void)free( (char *)removeOperationNodep->operAttrIds ) ;

	    /* Unlink the node.. */
	    *nextOperationNodepp = removeOperationNodep -> nextSnmpRMIBoperationNodep ;

	    /* Finally free the node storage.. */
	    (void)free( (char *)removeOperationNodep ) ;
	}

	return ;
}

void
SnmpRMIBfreeOperationStorage( snmpRMIBAgentDatap, operationId )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
int	operationId ;
{
struct  SnmpRMIBoperationNode * currOperNodep, ** prevOperNodepp ;

	/* Find the operation node that corresponds to this operation ID.. */
	if( (currOperNodep = snmpRMIBAgentDatap->snmpRMIBoperationNodep) != NULLSNMPRMIBOPERATIONNODE ) {

	    prevOperNodepp = &(snmpRMIBAgentDatap->snmpRMIBoperationNodep) ;

	    for(;
	    	currOperNodep  != NULLSNMPRMIBOPERATIONNODE &&
		currOperNodep->operationId != operationId  ;
		prevOperNodepp =  &(currOperNodep->nextSnmpRMIBoperationNodep),
		currOperNodep  =  currOperNodep -> nextSnmpRMIBoperationNodep   )
		; /* Empty loop */
	}

	if( currOperNodep != NULLSNMPRMIBOPERATIONNODE )
	    SnmpRMIBremoveOperationNode( prevOperNodepp ) ;

	
	return ;
}

/*
 * 
 *
 */
SnmpRMIBerrorEnum
SnmpRMIBcreateTxPduNode( superiorSnmpRMIBoperNodep, OIDvec, numOids, PEvec, authMsg, requestId, 
						baseIndexWithinOperationOIDvec, currTxPduNodepp,
 						attemptToRecoverFromPossibleMissingTooBigResponse )
struct	SnmpRMIBoperationNode * superiorSnmpRMIBoperNodep ;
OID  *	OIDvec  ;
int	numOids ;
PE   *	PEvec   ; /* May be NULL */
struct	type_SNMP_SnmpAuthMsg * authMsg ;
int	requestId, baseIndexWithinOperationOIDvec ;
struct	SnmpRMIBtxPduNode ** currTxPduNodepp ;
int	attemptToRecoverFromPossibleMissingTooBigResponse ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
struct 	SnmpRMIBtxPduNode * newTxPduNodep, * currTxPduNodep ;
int 	i,j ;

	if( (newTxPduNodep = (struct SnmpRMIBtxPduNode *)calloc( 1, 
				sizeof( struct SnmpRMIBtxPduNode ))) == NULLSNMPRMIBTXPDUNODE ) {
	    retVal = SnmpRMIBmemAllocFailure ;

	} else if( (newTxPduNodep -> OIDvec = (OID *)calloc( 1,
						sizeof( OID ) * numOids )) == (OID *)0 ) {
	    (void)free((char *)newTxPduNodep ) ;
	    retVal = SnmpRMIBmemAllocFailure ;

	} else if( PEvec != (PE *)0 &&
		   (newTxPduNodep -> PEvec = (PE *)calloc( 1,
		   				sizeof( PE ) * numOids )) == (PE *)0 ) {

	    (void)free((char *)newTxPduNodep -> OIDvec ) ;
	    (void)free((char *)newTxPduNodep ) ;
	    retVal = SnmpRMIBmemAllocFailure ;
							
	} else {

	    newTxPduNodep -> numOids = numOids ;
	    
	    for( i = 0 ; i<numOids ; i++ )
		if( (newTxPduNodep -> OIDvec[ i ] = oid_cpy( OIDvec[ i ] )) == NULLOID ) {
		    retVal = SnmpRMIBmemAllocFailure ;
		    break ;
		}

	    if( retVal != SnmpRMIBnoError ) {

		for( j=0 ; j<i ; j++ )
		    oid_free( newTxPduNodep -> OIDvec[ j ] );

		(void)free((char *)newTxPduNodep -> OIDvec ) ;
	    	(void)free((char *)newTxPduNodep ) ;

	    } else {

		if( PEvec != (PE *)0 ) {

		    for( i = 0 ; i<numOids ; i++ )
			if( (newTxPduNodep -> PEvec[ i ] = pe_cpy( PEvec[ i ] )) == NULLPE ) {
			    retVal = SnmpRMIBmemAllocFailure ;
			    break ;
		    	}

		    if( retVal != SnmpRMIBnoError ) {

			for( j=0 ; j<i ; j++ )
				pe_free( newTxPduNodep -> PEvec[ j ] );

			for( i=0 ; i<numOids ; i++ )
				oid_free( newTxPduNodep -> OIDvec[ i ] );

			(void)free((char *)newTxPduNodep -> OIDvec ) ;
			(void)free((char *)newTxPduNodep -> PEvec ) ;
		    	(void)free((char *)newTxPduNodep ) ;
		    }

		} else newTxPduNodep -> PEvec = (PE *)0 ;

		if( retVal == SnmpRMIBnoError ) {

		    newTxPduNodep->superiorSnmpRMIBoperNodep = superiorSnmpRMIBoperNodep ;
		    newTxPduNodep->version   = superiorSnmpRMIBoperNodep->version ;
		    newTxPduNodep->baseIndexWithinOperationOIDvec = baseIndexWithinOperationOIDvec ;
		    newTxPduNodep->requestId = requestId ;
	    	    newTxPduNodep->reTxCount = 0 ;
 	    	    newTxPduNodep->authMsg   = authMsg ;

	    	    newTxPduNodep->snmpErrorReceived = int_SNMP_error__status_noError ;
		    newTxPduNodep->snmpErrorIndex    = -1 ;

		    newTxPduNodep->attemptToRecoverFromPossibleMissingTooBigResponse = attemptToRecoverFromPossibleMissingTooBigResponse ;
			
		    newTxPduNodep->pduTxComplete = FALSE ;

		    newTxPduNodep->nextSnmpRMIBpeerTxPduNodep   = NULLSNMPRMIBTXPDUNODE ;
		    newTxPduNodep->nextSnmpRMIBsubordTxPduNodep = NULLSNMPRMIBTXPDUNODE ;

		    /* Link in the pdu transmission node structure into the current agent list */
		    if( *currTxPduNodepp == NULLSNMPRMIBTXPDUNODE )
			*currTxPduNodepp =  newTxPduNodep ;
		    else {

			for( currTxPduNodep = *currTxPduNodepp ;
		     	     currTxPduNodep -> nextSnmpRMIBpeerTxPduNodep ;
		     	     currTxPduNodep = currTxPduNodep->nextSnmpRMIBpeerTxPduNodep )
		     	     ; /* Empty loop */

		    	currTxPduNodep -> nextSnmpRMIBpeerTxPduNodep = newTxPduNodep ;
		    }
		}
	    }
	}

	return retVal ;
}


/*
 * Note: whilst a copy is taken of the OID parameters the authMsg 
 * pointer is buffered without copying.
 */

SnmpRMIBerrorEnum
SnmpRMIBcreateOperationAndaTxPduNode( snmpRMIBAgentDatap, numOperOids, OIDvec, numOids, PEvec, operAttrIds, 
					snmpOperation, authMsg, requestId, operationIdp, 
					tooBigMaxOidsPerPdu, operNodeTxPduNodepp )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
int	numOperOids ;
OID *	OIDvec  ; /* 0..(numOids-1) */
int	numOids ;
PE  *	PEvec   ; /* 0..(numOids-1) */
int *	operAttrIds ; /* 0..(numOids-1) */
SnmpOperationsEnum	snmpOperation ;
struct	type_SNMP_SnmpAuthMsg * authMsg ;
int	requestId, *operationIdp, tooBigMaxOidsPerPdu ;
struct  SnmpRMIBtxPduNode     ** operNodeTxPduNodepp ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
struct  SnmpRMIBoperationNode * newOperp = NULLSNMPRMIBOPERATIONNODE, * currOperp ;
int 	i ;

	if( (newOperp = (struct SnmpRMIBoperationNode *)calloc( 1, 
				sizeof( struct SnmpRMIBoperationNode ))) == NULLSNMPRMIBOPERATIONNODE ) {
	    retVal = SnmpRMIBmemAllocFailure ;

	} else if( (newOperp -> OIDspecificSNMPerrorp = (int *)malloc( 
						sizeof(int) * numOperOids )) == (int *)0) {

	    (void)free( (char *)newOperp ) ;
	    retVal = SnmpRMIBmemAllocFailure ;
	
	} else if( (newOperp -> operAttrIds = (int *)malloc(
						sizeof(int) * numOperOids )) == (int *)0) {
	    (void)free( (char *)newOperp ) ;
	    (void)free( (char *)newOperp -> OIDspecificSNMPerrorp ) ;
	    retVal = SnmpRMIBmemAllocFailure ;
	
	} else {

	    for( i=0 ; i<numOperOids ; i++ ) {
	    	newOperp -> OIDspecificSNMPerrorp[ i ] = int_SNMP_error__status_noError ;
		newOperp -> operAttrIds[ i ] = operAttrIds[ i ] ;
	    }

	    /* Buffer the parameters.. */
	    newOperp -> version = snmpRMIBAgentDatap -> version ;
	    newOperp -> snmpOperation = snmpOperation ;
	    *operationIdp = newOperp -> operationId = ++(snmpRMIBAgentDatap -> agentOperationId) ;
	    newOperp -> numOperOids = numOperOids ;

	    newOperp -> tooBigMaxOidsPerPdu = tooBigMaxOidsPerPdu ;

	    newOperp -> numberOfResponsesReceivedThisOper     = 0 ;
	    newOperp -> numberOfResponsesReceivedUpToThisOper =
	    			snmpRMIBAgentDatap->numOfResponsesReceived ;

	    newOperp -> snmpRMIBtxPduNodep = NULLSNMPRMIBTXPDUNODE ;		

	    newOperp -> nextSnmpRMIBoperationNodep= NULLSNMPRMIBOPERATIONNODE ;

	    /* Link in the operation structure into the current agent operations list */
	    if( snmpRMIBAgentDatap -> snmpRMIBoperationNodep == NULLSNMPRMIBOPERATIONNODE )
		snmpRMIBAgentDatap -> snmpRMIBoperationNodep =  newOperp ;

	    else {

	    	for( currOperp = snmpRMIBAgentDatap -> snmpRMIBoperationNodep ; 
		      currOperp -> nextSnmpRMIBoperationNodep ;
		      currOperp = currOperp -> nextSnmpRMIBoperationNodep )
			; /* Empty loop */
			  
		currOperp -> nextSnmpRMIBoperationNodep = newOperp ;
	    }
	    
	    /* Now form the TX pdu node.. */
	    if((retVal = SnmpRMIBcreateTxPduNode( newOperp, OIDvec, numOids, PEvec, authMsg,
						requestId, 0, &(newOperp -> snmpRMIBtxPduNodep ), FALSE )) == SnmpRMIBnoError )
		*operNodeTxPduNodepp = newOperp -> snmpRMIBtxPduNodep ; /* It must be the first one
									   in the txPdu list */
	}

	return retVal ;

} /* SnmpRMIBcreateOperationAndaTxPduNode() */

/*
 * Indicate that the respective PDU node and the request ID
 * have now been fully used. Remove the requestId from the
 * agents outstanding requests list (by shifting the remaining
 * requestIds down one position.)
 */
void
SnmpRMIBrequestIdComplete( snmpRMIBAgentDatap, thisTxPduNodep, requestId )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
struct  SnmpRMIBtxPduNode  * thisTxPduNodep ;
int	requestId ;
{
int	i ;

	thisTxPduNodep -> pduTxComplete = TRUE ;

	/* Remove the requestId from the agents list.. */
	for( i=0 ; i<snmpRMIBAgentDatap->numOutstandingRequestIDs ; i++ ) {

	    if( snmpRMIBAgentDatap->outstandingRequestIDs[i] == requestId ) {

#ifdef DEBUG
advise( NULLCP, "Completion of requestId %d TXed %ld S", requestId,
			snmpRMIBAgentDatap->pduTransmittedAtS[ i ] ) ;
#endif 
		/* Shuffle down the remaining request IDs and transmission times*/
		if( snmpRMIBAgentDatap->numOutstandingRequestIDs > 1 &&
		    (snmpRMIBAgentDatap->numOutstandingRequestIDs - 1 ) != i ) {

		    bcopy( (char*)&snmpRMIBAgentDatap->outstandingRequestIDs[i+1],
			   (char*)&snmpRMIBAgentDatap->outstandingRequestIDs[i],
			   sizeof( int ) * (snmpRMIBAgentDatap->numOutstandingRequestIDs - i - 1)) ;

		    bcopy( (char*)&snmpRMIBAgentDatap->pduTransmittedAtS[i+1],
			   (char*)&snmpRMIBAgentDatap->pduTransmittedAtS[i],
			   sizeof( long ) * (snmpRMIBAgentDatap->numOutstandingRequestIDs - i - 1)) ;
		}
		
		--snmpRMIBAgentDatap->numOutstandingRequestIDs ;

		break ; /* Exit from the for loop */
	    }
	}

	return ;
}

/*
 * SnmpRMIBsendFailed
 */

SnmpRMIBerrorEnum
SnmpRMIBsendPduToAgent( snmpRMIBAgentDatap, authMsg, requestId, reTxFlag )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
struct  type_SNMP_SnmpAuthMsg * authMsg ;
int	requestId, reTxFlag ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
PE	pe = NULLPE ;
struct 	timeval  tvs ;

	if( snmpRMIBAgentDatap -> agentParty -> party_tDomain == RFC1157DOMAIN ) {

	    if( send_v1_message (snmpRMIBAgentDatap->ps, (struct
					type_SNMP_Snmp1157Msg *) authMsg, &pe) == NOTOK )
		retVal = SnmpRMIBsendFailed ;

	} else if (send_v2_message ( snmpRMIBAgentDatap->ps, authMsg,
					snmpRMIBAgentDatap->managerParty, snmpRMIBAgentDatap->agentParty,
					snmpRMIBAgentDatap->mgmtContext, &pe) == NOTOK)
	    retVal = SnmpRMIBsendFailed ;


	/* Tidy up before leaving.. */
	if( pe != NULLPE ) {
	    pe_free( pe ), pe = NULLPE ;
	}

	if( retVal == SnmpRMIBnoError && reTxFlag == FALSE ) {

	    /* Add the reqestId to this agent's list since it is not already present.. */
	    
	    if( snmpRMIBAgentDatap -> numOutstandingRequestIDs != 0 &&
	    	snmpRMIBAgentDatap -> numOutstandingRequestIDs % SNMPRMIBOUTSANDINGIDSBLOCKSIZE == 0 ) {

		/* Extend the request ID array storage.. */
	    	snmpRMIBAgentDatap -> outstandingRequestIDs =
			(int *)realloc( (char*)snmpRMIBAgentDatap->outstandingRequestIDs,
			(size_t)sizeof(int)*(snmpRMIBAgentDatap->numOutstandingRequestIDs +
								SNMPRMIBOUTSANDINGIDSBLOCKSIZE) );

	    	snmpRMIBAgentDatap -> pduTransmittedAtS =
			(long *)realloc( (char*)snmpRMIBAgentDatap->pduTransmittedAtS,
			(size_t)sizeof(long)*(snmpRMIBAgentDatap->numOutstandingRequestIDs +
								SNMPRMIBOUTSANDINGIDSBLOCKSIZE) );
	    }	

	    snmpRMIBAgentDatap -> outstandingRequestIDs[ snmpRMIBAgentDatap->numOutstandingRequestIDs ] 
	    		= requestId ;

     	    (void)gettimeofday( &tvs, (struct timezone *) 0 ) ;
	    if( tvs.tv_usec >= 500000 ) tvs.tv_sec++ ;
	    snmpRMIBAgentDatap -> pduTransmittedAtS[ (snmpRMIBAgentDatap->numOutstandingRequestIDs)++ ]
			= tvs.tv_sec ;
	}

	return retVal ;

} /* SnmpRMIBsendPduToAgent() */


SnmpRMIBerrorEnum
SnmpRMIBformAndTxMessage( snmpRMIBAgentDatap, snmpOperation, OIDvec, numOids, PEvec, authMsgp, requestIdp )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
SnmpOperationsEnum snmpOperation ;
OID *	OIDvec  ;
int	numOids ;
PE  *	PEvec   ;
struct	type_SNMP_SnmpAuthMsg ** authMsgp ;
int *	requestIdp ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
struct	type_SNMP_SnmpAuthMsg * authmsg = NULLAUTHMSG ;
struct	type_SNMP_AuthInformation * authinfo;
struct	type_SNMP_SnmpMgmtCom * mgmtmsg;
struct	type_SNMP_PDUs * pdu;
struct	type_SNMP_PDU  * parm;
struct	type_SNMP_VarBindList **vp;

struct	snmp_party   * agentParty, * managerParty ;
int	offset = (int)snmpOperation, i ;

    if( snmpRMIBAgentDatap == NULLSNMPRMIBIFDATA )
    	return SnmpRMIBnullParameter ;

    if( (agentParty = snmpRMIBAgentDatap -> agentParty) == NULLPARTY )
	return SnmpRMIBnullAgentParty ;
 
    if( (managerParty = snmpRMIBAgentDatap -> managerParty) == NULLPARTY )
	return SnmpRMIBnullManagerParty ;

    if( agentParty -> party_tDomain != managerParty -> party_tDomain)
	return SnmpRMIBincompatibleTransDomains ;

    if( snmpRMIBAgentDatap -> mgmtContext == NULLCONTEXT )
	return SnmpRMIBnullContext ;

    if (agentParty -> party_tDomain == RFC1157DOMAIN) {

	struct type_SNMP_Snmp1157Msg *msg;

	switch (offset) {
	    case type_SNMP_PDUs_get__bulk__request:
	        offset = type_SNMP_PDUs_get__next__request;
		break;

	    case type_SNMP_PDUs_inform__request:
#ifdef DEBUG
		advise( NULLCP, "context held by SNMPv1 agent" ) ;
#endif
		return SnmpRMIBinvalidParameter ;

	    default:
		break;
	}

	if ((msg = (struct type_SNMP_Snmp1157Msg *) calloc (1, sizeof *msg)) == NULL)
	    goto no_mem;
	    
	msg -> version = int_SNMP_version_version__1 ;

	if ((msg -> community = str2qb ((char *) agentParty -> party_AuthPrivate,
				    agentParty -> party_AuthPrivate_len, 1)) == NULL) {
	    free ((char *) msg);
	    goto no_mem;
	}
	
	if ((pdu = (struct type_SNMP_PDUs *) calloc (1, sizeof *pdu)) == NULL)
	    goto no_mem;

	msg -> data   = pdu ;
	pdu -> offset = offset ;

	if ((parm = (struct type_SNMP_PDU *) calloc (1, sizeof *parm)) == NULL)
	    goto no_mem;

	authmsg = (struct type_SNMP_SnmpAuthMsg *) msg ;
    }
    else {

    	if ((authmsg = (struct type_SNMP_SnmpAuthMsg *)
					calloc (1, sizeof *authmsg)) == NULL) {
no_mem: ;
#ifdef DEBUG
	    advise (NULLCP, "out of memory");
#endif
	    if( authmsg )
	    	free_SNMP_SnmpAuthMsg( authmsg );

	    return SnmpRMIBmemAllocFailure ;
    	}

    	if ((authinfo = (struct type_SNMP_AuthInformation *)
					calloc (1, sizeof *authinfo)) == NULL)
	    goto no_mem;

    	authmsg -> authInfo = authinfo;   

    	if ((mgmtmsg = (struct type_SNMP_SnmpMgmtCom *) 
					calloc (1, sizeof *mgmtmsg)) == NULL)
	    goto no_mem;

   	authmsg -> authData = mgmtmsg;

    	if ((pdu = (struct type_SNMP_PDUs *) calloc (1, sizeof *pdu)) == NULL)
	    goto no_mem;

    	mgmtmsg -> pdu = pdu;

/* for now, always a PDU... */
    	pdu -> offset = offset;

	if ((parm = (struct type_SNMP_PDU *) calloc (1, sizeof *parm)) == NULL)
	    goto no_mem;

	if ((pdu -> offset = offset) == type_SNMP_PDUs_get__bulk__request) {
	    parm -> error__status = non_repeaters;
	    parm -> error__index = max_repetitions;
	}
    }

    pdu -> un.get__request = parm ;

    *requestIdp = parm -> request__id = ++(snmpRMIBAgentDatap -> agentRequestId) ;


    vp = &parm -> variable__bindings;

    for( i=0; i < numOids ; i++ ) {
 
	register struct type_SNMP_VarBindList *bind ;
	register struct type_SNMP_VarBind *v ;
	
	if ((bind = (struct type_SNMP_VarBindList *) calloc (1, sizeof *bind)) == NULL)
	    retVal = SnmpRMIBmemAllocFailure ;

	*vp = bind, vp = &bind -> next;

	if ((v = (struct type_SNMP_VarBind *) calloc (1, sizeof *v)) == NULL)
	    retVal = SnmpRMIBmemAllocFailure ;

	bind -> VarBind = v;

	/* Buffer the OID.. */
	if( (v->name  = oid_cpy(  OIDvec[ i ] )) == NULLOID )
	    retVal = SnmpRMIBmemAllocFailure ;
	else {
	    /* Buffer the PE (if available) else prepare response storage.. */
	    if( PEvec != (PE *)0 ) {
	    	if( (v->value = pe_cpy( PEvec[ i ] )) == NULLPE )
		    retVal = SnmpRMIBmemAllocFailure ;
	    } else {
	    	if( (v->value = pe_alloc( PE_CLASS_UNIV, PE_FORM_PRIM, PE_PRIM_NULL )) == NULLPE )
		    retVal = SnmpRMIBmemAllocFailure ;
	    }
	}
    }

    if( retVal == SnmpRMIBnoError && authmsg != NULLAUTHMSG ) {

    	/* 
    	 * Transmit the newly formed message.. 
    	 */
    	retVal = SnmpRMIBsendPduToAgent( snmpRMIBAgentDatap, authmsg, *requestIdp, FALSE ) ;
    
	if( retVal != SnmpRMIBnoError ) {

	    if (agentParty -> party_tDomain == RFC1157DOMAIN)
		free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *)authmsg) ;
	    else
		free_SNMP_SnmpAuthMsg (authmsg) ;
	}
	else *authMsgp = authmsg ;
    }    
    
    return retVal ;
}



/*
 * Depending on the maximum number of OIDs per request break down the request
 * between one or more SNMP pdus, transmit them and buffer their details to
 * the operational node and transmitted pdu linked lists.
 */
SnmpRMIBerrorEnum
SnmpRMIBbreakDownAndSendMessageToAgent( snmpRMIBAgentDatap, snmpOperation, OIDvec, numOids, PEvec, 
		operAttrIds, operationIdp, operNodeToBeCreated, txPduNodep,
		maxNumOidsPerRequest, attemptToRecoverFromPossibleMissingTooBigResponse )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
SnmpOperationsEnum snmpOperation ;
OID *	OIDvec  ;     /* 0..(numOids-1) */
int	numOids ;
PE  *	PEvec   ;     /* 0..(numOids-1) */
int *	operAttrIds ; /* 0..(numOids-1) */    	/* Note: NULL if operNodeToBeCreated == FALSE */
int *	operationIdp ;				/* Note: NULL if operNodeToBeCreated == FALSE */
int	operNodeToBeCreated ;			
struct  SnmpRMIBtxPduNode * txPduNodep ;/* Note: NULL if operNodeToBeCreated == TRUE  */
int 	maxNumOidsPerRequest, attemptToRecoverFromPossibleMissingTooBigResponse ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
struct	type_SNMP_SnmpAuthMsg * authmsg = NULLAUTHMSG ;
int	requestId, baseOidIndex, numOidsThisTime, numDiff ;
struct  SnmpRMIBtxPduNode * operNodeTxPduNodep = NULLSNMPRMIBTXPDUNODE ;
OID *	OIDp = (OID*)0 ;
PE  *	PEp  = (PE*)0  ;
int	usePeer ; /* as opposed to useSubordinate ! */

	usePeer = (operNodeToBeCreated==TRUE) ? TRUE : FALSE ;

	for( baseOidIndex = 0 ; 
	     baseOidIndex < numOids && retVal == SnmpRMIBnoError ; 
	     baseOidIndex += maxNumOidsPerRequest ) {

	    if( (numDiff = baseOidIndex + maxNumOidsPerRequest - numOids) > 0 )
	    	numOidsThisTime = maxNumOidsPerRequest - numDiff ;
	    else
		numOidsThisTime = maxNumOidsPerRequest ;

	    if( OIDvec )
	    	OIDp = &(OIDvec[ baseOidIndex ]) ;

	    if( PEvec )
	    	PEp  = &(PEvec[ baseOidIndex ]) ;
	    
	    if( (retVal = SnmpRMIBformAndTxMessage( snmpRMIBAgentDatap, snmpOperation,
		    				    OIDp, numOidsThisTime, PEp,
						    &authmsg, &requestId ) ) == SnmpRMIBnoError ) {
		if( operNodeToBeCreated == TRUE ) {

		    retVal = SnmpRMIBcreateOperationAndaTxPduNode( snmpRMIBAgentDatap, numOids,
		    		OIDp, numOidsThisTime, PEp, &(operAttrIds[baseOidIndex]),
	    			snmpOperation, authmsg, requestId, operationIdp, 
				numOidsThisTime, &operNodeTxPduNodep ) ;

		    txPduNodep = operNodeTxPduNodep ;
		    operNodeToBeCreated = FALSE ;

		} else {

		    retVal = SnmpRMIBcreateTxPduNode( txPduNodep->superiorSnmpRMIBoperNodep,
	    			OIDp, numOidsThisTime, PEp, authmsg, requestId, 
				txPduNodep->baseIndexWithinOperationOIDvec + baseOidIndex ,
				(usePeer == TRUE) ?
				&(txPduNodep->nextSnmpRMIBpeerTxPduNodep) :
				&(txPduNodep->nextSnmpRMIBsubordTxPduNodep), 
 				attemptToRecoverFromPossibleMissingTooBigResponse ) ;

		    attemptToRecoverFromPossibleMissingTooBigResponse = FALSE ; /* Only permitted to be TRUE
		    								   in first request */
		    usePeer = TRUE ;
		}

		if( retVal != SnmpRMIBnoError ) {
		    /* Free the unused SNMP message storage.. */
		    if( agentParty -> party_tDomain == RFC1157DOMAIN )
			free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *) authmsg);
		    else
			free_SNMP_SnmpAuthMsg (authmsg);
		}
	    }
	}
 
	return retVal ;
}
 
/*
 * SnmpRMIBnullParameter, SnmpRMIBnullAgentParty, SnmpRMIBnullManagerParty, 
 * SnmpRMIBincompatibleTransDomains, SnmpRMIBnullContext, SnmpRMIBinvalidParameter
 * 
 *
 * SnmpRMIBmemAllocFailure, SnmpRMIBsendFailed
 *
 */

SnmpRMIBerrorEnum
SnmpRMIBsendMessageToAgent( snmpRMIBAgentDatap, snmpOperation, OIDvec, numOids, PEvec, operAttrIds, 
								operationIdp, tooBigMaxOidsPerPdu )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
SnmpOperationsEnum snmpOperation ;
OID *	OIDvec  ;     /* 0..(numOids-1) */
int	numOids ;
PE  *	PEvec   ;     /* 0..(numOids-1) */
int *	operAttrIds ; /* 0..(numOids-1) */
int *	operationIdp, tooBigMaxOidsPerPdu ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;

	retVal = SnmpRMIBbreakDownAndSendMessageToAgent( snmpRMIBAgentDatap, snmpOperation, OIDvec, numOids, PEvec, operAttrIds,
			operationIdp, TRUE, NULLSNMPRMIBTXPDUNODE, tooBigMaxOidsPerPdu, FALSE ) ;
 
	return retVal ;

} /* SnmpRMIBsendMessageToAgent */



SnmpRMIBerrorEnum
SnmpRMIBformAndTxRetryPdu( snmpRMIBAgentDatap, snmpOperation, OIDvec, numOids, PEvec,
					baseIndexWithinOperationOIDvec, thisTxPduNodep )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
SnmpOperationsEnum snmpOperation ;
OID   *	OIDvec ;
int	numOids, baseIndexWithinOperationOIDvec ;
PE    *	PEvec  ;
struct	SnmpRMIBtxPduNode * thisTxPduNodep ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
struct	type_SNMP_SnmpAuthMsg * authmsg = NULLAUTHMSG ;
int	requestId ;

	if( (retVal = SnmpRMIBformAndTxMessage( snmpRMIBAgentDatap,
				snmpOperation, OIDvec, numOids, PEvec, &authmsg, &requestId ) ) == SnmpRMIBnoError ) {

	    if( (retVal = SnmpRMIBcreateTxPduNode( thisTxPduNodep->superiorSnmpRMIBoperNodep,
	    			OIDvec, numOids, PEvec, authmsg, requestId, 
				baseIndexWithinOperationOIDvec,
				&(thisTxPduNodep->nextSnmpRMIBsubordTxPduNodep), FALSE )) != SnmpRMIBnoError ) 
	    {
		if( agentParty -> party_tDomain == RFC1157DOMAIN )
		    free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *) authmsg);
		else
		    free_SNMP_SnmpAuthMsg (authmsg);
	    }
	}
 
	return retVal ;

} /* */


/*
 * After receiving an SNMP error when sending more than one OID
 * form individual PDUs for each OID in turn. This does not apply
 * to Set operations which must remain atomic.
 *
 * "tooBig" errors require each OID to be resent with fewer OIDs per PDU,
 * but the SNMP error does not need to be allocated to respective OIDs
 * due to the retry transmissions which will generate their own results
 * (unless the "tooBig" error was returned due to a remote agent fault.)
 *
 * Whilst "noSuchName" errors will lead to all but 
 * the OID defined by the error index being resent individually
 * This also needs to apply for Get-Next responses to account for
 * the very special case when the last object(s) in a MIB is/are not 
 * supported by the remote agent. If a Get-Next is applied to X.1 X.2 X.3, 
 * when X.3 is not supported a "noSuchName" error will be returned, which necessitates
 * individual Get-Nexts for X.0 and X.1 so that X.1.A and X.2.B can
 * be retrieved.
 *
 * "genErrors" merely require the allocation of the error message to
 * all of the OIDs, without any re-transmissions.
 * 
 * Note: this procedure is not used for timeout retries (hence the
 * code that prevent re-tries if only a single OID is involved.)
 */

SnmpRMIBerrorEnum
SnmpRMIBformAndTxRetryMessages( snmpRMIBAgentDatap, thisTxPduNodep, snmpErrorReceived, snmpErrorIndex )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
struct	SnmpRMIBtxPduNode  * thisTxPduNodep ;
int	snmpErrorReceived, snmpErrorIndex ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
int	i, numOids, maxNumOidsPerRequest ;
struct  SnmpRMIBoperationNode * thisOperNodep ;
SnmpOperationsEnum snmpOperation ;
PE *	PEp = (PE*)0  ;

	thisTxPduNodep -> snmpErrorReceived = snmpErrorReceived ;
	thisTxPduNodep -> snmpErrorIndex    = snmpErrorIndex ;


	thisOperNodep = thisTxPduNodep->superiorSnmpRMIBoperNodep ;
	
	if( thisTxPduNodep -> snmpErrorReceived != int_SNMP_error__status_noError ) {

	    /* Allocate the SNMP error to the respective OID(s) */
	    if( thisTxPduNodep -> snmpErrorReceived == int_SNMP_error__status_noSuchName ||
	        thisTxPduNodep -> snmpErrorReceived == int_SNMP_error__status_badValue     ) {

		/* Work-around for invalid SNMP agent implementations that return 
		   badValue errors with an ErrorIndex of 0 upon set operations !  */

		if( thisTxPduNodep -> snmpErrorIndex <= 0 || 
		    thisTxPduNodep -> snmpErrorIndex > thisTxPduNodep->numOids ) {
		    thisTxPduNodep -> snmpErrorIndex = snmpErrorIndex = 1 ;
		}

		thisOperNodep  -> OIDspecificSNMPerrorp[ 
			thisTxPduNodep->baseIndexWithinOperationOIDvec + thisTxPduNodep->snmpErrorIndex - 1 ] 
				= thisTxPduNodep -> snmpErrorReceived ;
	    } else {

	        /* Check if the Snmp error applies to all OIDs.. */
		if( thisTxPduNodep -> snmpErrorReceived == int_SNMP_error__status_genErr )
		    for( i = thisTxPduNodep->baseIndexWithinOperationOIDvec  ;
		    	 i < thisTxPduNodep->baseIndexWithinOperationOIDvec +
				      			thisTxPduNodep->numOids  ; i++ )
		        thisOperNodep -> OIDspecificSNMPerrorp[ i ] = thisTxPduNodep -> snmpErrorReceived ;
	    }
	}

	if( thisOperNodep -> snmpOperation == SetOperator ||
			(numOids = thisTxPduNodep -> numOids) <= 1 )
	    SnmpRMIBrequestIdComplete( snmpRMIBAgentDatap, thisTxPduNodep, thisTxPduNodep->requestId ) ;
	else {

	    if( thisTxPduNodep -> snmpErrorReceived == int_SNMP_error__status_tooBig ||
	        thisTxPduNodep -> snmpErrorReceived == int_SNMP_error__status_noSuchName  ) {

	    snmpOperation = thisOperNodep->snmpOperation ;

	    if( thisTxPduNodep -> snmpErrorReceived == int_SNMP_error__status_tooBig ) {

	    	/* We must reduce the number of OIDs sent per request.. */
		maxNumOidsPerRequest = (*snmpRMIBAgentDatap->tooBigReduceRetryFN)( numOids ) ;

		if( maxNumOidsPerRequest > 0 ) {

		    thisOperNodep -> tooBigMaxOidsPerPdu = maxNumOidsPerRequest ;

		    retVal = SnmpRMIBbreakDownAndSendMessageToAgent( snmpRMIBAgentDatap, snmpOperation,
				thisTxPduNodep->OIDvec, thisTxPduNodep->numOids, thisTxPduNodep->PEvec,
				(int*)0, (int*)0, FALSE, thisTxPduNodep, maxNumOidsPerRequest, FALSE ) ;
		} else {

		    /* Suspect a serious remote agent fault ! */
		    for( i = thisTxPduNodep->baseIndexWithinOperationOIDvec  ;
		    	 i < thisTxPduNodep->baseIndexWithinOperationOIDvec +
				      			thisTxPduNodep->numOids  ; i++ )
		        thisOperNodep -> OIDspecificSNMPerrorp[ i ] = thisTxPduNodep -> snmpErrorReceived ;

		    SnmpRMIBrequestIdComplete( snmpRMIBAgentDatap, thisTxPduNodep, thisTxPduNodep->requestId ) ;
		}

	    } else /* ..noSuchName error */
/* STATEMENT REMOVED if( thisTxPduNodep -> superiorSnmpRMIBoperNodep -> snmpOperation != GetNextOperator )*/ {
	    
		    /* Re-transmit the request with one OID per PDU.. */
	    	    for( i=0 ; i<numOids && retVal == SnmpRMIBnoError ; i++ ) {

			if( i != (thisTxPduNodep->snmpErrorIndex-1) ) {

			    if( thisTxPduNodep->PEvec )
				PEp = &(thisTxPduNodep->PEvec[ i ]) ;

			    retVal = SnmpRMIBformAndTxRetryPdu( snmpRMIBAgentDatap,
					snmpOperation, &(thisTxPduNodep->OIDvec[ i ]), 1, PEp,
					thisTxPduNodep->baseIndexWithinOperationOIDvec + i,
					thisTxPduNodep ) ;
			}
		    }
	    	}
	    }
	}
	
	return retVal ;
}


/*
 *
 *
 * Note: This procedure also tries to handle the case when a remote
 *	 agent completely ignores requests that it should be returning
 *	 a "tooBig" error for.
 */

void
SnmpRMIBhandleTimeout( snmpRMIBAgentDatap )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
{
int	operationId, requestId, timeoutIndex, i ;
struct  SnmpRMIBtxPduNode * thisTxPduNodep ;
struct 	timeval tvs ;
int	maxNumOidsPerRequest ;

	(void)gettimeofday( &tvs, (struct timezone *) 0 ) ;
	if( tvs.tv_usec >= 500000 ) tvs.tv_sec++ ;

	/*
	 * Note: when individual requestIds are determined to have completed
	 *       SnmpRMIBrequestIdComplete() causes the arrays pduTransmittedAtS[]
	 *	 and outstandingRequestIDs[] to shrink downwards by one index. Hence
	 *	 the array is processed from top down.
	 */

	/* Find those transmitted PDUs that have now been timed out... */	
	for( timeoutIndex = snmpRMIBAgentDatap->numOutstandingRequestIDs-1 ; timeoutIndex>=0 ; timeoutIndex-- ) {

	    if( snmpRMIBAgentDatap->pduTransmittedAtS[ timeoutIndex ] + snmpRMIBAgentDatap->timeoutS <= tvs.tv_sec ) {

		requestId = snmpRMIBAgentDatap->outstandingRequestIDs[ timeoutIndex ] ;
#ifdef DEBUG
advise( NULLCP, "Timeout of requestId %d %ld S + %ld S <= %ld S", requestId,
			snmpRMIBAgentDatap->pduTransmittedAtS[ timeoutIndex ],
 			snmpRMIBAgentDatap->timeoutS, tvs.tv_sec ) ;
#endif
		
		/* Find the pdu TX node that corresponds to this requst ID.. */
		if( SnmpRMIBgetOperationIdforRequestId( snmpRMIBAgentDatap,
				requestId, &operationId, &thisTxPduNodep ) == SnmpRMIBnoError ) {

		    /* Check if the maximum number of retries have been exceeded.. */
		    if( thisTxPduNodep->reTxCount == snmpRMIBAgentDatap->maxRetries ) {
#ifdef DEBUG
advise( NULLCP, "MAX RETRIES OF %d REACHED FOR request Id %d oper Id %d",  snmpRMIBAgentDatap->maxRetries,
						thisTxPduNodep->requestId, operationId ) ;
#endif
			/* 
			 * If no PDUs have been received from the remote agent for this operation, 
			 * when the agent has previously responded, assume that the 
			 * remote agent is failing to return a "tooBig" error.
			 */
			SnmpRMIBrequestIdComplete( snmpRMIBAgentDatap, thisTxPduNodep, requestId ) ;

			if( snmpRMIBAgentDatap -> remoteAgentMayIgnoreTooBigMessages == FALSE ||
			    thisTxPduNodep->superiorSnmpRMIBoperNodep->numberOfResponsesReceivedUpToThisOper == 0 ||
/*			    thisTxPduNodep->superiorSnmpRMIBoperNodep->numberOfResponsesReceivedThisOper     != 0 ||*/
			    (maxNumOidsPerRequest = (*snmpRMIBAgentDatap->tooBigReduceRetryFN)( thisTxPduNodep -> numOids )) <= 0 ||
			    SnmpRMIBbreakDownAndSendMessageToAgent( snmpRMIBAgentDatap, thisTxPduNodep->superiorSnmpRMIBoperNodep->snmpOperation,
						thisTxPduNodep->OIDvec, thisTxPduNodep->numOids, thisTxPduNodep->PEvec,
						(int*)0, (int*)0, FALSE, thisTxPduNodep,
						maxNumOidsPerRequest, TRUE ) != SnmpRMIBnoError ) {

			    /* Indicate the reason for the failure - multiple timeouts.. */
		    	    for( i = thisTxPduNodep->baseIndexWithinOperationOIDvec  ; 
			         i < thisTxPduNodep->baseIndexWithinOperationOIDvec +
			      				thisTxPduNodep->numOids  ; i++ )
			    	thisTxPduNodep->superiorSnmpRMIBoperNodep->
			    		OIDspecificSNMPerrorp[ i ] = int_SNMP_error__status_timeOut ;

			}

		    } else {

		    	/* 
			 * Further retransmission required.. 
			 */
			 
			/* Update the requestId */
			requestId = thisTxPduNodep->requestId =
				snmpRMIBAgentDatap->outstandingRequestIDs[ timeoutIndex ] =
								++(snmpRMIBAgentDatap -> agentRequestId) ;

			if( snmpRMIBAgentDatap->version == SnmpV1 )
			    ((struct type_SNMP_Snmp1157Msg *)thisTxPduNodep->authMsg)->data->un.get__request->request__id = requestId ;
			else
			    thisTxPduNodep->authMsg->authData->pdu->un.get__request->request__id = requestId ;
			

			if( SnmpRMIBsendPduToAgent( snmpRMIBAgentDatap, thisTxPduNodep->authMsg, requestId, TRUE ) == SnmpRMIBnoError ) {

			    ++thisTxPduNodep -> reTxCount ;
			    /* Update the transmission time... */
			    snmpRMIBAgentDatap->pduTransmittedAtS[ timeoutIndex ] = tvs.tv_sec ;
			
			} else {
			
			    SnmpRMIBrequestIdComplete( snmpRMIBAgentDatap, thisTxPduNodep, requestId ) ;
			}
		    }
		}
	    }
	}

	return ;
}


/* 
 * Assumes depth first linked list traversal..
 */
int
SnmpRMIBcheckForPendingTxPdu( checkTxPduNodep, operationId, dummyintp, dummyTxPduNodepp )
struct	SnmpRMIBtxPduNode * checkTxPduNodep ;
int	operationId, *dummyintp ;
struct  SnmpRMIBtxPduNode ** dummyTxPduNodepp ;
{
	if( checkTxPduNodep -> superiorSnmpRMIBoperNodep -> operationId == operationId &&
	    checkTxPduNodep -> nextSnmpRMIBsubordTxPduNodep == NULLSNMPRMIBTXPDUNODE     ) {

	    if( checkTxPduNodep -> pduTxComplete == FALSE )
	    	return TRUE ; /* At least one transmission is still pending */
	}

	return FALSE ;
}

/*
 * Check through the linked lists of SnmpRMIBtxPduNode nodes to see
 * if any transmissions are still in progress..
 */
int
SnmpRMIBisOperationStillNotCompleted( currOperNodep, requiredOperationId )
struct  SnmpRMIBoperationNode * currOperNodep ;
int	requiredOperationId ;
{
int	operationNotYetCompleted = FALSE ;

	if( currOperNodep -> snmpRMIBtxPduNodep != NULLSNMPRMIBTXPDUNODE )
	    SnmpRMIBrecursiveTxPduNodeOperation( currOperNodep -> snmpRMIBtxPduNodep, requiredOperationId,
		&operationNotYetCompleted, (int *)0, (struct SnmpRMIBtxPduNode**)0,
		SnmpRMIBcheckForPendingTxPdu ) ;

	return operationNotYetCompleted ;
}

/*
 * Note: no guarantee that PDUs are receive for the required operationId if
 * multiple asynch operations are in progress.
 *
 * SnmpRMIBnoPDUreceived (xselect)
 * SnmpRMIBreceiveFailed, SnmpRMIBunexpectedResponsePDUtype, SnmpRMIBreceivePDUerror
 * SnmpRMIBunknownRequestId
 */
SnmpRMIBerrorEnum
SnmpRMIBreceivePduFromAgent( snmpRMIBAgentDatap )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
int	PDUavailable = FALSE, requestId = -1, operationId, TimeoutOccurred = FALSE ;
struct  SnmpRMIBoperationNode * thisOperNodep ;
struct  SnmpRMIBtxPduNode     * thisTxPduNodep ;
struct	snmp_party *agentParty, *srcParty, *dstParty ;
struct	type_SNMP_SnmpAuthMsg *newmsg = NULL;
struct	type_SNMP_PDU *parm = NULL;
struct	qbuf *qb = NULL;
PE	pe = NULLPE ;
fd_set	rfds;

	FD_ZERO (&rfds);
	FD_SET ( snmpRMIBAgentDatap -> sd, &rfds);

	switch (xselect (snmpRMIBAgentDatap->sd + 1, &rfds, NULLFD, NULLFD, 
 							snmpRMIBAgentDatap->timeoutS )) {
	
	    case NOTOK:
	        advise( "failed", "xselect" );
		break ;

	    default:
		if (FD_ISSET( snmpRMIBAgentDatap->sd, &rfds )) {
		    PDUavailable = TRUE ;
		    break;
		}
		/* else fall... */
	    case OK:
		fprintf (stderr, "Timeout...\n");
		TimeoutOccurred = TRUE ;
		break ;
	}

	if( PDUavailable == FALSE ) {

	    if( TimeoutOccurred == TRUE )
		SnmpRMIBhandleTimeout( snmpRMIBAgentDatap ) ;

	    else retVal = SnmpRMIBnoPDUreceived ;

	} else {

	    agentParty = snmpRMIBAgentDatap->agentParty ;

	    if( agentParty -> party_tDomain == RFC1157DOMAIN) {
		struct type_SNMP_Snmp1157Msg *msg;

		if (recv_v1_message ( snmpRMIBAgentDatap->ps, (struct type_SNMP_Snmp1157Msg **) &newmsg, &pe)
		        == NOTOK)
		    retVal = SnmpRMIBreceiveFailed ;

		else if ((msg = (struct type_SNMP_Snmp1157Msg *) newmsg) -> data -> offset
	        			!= type_SNMP_PDUs_response) {
#ifdef DEBUG
		    fprintf (stderr, "unexpected message type %d\n", msg -> data -> offset);
#endif
		    retVal = SnmpRMIBunexpectedResponsePDUtype ;

	    	} else requestId = (parm = msg -> data -> un.response) -> request__id;

	    } else {

		/* SNMPv2 */
    
		if (recv_v2_message ( &newmsg, &srcParty, &dstParty, &pe, &qb) == NOTOK)
		    retVal = SnmpRMIBreceiveFailed ;

		else if (newmsg -> authData -> pdu -> offset != type_SNMP_PDUs_response) {
#ifdef DEBUG
		    fprintf (stderr, "unexpected message type %d\n", newmsg -> authData -> pdu -> offset );
#endif
		    retVal = SnmpRMIBunexpectedResponsePDUtype ;

		} else requestId = (parm = newmsg -> authData -> pdu -> un.response) -> request__id;
	    }

	    if( retVal == SnmpRMIBnoError ) {

		if( (retVal = SnmpRMIBgetOperationIdforRequestId( snmpRMIBAgentDatap,
		                                            requestId, &operationId, &thisTxPduNodep )) == SnmpRMIBnoError ) {

		    thisOperNodep = thisTxPduNodep -> superiorSnmpRMIBoperNodep ;

		    /* Increment the number of successfully received response PDUs.. */
		    ++thisOperNodep -> numberOfResponsesReceivedThisOper ;

		    /* Check for the "ignore too big errors" remote agent handling case.. */
		    if( thisTxPduNodep -> attemptToRecoverFromPossibleMissingTooBigResponse == TRUE ) {
			/* Reduce the number of OIDs permitted per request.. */
			thisOperNodep->tooBigMaxOidsPerPdu = thisTxPduNodep->numOids ;
		    }

		    if( parm -> error__status != int_SNMP_error__status_noError) {
#ifdef DEBUG
			fprintf (stderr, "%s at position %d\n",
				snmp_error (parm -> error__status), parm -> error__index);
#endif
			retVal = SnmpRMIBformAndTxRetryMessages( snmpRMIBAgentDatap, thisTxPduNodep,
		 					(int)parm -> error__status, (int)parm -> error__index ) ;

		    } else {

			/* Successful request */

			SnmpRMIBrequestIdComplete( snmpRMIBAgentDatap, thisTxPduNodep, requestId ) ;

			retVal = SnmpRMIBcreateOrUpdateRxedResultsNode(
						thisOperNodep, thisTxPduNodep, parm->variable__bindings ) ;
		    }

	    	} else if( retVal == SnmpRMIBunknownRequestId ) {
		    /* Looks like an old response has finally arrived ! */
		    retVal = SnmpRMIBnoError ;
#ifdef DEBUG
		    advise( NULLCP, "Warning: Unexpected requestId %d", requestId ) ;
#endif
		}
	    }

    	    if (pe)
		pe_free (pe);
	    if (qb)
		qb_free (qb);

	    if (newmsg) {
		/* Free the received message.. */
		if( agentParty -> party_tDomain == RFC1157DOMAIN )
		    free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *) newmsg );
		else
		    free_SNMP_SnmpAuthMsg (newmsg);
	    }
	}

	return retVal ;
}

/*
 * Cause the collation of incoming PDUs until all responses
 * have been received for the specified operation.
 *
 * SnmpRMIBnoKnownOperation
 */
SnmpRMIBerrorEnum
SnmpRMIBreceiveMessageFromAgent( snmpRMIBAgentDatap, requiredOperationId, 
					currRXedResultspp, tooBigMaxOidsPerPdup )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
int	requiredOperationId ;
struct	SnmpRMIBrxedResults ** currRXedResultspp ;
int *	tooBigMaxOidsPerPdup ;
{
SnmpRMIBerrorEnum retVal = SnmpRMIBnoError ;
int	operationHasNotYetCompleted = TRUE, failureDueToMultipleTimeouts, i ;
struct  SnmpRMIBoperationNode * currOperNodep ;

	*currRXedResultspp = NULLSNMPRMIBRXEDRESULTS ;

	for( currOperNodep = snmpRMIBAgentDatap -> snmpRMIBoperationNodep ;
	     currOperNodep != NULLSNMPRMIBOPERATIONNODE && 
	     	currOperNodep -> operationId != requiredOperationId ;
	     currOperNodep =  currOperNodep -> nextSnmpRMIBoperationNodep )
		; /* DUMMY LOOP BODY */

	if( currOperNodep == NULLSNMPRMIBOPERATIONNODE )
	    retVal = SnmpRMIBnoKnownOperation ;

	else {
	
	    /* .. Check if we already have a full set of results for this operation... */
	    operationHasNotYetCompleted = SnmpRMIBisOperationStillNotCompleted( currOperNodep, 
 									requiredOperationId ) ;

	    while( retVal == SnmpRMIBnoError && operationHasNotYetCompleted == TRUE ) {

		if( (retVal = SnmpRMIBreceivePduFromAgent( snmpRMIBAgentDatap )) == SnmpRMIBnoError )
		    operationHasNotYetCompleted = SnmpRMIBisOperationStillNotCompleted( currOperNodep, requiredOperationId ) ;
	    }

	    if( operationHasNotYetCompleted == FALSE ) {

		if( currOperNodep -> snmpRMIBreceivedResultsNodep == NULLSNMPRMIBRXEDRESULTS ) {

		    if( retVal == SnmpRMIBnoError ) {

			failureDueToMultipleTimeouts = TRUE ;

			/* Check for a multiple timeout failure.. */
			for( i=0 ; i < currOperNodep->numOperOids ; i++ )
			    if( currOperNodep->OIDspecificSNMPerrorp[ i ] != int_SNMP_error__status_timeOut ) {
			    	failureDueToMultipleTimeouts = FALSE ;
				break ;
			    }

		    	if( failureDueToMultipleTimeouts == TRUE )
			    retVal = SnmpRMIBmultipleTimeouts ;
			else {
			    /* We still need to make any Snmp specific errors available.. */
			    retVal = SnmpRMIBcreateOrUpdateRxedResultsNode( currOperNodep, NULLSNMPRMIBTXPDUNODE, NULL ) ;			    
			}
		    }
		} 
		
		if( currOperNodep -> snmpRMIBreceivedResultsNodep != NULLSNMPRMIBRXEDRESULTS &&
			retVal == SnmpRMIBnoError ) {
		
		    /* Transfer the OID specific errors across to the results node.. */
		    currOperNodep -> snmpRMIBreceivedResultsNodep -> OIDspecificSNMPerrorp = 
							currOperNodep ->  OIDspecificSNMPerrorp ;

		    currOperNodep -> OIDspecificSNMPerrorp = (int *)0 ;
			
		    currOperNodep -> snmpRMIBreceivedResultsNodep -> operAttrIds = 
							currOperNodep -> operAttrIds ;
		    currOperNodep -> operAttrIds = (int *)0 ;
			
		    *currRXedResultspp = currOperNodep -> snmpRMIBreceivedResultsNodep ;
		    currOperNodep -> snmpRMIBreceivedResultsNodep = NULLSNMPRMIBRXEDRESULTS ;

		    if( tooBigMaxOidsPerPdup )
		    	*tooBigMaxOidsPerPdup = currOperNodep->tooBigMaxOidsPerPdu ;

		    snmpRMIBAgentDatap -> numOfResponsesReceived += currOperNodep -> numberOfResponsesReceivedThisOper ;
		}
	    }

	    /* Free up the non-results operation storage.. */
    	    SnmpRMIBfreeOperationStorage( snmpRMIBAgentDatap, requiredOperationId ) ;
	}

	return retVal ;
}


/* */

static int  process ( authmsg, learning )
struct type_SNMP_SnmpAuthMsg *authmsg;
int	learning ;
{
    int	    gotone,
	    request_id,
	    offset,
    	    tries;
    PE	    pe = NULLPE;
    struct qbuf *qb = NULL;
    struct type_SNMP_SnmpAuthMsg *newmsg = NULL;
    struct snmp_party *srcParty,
		      *dstParty;
    register struct type_SNMP_PDU *parm = NULL;
    register struct type_SNMP_VarBindList *vp = NULL;

    if (authmsg == NULL)
	return OK;

    if (agentParty -> party_tDomain == RFC1157DOMAIN) {
	struct type_SNMP_Snmp1157Msg *msg;

    	request_id = (msg = (struct type_SNMP_Snmp1157Msg *) authmsg) 
				-> data -> un.get__request -> request__id;

    	offset = msg -> data -> offset;

	if( offset == type_SNMP_PDUs_set__request )
	    tries = 0;
    }
    else {

    	request_id = authmsg -> authData -> pdu -> un.get__request -> request__id;

    	offset = authmsg -> authData -> pdu -> offset ;
    }

    if( offset == type_SNMP_PDUs_set__request)
	tries = 0;

    gotone = 0, tries = retries;

again: ;

    if (agentParty -> party_tDomain == RFC1157DOMAIN) {
	if (send_v1_message (ps, (struct type_SNMP_Snmp1157Msg *) authmsg, &pe)
	        == NOTOK )
	    goto out;
    }
    else
	if (send_v2_message ( ps, authmsg, managerParty, agentParty,
			   mgmtContext, &pe) == NOTOK)

try_again: ;

    if (pe)
	pe_free (pe), pe = NULL;
    if (qb)
	qb_free (qb), qb = NULL;
    if (newmsg) {
	if (agentParty -> party_tDomain == RFC1157DOMAIN)
	    free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *) newmsg);
	else
	    free_SNMP_SnmpAuthMsg (newmsg);

	newmsg = NULL;
    }

    {
	fd_set	rfds;

	FD_ZERO (&rfds);
	FD_SET ( sd, &rfds);

	switch (xselect ( sd + 1, &rfds, NULLFD, NULLFD, timeout)) {
	    case NOTOK:
	        advise ("failed", "xselect");
		goto out;

	    default:
		if (FD_ISSET ( sd, &rfds))
		    break;
		/* else fall... */
	    case OK:
		if (watch)
		    fprintf (stderr, "timeout...\n");
		if (offset == type_SNMP_PDUs_set__request) {
		    advise (NULLCP,
			    "no response within %d second%s", timeout,
			    timeout != 1 ? "s" : "");
		    goto out;
		}
		if (tries-- > 0)
		    goto again;
		advise (NULLCP,
			"no %sresponse within %d retries of %s%d second%s each",
			gotone ? "acceptable " : "", retries + gotone,
			gotone ? "upto " : "", timeout,
			timeout != 1 ? "s" : "");
		goto out;
	}
    }

    if (agentParty -> party_tDomain == RFC1157DOMAIN) {
	struct type_SNMP_Snmp1157Msg *msg;

	if (recv_v1_message ( ps, (struct type_SNMP_Snmp1157Msg **) &newmsg, &pe)
	        == NOTOK)
	    goto out;
	gotone++;

	if ((msg = (struct type_SNMP_Snmp1157Msg *) newmsg) -> data -> offset
	        != type_SNMP_PDUs_response) {
	    if (watch)
		fprintf (stderr, "unexpected message type %d\n",
			 msg -> data -> offset);
	    goto try_again;
	}

    	if ((parm = msg -> data -> un.response) -> request__id
	    != request_id) {
	    fprintf (stderr, "request-id mismatch (got %d, wanted %d)\n",
		 parm -> request__id, request_id);
	    goto try_again;
	}
	    
	request_id = (parm = msg -> data -> un.response) -> request__id;
    }
    else {
	if (recv_v2_message ( &newmsg, &srcParty, &dstParty, &pe, &qb)
	        == NOTOK)
	    goto out;

	gotone++;

	if (newmsg -> authData -> pdu -> offset != type_SNMP_PDUs_response) {
	    if (watch)
		fprintf (stderr, "unexpected message type %d\n",
			 newmsg -> authData -> pdu -> offset);
	    goto try_again; /* KMTM Note goto out in snmpi.c */
	}

    	if ((parm = newmsg -> authData -> pdu -> un.response) -> request__id
	    != request_id) {
	    fprintf (stderr, "request-id mismatch (got %d, wanted %d)\n",
		 parm -> request__id, request_id);
	    goto try_again;
    	}

	request_id = (parm = newmsg -> authData -> pdu -> un.response) -> request__id;
    }
	
    if ((lasterr = parm -> error__status) != int_SNMP_error__status_noError) {
	fprintf (stderr, "%s at position %d\n",
		snmp_error (parm -> error__status), parm -> error__index);
	goto out;
    }

    if (learning
	    && srcParty -> party_authProt != NOAUTH
	    && learn_party (parm -> variable__bindings) == NOTOK)
	advise (NULLCP, "in-core party database suspect");

    for (vp = parm -> variable__bindings; vp; vp = vp -> next) {
	caddr_t	 value;
	register OI	oi;
	register OS	os;
	register struct type_SNMP_VarBind *v = vp -> VarBind;

	if (v -> value -> pe_class == PE_CLASS_CONT) {
	    printf ("%s=%s\n", oid2ode (v -> name),
		    snmpV2_exception (v -> value -> pe_id));
	    continue;
	}
	if ((oi = name2inst (v -> name)) == NULL) {
/*
	    advise (NULLCP, "unknown variable \"%s\"", oid2ode (v -> name));
 */
no_dice: ;
	    printf ("%s=", oid2ode (v -> name));
	    vunknown (v -> value);
	    continue;
	}
	if ((os = oi -> oi_type -> ot_syntax) == NULL) {
	    advise (NULLCP, "unknown syntax for object \"%s\"",
		    oi -> oi_type -> ot_text);
	    goto no_dice;
	}
	if ((*os -> os_decode) (&value, v -> value) == NOTOK) {
	    advise (NULLCP, "decode error for variable \"%s\"",
		    oid2ode (v -> name));
	    goto no_dice;
	}

	printf ("%s=", oid2ode (v -> name));
	(*os -> os_print) (value, os);
	printf ("\n");

	(*os -> os_free) (value);
    }

out: ;
    if (authmsg) {
	if (agentParty -> party_tDomain == RFC1157DOMAIN)
	    free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *) authmsg);
	else
	    free_SNMP_SnmpAuthMsg (authmsg);
    }
    if (newmsg) {
	if (agentParty -> party_tDomain == RFC1157DOMAIN)
	    free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *) newmsg);
	else
	    free_SNMP_SnmpAuthMsg (newmsg);
    }

    return OK;
}

/* snmpV1 */

static int  send_v1_message ( sp, msg, pe)
PS	sp;
register struct type_SNMP_Snmp1157Msg *msg;
PE     *pe;
{
    int	    result = NOTOK;
    char    buffer[BUFSIZ];

    if (*pe)
	pe_free (*pe), *pe = NULL;

    if (encode_SNMP_Snmp1157Msg (pe, 1, 0, NULLCP, msg) != NOTOK) {

	if (watch)
	    (void) print_SNMP_Snmp1157Msg (*pe, 1, NULLIP, NULLVP, NULLCP);

	if (pe2ps (sp, *pe) == NOTOK) {
	    (void) sprintf (buffer, "failed: pe2ps loses {%s}",
			    ps_error (sp -> ps_errno));
snmpV1_result: ;
	    fprintf( stderr, "%s\n", buffer );
	    goto out;
	}
	else
	    result = OK;
    }
    else {
	(void) sprintf (buffer, "failed: encode_SNMP_Snmp1157Msg loses {%s}",
			PY_pepy);
	goto snmpV1_result;
    }

out: ;
    if (*pe && result != OK)
	pe_free (*pe), *pe = NULL;

    ps_flush( sp ) ;

    return result;
}


/* snmpV2 */

int	send_v2_message (sp, msg, srcParty, dstParty, ourContext, pe)
PS	sp;
register struct type_SNMP_SnmpAuthMsg *msg;
register struct snmp_party *srcParty,
			   *dstParty;
struct snmp_context *ourContext;
PE     *pe;
{
    int	    result = NOTOK;
    struct type_SNMP_SnmpPrivMsg *privmsg = NULL;

    if (srcParty -> party_authProt != dstParty -> party_authProt) {
	advise (NULLCP,
		"authentication mismatch on source and destination parties");
	return NOTOK;
    }
    if (srcParty -> party_privProt != dstParty -> party_privProt) {
	advise (NULLCP, "privacy mismatch on source and destination parties");
	return NOTOK;
    }

    if (*pe)
	pe_free (*pe), *pe = NULL;
    if (msg -> authData -> srcParty)
	oid_free (msg -> authData -> srcParty);
    if ((msg -> authData -> srcParty = oid_cpy (srcParty -> party_Identity))
	    == NULL) {
no_mem: ;
	advise (NULLCP, "out of memory");
	goto out;
    }
    if (msg -> authData -> dstParty)
	oid_free (msg -> authData -> dstParty);
    if ((msg -> authData -> dstParty = oid_cpy (dstParty -> party_Identity))
	    == NULL)
	goto no_mem;
    if (msg -> authData -> context)
	oid_free (msg -> authData -> context);
    if ((msg -> authData -> context = oid_cpy (ourContext -> sc_identity))
	    == NULL)
	goto no_mem;

    if ((privmsg = (struct type_SNMP_SnmpPrivMsg *) calloc (1,
							    sizeof *privmsg))
		== NULL
	    || (privmsg -> privDst = oid_cpy (dstParty -> party_Identity))
		    == NULLOID)
	goto no_mem;
    if ((privmsg -> privData =
	     generate_outgoing_auth (msg, srcParty, dstParty,
				     (caddr_t) (watch ? -1 : 0))) == NULL) {
	advise (NULLCP, "error generating outgoing authentication: %s",
		PY_pepy);
	goto out;
    }

    if (do_encryption (privmsg, dstParty) == NOTOK) {
	advise (NULLCP, "error encrypting: %s", PY_pepy);
	goto out;
    }

    if (encode_SNMP_SnmpPrivMsg (pe, 1, 0, NULLCP, privmsg) != NOTOK) {
	if (watch)
	    (void) print_SNMP_SnmpPrivMsg (*pe, 1, NULLIP, NULLVP, NULLCP);

       if (pe2ps (sp, *pe) == NOTOK) 
	   advise (NULLCP, "pe2ps: %s", ps_error (sp -> ps_errno));
       else
	   result = OK;
    }
    else
	advise (NULLCP, "encode_SNMP_SnmpPrivMsg: %s", PY_pepy);

out: ;
    if (*pe && result != OK)
	pe_free (*pe), *pe = NULL;

    ps_flush( sp ) ;

    if (privmsg)
	free_SNMP_SnmpPrivMsg (privmsg);

    return result;
}

/* SNMPv1 */

static int  recv_v1_message ( sp, msg, pe )
PS	sp ;
register struct type_SNMP_Snmp1157Msg **msg;
PE	*pe;
{
char	buffer[BUFSIZ];

    *msg = NULL;

    if ((*pe = ps2pe (sp)) == NULLPE) {
	(void) sprintf (buffer, "failed: ps2pe loses {%s}",
			ps_error (sp -> ps_errno));
snmp_result: ;
	fprintf( stderr, "%s\n", buffer );
	goto out;
    }

    if (decode_SNMP_Snmp1157Msg (*pe, 1, NULLIP, NULLVP, msg) == NOTOK) {
	(void) sprintf (buffer, "failed: decode_SNMP_Snmp1157Msg loses {%s}",
			PY_pepy);
	goto snmp_result;
    }

    if (watch)
	(void) print_SNMP_Snmp1157Msg (*pe, 1, NULLIP, NULLVP, NULLCP);

    return OK;

out: ;
    if (*pe)
	pe_free (*pe), *pe = NULLPE;

    ps_flush( sp ) ;
	
    if (*msg)
	free_SNMP_Snmp1157Msg (*msg), *msg = NULL;

    return NOTOK;
}


/* SNMPv2 */

int	recv_v2_message( msg, srcParty, dstParty, pe, qb )
register struct type_SNMP_SnmpAuthMsg **msg;
register struct snmp_party **srcParty,
			   **dstParty;
PE	*pe;
struct qbuf **qb;
{
    struct type_SNMP_SnmpPrivMsg *privmsg;
    PS	    qbufPS;

    privmsg = NULL, *msg = NULL;
    *srcParty = NULL, *dstParty = NULL;
    *pe = NULL, *qb = NULL;
    qbufPS = NULL;

    if ((*pe = ps2pe (ps)) == NULLPE) {
bad_parse: ;
	advise (NULLCP, "ps2pe: %s", ps_error (ps -> ps_errno));
	goto out;
    }

    if (decode_SNMP_SnmpPrivMsg (*pe, 1, NULLIP, NULLVP, &privmsg) == NOTOK) {
	advise (NULLCP, "decode_SNMP_SnmpPrivMsg: %s", PY_pepy);
	goto out;
    }

    if (watch)
	(void) print_SNMP_SnmpPrivMsg (*pe, 1, NULLIP, NULLVP, NULLCP);

    if ((*dstParty = lookup_party (privmsg -> privDst, MANAGER)) == NULLPARTY){
	advise (NULLCP, "badParty: destination=%s",
		oid2ode (privmsg -> privDst));
	goto out;
    }

    if ((*qb = do_decryption (privmsg, *dstParty)) == NULL) {
	advise (NULLCP, "error decrypting: %s", PY_pepy);
	goto out;
    }

    if ((qbufPS = ps_alloc (qbuf_open)) == NULLPS
	    || qbuf_setup (qbufPS, *qb) == NOTOK) {
	if (qbufPS == NULLPS)
	    advise (NULLCP, "ps_alloc(qbuf_open): failed");
	else
	    advise (NULLCP, "qbuf_setup: %s", ps_error (ps -> ps_errno));

	goto out;
    }

    if (*pe)
	pe_free (*pe), *pe = NULLPE;

    if ((*pe = ps2pe (qbufPS)) == NULLPE)
	goto bad_parse;
    do_decryption_fixup (privmsg, *qb);

    if (decode_SNMP_SnmpAuthMsg (*pe, 1, NULLIP, NULLVP, msg) == NOTOK) {
	advise (NULLCP, "decode_SNMP_SnmpAuthMsg: %s", PY_pepy);
	goto out;
    }

    if (watch)
	(void) print_SNMP_SnmpAuthMsg (*pe, 1, NULLIP, NULLVP, NULLCP);

    if ((*srcParty = lookup_party ((*msg) -> authData -> srcParty, AGENT))
	    == NULLPARTY) {
	advise (NULLCP, "badParty: source=%s",
		oid2ode ((*msg) -> authData -> srcParty));
	goto out;
    }

    if (authenticate_incoming (privmsg, *msg, *srcParty, *dstParty, NULLIP)
	    == NOTOK) {
	advise (NULLCP, "authentication failure: %s", PY_pepy);
	goto out;
    }

    ps_free (qbufPS);
    free_SNMP_SnmpPrivMsg (privmsg);
    return OK;

out: ;
    if (*pe)
	pe_free (*pe), *pe = NULLPE;
    if (qbufPS)
	ps_free (qbufPS);
    if (*qb)
	qb_free (*qb), *qb = NULL;
    if (privmsg)
	free_SNMP_SnmpPrivMsg (privmsg);
    if (*msg)
	free_SNMP_SnmpAuthMsg (*msg), *msg = NULL;

    return NOTOK;
}

/* */

static char *access_t[] = { "not-accessible",
			    "read-only",
			    "write-only",
			    "read-write"};

static char *status_t[] = { "obsolete",
			    "mandatory",
			    "optional",
			    "deprecated" };
			     

static int  f_compile (vec)
char  **vec;
{
    register int    i,
		    j,
		    k;
    int	    fast,
	    source;
    char   *cp,
	   *file,
    	    sysout[BUFSIZ];
    register OS	    os;
    OS	    first,
	    last;
    register OT	    ot;
    FILE   *fp;

    fast = source = 0;
    file = defs ? defs : "./objects.defs";
    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("compile [-f|-s] [file]\n");
	printf ("    -f:   brief output (default to stdout)\n");
	printf ("    -s:   source output (C language bindings)\n");
	printf ("    file: output file (default %s)\n", file);

	return OK;
    }
    while (cp = *vec++) {
	if (strcmp (cp, "-f") == 0) {
	    fast = 1, source = 0, file = NULLCP;
	    continue;
	}
	if (strcmp (cp, "-s") == 0) {
	    source = 1, fast = 0;
	    continue;
	}

	file = cp;
    }

    if (file) {
	char   *sp;

	if (source && strcmp (file, "./objects.defs") == 0)
	    file = "./loadobjects.defs";
	if (sp = rindex (file, '/'))
	    sp++;
	if (sp == NULL || *sp == NULL)
	    sp = file;
	sp += strlen (sp) - 5;
	if (sp > file && strcmp (sp, ".defs") == 0)
	    (void) sprintf (sysout, "%.*s.%s", sp - file, file,
			    source ? "c" : "defs");
	else
	    (void) sprintf (sysout, "%s.%s", file,
			    source ? "c" : "defs");

	if ((fp = fopen (sysout, "w")) == NULL) {
	    advise (sysout, "unable to write");
	    return OK;
	}
    }
    else
	fp = stdout;

    i = j = k = 0;
    if (source) {
	int	v;

	fprintf (fp, "/* compiled */\n\n");
	fprintf (fp, "#include <stdio.h>\n#include \"objects.h\"\n\n");

	fprintf (fp, "static unsigned int _elems[] = {\n");
	v = 0;
	for (ot = anchor; ot; ot = ot -> ot_next) {
	    register unsigned int *ip = ot -> ot_name -> oid_elements;

	    fprintf (fp, "     ");
	    for (j = ot -> ot_name -> oid_nelem; j > 0; j--)
		fprintf (fp, " %u,", *ip++);
	    fprintf (fp, " 0, /* %s */ \n", ot -> ot_text);

	    ot -> ot_views = v++;
	}
	fprintf (fp, "};\n\n");

	fprintf (fp, "static OIDentifier _names[] = {\n");
	for (ot = anchor; ot; ot = ot -> ot_next) {
	    fprintf (fp, "    { %d, &_elems[%d] }, \n",
		     ot -> ot_name -> oid_nelem, k);
	    k += ot -> ot_name -> oid_nelem + 1;
	}
	fprintf (fp, "};\n\n");

	flsyntax (&first, &last);
	fprintf (fp, "static struct _syntax {\n    char   *name;\n");
	fprintf (fp, "    OS      value;\n    char   *textual;\n");
	fprintf (fp, "    char   *display;\n}\t_syntaxes[] = {\n");
	for (os = first; os <= last; os++) {
	    fprintf (fp, "    { \"%s\", NULLOS,", os -> os_name);
	    fprintf (fp, os -> os_textual ? " \"%s\"," : " NULL,",
		     os -> os_textual);
	    fprintf (fp, os -> os_display ? " \"%s\"" : " NULL",
		     os -> os_display);
	    fprintf (fp, " }, \n");
	}
	fprintf (fp, "    NULL\n};\n\n");
    }
    else
	if (!fast) {
	    for (ot = anchor; ot; ot = ot -> ot_next) {
		i++;
		j += strlen (ot -> ot_text)
		        + strlen (sprintoid (ot -> ot_name));
		k += ot -> ot_name -> oid_nelem;
	    }
	    j += i << 1, k += i;
	    fprintf (fp, "--* compiled %d %d %d\n", i, j, k);

	    flsyntax (&first, &last);
	    for (os = first; os <= last; os++)
		if (os -> os_textual)
		    fprintf (fp, "%%%-19s %-16s %-15s \"%s\"\n", "tc",
			     os -> os_name, os -> os_textual,
			     os -> os_display ? os -> os_display : "");
	}

    if (source) {
	i = 0;

	fprintf (fp, "static object_type _types[] = {\n");
    }
    for (ot = anchor; ot; ot = ot -> ot_next) {
	if (fast) {
	    fprintf (fp, "%s=%s\n", ot -> ot_text, sprintoid (ot -> ot_name));
	    continue;
	}

	if (source) {
	    fprintf (fp, "    { \"%s\", \"%s\", &_names[%d],\n      (OS) ",
		     ot -> ot_text, sprintoid (ot -> ot_name), i++);
	    if (ot -> ot_syntax)
		fprintf (fp, "%d", ot -> ot_syntax - first);
	    else
		fprintf (fp, "-1");
	    fprintf (fp, ", %d, (u_long) 0, %d,\n",
		     ot -> ot_access & OT_RDWRITE,
		     ot -> ot_status);
	    fprintf (fp,
		     "      (caddr_t) 0, NULLIFP, NULLIFP, (caddr_t) 0,\n");
	    fprintf (fp, "      ");
#define	OT_OFFSET(x,y) \
	    if (ot -> x) \
		fprintf (fp, "&_types[%d]%s", ot -> x -> ot_views, y); \
	    else \
		fprintf (fp, "NULLOT%s", y)

	    OT_OFFSET (ot_chain, ", ");
	    OT_OFFSET (ot_sibling, ", ");
	    OT_OFFSET (ot_children, ", ");
	    OT_OFFSET (ot_next, " },\n");

#undef	OT_OFFSET

	    continue;
	}

	fprintf (fp, "%-20s %-20s", ot -> ot_text, sprintoid (ot -> ot_name));
	if (os = ot -> ot_syntax)
	    fprintf (fp, " %-15s %-15s %s",
		     os ? os -> os_name : "Aggregate",
		     access_t[ot -> ot_access & OT_RDWRITE],
		     status_t[ot -> ot_status & OT_DEPRECATED]);
	fprintf (fp, "\n");
    }
    if (source) {
	fprintf (fp, "    NULL\n};\n\n");
	fprintf (fp, "OT\tanchor = _types;\n");
	fprintf (fp, "OT\tchain = &_types[%d];\n\n", i - 1);

	flobjects (fp);
    }

    if (file)
	(void) fclose (fp);

    if (!fast)
	advise (NULLCP, "%d definitions written to %s", i,
		file ? sysout : "stdout");

    return OK;
}

/* */

static int  f_help (vec)
char  **vec;
{
    register int    i,
                    j,
                    w;
    int     columns,
            width,
            lines;
    register struct dispatch   *ds,
                               *es;

    for (es = dispatches; es -> ds_name; es++)
	continue;
    width = helpwidth;

    if (*++vec == NULL) {
	if ((columns = ncols (stdout) / (width = (width + 8) & ~7)) == 0)
	    columns = 1;
	lines = ((es - dispatches) + columns - 1) / columns;

	printf ("Operations:\n");
	for (i = 0; i < lines; i++)
	    for (j = 0; j < columns; j++) {
		ds = dispatches + j * lines + i;
		printf ("%s", ds -> ds_name);
		if (ds + lines >= es) {
		    printf ("\n");
		    break;
		}
		for (w = strlen (ds -> ds_name); w < width; w = (w + 8) & ~7)
		    (void) putchar ('\t');
	    }
	printf ("\n");

	return OK;
    }

    if (strcmp (*vec, "-help") == 0) {
	printf ("help [commands ...]\n");
	printf ("    with no arguments, lists operations which may be invoked\n");
	printf ("    otherwise prints help for each operation given\n");

	return OK;
    }
    
    for (; *vec; vec++)
	if (strcmp (*vec, "?") == 0) {
	    for (ds = dispatches; ds -> ds_name; ds++)
		printf ("%-*s\t- %s\n", width, ds -> ds_name, ds -> ds_help);

	    break;
	}
	else
	    if (ds = getds (*vec))
		printf ("%-*s\t- %s\n", width, ds -> ds_name, ds -> ds_help);

    return OK;
}

/* */

static int  f_quit (vec)
char  **vec;
{
    int	    nosync = 0;

    if (vec && *++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("quit [-nosync]\n");
	printf ("    terminate %s\n", myname);

	return OK;
    }

    if (vec)
	while (*vec) {
	    if (strcmp (*vec, "-nosync") == 0) {
		nosync = 1;
		vec++;
		continue;
	    }
	    break;
	}

    if (!nosync && (party_changed || context_changed || access_changed)) {
	char    file[BUFSIZ];

	(void) sprintf (file, "%spublic", pParty);
	if (access (file, 2) != NOTOK)
	    (void) f_sync (NULLVP);
    }

    return DONE;
}

/* */

static int  f_adjust_clock (vec)
char  **vec;
{
    int	    narg;
    char   *args[4];
    register struct snmp_party *p;
    struct snmp_party *agent,
		      *manager;
    struct snmp_context *resources;
    struct snmp_access *sa;

    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("adjust\n");
	printf ("    adjust clocks\n");

	return OK;
    }

    narg = 0;
    args[narg++] = "bulk";
    args[narg++] = "-alg3";
    args[narg++] = "partyAuthClock";
    args[narg] = NULL;
    clocking = 1;

    agent = agentParty;
    manager = managerParty;
    resources = mgmtContext;

    if (agent -> party_authProt != NOAUTH) {
	for (p = partyHead; p; p = p -> party_next)
	    if (p -> party_location == AGENT
		    && p -> party_authProt == NOAUTH
		    && p -> party_TAddress_len
			    == agentParty -> party_TAddress_len
		    && bcmp ((char *) p -> party_TAddress,
			     (char *) agentParty -> party_TAddress,
			     p -> party_TAddress_len) == 0)
		break;
	if (p == NULL) {
	    advise (NULLCP, "unable to find an unauthenticated agent party");
	    goto out;
	}
	agentParty = p;
    }

    if (manager -> party_authProt != NOAUTH) {
	for (p = partyHead; p; p = p -> party_next)
	    if (p -> party_location == MANAGER
		    && p -> party_authProt == agentParty -> party_authProt
		    && p -> party_tDomain == agentParty -> party_tDomain)
		break;
	if (p == NULL) {
	    advise (NULLCP, "unable to find an unauthenticated manager party");
	    goto out;
	}
	managerParty = p;
    }

    if (accessHead)
	for (sa = agentParty -> party_access; sa; )
	    if (sa -> sa_status == ROW_ACTIVE
		    && sa -> sa_subject == managerParty) {
		mgmtContext = sa -> sa_resources;
		break;
	    }
	    else
		if ((sa = sa -> sa_next) && sa -> sa_target != agentParty)
		    break;

    for (p = partyHead; p; p = p -> party_next) {
	p -> party_old_AuthStart = p -> party_AuthStart;
	p -> party_new_AuthStart = 0;
    }

    printf ("Pass 1: Examine...\n");
    (void) f_bulk (args);
    printf ("\n");

    agentParty = agent;
    managerParty = manager;
    mgmtContext = resources;

    if (agent -> party_authProt == NOAUTH) {
	for (p = partyHead; p; p = p -> party_next)
	    if (p -> party_location == AGENT
		    && p -> party_authProt != NOAUTH
		    && p -> party_TAddress_len
			    == agentParty -> party_TAddress_len
		    && bcmp ((char *) p -> party_TAddress,
			     (char *) agentParty -> party_TAddress,
			     p -> party_TAddress_len) == 0)
		break;
	if (p == NULL) {
	    advise (NULLCP, "unable to find an authenticated agent party");
	    goto out;
	}
	agentParty = p;
    }

    if (manager -> party_authProt == NOAUTH) {
	for (p = partyHead; p; p = p -> party_next)
	    if (p -> party_location == MANAGER
		    && p -> party_authProt == agentParty -> party_authProt
		    && p -> party_privProt == agentParty -> party_privProt
		    && p -> party_tDomain == agentParty -> party_tDomain)
		break;
	if (p == NULL) {
	    advise (NULLCP, "unable to find an authenticated manager party");
	    goto out;
	}
	managerParty = p;
    }

    if (accessHead)
	for (sa = agentParty -> party_access; sa; )
	    if (sa -> sa_status == ROW_ACTIVE
		    && sa -> sa_subject == managerParty) {
		mgmtContext = sa -> sa_resources;
		break;
	    }
	    else
		if ((sa = sa -> sa_next) && sa -> sa_target != agentParty)
		    break;

    for (p = partyHead; p; p = p -> party_next)
	if (p -> party_new_AuthStart)
	    break;
    if (!p)
	goto out;

    for (p = partyHead; p; p = p -> party_next)
	if (p -> party_new_AuthStart
	        && p -> party_location == MANAGER
	        && p -> party_AuthStart > p -> party_new_AuthStart) {
								/* case #4 */

	    p -> party_AuthStart = p -> party_new_AuthStart;
	    party_changed = 1;
	    printf ("party %s clock changed to %ld\n",
		    oid2ode (p -> party_Identity),
		    party_time_now () - p -> party_AuthStart);
	}

    for (p = partyHead; p; p = p -> party_next)
	p -> party_new_AuthStart = 0;

    printf ("\nPass 2: Verify...\n");
    (void) f_bulk (args);
    printf ("\n");


    narg = 0;
    for (p = partyHead; p; p = p -> party_next)
	if (p -> party_new_AuthStart) {
	    narg = 1;
	    break;
	}
    if (narg)
	(void) f_sync (NULLVP);
    else {
	advise (NULLCP, "reverting clocks back to original values...");
	for (p = partyHead; p; p = p -> party_next)
	    p -> party_AuthStart = p -> party_old_AuthStart;
    }

out: ;
    agentParty = agent;
    managerParty = manager;
    mgmtContext = resources;
    clocking = 0;

    return OK;
}

/* */

static	int	doing_f_secrets = 0;


static int  f_secrets (vec)
char  **vec;
{
    u_long  now;
    char   *auth,
	   *priv;
    register struct snmp_party *p;
    struct snmp_party *agent,
		      *manager,
		      *myParty;
    struct snmp_context *resources;
    struct snmp_access *sa;
    struct type_SNMP_SnmpAuthMsg *authmsg;

    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("secrets party auth-secret [priv-secret]\n");
	printf ("    change secrets\n");

	return OK;
    }
    if (*vec == NULL)
	return OK;

    if ((myParty = find_party (*vec++, EITHER, 0)) == NULL) {
	print_parties (EITHER);

	return OK;
    }
    if (*vec == NULL)
	return OK;

    auth = *vec++;
    if (myParty -> party_authProt == NOAUTH) {
	advise (NULLCP, "party %s is unauthenticated",
		oid2ode (myParty -> party_Identity));
	return OK;
    }

    if ((priv = *vec ? *vec++ : NULL) && myParty -> party_privProt == NOPRIV) {
	advise (NULLCP, "party %s is not privacy-enabled",
		oid2ode (myParty -> party_Identity));
	return OK;
    }

    {
	int     narg;
	char    buffer1[BUFSIZ],
	        buffer2[BUFSIZ],
	        buffer3[BUFSIZ],
	       *args[5];
	register struct type_SNMP_VarBindList *vp;

	narg = 0;
	args[narg++] = "secret";
	(void) sprintf (buffer1, "partyAuthPrivate.%d.%s=%s",
			myParty -> party_Identity -> oid_nelem,
			sprintoid (myParty -> party_Identity), auth);
	args[narg++] = buffer1;
	if (priv) {
	    (void) sprintf (buffer2, "partyPrivPrivate.%d.%s=%s",
			    myParty -> party_Identity -> oid_nelem,
			    sprintoid (myParty -> party_Identity), priv);
	    args[narg++] = buffer2;
	}
	(void) sprintf (buffer3, "partyAuthClock.%d.%s=0",
			    myParty -> party_Identity -> oid_nelem,
			    sprintoid (myParty -> party_Identity));
	args[narg++] = buffer3;
	args[narg] = NULL;

	if ((authmsg = new_message (type_SNMP_PDUs_set__request, args))
	        == NULL)
	    return OK;
	vp = authmsg -> authData -> pdu -> un.set__request
	    			        -> variable__bindings;
	if (secretize (vp -> VarBind,
		       myParty -> party_AuthPrivate,
		       myParty -> party_AuthPrivate_len,
		       &myParty -> party_new_AuthPublic,
		       &myParty -> party_new_AuthPublic_len) == NOTOK)
	    goto out;
	if (priv
	        && secretize (vp -> next -> VarBind,
			      myParty -> party_PrivPrivate,
			      myParty -> party_PrivPrivate_len,
			      &myParty -> party_new_PrivPublic,
			      &myParty -> party_new_PrivPublic_len)
	    		== NOTOK)
	    goto out;
    }

    agent = agentParty;
    manager = managerParty;
    resources = mgmtContext;

    if (agent -> party_privProt == NOPRIV) {
	for (p = partyHead; p; p = p -> party_next)
	    if (p -> party_location == AGENT
		    && p -> party_privProt != NOPRIV
		    && p -> party_TAddress_len
			    == agentParty -> party_TAddress_len
		    && bcmp ((char *) p -> party_TAddress,
			     (char *) agentParty -> party_TAddress,
			     p -> party_TAddress_len) == 0)
		break;
	if (p == NULL) {
	    advise (NULLCP, "unable to find a privacy-enabled agent party");
	    goto out;
	}
	agentParty = p;
    }

    if (manager -> party_privProt == NOPRIV) {
	for (p = partyHead; p; p = p -> party_next)
	    if (p -> party_location == MANAGER
		    && p -> party_authProt == agentParty -> party_authProt
		    && p -> party_privProt == agentParty -> party_privProt
		    && p -> party_tDomain == agentParty -> party_tDomain)
		break;
	if (p == NULL) {
	    advise (NULLCP, "unable to find a privacy-enabled manager party");
	    goto out;
	}
	managerParty = p;
    }

    if (accessHead)
	for (sa = agentParty -> party_access; sa; )
	    if (sa -> sa_status == ROW_ACTIVE
		    && sa -> sa_subject == managerParty) {
		mgmtContext = sa -> sa_resources;
		break;
	    }
	    else
		if ((sa = sa -> sa_next) && sa -> sa_target != agentParty)
		    break;

    now = party_time_now ();

    if (myParty == managerParty && priv) {
	myParty -> party_new_PrivPrivate
				    = myParty -> party_PrivPrivate;
	myParty -> party_new_PrivPrivate_len
				    = myParty -> party_PrivPrivate_len;

	myParty -> party_PrivPrivate = myParty -> party_new_PrivPublic;
	myParty -> party_PrivPrivate_len
					= myParty -> party_new_PrivPublic_len;

	myParty -> party_new_PrivPublic = NULL;
	myParty -> party_new_PrivPublic_len = 0;
    }
    else
	if (myParty == agentParty) {
	    myParty -> party_new_AuthStart = myParty -> party_AuthStart;

	    myParty -> party_AuthStart = now;
	    
	    myParty -> party_new_AuthPrivate
				    = myParty -> party_AuthPrivate;
	    myParty -> party_new_AuthPrivate_len
				    = myParty -> party_AuthPrivate_len;

	    myParty -> party_AuthPrivate
					    = myParty -> party_new_AuthPublic;
	    myParty -> party_AuthPrivate_len
					= myParty -> party_new_AuthPublic_len;

	    myParty -> party_new_AuthPublic = NULL;
	    myParty -> party_new_AuthPublic_len = 0;
	}

    lasterr = -1;

    doing_f_secrets = 1;
    (void) process (authmsg, 1);
    doing_f_secrets = 0;
    authmsg = NULL;

    if (lasterr == int_SNMP_error__status_noError) {
	if (myParty -> party_new_AuthPublic) {
	    if (myParty -> party_AuthPrivate)
		free ((char *) myParty -> party_AuthPrivate);
	    myParty -> party_AuthPrivate
					= myParty -> party_new_AuthPublic;
	    myParty -> party_AuthPrivate_len
					= myParty -> party_new_AuthPublic_len;

	    myParty -> party_new_AuthPublic = NULL;
	    myParty -> party_new_AuthPublic_len = 0;
	}

	if (myParty -> party_new_PrivPublic) {
	    if (myParty -> party_PrivPrivate)
		free ((char *) myParty -> party_PrivPrivate);
	    myParty -> party_PrivPrivate
					= myParty -> party_new_PrivPublic;
	    myParty -> party_PrivPrivate_len
					= myParty -> party_new_PrivPublic_len;

	    myParty -> party_new_PrivPublic = NULL;
	    myParty -> party_new_PrivPublic_len = 0;
	}

	myParty -> party_status = ROW_ACTIVE;

	if (myParty != agentParty)
	    myParty -> party_AuthStart = now;

	party_changed = 1;

	(void) f_sync (NULLVP);
    }
    else {
	advise (NULLCP, "reverting secrets back to original values...\n\
\tThere is a possibility that the set was OK, but the response got lost.\n\
\tIn that case, you may have to edit the %sprivate file and replace\n\
\tthe secrets there with the new ones you just supplied.\n\
\tWhen connectivity is restored, you should try using the parties.\n\
\tIf they don't work, then try editing the %sprivate file.", pParty, pParty);

	if (myParty == managerParty && priv) {
	    free ((char *) myParty -> party_PrivPrivate);

	    myParty -> party_PrivPrivate
				    = myParty -> party_new_PrivPrivate;
	    myParty -> party_PrivPrivate_len
				= myParty -> party_new_PrivPrivate_len;

	    myParty -> party_PrivPrivate = NULL;
	    myParty -> party_PrivPrivate_len = 0;
	}
	else
	    if (myParty == agentParty) {
		myParty -> party_AuthStart = myParty -> party_new_AuthStart;

		myParty -> party_new_AuthStart = 0;

		free ((char *) myParty -> party_AuthPrivate);

		myParty -> party_AuthPrivate
				    = myParty -> party_new_AuthPrivate;
		myParty -> party_AuthPrivate_len
				= myParty -> party_new_AuthPrivate_len;

		myParty -> party_new_AuthPrivate = NULL;
		myParty -> party_new_AuthPrivate_len = 0;
	    }
    }

out: ;
    agentParty = agent;
    managerParty = manager;
    mgmtContext = resources;

    if (myParty -> party_new_AuthPublic) {
	free ((char *) myParty -> party_new_AuthPublic);
	myParty -> party_new_AuthPublic = NULL;
	myParty -> party_new_AuthPublic_len = 0;
    }
    if (myParty -> party_new_PrivPublic) {
	free ((char *) myParty -> party_new_PrivPublic);
	myParty -> party_new_PrivPublic = NULL;
	myParty -> party_new_PrivPublic_len = 0;
    }
    if (authmsg)
	free_SNMP_SnmpAuthMsg (authmsg);

    return OK;
}


static  secretize (vb, cp, len, sp, sl)
struct type_SNMP_VarBind *vb;
u_char  *cp,
       **sp;
int	 len,
        *sl;
{
    int	    result;
    register OI	oi;
    register OS	os;
    register OT	ot;
    caddr_t value;

    if (vb -> value -> pe_class == PE_CLASS_CONT) {
	advise (NULLCP, "%s=%s, you lose extraordinarily big",
		oid2ode (vb -> name), snmpV2_exception (vb -> value -> pe_id));
	return NOTOK;
    }
    if ((oi = name2inst (vb -> name)) == NULL
	    || (ot = oi -> oi_type) == NULL
	    || (os = ot -> ot_syntax) == NULL) {
	advise (NULLCP, "name2inst %s, you lose extraordinarily big",
		oi ? (ot ? "lost OS" : "lost OT") : "failed");
	return NOTOK;
    }

    if ((*os -> os_decode) (&value, vb -> value) == NOTOK) {
	advise (NULLCP, "os_decode failed, you lose extraordinarily big");
	return NOTOK;
    }

    {
	struct qbuf *qb = (struct qbuf *) value;

	if ((*sp = (u_char *) qb2str (qb)) == NULL) {
	    advise (NULLCP, "qb2str failed, you lose extraordinarily big");
	    result = NOTOK;
	    goto out;
	}
	*sl = qb -> qb_len;

	if (cp)
	    exor ((u_char *) qb -> qb_forw -> qb_data, qb -> qb_len, cp, len);
    }

    pe_free (vb -> value);
    if ((result = (*os -> os_encode) (value, &vb -> value)) == NOTOK)
	advise (NULLCP, "os_encode failed, you lose extraordinarily big");

out: ;
    (*os -> os_free) (value);

    return result;
}

/* */

static int  f_tune (vec)
char  **vec;
{
    int	    narg;
    char    buffer1[BUFSIZ],
	    buffer2[BUFSIZ],
	    buffer3[BUFSIZ],
	    buffer4[BUFSIZ],
	    buffer5[BUFSIZ],
	    buffer6[BUFSIZ],
	    buffer7[BUFSIZ],
	    buffer8[BUFSIZ],
	   *args[9];
    register struct snmp_party *p;
    struct snmp_party *agent,
		      *manager,
		      *myParty;

    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("tune agent-party\n");
	printf ("    tune agent parameters\n");

	return OK;
    }
    if (*vec == NULL)
	return OK;

    if ((myParty = find_party (*vec++, AGENT, 0)) == NULL) {
	print_parties (AGENT);

	return OK;
    }

    (void) sprintf (buffer1, "%d.%s", myParty -> party_Identity -> oid_nelem,
		    sprintoid (myParty -> party_Identity));
    (void) sprintf (buffer2, "partyTDomain.%s", buffer1);
    (void) sprintf (buffer3, "partyTAddress.%s", buffer1);
    (void) sprintf (buffer4, "partyMaxMessageSize.%s", buffer1);
    (void) sprintf (buffer5, "partyLocal.%s", buffer1);
    (void) sprintf (buffer6, "partyAuthPublic.%s", buffer1);
    (void) sprintf (buffer7, "partyAuthLifetime.%s", buffer1);
    (void) sprintf (buffer8, "partyPrivPublic.%s", buffer1);

    narg = 0;
    args[narg++] = "get";
    args[narg++] = buffer2;
    args[narg++] = buffer3;
    args[narg++] = buffer4;
    args[narg++] = buffer5;
    args[narg++] = buffer6;
    args[narg++] = buffer7;
    args[narg++] = buffer8;
    args[narg] = NULL;

    agent = agentParty;
    manager = managerParty;

    if (agent -> party_authProt == NOAUTH) {
	for (p = partyHead; p; p = p -> party_next)
	    if (p -> party_location == AGENT
		    && p -> party_authProt != NOAUTH
		    && p -> party_TAddress_len
			    == myParty -> party_TAddress_len
		    && bcmp ((char *) p -> party_TAddress,
			     (char *) myParty -> party_TAddress,
			     p -> party_TAddress_len) == 0)
		break;
	if (p == NULL) {
	    advise (NULLCP, "unable to find an authenticated agent party");
	    goto out;
	}
	agentParty = p;
    }

    if (manager -> party_authProt == NOAUTH) {
	for (p = partyHead; p; p = p -> party_next)
	    if (p -> party_location == MANAGER
		    && p -> party_authProt != NOAUTH
		    && p -> party_tDomain == agentParty -> party_tDomain)
		break;
	if (p == NULL) {
	    advise (NULLCP, "unable to find an unauthenticated manager party");
	    goto out;
	}
	managerParty = p;
    }


    lasterr = -1;

    (void) process (new_message (type_SNMP_PDUs_get__request, args), 1);

    if (lasterr == int_SNMP_error__status_noError)
	(void) f_sync (NULLVP);

out: ;
    agentParty = agent;
    managerParty = manager;

    return OK;
}

/* */

static int  sp_compar (a, b)
struct snmp_party **a, **b;
{
    if ((*a) -> party_index <= 6 || (*b) -> party_index <= 6)
	return ((*a) -> party_index - (*b) -> party_index);
	
    return oid_cmp (((*a) -> party_Identity), ((*b) -> party_Identity));
}


static int  sc_compar (a, b)
struct snmp_context **a, **b;
{
    if ((*a) -> sc_index <= 2 || (*b) -> sc_index <= 2)
	return ((*a) -> sc_index - (*b) -> sc_index);
	
    return oid_cmp (((*a) -> sc_identity), ((*b) -> sc_identity));
}


static int  f_sync (vec)
char  **vec;
{
    int	    force = 0,
	    sortit = 1;

    if (vec && *++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("sync [-force] [-nosort]\n");
	printf ("    synchronize party information to \"stable-storage\"\n");

	return OK;
    }

    if (vec)
	while (*vec) {
	    if (strcmp (*vec, "-force") == 0) {
		force = 1;
		vec++;
		continue;
	    }
	    if (strcmp (*vec, "-nosort") == 0) {
		sortit = 0;
		vec++;
		continue;
	    }
	    break;
	}

    if (sortit) {
	int	i;
	register struct snmp_party *p;
	register struct snmp_context *sc;

	i = 0;
	for (p = partyHead; p; p = p -> party_next)
	    i++;

	if (i > 1) {
	    register struct snmp_party **base,
				       **spe,
				       **spp;

	    if ((base = (struct snmp_party **)
			    malloc ((unsigned) (i * sizeof *base))) == NULL)
		goto out;

	    spe = base;
	    for (p = partyHead; p; p = p -> party_next)
		*spe++ = p;

	    qsort ((char *) base, i, sizeof *base, (IFP) sp_compar);

	    spp = base;
	    p = partyHead = *spp++;

	    while (spp < spe) {
		p -> party_next = *spp;
		p = *spp++;
	    }
	    p -> party_next = NULL;

	    free ((char *) base);
	}

	i = 0;
	for (sc = contextHead; sc; sc  = sc -> sc_next)
	    i++;

	if (i > 1) {
	    register struct snmp_context **base,
				         **sce,
				         **scp;

	    if ((base = (struct snmp_context **)
			    malloc ((unsigned) (i * sizeof *base))) == NULL)
		goto out;

	    sce = base;
	    for (sc = contextHead; sc; sc = sc -> sc_next)
		*sce++ = sc;

	    qsort ((char *) base, i, sizeof *base, (IFP) sc_compar);

	    scp = base;
	    sc = contextHead = *scp++;

	    while (scp < sce) {
		sc -> sc_next = *scp;
		sc = *scp++;
	    }
	    sc -> sc_next = NULL;

	    free ((char *) base);
	}

out: ;
    }

    if (sync_party (pParty, MANAGER, force) == NOTOK)
	advise (NULLCP, "stable storage suspect: %s", PY_pepy);

    return OK;
}

/* */

static int  f_agent (vec)
char  **vec;
{
    return change_party (vec, "agent", "agent (destination)", AGENT,
			   &agentParty);
}


static int  f_manager (vec)
char  **vec;
{
    return change_party (vec, "manager", "manager (source)", MANAGER,
			   &managerParty);
}


static int  change_party (vec, cmd, help, who, party)
char  **vec,
       *cmd,
       *help;
int	who;
struct snmp_party **party;
{
    struct snmp_party *myParty;

    if (*++vec != NULL) {
	if (strcmp (*vec, "-help") == 0) {
	    fprintf(stderr, "%s party\n", cmd);
	    fprintf(stderr, "   change the %s party\n", help);

	    return OK;
	}
    }
    else {
	print_parties (who);

	return OK;
    }

    if ((myParty = find_party (*vec, who, 1)) == NULL) {
	print_parties (who);

	return OK;
    }
    
    if (who != AGENT || party_connect (myParty) != NOTOK)
	*party = myParty;

    return OK;
}

/* */

static int  f_print_all_parties (vec)
char **vec;
{
    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("pparties\n");
	printf ("    print out a table of all known parties\n");

	return OK;
    }

    print_parties (EITHER);

    return OK;
}

/* */

static	print_parties (who)
int	who;
{
    register struct snmp_party *p;

    printf ("No.   Identity\t\t\t\tAuthentication\t      Privacy\n");
    for (p = partyHead; p; p = p -> party_next)
	if (who == EITHER
	        || (p -> party_tDomain == SNMPUDPDOMAIN
		        && who == p -> party_location))
	    print_party (p);
}


static	print_party (party)
register struct snmp_party *party;
{
    char   *cp;

    printf ("%3d%c%c%c%s\t", party -> party_index,
	    party == agentParty || party == managerParty ? '*' : ' ',
	    party -> party_tDomain != SNMPUDPDOMAIN ? ' '
		: party -> party_location == AGENT ? 'A' : 'M',
	    party -> party_status != ROW_ACTIVE ? '!' : ' ',
	    cp = oid2ode (party -> party_Identity));
    if (strlen (cp) < 26)
	printf("\t");
    printf ("%-21s ", cp =  oid2ode (party -> party_AuthProtocol));
    printf ("%s\n", oid2ode (party -> party_PrivProtocol));

    if (party -> party_location == AGENT)
	printf("      residing at %s\n", cp = party_address (party));
}

/* */

static int  f_context (vec)
char  **vec;
{
    struct snmp_context *sc;

    if (*++vec != NULL) {
	if (strcmp (*vec, "-help") == 0) {
	    fprintf(stderr, "context name\n");
	    fprintf(stderr, "   change the management context\n");

	    return OK;
	}
    }
    else {
print_contexts: ;
	printf ("No.   Identity\t\t\t\tInformation\n");
	for (sc = contextHead; sc; sc = sc -> sc_next)
	    if (sc -> sc_location == AGENT)
		print_context (sc, 0);

	return OK;
    }

    if ((sc = find_context (*vec)) == NULL)
	goto print_contexts;
	
    mgmtContext = sc;

    return OK;
}

/* */

static	print_context (sc, doC)
register struct snmp_context *sc;
int	doC;
{
    char   *cp;

    printf ("%3d%c%c%c%s\t", sc -> sc_index,
	    sc == mgmtContext ? '*' : ' ',
	    doC ? 'C' : ' ',
	    sc -> sc_status != ROW_ACTIVE ? '!' : ' ',
	    cp = oid2ode (sc -> sc_identity));
    if (strlen (cp) < 26)
	printf ("\t");
    printf ("%s\n", u_char2str (sc -> sc_name, sc -> sc_name_len));
    if (sc -> sc_view > 0) {
	printf ("\tview  %d", sc -> sc_view);
	if (sc -> sc_entity_len > 0)
	    printf (", entity %s",
		    u_char2str (sc -> sc_entity, sc -> sc_entity_len));
	if (oid_cmp (sc -> sc_time, currentTime))
	    printf (", time %s", oid2ode (sc -> sc_time));
    }
    else {
	printf ("\tproxy");
	if (sc -> sc_dst_oid) {
	    printf (" %s", oid2ode (sc -> sc_dst_oid));
	    if (sc -> sc_src_oid)
		printf ("\n\t  src %s", oid2ode (sc -> sc_src_oid));
	    if (sc -> sc_ctx)
		printf ("\n\t  ctx %s", oid2ode (sc -> sc_ctx));
	}
    }
    printf ("\n");
}

/* */

static int  f_access (vec)
char  **vec;
{
    register struct snmp_access *sa;

    if (*++vec != NULL) {
	fprintf(stderr, "access\n");
	fprintf(stderr, "   print out access control policy\n");

	return OK;
    }

    printf (" Dst Src Ctx Privs\n");
    for (sa = accessHead; sa; sa = sa -> sa_next)
	printf ("%c%3d %3d %3d %s\n",
		sa -> sa_status != ROW_ACTIVE ? '!' : ' ',
		sa -> sa_target_index, sa -> sa_subject_index,
		sa -> sa_resources_index,
		sprintb (sa -> sa_privileges,
			 "\020\01get\02next\03response\04set\05v1trap\06bulk\07inform\010trap"));

    return OK;
}

/* */

static int  f_status (vec)
char  **vec;
{
    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("status\n");
	printf ("    report status\n");

	return OK;
    }

    if (agentParty)
	print_party (agentParty);
    else
	printf ("No agent (destination) specified\n");
    if (managerParty)
	print_party (managerParty);
    else
	printf ("No manager (source) specified\n");
    if (mgmtContext)
	print_context (mgmtContext, 1);
    else
	printf ("No context specified\n");

    return OK;
}

/* */

static int  f_retries (vec)
char  **vec;
{
    int	    newretries;

    if ((*++vec == NULL) || (*vec != NULL && strcmp (*vec, "-help") == 0)) {
        printf ("retries value\n");
        printf ("    set the number of retries for each operation\n");
        printf ("    (currently %d)\n", retries);
        return OK;
    }

    if (sscanf (*vec, "%d", &newretries) < 1 || newretries < 1) {
	printf ("value must be a positive number\n");
	return OK;
    }

    retries = newretries;

    return OK;
}

/* */

static int  f_timeout (vec)
char  **vec;
{
    int	    newtimeout;

    if ((*++vec == NULL) || (*vec != NULL && strcmp (*vec, "-help") == 0)) {
        printf ("timeout value\n");
        printf ("    set the number of seconds between retries\n");
        printf ("    (currently %d)\n", timeout);
        return OK;
    }

    if (sscanf (*vec, "%d", &newtimeout) < 1 || newtimeout < 1) {
	printf ("value must be a positive number of seconds\n");
	return OK;
    }

    timeout = newtimeout;

    return OK;
}

/* */

static int  f_watch (vec)
char  **vec;
{
    if (*++vec != NULL && strcmp (*vec, "-help") == 0) {
	printf ("watch\n");
	printf ("    toggle watching (currently %s)\n",
		watch ? "on" : "off");

	return OK;
    }

    watch = !watch;
    printf ("Turning %s watching\n", watch ? "on" : "off");

    return OK;
}

/*   SYNTAX */

static char *ifType[] = {
    "other", "regular1822", "hdh1822", "ddn-x25", "rfc877-x25",
    "ethernet-csmacd", "iso88023-csmacd", "iso88024-tokenBus",
    "iso88025-tokenRing", "iso88026-man", "starLan", "proteon-10Mbit",
    "proteon-80Mbit", "hyperchannel", "fddi", "lapb", "sdlc", "ds1",
     "e1", "basicISDN", "primaryISDN", "propPointToPointSerial",
    "ppp", "softwareLoopback", "eon", "ethernet-3Mbit",
    "nsip", "slip", "ultra", "ds3", "sip", "frame-relay"
};

static char *ifStatus[] = {
    "up", "down", "testing"
};

static char *ipForwarding[] = {
    "gateway", "host"
};

static char *routeType[] = {
    "other", "invalid", "direct", "remote"
};

static char *ipRouteProto[] = {
    "other", "local", "netmgmt", "icmp", "egp", "ggp", "hello", "rip", "is-is",
    "es-is", "ciscoIgrp", "bbnSpfIgp", "ospf", "bgp"
};

static char *netToMediaType[] = {
    "other", "invalid", "dynamic", "static"
};

static char *tcpRtoAlgorithm[] = {
    "other", "constant", "rsre", "vanj"
};

static char *tcpConnState[] = {
    "closed", "listen", "synSent", "synReceived", "established", "finWait1",
    "finWait2", "closeWait", "lastAck", "closing", "timewait"
};

static char *identStatus[] = {
    "noError", "unknownError"
};

static char *egpNeighState[] = {
    "idle", "acquisition", "down", "up", "cease"
};

static char *egpNeighMode[] = {
    "active", "passive"
};

static char *egpNeighEventTrigger[] = {
    "start", "stop"
};

static char *enabled[] = {
    "enabled", "disabled"
};

static char *truthValue[] = {
    "true", "false"
};

static char *rowStatus[] = {
    "active", "notInService", "notReady", "createAndGo", "createAndWait",
    "destroy"
};

static char *storageType[] = {
    "other", "volatile", "nonVolatile", "permanent"
};

static char *viewType[] = {
    "included", "excluded"
};

static char *agentAction[] = {
    "other", "coldStart", "dumpPartyDB", "terminate", "warmStart", "debugOn",
    "debugOff"
};


static struct ivar {
    char   *iv_object;

    char  **iv_values;
    int	    iv_nvalue;
} ivars[] = {
    "ifType", ifType,
	sizeof ifType / sizeof ifType[0],
    
    "ifAdminStatus", ifStatus,
	sizeof ifStatus / sizeof ifStatus[0],
    
    "ifOperStatus", ifStatus,
	sizeof ifStatus / sizeof ifStatus[0],

    "ipForwarding", ipForwarding,
	sizeof ipForwarding / sizeof ipForwarding[0],

    "ipRouteType", routeType,
	sizeof routeType / sizeof routeType[0],

    "ipRouteProto", ipRouteProto,
	sizeof ipRouteProto / sizeof ipRouteProto[0],

    "ipNetToMediaType", netToMediaType,
	sizeof netToMediaType / sizeof netToMediaType[0],

    "tcpRtoAlgorithm", tcpRtoAlgorithm,
	sizeof tcpRtoAlgorithm / sizeof tcpRtoAlgorithm[0],

    "tcpConnState", tcpConnState,
	sizeof tcpConnState / sizeof tcpConnState[0],

    "identStatus", identStatus,
	sizeof identStatus / sizeof identStatus[0],

    "egpNeighState", egpNeighState,
	sizeof egpNeighState / sizeof egpNeighState[0],

    "egpNeighMode", egpNeighMode,
	sizeof egpNeighMode / sizeof egpNeighMode[0],

    "egpNeighEventTrigger", egpNeighEventTrigger,
	sizeof egpNeighEventTrigger / sizeof egpNeighEventTrigger[0],

    "snmpEnableAuthenTraps", enabled,
	sizeof enabled / sizeof enabled[0],

    "unixNetstat", enabled,
	sizeof enabled / sizeof enabled[0],

    "snmpV2EnableAuthenTraps", truthValue,
	sizeof truthValue / sizeof truthValue[0],
    
    "partyLocal", truthValue,
	sizeof truthValue / sizeof truthValue[0],
    
    "partyStorageType", storageType,
	sizeof storageType / sizeof storageType[0],

    "partyStatus", rowStatus,
	sizeof rowStatus / sizeof rowStatus[0],

    "contextLocal", truthValue,
	sizeof truthValue / sizeof truthValue[0],
    
    "contextStorageType", storageType,
	sizeof storageType / sizeof storageType[0],

    "contextStatus", rowStatus,
	sizeof rowStatus / sizeof rowStatus[0],

    "aclStorageType", storageType,
	sizeof storageType / sizeof storageType[0],

    "aclStatus", rowStatus,
	sizeof rowStatus / sizeof rowStatus[0],

    "viewType", viewType,
	sizeof viewType / sizeof viewType[0],

    "viewStorageType", storageType,
	sizeof storageType / sizeof storageType[0],

    "viewStatus", rowStatus,
	sizeof rowStatus / sizeof rowStatus[0],

    "agentAction", agentAction,
	sizeof agentAction / sizeof agentAction[0],

    NULL
};

/* */

static int  enum_parse (x, s, os)
integer **x;
char   *s;
OS	os;
{
    register char  **ap,
		   **ep;
    long    l;

    for (ep = (ap = os -> os_data1) + os -> os_data2; ap < ep; ap++)
	if (lexequ (s, *ap) == 0) {
	    l = (ap - os -> os_data1) + 1;
	    goto do_assign;
	}

    if (sscanf (s, "%ld", &l) != 1)
	return NOTOK;

do_assign: ;
    if ((*x = (integer *) malloc (sizeof **x)) == NULL)
	return NOTOK;
    **x = (integer) l;

    return OK;
}


static int  enum_print (x, os)
integer *x;
OS	os;
{
    int	    i = *x;

    if (i <= 0 || i > os -> os_data2)
	printf ("unknown(%d)", i);
    else
	printf ("%s(%d)", os -> os_data1[i - 1], i);
}


static	moresyntax (check)
int	check;
{
    register struct ivar *iv;
    register OT	   ot;
    register OS	   os;

    for (iv = ivars; iv -> iv_object; iv++)
	if (ot = text2obj (iv -> iv_object)) {
	    char   *name;

	    if ((os = ot -> ot_syntax) == NULL) {
		advise (NULLCP, "no syntax defined for object \"%s\"",
			iv -> iv_object);
		continue;
	    }
	    name = os -> os_name;

	    (void) add_syntax (iv -> iv_object, os -> os_encode,
			       os -> os_decode, os -> os_free, enum_parse,
			       enum_print);
	    if ((os = text2syn (iv -> iv_object)) == NULL)
		adios (NULLCP, "lost syntax for object \"%s\"",
		       iv -> iv_object);
	    ot -> ot_syntax = os;
	    os -> os_name = name;
	    os -> os_data1 = iv -> iv_values;
	    os -> os_data2 = iv -> iv_nvalue;
	}
	else
	    if (check)
		advise (NULLCP, "no \"%s\" object", iv -> iv_object);

    if (os = text2syn ("ObjectID"))
	os -> os_parse = object_parse, os -> os_print = object_print;

    if (ot = text2obj ("viewSubtree"))
	ot -> ot_getfnx = (IFP) 1;
    if (ot = text2obj ("viewMask"))
	ot -> ot_getfnx = (IFP) 1;
    if (ot = text2obj ("viewType"))
	ot -> ot_getfnx = (IFP) 1;
    if (ot = text2obj ("viewStorageType"))
	ot -> ot_getfnx = (IFP) 1;
    if (ot = text2obj ("viewStatus"))
	ot -> ot_getfnx = (IFP) 1;
}

/*   MISCELLANY */

static	arginit (vec)
char    **vec;
{
    int	    w;
    register char  *ap,
		   *pp;
    char   *aParty,
	   *mParty,
	   *tContext;
    register struct dispatch *ds;

    if (myname = rindex (*vec, '/'))
	myname++;
    if (myname == NULL || *myname == NULL)
	myname = *vec;

/*    _tailor_logging_only = 1;*/
    isodetailor (myname, 1);

    if (ontty = isatty (fileno (stdin)))
	verbose++;

    aParty = mParty = tContext = NULL;
    for (vec++; ap = *vec; vec++) {
	if (*ap == '-') {
	    while (*++ap)
		switch (*ap) {
		    case 'a':
		        if ((aParty = *++vec) == NULL || *aParty == '-')
			    adios (NULLCP, "usage: %s -a agentParty", myname);
			break;

		    case 'c':
		        if ((tContext = *++vec) == NULL || *tContext == '-')
			    adios (NULLCP, "usage: %s -c mgmtContext", myname);
			break;

		    case 'd':
		        debug++;
			break;

		    case 'f':
			if ((pp = *++vec) == NULL || *pp == '-')
			    adios (NULLCP, "usage: %s -f objects-file",
				   myname);
			defs = pp;
			break;
			
		    case 'm':
		        if ((mParty = *++vec) == NULL || *mParty == '-')
			    adios (NULLCP, "usage: %s -m managerParty",
				   myname);
			break;

		    case 'p':
		        if ((pParty = *++vec) == NULL || *pParty == '-')
			    adios (NULLCP, "usage: %s -p party-file-prefix",
				   myname);
			break;

		    case 'v':
		        verbose++;
			break;

		    case 'w':
		        watch++;
			break;

		    case 'Z':
			{
			    struct type_SNMP_SnmpPrivMsg *privmsg;
			    PE	    stdpe;
			    PS	    stdps;

			    if ((stdps = ps_alloc (std_open)) == NULLPS)
				adios (NULLCP, "ps_alloc(stdin): you lose");
			    if (std_setup (stdps, stdin) == NOTOK)
				adios (NULLCP,
				       "std_setup(stdin): you lose [%s]",
				       ps_error (stdps -> ps_errno));
			    if ((stdpe = ps2pe (stdps)) == NULLPE)
				adios (NULLCP,
				       "ps2pe(stdin): you lose [%s]",
				       ps_error (stdps -> ps_errno));
			    if (decode_SNMP_SnmpPrivMsg (stdpe, 1, NULLIP,
							 NULLVP, &privmsg)
				    == NOTOK)
				adios (NULLCP, "decode_SNMP_SnmpPrivMsg: %s",
				       PY_pepy);
			    (void) print_SNMP_SnmpPrivMsg (stdpe, 1, NULLIP,
							   NULLVP, NULLCP);
			}
			exit (0);

		    default:
			adios (NULLCP, "unknown switch -%c", *ap);
		}
	    continue;
	}
	if (op == NULL) {
	    op = vec;
	    break;
	}
    }

    helpwidth = 0;
    for (ds = dispatches; ds -> ds_name; ds++)
	if ((w = strlen (ds -> ds_name)) > helpwidth)
	    helpwidth = w;

#ifdef	SYS5
    (void) srand ((unsigned int) time ((long *) 0));
#else
    (void) srandom ((int) time ((long *) 0));
#endif

    ps_len_strategy = PS_LEN_LONG;

    if (loadobjects (defs) == NOTOK)
	adios (NULLCP, "loadobjects: %s", PY_pepy);
    if (defs && (ap = rindex (defs, '/')))
	ap++;
    else
	ap = defs ? defs : "";

    if (!op || (strcmp (*op, "compile") && strcmp (*op, "quit"))) {
	moresyntax (strcmp (ap, "objects.defs") == 0);

	if (init_party (pParty, MANAGER, 0, &pParty) == NOTOK)
	    adios (NULLCP, "%s", PY_pepy);
	remember_party ();
    }

    if (aParty) {
	if ((agentParty = find_party (aParty, AGENT, 1)) == NULL)
	    exit (1);
	if (party_connect (agentParty) == NOTOK)
	    agentParty = NULL;
    }
    else
	for (agentParty = partyHead;
	         agentParty;
	         agentParty = agentParty -> party_next)
	    if (agentParty -> party_status == ROW_ACTIVE
		    && agentParty -> party_location == AGENT
		    && agentParty -> party_tDomain == SNMPUDPDOMAIN
		    && agentParty -> party_authProt == NOAUTH
		    && agentParty -> party_privProt == NOPRIV
		    && party_connect (agentParty) != NOTOK)
		break;

    if (mParty) {
	if ((managerParty = find_party (mParty, MANAGER, 1)) == NULL)
	    exit (1);
    }
    else
	if (agentParty)
	    for (managerParty = partyHead;
	             managerParty;
	             managerParty = managerParty -> party_next)
		if (managerParty -> party_status == ROW_ACTIVE
		        && managerParty -> party_location == MANAGER
		        && managerParty -> party_tDomain == SNMPUDPDOMAIN
		        && managerParty -> party_authProt
				== agentParty -> party_authProt
		        && managerParty -> party_privProt
				== agentParty -> party_privProt
		        && managerParty -> party_tDomain
				== agentParty -> party_tDomain)
		    break;

    if (tContext) {
	if ((mgmtContext = find_context (tContext)) == NULL)
	    exit (1);
    }
    else
	if (agentParty && managerParty && accessHead) {
	    register struct snmp_access *sa;

	    for (sa = agentParty -> party_access; sa; )
		if (sa -> sa_status == ROW_ACTIVE
		        && sa -> sa_subject == managerParty) {
		    mgmtContext = sa -> sa_resources;
		    break;
		}
		else
		    if ((sa = sa -> sa_next) && sa -> sa_target != agentParty)
			break;
	}
	else
	    for (mgmtContext = contextHead;
	             mgmtContext;
	             mgmtContext = mgmtContext -> sc_next)
		if (mgmtContext -> sc_status == ROW_ACTIVE
		        && mgmtContext -> sc_location == AGENT)
		    break;

    if ((nullSpecific = text2oid ("0.0")) == NULLOID)
	adios (NULLCP, "text2oid (\"0.0\") failed!");
    if ((unSpecified = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM, PE_PRIM_NULL))
		    == NULL)
	adios (NULLCP, "out of memory");
}

/* */

SnmpRMIBerrorEnum SnmpRMIBpartyConnect( snmpRMIBAgentDatap )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
{
struct	snmp_party * agent = snmpRMIBAgentDatap -> agentParty ;
#ifdef	TCP
    char   *pp;
    struct sockaddr_in in_socket,
		       lo_socket;
    static struct sockaddr_in my_socket;
    register struct sockaddr_in *isock = &in_socket,
				*lsock;

    if (agent -> party_tDomain != SNMPUDPDOMAIN
	    && agent -> party_tDomain != RFC1157DOMAIN) {

	advise (NULLCP, "party %s is not reachable over UDP",
		oid2ode (agent -> party_Identity));

	return SnmpRMIBinvalidTransDomain ;
    }

    if(agent -> party_TAddress[0] == 127
	    && agent -> party_TAddress[1] == 0
	    && agent -> party_TAddress[2] == 0
	    && agent -> party_TAddress[3] == 1) {
	lsock = &lo_socket;

	bzero ((char *) lsock, sizeof *lsock);
	lsock -> sin_family = AF_INET;
	bcopy ((char *) agent -> party_TAddress, (char *) &lsock -> sin_addr,
	       sizeof lsock -> sin_addr);
    }
    else {
	static int once_only = 1;

	lsock = &my_socket;

	if (once_only) {
	    register struct hostent *hp;

	    once_only = 0;

	    if ((hp = gethostbystring (pp = getlocalhost ())) == NULL)
		adios (NULLCP, "%s: unknown host", pp);

	    bzero ((char *) lsock, sizeof *lsock);
	    lsock -> sin_family = hp -> h_addrtype;
	    inaddr_copy (hp, lsock);
	} else {
	    /* KMTM reset the sin_port to zero, so that an unused port will be selected ! */
	    lsock->sin_port = 0 ;
	}
    }

    bzero ((char *) isock, sizeof *isock);
    isock -> sin_family = AF_INET;
    bcopy ((char *) agent -> party_TAddress, (char *) &isock -> sin_addr,
	   sizeof isock -> sin_addr);
    bcopy ((char *) (agent -> party_TAddress + sizeof isock -> sin_addr),
	   (char *) &isock -> sin_port, sizeof isock -> sin_port);

    /* Is the socket already open.. */
    if( snmpRMIBAgentDatap -> sd >= 0 )
	(void) close_udp_socket( snmpRMIBAgentDatap->sd );

    if ((snmpRMIBAgentDatap -> sd = start_udp_client (lsock, 0, 0, 0)) == NOTOK)
	return SnmpRMIBstartUDPclientFailed ; 

    if (join_udp_server ( snmpRMIBAgentDatap -> sd, isock) == NOTOK)
    	return SnmpRMIBjoinUDPserverFailed ;

    /* Does the PS already exist.. */
    if( snmpRMIBAgentDatap -> ps )
	ps_free( snmpRMIBAgentDatap -> ps );

    if ((snmpRMIBAgentDatap -> ps = ps_alloc (dg_open)) == NULLPS
	    || dg_setup ( snmpRMIBAgentDatap -> ps, snmpRMIBAgentDatap -> sd, agent -> party_MaxMessageSize,
			 read_udp_socket, write_udp_socket,
			 check_udp_socket) == NOTOK) {

	if( snmpRMIBAgentDatap -> ps == NULLPS )
	    return SnmpRMIBPSallocFailure ;
	else
	{
	    advise (NULLCP, "dg_setup: %s", ps_error( snmpRMIBAgentDatap -> ps -> ps_errno ));
	    ps_flush( snmpRMIBAgentDatap -> ps ) ;

	    return SnmpRMIBPSerror ;
	}
    }

    ps_len_strategy = PS_LEN_LONG;

    return SnmpRMIBnoError ;
#else
    return SnmpRMIBUDPsupportNotConfig ;
#endif

} /* SnmpRMIBpartyConnect() */


void
SnmpRMIBpartyDisconnect( snmpRMIBAgentDatap )
struct  SnmpRMIBsnmpifData * snmpRMIBAgentDatap ;
{
	(void)close_udp_socket( snmpRMIBAgentDatap->sd );
	if( snmpRMIBAgentDatap->ps )
	    ps_free( snmpRMIBAgentDatap->ps ) ;

	snmpRMIBAgentDatap->ps = NULLPS ;

	snmpRMIBAgentDatap->sd = -1 ;
	
	return ;
}

struct 	type_SNMP_Snmp1157Msg *
SnmpRMIBdecodeTrap( trappe )
PE	trappe ;
{
struct	type_SNMP_Snmp1157Msg *retmsg = (struct type_SNMP_Snmp1157Msg*)0,
			      *msg    = (struct type_SNMP_Snmp1157Msg*)0 ;
struct	type_SNMP_PDUs 	      *md ;
char	buffer[BUFSIZ];

    	if (decode_SNMP_Snmp1157Msg (trappe, 1, NULLIP, NULLVP, &msg) == NOTOK) {
		(void) sprintf (buffer, "failed: decode_SNMP_Snmp1157Msg loses {%s}",
			PY_pepy);
     	} else {

	    /* Check that this is a Trap message and extract off the relevant part.. */
	    if( msg->version 		== int_SNMP_version_version__1 &&
	    	msg->community		!= (struct qbuf*)0  &&
	    	(md = msg->data) 	!= (struct type_SNMP_PDUs*)0 &&
		md->offset 		== type_SNMP_PDUs_trap &&
 		md->un.trap 		!= (struct type_SNMP_Trap__PDU*)0 ) {

		retmsg = msg, msg = (struct type_SNMP_Snmp1157Msg*)0 ;
	    }

	    /* Tidy up... */
	    if( msg != (struct type_SNMP_Snmp1157Msg*)0 )
	    	free_SNMP_Snmp1157Msg( msg ) ;
	}

	return retmsg ;
}

static int party_connect (agent)
struct snmp_party *agent;
{
#ifdef	TCP
    char   *pp;
    struct sockaddr_in in_socket,
		       lo_socket;
    static struct sockaddr_in my_socket;
    register struct sockaddr_in *isock = &in_socket,
				*lsock;

    if (agent -> party_tDomain != SNMPUDPDOMAIN
	    && agent -> party_tDomain != RFC1157DOMAIN) {
	advise (NULLCP, "party %s is not reachable via SNMPv1/2 over UDP",
		(agent -> party_Identity != NULLOID) ? oid2ode (agent -> party_Identity) : "UNKNOWN PARTY" );
	return NOTOK;
    }

    if (agent -> party_TAddress[0] == 127
	    && agent -> party_TAddress[1] == 0
	    && agent -> party_TAddress[2] == 0
	    && agent -> party_TAddress[3] == 1) {
	lsock = &lo_socket;

	bzero ((char *) lsock, sizeof *lsock);
	lsock -> sin_family = AF_INET;
	bcopy ((char *) agent -> party_TAddress, (char *) &lsock -> sin_addr,
	       sizeof lsock -> sin_addr);
    }
    else {
	static int once_only = 1;

	lsock = &my_socket;

	if (once_only) {
	    register struct hostent *hp;

	    once_only = 0;

	    if ((hp = gethostbystring (pp = getlocalhost ())) == NULL)
		adios (NULLCP, "%s: unknown host", pp);

	    bzero ((char *) lsock, sizeof *lsock);
	    lsock -> sin_family = hp -> h_addrtype;
	    inaddr_copy (hp, lsock);
	}
    }

    bzero ((char *) isock, sizeof *isock);
    isock -> sin_family = AF_INET;
    bcopy ((char *) agent -> party_TAddress, (char *) &isock -> sin_addr,
	   sizeof isock -> sin_addr);
    bcopy ((char *) (agent -> party_TAddress + sizeof isock -> sin_addr),
	   (char *) &isock -> sin_port, sizeof isock -> sin_port);

    if (sd != NOTOK)
	(void) close_udp_socket (sd);
    if ((sd = start_udp_client (lsock, 0, 0, 0)) == NOTOK)
	adios ("failed", "start_udp_client");
    if (join_udp_server (sd, isock) == NOTOK)
	adios ("failed", "join_udp_server");

    if (ps)
	ps_free (ps);
    if ((ps = ps_alloc (dg_open)) == NULLPS
	    || dg_setup (ps, sd, agent -> party_MaxMessageSize,
			 read_udp_socket, write_udp_socket,
			 check_udp_socket) == NOTOK) {
	if (ps == NULLPS)
	    adios (NULLCP, "ps_alloc(dg_open): failed");
	else
	    adios (NULLCP, "dg_setup: %s", ps_error (ps -> ps_errno));
    }

    return OK;
#else
    adios (NULLCP, "UDP support not configured");
#endif
}

/* */

static struct snmp_party *find_party (cp, who, warning)
char   *cp;
int	who,
    	warning;
{
    int	    i;
    register char *dp;
    register struct snmp_party *p;
    OID	    oid;

    for (dp = cp; isdigit (*dp); dp++)
	continue;
    if (*dp == NULL && sscanf (cp, "%d", &i) == 1) {
	if (!(p = index_party_aux (i, EITHER, 0))) {
	    advise (NULLCP, "no party with index %d", i);
	    return NULL;
	}

	goto test_party;
    }

    if (!(oid = text2oid (cp))) {
	advise (NULLCP, "no such object as %s", cp);
	return NULL;
    }

    p = lookup_party_aux (oid, EITHER, 0);
    oid_free (oid);

    if (!p) {
	advise (NULLCP, "no such party as %s", cp);
	return NULL;
    }

test_party: ;
    if (who != EITHER && who != p -> party_location) {
	advise (NULLCP, "party %s is a%s party", cp,
		p -> party_location == AGENT ? "n agent (destination)"
					     : " manager (source)");
	return NULL;
    }
    if (p -> party_tDomain != SNMPUDPDOMAIN &&
    	p -> party_tDomain != RFC1157DOMAIN ) {
	advise (NULLCP, "party %s isn't an SNMPv2/UDP party", cp);

	return NULL;
    }
    
    if (p -> party_status != ROW_ACTIVE && warning)
	advise (NULLCP, "party number %d has invalid secrets (warning)",
		p -> party_index);

    return p;
}

/* */

static struct snmp_context *find_context (cp)
char   *cp;
{
    int	    i;
    register char *dp;
    register struct snmp_context *sc;
    OID	    oid;

    for (dp = cp; isdigit (*dp); dp++)
	continue;
    if (*dp == NULL && sscanf (cp, "%d", &i) == 1) {
	if (!(sc = index_context_aux (i, 0))) {
	    advise (NULLCP, "no context with index %d", i);
	    return NULL;
	}

	goto test_context;
    }

    i = strlen (cp);
    for (sc = contextHead; sc; sc = sc -> sc_next)
	if (sc -> sc_name
	        && sc -> sc_name_len == i
	        && !lexequ ((char *) sc -> sc_name, cp))
	    goto test_context;

    if (!(oid = text2oid (cp))) {
	advise (NULLCP, "no such object as %s", cp);
	return NULL;
    }

    sc = lookup_context_aux (oid, 0);
    oid_free (oid);

    if (!sc) {
	advise (NULLCP, "no such context as %s", cp);
	return NULL;
    }

test_context: ;
    if (sc -> sc_location != AGENT) {
	advise (NULLCP, "context %s is local", cp);
	return NULL;
    }
    if (sc -> sc_status != ROW_ACTIVE)
	advise (NULLCP, "context number %d isn't active (warning)",
		sc -> sc_index);

    return sc;
}

/* */

static struct objreg party_objects[] = {
    "partyIndex",		(caddr_t) partyIndex,
    "partyTDomain",		(caddr_t) partyTDomain,
    "partyTAddress",		(caddr_t) partyTAddress,
    "partyMaxMessageSize",	(caddr_t) partyMaxMessageSize,
    "partyLocal",		(caddr_t) partyLocal,
    "partyAuthProtocol",	(caddr_t) partyAuthProtocol,
    "partyAuthClock",		(caddr_t) partyAuthClock,
    "partyAuthPrivate",		(caddr_t) partyAuthPrivate,
    "partyAuthPublic",		(caddr_t) partyAuthPublic,
    "partyAuthLifetime",	(caddr_t) partyAuthLifetime,
    "partyPrivProtocol",	(caddr_t) partyPrivProtocol,
    "partyPrivPrivate",		(caddr_t) partyPrivPrivate,
    "partyPrivPublic",		(caddr_t) partyPrivPublic,
    "partyCloneFrom",		(caddr_t) partyCloneFrom,
    "partyStatus",		(caddr_t) partyStatus,
    "partyStorageType",		(caddr_t) partyStorageType,

    "iSODEContextName",		(caddr_t) contextName,
    "contextIndex",		(caddr_t) contextIndex,
    "contextLocal",		(caddr_t) contextLocal,
    "contextViewIndex",		(caddr_t) contextViewIndex,
    "contextLocalEntity",	(caddr_t) contextLocalEntity,
    "contextLocalTime",		(caddr_t) contextLocalTime,
    "contextProxyDstParty",	(caddr_t) contextProxyDstParty,
    "contextProxySrcParty",	(caddr_t) contextProxySrcParty,
    "contextProxyContext",	(caddr_t) contextProxyContext,
    "contextStatus",		(caddr_t) contextStatus,
    "contextStorageType",	(caddr_t) contextStorageType,

    "aclPrivileges",		(caddr_t) aclPrivileges,
    "aclStatus",		(caddr_t) aclStatus,
    "aclStorageType",		(caddr_t) aclStorageType,

    NULL
};

static	remember_party ()
{
    register_objects (party_objects, NULLIFP, learn_party);
}

/* */

/* call this when an authenticated response, without errors, is received */


static int  learn_party (vp)
register struct type_SNMP_VarBindList *vp;
{
    int	    result = OK;

    for (; vp; vp = vp -> next) {
	int	i,
		ifvar;
	register struct snmp_party *p = NULL;
	register struct snmp_context *sc = NULL;
	register struct snmp_access *sa = NULL;
	register OI	oi;
	register OS	os;
	register OT	ot;
	caddr_t value;

	if ((oi = name2inst (vp -> VarBind -> name)) == NULL
	        || (ot = oi -> oi_type) == NULL
	        || ot -> ot_setfnx != learn_party
	        || (os = ot -> ot_syntax) == NULL)
	    continue;

	switch (ifvar = (int) ot -> ot_info) {
	    case partyIndex:
	    case partyAuthPrivate:
	    case partyPrivPrivate:
		if (doing_f_secrets)
		    continue;
		/* else fall... */
	    case partyTDomain:
	    case partyTAddress:
	    case partyMaxMessageSize:
	    case partyLocal:
	    case partyAuthProtocol:
	    case partyAuthClock:
	    case partyAuthPublic:
	    case partyAuthLifetime:
	    case partyPrivProtocol:
	    case partyPrivPublic:
	    case partyCloneFrom:
	    case partyStatus:
	    case partyStorageType:
		{
		    unsigned int *ip;
		    struct OIDentifier oids;
		    OID	    oid;

		    if ((i = (oid = oi -> oi_name) -> oid_nelem
		                           - ot -> ot_name -> oid_nelem) < 2)
			continue;
		    ip = oid -> oid_elements + ot -> ot_name -> oid_nelem;
		    oids.oid_elements = ip, oids.oid_nelem = i;

		    for (p = partyHead; p; p = p -> party_next)
			if (oid_cmp (p -> party_Identity, &oids) == 0)
			    break;
		    if (!p) {
			register struct snmp_party **pp;
		
			if ((p = (struct snmp_party *) calloc (1, sizeof *p))
			        == NULL)
			    goto losing;
			p -> party_Identity = oid_cpy (&oids);
			p -> party_TDomain = oid_cpy (snmpUDPDomain);
			p -> party_tDomain = SNMPUDPDOMAIN;
			p -> party_TAddress =
			    (u_char *) calloc ((unsigned) p -> party_TAddress_len = 6,
					       sizeof *p -> party_TAddress);
			p -> party_MaxMessageSize = MAXSNMP;
			p -> party_location = MANAGER;
			p -> party_AuthProtocol = oid_cpy (md5Auth);
			p -> party_authProt = MD5AUTHPROT;
			p -> party_AuthStart = party_time_now ();
			p -> party_AuthLifetime = 300;
			p -> party_PrivProtocol = oid_cpy (noPriv);
			p -> party_privProt = NOPRIV;
			p -> party_status = ROW_ACTIVE;
			p -> party_storage = STORAGE_NONVOLATILE;

			if (!p -> party_Identity
			        || !p -> party_TDomain
			        || !p -> party_TAddress
			        || !p -> party_AuthProtocol
			        || !p -> party_PrivProtocol) {
			    free_party_aux (p, 0);
losing: ;
			    result = NOTOK;
			    continue;
			}

			i = 1;
			for (pp = &partyHead; *pp; pp = &((*pp) -> party_next))
			    if ((*pp) -> party_index >= i)
				i = (*pp) -> party_index + 1;
			*pp = p;
			p -> party_index = i;

			party_changed = 1;
		    }
		}
		break;

	    case contextName:
	    case contextIndex:
	    case contextLocal:
	    case contextViewIndex:
	    case contextLocalEntity:
	    case contextLocalTime:
	    case contextProxyDstParty:
	    case contextProxySrcParty:
	    case contextProxyContext:
	    case contextStatus:
	    case contextStorageType:
		{
		    unsigned int *ip;
		    struct OIDentifier oids;
		    OID	    oid;

		    if ((i = (oid = oi -> oi_name) -> oid_nelem
		                           - ot -> ot_name -> oid_nelem) < 2)
			continue;
		    ip = oid -> oid_elements + ot -> ot_name -> oid_nelem;
		    oids.oid_elements = ip, oids.oid_nelem = i;

		    for (sc = contextHead; sc; sc = sc -> sc_next)
			if (oid_cmp (sc -> sc_identity, &oids) == 0)
			    break;
		    if (!sc) {
			register struct snmp_context **pc;
		
			if ((sc = (struct snmp_context *) calloc (1,
								  sizeof *sc))
			        == NULL)
			    goto losing;
			if (sc -> sc_identity = oid_cpy (&oids))
			    sc -> sc_name = (u_char *)
					strdup (sprintoid (sc -> sc_identity));
			sc -> sc_location = AGENT;
			sc -> sc_entity =
			    (u_char *) calloc (1, sizeof *sc -> sc_entity);
			sc -> sc_time = oid_cpy (currentTime);
			sc -> sc_dst_oid = oid_cpy (nullSpecific);
			sc -> sc_src_oid = oid_cpy (nullSpecific);
			sc -> sc_ctx = oid_cpy (nullSpecific);
			sc -> sc_status = ROW_ACTIVE;
			sc -> sc_storage = STORAGE_NONVOLATILE;

			if (!sc -> sc_identity
			        || !sc -> sc_name
			        || !sc -> sc_entity
			        || !sc -> sc_time
			        || !sc -> sc_dst_oid
			        || !sc -> sc_src_oid
			        || !sc -> sc_ctx) {
			    free_context_aux (sc, 0);
			    goto losing;
			}
			sc -> sc_name_len = strlen ((char *) sc -> sc_name);

			i = 1;
			for (pc = &contextHead; *pc; pc = &((*pc) -> sc_next))
			    if ((*pc) -> sc_index >= i)
				i = (*pc) -> sc_index + 1;
			*pc = sc;
			sc -> sc_index = i;

			context_changed = 1;
		    }
		}
		break;

	    case aclPrivileges:
	    case aclStatus:
	    case aclStorageType:
		{
		    unsigned int *ip;
		    struct OIDentifier oids;
		    OID	    oid;

		    if ((i = (oid = oi -> oi_name) -> oid_nelem
		                           - ot -> ot_name -> oid_nelem) != 3)
			continue;
		    ip = oid -> oid_elements + ot -> ot_name -> oid_nelem;
		    oids.oid_elements = ip, oids.oid_nelem = i;

		    for (sa = accessHead; sa; sa = sa -> sa_next)
			if (oid_cmp (&sa -> sa_instance, &oids) == 0)
			    break;
		    if (!sa) {
			if ((sa = (struct snmp_access *) calloc (1,
								 sizeof *sa))
			            == NULL
			        || (oid = oid_cpy (&oids)) == NULL) {
			    free_access_aux (sa, 0);
			    goto losing;
			}
			sa -> sa_instance = *oid;	/* struct copy */
			free ((char *) oid);

			sa -> sa_target =
			    index_party_aux (sa -> sa_target_index = *ip++,
					     EITHER, 0);
			sa -> sa_subject =
			    index_party_aux (sa -> sa_subject_index = *ip++,
					     EITHER, 0);
			sa -> sa_resources =
			    index_context_aux (sa -> sa_resources_index = *ip,
					       0);

			if (!sa -> sa_target
			        || !sa -> sa_subject
			        || !sa -> sa_resources) {
			    free_access_aux (sa, 0);
			    goto losing;
			}

			sa -> sa_privileges = sa -> sa_new_privileges =
							  PARTY_PRIV_GET
							| PARTY_PRIV_NEXT
							| PARTY_PRIV_BULK;
			sa -> sa_status = ROW_ACTIVE;
			sa -> sa_storage = STORAGE_NONVOLATILE;

			sa -> sa_next = accessHead;
			accessHead = sa;

			access_changed = 1;
		    }
		}
		break;

	    default:
		continue;
	}
	
	if (debug) {
	    fprintf (stderr, "learning about ");
	    if (p)
		fprintf (stderr, "party %s", oid2ode (p -> party_Identity));
	    else
		if (sc)
		    fprintf (stderr, "context %s",
			     oid2ode (sc -> sc_identity));
	        else
		    fprintf (stderr, "access %s",
			     oid2ode (&sa -> sa_instance));
	    fprintf (stderr, " (variable %s)\n", ot -> ot_text);
	}

	if (vp -> VarBind -> value -> pe_class == PE_CLASS_CONT
	        || (*os -> os_decode) (&value, vp -> VarBind -> value)
			== NOTOK)
	    continue;

	switch (ifvar) {
	    case partyIndex:
		if ((i = *((integer *) value)) < 0 || p -> party_index == i)
		    goto ignore_value;
		p -> party_index = i;
		break;

	    case partyTDomain:
	        if (oid_cmp (p -> party_TDomain, (OID) value) == 0)
		    goto ignore_value;
		oid_free (p -> party_TDomain);
		p -> party_TDomain = (OID) value;
		p -> party_tDomain =
		        oid_cmp (p -> party_TDomain, snmpUDPDomain) == 0
			    ? SNMPUDPDOMAIN
		            : oid_cmp (p -> party_TDomain, rfc1157Domain) == 0
				 ? RFC1157DOMAIN : BADOMAIN;
		if (p -> party_tDomain == RFC1157DOMAIN) {
		    oid_free (p -> party_AuthProtocol);
		    p -> party_AuthProtocol = oid_cpy (rfc1157noAuth);
		    p -> party_authProt = RFC1157NOAUTH;
		    p -> party_AuthLifetime = 0;
		    p -> party_location = MANAGER;
		    p -> party_MaxMessageSize = MAXSNMP;
		}
nullify_value: ;
		value = NULL;
		break;

	    case partyTAddress:
		{
		    char   *cp;
		    struct qbuf *qb = (struct qbuf *) value;

		    if (qb -> qb_forw -> qb_forw != qb
			    || (p -> party_TAddress_len == qb -> qb_len
				    && bcmp ((char *) p -> party_TAddress,
					     qb -> qb_forw -> qb_data,
					     p -> party_TAddress_len) == 0))
			goto ignore_value;
		    if ((cp = qb2str (qb)) == NULL) {
			result = NOTOK;
			goto ignore_value;
		    }
		    free ((char *) p -> party_TAddress);
		    p -> party_TAddress = (u_char *) cp;
		    p -> party_TAddress_len = qb -> qb_len;
		    {
			register struct snmp_party *q;

			for (q = partyHead; q; q = q -> party_next)
			    if (q -> party_location == AGENT
				    && bcmp ((char *) q -> party_TAddress,
					     (char *) p -> party_TAddress,
					     q -> party_TAddress_len) == 0)
				p -> party_location = AGENT;
		    }
		}
		break;

	    case partyMaxMessageSize:
		if ((i = *((integer *) value)) < MAXSNMP
		        || i > BIGSNMP
			|| p -> party_MaxMessageSize == i)
		    goto ignore_value;
		p -> party_MaxMessageSize = i;
		break;

	    case partyLocal:
		if ((i = *((integer *) value)) < AGENT
		        || i > MANAGER
			|| p -> party_location == i)
		    goto ignore_value;
		p -> party_location = i;
		break;

	    case partyAuthProtocol:
	        if (oid_cmp (p -> party_AuthProtocol, (OID) value) == 0)
		    goto ignore_value;
		oid_free (p -> party_AuthProtocol);
		p -> party_clone |= PARTY_AUTHPROT;
		p -> party_AuthProtocol = (OID) value;
		p -> party_authProt =
			oid_cmp (p -> party_AuthProtocol, noAuth) == 0
			    ? NOAUTH
			    : oid_cmp (p -> party_AuthProtocol, md5Auth) ==0
				  ? MD5AUTHPROT
			          : oid_cmp (p -> party_AuthProtocol,
					     rfc1157noAuth) == 0
						 ? RFC1157NOAUTH : BADAUTH;
		if (p -> party_authProt != MD5AUTHPROT)
		    p -> party_AuthLifetime = 0;
		goto nullify_value;

	    case partyAuthClock:
		{
		    integer big = *((integer *) value);
		    long    now = party_time_now () - big;

		    if (big < 0 || p -> party_AuthStart <= now)
			goto ignore_value;
		    p -> party_AuthStart = now;
		}
		break;

	    case partyAuthPrivate:
		{
		    struct qbuf *qb = (struct qbuf *) value;

		    if (qb -> qb_forw -> qb_forw != qb)
			goto ignore_value;
		    if (p -> party_AuthPrivate)
			exor (p -> party_AuthPrivate,
			      p -> party_AuthPrivate_len,
			      (u_char *) qb -> qb_forw -> qb_data,
			      qb -> qb_len);
		    else {
			char   *cp;

			if ((cp = qb2str (qb)) == NULL) {
			    result = NOTOK;
			    goto ignore_value;
			}
			p -> party_AuthPrivate = (u_char *) cp;
			p -> party_AuthPrivate_len = qb -> qb_len;
		    }
		}
		break;

	    case partyAuthPublic:
		{
		    char   *cp;
		    struct qbuf *qb = (struct qbuf *) value;

		    if (qb -> qb_forw -> qb_forw != qb
			    || (p -> party_AuthPublic
				    && p -> party_AuthPublic_len
								== qb -> qb_len
				    && bcmp ((char *) p -> party_AuthPublic,
					     qb -> qb_forw -> qb_data,
					     p -> party_AuthPublic_len) == 0)
			    || (!p -> party_AuthPublic && qb -> qb_len == 0))
			goto ignore_value;
		    if ((cp = qb2str (qb)) == NULL) {
			result = NOTOK;
			goto ignore_value;
		    }
		    if (p -> party_AuthPublic)
			free ((char *) p -> party_AuthPublic);
		    p -> party_AuthPublic = (u_char *) cp;
		    p -> party_AuthPublic_len = qb -> qb_len;
		}
		break;

	    case partyAuthLifetime:
		if ((i = *((integer *) value)) < 0
		        || i > 2147483647
			|| p -> party_AuthLifetime == i)
		    goto ignore_value;
		p -> party_AuthLifetime = i;
		break;

	    case partyPrivProtocol:
	        if (oid_cmp (p -> party_PrivProtocol, (OID) value) == 0)
		    goto ignore_value;
		oid_free (p -> party_PrivProtocol);
		p -> party_clone |= PARTY_PRIVPROT;
		p -> party_PrivProtocol = (OID) value;
		p -> party_privProt =
			oid_cmp (p -> party_PrivProtocol, noPriv) == 0
			    ? NOPRIV
			    : oid_cmp (p -> party_PrivProtocol, desPriv) == 0
				  ? DESPRIVPROT : BADPRIV;
		goto nullify_value;

	    case partyPrivPrivate:
		{
		    struct qbuf *qb = (struct qbuf *) value;

		    if (qb -> qb_forw -> qb_forw != qb)
			goto ignore_value;
		    if (p -> party_PrivPrivate)
			exor (p -> party_PrivPrivate,
			      p -> party_PrivPrivate_len,
			      (u_char *) qb -> qb_forw -> qb_data,
			      qb -> qb_len);
		    else {
			char   *cp;

			if ((cp = qb2str (qb)) == NULL) {
			    result = NOTOK;
			    goto ignore_value;
			}
			p -> party_PrivPrivate = (u_char *) cp;
			p -> party_PrivPrivate_len = qb -> qb_len;
		    }
		}
		break;

	    case partyPrivPublic:
		{
		    char   *cp;
		    struct qbuf *qb = (struct qbuf *) value;

		    if (qb -> qb_forw -> qb_forw != qb
			    || (p -> party_PrivPublic
				    && p -> party_PrivPublic_len
								== qb -> qb_len
				    && bcmp ((char *) p -> party_PrivPublic,
					     qb -> qb_forw -> qb_data,
					     p -> party_PrivPublic_len) == 0)
			    || (!p -> party_PrivPublic && qb -> qb_len == 0))
			goto ignore_value;
		    if ((cp = qb2str (qb)) == NULL) {
			result = NOTOK;
			goto ignore_value;
		    }
		    if (p -> party_PrivPublic)
			free ((char *) p -> party_PrivPublic);
		    p -> party_PrivPublic = (u_char *) cp;
		    p -> party_PrivPublic_len = qb -> qb_len;
		}
		break;

	    case partyCloneFrom:
		{
		    register struct snmp_party *q;

		    if (!(q = lookup_party ((OID) value, EITHER))) {
			result = NOTOK;
			goto ignore_value;
		    }

		    if (!(p -> party_clone & PARTY_AUTHPROT)
			    && p -> party_authProt != q -> party_authProt) {
			oid_free (p -> party_AuthProtocol);
			p -> party_AuthProtocol =
					    oid_cpy (q -> party_AuthProtocol);
			p -> party_authProt = q -> party_authProt;
		    }
		    if (p -> party_authProt != NOAUTH) {
			if (p -> party_AuthPrivate)
			    exor (p -> party_AuthPrivate,
				  p -> party_AuthPrivate_len,
				  q -> party_AuthPrivate,
				  q -> party_AuthPrivate_len);
		        else
			    if (q -> party_AuthPrivate) {
				if ((p -> party_AuthPrivate =
				     (unsigned char *) malloc ((unsigned int)
						   q -> party_AuthPrivate_len))
				        == NULL) {
				    result = NOTOK;
				    goto ignore_value;
				}
				bcopy ((char *) q -> party_AuthPrivate,
				       (char *) p -> party_AuthPrivate,
				       p -> party_AuthPrivate_len
				               = q -> party_AuthPrivate_len);
			    }
		    }

		    if (!(p -> party_clone & PARTY_PRIVPROT)
			    && p -> party_privProt != q -> party_privProt) {
			oid_free (p -> party_PrivProtocol);
			p -> party_PrivProtocol =
					    oid_cpy (q -> party_PrivProtocol);
			p -> party_privProt = q -> party_privProt;
		    }
		    if (p -> party_privProt != NOPRIV) {
			if (p -> party_PrivPrivate)
			    exor (p -> party_PrivPrivate,
				  p -> party_PrivPrivate_len,
				  q -> party_PrivPrivate,
				  q -> party_PrivPrivate_len);
		        else
			    if (q -> party_PrivPrivate) {
				if ((p -> party_PrivPrivate =
				     (unsigned char *) malloc ((unsigned int)
						   q -> party_PrivPrivate_len))
				        == NULL) {
				    result = NOTOK;
				    goto ignore_value;
				}
				bcopy ((char *) q -> party_PrivPrivate,
				       (char *) p -> party_PrivPrivate,
				       p -> party_PrivPrivate_len
				               = q -> party_PrivPrivate_len);
			    }
		    }
		}
		goto ignore_value;

	    case partyStatus:
		switch (i = *((integer *) value)) {
		    case ROW_ACTIVE:
		    case ROW_NOTINSVC:
		    case ROW_DESTROY:
		        p -> party_status = i;
			break;

		    case ROW_CREATEGO:
			p -> party_status = ROW_ACTIVE;
			break;

		    case ROW_CREATEWAIT:
			p -> party_status = ROW_NOTINSVC;
			break;

		    default:
		        goto ignore_value;
		}
		break;

	    case partyStorageType:
		switch (i = *((integer *) value)) {
		    case STORAGE_OTHER:
		    case STORAGE_VOLATILE:
		    case STORAGE_NONVOLATILE:
		    case STORAGE_PERMANENT:
		        p -> party_storage = i;
			break;

		    default:
		        goto ignore_value;
		}
		break;
		
	    case contextName:
		{
		    char   *cp;
		    struct qbuf *qb = (struct qbuf *) value;

		    if (qb -> qb_forw -> qb_forw != qb
			    || (sc -> sc_name
				    && sc -> sc_name_len == qb -> qb_len
				    && bcmp ((char *) sc -> sc_name,
					     qb -> qb_forw -> qb_data,
					     sc -> sc_name_len) == 0)
			    || (!sc -> sc_name && qb -> qb_len == 0))
			goto ignore_value;
		    if ((cp = qb2str (qb)) == NULL) {
			result = NOTOK;
			goto ignore_value;
		    }
		    if (sc -> sc_name)
			free ((char *) sc -> sc_name);
		    sc -> sc_name = (u_char *) cp;
		    sc -> sc_name_len = qb -> qb_len;
		}
		break;

	    case contextIndex:
		if ((i = *((integer *) value)) < 0 || sc -> sc_index == i)
		    goto ignore_value;
		sc -> sc_index = i;
		break;

	    case contextLocal:
		if ((i = *((integer *) value)) < AGENT
		        || i > MANAGER
			|| sc -> sc_location == i)
		    goto ignore_value;
		sc -> sc_location = i;
		break;

	    case contextViewIndex:
		if ((i = *((integer *) value)) < 0
			|| sc -> sc_view == i)
		    goto ignore_value;
		sc -> sc_view = i;
		break;

	    case contextLocalEntity:
		{
		    char   *cp;
		    struct qbuf *qb = (struct qbuf *) value;

		    if (qb -> qb_forw -> qb_forw != qb
			    || (sc -> sc_entity
				    && sc -> sc_entity_len == qb -> qb_len
				    && bcmp ((char *) sc -> sc_entity,
					     qb -> qb_forw -> qb_data,
					     sc -> sc_entity_len) == 0)
			    || (!sc -> sc_entity && qb -> qb_len == 0))
			goto ignore_value;
		    if ((cp = qb2str (qb)) == NULL) {
			result = NOTOK;
			goto ignore_value;
		    }
		    if (sc -> sc_entity)
			free ((char *) sc -> sc_entity);
		    sc -> sc_entity = (u_char *) cp;
		    sc -> sc_entity_len = qb -> qb_len;
		}
		break;

	    case contextLocalTime:
		if (sc -> sc_time) {
		    if (oid_cmp (sc -> sc_time, (OID) value) == 0)
			goto ignore_value;
		    oid_free (sc -> sc_time);
		}
		sc -> sc_time = (OID) value;
		goto nullify_value;

	    case contextProxyDstParty:
		if (sc -> sc_dst_oid) {
		    if (oid_cmp (sc -> sc_dst_oid, (OID) value) == 0)
			goto ignore_value;
		    oid_free (sc -> sc_dst_oid);
		}
		sc -> sc_dst_oid = (OID) value;
		goto nullify_value;

	    case contextProxySrcParty:
		if (sc -> sc_src_oid) {
		    if (oid_cmp (sc -> sc_src_oid, (OID) value) == 0)
			goto ignore_value;
		    oid_free (sc -> sc_src_oid);
		}
		sc -> sc_src_oid = (OID) value;
		goto nullify_value;

	    case contextProxyContext:
		if (sc -> sc_ctx) {
		    if (oid_cmp (sc -> sc_ctx, (OID) value) == 0)
			goto ignore_value;
		    oid_free (sc -> sc_ctx);
		}
		sc -> sc_ctx = (OID) value;
		goto nullify_value;

	    case contextStatus:
		switch (i = *((integer *) value)) {
		    case ROW_ACTIVE:
		    case ROW_NOTINSVC:
		    case ROW_DESTROY:
		        sc -> sc_status = i;
			break;

		    case ROW_CREATEGO:
			sc -> sc_status = ROW_ACTIVE;
			break;

		    case ROW_CREATEWAIT:
			sc -> sc_status = ROW_NOTINSVC;
			break;

		    default:
		        goto ignore_value;
		}
		break;

	    case contextStorageType:
		switch (i = *((integer *) value)) {
		    case STORAGE_OTHER:
		    case STORAGE_VOLATILE:
		    case STORAGE_NONVOLATILE:
		    case STORAGE_PERMANENT:
		        sc -> sc_storage = i;
			break;

		    default:
		        goto ignore_value;
		}
		break;
		
	    case aclPrivileges:
		if ((i = *((integer *) value)) < 0
		        || i > 255
			|| sa -> sa_privileges == i)
		    goto ignore_value;
		sa -> sa_privileges = sa -> sa_new_privileges = i;
		break;

	    case aclStatus:
		switch (i = *((integer *) value)) {
		    case ROW_ACTIVE:
		    case ROW_NOTINSVC:
		    case ROW_DESTROY:
		        sa -> sa_status = i;
			break;

		    case ROW_CREATEGO:
			sa -> sa_status = ROW_ACTIVE;
			break;

		    case ROW_CREATEWAIT:
			sa -> sa_status = ROW_NOTINSVC;
			break;

		    default:
		        goto ignore_value;
		}
		break;

	    case aclStorageType:
		switch (i = *((integer *) value)) {
		    case STORAGE_OTHER:
		    case STORAGE_VOLATILE:
		    case STORAGE_NONVOLATILE:
		    case STORAGE_PERMANENT:
		        sa -> sa_storage = i;
			break;

		    default:
		        goto ignore_value;
		}
		break;
		
	    default:
ignore_value: ;
		(*os -> os_free) (value);
		if (debug)
		    fprintf (stderr, "\tignored...\n");
		continue;
	}

	if (value)
	    (*os -> os_free) (value);

	if (debug)
	    fprintf (stderr, "\tchanged...\n");
	switch (ifvar) {
	    case partyTDomain:
	    case partyTAddress:
	    case partyMaxMessageSize:
	    case partyLocal:
	    case partyAuthProtocol:
	    case partyAuthClock:
	    case partyAuthPrivate:
	    case partyAuthPublic:
	    case partyAuthLifetime:
	    case partyPrivProtocol:
	    case partyPrivPrivate:
	    case partyPrivPublic:
	    case partyCloneFrom:
	    case partyStatus:
	    case partyStorageType:
	        party_changed = 1;
		break;

	    case contextName:
	    case contextLocal:
	    case contextViewIndex:
	    case contextLocalEntity:
	    case contextLocalTime:
	    case contextProxyDstParty:
	    case contextProxySrcParty:
	    case contextProxyContext:
	    case contextStatus:
	    case contextStorageType:
		context_changed = 1;
		break;

	    case aclPrivileges:
	    case aclStatus:
	    case aclStorageType:
		access_changed = 1;
		break;
	}
    }

    return result;
}

/* */

static	check_party ()
{
    register struct snmp_party *p;
    register struct snmp_context *sc;
    register struct snmp_access *sa;
    
    if (agentParty && agentParty -> party_status == ROW_DESTROY)
	agentParty = NULLPARTY;
    if (managerParty && managerParty -> party_status == ROW_DESTROY)
	managerParty = NULLPARTY;
    if (mgmtContext && mgmtContext -> sc_status == ROW_DESTROY)
	mgmtContext = NULL;

    for (p = partyHead; p; p = p -> party_next)
	if (p -> party_status == ROW_DESTROY)
	    break;
    if (p)
	sc = NULL, sa = NULL;
    else {
	for (sc = contextHead; sc; sc = sc -> sc_next)
	    if (sc -> sc_status == ROW_DESTROY)
		break;
	if (sc)
	    sa = NULL;
	else
	    for (sa = accessHead; sa; sa = sa -> sa_next)
		if (sa -> sa_status == ROW_DESTROY)
		    break;
    }

    if (p || sc || sa)
	(void) f_sync (NULLVP);
}

/* */

#include <ctype.h>


/* ARGSUSED */

static int  object_parse (x, s, os)
OID    *x;
char   *s;
OS	os;
{
    char   *name;
    OID	    oid = NULL;

    if ((name = strdup (s)) == NULL)
	return NOTOK;
    s = name;

    if (isdigit ((u_char) *s)) {
	int	result;

	if (oid = str2oid (s)) {
okay: ;
	    *x = oid;
	    result = OK;
	}
	else {	
out: ;
	    if (oid)
		oid_free (oid);
	    result = NOTOK;
	}

	free (name);
	return result;
    }
    
    for (;;) {
	int	i;
	register char *cp,
		      *dp;
	OID	new;
	OT	ot;

	if (cp = index (s, '.'))
	    *cp++ = NULL;

	if (lexnequ (s, "party#", sizeof "party#" - 1) == 0) {
	    register struct snmp_party *p;

	    if (sscanf (s + sizeof "party#" - 1, "%d", &i) != 1)
		goto out;
	    if (!(p = index_party_aux (i, EITHER, 0)))
		goto out;
	    new = p -> party_Identity;
	    goto insert;
	}

	if (lexnequ (s, "context#", sizeof "context#" - 1) == 0) {
	    register struct snmp_context *sc;

	    if (sscanf (s + sizeof "context#" - 1, "%d", &i) != 1)
		goto out;
	    if (!(sc = index_context_aux (i, 0)))
		goto out;
	    new = sc -> sc_identity;
	    goto insert;
	}

	if ((ot = text2obj (s)) == NULL)
	    goto out;
	new = ot -> ot_name;
	/* and fall... */

insert: ;
	i = 0;
	if (dp = cp)
	    for (s = cp; isdigit ((u_char) *cp); i++) {
		if (s = index (cp, '.'))
		    cp = s + 1;
		else
		    cp += strlen (cp);
	    }

	{
	    register unsigned int *ip;
	    OID	    q;

	    if (oid == NULL) {
		if ((oid = oid_cpy (new)) == NULL)
		    goto out;

		if (i > 0) {
		    if ((q = oid_extend (oid, i)) == NULL)
			goto out;
		    oid_free (oid);
		    oid = q;

		    ip = oid -> oid_elements + oid -> oid_nelem - i;
		}
	    }
	    else {
		int	j;
		register unsigned int *jp;

		if ((q = oid_extend (oid, new -> oid_nelem + i)) == NULL)
		    goto out;
		j = oid -> oid_nelem;
		oid_free (oid);
		oid = q;

		ip = oid -> oid_elements + j;
		for (jp = new -> oid_elements, j = new -> oid_nelem;
		         j > 0;
		         j--)
		    *ip++ = *jp++;
	    }

	    if (i > 0)
		for (s = dp; isdigit ((u_char) *dp); i--) {
		    if (sscanf (dp, "%u", ip++) != 1)
			goto out;
		    if (s = index (dp, '.'))
			dp = s + 1;
		    else
			dp += strlen (dp);
		}
	}

	if (!(s = cp) || !*s)
	    goto okay;
    }
}


/* ARGSUSED */

static int  object_print (x, os)
OID	x;
OS	os;
{
    register struct snmp_party *p;
    register struct snmp_context *sc;

    if ((p = lookup_party_aux (x, EITHER, 0)) && p -> party_index > 0)
	printf ("party#%d", p -> party_index);
    else
	if ((sc = lookup_context_aux (x, 0)) && sc -> sc_index > 0)
	    printf ("context#%d", sc -> sc_index);
	else
	    printf ("%s", oid2ode (x));
}


/* ARGSUSED */

static char *oid2ode_aux (oid, quoted)
OID	oid;
int	quoted;
{
    int	    fancy;
    register int    i;
    register char  *bp;
    register unsigned int *ip;
    OT	    ot;
    static char buffer[BUFSIZ];

    if ((oid -> oid_nelem == 2		       /* XXX: 0.0 is a special case */
		&& oid -> oid_elements[0] == 0
		&& oid -> oid_elements[1] == 0)
	    || (ot = name2obj (oid)) == NULLOT)
	return sprintoid (oid);

    (void) strcpy (bp = buffer, ot -> ot_text);
    bp += strlen (bp);

    ip = oid -> oid_elements + ot -> ot_name -> oid_nelem;
    i = oid -> oid_nelem - ot -> ot_name -> oid_nelem;

/* ultra-special hack for the viewTable */
    if (i > 2 && ot -> ot_getfnx == (IFP) 1) {
	(void) sprintf (bp, ".%u", *ip++);
	bp += strlen (bp);
	i--;
    }

    fancy = 0;
    if (i > 1) {
	int	j;
	struct OIDentifier oids;
	register struct snmp_party *p;
	register struct snmp_context *sc;
	OT	ot2;

	oids.oid_elements = ip, oids.oid_nelem = i;

	if ((p = lookup_party_aux (&oids, EITHER, 0))
	        && p -> party_index > 0) {
	    (void) sprintf (bp, ".party#%d", p -> party_index);
	    bp += strlen (bp);

	    ip += (j = p -> party_Identity -> oid_nelem + 1);
	    i -= j;
	    fancy = 1;
	}
	else
	if ((sc = lookup_context_aux (&oids, 0)) && sc -> sc_index > 0) {
	    (void) sprintf (bp, ".context#%d", sc -> sc_index);
	    bp += strlen (bp);

	    ip += (j = sc -> sc_identity -> oid_nelem + 1);
	    i -= j;
	    fancy = 1;
	}
	else
	if (ot2 = name2obj (&oids)) {
	    int	    k = *ip + 1;

	    (void) sprintf (bp, ".%s", ot2 -> ot_text);
	    bp += strlen (bp);

	    ip += (j = ot2 -> ot_name -> oid_nelem + 1);
	    for (; j < k; ip++, j++) {
		(void) sprintf (bp, ".%u", *ip);
		bp += strlen (bp);
	    }
	    i -= j;
	    fancy = 1;
	}
    }

    if (fancy && i > 0) {
	(void) strcpy (bp = buffer, ot -> ot_text);
	bp += strlen (bp);

	ip = oid -> oid_elements + ot -> ot_name -> oid_nelem;
	i = oid -> oid_nelem - ot -> ot_name -> oid_nelem;
    }
    for (; i-- > 0; ip++) {
	(void) sprintf (bp, ".%u", *ip);
	bp += strlen (bp);
    }

    return buffer;
}

/*   INTERACTIVE */

static int  getline (prompt, buffer)
char   *prompt,
       *buffer;
{
    register int    i;
    register char  *cp,
                   *ep;
    static int  sticky = 0;

    if (interrupted) {
	interrupted = 0;
	return NOTOK;
    }

    if (sticky) {
	sticky = 0;
	return NOTOK;
    }

    switch (setjmp (intrenv)) {
	case OK:
	    armed++;
	    break;

	case NOTOK:
	    if (ontty)
		printf ("\n");	/* and fall */
	default:
	    armed = 0;
	    return NOTOK;
    }
	
    if (ontty) {
	printf (prompt, myname);
	(void) fflush (stdout);
    }

    for (ep = (cp = buffer) + BUFSIZ - 1; (i = getchar ()) != '\n';) {
	if (i == EOF) {
	    if (ontty)
		printf ("\n");
	    clearerr (stdin);
	    if (cp == buffer)
		longjmp (intrenv, DONE);

	    sticky++;
	    break;
	}

	if (cp < ep)
	    *cp++ = i;
    }
    *cp = NULL;

    armed = 0;
    
    return OK;
}

/* */

/* ARGSUSED */

static	SFD intrser (sig)
int	sig;
{
#ifndef	BSDSIGS
    (void) signal (SIGINT, intrser);
#endif

    if (armed)
	longjmp (intrenv, NOTOK);

    interrupted++;
}

/* */

#ifndef	TIOCGWINSZ
/* ARGSUSED */
#endif

static int  ncols (fp)
FILE *fp;
{
#ifdef	TIOCGWINSZ
    int	    i;
    struct winsize ws;

    if (ioctl (fileno (fp), TIOCGWINSZ, (char *) &ws) != NOTOK
	    && (i = ws.ws_col) > 0)
	return i;
#endif

    return 80;
}

/*   ERRORS */

#ifndef	lint
static void	_advise ();


void	adios (va_alist)
va_dcl
{
    va_list ap;

    va_start (ap);

    _advise (ap);

    va_end (ap);

    _exit (1);
}
#else
/* VARARGS */

void	adios (what, fmt)
char   *what,
       *fmt;
{
    adios (what, fmt);
}
#endif


#ifndef	lint
/*void	advise (va_alist)
va_dcl
{
    va_list ap;

    va_start (ap);

    _advise (ap);

    va_end (ap);
}*/


static void  _advise (ap)
va_list	ap;
{
    char    buffer[BUFSIZ];

    asprintf (buffer, ap);

    (void) fflush (stdout);

    fprintf (stderr, "%s: ", myname);
    (void) fputs (buffer, stderr);
    (void) fputc ('\n', stderr);

    (void) fflush (stderr);
}
#else
/* VARARGS */

/*KMTM void	advise (what, fmt)
char   *what,
       *fmt;
{
    advise (what, fmt);
}*/
#endif
